
/*************   Method 1   *************/ 

@Override
    public boolean canAddEdge(Object edge)  {
        if (edge == null) {
            return false;
        }
        if (containsEdge(edge)) {
            return false;
        }
        Object end0 = null, end1 = null;
        if (edge instanceof CommentEdge) {
            end0 = ((CommentEdge) edge).getSource();
            end1 = ((CommentEdge) edge).getDestination();
        } else if (Model.getFacade().isAAssociationEnd(edge)) {
            end0 = Model.getFacade().getAssociation(edge);
            end1 = Model.getFacade().getType(edge);

            return (end0 != null
                    && end1 != null
                    && (containsEdge(end0) || containsNode(end0))
                    && containsNode(end1));
        } else if (Model.getFacade().isARelationship(edge)) {
            end0 = Model.getCoreHelper().getSource(edge);
            end1 = Model.getCoreHelper().getDestination(edge);
        } else if (Model.getFacade().isALink(edge)) {
            end0 = Model.getCommonBehaviorHelper().getSource(edge);
            end1 =
                    Model.getCommonBehaviorHelper().getDestination(edge);
        } else if (edge instanceof CommentEdge) {
            end0 = ((CommentEdge) edge).getSource();
            end1 = ((CommentEdge) edge).getDestination();
        } else {
            return false;
        }

        if (end0 == null || end1 == null) {
            LOG.error("Edge rejected. Its ends are not attached to anything");
            return false;
        }

        if (!c1ontainsNode(end0)
                && !containsEdge(end0)) {
            LOG.error("Edge rejected. Its source end is attached to "
                    + end0
                    + " but this is not in the graph model");
            return false;
        }
        if (!containsNode(end1)
                && !containsEdge(end1)) {
            LOG.error("Edge rejected. Its destination end is attached to "
                    + end1
                    + " but this is not in the graph model");
            return false;
        }

        return true;
    }

   
/*************   Method 2   *************/ 

        public final boolean saveDefault(boolean force) {
        if (force) {
            File toFile = new File(getDefaultPath());
            boolean saved = saveFile(toFile);
            if (saved) {
                loadedFromFile = toFile;
            }
            return saved;
        }
        if (!loaded) {
            return false;
        }

        if (loadedFromFile != null) {
            return saveFile(loadedFromFile);
        }
        if (loadedFromURL != null) {
            return saveURL(loadedFromURL);
        }
        return false;
    }

/*************   Method 3   *************/ 

        public Iterator childIterator(Object parent) {
        List res = new ArrayList();
        if (parent instanceof Project) {
            Project p = (Project) parent;
            res.addAll(p.getUserDefinedModelList());
            res.addAll(p.getDiagramList());
        } else if (parent instanceof ArgoDiagram) {
            ArgoDiagram d = (ArgoDiagram) parent;
            res.addAll(d.getGraphModel().getNodes());
            res.addAll(d.getGraphModel().getEdges());
        } else if (Model.getFacade().isAModelElement(parent)) {
            res.addAll(Model.getFacade().getModelElementContents(parent));
        }
        
	return res.iterator();
    }

/*************   Method 4   *************/ 

    public void init() {
        Object classCls = Model.getMetaTypes().getUMLClass();
        Agency.register(crConsiderSingleton, classCls);
        Agency.register(crSingletonViolatedMSA, classCls);
        Agency.register(crSingletonViolatedOPC, classCls);
    }

/*************   Method 5   *************/ 

        @Override
    public void addNodeRelatedEdges(Object node) {
        super.addNodeRelatedEdges(node);

	if (Model.getFacade().isAClassifier(node)) {
	    Collection ends = Model.getFacade().getAssociationEnds(node);
	    for (Object end : ends) {
		if (canAddEdge(Model.getFacade().getAssociation(end))) {
                    addEdge(Model.getFacade().getAssociation(end));
                }
	    }
	}
	if (Model.getFacade().isAGeneralizableElement(node)) {
	    Collection generalizations = 
	        Model.getFacade().getGeneralizations(node);
	    for (Object generalization : generalizations) {
		if (canAddEdge(generalization)) {
		    addEdge(generalization);
		    return;
		}
	    }
	    Collection specializations = Model.getFacade().getSpecializations(node);
	    for (Object specialization : specializations) {
		if (canAddEdge(specialization)) {
		    addEdge(specialization);
		    return;
		}
	    }
	}
	if (Model.getFacade().isAModelElement(node)) {
	    Collection dependencies =
		new ArrayList(Model.getFacade().getClientDependencies(node));
	    dependencies.addAll(Model.getFacade().getSupplierDependencies(node));
	    for (Object dependency : dependencies) {
		if (canAddEdge(dependency)) {
		    addEdge(dependency);
		    return;
		}
	    }
	}
    }

/*************   Method 6   *************/ 


	private void arrangeDialog(final JTextArea textArea,
			final JPanel patternPreviewPanel,
			final JPanel availablePatternsPanel, final JButton okButton,
			final JPanel auxPanel) {
		JPanel panel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.insets = new Insets(10, 10, 10, 10);
		panel.add(textArea, c);
		c.gridy = 1;
		c.insets = new Insets(5, 30, 5, 30);
		panel.add(this.firstElementLabel, c);
		c.gridy = 2;
		panel.add(this.patternComboBox, c);
		c.gridx = 0;
		c.gridy = 3;
		c.weightx = 0.7;
		c.weighty = 1;
		c.gridwidth = 1;
		c.fill = GridBagConstraints.BOTH;
		c.insets = new Insets(5, 5, 5, 5);
		panel.add(patternPreviewPanel, c);
		c.gridx = 1;
		c.weightx = 0.3;
		panel.add(availablePatternsPanel, c);
		c.gridx = 0;
		c.gridy = 4;
		c.gridwidth = 2;
		c.weighty = 0;
		c.fill = GridBagConstraints.NONE;
		c.anchor = GridBagConstraints.CENTER;
		c.insets = new Insets(10, 10, 10, 10);
		panel.add(auxPanel, c);
		add(panel);
		getRootPane().setDefaultButton(okButton);
	}

/*************   Method 7   *************/ 

    @Override
    public boolean equals(final Object obj) {
        if (this == obj) {
            return true;
        }
        if (obj == null) {
            return false;
        }
        if (getClass() != obj.getClass()) {
            return false;
        }
        FontSettings other = (FontSettings) obj;
        if (this.font == null) {
            if (other.font != null) {
                return false;
            }
        } else if (!this.font.equals(other.font)) {
            return false;
        }
        if (this.useFontSmoothing != other.useFontSmoothing) {
            return false;
        }
        if (this.useFontSmoothingSettingsFromOs != other.useFontSmoothingSettingsFromOs) {
            return false;
        }
        return true;
    }

/*************   Method 8   *************/ 

    ApplicationVersion getApplicationVersionFromXml(final Document xml) {
        Element element = (Element) xml.getElementsByTagName("latest").item(0);
        String date = XMLUtils.getChildElementContent(element, "date");
        int major = Integer.parseInt(XMLUtils.getChildElementContent(element,
                "majorNumber"));
        int minor = Integer.parseInt(XMLUtils.getChildElementContent(element,
                "minorNumber"));
        int revision = Integer.parseInt(XMLUtils.getChildElementContent(
                element, "revisionNumber"));
        String url = element.getAttribute("url");

        String directDownloadURL = applyVersion(
                XMLUtils.getChildElementContent(element, getElementNameForOS()),
                major, minor, revision);

        String changes = XMLUtils.getChildElementContent(element, "changes");

        return new ApplicationVersion(date, major, minor, revision,
                VersionType.FINAL, "", url, directDownloadURL, changes);
    }

/*************   Method 9   *************/ 


    private List<Entry<T, Integer>> getElementsSorted() {
        List<Entry<T, Integer>> list = new ArrayList<Map.Entry<T, Integer>>(
                this.count.entrySet());
        Collections.sort(list, new Comparator<Entry<T, Integer>>() {
            @Override
            public int compare(final Entry<T, Integer> o1,
                    final Entry<T, Integer> o2) {
                return -o1.getValue().compareTo(o2.getValue());
            }
        });
        return list;
    }

/*************   Method 10   *************/ 

    @Override
    public JPopupMenu getTreePopupMenu() {
        if (this.deviceTreeMenu == null) {
            this.deviceTreeMenu = new JPopupMenu();
            AbstractActionOverSelectedObjects<IAudioObject> addToPlayListAction = getBeanFactory()
                    .getBean("addToPlayListFromDeviceNavigationView",
                            AddToPlayListAction.class);
            addToPlayListAction.setAudioObjectsSource(this);
            this.deviceTreeMenu.add(addToPlayListAction);

            SetAsPlayListAction setAsPlayListAction = getBeanFactory().getBean(
                    "setAsPlaylistFromDeviceNavigationView",
                    SetAsPlayListAction.class);
            setAsPlayListAction.setAudioObjectsSource(this);
            this.deviceTreeMenu.add(setAsPlayListAction);

            this.deviceTreeMenu.add(new JSeparator());

            AbstractActionOverSelectedTreeObjects<IFolder> openFolder = getBeanFactory()
                    .getBean("openFolderFromDeviceNavigationTree",
                            OpenFolderFromNavigatorTreeAction.class);
            openFolder.setTreeObjectsSource(this);
            this.deviceTreeMenu.add(openFolder);

            this.deviceTreeMenu.add(new JSeparator());
            this.deviceTreeMenu.add(new EditTagMenu(false, this,
                    getBeanFactory()));

            AbstractActionOverSelectedTreeObjects<IAlbum> editTitles = getBeanFactory()
                    .getBean("editTitlesFromDeviceViewAction",
                            EditTitlesAction.class);
            editTitles.setTreeObjectsSource(this);
            this.deviceTreeMenu.add(editTitles);
            this.deviceTreeMenu.add(new JSeparator());
            this.deviceTreeMenu.add(getBeanFactory().getBean(
                    RemoveFromDiskAction.class));
            this.deviceTreeMenu.add(new JSeparator());

            CopyToRepositoryAction copyToRepositoryAction = getBeanFactory()
                    .getBean(CopyToRepositoryAction.class);
            copyToRepositoryAction.setAudioObjectsSource(this);
            this.deviceTreeMenu.add(copyToRepositoryAction);

            this.deviceTreeMenu.add(getBeanFactory().getBean(
                    FillDeviceWithRandomSongsAction.class));
            this.deviceTreeMenu.add(new JSeparator());
            this.deviceTreeMenu.add(getBeanFactory().getBean(
                    SearchArtistAction.class));
            this.deviceTreeMenu.add(getBeanFactory().getBean(
                    SearchArtistAtAction.class));
        }
        return this.deviceTreeMenu;
    }

/*************   Method 11   *************/ 

    private void moveToBottom(final IPlayList playList, final int[] rows) {
        int j = 0;
        for (int i = rows.length - 1; i >= 0; i--) {
            IAudioObject aux = playList.get(rows[i]);
            playList.remove(rows[i]);
            playList.add(playList.size() - j++, aux);
        }
        if (rows[rows.length - 1] < playList.getCurrentAudioObjectIndex()) {
            playList.setCurrentAudioObjectIndex(playList
                    .getCurrentAudioObjectIndex() - rows.length);
        } else if (rows[0] <= playList.getCurrentAudioObjectIndex()
                && playList.getCurrentAudioObjectIndex() <= rows[rows.length - 1]) {
            playList.setCurrentAudioObjectIndex(playList
                    .getCurrentAudioObjectIndex()
                    + playList.size()
                    - rows[rows.length - 1] - 1);
        }
    }

/*************   Method 12   *************/ 

    private void fillCdInfo() {
        CDInfo info = this.cdda2wav.getCDInfo();
        info.setTracks(this.tracks);
        info.setDurations(this.durations);
        info.setDuration(this.totalDuration);
        info.setID(this.id);
        if (this.album != null && !this.album.equals("")) {
            info.setAlbum(this.album);
        }

        if (this.artist != null && !this.artist.equals("")) {
            info.setArtist(this.artist);
        }

        info.setTitles(this.titles);
        info.setArtists(this.artists);
        info.setComposers(this.composers);
    }

/*************   Method 13   *************/ 

    @Override
    public int hashCode() {
        final int prime = 31;
        int result = 1;
        result = prime * result + this.order;
        result = prime * result
                + ((this.sort == null) ? 0 : this.sort.hashCode());
        result = prime * result + (this.visible ? 1231 : 1237);
        result = prime * result + this.width;
        return result;
    }

/*************   Method 14   *************/ 
	    @Override
    public boolean predicate2(Object dm, Designer dsgr) {

        if (!(Model.getFacade().isAClass(dm))) {
            return NO_PROBLEM;
        }

        if (Model.getFacade().isAAssociationClass(dm)) {
            return NO_PROBLEM;
        }

        if (Model.getFacade().getName(dm) == null
                || "".equals(Model.getFacade().getName(dm))) {
            return NO_PROBLEM;
        }

        if (!(Model.getFacade().isPrimaryObject(dm))) {
            return NO_PROBLEM;
        }

        if (Model.getFacade().isAbstract(dm)) {
            return NO_PROBLEM;
        }

        if (Model.getFacade().isSingleton(dm)) {
            return NO_PROBLEM;
        }

	if (Model.getFacade().isUtility(dm)) {
	    return NO_PROBLEM;
	}

	Iterator iter = Model.getFacade().getAttributes(dm).iterator();

	while (iter.hasNext()) {
	    if (!Model.getFacade().isStatic(iter.next())) {
	        return NO_PROBLEM;
	    }
	}

	Iterator ends = Model.getFacade().getAssociationEnds(dm).iterator();

	while (ends.hasNext()) {
	    Iterator otherends =
		Model.getFacade()
			.getOtherAssociationEnds(ends.next()).iterator();

	    while (otherends.hasNext()) {
		if (Model.getFacade().isNavigable(otherends.next())) {
		    return NO_PROBLEM;
		}
	    }
	}

	return PROBLEM_FOUND;
    }

/*************   Method 15   *************/ 

private void addUserDefinedProfile(String fileName, StringBuffer xmi,
            ProfileManager profileManager) throws IOException {
        File profilesDirectory = getProfilesDirectory(profileManager);
        File profileFile = new File(profilesDirectory, fileName);
        OutputStreamWriter writer = new OutputStreamWriter(
                new FileOutputStream(profileFile), 
                Argo.getEncoding());
        writer.write(xmi.toString());
        writer.close();
        LOG.info("Wrote user defined profile \"" + profileFile 
            + "\", with size " + xmi.length() + ".");
        if (isSomeProfileDirectoryConfigured(profileManager)) {
            profileManager.refreshRegisteredProfiles();
        } else {
            profileManager.addSearchPathDirectory(
                profilesDirectory.getAbsolutePath());
        }
    }


/*************   Method 16   *************/ 

 @Override
    protected void setStandardBounds(int x, int y, int w, int h) {
        if (getNameFig() == null) {
            return;
        }

        Rectangle oldBounds = getBounds();

        Dimension nameMin = getNameFig().getMinimumSize();

        getBigPort().setBounds(x, y, w, h);
        cover.setBounds(x, y, w, h);
        getNameFig().setBounds(x, y, nameMin.width + 10, nameMin.height + 4);

        
        _x = x; _y = y; _w = w; _h = h;

        firePropChange("bounds", oldBounds, getBounds());
        calcBounds(); 
        updateEdges();
    }

/*************   Method 17   *************/ 

      @Override
    public void addEdge(Object edge) {
        LOG.debug("adding class edge!!!!!!");
        if (!canAddEdge(edge)) {
            return;
        }
        getEdges().add(edge);
        
        if (Model.getFacade().isAModelElement(edge)
	    && Model.getFacade().getNamespace(edge) == null) {
            Model.getCoreHelper().addOwnedElement(getHomeModel(), edge);
        }
        fireEdgeAdded(edge);
    }


/*************   Method 18   *************/ 

    protected final void removeAllElementListeners(
            PropertyChangeListener ) {
        for (Object[] lis : listeners) {
            Object property = lilisteners[1];
            if (property == null) {
                Model.getPump().removeModelEventListener(listener, lis[0]);
            } else if (property instanceof String[]) {
                Model.getPump().removeModelEventListener(listener, lis[0],
                        (String[]) property);
            } else if (property instanceof String) {
                Model.getPump().removeModelEventListener(listener, lis[0],
                        (String) property);
            } else {
                throw new RuntimeException(
                        "Internal error in removeAllElementListeners");
            }
        }
        listeners.clear();
    }


/*************   Method 19   *************/ 

       public Collection loadModel(ProfileReference reference) 
        throws ProfileException {
        
        if (reader == null) {
            LOG.error("Profile not found");
            throw new ProfileException("Profile not found!");
        }
        
        try {
            XmiReader xmiReader = Model.getXmiReader();
            InputSource inputSource = new InputSource(reader);
            inputSource.setSystemId(reference.getPath());
            inputSource.setPublicId(
                    reference.getPublicReference().toString());
            Collection elements = xmiReader.parse(inputSource, true);
            return elements;
        } catch (UmlException e) {
            throw new ProfileException("Invalid XMI data!", e);
        }
    }


/*************   Method 20   *************/ 
      @Override
  public String showSaveDialog(View        parentView,
                               String      dialogTitle,
                               ContentType contentType,
                               String      name) {
    if (contentType == ContentType.SWEET_HOME_3D) {
      String message = this.preferences.getLocalizedString(
          AppletContentManager.class, "showSaveDialog.message");
      String savedName = (String)JOptionPane.showInputDialog(SwingUtilities.getRootPane((JComponent)parentView), 
          message, getFileDialogTitle(true), JOptionPane.QUESTION_MESSAGE, null, null, name);
      if (savedName == null) {
        return null;
      }
      savedName = savedName.trim();
  
      try {
        
        if (this.recorder.exists(savedName)
            && !confirmOverwrite(parentView, savedName)) {
          return showSaveDialog(parentView, dialogTitle, contentType, savedName);
        
        } else if (savedName.length() == 0) {
          return showSaveDialog(parentView, dialogTitle, contentType, savedName);
        }
        return savedName;
      } catch (RecorderException ex) {
        String errorMessage = this.preferences.getLocalizedString(
            AppletContentManager.class, "showSaveDialog.checkHomeError");
        showError(parentView, errorMessage);
        return null;
      }
    } else {
      return super.showSaveDialog(parentView, dialogTitle, contentType, name);
    }
  }


/*************   Method 21   *************/ 

  private void computeBounds(Node node, BoundingBox bounds, 
                             Transform3D parentTransformations, boolean transformShapeGeometry) {
    if (node instanceof Group) {
      if (node instanceof TransformGroup) {
        parentTransformations = new Transform3D(parentTransformations);
        Transform3D transform = new Transform3D();
        ((TransformGroup)node).getTransform(transform);
        parentTransformations.mul(transform);
      }
      
      Enumeration<?> enumeration = ((Group)node).getAllChildren();
      while (enumeration.hasMoreElements ()) {
        computeBounds((Node)enumeration.nextElement(), bounds, parentTransformations, transformShapeGeometry);
      }
    } else if (node instanceof Link) {
      computeBounds(((Link)node).getSharedGroup(), bounds, parentTransformations, transformShapeGeometry);
    } else if (node instanceof Shape3D) {
      Shape3D shape = (Shape3D)node;
      Bounds shapeBounds;
      if (transformShapeGeometry) {
        shapeBounds = computeTransformedGeometryBounds(shape, parentTransformations);
      } else {
        shapeBounds = shape.getBounds();
        shapeBounds.transform(parentTransformations);
      }
      bounds.combine(shapeBounds);
    }
  }

/*************   Method 22   *************/ 

    private GraphicsConfigTemplate3D createGraphicsConfigurationTemplate3D() {
    if (System.getProperty("j3d.implicitAntialiasing") == null) {
      System.setProperty("j3d.implicitAntialiasing", "true");
    }
    GraphicsConfigTemplate3D template = new GraphicsConfigTemplate3D();
    
    template.setSceneAntialiasing(GraphicsConfigTemplate3D.PREFERRED);
    
    String stereo = System.getProperty("j3d.stereo");
    if (stereo != null) {
      if ("REQUIRED".equals(stereo))
        template.setStereo(GraphicsConfigTemplate.REQUIRED);
      else if ("PREFERRED".equals(stereo))
        template.setStereo(GraphicsConfigTemplate.PREFERRED);
    }
    return template;
  }


/*************   Method 23   *************/ 

    public float getXCenter() {
    float xMin = this.points [0][0]; 
    float xMax = this.points [0][0]; 
    for (int i = 1; i < this.points.length; i++) {
      xMin = Math.min(xMin, this.points [i][0]);
      xMax = Math.max(xMax, this.points [i][0]);
    }
    return (xMin + xMax) / 2;
  }

/*************   Method 24   *************/ 

    private void getAreaPoints(Area area, 
                             boolean reversed, 
                             List<float [][]> areaPoints,
                             Map<Integer, List<float [][]>> areaHoles) {
    
    List<float []> currentPathPoints = new ArrayList<float[]>();
    float [] previousRoomPoint = null;
    int i = 0;
    for (PathIterator it = area.getPathIterator(null, 1); !it.isDone(); it.next()) {
      float [] roomPoint = new float[2];
      switch (it.currentSegment(roomPoint)) {
        case PathIterator.SEG_MOVETO :
        case PathIterator.SEG_LINETO : 
          if (previousRoomPoint == null
              || roomPoint [0] != previousRoomPoint [0] 
              || roomPoint [1] != previousRoomPoint [1]) {
            currentPathPoints.add(roomPoint);
          }
          previousRoomPoint = roomPoint;
          break;
        case PathIterator.SEG_CLOSE :
          if (currentPathPoints.get(0) [0] == previousRoomPoint [0] 
              && currentPathPoints.get(0) [1] == previousRoomPoint [1]) {
            currentPathPoints.remove(currentPathPoints.size() - 1);
          }
          if (currentPathPoints.size() > 2) {
            float [][] pathPoints = 
                currentPathPoints.toArray(new float [currentPathPoints.size()][]);
            Room subRoom = new Room(pathPoints);
            if (subRoom.getArea() > 0) {
              boolean pathPointsClockwise = subRoom.isClockwise();
              if (pathPointsClockwise) {
                
                if (!reversed) {
                  pathPoints = getReversedArray(pathPoints);
                }
                List<float [][]> holes = areaHoles.get(i);
                if (holes == null) {
                  holes = new ArrayList<float [][]>(1);
                  areaHoles.put(i, holes);
                }
                holes.add(pathPoints);
              } else {
                if (reversed) {
                  pathPoints = getReversedArray(pathPoints);
                }
                areaPoints.add(pathPoints);
                i++;
              }
            }
          }
          currentPathPoints.clear();
          previousRoomPoint = null;
          break;
      }
    }
  }

/*************   Method 25   *************/ 

    private void createSwtMenu(Shell shell, UserPreferences preferences,
                             Menu menuBar, AbstractMenuItem menuItem) {
    String menuName = menuItem.getId();
    if (menuName != null) {
      AbstractMenuItem [] subMenuItems = menuItem.getChildren();
      MenuItem menuHeader;
      if (subMenuItems != null) {
        Menu currentMenu = new Menu(shell, SWT.DROP_DOWN);
        menuHeader = new MenuItem(menuBar, SWT.CASCADE);
        menuHeader.setMenu(currentMenu);
        for (AbstractMenuItem subMenuItem : subMenuItems) {
          createSwtMenu(shell, preferences, currentMenu, subMenuItem);
        }
      } else {
        menuHeader = new MenuItem(menuBar, SWT.PUSH);
      }
      menuHeader.setText(getMenuLabel(menuName, preferences));
      Image image = getIcon(shell, preferences, menuName);
      if(image != null) {
        menuHeader.setImage(image);
      }
    }
    else {
      new MenuItem(menuBar, SWT.SEPARATOR);
    }
  }

/*************   Method 26   *************/ 

  private String getOptionalString(UserPreferences preferences, 
                                   Class<?> resourceClass, 
                                   String propertyKey,
                                   boolean label) {
    try {
      String localizedText = label 
          ? SwingTools.getLocalizedLabelText(preferences, resourceClass, propertyKey)
          : preferences.getLocalizedString(resourceClass, propertyKey);
      if (localizedText != null && localizedText.length() > 0) {
        return localizedText;
      } else {
        return null;
      }
    } catch (IllegalArgumentException ex) {
      return null;
    }
  }

/*************   Method 27   *************/

    public Table addTable(Class<?> clazz) {
        if(Modifier.isAbstract(clazz.getModifiers())){
            return null;
        }
        Table table = new Table();
        ClassInfo info = ClassInfo.getClassInfo(clazz);
        table.setName(info.tableName);
        table.setType("MyISAM");
        database.addTable(table);
        
        Map<String, UniqueIndex> uniques = new HashMap<String, UniqueIndex>();
        Map<String, NonUniqueIndex> indexes = new HashMap<String, NonUniqueIndex>();
        
        for (Field field : info.allFields) {
            String[] columns = ClassInfo.getColumnNames(field);
            boolean notNull = field.getAnnotation(NotNull.class) != null;
            
            Class<?> type = field.getType();
            if(!ClassInfo.isModel(type) || (ClassInfo.isModel(type) && ClassInfo.isEmbedded(field))) {
                Column column = createColumn(clazz, field, columns[0]);
                
                if(notNull || type.isPrimitive()) {
                    column.setRequired(true);
                    
                    if(type.isPrimitive() && !ClassInfo.isId(field)) { 
                        if(type == Boolean.TYPE) {
                            column.setDefaultValue("false");
                        } else {
                            column.setDefaultValue("0");
                        }
                    }
                }
                
                Id id = field.getAnnotation(Id.class);
                if(id != null) {
                    column.setPrimaryKey(true);
                    column.setRequired(true);
                    
                    if(id.value() == Generator.AUTO_INCREMENT 
                            && (Long.TYPE == type || Long.class.isAssignableFrom(type)))
                        column.setAutoIncrement(true);
                }
                
                table.addColumn(column);
            } else {
                List<Field> keys = ClassInfo.getClassInfo(type).keys;
                
                for (int i = 0; i < columns.length; i++) {
                    Field f = keys.get(i);
                    Column column = createColumn(clazz, f, columns[i]);

                    if(notNull)
                        column.setRequired(true);
                    
                    table.addColumn(column);
                }
            }
        }

        for (Field field : info.updateFields) {
            Index index = field.getAnnotation(Index.class);
            if(index != null) {
                String[] names = index.value();
                for (String name : names) {
                    NonUniqueIndex i = indexes.get(name);
                    if(i == null) {
                        i = new NonUniqueIndex();
                        i.setName(name);
                        indexes.put(name, i);
                        table.addIndex(i);
                    }
                    fillIndex(i, field);
                }
            }
            Unique unique = field.getAnnotation(Unique.class);
            if(unique != null) {
                String[] names = unique.value();
                for (String name : names) {
                    UniqueIndex i = uniques.get(name);
                    if(i == null) {
                        i = new UniqueIndex();
                        i.setName(name);
                        uniques.put(name, i);
                        table.addIndex(i);
                    }
                    fillIndex(i, field);
                }
            }
        }
        tables.put(table.getName(), table);
        return table;
    }

/*************   Method 28   *************/ 

        public List<String> getUpdateFieldsColumnNames() {
        List<String> strs = new ArrayList<String>(this.updateFields.size());
        for(Field field: this.updateFields){
            Column c = field.getAnnotation(Column.class);
            if(c != null && c.value().length > 0) {
                strs.add(c.value()[0]);
            }
    
            else if(isModel(field.getType())) {
                ClassInfo ci = getClassInfo(field.getType());
                for (Field key : ci.keys) {
                    Collections.addAll(strs, getColumnNames(key));
                }
            }
            else {
                strs.add(field.getName());
            }
        }
        return strs;
    }

/*************   Method 29   *************/ 

    public static void encode( ByteBuffer raw, CharBuffer encoded ){
        byte[] raw3 = new byte[3];
        byte[] enc4 = new byte[4];

        while( raw.hasRemaining() ){
            int rem = Math.min(3,raw.remaining());
            raw.get(raw3,0,rem);
            Base64.encode3to4(enc4, raw3, rem, Base64.NO_OPTIONS );
            for( int i = 0; i < 4; i++ ){
                encoded.put( (char)(enc4[i] & 0xFF) );
            }
        }   
    }

/*************   Method 30   *************/ 

        @Override
    public int save(Iterable<?> objects) {
        List<Object> entities2Insert = new ArrayList<Object>();
        List<Object> entities2Update = new ArrayList<Object>();

        for(Object obj:objects){
            Class<?> clazz = obj.getClass();
            ClassInfo info = ClassInfo.getClassInfo(clazz);
            Field idField = info.getIdField();
            
            Object idVal = Util.readField(obj, idField);
            
            if(idVal == null){
                entities2Insert.add(obj);
            }
            else{
                entities2Update.add(obj);
            }
        }
        return insert(entities2Insert) + update(entities2Update);
    }

/*************   Method 31   *************/ 

        public static Object readField(Object object, Field field) {
        boolean wasAccess = true;
        if(!field.isAccessible()){
            field.setAccessible(true);
            wasAccess = false;
        }
        try {
            return field.get(object);
        } catch (Exception e) {
            throw new SienaException(e);
        } finally {
            if(!wasAccess){
                field.setAccessible(false);
            }
        }
    }

/*************   Method 32   *************/ 

        public Victory.Result victory(IGame game, HashMap<String, Object> ctx) {
        boolean victory = false;
        VictoryResult vr = new VictoryResult(true);
        
        Hashtable<Integer,Integer> killsTeam = new Hashtable<Integer,Integer>();
        
        Hashtable<Integer,Integer> killsPlayer = new Hashtable<Integer,Integer>();
        
        updateKillTables(game, killsTeam, killsPlayer, game.getWreckedEntities());
        updateKillTables(game, killsTeam, killsPlayer, game.getCarcassEntities());
        
        boolean teamHasHighestKills = true;
        int highestKillsId = -1;
        int killCount = 0;
        for (Integer killer : killsTeam.keySet()){
            if (killsTeam.get(killer) > killCount){
                highestKillsId = killer;
                killCount = killsTeam.get(killer);
            }
        }
        
        for (Integer killer : killsPlayer.keySet()){
            if (killsTeam.get(killer) > killCount){
                highestKillsId = killer;
                killCount = killsPlayer.get(killer);
                teamHasHighestKills = false;
            }
        }
        
        if (killCount >= killCondition){
            Report r = new Report(7106, Report.PUBLIC);
            victory = true;
            if (teamHasHighestKills) {
                r.add("Team " + highestKillsId);
                vr.addTeamScore(highestKillsId, 1.0);                
            } else {
                IPlayer winner = game.getPlayer(highestKillsId);
                r.add(winner.getName());
                vr.addPlayerScore(winner.getId(), 1.0);
            }
            r.add(killCount);
            vr.addReport(r);
        }
        
        if (victory)
            return vr;
        return new SimpleNoResult();
    }

/*************   Method 33   *************/ 

    private void parseAdvantages(Entity entity, String adv) {
        StringTokenizer st = new StringTokenizer(adv);

        while (st.hasMoreTokens()) {
            String curAdv = st.nextToken();
            int curParameter = 0;
            boolean bParameterDetected = false;

            StringTokenizer advantageParameterTokenizer = new StringTokenizer(
                    curAdv, ":");
            if (advantageParameterTokenizer.countTokens() > 1) {
                
                curAdv = advantageParameterTokenizer.nextToken();
                String curParam = advantageParameterTokenizer.nextToken();
                curParameter = Integer.parseInt(curParam);
                bParameterDetected = true;
            }

            IOption option = entity.getCrew().getOptions().getOption(curAdv);

            if (null == option) {
                System.out.println("Ignoring invalid pilot advantage: "
                        + curAdv);
            } else {
                System.out.println("Adding pilot advantage '" + curAdv
                        + "' to " + entity.getDisplayName());
                
                if (bParameterDetected) {
                    option.setValue(curParameter);
                } else {
                    option.setValue(true);
                }
            }
        }
    }

/*************   Method 34   *************/ 

    private void roll(int connId, int dice, int sides) {
        StringBuffer diceBuffer = new StringBuffer();
        int total = 0;
        for (int i = 0; i < dice; i++) {
            int roll = Compute.randomInt(sides) + 1;
            total += roll;

            if (dice < 2) {
                diceBuffer.append(roll);
                continue;
            }
            
            if (i < dice - 1) {
                diceBuffer.append(roll);
                diceBuffer.append(", ");
            } else {
                diceBuffer.append("and ");
                diceBuffer.append(roll);
            }
        }
        server.sendServerChat(server.getPlayer(connId).getName()
                + " has rolled " + diceBuffer + " for a total of " + total
                + ", using " + dice + "d" + sides);
    }

/*************   Method 35   *************/ 

        public void deploy(int id, Coords c, int nFacing, int elevation,
                       List<Entity> loadedUnits, boolean assaultDrop) {
        int packetCount = 6 + loadedUnits.size();
        int index = 0;
        Object[] data = new Object[packetCount];
        data[index++] = new Integer(id);
        data[index++] = c;
        data[index++] = new Integer(nFacing);
        data[index++] = new Integer(elevation);
        data[index++] = new Integer(loadedUnits.size());
        data[index++] = new Boolean(assaultDrop);

        for (Entity ent : loadedUnits) {
            data[index++] = new Integer(ent.getId());
        }

        send(new Packet(Packet.COMMAND_ENTITY_DEPLOY, data));
        flushConn();
    }

/*************   Method 36   *************/ 

    public static boolean canMechFindClub(IGame game, int entityId) {
        final Entity entity = game.getEntity(entityId);
        if (null == entity.getPosition()) {
            return false;
        }
        final IHex hex = game.getBoard().getHex(entity.getPosition());

        if (!(entity instanceof BipedMech || entity instanceof TripodMech)) {
            return false;
        }

        if (entity.isShutDown() || !entity.getCrew().isActive()) {
            return false;
        }

        if (game.getOptions().booleanOption("no_clan_physical")
            && entity.isClan()) {
            return false;
        }

        if ((hex.terrainLevel(Terrains.WOODS) < 1)
            && (hex.terrainLevel(Terrains.JUNGLE) < 1)
            && (hex.terrainLevel(Terrains.RUBBLE) < Building.MEDIUM)
            && (hex.terrainLevel(Terrains.ARMS) < 1)
            && (hex.terrainLevel(Terrains.LEGS) < 1)) {
            return false;
        }

        if (!entity.hasWorkingSystem(Mech.ACTUATOR_SHOULDER, Mech.LOC_RARM)
            || !entity.hasWorkingSystem(Mech.ACTUATOR_SHOULDER,
                                        Mech.LOC_LARM)
            || (!entity.hasWorkingSystem(Mech.ACTUATOR_HAND, Mech.LOC_RARM) && !((Mech) entity)
                .hasClaw(Mech.LOC_RARM))
            || (!entity.hasWorkingSystem(Mech.ACTUATOR_HAND, Mech.LOC_LARM) && !((Mech) entity)
                .hasClaw(Mech.LOC_LARM))) {
            return false;
        }

        if (entity.hasQuirk(OptionsConstants.QUIRK_NEG_NO_ARMS)) {
            return false;
        }

        if (entity.getClubs().size() > 0) {
            return false;
        }

        return true;
    }

/*************   Method 37   *************/ 

        private void checkReady() {
        
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer player = i.nextElement();
            if (!player.isGhost() && !player.isObserver() && !player.isDone()) {
                return;
            }
        }

        if (game.getNoOfInitiativeRerollRequests() > 0) {
            resetActivePlayersDone();
            game.rollInitAndResolveTies();

            determineTurnOrder(IGame.Phase.PHASE_INITIATIVE);
            clearReports();
            writeInitiativeReport(true);
            sendReport(true);
            return; 
        }

        if (!game.phaseHasTurns(game.getPhase())
            && ((game.getPhase() != IGame.Phase.PHASE_LOUNGE) || (game.getNoOfEntities() > 0))) {
            endCurrentPhase();
        }
    }

/*************   Method 38   *************/ 

    public final boolean unload(Entity unit) {
        
        Entity trooper = game.getEntity(troopers);
        if ((trooper == null) || !trooper.equals(unit)) {
            
            return false;
        }
        troopers = Entity.NONE;
        return true;
    }

/*************   Method 39   *************/ 

    public int getEnemyInitialBV(IGame game, IPlayer player) {
        int ret = 0;
        for (Enumeration<IPlayer> f = game.getPlayers(); f.hasMoreElements();) {
            IPlayer other = f.nextElement();
            if (other.isObserver())
                continue;
            if (other.isEnemyOf(player)) {
                ret += other.getInitialBV();
            }
        }
        return ret;
    }

/*************   Method 40   *************/ 

        public static Entity createEntityInstance(Field idField, ClassInfo info, Object obj){
        Entity entity = null;
        Id id = idField.getAnnotation(Id.class);
        Class<?> type = idField.getType();

        if(id != null){
            switch(id.value()) {
            case NONE:
                Object idVal = null;
                idVal = Util.readField(obj, idField);
                if(idVal == null)
                    throw new SienaException("Id Field " + idField.getName() + " value null");
                String keyVal = Util.toString(idField, idVal);              
                entity = new Entity(info.tableName, keyVal);
                break;
            case AUTO_INCREMENT:
                
                if(Long.TYPE == type || Long.class.isAssignableFrom(type)){
                    entity = new Entity(info.tableName);
                }else {
                    Object idStringVal = null;
                    idStringVal = Util.readField(obj, idField);
                    if(idStringVal == null)
                        throw new SienaException("Id Field " + idField.getName() + " value null");
                    String keyStringVal = Util.toString(idField, idStringVal);              
                    entity = new Entity(info.tableName, keyStringVal);
                }
                break;
            case UUID:
                entity = new Entity(info.tableName, UUID.randomUUID().toString());
                break;
            default:
                throw new SienaRestrictedApiException("DB", "createEntityInstance", "Id Generator "+id.value()+ " not supported");
            }
        }
        else throw new SienaException("Field " + idField.getName() + " is not an @Id field");
        
        return entity;
    }
    

/*************   Method 41   *************/ 

        public static void fillRequestElement(Object obj, Element element, boolean ids) {
        Class<?> clazz = obj.getClass();
        element.addAttribute("class", clazz.getName());
        
        Field[] fields = clazz.getDeclaredFields();
        for (Field field : fields) {
            if(field.getType() == Class.class) continue;
            if(ids && !ClassInfo.isId(field)) continue;
            field.setAccessible(true);
            Object value;
            try {
                value = field.get(obj);
            } catch (Exception e) {
                throw new SienaException(e);
            }
            Class<?> type = field.getType();
            if(ClassInfo.isModel(type)) {
                Element f = element.addElement("object");
                f.addAttribute("name", field.getName());
                if(value != null) {
                    fillRequestElement((Model) value, f, true);
                }
            } else {
                Element f = element.addElement("field");
                f.addAttribute("name", field.getName());
                if(value != null) {
                    f.setText(Util.toString(field, value));
                }
            }
        }
    }


/*************   Method 42   *************/ 

        public static <T> int mapSelectResult(SelectResult res, Iterable<T> objects) {
        List<Item> items = res.getItems();
        
        Class<?> clazz = null;
        ClassInfo info = null;
        int nb = 0;
        for(T obj: objects){
            if(clazz == null){
                clazz = obj.getClass();
                info = ClassInfo.getClassInfo(clazz);               
            }
            String itemName = getItemName(clazz, obj);
            Item theItem = null;
            for(Item item:items){
                if(item.getName().equals(itemName)){
                    theItem = item;
                    items.remove(item);
                    break;
                }
            }
            if(theItem != null){
                fillModel(theItem, clazz, info, obj);
                nb++;
            }
        }
        return nb;
    }

/*************   Method 43   *************/ 

public void addAndMoveCursor(String cursor){
        
        if(cursorIdx < cursors.size()-1 && cursorIdx>=0){
            cursors.set(++cursorIdx, cursor);
        }
        else{
            cursors.add(++cursorIdx, cursor);
        }
    }

/*************   Method 44   *************/ 

        public List<T> get() {
        List<T> results;
        switch(mapType){
        case KEYS_ONLY:
            results = GaeMappingUtils.mapEntitiesKeysOnly(entities, query.getQueriedClass());
            break;
        case ALL:
        default:
            results = pm.map(query, entities);
            break;
        }
        
        QueryOptionPage pag = (QueryOptionPage)query.option(QueryOptionPage.ID);
        QueryOptionGaeContext gaeCtx = (QueryOptionGaeContext)query.option(QueryOptionGaeContext.ID);
        if(pag.isPaginating()){
            if(results.size() == 0){
                gaeCtx.noMoreDataAfter = true;
            }else {
                gaeCtx.noMoreDataAfter = false;
            }
        }
        return results;
    }


/*************   Method 45   *************/ 

    public static void embed(ReplaceableItem item, String embeddingColumnName, Object embeddedObj){
        Class<?> clazz = embeddedObj.getClass();
        if(clazz.isArray() || Collection.class.isAssignableFrom(clazz)){
            throw new SienaException("can't serializer Array/Collection in native mode");
        }
        
        for (Field f : ClassInfo.getClassInfo(clazz).updateFields) {
            String propValue = SdbMappingUtils.objectFieldToString(embeddedObj, f);
            if(propValue != null){
                ReplaceableAttribute attr = 
                    new ReplaceableAttribute(
                            getEmbeddedAttributeName(embeddingColumnName, f), propValue, true);
                item.withAttributes(attr);
            }else {
                if (ClassInfo.isEmbeddedNative(f)){
                    SdbNativeSerializer.embed(
                        item, 
                        getEmbeddedAttributeName(embeddingColumnName, f), 
                        Util.readField(embeddedObj, f));
                }
            }
        }
    }


/*************   Method 46   *************/ 

        public Response putAttributes(String domain, Item item) {
        TreeMap<String, String> parameters = new TreeMap<String, String>();
        parameters.put("Action", "PutAttributes");
        parameters.put("DomainName", domain);
        parameters.put("ItemName", item.name);
        
        int i = 0;
        for (Map.Entry<String, List<String>> entry : item.attributes.entrySet()) {
            List<String> values = entry.getValue();
            for (String value : values) {
                parameters.put("Attribute."+i+".Name", entry.getKey());
                parameters.put("Attribute."+i+".Value", value);
                parameters.put("Attribute."+i+".Replace", "true"); 
                i++;
            }
        }
        return request(parameters, new PlainHandler());
    }


/*************   Method 47   *************/ 

        public synchronized void addCommand(Command command) {

        ProjectManager.getManager().setSaveEnabled(true);
        
        if (undoMax == 0) {
            return;
        }
        
        if (!command.isUndoable()) {
            undoStack.clear();
            newInteraction = true;
        }
        
        final Interaction macroCommand;
        if (newInteraction || undoStack.isEmpty()) {
            redoStack.clear();
            newInteraction = false;
            if (undoStack.size() > undoMax) {
                undoStack.remove(0);
            }
            macroCommand = new Interaction(newInteractionLabel);
            undoStack.push(macroCommand);
        } else {
            macroCommand = undoStack.peek();
        }
        macroCommand.addCommand(command);
    }

/*************   Method 48   *************/ 

        public void vetoableChange(PropertyChangeEvent pce) {
    
    if ("ownedElement".equals(pce.getPropertyName())) {
        List oldOwned = (List) pce.getOldValue();
        Object eo =  pce.getNewValue();
        Object me = Model.getFacade().getModelElement(eo);
        if (oldOwned.contains(eo)) {
        LOG.debug("model removed " + me);
        if (Model.getFacade().isAClassifier(me)) {
                    removeNode(me);
                }
        if (Model.getFacade().isAMessage(me)) {
                    removeNode(me);
                }
        if (Model.getFacade().isAAssociation(me)) {
                    removeEdge(me);
                }
        } else {
        LOG.debug("model added " + me);
        }
    }
    }


/*************   Method 49   *************/ 

    public void eventDispatched(AWTEvent event) {
        Object source = event.getSource();
        if (event instanceof KeyEvent
                && source instanceof Component) {
            
            if ((SwingUtilities.windowForComponent((Component) source) == _window)) {
                ((KeyEvent) event).consume();
            }
        }
    }

/*************   Method 50   *************/ 

    private boolean isSelectedInternal(String name) {
    Map.Entry<ModuleInterface, ModuleStatus> entry = findModule(name);

    if (entry != null) {
        ModuleStatus status = entry.getValue();

        if (status == null) {
        return false;
        }

        return status.isSelected();
    }
    return false;
    }

/*************   Method 51   *************/ 

        public Object execute() {
            final Iterator<Command> it = commands.iterator();
            while (it.hasNext()) {
                it.next().execute();
            }
            return null;
        }

/*************   Method 52   *************/ 

        @Override
    public void run(int connId, String[] args) {
        int kickArg = server.isPassworded() ? 2 : 1;

        if (!canRunRestrictedCommand(connId)) {
            server.sendServerChat(connId,
                    "Observers are restricted from kicking others.");
            return;
        }
        if (server.isPassworded()
                && (args.length < 3 || !server.isPassword(args[1]))) {
            server
                    .sendServerChat(connId,
                            "The password is incorrect.  Usage: /kick <password> [id#]");
        } else
            try {
                int kickedId = Integer.parseInt(args[kickArg]);

                if (kickedId == connId) {
                    server.sendServerChat("Don't be silly.");
                    return;
                }

                server.sendServerChat(server.getPlayer(connId).getName()
                        + " attempts to kick player #" + kickedId + " ("
                        + server.getPlayer(kickedId).getName() + ")...");
                
                server.send(kickedId, new Packet(Packet.COMMAND_CLOSE_CONNECTION));
                server.getConnection(kickedId).close();

            } catch (ArrayIndexOutOfBoundsException ex) {
                server
                        .sendServerChat("/kick : kick failed.  Type /who for a list of players with id #s.");
            } catch (NumberFormatException ex) {
                server
                        .sendServerChat("/kick : kick failed.  Type /who for a list of players with id #s.");
            } catch (NullPointerException ex) {
                server
                        .sendServerChat("/kick : kick failed.  Type /who for a list of players with id #s.");
            }
    }

/*************   Method 53   *************/ 

        public float getWeightArmor() {
        return (float) getEntity().getLabArmorTonnage();
    }

    public float getWeightAllocatedArmor() {

        float armorWeight = 0;
        if (!getEntity().hasPatchworkArmor()) {
            armorWeight += armor[0].getWeightArmor(getTotalOArmor(),
                    getWeightCeilingArmor());
        } else {
            for (int i = 0; i < armor.length; i++) {
                int points = getEntity().getOArmor(i);
                if (getEntity().hasRearArmor(i) &&
                        (getEntity().getOArmor(i, true) > 0)) {
                    points += getEntity().getOArmor(i, true);
                }
                armorWeight += armor[i].getWeightArmor(points,
                        getWeightCeilingArmor());
            }
        }
        return armorWeight;
    }

/*************   Method 54   *************/ 

       protected int calcAttackValue() {
        int av = 0;
        
        double damage = ((InfantryWeapon)wtype).getInfantryDamage();
        if((ae instanceof Infantry) && !(ae instanceof BattleArmor)) {
            damage = ((Infantry)ae).getDamagePerTrooper();
            av = (int) Math.round(damage * 0.6 * ((Infantry)ae).getShootingStrength());
        }
        if(bDirect) {
            av = Math.min(av+(toHit.getMoS()/3), av*2);
        }
        if(bGlancing) {
            av = (int) Math.floor(av / 2.0);
        }
        return av;
    }

/*************   Method 55   *************/ 

    protected void setMekHitLocLog() {
        String name = store.getString(MEK_HIT_LOC_LOG);
        if (name.length() != 0) {
            try {
                mekHitLocLog = new PrintWriter(new BufferedWriter(
                        new FileWriter(name)));
                mekHitLocLog.println("Table\tSide\tRoll");
            } catch (Throwable thrown) {
                thrown.printStackTrace();
                mekHitLocLog = null;
            }
        }
    }

/*************   Method 56   *************/ 


        protected double updateAVforAmmo(double current_av, AmmoType atype,
            WeaponType bayWType, int range, int wId) {

        Mounted mLinker = weapon.getLinkedBy();
        int bonus = 0;
        if ((mLinker != null && mLinker.getType() instanceof MiscType
                && !mLinker.isDestroyed() && !mLinker.isMissing()
                && !mLinker.isBreached() && mLinker.getType().hasFlag(
                MiscType.F_ARTEMIS))
                && atype.getMunitionType() == AmmoType.M_ARTEMIS_CAPABLE) {
            bonus = (int) Math.ceil(atype.getRackSize() / 5.0);
            if (atype.getAmmoType() == AmmoType.T_SRM) {
                bonus = 2;
            }
            current_av = current_av + bonus;
        }
        
        if (((mLinker != null) && (mLinker.getType() instanceof MiscType)
                && !mLinker.isDestroyed() && !mLinker.isMissing()
                && !mLinker.isBreached() && mLinker.getType().hasFlag(
                MiscType.F_ARTEMIS_V))
                && (atype.getMunitionType() == AmmoType.M_ARTEMIS_V_CAPABLE)) {
            
            bonus = (int) Math.ceil(atype.getRackSize() / 5.0);
            if (atype.getAmmoType() == AmmoType.T_SRM) {
                bonus = 2;
            }
        }

        if (atype.getMunitionType() == AmmoType.M_CLUSTER) {
            current_av = Math.floor(0.6 * current_av);
        } else if (AmmoType.T_ATM == atype.getAmmoType()) {
            if (atype.getMunitionType() == AmmoType.M_EXTENDED_RANGE) {
                current_av = bayWType.getShortAV() / 2;
            } else if (atype.getMunitionType() == AmmoType.M_HIGH_EXPLOSIVE) {
                current_av = 1.5 * current_av;
                if (range > WeaponType.RANGE_SHORT) {
                    current_av = 0.0;
                }
            }
        } else if (atype.getAmmoType() == AmmoType.T_MML
                && !atype.hasFlag(AmmoType.F_MML_LRM)) {
            current_av = 2 * current_av;
            if (range > WeaponType.RANGE_SHORT) {
                current_av = 0;
            }
        } else if (atype.getAmmoType() == AmmoType.T_AR10) {
            if (atype.hasFlag(AmmoType.F_AR10_KILLER_WHALE)) {
                current_av = 4;
            } else if (atype.hasFlag(AmmoType.F_AR10_WHITE_SHARK)) {
                current_av = 3;
            } else {
                current_av = 2;
            }
        }
        return current_av;
    }

/*************   Method 57   *************/ 

    public boolean correctHeatSinks(StringBuffer buff) {
        if ((aero.getHeatType() != Aero.HEAT_SINGLE) 
                && (aero.getHeatType() != Aero.HEAT_DOUBLE)) {
            buff.append("Invalid heatsink type!  Valid types are "
                    + Aero.HEAT_SINGLE + " and " + Aero.HEAT_DOUBLE
                    + ".  Found " + aero.getHeatType() + ".");
        }
        
        if (aero.getEntityType() == Entity.ETYPE_CONV_FIGHTER){
            int maxWeapHeat = countHeatEnergyWeapons();
            int heatDissipation = 0;
            if (aero.getHeatType() == Aero.HEAT_DOUBLE){
                buff.append("Conventional fighters may only use single " +
                        "heatsinks!");
                return false;
            } 
            heatDissipation = aero.getHeatSinks();
            
            if(maxWeapHeat > heatDissipation) {
                buff.append("Conventional fighters must be able to " +
                        "dissipate all heat from energy weapons! \n" +
                        "Max energy heat: " + maxWeapHeat + 
                        ", max dissipation: " + heatDissipation);
                return false;
            } else {
                return true;
            }
        } else {
            return true;
        }        
    }

/*************   Method 58   *************/ 

      public void run() {
        int port = 0;
        try {
            port = Integer.parseInt(hostPort.getText());
            ServerSocket serverSocket = new ServerSocket(port);
            Socket s = serverSocket.accept();
            serverSocket.close();

            System.out.println("Accepted peer connection.");

            conn = ConnectionFactory.getInstance().createServerConnection(s, 0);
            
            conn.addConnectionListener(connectionListener);

            board = new Board();
            panConnect.setEnabled(false);
            panXmit.setEnabled(true);

        } catch (Throwable err) {
            err.printStackTrace();
        }
    }

/*************   Method 59   *************/ 

    protected void receivePlayerInfo(Packet c) {
        int pindex = c.getIntValue(0);
        IPlayer newPlayer = (IPlayer) c.getObject(1);
        if (getPlayer(newPlayer.getId()) == null) {
            game.addPlayer(pindex, newPlayer);
        } else {
            game.setPlayer(pindex, newPlayer);
        }

        PreferenceManager.getClientPreferences().setLastPlayerColor(
                newPlayer.getColorIndex());
        PreferenceManager.getClientPreferences().setLastPlayerCategory(
                newPlayer.getCamoCategory());
        PreferenceManager.getClientPreferences().setLastPlayerCamoName(
                newPlayer.getCamoFileName());
    }

/*************   Method 60   *************/ 

  private List<HelpDocument> searchInHelpDocuments(URL helpIndex, String [] searchedWords) {
    List<URL> parsedDocuments = new ArrayList<URL>(); 
    parsedDocuments.add(helpIndex);
    
    List<HelpDocument> helpDocuments = new ArrayList<HelpDocument>();
    
    for (int i = 0; i < parsedDocuments.size(); i++) {
      try {
        
        URL helpDocumentUrl = parsedDocuments.get(i);
        HelpDocument helpDocument = new HelpDocument(helpDocumentUrl, searchedWords);
        helpDocument.parse();
        
        if (helpDocument.getRelevance() > 0) {
          helpDocuments.add(helpDocument);
        }
        
        for (URL url : helpDocument.getReferencedDocuments()) {
          String lowerCaseFile = url.getFile().toLowerCase();
          if (lowerCaseFile.endsWith(".html")
              && !parsedDocuments.contains(url)) {
            parsedDocuments.add(url);
          } 
        } 
      } catch (IOException ex) {
      }
    }
    Collections.sort(helpDocuments, new Comparator<HelpDocument>() {
        public int compare(HelpDocument document1, HelpDocument document2) {
          return document2.getRelevance() - document1.getRelevance();
        }
      });
    return helpDocuments;
  }

/*************   Method 61   *************/ 

    private void setLength(Float length, boolean updateEndPoint) {
    if (length != this.length) {
      Float oldLength = this.length;
      this.length = length;
      this.propertyChangeSupport.firePropertyChange(Property.LENGTH.name(), oldLength, length);
      
      if (updateEndPoint) {
        Float xStart = getXStart();
        Float yStart = getYStart();
        Float xEnd = getXEnd();
        Float yEnd = getYEnd();
        if (xStart != null && yStart != null && xEnd != null && yEnd != null && length != null) {
          if (getArcExtentInDegrees() != null && getArcExtentInDegrees().floatValue() == 0) {
            double wallAngle = Math.atan2(yStart - yEnd, xEnd - xStart);
            setXEnd((float)(xStart + length * Math.cos(wallAngle)));
            setYEnd((float)(yStart - length * Math.sin(wallAngle)));
          } else {
            throw new UnsupportedOperationException(
                "Computing end point of a round wall from its length not supported");
          }
        } else {
          setXEnd(null);
          setYEnd(null);
        }
      }
    }
  }

/*************   Method 62   *************/ 

  private void setBackFaceNormalFlip(Node node, boolean backFaceNormalFlip) {
    if (node instanceof Group) {
      
      Enumeration<?> enumeration = ((Group)node).getAllChildren(); 
      while (enumeration.hasMoreElements()) {
        setBackFaceNormalFlip((Node)enumeration.nextElement(), backFaceNormalFlip);
      }
    } else if (node instanceof Link) {
      setBackFaceNormalFlip(((Link)node).getSharedGroup(), backFaceNormalFlip);
    } else if (node instanceof Shape3D) {
      Appearance appearance = ((Shape3D)node).getAppearance();
      if (appearance == null) {
        appearance = createAppearanceWithChangeCapabilities();
        ((Shape3D)node).setAppearance(appearance);
      }
      PolygonAttributes polygonAttributes = appearance.getPolygonAttributes();
      if (polygonAttributes == null) {
        polygonAttributes = createPolygonAttributesWithChangeCapabilities();
        appearance.setPolygonAttributes(polygonAttributes);
      }
      polygonAttributes.setBackFaceNormalFlip(backFaceNormalFlip);
    }
  }
  
/*************   Method 63   *************/ 

  public void add(TexturesCategory category, CatalogTexture texture) {
    int index = this.categories.indexOf(category);
    
    if (index == -1) {
      category = new TexturesCategory(category.getName());
      add(category);
    } else {
      category = this.categories.get(index);
    }    
    
    category.add(texture);
    
    this.texturesChangeSupport.fireCollectionChanged(texture, 
        category.getIndexOfTexture(texture), CollectionEvent.Type.ADD);
  }

/*************   Method 64   *************/  
  
    private void updateAdvancedComponents() {
    Component root = SwingUtilities.getRoot(this);
    if (root != null) {
      boolean highQuality = controller.getQuality() >= 2;
      boolean advancedComponentsVisible = this.advancedComponentsSeparator.isVisible();
      if (advancedComponentsVisible != highQuality) {
        int componentsHeight = this.advancedComponentsSeparator.getPreferredSize().height + 6
            + this.dateSpinner.getPreferredSize().height + 5
            + this.ceilingLightEnabledCheckBox.getPreferredSize().height;
        this.advancedComponentsSeparator.setVisible(highQuality);
        this.dateLabel.setVisible(highQuality);
        this.dateSpinner.setVisible(highQuality);
        this.timeLabel.setVisible(highQuality);
        this.timeSpinner.setVisible(highQuality);
        this.dayNightLabel.setVisible(highQuality);
        this.ceilingLightEnabledCheckBox.setVisible(highQuality);
        root.setSize(root.getWidth(), 
            root.getHeight() + (advancedComponentsVisible ? -componentsHeight : componentsHeight));
      }
    }   
  }
