package org.argouml.notation;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.Collection;

import org.apache.log4j.Logger;
import org.argouml.model.Model;

public abstract class NotationProvider {

    private static final Logger LOG = Logger.getLogger(NotationProvider.class);
    
    private static final String LIST_SEPARATOR = ", ";
    
    private final Collection<Object[]> listeners = new ArrayList<Object[]>();

    public abstract String getParsingHelp();
    
    public abstract void parse(Object modelElement, String text);

    public abstract String toString(Object modelElement,
            NotationSettings settings);
    
    public void initialiseListener(PropertyChangeListener listener, 
            Object modelElement) {
        addElementListener(listener, modelElement);
    }
    
    public void cleanListener(final PropertyChangeListener listener, 
            final Object modelElement) {
        removeAllElementListeners(listener);
    }
    
    public void updateListener(final PropertyChangeListener listener, 
            Object modelElement,
            PropertyChangeEvent pce) {
      
        if (Model.getUmlFactory().isRemoved(modelElement)) {
            LOG.warn("Encountered deleted object during delete of " 
                    + modelElement);
            return;
        }
        cleanListener(listener, modelElement);
        initialiseListener(listener, modelElement);
    }
    
    protected final void addElementListener(PropertyChangeListener listener, 
            Object element) {
        if (Model.getUmlFactory().isRemoved(element)) {
            LOG.warn("Encountered deleted object during delete of " + element);
            return;
        }
        Object[] entry = new Object[] {element, null};
        if (!listeners.contains(entry)) {
            listeners.add(entry);
            Model.getPump().addModelEventListener(listener, element);
        } else {
            LOG.warn("Attempted duplicate registration of event listener"
                    + " - Element: " + element + " Listener: " + listener);
        }
    }
    
    protected final void addElementListener(PropertyChangeListener listener, 
            Object element, String property) {
        if (Model.getUmlFactory().isRemoved(element)) {
            LOG.warn("Encountered deleted object during delete of " + element);
            return;
        }
        Object[] entry = new Object[] {element, property};
        if (!listeners.contains(entry)) {
            listeners.add(entry);
            Model.getPump().addModelEventListener(listener, element, property);
        } else {
            LOG.debug("Attempted duplicate registration of event listener"
                    + " - Element: " + element + " Listener: " + listener);
        }
    }

    protected final void addElementListener(PropertyChangeListener listener, 
            Object element, String[] property) {
        if (Model.getUmlFactory().isRemoved(element)) {
            LOG.warn("Encountered deleted object during delete of " + element);
            return;
        }
        Object[] entry = new Object[] {element, property};
        if (!listeners.contains(entry)) {
            listeners.add(entry);
            Model.getPump().addModelEventListener(listener, element, property);
        } else {
            LOG.debug("Attempted duplicate registration of event listener"
                    + " - Element: " + element + " Listener: " + listener);
        }
    }
    
    protected final void removeElementListener(PropertyChangeListener listener, 
            Object element) {
        listeners.remove(new Object[] {element, null});
        Model.getPump().removeModelEventListener(listener, element);
    }
    
    protected final void removeAllElementListeners(
            PropertyChangeListener ) {
        for (Object[] lis : listeners) {
            Object property = lilisteners[1];
            if (property == null) {
                Model.getPump().removeModelEventListener(listener, lis[0]);
            } else if (property instanceof String[]) {
                Model.getPump().removeModelEventListener(listener, lis[0],
                        (String[]) property);
            } else if (property instanceof String) {
                Model.getPump().removeModelEventListener(listener, lis[0],
                        (String) property);
            } else {
                throw new RuntimeException(
                        "Internal error in removeAllElementListeners");
            }
        }
        listeners.clear();
    }
    
    protected StringBuilder formatNameList(Collection modelElements) {
        return formatNameList(modelElements, LIST_SEPARATOR);
    }

    protected StringBuilder formatNameList(Collection modelElements, 
            String separator) {
        StringBuilder result = new StringBuilder();
        for (Object element : modelElements) {
            String name = Model.getFacade().getName(element);
            
            result.append(name).append(separator);
        }
        if (result.length() >= separator.length()) {
            result.delete(result.length() - separator.length(), 
                    result.length());
        }
        return result;
    }
}
