package megamek.common.weapons.infantry;

import java.util.Vector;

import megamek.common.BattleArmor;
import megamek.common.Building;
import megamek.common.Compute;
import megamek.common.Entity;
import megamek.common.IGame;
import megamek.common.Infantry;
import megamek.common.Report;
import megamek.common.ToHitData;
import megamek.common.WeaponType;
import megamek.common.actions.WeaponAttackAction;
import megamek.common.weapons.WeaponHandler;
import megamek.server.Server;
import megamek.server.Server.DamageType;

public class InfantryWeaponHandler extends WeaponHandler {
    
    private static final long serialVersionUID = 1425176802065536326L;
    
    public InfantryWeaponHandler(ToHitData t, WeaponAttackAction w, IGame g,
            Server s) {
        super(t, w, g, s);
        bSalvo = true;
    }

    @Override
    protected int calcDamagePerHit() {
        return 1;
    }

    
    @Override
    protected int calcnCluster() {
        return 2;
    }
    
    @Override
    protected int calcHits(Vector<Report> vPhaseReport) {
        int nHitMod = 0;
        if (bGlancing) {
            nHitMod -= 4;
        }
        int troopersHit = 0;
        
        if (ae.getSwarmTargetId() == target.getTargetId()) {
            troopersHit = ((Infantry)ae).getShootingStrength();
        } else if (!(ae instanceof Infantry)) {
            troopersHit = 1;
        } else {
        	troopersHit = Compute.missilesHit(((Infantry) ae)
                .getShootingStrength(), nHitMod, bGlancing);
        }
        double damage = ((InfantryWeapon)wtype).getInfantryDamage();
        if((ae instanceof Infantry) && !(ae instanceof BattleArmor)) {
            
            damage = ((Infantry)ae).getDamagePerTrooper();
        }
        if((ae instanceof Infantry)
                && nRange == 0
                && ae.getCrew().getOptions().booleanOption("tsm_implant")) {
            damage += 0.14;
        }
        int damageDealt = (int) Math.round(damage * troopersHit);
        if((target instanceof Infantry) && !(target instanceof BattleArmor) && wtype.hasFlag(WeaponType.F_INF_BURST)) {
            damageDealt += Compute.d6();
        }
        if((ae instanceof Infantry)
                && nRange == 0
                && ae.getCrew().getOptions().booleanOption("tsm_implant")) {

        }
        if ((target instanceof Infantry) && ((Infantry)target).isMechanized()) {
            damageDealt /= 2;
        }
        
        if ((target instanceof Building) && (wtype.hasFlag(WeaponType.F_INF_NONPENETRATING))) {
            damageDealt = 0;
        }
        if (wtype.hasFlag(WeaponType.F_INF_NONPENETRATING)) {
            damageType = DamageType.NONPENETRATING;
        }
        Report r = new Report(3325);
        r.subject = subjectId;
        if (ae instanceof Infantry) {
            r.add(troopersHit);
            r.add(" troopers ");
        } else { 
            r.add("");
            r.add("");
        }
        r.add(toHit.getTableDesc() + ", causing " + damageDealt
                + " damage.");
        r.newlines = 0;
        vPhaseReport.addElement(r);
        if((target instanceof Infantry) && !(target instanceof BattleArmor)) {
            
            
            nDamPerHit = damageDealt;
            return 1;
        }
        return damageDealt;
    }

    protected int calcnClusterAero(Entity entityTarget) {
        return 5;
    }

    protected int calcAttackValue() {
        int av = 0;
        
        double damage = ((InfantryWeapon)wtype).getInfantryDamage();
        if((ae instanceof Infantry) && !(ae instanceof BattleArmor)) {
            damage = ((Infantry)ae).getDamagePerTrooper();
            av = (int) Math.round(damage * 0.6 * ((Infantry)ae).getShootingStrength());
        }
        if(bDirect) {
            av = Math.min(av+(toHit.getMoS()/3), av*2);
        }
        if(bGlancing) {
            av = (int) Math.floor(av / 2.0);
        }
        return av;
    }

}
