package megamek.test;

import java.awt.Button;
import java.awt.FileDialog;
import java.awt.Frame;
import java.awt.GridLayout;
import java.awt.Label;
import java.awt.Panel;
import java.awt.TextField;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.WindowAdapter;
import java.awt.event.WindowEvent;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Timer;
import java.util.TimerTask;

import javax.swing.SwingUtilities;

import megamek.common.Board;
import megamek.common.net.ConnectionFactory;
import megamek.common.net.ConnectionListenerAdapter;
import megamek.common.net.DisconnectedEvent;
import megamek.common.net.IConnection;
import megamek.common.net.Packet;
import megamek.common.net.PacketReceivedEvent;

public class PacketTool extends Frame implements Runnable {

    private static final long serialVersionUID = 3249150110009720658L;

    private Board board = null;

    private Panel panConnect = null;

    private Panel panXmit = null;

    private TextField hostName = null;

    private TextField hostPort = null;

    private Label boardName = null;

    private Button butSend = null;

    IConnection conn = null;

    public static void main(String[] args) {
        Frame frame = new PacketTool();

        frame.pack();
        frame.setLocationRelativeTo(null);

        frame.setVisible(true);

    }

    public PacketTool() {
        super("Board Transmition");
        Button button = null;
        Panel main = null;
        addWindowListener(new WindowAdapter() {
            @Override
            public void windowClosing(WindowEvent e) {
                quit();
            }
        });

        main = new Panel();
        main.setLayout(new GridLayout(0, 1));
        panConnect = new Panel();
        panConnect.setLayout(new GridLayout(0, 2));
        main.add(panConnect);
        panXmit = new Panel();
        panXmit.setLayout(new GridLayout(0, 1));
        panXmit.setEnabled(false);
        main.add(panXmit);
        this.add(main);

        panConnect.add(new Label(" Connect To:"));
        hostName = new TextField("localhost", 10);
        panConnect.add(hostName);
        panConnect.add(new Label("Port Number:"));
        hostPort = new TextField("2346", 10);
        panConnect.add(hostPort);
        button = new Button("Listen");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                (new Thread(PacketTool.this, "Packet Reader")).start();
            }
        });
        panConnect.add(button);
        button = new Button("Connect");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                connect();
            }
        });
        panConnect.add(button);

        button = new Button("Load Board");
        button.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                boardLoad();
            }
        });
        panXmit.add(button);
        boardName = new Label();
        boardName.setAlignment(Label.CENTER);
        panXmit.add(boardName);
        butSend = new Button("Send");
        butSend.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                send();
            }
        });
        butSend.setEnabled(false);
        panXmit.add(butSend);

    }

    public synchronized void quit() {
        if (null != conn) {
            conn.close();
        }
        System.exit(0);
    }

    public void connect() {
        String host = hostName.getText();
        int port = 0;
        try {
            port = Integer.parseInt(hostPort.getText());
            conn = ConnectionFactory.getInstance().createServerConnection(
                    new Socket(host, port), 1);
            Timer t = new Timer(true);
            final Runnable packetUpdate = new Runnable() {
                public void run() {
                    IConnection connection = conn;
                    if (connection != null) {
                        connection.update();
                    }
                }
            };
            final TimerTask packetUpdate2 = new TimerTask() {
                @Override
                public void run() {
                    try {
                        SwingUtilities.invokeAndWait(packetUpdate);
                    } catch (Exception ie) {
                        
                    }
                }
            };
            t.schedule(packetUpdate2, 500, 150);

            
            System.out.println("Connected to peer.");
            conn.addConnectionListener(connectionListener);

            board = new Board();
            panConnect.setEnabled(false);
            panXmit.setEnabled(true);
        } catch (Throwable err) {
            err.printStackTrace();
        }
    }

    public void boardLoad() {
        FileDialog fd = new FileDialog(this, "Load Board...", FileDialog.LOAD);
        fd.setDirectory("data" + File.separator + "boards");
        if (boardName.getText().length() > 0) {
            fd.setFile(boardName.getText());
        }
        fd.setLocation(this.getLocation().x + 150, this.getLocation().y + 100);
        fd.setVisible(true);

        if (fd.getFile() == null) {
            
            return;
        }
        String curpath = fd.getDirectory();
        String curfile = fd.getFile();
        
        try {
            InputStream is = new FileInputStream(new File(curpath, curfile));
            board.load(is);
            
            is.close();

            boardName.setText(curfile);
            
            butSend.setEnabled(true);
        } catch (IOException ex) {
            System.err.println("error opening file to save!");
            System.err.println(ex);
        }
    }

    public void send() {
        long start = conn.bytesSent();
        Packet packet = new Packet(Packet.COMMAND_SENDING_BOARD, board);
        
        conn.send(packet);
        System.out.print("Bytes sent: ");
        System.out.print(conn.bytesSent() - start);
        System.out.print(", total sent: ");
        System.out.println(conn.bytesSent());
    }

    public void run() {
        int port = 0;
        try {
            port = Integer.parseInt(hostPort.getText());
            ServerSocket serverSocket = new ServerSocket(port);
            Socket s = serverSocket.accept();
            serverSocket.close();

            System.out.println("Accepted peer connection.");

            conn = ConnectionFactory.getInstance().createServerConnection(s, 0);
            
            conn.addConnectionListener(connectionListener);

            board = new Board();
            panConnect.setEnabled(false);
            panXmit.setEnabled(true);

        } catch (Throwable err) {
            err.printStackTrace();
        }
    }

    public synchronized void handle(int id, Packet packet) {
        System.out.print("Connection #");
        System.out.print(id);
        System.out.print(", received a ");
        if (null == packet) {
            System.out.print("null");
        } else {
            switch (packet.getCommand()) {
                case Packet.COMMAND_SERVER_GREETING:
                    System.out.print("COMMAND_SERVER_GREETING");
                    break;
                case Packet.COMMAND_CLIENT_NAME:
                    System.out.print("COMMAND_CLIENT_NAME");
                    break;
                case Packet.COMMAND_LOCAL_PN:
                    System.out.print("COMMAND_LOCAL_PN");
                    break;
                case Packet.COMMAND_PLAYER_ADD:
                    System.out.print("COMMAND_PLAYER_ADD");
                    break;
                case Packet.COMMAND_PLAYER_REMOVE:
                    System.out.print("COMMAND_PLAYER_REMOVE");
                    break;
                case Packet.COMMAND_PLAYER_UPDATE:
                    System.out.print("COMMAND_PLAYER_UPDATE");
                    break;
                case Packet.COMMAND_PLAYER_READY:
                    System.out.print("COMMAND_PLAYER_READY");
                    break;
                case Packet.COMMAND_CHAT:
                    System.out.print("COMMAND_CHAT");
                    break;
                case Packet.COMMAND_ENTITY_ADD:
                    System.out.print("COMMAND_ENTITY_ADD");
                    break;
                case Packet.COMMAND_ENTITY_REMOVE:
                    System.out.print("COMMAND_ENTITY_REMOVE");
                    break;
                case Packet.COMMAND_ENTITY_MOVE:
                    System.out.print("COMMAND_ENTITY_MOVE");
                    break;
                case Packet.COMMAND_ENTITY_DEPLOY:
                    System.out.print("COMMAND_ENTITY_DEPLOY");
                    break;
                case Packet.COMMAND_ENTITY_ATTACK:
                    System.out.print("COMMAND_ENTITY_ATTACK");
                    break;
                case Packet.COMMAND_ENTITY_UPDATE:
                    System.out.print("COMMAND_ENTITY_UPDATE");
                    break;
                case Packet.COMMAND_ENTITY_MODECHANGE:
                    System.out.print("COMMAND_ENTITY_MODECHANGE");
                    break;
                case Packet.COMMAND_ENTITY_MOUNTED_FACINGCHANGE:
                    System.out.print("COMMAND_ENTITY_MOUNTED_FACINGCHANGE");
                    break;
                case Packet.COMMAND_ENTITY_AMMOCHANGE:
                    System.out.print("COMMAND_ENTITY_AMMOCHANGE");
                    break;
                case Packet.COMMAND_ENTITY_VISIBILITY_INDICATOR:
                    System.out.print("COMMAND_ENTITY_VISIBILITY_INDICATOR");
                    break;
                case Packet.COMMAND_CHANGE_HEX:
                    System.out.print("COMMAND_CHANGE_HEX");
                    break;
                case Packet.COMMAND_BLDG_ADD:
                    System.out.print("COMMAND_BLDG_ADD");
                    break;
                case Packet.COMMAND_BLDG_REMOVE:
                    System.out.print("COMMAND_BLDG_REMOVE");
                    break;
                case Packet.COMMAND_BLDG_UPDATE:
                    System.out.print("COMMAND_BLDG_UPDATE_CF");
                    break;
                case Packet.COMMAND_BLDG_COLLAPSE:
                    System.out.print("COMMAND_BLDG_COLLAPSE");
                    break;
                case Packet.COMMAND_PHASE_CHANGE:
                    System.out.print("COMMAND_PHASE_CHANGE");
                    break;
                case Packet.COMMAND_TURN:
                    System.out.print("COMMAND_TURN");
                    break;
                case Packet.COMMAND_ROUND_UPDATE:
                    System.out.print("COMMAND_ROUND_UPDATE");
                    break;
                case Packet.COMMAND_SENDING_BOARD:
                    System.out.print("COMMAND_SENDING_BOARD");
                    
                    Board recvBoard = (Board) packet.getObject(0);
                    try {
                        recvBoard.save(new FileOutputStream("xmit.board"));
                    } catch (IOException ioErr) {
                        ioErr.printStackTrace();
                    }
                    break;
                case Packet.COMMAND_SENDING_ENTITIES:
                    System.out.print("COMMAND_SENDING_ENTITIES");
                    break;
                case Packet.COMMAND_SENDING_PLAYERS:
                    System.out.print("COMMAND_SENDING_PLAYERS");
                    break;
                case Packet.COMMAND_SENDING_TURNS:
                    System.out.print("COMMAND_SENDING_TURNS");
                    break;
                case Packet.COMMAND_SENDING_REPORTS:
                    System.out.print("COMMAND_SENDING_REPORTS");
                    break;
                case Packet.COMMAND_SENDING_GAME_SETTINGS:
                    System.out.print("COMMAND_SENDING_GAME_SETTINGS");
                    break;
                case Packet.COMMAND_SENDING_MAP_SETTINGS:
                    System.out.print("COMMAND_SENDING_MAP_SETTINGS");
                    break;
                case Packet.COMMAND_SENDING_MAP_DIMENSIONS:
                    System.out.print("COMMAND_SENDING_MAP_SETTINGS");
                    break;
                case Packet.COMMAND_END_OF_GAME:
                    System.out.print("COMMAND_END_OF_GAME");
                    break;
                case Packet.COMMAND_DEPLOY_MINEFIELDS:
                    System.out.print("COMMAND_DEPLOY_MINEFIELDS");
                    break;
                case Packet.COMMAND_REVEAL_MINEFIELD:
                    System.out.print("COMMAND_REVEAL_MINEFIELD");
                    break;
                case Packet.COMMAND_REMOVE_MINEFIELD:
                    System.out.print("COMMAND_REMOVE_MINEFIELD");
                    break;
                case Packet.COMMAND_SENDING_MINEFIELDS:
                    System.out.print("COMMAND_SENDING_MINEFIELDS");
                    break;
                case Packet.COMMAND_REROLL_INITIATIVE:
                    System.out.print("COMMAND_REROLL_INITIATIVE");
                    break;
                case Packet.COMMAND_SET_ARTYAUTOHITHEXES:
                    System.out.print("COMMAND_SET_ARTYAUTOHITHEXES");
                    break;
                default:
                    System.out.print("unknown");
                    break;
            }
        }
        System.out.println(" packet.");
    }

    public synchronized void disconnected(IConnection deadConn) {
        
        System.out.println("s: connection " + deadConn.getId() + " disconnected");

        
        panXmit.setEnabled(false);
        butSend.setEnabled(false);
        boardName.setText("");
        board = null;
        deadConn = null;
        panConnect.setEnabled(true);
    }

    private ConnectionListenerAdapter connectionListener = new ConnectionListenerAdapter() {

        @Override
        public void disconnected(DisconnectedEvent e) {
            PacketTool.this.disconnected(e.getConnection());
        }

        @Override
        public void packetReceived(PacketReceivedEvent e) {
            PacketTool.this.handle(e.getConnection().getId(), e.getPacket());
        }

    };
}
