package net.sourceforge.atunes.kernel.modules.pattern;

import java.awt.BorderLayout;
import java.awt.GridBagConstraints;
import java.awt.GridBagLayout;
import java.awt.GridLayout;
import java.awt.Insets;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.KeyAdapter;
import java.awt.event.KeyEvent;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;

import javax.swing.BorderFactory;
import javax.swing.JButton;
import javax.swing.JComboBox;
import javax.swing.JLabel;
import javax.swing.JPanel;
import javax.swing.JScrollPane;
import javax.swing.JTable;
import javax.swing.JTextArea;
import javax.swing.JTextField;
import javax.swing.SwingUtilities;
import javax.swing.table.DefaultTableModel;

import net.sourceforge.atunes.gui.images.Images;
import net.sourceforge.atunes.gui.views.controls.AbstractCustomDialog;
import net.sourceforge.atunes.model.IControlsBuilder;
import net.sourceforge.atunes.model.IFrame;
import net.sourceforge.atunes.model.IStateRepository;
import net.sourceforge.atunes.utils.I18nUtils;
import net.sourceforge.atunes.utils.StringUtils;

public final class PatternInputDialog extends AbstractCustomDialog {

	private final class OkActionListener implements ActionListener {
		private final boolean massiveRecognition;

		private OkActionListener(final boolean massiveRecognition) {
			this.massiveRecognition = massiveRecognition;
		}

		@Override
		public void actionPerformed(final ActionEvent e) {
			PatternInputDialog.this.result = (String) PatternInputDialog.this.patternComboBox
					.getSelectedItem();

			if (PatternInputDialog.this.result != null
					&& !PatternInputDialog.this.result.trim().equals("")) {
				
				for (AbstractPattern pattern : PatternInputDialog.this.patterns
						.getPatternsList()) {
					PatternInputDialog.this.result = PatternInputDialog.this.result
							.replace(pattern.getPattern().toLowerCase(),
									pattern.getPattern());
				}

				List<String> previousPatterns = null;
				if (this.massiveRecognition) {
					previousPatterns = PatternInputDialog.this.stateRepository
							.getMassiveRecognitionPatterns();
				} else {
					previousPatterns = PatternInputDialog.this.stateRepository
							.getRecognitionPatterns();
				}

				if (previousPatterns == null) {
					previousPatterns = new ArrayList<String>();
					if (this.massiveRecognition) {
						PatternInputDialog.this.stateRepository
								.setMassiveRecognitionPatterns(previousPatterns);
					} else {
						PatternInputDialog.this.stateRepository
								.setRecognitionPatterns(previousPatterns);
					}
				}

				
				if (!previousPatterns.contains(PatternInputDialog.this.result)) {
					previousPatterns.add(PatternInputDialog.this.result);
				}
			}
			dispose();
		}
	}

	private static final long serialVersionUID = -5789081662254435503L;

	private JLabel firstElementLabel;

	private JComboBox patternComboBox;

	private JTable patternPreviewTable;

	private JTable availablePatternsTable;

	private String previewString = null;

	private String result = null;

	private static final String[] PREVIEW_COLUMN_NAMES = new String[] {
			I18nUtils.getString("NAME"), I18nUtils.getString("VALUE") };

	private IStateRepository stateRepository;

	private final boolean massiveRecognition;

	private Patterns patterns;

	private PatternMatcher patternMatcher;

	public void setPatternMatcher(final PatternMatcher patternMatcher) {
		this.patternMatcher = patternMatcher;
	}

	public void setPatterns(final Patterns patterns) {
		this.patterns = patterns;
	}

	public void setStateRepository(final IStateRepository stateRepository) {
		this.stateRepository = stateRepository;
	}

	public PatternInputDialog(final IFrame frame,
			final boolean massiveRecognition,
			final IControlsBuilder controlsBuilder) {
		super(frame, 700, 450, controlsBuilder);
		this.massiveRecognition = massiveRecognition;
	}

	@Override
	public void initialize() {
		setResizable(false);
		setIconImage(Images.getImage(Images.APP_LOGO_16).getImage());
		setTitle(I18nUtils.getString("PATTERN_INPUT"));

		
		JTextArea textArea = getControlsBuilder().createTextArea();
		textArea.setText(I18nUtils.getString("PATTERN_INPUT_INSTRUCTIONS"));
		textArea.setBorder(BorderFactory.createEmptyBorder());
		textArea.setWrapStyleWord(true);
		textArea.setEditable(false);
		textArea.setLineWrap(true);
		textArea.setOpaque(false);

		this.firstElementLabel = new JLabel();

		List<String> previousPatterns = null;
		if (this.massiveRecognition) {
			previousPatterns = this.stateRepository
					.getMassiveRecognitionPatterns();
		} else {
			previousPatterns = this.stateRepository.getRecognitionPatterns();
		}
		
		if (previousPatterns != null) {
			Collections.sort(previousPatterns);
		}
		this.patternComboBox = new JComboBox(
				previousPatterns != null ? previousPatterns
						.toArray(new String[previousPatterns.size()])
						: new String[0]);
		this.patternComboBox.setEditable(true);

		JPanel patternPreviewPanel = new JPanel(new BorderLayout());
		this.patternPreviewTable = getLookAndFeel().getTable();
		JScrollPane patternPreviewTableScrollPane = getControlsBuilder()
				.createScrollPane(this.patternPreviewTable);
		patternPreviewPanel.add(patternPreviewTableScrollPane,
				BorderLayout.CENTER);
		patternPreviewPanel.setBorder(BorderFactory.createTitledBorder(
				BorderFactory.createEmptyBorder(),
				I18nUtils.getString("PREVIEW")));

		JPanel availablePatternsPanel = new JPanel(new BorderLayout());
		this.availablePatternsTable = getLookAndFeel().getTable();
		this.availablePatternsTable.addMouseListener(new MouseAdapter() {
			@Override
			public void mouseClicked(final MouseEvent e) {
				if (e.getClickCount() == 2) {
					
					int row = PatternInputDialog.this.availablePatternsTable
							.rowAtPoint(e.getPoint());
					
					String pattern = ((AvailablePatternsDefaultTableModel) PatternInputDialog.this.availablePatternsTable
							.getModel()).getPatternAtRow(row);
					
					String newPattern = StringUtils.getString(
							PatternInputDialog.this.patternComboBox.getEditor()
									.getItem(), pattern);
					PatternInputDialog.this.patternComboBox.getEditor()
							.setItem(newPattern);
					PatternInputDialog.this.patternComboBox
							.setSelectedItem(newPattern);
					
					previewPattern(PatternInputDialog.this.massiveRecognition);
				}
			}
		});
		JScrollPane availablePatternsScrollPane = getControlsBuilder()
				.createScrollPane(this.availablePatternsTable);
		availablePatternsPanel.add(availablePatternsScrollPane,
				BorderLayout.CENTER);
		availablePatternsPanel.setBorder(BorderFactory.createTitledBorder(
				BorderFactory.createEmptyBorder(),
				I18nUtils.getString("AVAILABLE_PATTERNS")));

		JButton okButton = new JButton(I18nUtils.getString("OK"));
		ActionListener okListener = new OkActionListener(
				this.massiveRecognition);
		okButton.addActionListener(okListener);
		JButton cancelButton = new JButton(I18nUtils.getString("CANCEL"));
		cancelButton.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(final ActionEvent e) {
				PatternInputDialog.this.result = null;
				dispose();
			}
		});

		this.patternComboBox.addActionListener(new ActionListener() {
			@Override
			public void actionPerformed(final ActionEvent e) {
				previewPattern(PatternInputDialog.this.massiveRecognition);
			}
		});

		this.patternComboBox.getEditor().getEditorComponent()
				.addKeyListener(new KeyAdapter() {
					@Override
					public void keyTyped(final KeyEvent e) {
						super.keyTyped(e);

						SwingUtilities.invokeLater(new Runnable() {
							@Override
							public void run() {
								previewPattern(PatternInputDialog.this.massiveRecognition);
							}
						});
					}
				});

		GridLayout gl = new GridLayout(1, 2, 30, 0);
		JPanel auxPanel = new JPanel(gl);
		auxPanel.add(okButton);
		auxPanel.add(cancelButton);

		arrangeDialog(textArea, patternPreviewPanel, availablePatternsPanel,
				okButton, auxPanel);
	}

	private void arrangeDialog(final JTextArea textArea,
			final JPanel patternPreviewPanel,
			final JPanel availablePatternsPanel, final JButton okButton,
			final JPanel auxPanel) {
		JPanel panel = new JPanel(new GridBagLayout());
		GridBagConstraints c = new GridBagConstraints();
		c.gridx = 0;
		c.gridy = 0;
		c.weightx = 1;
		c.gridwidth = 2;
		c.fill = GridBagConstraints.HORIZONTAL;
		c.insets = new Insets(10, 10, 10, 10);
		panel.add(textArea, c);
		c.gridy = 1;
		c.insets = new Insets(5, 30, 5, 30);
		panel.add(this.firstElementLabel, c);
		c.gridy = 2;
		panel.add(this.patternComboBox, c);
		c.gridx = 0;
		c.gridy = 3;
		c.weightx = 0.7;
		c.weighty = 1;
		c.gridwidth = 1;
		c.fill = GridBagConstraints.BOTH;
		c.insets = new Insets(5, 5, 5, 5);
		panel.add(patternPreviewPanel, c);
		c.gridx = 1;
		c.weightx = 0.3;
		panel.add(availablePatternsPanel, c);
		c.gridx = 0;
		c.gridy = 4;
		c.gridwidth = 2;
		c.weighty = 0;
		c.fill = GridBagConstraints.NONE;
		c.anchor = GridBagConstraints.CENTER;
		c.insets = new Insets(10, 10, 10, 10);
		panel.add(auxPanel, c);
		add(panel);
		getRootPane().setDefaultButton(okButton);
	}

	void previewPattern(final boolean massiveRecognition) {
		Map<String, String> matches = this.patternMatcher.getPatternMatches(
				((JTextField) this.patternComboBox.getEditor()
						.getEditorComponent()).getText(), this.previewString,
				massiveRecognition);

		String[][] data = new String[matches.size()][2];
		int i = 0;
		for (Entry<String, String> entry : matches.entrySet()) {
			data[i][0] = I18nUtils.getString(entry.getKey());
			data[i][1] = entry.getValue();
			i++;
		}

		((DefaultTableModel) this.patternPreviewTable.getModel())
				.setDataVector(data, PREVIEW_COLUMN_NAMES);
	}

	public String getResult() {
		return this.result;
	}

	public void show(final List<AbstractPattern> availablePatterns,
			final String previewString) {
		this.previewString = previewString;
		this.patternComboBox.setSelectedIndex(-1);
		this.firstElementLabel.setText(previewString);

		String[][] patternsMatrix = new String[availablePatterns.size()][2];
		int i = 0;
		for (AbstractPattern p : availablePatterns) {
			patternsMatrix[i][0] = p.getPattern();
			patternsMatrix[i][1] = p.getDescription();
			i++;
		}
		
		this.patternPreviewTable.setAutoResizeMode(JTable.AUTO_RESIZE_OFF);
		this.patternPreviewTable
				.setColumnModel(new PatternPreviewDefaultTableColumnModel());
		DefaultTableModel patternPreviewTableModel = new PatternPreviewDefaultTableModel(
				new String[0][2], PREVIEW_COLUMN_NAMES);
		this.patternPreviewTable.setModel(patternPreviewTableModel);

		DefaultTableModel availablePatternsTableModel = new AvailablePatternsDefaultTableModel(
				patternsMatrix, new String[] { I18nUtils.getString("PATTERN"),
						I18nUtils.getString("DESCRIPTION") });
		this.availablePatternsTable.setModel(availablePatternsTableModel);
		setVisible(true);
		this.patternComboBox.requestFocus();
	}
}
