package megamek.server;

import java.io.BufferedInputStream;
import java.io.BufferedReader;
import java.io.DataOutputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.InterruptedIOException;
import java.net.HttpURLConnection;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.net.URL;
import java.net.URLEncoder;
import java.net.UnknownHostException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.Date;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TreeMap;
import java.util.TreeSet;
import java.util.Vector;
import java.util.concurrent.ConcurrentLinkedQueue;
import java.util.zip.GZIPInputStream;
import java.util.zip.GZIPOutputStream;

import megamek.MegaMek;
import megamek.client.ui.swing.util.PlayerColors;
import megamek.common.Aero;
import megamek.common.AmmoType;
import megamek.common.BattleArmor;
import megamek.common.Bay;
import megamek.common.BipedMech;
import megamek.common.Board;
import megamek.common.BoardDimensions;
import megamek.common.BombType;
import megamek.common.Building;
import megamek.common.Building.BasementType;
import megamek.common.BuildingTarget;
import megamek.common.CalledShot;
import megamek.common.CommonConstants;
import megamek.common.Compute;
import megamek.common.ComputeECM;
import megamek.common.Configuration;
import megamek.common.ConvFighter;
import megamek.common.Coords;
import megamek.common.Crew;
import megamek.common.CriticalSlot;
import megamek.common.Dropship;
import megamek.common.ECMInfo;
import megamek.common.EjectedCrew;
import megamek.common.Engine;
import megamek.common.Entity;
import megamek.common.EntityMovementMode;
import megamek.common.EntityMovementType;
import megamek.common.EntitySelector;
import megamek.common.EntityWeightClass;
import megamek.common.EquipmentMode;
import megamek.common.EquipmentType;
import megamek.common.FighterSquadron;
import megamek.common.Flare;
import megamek.common.FuelTank;
import megamek.common.Game;
import megamek.common.GameTurn;
import megamek.common.GunEmplacement;
import megamek.common.HexTarget;
import megamek.common.HitData;
import megamek.common.IArmorState;
import megamek.common.IBoard;
import megamek.common.IEntityRemovalConditions;
import megamek.common.IGame;
import megamek.common.IGame.Phase;
import megamek.common.IHex;
import megamek.common.ILocationExposureStatus;
import megamek.common.INarcPod;
import megamek.common.IPlayer;
import megamek.common.ITerrain;
import megamek.common.Infantry;
import megamek.common.InfernoTracker;
import megamek.common.Jumpship;
import megamek.common.LargeSupportTank;
import megamek.common.LocationFullException;
import megamek.common.LosEffects;
import megamek.common.MapSettings;
import megamek.common.Mech;
import megamek.common.MechWarrior;
import megamek.common.Minefield;
import megamek.common.MiscType;
import megamek.common.Mounted;
import megamek.common.MovePath;
import megamek.common.MovePath.MoveStepType;
import megamek.common.MoveStep;
import megamek.common.OffBoardDirection;
import megamek.common.PhysicalResult;
import megamek.common.PilotingRollData;
import megamek.common.PlanetaryConditions;
import megamek.common.Player;
import megamek.common.Protomech;
import megamek.common.QuadMech;
import megamek.common.Report;
import megamek.common.Roll;
import megamek.common.SmallCraft;
import megamek.common.SpaceStation;
import megamek.common.SpecialHexDisplay;
import megamek.common.SuperHeavyTank;
import megamek.common.SupportTank;
import megamek.common.SupportVTOL;
import megamek.common.Tank;
import megamek.common.TargetRoll;
import megamek.common.Targetable;
import megamek.common.Team;
import megamek.common.TeleMissile;
import megamek.common.Terrain;
import megamek.common.Terrains;
import megamek.common.ToHitData;
import megamek.common.TripodMech;
import megamek.common.TurnOrdered;
import megamek.common.TurnVectors;
import megamek.common.UnitLocation;
import megamek.common.VTOL;
import megamek.common.Warship;
import megamek.common.WeaponComparatorBV;
import megamek.common.WeaponType;
import megamek.common.actions.AbstractAttackAction;
import megamek.common.actions.ArtilleryAttackAction;
import megamek.common.actions.AttackAction;
import megamek.common.actions.BAVibroClawAttackAction;
import megamek.common.actions.BreakGrappleAttackAction;
import megamek.common.actions.BrushOffAttackAction;
import megamek.common.actions.ChargeAttackAction;
import megamek.common.actions.ClearMinefieldAction;
import megamek.common.actions.ClubAttackAction;
import megamek.common.actions.DfaAttackAction;
import megamek.common.actions.DodgeAction;
import megamek.common.actions.EntityAction;
import megamek.common.actions.FindClubAction;
import megamek.common.actions.FlipArmsAction;
import megamek.common.actions.GrappleAttackAction;
import megamek.common.actions.JumpJetAttackAction;
import megamek.common.actions.KickAttackAction;
import megamek.common.actions.LayExplosivesAttackAction;
import megamek.common.actions.ProtomechPhysicalAttackAction;
import megamek.common.actions.PunchAttackAction;
import megamek.common.actions.PushAttackAction;
import megamek.common.actions.RamAttackAction;
import megamek.common.actions.RepairWeaponMalfunctionAction;
import megamek.common.actions.SearchlightAttackAction;
import megamek.common.actions.SpotAction;
import megamek.common.actions.TeleMissileAttackAction;
import megamek.common.actions.ThrashAttackAction;
import megamek.common.actions.TorsoTwistAction;
import megamek.common.actions.TriggerAPPodAction;
import megamek.common.actions.TriggerBPodAction;
import megamek.common.actions.TripAttackAction;
import megamek.common.actions.UnjamAction;
import megamek.common.actions.UnjamTurretAction;
import megamek.common.actions.UnloadStrandedAction;
import megamek.common.actions.WeaponAttackAction;
import megamek.common.containers.PlayerIDandList;
import megamek.common.event.GameListener;
import megamek.common.event.GameVictoryEvent;
import megamek.common.net.ConnectionFactory;
import megamek.common.net.ConnectionListenerAdapter;
import megamek.common.net.DisconnectedEvent;
import megamek.common.net.IConnection;
import megamek.common.net.Packet;
import megamek.common.net.PacketReceivedEvent;
import megamek.common.options.GameOptions;
import megamek.common.options.IBasicOption;
import megamek.common.options.IOption;
import megamek.common.options.OptionsConstants;
import megamek.common.preference.PreferenceManager;
import megamek.common.util.BoardUtilities;
import megamek.common.util.HashCodeUtil;
import megamek.common.util.StringUtil;
import megamek.common.verifier.EntityVerifier;
import megamek.common.verifier.TestAero;
import megamek.common.verifier.TestBattleArmor;
import megamek.common.verifier.TestEntity;
import megamek.common.verifier.TestMech;
import megamek.common.verifier.TestSupportVehicle;
import megamek.common.verifier.TestTank;
import megamek.common.weapons.AttackHandler;
import megamek.common.weapons.BPodWeapon;
import megamek.common.weapons.HVACWeapon;
import megamek.common.weapons.MPodWeapon;
import megamek.common.weapons.PPCWeapon;
import megamek.common.weapons.TAGHandler;
import megamek.common.weapons.TSEMPWeapon;
import megamek.common.weapons.Weapon;
import megamek.common.weapons.WeaponHandler;
import megamek.server.commands.AddBotCommand;
import megamek.server.commands.AllowTeamChangeCommand;
import megamek.server.commands.AssignNovaNetServerCommand;
import megamek.server.commands.CheckBVCommand;
import megamek.server.commands.CheckBVTeamCommand;
import megamek.server.commands.DefeatCommand;
import megamek.server.commands.ExportListCommand;
import megamek.server.commands.FixElevationCommand;
import megamek.server.commands.HelpCommand;
import megamek.server.commands.JoinTeamCommand;
import megamek.server.commands.KickCommand;
import megamek.server.commands.ListEntitiesCommand;
import megamek.server.commands.ListSavesCommand;
import megamek.server.commands.LoadGameCommand;
import megamek.server.commands.LocalLoadGameCommand;
import megamek.server.commands.LocalSaveGameCommand;
import megamek.server.commands.NukeCommand;
import megamek.server.commands.ResetCommand;
import megamek.server.commands.RollCommand;
import megamek.server.commands.RulerCommand;
import megamek.server.commands.SaveGameCommand;
import megamek.server.commands.SeeAllCommand;
import megamek.server.commands.ServerCommand;
import megamek.server.commands.ShowEntityCommand;
import megamek.server.commands.ShowTileCommand;
import megamek.server.commands.ShowValidTargetsCommand;
import megamek.server.commands.SkipCommand;
import megamek.server.commands.TeamCommand;
import megamek.server.commands.TraitorCommand;
import megamek.server.commands.VictoryCommand;
import megamek.server.commands.WhoCommand;
import megamek.server.victory.Victory;

import com.thoughtworks.xstream.XStream;

public class Server implements Runnable {

    private static class EntityTargetPair {
        Entity ent;

        Targetable target;

        EntityTargetPair (Entity e, Targetable t) {
            ent = e;
            target = t;
        }

        @Override
        public boolean equals(Object o) {
            if (!(o instanceof EntityTargetPair)) {
                return false;
            }
            EntityTargetPair other = (EntityTargetPair)o;
            return ent.equals(other.ent)
                    && (((target != null) && target.equals(other.target))
                            || ((target == null) && (other.target == null)));
        }

        @Override
        public int hashCode() {
            int hashCode = HashCodeUtil.SEED;
            hashCode = HashCodeUtil.hash(hashCode, ent.getId());
            hashCode = HashCodeUtil.hash(hashCode, target);
            return hashCode;
        }

    }
    
    public enum DamageType {
        NONE, FRAGMENTATION, FLECHETTE, ACID, INCENDIARY, IGNORE_PASSENGER, ANTI_TSM, ANTI_INFANTRY, NAIL_RIVET,
        NONPENETRATING
    }

    public final static String DEFAULT_BOARD = MapSettings.BOARD_SURPRISE;

    private String password;

    private final String metaServerUrl;

    private ServerSocket serverSocket;

    private String motd;

    private class ReceivedPacket {
        public int connId;
        public Packet packet;

        ReceivedPacket(int cid, Packet p) {
            packet = p;
            connId = cid;
        }

    }

    private class PacketPump implements Runnable {

        boolean shouldStop;

        PacketPump() {
            shouldStop = false;
        }

        public void signalEnd() {
            shouldStop = true;
        }

        @Override
        public void run() {
            while (!shouldStop) {
                while (!packetQueue.isEmpty()) {
                    ReceivedPacket rp = packetQueue.poll();
                    synchronized (serverLock) {
                        handle(rp.connId, rp.packet);
                    }
                }
                try {
                    synchronized (packetQueue) {
                        packetQueue.wait();
                    }
                } catch (InterruptedException e) {
                    
                    
                }
            }
        }

    }

    Vector<IConnection> connections = new Vector<IConnection>(4);

    Hashtable<Integer, ConnectionHandler> connectionHandlers = new Hashtable<Integer, ConnectionHandler>();

    ConcurrentLinkedQueue<ReceivedPacket> packetQueue = new ConcurrentLinkedQueue<ReceivedPacket>();

    ConcurrentLinkedQueue<ReceivedPacket> cfrPacketQueue = new ConcurrentLinkedQueue<ReceivedPacket>();

    Vector<IConnection> connectionsPending = new Vector<IConnection>(4);

    Hashtable<Integer, IConnection> connectionIds = new Hashtable<Integer, IConnection>();

    private int connectionCounter;

    IGame game = new Game();

    private Vector<Report> vPhaseReport = new Vector<Report>();

    public Vector<Report> getvPhaseReport() {
        return vPhaseReport;
    }

    private MapSettings mapSettings = new MapSettings();

    private Hashtable<String, ServerCommand> commandsHash = new Hashtable<String, ServerCommand>();

    private Thread connector;

    private PacketPump packetPump;
    private Thread packetPumpThread;

    private Hashtable<Building, Boolean> affectedBldgs = new Hashtable<Building, Boolean>();

    private Vector<PhysicalResult> physicalResults = new Vector<PhysicalResult>();

    private Vector<DynamicTerrainProcessor> terrainProcessors = new Vector<DynamicTerrainProcessor>();

    private Timer watchdogTimer = new Timer("Watchdog Timer");

    private static EntityVerifier entityVerifier;

    private ArrayList<int[]> scheduledNukes = new ArrayList<int[]>();

    private static Server serverInstance = null;

    private String serverAccessKey = null;

    private Timer serverBrowserUpdateTimer = null;

    private int requestedTeam = IPlayer.TEAM_NONE;

    private IPlayer playerChangingTeam = null;

    private boolean changePlayersTeam = false;

    private Set<Coords> hexUpdateSet = new LinkedHashSet<Coords>();

    private ConnectionListenerAdapter connectionListener = new ConnectionListenerAdapter() {

        @Override
        public void disconnected(DisconnectedEvent e) {
            synchronized (serverLock) {
                IConnection conn = e.getConnection();

                System.out.println("s: connection " + conn.getId()
                                   + " disconnected");

                connections.removeElement(conn);
                connectionsPending.removeElement(conn);
                connectionIds.remove(conn.getId());
                ConnectionHandler ch = connectionHandlers.get(conn.getId());
                if (ch != null) {
                    ch.signalStop();
                    connectionHandlers.remove(conn.getId());
                }

                IPlayer player = getPlayer(conn.getId());
                if (null != player) {
                    Server.this.disconnected(player);
                }
            }
        }

        @Override
        public void packetReceived(PacketReceivedEvent e) {
            ReceivedPacket rp = new ReceivedPacket(e.getConnection().getId(),
                                                   e.getPacket());
            
            if (e.getPacket().getCommand() == Packet.COMMAND_CLIENT_FEEDBACK_REQUEST) {
                synchronized (cfrPacketQueue) {
                    cfrPacketQueue.add(rp);
                    cfrPacketQueue.notifyAll();
                }
            } else {
                synchronized (packetQueue) {
                    packetQueue.add(rp);
                    packetQueue.notifyAll();
                }
            }
        }

    };

    private final Object serverLock = new Object();

    public Server(String password, int port) throws IOException {
        this(password, port, false, "");
    }

    public Server(String password, int port, boolean registerWithServerBrowser,
                  String metaServerUrl) throws IOException {
        this.metaServerUrl = metaServerUrl;
        this.password = password.length() > 0 ? password : null;
        
        serverSocket = new ServerSocket(port);

        motd = createMotd();

        game.getOptions().initialize();
        game.getOptions().loadOptions();

        changePhase(IGame.Phase.PHASE_LOUNGE);

        System.out.println("s: starting a new server...");

        try {
            String host = InetAddress.getLocalHost().getHostName();
            System.out.print("s: hostname = '");
            System.out.print(host);
            System.out.print("' port = ");
            System.out.println(serverSocket.getLocalPort());
            InetAddress[] addresses = InetAddress.getAllByName(host);
            for (InetAddress addresse : addresses) {
                System.out.println("s: hosting on address = "
                                   + addresse.getHostAddress());
            }
        } catch (UnknownHostException e) {
            
        }

        System.out.println("s: password = " + this.password);

        registerCommand(new DefeatCommand(this));
        registerCommand(new ExportListCommand(this));
        registerCommand(new FixElevationCommand(this));
        registerCommand(new HelpCommand(this));
        registerCommand(new KickCommand(this));
        registerCommand(new ListSavesCommand(this));
        registerCommand(new LocalSaveGameCommand(this));
        registerCommand(new LocalLoadGameCommand(this));
        registerCommand(new ResetCommand(this));
        registerCommand(new RollCommand(this));
        registerCommand(new SaveGameCommand(this));
        registerCommand(new LoadGameCommand(this));
        registerCommand(new SeeAllCommand(this));
        registerCommand(new SkipCommand(this));
        registerCommand(new VictoryCommand(this));
        registerCommand(new WhoCommand(this));
        registerCommand(new TeamCommand(this));
        registerCommand(new ShowTileCommand(this));
        registerCommand(new ShowEntityCommand(this));
        registerCommand(new RulerCommand(this));
        registerCommand(new ShowValidTargetsCommand(this));
        registerCommand(new AddBotCommand(this));
        registerCommand(new CheckBVCommand(this));
        registerCommand(new CheckBVTeamCommand(this));
        registerCommand(new NukeCommand(this));
        registerCommand(new TraitorCommand(this));
        registerCommand(new ListEntitiesCommand(this));
        registerCommand(new AssignNovaNetServerCommand(this));
        registerCommand(new AllowTeamChangeCommand(this));
        registerCommand(new JoinTeamCommand(this));

        terrainProcessors.add(new FireProcessor(this));
        terrainProcessors.add(new SmokeProcessor(this));
        terrainProcessors.add(new GeyserProcessor(this));
        terrainProcessors.add(new ElevatorProcessor(this));
        terrainProcessors.add(new ScreenProcessor(this));
        terrainProcessors.add(new WeatherProcessor(this));
        terrainProcessors.add(new QuicksandProcessor(this));

        packetPump = new PacketPump();
        packetPumpThread = new Thread(packetPump, "Packet Pump");
        packetPumpThread.start();

        if (registerWithServerBrowser) {

            final TimerTask register = new TimerTask() {
                @Override
                public void run() {
                    registerWithServerBrowser(true,
                                              Server.getServerInstance().metaServerUrl);
                }
            };
            serverBrowserUpdateTimer = new Timer(
                    "Server Browser Register Timer", true);
            serverBrowserUpdateTimer.schedule(register, 1, 40000);
        }

        connector = new Thread(this, "Connection Listener");
        connector.start();

        serverInstance = this;
    }

    public void setGame(IGame g) {
        
        Vector<GameListener> gameListenersClone = new Vector<GameListener>();
        for (GameListener listener : getGame().getGameListeners()) {
            gameListenersClone.add(listener);
        }

        game = g;

        for (GameListener listener : gameListenersClone) {
            getGame().addGameListener(listener);
        }

        for (Iterator<Entity> e = game.getEntities(); e.hasNext(); ) {
            Entity ent = e.next();
            ent.setGame(game);
            if (ent instanceof Mech) {
                ((Mech) ent).setBAGrabBars();
            }
            if (ent instanceof Tank) {
                ((Tank) ent).setBAGrabBars();
            }
        }
        game.setOutOfGameEntitiesVector(game.getOutOfGameEntitiesVector());
        for (Enumeration<IPlayer> e = game.getPlayers(); e.hasMoreElements(); ) {
            IPlayer p = e.nextElement();
            p.setGame(game);
            p.setGhost(true);
        }
        
        for (Enumeration<AttackHandler> a = game.getAttacks(); a
                .hasMoreElements(); ) {
            AttackHandler handler = a.nextElement();
            if (handler instanceof WeaponHandler) {
                ((WeaponHandler) handler).restore();
            }
        }
    }
    public IGame getGame() {
        return game;
    }

    private String createMotd() {
        StringBuffer buf = new StringBuffer();
        buf.append("Welcome to MegaMek.  Server is running version ");
        buf.append(MegaMek.VERSION);
        buf.append(", build date ");
        if (MegaMek.TIMESTAMP > 0L) {
            buf.append(new Date(MegaMek.TIMESTAMP).toString());
        } else {
            buf.append("unknown");
        }
        buf.append('.');

        return buf.toString();
    }

    public boolean isPassworded() {
        return password != null;
    }

    public boolean isPassword(Object guess) {
        return password.equals(guess);
    }

    private void registerCommand(ServerCommand command) {
        commandsHash.put(command.getName(), command);
    }

    public ServerCommand getCommand(String name) {
        return commandsHash.get(name);
    }

    public void die() {
        watchdogTimer.cancel();

        connector = null;
        packetPump.signalEnd();
        packetPumpThread.interrupt();
        packetPumpThread = null;

        try {
            serverSocket.close();
        } catch (IOException ex) {
        }

        for (Enumeration<IConnection> connEnum = connectionsPending.elements(); connEnum
                .hasMoreElements(); ) {
            IConnection conn = connEnum.nextElement();
            conn.close();
        }
        connectionsPending.removeAllElements();

        for (Enumeration<IConnection> connEnum = connections.elements(); connEnum
                .hasMoreElements(); ) {
            IConnection conn = connEnum.nextElement();
            send(conn.getId(), new Packet(Packet.COMMAND_CLOSE_CONNECTION));
        }

        for (Enumeration<IConnection> connEnum = connections.elements(); connEnum
                .hasMoreElements(); ) {
            IConnection conn = connEnum.nextElement();
            conn.close();
        }

        connections.removeAllElements();
        connectionIds.clear();
        if (serverBrowserUpdateTimer != null) {
            serverBrowserUpdateTimer.cancel();
        }
        if (metaServerUrl != "") {
            registerWithServerBrowser(false, metaServerUrl);
        }

        System.out.flush();
    }

    public Enumeration<String> getAllCommandNames() {
        return commandsHash.keys();
    }

    void greeting(int cn) {
        
        sendToPending(cn, new Packet(Packet.COMMAND_SERVER_GREETING));
    }

    public int getFreeConnectionId() {
        while ((getPendingConnection(connectionCounter) != null)
               || (getConnection(connectionCounter) != null)
               || (getPlayer(connectionCounter) != null)) {
            connectionCounter++;
        }
        return connectionCounter;
    }

    public int getFreeEntityId() {
        return game.getNextEntityId();
    }

    private void receivePlayerInfo(Packet packet, int connId) {
        IPlayer player = (IPlayer) packet.getObject(0);
        IPlayer gamePlayer = game.getPlayer(connId);
        if (null != gamePlayer) {
            gamePlayer.setColorIndex(player.getColorIndex());
            gamePlayer.setStartingPos(player.getStartingPos());
            gamePlayer.setTeam(player.getTeam());
            gamePlayer.setCamoCategory(player.getCamoCategory());
            gamePlayer.setCamoFileName(player.getCamoFileName());
            gamePlayer.setNbrMFConventional(player.getNbrMFConventional());
            gamePlayer.setNbrMFCommand(player.getNbrMFCommand());
            gamePlayer.setNbrMFVibra(player.getNbrMFVibra());
            gamePlayer.setNbrMFActive(player.getNbrMFActive());
            gamePlayer.setNbrMFInferno(player.getNbrMFInferno());
            if (gamePlayer.getConstantInitBonus()
                != player.getConstantInitBonus()) {
                sendServerChat("Player " + gamePlayer.getName()
                               + " changed their initiative bonus from "
                               + gamePlayer.getConstantInitBonus()
                               + " to " + player.getConstantInitBonus() + ".");
            }
            gamePlayer.setConstantInitBonus(player.getConstantInitBonus());
        }
    }

    private String correctDupeName(String oldName) {
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            IPlayer player = i.nextElement();
            if (player.getName().equals(oldName)) {
                
                String newName = oldName;
                int dupNum = 2;
                try {
                    dupNum = Integer.parseInt(oldName.substring(oldName
                                                                        .lastIndexOf(".") + 1));
                    dupNum++;
                    newName = oldName.substring(0, oldName.lastIndexOf("."));
                } catch (Exception e) {
                    
                    dupNum = 2;
                }
                newName = newName.concat(".").concat(Integer.toString(dupNum));
                return correctDupeName(newName);
            }
        }
        return oldName;
    }

    private void receivePlayerVersion(Packet packet, int connId) {
        String version = (String) packet.getObject(0);
        String clientChecksum = (String) packet.getObject(1);
        String serverChecksum = MegaMek.getMegaMekSHA256();
        StringBuffer buf = new StringBuffer();
        boolean needs = false;
        if (!version.equals(MegaMek.VERSION)) {
            buf.append("Client/Server version mismatch. Server reports: "
                       + MegaMek.VERSION + ", Client reports: " + version);
            System.out.println("ERROR: Client/Server Version Mismatch -- Client: "+version+" Server: "+MegaMek.VERSION);
            needs = true;
        }
        if (!clientChecksum.equals(serverChecksum)) {
            if (!version.equals(MegaMek.VERSION)) {
                buf.append(System.lineSeparator());
            }
            buf.append("Client/Server checksum mismatch. Server reports: "
                    + serverChecksum + ", Client reports: " + clientChecksum);
            System.out.println("ERROR: Client/Server Checksum Mismatch -- Client: "+clientChecksum+" Server: "+serverChecksum);
            needs = true;
        }

        if (needs) {
            IPlayer player = getPlayer(connId);
            if (null != player) {
                sendServerChat("For " + player.getName() + " Server reports:"
                        + System.lineSeparator()
                        + buf.toString());
            }
        } else {
            System.out.println("SUCCESS: Client/Server Version ("+version+") and Checksum ("+clientChecksum+") matched");
        }
    }

    private void receivePlayerName(Packet packet, int connId) {
        final IConnection conn = getPendingConnection(connId);
        String name = (String) packet.getObject(0);
        boolean returning = false;

        if (conn == null) {
            System.out.println("server: got a client name from a non-pending"
                               + " connection");
            return;
        }

        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            IPlayer player = i.nextElement();
            if (player.getName().equals(name)) {
                if (player.isGhost()) {
                    returning = true;
                    player.setGhost(false);
                    
                    connId = player.getId();
                    conn.setId(connId);
                }
            }
        }

        if (!returning) {
            
            name = correctDupeName(name);
            sendToPending(connId, new Packet(Packet.COMMAND_SERVER_CORRECT_NAME, name));
        }

        connectionsPending.removeElement(conn);
        connections.addElement(conn);
        connectionIds.put(new Integer(conn.getId()), conn);

        if (!returning) {
            int team = IPlayer.TEAM_UNASSIGNED;
            if (game.getPhase() == Phase.PHASE_LOUNGE) {
                team = IPlayer.TEAM_NONE;
                for (IPlayer p : game.getPlayersVector()) {
                    if (p.getTeam() > team) {
                        team = p.getTeam();
                    }
                }
                team++;
            }
            IPlayer newPlayer = new Player(connId, name);
            int colorInd = newPlayer.getColorIndex();
            Enumeration<IPlayer> players = game.getPlayers();
            while (players.hasMoreElements()
                   && (colorInd < IPlayer.colorNames.length)) {
                final IPlayer p = players.nextElement();
                if (p.getId() == newPlayer.getId()) {
                    continue;
                }
                if (p.getColorIndex() == colorInd) {
                    colorInd++;
                }
            }
            if (colorInd == -1) {
                colorInd = 0;
            }
            newPlayer.setColorIndex(colorInd);
            newPlayer.setCamoCategory(IPlayer.NO_CAMO);
            newPlayer.setCamoFileName(IPlayer.colorNames[colorInd]);
            newPlayer.setTeam(Math.min(team, 5));
            game.addPlayer(connId, newPlayer);
            validatePlayerInfo(connId);
        }

        IPlayer player = getPlayer(connId);
        if ((game.getPhase() != IGame.Phase.PHASE_LOUNGE) && (null != player)
            && (game.getEntitiesOwnedBy(player) < 1)) {
            player.setObserver(true);
        }

        sendServerChat(connId, motd);

        send(createPlayerConnectPacket(connId));
        
        send(connId, new Packet(Packet.COMMAND_LOCAL_PN, new Integer(connId)));

        sendCurrentInfo(connId);

        try {
            InetAddress[] addresses = InetAddress.getAllByName(InetAddress
                                                                       .getLocalHost().getHostName());
            for (InetAddress addresse : addresses) {
                sendServerChat(connId,
                               "Machine IP is " + addresse.getHostAddress());
            }
        } catch (UnknownHostException e) {
            
        }

        sendServerChat(connId,
                       "Listening on port " + serverSocket.getLocalPort());

        player = getPlayer(connId);
        if (null != player) {
            StringBuffer buff = new StringBuffer();
            buff.append(player.getName()).append(" connected from ")
                .append(getClient(connId).getInetAddress());
            String who = buff.toString();
            System.out.print("s: player #");
            System.out.print(connId);
            System.out.print(", ");
            System.out.println(who);

            sendServerChat(who);
        } 
    }

    private void sendCurrentInfo(int connId) {
        
        transmitAllPlayerConnects(connId);
        send(connId, createGameSettingsPacket());
        send(connId, createPlanetaryConditionsPacket());

        IPlayer player = game.getPlayer(connId);
        if (null != player) {
            send(connId,
                 new Packet(Packet.COMMAND_SENDING_MINEFIELDS, player
                         .getMinefields()));

            switch (game.getPhase()) {
                case PHASE_LOUNGE:
                    send(connId, createMapSettingsPacket());
                    send(createMapSizesPacket());
                    
                    send(connId,
                         new Packet(Packet.COMMAND_PHASE_CHANGE, game.getPhase()));
                    if (doBlind()) {
                        send(connId, createFilteredFullEntitiesPacket(player));
                    } else {
                        send(connId, createFullEntitiesPacket());
                    }
                    break;
                default:
                    send(connId, new Packet(Packet.COMMAND_ROUND_UPDATE,
                                            new Integer(game.getRoundCount())));
                    send(connId, createBoardPacket());
                    send(connId, createAllReportsPacket(player));

                    
                    if (doBlind()) {
                        send(connId, createFilteredFullEntitiesPacket(player));
                    } else {
                        send(connId, createFullEntitiesPacket());
                    }
                    player.setDone(game.getEntitiesOwnedBy(player) <= 0);
                    send(connId,
                         new Packet(Packet.COMMAND_PHASE_CHANGE, game.getPhase()));
                    break;
            }
            if ((game.getPhase() == IGame.Phase.PHASE_FIRING)
                || (game.getPhase() == IGame.Phase.PHASE_TARGETING)
                || (game.getPhase() == IGame.Phase.PHASE_OFFBOARD)
                || (game.getPhase() == IGame.Phase.PHASE_PHYSICAL)) {
                
                send(connId, createAttackPacket(game.getActionsVector(), 0));
                send(connId, createAttackPacket(game.getChargesVector(), 1));
                send(connId, createAttackPacket(game.getRamsVector(), 1));
                send(connId,
                     createAttackPacket(game.getTeleMissileAttacksVector(),
                                        1));
            }
            if (game.phaseHasTurns(game.getPhase())) {
                send(connId, createTurnVectorPacket());
                send(connId, createTurnIndexPacket());
            }

            send(connId, createArtilleryPacket(player));
            send(connId, createFlarePacket());
            send(connId, createSpecialHexDisplayPacket(connId));
        } 
    }

    public void sendEntities(int connId) {
        if (doBlind()) {
            send(connId, createFilteredEntitiesPacket(getPlayer(connId), null));
        } else {
            send(connId, createEntitiesPacket());
        }
    }

    public void validatePlayerInfo(int playerId) {
        final IPlayer player = getPlayer(playerId);

        boolean[] colorUsed = new boolean[IPlayer.colorNames.length];
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer otherPlayer = i.nextElement();
            if (otherPlayer.getId() != playerId) {
                colorUsed[otherPlayer.getColorIndex()] = true;
            }
        }
        if ((null != player) && colorUsed[player.getColorIndex()]) {
            
            for (int i = 0; i < colorUsed.length; i++) {
                if (!colorUsed[i]) {
                    player.setColorIndex(i);
                    break;
                }
            }
        }

    }

    void disconnected(IPlayer player) {
        IGame.Phase phase = game.getPhase();

        if (phase == IGame.Phase.PHASE_LOUNGE) {
            removeAllEntitesOwnedBy(player);
        }

        if ((phase == IGame.Phase.PHASE_VICTORY)
            || (phase == IGame.Phase.PHASE_LOUNGE) || player.isObserver()) {
            game.removePlayer(player.getId());
            send(new Packet(Packet.COMMAND_PLAYER_REMOVE, new Integer(
                    player.getId())));
            
            if (phase == IGame.Phase.PHASE_LOUNGE) {
                resetActivePlayersDone();
            }
        } else {
            player.setGhost(true);
            player.setDone(true);
            send(createPlayerUpdatePacket(player.getId()));
        }

        if (game.phaseHasTurns(game.getPhase()) && (null != game.getTurn())) {
            if (game.getTurn().isValid(player.getId(), game)) {
                sendGhostSkipMessage(player);
            }
        } else {
            checkReady();
        }

        sendServerChat(player.getName() + " disconnected.");

        System.out.println("s: removed player " + player.getName());

        if (0 == game.getNoOfPlayers()) {
            resetGame();
        }
    }

    public void checkForObservers() {
        for (Enumeration<IPlayer> e = game.getPlayers(); e.hasMoreElements(); ) {
            IPlayer p = e.nextElement();
            p.setObserver((game.getEntitiesOwnedBy(p) < 1)
                          && (game.getPhase() != IGame.Phase.PHASE_LOUNGE));
        }
    }

    public void resetGame() {
        
        game.reset();
        send(createEntitiesPacket());
        send(new Packet(Packet.COMMAND_SENDING_MINEFIELDS, new Vector<Object>()));

        ArrayList<IPlayer> ghosts = new ArrayList<IPlayer>();
        for (Enumeration<IPlayer> players = game.getPlayers(); players
                .hasMoreElements(); ) {
            IPlayer p = players.nextElement();
            if (p.isGhost()) {
                ghosts.add(p);
            } else {
                
                p.setStartingPos(0);
                send(createPlayerUpdatePacket(p.getId()));
            }
        }
        for (IPlayer p : ghosts) {
            game.removePlayer(p.getId());
            send(new Packet(Packet.COMMAND_PLAYER_REMOVE,
                            new Integer(p.getId())));
        }

        resetPlayersDone();
        transmitAllPlayerDones();

        SimpleDateFormat format = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss z");
        System.out.print(format.format(new Date()));
        System.out.println(" END OF GAME");

        changePhase(IGame.Phase.PHASE_LOUNGE);
    }

    public void autoSave() {
        String fileName = "autosave";
        if (PreferenceManager.getClientPreferences().stampFilenames()) {
            fileName = StringUtil.addDateTimeStamp(fileName);
        }
        saveGame(fileName, game.getOptions().booleanOption("autosave_msg"));
    }

    public void sendSaveGame(int connId, String sFile, String sLocalPath) {
        saveGame(sFile, false);
        String sFinalFile = sFile;
        if (!sFinalFile.endsWith(".sav.gz")) {
            if (sFinalFile.endsWith(".sav")) {
                sFinalFile = sFile + ".gz";
            } else {
                sFinalFile = sFile + ".sav.gz";
            }
        }
        sLocalPath = sLocalPath.replaceAll("\\|", " ");
        String localFile = "savegames" + File.separator + sFinalFile;
        try {
            ArrayList<Integer> data = new ArrayList<Integer>();
            BufferedInputStream fin = new BufferedInputStream(
                    new FileInputStream(localFile));
            int input;
            while ((input = fin.read()) != -1) {
                data.add(input);
            }
            send(connId, new Packet(Packet.COMMAND_SEND_SAVEGAME, new Object[]{
                    sFinalFile, data, sLocalPath}));
            sendChat(connId, "***Server", "Savegame has been sent to you.");
            fin.close();
        } catch (Exception e) {
            System.err.println("Unable to load file: " + localFile);
            e.printStackTrace();
        }
    }

    public void saveGame(String sFile, boolean sendChat) {
        
        if (sFile.endsWith(".gz")) {
            sFile = sFile.replace(".gz", "");
        }
        XStream xstream = new XStream();
        String sFinalFile = sFile;
        if (!sFinalFile.endsWith(".sav")) {
            sFinalFile = sFile + ".sav";
        }
        try {
            File sDir = new File("savegames");
            if (!sDir.exists()) {
                sDir.mkdir();
            }

            sFinalFile = sDir + File.separator + sFinalFile;
            GZIPOutputStream gzo = new GZIPOutputStream(new FileOutputStream(
                    sFinalFile + ".gz"));
            xstream.toXML(game, gzo);
            gzo.close();
        } catch (Exception e) {
            System.err.println("Unable to save file: " + sFinalFile);
            e.printStackTrace();
        }

        if (sendChat) {
            sendChat("MegaMek", "Game saved to " + sFinalFile);
        }
    }

    public void saveGame(String sFile) {
        saveGame(sFile, true);
    }

    public void sendLoadGame(int connId, String sFile) {
        String sFinalFile = sFile;
        if (!sFinalFile.endsWith(".sav") && !sFinalFile.endsWith(".sav.gz")) {
            sFinalFile = sFile + ".sav";
        }
        if (!sFinalFile.endsWith(".gz")) {
            sFinalFile = sFinalFile + ".gz";
        }
        send(connId, new Packet(Packet.COMMAND_LOAD_SAVEGAME,
                                new Object[]{sFinalFile}));
    }

    public boolean loadGame(File f) {
        System.out.println("s: loading saved game file '" + f + '\'');
        IGame newGame;
        try {
            XStream xstream = new XStream();
            newGame = (IGame) xstream.fromXML(new GZIPInputStream(
                    new FileInputStream(f)));
        } catch (Exception e) {
            System.err.println("Unable to load file: " + f);
            e.printStackTrace();
            return false;
        }
        setGame(newGame);
        
        for (IConnection conn : connections) {
            sendCurrentInfo(conn.getId());
        }
        return true;
    }

    public IPlayer getPlayer(int id) {
        return game.getPlayer(id);
    }

    private void removeAllEntitesOwnedBy(IPlayer player) {
        Vector<Entity> toRemove = new Vector<Entity>();

        for (Iterator<Entity> e = game.getEntities(); e.hasNext(); ) {
            final Entity entity = e.next();

            if (entity.getOwner().equals(player)) {
                toRemove.addElement(entity);
            }
        }

        for (Entity entity : toRemove) {
            int id = entity.getId();
            game.removeEntity(id, IEntityRemovalConditions.REMOVE_NEVER_JOINED);
            send(createRemoveEntityPacket(id,
                                          IEntityRemovalConditions.REMOVE_NEVER_JOINED));
        }
    }

    private IConnection getClient(int connId) {
        return getConnection(connId);
    }

    public Enumeration<IConnection> getConnections() {
        return connections.elements();
    }

    public IConnection getConnection(int connId) {
        return connectionIds.get(new Integer(connId));
    }

    IConnection getPendingConnection(int connId) {
        for (IConnection conn : connectionsPending) {
            if (conn.getId() == connId) {
                return conn;
            }
        }
        return null;
    }

    private void resetEntityRound() {
        for (Iterator<Entity> e = game.getEntities(); e.hasNext(); ) {
            Entity entity = e.next();

            entity.newRound(game.getRoundCount());
        }
    }

    private void destroyDoomedEntities(Vector<Integer> entityIds) {
        Vector<Entity> toRemove = new Vector<Entity>(0, 10);
        for (Integer entityId : entityIds) {
            Entity entity = game.getEntity(entityId);
            if (entity.isDoomed()) {
                entity.setDestroyed(true);

                final int swarmedId = entity.getSwarmTargetId();
                if (Entity.NONE != swarmedId) {
                    final Entity swarmed = game.getEntity(swarmedId);
                    swarmed.setSwarmAttackerId(Entity.NONE);
                    entity.setSwarmTargetId(Entity.NONE);
                    Report r = new Report(5165);
                    r.subject = swarmedId;
                    r.addDesc(swarmed);
                    addReport(r);
                    entityUpdate(swarmedId);
                }
            }

            if (entity.isDestroyed()) {
                toRemove.addElement(entity);
            }
        }

        for (Entity entity : toRemove) {
            int condition = IEntityRemovalConditions.REMOVE_SALVAGEABLE;
            if (!entity.isSalvage()) {
                condition = IEntityRemovalConditions.REMOVE_DEVASTATED;
            }
            
            if ((game.getPhase() == Phase.PHASE_MOVEMENT)
                && entity.isSelectableThisTurn()) {
                game.removeTurnFor(entity);
                send(createTurnVectorPacket());
            }
            entityUpdate(entity.getId());
            game.removeEntity(entity.getId(), condition);
            send(createRemoveEntityPacket(entity.getId(), condition));
        }
    }

    private void resetEntityPhase(IGame.Phase phase) {
        
        Vector<Entity> toRemove = new Vector<Entity>(0, 10);
        for (Iterator<Entity> e = game.getEntities(); e.hasNext(); ) {
            final Entity entity = e.next();
            entity.newPhase(phase);
            if (entity.isDoomed()) {
                entity.setDestroyed(true);

                final int swarmedId = entity.getSwarmTargetId();
                if (Entity.NONE != swarmedId) {
                    final Entity swarmed = game.getEntity(swarmedId);
                    swarmed.setSwarmAttackerId(Entity.NONE);
                    entity.setSwarmTargetId(Entity.NONE);
                    Report r = new Report(5165);
                    r.subject = swarmedId;
                    r.addDesc(swarmed);
                    addReport(r);
                    entityUpdate(swarmedId);
                }
            }

            if (entity.isDestroyed()) {
                toRemove.addElement(entity);
            }
        }

        for (Entity entity : toRemove) {
            int condition = IEntityRemovalConditions.REMOVE_SALVAGEABLE;
            if (!entity.isSalvage()) {
                condition = IEntityRemovalConditions.REMOVE_DEVASTATED;
            }

            entityUpdate(entity.getId());
            game.removeEntity(entity.getId(), condition);
            send(createRemoveEntityPacket(entity.getId(), condition));
        }

        for (Iterator<Entity> e = game.getEntities(); e.hasNext(); ) {
            final Entity entity = e.next();

            entity.applyDamage();

            entity.reloadEmptyWeapons();

            entity.damageThisRound += entity.damageThisPhase;
            entity.damageThisPhase = 0;
            entity.engineHitsThisPhase = 0;
            entity.rolledForEngineExplosion = false;
            entity.dodging = false;
            entity.setShutDownThisPhase(false);
            entity.setStartupThisPhase(false);

            if (phase == IGame.Phase.PHASE_DEPLOYMENT) {
                entity.setDone(!entity.shouldDeploy(game.getRoundCount()));
            } else {
                entity.setDone(false);
            }

            entity.setIlluminated(false);
            entity.setUsedSearchlight(false);
            entity.setCarefulStand(false);

            if (entity instanceof MechWarrior) {
                ((MechWarrior) entity).setLanded(true);
            }
        }
        game.clearIlluminatedPositions();
        send(new Packet(Packet.COMMAND_CLEAR_ILLUM_HEXES));
    }

    private boolean isReportingPhase() {

        if ((game.getPhase() == IGame.Phase.PHASE_FIRING_REPORT)
            || (game.getPhase() == IGame.Phase.PHASE_INITIATIVE_REPORT)
            || (game.getPhase() == IGame.Phase.PHASE_MOVEMENT_REPORT)
            || (game.getPhase() == IGame.Phase.PHASE_OFFBOARD_REPORT)
            || (game.getPhase() == IGame.Phase.PHASE_PHYSICAL_REPORT)) {
            return true;
        }

        return false;
    }

    private void resetPlayersDone() {
        if (isReportingPhase()) {
            return;
        }
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer player = i.nextElement();
            player.setDone(false);
        }
        transmitAllPlayerDones();
    }

    private void resetActivePlayersDone() {
        
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer player = i.nextElement();

            player.setDone(game.getEntitiesOwnedBy(player) <= 0);
        }
        transmitAllPlayerDones();
    }

    private void prepareVictoryReport() {
        Report r;

        Vector<Entity> toRemove = new Vector<Entity>();
        for (Entity e : game.getEntitiesVector()) {
            if (e.isCarcass() && !e.isDestroyed()) {
                toRemove.add(e);
            }
        }
        for (Entity e : toRemove) {
            destroyEntity(e, "crew death", false, true);
            game.removeEntity(e.getId(),
                              IEntityRemovalConditions.REMOVE_SALVAGEABLE);
            e.setDestroyed(true);
        }
        addReport(new Report(7000, Report.PUBLIC));

        r = new Report(1210);
        r.type = Report.PUBLIC;
        if (game.getVictoryTeam() == IPlayer.TEAM_NONE) {
            IPlayer player = getPlayer(game.getVictoryPlayerId());
            if (null == player) {
                r.messageId = 7005;
            } else {
                r.messageId = 7010;
                r.add(Server.getColorForPlayer(player));
            }
        } else {
            
            r.messageId = 7015;
            r.add(game.getVictoryTeam());
        }
        addReport(r);

        Enumeration<IPlayer> players = game.getPlayers();
        while (players.hasMoreElements()) {
            IPlayer player = players.nextElement();
            r = new Report();
            r.type = Report.PUBLIC;
            r.messageId = 7016;
            r.add(Server.getColorForPlayer(player));
            r.add(player.getBV());
            r.add(player.getInitialBV());
            r.add(player.getFledBV());
            addReport(r);
        }

        Iterator<Entity> survivors = game.getEntities();
        if (survivors.hasNext()) {
            addReport(new Report(7020, Report.PUBLIC));
            while (survivors.hasNext()) {
                Entity entity = survivors.next();

                if (!entity.isDeployed()) {
                    continue;
                }

                addReport(entity.victoryReport());
            }
        }
        
        Iterator<Entity> undeployed = game.getEntities();
        if (undeployed.hasNext()) {
            boolean wroteHeader = false;

            while (undeployed.hasNext()) {
                Entity entity = undeployed.next();

                if (entity.isDeployed()) {
                    continue;
                }

                if (!wroteHeader) {
                    addReport(new Report(7075, Report.PUBLIC));
                    wroteHeader = true;
                }

                addReport(entity.victoryReport());
            }
        }
        
        Enumeration<Entity> retreat = game.getRetreatedEntities();
        if (retreat.hasMoreElements()) {
            addReport(new Report(7080, Report.PUBLIC));
            while (retreat.hasMoreElements()) {
                Entity entity = retreat.nextElement();
                addReport(entity.victoryReport());
            }
        }
        
        Enumeration<Entity> graveyard = game.getGraveyardEntities();
        if (graveyard.hasMoreElements()) {
            addReport(new Report(7085, Report.PUBLIC));
            while (graveyard.hasMoreElements()) {
                Entity entity = graveyard.nextElement();
                addReport(entity.victoryReport());
            }
        }
        
        Enumeration<Entity> devastated = game.getDevastatedEntities();
        if (devastated.hasMoreElements()) {
            addReport(new Report(7090, Report.PUBLIC));

            while (devastated.hasMoreElements()) {
                Entity entity = devastated.nextElement();
                addReport(entity.victoryReport());
            }
        }
        
        addReport(new Report(7095, Report.PUBLIC));
    }

    private String getDetailedVictoryReport() {
        StringBuffer sb = new StringBuffer();

        Vector<Entity> vAllUnits = new Vector<Entity>();
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            vAllUnits.addElement(i.next());
        }

        for (Enumeration<Entity> i = game.getRetreatedEntities(); i
                .hasMoreElements(); ) {
            vAllUnits.addElement(i.nextElement());
        }

        for (Enumeration<Entity> i = game.getGraveyardEntities(); i
                .hasMoreElements(); ) {
            vAllUnits.addElement(i.nextElement());
        }

        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {

            IPlayer p = i.nextElement();
            sb.append("++++++++++ ").append(p.getName()).append(" ++++++++++");
            sb.append(CommonConstants.NL);

            for (int x = 0; x < vAllUnits.size(); x++) {
                Entity e = vAllUnits.elementAt(x);
                if (e.getOwner() == p) {
                    sb.append(UnitStatusFormatter.format(e));
                }
            }

            Enumeration<Entity> devastated = game.getDevastatedEntities();
            if (devastated.hasMoreElements()) {
                sb.append("=============================================================");
                sb.append(CommonConstants.NL);
                sb.append("The following utterly destroyed units are not available for salvage:");
                sb.append(CommonConstants.NL);
                while (devastated.hasMoreElements()) {
                    Entity e = devastated.nextElement();
                    if (e.getOwner() == p) {
                        sb.append(e.getShortName()).append(", Pilot: ")
                          .append(e.getCrew().getName()).append(" (")
                          .append(e.getCrew().getGunnery()).append('/')
                          .append(e.getCrew().getPiloting()).append(')');
                        sb.append(CommonConstants.NL);
                    }
                } 
                sb.append("=============================================================");
                sb.append(CommonConstants.NL);
            }
        } 
        return sb.toString();
    }

    public void forceVictory(IPlayer victor) {
        game.setForceVictory(true);
        if (victor.getTeam() == IPlayer.TEAM_NONE) {
            game.setVictoryPlayerId(victor.getId());
            game.setVictoryTeam(IPlayer.TEAM_NONE);
        } else {
            game.setVictoryPlayerId(IPlayer.PLAYER_NONE);
            game.setVictoryTeam(victor.getTeam());
        }

        Vector<IPlayer> playersVector = game.getPlayersVector();
        for (int i = 0; i < playersVector.size(); i++) {
            IPlayer player = playersVector.elementAt(i);
            player.setAdmitsDefeat(false);
        }
    }

    public void cancelVictory() {
        game.setForceVictory(false);
        game.setVictoryPlayerId(IPlayer.PLAYER_NONE);
        game.setVictoryTeam(IPlayer.TEAM_NONE);
    }

    public void requestTeamChange(int team, IPlayer player) {
        requestedTeam = team;
        playerChangingTeam = player;
        changePlayersTeam = false;
    }

    public void allowTeamChange() {
        changePlayersTeam = true;
    }

    public boolean isTeamChangeRequestInProgress() {
        return playerChangingTeam != null;
    }

    public IPlayer getPlayerRequestingTeamChange() {
        return playerChangingTeam;
    }

    public int getRequestedTeam() {
        return requestedTeam;
    }

    private void processTeamChange() {
        if (playerChangingTeam != null) {
            playerChangingTeam.setTeam(requestedTeam);
            game.setupTeams();
            send(createPlayerUpdatePacket(playerChangingTeam.getId()));
            String teamString = "Team " + requestedTeam + "!";
            if (requestedTeam == IPlayer.TEAM_UNASSIGNED) {
                teamString = " unassigned!";
            } else if (requestedTeam == IPlayer.TEAM_NONE) {
                teamString = " lone wolf!";
            }
            sendServerChat(playerChangingTeam.getName()
                           + " has changed teams to " + teamString);
            playerChangingTeam = null;
        }
        changePlayersTeam = false;
    }

    private void checkReady() {
        
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer player = i.nextElement();
            if (!player.isGhost() && !player.isObserver() && !player.isDone()) {
                return;
            }
        }

        if (game.getNoOfInitiativeRerollRequests() > 0) {
            resetActivePlayersDone();
            game.rollInitAndResolveTies();

            determineTurnOrder(IGame.Phase.PHASE_INITIATIVE);
            clearReports();
            writeInitiativeReport(true);
            sendReport(true);
            return; 
        }

        if (!game.phaseHasTurns(game.getPhase())
            && ((game.getPhase() != IGame.Phase.PHASE_LOUNGE) || (game.getNoOfEntities() > 0))) {
            endCurrentPhase();
        }
    }

    private void endCurrentTurn(Entity entityUsed) {

        boolean turnsChanged = false;
        boolean outOfOrder = false;
        GameTurn turn = game.getTurn();
        if (game.isPhaseSimultaneous() && (entityUsed != null)
            && (entityUsed != null)
            && !turn.isValid(entityUsed.getOwnerId(), game)) {
            
            outOfOrder = true;
            entityUsed.setDone(false);
            GameTurn removed = game.removeFirstTurnFor(entityUsed);
            entityUsed.setDone(true);
            turnsChanged = true;
            if (removed != null) {
                turn = removed;
            }
        }
        final Phase currPhase = game.getPhase();
        final GameOptions gameOpts = game.getOptions();
        final int playerId = null == entityUsed ? IPlayer.PLAYER_NONE
                                                : entityUsed.getOwnerId();
        boolean infMoved = entityUsed instanceof Infantry;
        boolean infMoveMulti = gameOpts.booleanOption("inf_move_multi")
               && ((currPhase == IGame.Phase.PHASE_MOVEMENT)
                   || (currPhase == IGame.Phase.PHASE_DEPLOYMENT)
                   || (currPhase == IGame.Phase.PHASE_INITIATIVE));
        boolean protosMoved = entityUsed instanceof Protomech;
        boolean protosMoveMulti = gameOpts.booleanOption("protos_move_multi");
        boolean tanksMoved = entityUsed instanceof Tank;
        boolean tanksMoveMulti = gameOpts.booleanOption(
                "vehicle_lance_movement")
                && ((currPhase == IGame.Phase.PHASE_MOVEMENT)
                    || (currPhase == IGame.Phase.PHASE_DEPLOYMENT)
                    || (currPhase == IGame.Phase.PHASE_INITIATIVE));
        boolean meksMoved = entityUsed instanceof Mech;
        boolean meksMoveMulti = gameOpts.booleanOption("mek_lance_movement")
                && ((currPhase == IGame.Phase.PHASE_MOVEMENT)
                    || (currPhase == IGame.Phase.PHASE_DEPLOYMENT)
                    || (currPhase == IGame.Phase.PHASE_INITIATIVE));

        int multiMask = 0;
        if (infMoveMulti && infMoved) {
            multiMask = GameTurn.CLASS_INFANTRY;
        } else if (protosMoveMulti && protosMoved) {
            multiMask = GameTurn.CLASS_PROTOMECH;
        } else if (tanksMoveMulti && tanksMoved) {
            multiMask = GameTurn.CLASS_TANK;
        } else if (meksMoveMulti && meksMoved) {
            multiMask = GameTurn.CLASS_MECH;
        }

        List<GameTurn> turnVector = game.getTurnVector();
        int turnIndex = game.getTurnIndex();
        boolean usedEntityNotDone = false;
        if ((turnIndex + 1) < turnVector.size()) {
            GameTurn nextTurn = turnVector.get(turnIndex + 1);
            if (nextTurn instanceof GameTurn.SpecificEntityTurn) {
                GameTurn.SpecificEntityTurn seTurn =
                        (GameTurn.SpecificEntityTurn) nextTurn;
                if ((entityUsed != null)
                        && (seTurn.getEntityNum() == entityUsed.getId())) {
                    turnIndex++;
                    usedEntityNotDone = true;
                }
            }
        }

        boolean isMultiTurn = turn.isMultiTurn();

        if (protosMoved && !protosMoveMulti && !isMultiTurn
            && (entityUsed != null)) {

            final char movingUnit = entityUsed.getUnitNumber();
            final int movingId = entityUsed.getId();

            int protoTurns = game.getSelectedEntityCount(new EntitySelector() {
                private final int ownerId = playerId;

                private final int entityId = movingId;

                private final char unitNum = movingUnit;

                public boolean accept(Entity entity) {
                    if ((entity instanceof Protomech)
                        && entity.isSelectableThisTurn()
                        && (ownerId == entity.getOwnerId())
                        && (entityId != entity.getId())
                        && (unitNum == entity.getUnitNumber())) {
                        return true;
                    }
                    return false;
                }
            });

            for (int i = 0; i < protoTurns; i++) {
                GameTurn newTurn = new GameTurn.UnitNumberTurn(playerId,
                                                               movingUnit);
                newTurn.setMultiTurn(true);
                game.insertTurnAfter(newTurn, turnIndex);
                turnsChanged = true;
            }
        }
        
        else if (((infMoved && infMoveMulti)
                  || (protosMoved && protosMoveMulti)) && !isMultiTurn) {
            int remaining = 0;

            
            if (infMoveMulti && infMoved) {
                remaining += game.getInfantryLeft(playerId);
            }
            if (protosMoveMulti && protosMoved) {
                remaining += game.getProtomechsLeft(playerId);
            }
            if (usedEntityNotDone) {
                remaining--;
            }
            int moreInfAndProtoTurns = Math.min(
                    gameOpts.intOption("inf_proto_move_multi") - 1, remaining);

            
            for (int i = 0; i < moreInfAndProtoTurns; i++) {
                GameTurn newTurn = new GameTurn.EntityClassTurn(playerId,
                                                                multiMask);
                newTurn.setMultiTurn(true);
                game.insertTurnAfter(newTurn, turnIndex);
                turnsChanged = true;
            }
        }

        if (tanksMoved && tanksMoveMulti && !isMultiTurn) {
            int remaining = game.getVehiclesLeft(playerId);
            if (usedEntityNotDone) {
                remaining--;
            }
            int moreVeeTurns = Math.min(
                    gameOpts.intOption("vehicle_lance_movement_number") - 1,
                    remaining);

            for (int i = 0; i < moreVeeTurns; i++) {
                GameTurn newTurn = new GameTurn.EntityClassTurn(playerId,
                                                                multiMask);
                newTurn.setMultiTurn(true);
                game.insertTurnAfter(newTurn, turnIndex);
                turnsChanged = true;
            }
        }

        if (meksMoved && meksMoveMulti && !isMultiTurn) {
            int remaining = game.getMechsLeft(playerId);
            if (usedEntityNotDone) {
                remaining--;
            }
            int moreMekTurns = Math.min(
                    gameOpts.intOption("mek_lance_movement_number") - 1,
                    remaining);

            for (int i = 0; i < moreMekTurns; i++) {
                GameTurn newTurn = new GameTurn.EntityClassTurn(playerId,
                                                                multiMask);
                newTurn.setMultiTurn(true);
                game.insertTurnAfter(newTurn, turnIndex);
                turnsChanged = true;
            }
        }

        if (turnsChanged) {
            send(createTurnVectorPacket());
        }
        if (outOfOrder) {
            send(createTurnIndexPacket());
        } else {
            changeToNextTurn();
        }
    }

    private void changePhase(IGame.Phase phase) {
        game.setLastPhase(game.getPhase());
        game.setPhase(phase);

        prepareForPhase(phase);

        if (isPhasePlayable(phase)) {
            
            send(new Packet(Packet.COMMAND_PHASE_CHANGE, phase));

            executePhase(phase);
        } else {
            endCurrentPhase();
        }
    }

    private void prepareForPhase(IGame.Phase phase) {
        switch (phase) {
            case PHASE_LOUNGE:
                clearReports();
                mapSettings.setBoardsAvailableVector(scanForBoards(
                        mapSettings.getBoardWidth(), mapSettings.getBoardHeight()));
                mapSettings.setNullBoards(DEFAULT_BOARD);
                send(createMapSettingsPacket());
                checkForObservers();
                transmitAllPlayerUpdates();
                break;
            case PHASE_INITIATIVE:
                
                game.resetActions();
                game.resetTagInfo();
                sendTagInfoReset();
                clearReports();
                resetEntityRound();
                resetEntityPhase(phase);
                checkForObservers();
                transmitAllPlayerUpdates();

                resetActivePlayersDone();
                rollInitiative();

                if (!game.shouldDeployThisRound()) {
                    incrementAndSendGameRound();
                }

                determineTurnOrder(phase);
                writeInitiativeReport(false);
                
                checkForConditionDeath();

                checkForBlueShieldDamage();
                if (game.getBoard().inAtmosphere()) {
                    checkForAtmosphereDeath();
                }
                if (game.getBoard().inSpace()) {
                    checkForSpaceDeath();
                }

                System.out.println("Round " + game.getRoundCount()
                                   + " memory usage: " + MegaMek.getMemoryUsed());
                break;
            case PHASE_DEPLOY_MINEFIELDS:
                checkForObservers();
                transmitAllPlayerUpdates();
                resetActivePlayersDone();
                setIneligible(phase);

                Enumeration<IPlayer> e = game.getPlayers();
                Vector<GameTurn> turns = new Vector<GameTurn>();
                while (e.hasMoreElements()) {
                    IPlayer p = e.nextElement();
                    if (p.hasMinefields() && game.getBoard().onGround()) {
                        GameTurn gt = new GameTurn(p.getId());
                        turns.addElement(gt);
                    }
                }
                game.setTurnVector(turns);
                game.resetTurnIndex();

                send(createTurnVectorPacket());
                break;
            case PHASE_SET_ARTYAUTOHITHEXES:
                
                Iterator<Entity> entities = game.getEntities();
                while (entities.hasNext()) {
                    Entity en = entities.next();
                    en.deployOffBoard();
                }
                checkForObservers();
                transmitAllPlayerUpdates();
                resetActivePlayersDone();
                setIneligible(phase);

                Enumeration<IPlayer> players = game.getPlayers();
                Vector<GameTurn> turn = new Vector<GameTurn>();

                while (players.hasMoreElements()) {

                    final IPlayer p = players.nextElement();

                    EntitySelector playerArtySelector = new EntitySelector() {
                        private IPlayer owner = p;

                        public boolean accept(Entity entity) {
                            if (owner.equals(entity.getOwner())
                                && entity.isEligibleForArtyAutoHitHexes()) {
                                return true;
                            }
                            return false;
                        }
                    };
                if (game.getSelectedEntities(playerArtySelector).hasNext()) {
                        
                        GameTurn gt = new GameTurn(p.getId());
                        turn.addElement(gt);
                    }
                }
                game.setTurnVector(turn);
                game.resetTurnIndex();

                send(createTurnVectorPacket());
                break;
            case PHASE_MOVEMENT:
            case PHASE_DEPLOYMENT:
            case PHASE_FIRING:
            case PHASE_PHYSICAL:
            case PHASE_TARGETING:
            case PHASE_OFFBOARD:
                
                if (doBlind()) {
                    updateVisibilityIndicator(null);
                }
                resetEntityPhase(phase);
                checkForObservers();
                transmitAllPlayerUpdates();
                setIneligible(phase);
                determineTurnOrder(phase);
                resetActivePlayersDone();
                
                entityAllUpdate();
                clearReports();
                doTryUnstuck();
                break;
            case PHASE_END:
                resetEntityPhase(phase);
                clearReports();
                resolveHeat();
                if (game.getPlanetaryConditions().isSandBlowing()
                    && (game.getPlanetaryConditions().getWindStrength() > PlanetaryConditions.WI_LIGHT_GALE)) {
                    addReport(resolveBlowingSandDamage());
                }
                addReport(resolveControlRolls());
                addReport(checkForTraitors());
                
                addReport(new Report(5005, Report.PUBLIC));
                resolveHarjelRepairs();
                checkForSuffocation();
                game.getPlanetaryConditions().determineWind();
                send(createPlanetaryConditionsPacket());

                hexUpdateSet.clear();
                for (DynamicTerrainProcessor tp : terrainProcessors) {
                    tp.doEndPhaseChanges(vPhaseReport);
                }
                sendChangedHexes(hexUpdateSet);

                applyBuildingDamage();
                addReport(game.ageFlares());
                send(createFlarePacket());
                resolveAmmoDumps();
                resolveCrewWakeUp();
                resolveSelfDestruct();
                resolveShutdownCrashes();
                checkForIndustrialEndOfTurn();
                resolveMechWarriorPickUp();
                resolveVeeINarcPodRemoval();
                resolveFortify();
                checkForObservers();
                transmitAllPlayerUpdates();
                entityAllUpdate();
                break;
            case PHASE_INITIATIVE_REPORT:
                autoSave();
                
                Enumeration<IPlayer> players2 = game.getPlayers();
                while (players2.hasMoreElements()) {
                    IPlayer player = players2.nextElement();
                    Report r = new Report();
                    r.type = Report.PUBLIC;
                    if (doBlind()) {
                        r.type = Report.PLAYER;
                        r.player = player.getId();
                    }
                    r.messageId = 7016;
                    r.add(Server.getColorForPlayer(player));
                    r.add(player.getBV());
                    r.add(player.getInitialBV());
                    r.add(player.getFledBV());
                    addReport(r);
                }
            case PHASE_TARGETING_REPORT:
            case PHASE_MOVEMENT_REPORT:
            case PHASE_OFFBOARD_REPORT:
            case PHASE_FIRING_REPORT:
            case PHASE_PHYSICAL_REPORT:
            case PHASE_END_REPORT:
                resetActivePlayersDone();
                sendReport();
                if (game.getOptions().booleanOption("paranoid_autosave")) {
                    autoSave();
                }
                break;
            case PHASE_VICTORY:
                resetPlayersDone();
                clearReports();
                prepareVictoryReport();
                game.addReports(vPhaseReport);
                
                for (Iterator<Entity> ents = game.getEntities(); ents.hasNext(); ) {
                    Entity entity = ents.next();
                    if ((entity instanceof Aero)
                        && !(entity instanceof FighterSquadron)) {
                        Aero a = (Aero) entity;
                        if (a.isPartOfFighterSquadron() || a.isCapitalFighter()) {
                            a.doDisbandDamage();
                        }
                    }
                    
                    if (game.getOptions().booleanOption("aero_sanity") && (entity instanceof Aero)) {
                        
                        int scale = 1;
                        if (entity.isCapitalScale()) {
                            scale = 10;
                        }
                        Aero a = (Aero) entity;
                        int currentSI = a.getSI() / (2 * scale);
                        a.set0SI(a.get0SI() / (2 * scale));
                        if (currentSI > 0) {
                            a.setSI(currentSI);
                        }
                        if (scale > 1) {
                            for (int loc = 0; loc < entity.locations(); loc++) {
                                int currentArmor = entity.getArmor(loc) / scale;
                                if (entity.getOArmor(loc) > 0) {
                                    entity.initializeArmor(entity.getOArmor(loc) / scale, loc);
                                }
                                if (entity.getArmor(loc) > 0) {
                                    entity.setArmor(currentArmor, loc);
                                }
                            }
                        }
                    }
                }
                send(createFullEntitiesPacket());
                send(createReportPacket(null));
                send(createEndOfGamePacket());
                break;
            default:
        }
    }

    private boolean isPhasePlayable(IGame.Phase phase) {
        switch (phase) {
            case PHASE_INITIATIVE:
            case PHASE_END:
                return false;
            case PHASE_SET_ARTYAUTOHITHEXES:
            case PHASE_DEPLOY_MINEFIELDS:
            case PHASE_DEPLOYMENT:
            case PHASE_MOVEMENT:
            case PHASE_FIRING:
            case PHASE_PHYSICAL:
            case PHASE_TARGETING:
                return game.hasMoreTurns();
            case PHASE_OFFBOARD:
                return isOffboardPlayable();
            default:
                return true;
        }
    }

    private boolean isOffboardPlayable() {
        if (!game.hasMoreTurns()) {
            return false;
        }
        for (Iterator<Entity> e = game.getEntities(); e.hasNext();) {
            Entity entity = e.next();
            for (Mounted m : entity.getAmmo()) {
                AmmoType atype = (AmmoType) m.getType();
                if (((atype.getAmmoType() == AmmoType.T_LRM)
                        || (atype.getAmmoType() == AmmoType.T_MML)
                        || (atype.getAmmoType() == AmmoType.T_NLRM)
                        || (atype.getAmmoType() == AmmoType.T_MEK_MORTAR))
                        && (atype.getMunitionType() == AmmoType.M_SEMIGUIDED)) {
                    return true;
                }
                if (((atype.getAmmoType() == AmmoType.T_ARROW_IV)
                        || (atype.getAmmoType() == AmmoType.T_LONG_TOM) || (atype
                        .getAmmoType() == AmmoType.T_SNIPER))
                        && (atype.getMunitionType() == AmmoType.M_HOMING)) {
                    return true;
                }
            }
            for (Mounted b : entity.getBombs()) {
                if (!b.isDestroyed()
                    && (b.getUsableShotsLeft() > 0)
                    && (((BombType) b.getType()).getBombType() == BombType.B_LG)) {
                    return true;
                }
            }
        }
        
        for (Enumeration<AttackHandler> attacks = game.getAttacks(); attacks
                .hasMoreElements(); ) {
            AttackHandler ah = attacks.nextElement();
            Mounted ammo = ah.getWaa().getEntity(game)
                             .getEquipment(ah.getWaa().getAmmoId());
            if (ammo != null) {
                AmmoType atype = (AmmoType) ammo.getType();
                if (atype.getMunitionType() == AmmoType.M_HOMING) {
                    return true;
                }
            }
        }
        return false;
    }

    private void executePhase(IGame.Phase phase) {
        switch (phase) {
            case PHASE_EXCHANGE:
                resetPlayersDone();
                calculatePlayerBVs();
                
                game.setupTeams();
                applyBoardSettings();
                game.getPlanetaryConditions().determineWind();
                send(createPlanetaryConditionsPacket());
                
                send(createBoardPacket());
                game.setupRoundDeployment();
                game.setVictoryContext(new HashMap<String, Object>());
                game.createVictoryConditions();
                
                checkEntityExchange();
                break;
            case PHASE_MOVEMENT:
                
                addReport(new Report(2000, Report.PUBLIC));
            case PHASE_SET_ARTYAUTOHITHEXES:
            case PHASE_DEPLOY_MINEFIELDS:
            case PHASE_DEPLOYMENT:
            case PHASE_FIRING:
            case PHASE_PHYSICAL:
            case PHASE_TARGETING:
            case PHASE_OFFBOARD:
                changeToNextTurn();
                if (game.getOptions().booleanOption("paranoid_autosave")) {
                    autoSave();
                }
                break;
            default:
        }
    }

    public void calculatePlayerBVs() {
        for (Enumeration<IPlayer> players = game.getPlayers(); players
                .hasMoreElements(); ) {
            players.nextElement().setInitialBV();
        }
    }

    public void checkEntityExchange() {
        for (Iterator<Entity> entities = game.getEntities(); entities.hasNext();) {
            Entity entity = entities.next();
            if (entity instanceof Aero) {
                Aero a = (Aero) entity;
                if (game.getBoard().inSpace()) {
                    
                    a.liftOff(0);
                } else {
                    
                    if (game.getBoard().inAtmosphere() && !entity.isAirborne()) {
                        
                        a.liftOff(a.getAltitude());
                    }
                }
                
                if (game.getOptions().booleanOption("at2_nukes")
                    && ((entity instanceof Dropship) || (entity instanceof Jumpship))) {
                    entity.applySantaAnna();
                }
                if (!((entity instanceof SmallCraft) || (entity instanceof Jumpship))) {
                    ((Aero) entity).applyBombs();
                }
            }

            if (entity.getLoadedUnits().size() > 0) {
                Vector<Integer> v = new Vector<Integer>();
                for (Entity en : entity.getLoadedUnits()) {
                    v.add(en.getId());
                }
                entity.setLoadedKeepers(v);
            }

            if (game.getOptions().booleanOption("aero_sanity")
                && (entity instanceof Aero)) {
                Aero a = (Aero) entity;
                if (entity.isCapitalScale()) {
                    int currentSI = a.getSI() * 20;
                    a.initializeSI(a.get0SI() * 20);
                    a.setSI(currentSI);
                    if (entity.isCapitalFighter()) {
                        a.autoSetCapArmor();
                        a.autoSetFatalThresh();
                    } else {
                        
                        
                        for (int loc = 0; loc < entity.locations(); loc++) {
                            if (entity.getArmor(loc) > 0) {
                                int currentArmor = entity.getArmor(loc) * 10;
                                entity.initializeArmor(
                                        entity.getOArmor(loc) * 10, loc);
                                entity.setArmor(currentArmor, loc);

                            }
                        }
                    }
                } else {
                    int currentSI = a.getSI() * 2;
                    a.initializeSI(a.get0SI() * 2);
                    a.setSI(currentSI);
                }
            }
            
            entity.setExternalSpotlight(entity.hasExternaSpotlight()
                                        || entity.hasQuirk(OptionsConstants.QUIRK_POS_SEARCHLIGHT));
            entityUpdate(entity.getId());
        }
    }

    private void endCurrentPhase() {
        switch (game.getPhase()) {
            case PHASE_LOUNGE:
                changePhase(IGame.Phase.PHASE_EXCHANGE);
                break;
            case PHASE_EXCHANGE:
                changePhase(IGame.Phase.PHASE_SET_ARTYAUTOHITHEXES);
                break;
            case PHASE_STARTING_SCENARIO:
                changePhase(IGame.Phase.PHASE_SET_ARTYAUTOHITHEXES);
                break;
            case PHASE_SET_ARTYAUTOHITHEXES:
                sendSpecialHexDisplayPackets();
                Enumeration<IPlayer> e = game.getPlayers();
                boolean mines = false;
                while (e.hasMoreElements() && !mines) {
                    IPlayer p = e.nextElement();
                    if (p.hasMinefields()) {
                        mines = true;
                    }
                }
                if (mines) {
                    changePhase(IGame.Phase.PHASE_DEPLOY_MINEFIELDS);
                } else {
                    changePhase(IGame.Phase.PHASE_INITIATIVE);
                }
                break;
            case PHASE_DEPLOY_MINEFIELDS:
                changePhase(IGame.Phase.PHASE_INITIATIVE);
                break;
            case PHASE_DEPLOYMENT:
                game.clearDeploymentThisRound();
                game.checkForCompleteDeployment();
                Enumeration<IPlayer> pls = game.getPlayers();
                while (pls.hasMoreElements()) {
                    IPlayer p = pls.nextElement();
                    p.adjustStartingPosForReinforcements();
                }

                if (game.getRoundCount() < 1) {
                    changePhase(IGame.Phase.PHASE_INITIATIVE);
                } else {
                    changePhase(IGame.Phase.PHASE_TARGETING);
                }
                break;
            case PHASE_INITIATIVE:
                resolveWhatPlayersCanSeeWhatUnits();
                game.addReports(vPhaseReport);
                changePhase(IGame.Phase.PHASE_INITIATIVE_REPORT);
                break;
            case PHASE_INITIATIVE_REPORT:
                
                game.setupRoundDeployment();
                
                if (game.shouldDeployThisRound()) {
                    changePhase(IGame.Phase.PHASE_DEPLOYMENT);
                } else {
                    changePhase(IGame.Phase.PHASE_TARGETING);
                }
                break;
            case PHASE_MOVEMENT:
                resolveWhatPlayersCanSeeWhatUnits();
                doAllAssaultDrops();
                addMovementHeat();
                applyBuildingDamage();
                checkForPSRFromDamage();
                addReport(resolvePilotingRolls()); 
                
                checkForFlamingDamage();
                checkForTeleMissileAttacks();
                cleanupDestroyedNarcPods();
                checkForFlawedCooling();
                
                if (vPhaseReport.size() > 1) {
                    game.addReports(vPhaseReport);
                    changePhase(IGame.Phase.PHASE_MOVEMENT_REPORT);
                } else {
                    
                    addReport(new Report(1205, Report.PUBLIC));
                    game.addReports(vPhaseReport);
                    sendReport();
                    changePhase(IGame.Phase.PHASE_OFFBOARD);
                }
                break;
            case PHASE_MOVEMENT_REPORT:
                changePhase(IGame.Phase.PHASE_OFFBOARD);
                break;
            case PHASE_FIRING:
                addReport(new Report(3000, Report.PUBLIC));
                resolveWhatPlayersCanSeeWhatUnits();
                resolveAllButWeaponAttacks();
                resolveSelfDestructions();
                reportGhostTargetRolls();
                reportLargeCraftECCMRolls();
                resolveOnlyWeaponAttacks();
                assignAMS();
                handleAttacks();
                resolveAeroElevationLoss();
                resolveScheduledNukes();
                applyBuildingDamage();
                checkForPSRFromDamage();
                cleanupDestroyedNarcPods();
                addReport(resolvePilotingRolls());
                checkForFlawedCooling();
                
                if (vPhaseReport.size() > 1) {
                    game.addReports(vPhaseReport);
                    changePhase(IGame.Phase.PHASE_FIRING_REPORT);
                } else {
                    
                    addReport(new Report(1205, Report.PUBLIC));
                    sendReport();
                    game.addReports(vPhaseReport);
                    changePhase(IGame.Phase.PHASE_PHYSICAL);
                }
                break;
            case PHASE_FIRING_REPORT:
                changePhase(IGame.Phase.PHASE_PHYSICAL);
                break;
            case PHASE_PHYSICAL:
                resolveWhatPlayersCanSeeWhatUnits();
                resolvePhysicalAttacks();
                applyBuildingDamage();
                checkForPSRFromDamage();
                addReport(resolvePilotingRolls());
                resolveSinkVees();
                cleanupDestroyedNarcPods();
                checkForFlawedCooling();
                checkForChainWhipGrappleChecks();
                
                if (vPhaseReport.size() > 1) {
                    game.addReports(vPhaseReport);
                    changePhase(IGame.Phase.PHASE_PHYSICAL_REPORT);
                } else {
                    
                    addReport(new Report(1205, Report.PUBLIC));
                    game.addReports(vPhaseReport);
                    sendReport();
                    changePhase(IGame.Phase.PHASE_END);
                }
                break;
            case PHASE_PHYSICAL_REPORT:
                changePhase(IGame.Phase.PHASE_END);
                break;
            case PHASE_TARGETING:
                vPhaseReport.addElement(new Report(1035, Report.PUBLIC));
                resolveAllButWeaponAttacks();
                resolveOnlyWeaponAttacks();
                handleAttacks();
                
                if (vPhaseReport.size() > 1) {
                    game.addReports(vPhaseReport);
                    changePhase(IGame.Phase.PHASE_TARGETING_REPORT);
                } else {
                    
                    vPhaseReport.addElement(new Report(1205, Report.PUBLIC));
                    game.addReports(vPhaseReport);
                    sendReport();
                    changePhase(IGame.Phase.PHASE_MOVEMENT);
                }

                sendSpecialHexDisplayPackets();
                for (Enumeration<IPlayer> i = game.getPlayers(); i
                        .hasMoreElements(); ) {
                    IPlayer player = i.nextElement();
                    int connId = player.getId();
                    send(connId, createArtilleryPacket(player));
                }

                break;
            case PHASE_OFFBOARD:
                
                addReport(new Report(1100, Report.PUBLIC));
                resolveAllButWeaponAttacks(); 
                
                resolveOnlyWeaponAttacks(); 
                handleAttacks();
                for (Enumeration<IPlayer> i = game.getPlayers(); i
                        .hasMoreElements(); ) {
                    IPlayer player = i.nextElement();
                    int connId = player.getId();
                    send(connId, createArtilleryPacket(player));
                }
                applyBuildingDamage();
                checkForPSRFromDamage();
                addReport(resolvePilotingRolls());

                cleanupDestroyedNarcPods();
                checkForFlawedCooling();

                sendSpecialHexDisplayPackets();
                sendTagInfoUpdates();

                if (vPhaseReport.size() > 1) {
                    game.addReports(vPhaseReport);
                    changePhase(IGame.Phase.PHASE_OFFBOARD_REPORT);
                } else {
                    addReport(new Report(1205, Report.PUBLIC));
                    game.addReports(vPhaseReport);
                    sendReport();
                    changePhase(IGame.Phase.PHASE_FIRING);
                }
                break;
            case PHASE_OFFBOARD_REPORT:
                sendSpecialHexDisplayPackets();
                changePhase(IGame.Phase.PHASE_FIRING);
                break;
            case PHASE_TARGETING_REPORT:
                changePhase(IGame.Phase.PHASE_MOVEMENT);
                break;
            case PHASE_END:
                
                resetEntityPhase(IGame.Phase.PHASE_END);
                boolean victory = victory(); 
                
                if ((vPhaseReport.size() > 3)
                    || ((vPhaseReport.size() > 1) && (vPhaseReport.elementAt(1).messageId != 1205))) {
                    game.addReports(vPhaseReport);
                    changePhase(IGame.Phase.PHASE_END_REPORT);
                } else {
                   
                    addReport(new Report(1205, Report.PUBLIC));
                    game.addReports(vPhaseReport);
                    sendReport();
                    if (victory) {
                        changePhase(IGame.Phase.PHASE_VICTORY);
                    } else {
                        changePhase(IGame.Phase.PHASE_INITIATIVE);
                    }
                }
                break;
            case PHASE_END_REPORT:
                if (changePlayersTeam) {
                    processTeamChange();
                }
                if (victory()) {
                    changePhase(IGame.Phase.PHASE_VICTORY);
                } else {
                    changePhase(IGame.Phase.PHASE_INITIATIVE);
                }
                break;
            case PHASE_VICTORY:
                GameVictoryEvent gve = new GameVictoryEvent(this, game);
                game.processGameEvent(gve);
                transmitGameVictoryEventToAll();
                resetGame();
                break;
            default:
        }
    }

    private void sendSpecialHexDisplayPackets() {
        if (connections == null) {
            return;
        }
        for (int i = 0; i < connections.size(); i++) {
            if (connections.get(i) != null) {
                connections.get(i).send(createSpecialHexDisplayPacket(i));
            }
        }
    }

    private void sendTagInfoUpdates() {
        if (connections == null) {
            return;
        }
        for (int i = 0; i < connections.size(); i++) {
            if (connections.get(i) != null) {
                connections.get(i).send(createTagInfoUpdatesPacket());
            }
        }
    }

    public void sendTagInfoReset() {
        if (connections == null) {
            return;
        }
        for (int i = 0; i < connections.size(); i++) {
            if (connections.get(i) != null) {
                connections.get(i).send(
                        new Packet(Packet.COMMAND_RESET_TAGINFO));
            }
        }
    }

    private void incrementAndSendGameRound() {
        game.incrementRoundCount();
        send(new Packet(Packet.COMMAND_ROUND_UPDATE, new Integer(
                game.getRoundCount())));
    }

    private void receiveForwardIni(Packet packet, int connid) {
        

        IPlayer current = getPlayer(connid);

        if (game.getTurn().getPlayerNum() != current.getId()) {
            
            return;
        }
        
        if (game.getOptions().booleanOption("individual_initiative")) {
            return;
        }
        
        Team currentteam = game.getTeamForPlayer(current);
        IPlayer next = currentteam.getNextValidPlayer(current, game);
        
        
        if ((next != null) && (game.getEntitiesOwnedBy(next) != 0)
            && (game.getTurnForPlayer(next.getId()) != null)) {

            int currentturnindex = game.getTurnIndex();
            
            List<GameTurn> turns = game.getTurnVector();
            GameTurn turn = game.getTurn();
            
            
            boolean isGeneralMoveTurn = (!(turn instanceof GameTurn.SpecificEntityTurn)
                                         && !(turn instanceof GameTurn.UnitNumberTurn)
                                         && !(turn instanceof GameTurn.UnloadStrandedTurn)
                                         && !(turn instanceof GameTurn.TriggerBPodTurn) && !(turn instanceof GameTurn
                    .TriggerAPPodTurn));
            if (!isGeneralMoveTurn) {
                
                
                return;
            }
            
            boolean isEntityClassTurn = (turn instanceof GameTurn.EntityClassTurn);
            int classmask = 0;
            if (isEntityClassTurn) {
                GameTurn.EntityClassTurn tempturn = (GameTurn.EntityClassTurn) turn;
                classmask = tempturn.getTurnCode();
            }

            boolean switched = false;
            int nextturnid = 0;
            for (int i = currentturnindex; i < turns.size(); i++) {
                
                if (turns.get(i).isValid(next.getId(), game)) {
                    nextturnid = i;
                    if (isEntityClassTurn) {
                        
                        if ((turns.get(i) instanceof GameTurn.EntityClassTurn)) {
                            
                            if (!(((GameTurn.EntityClassTurn) turns.get(i))
                                          .getTurnCode() == classmask)) {
                                
                                continue;
                            }
                        } else {
                            continue;
                        }
                    }
                    switched = true;
                    break;
                }
            }

            if (switched) {
                game.swapTurnOrder(currentturnindex, nextturnid);
                
                send(createTurnVectorPacket());
                send(createTurnIndexPacket());
            } else {
                
                return;
            }

        }
    }

    private void changeToNextTurn() {
        
        if (!game.hasMoreTurns()) {
            endCurrentPhase();
            return;
        }

        GameTurn nextTurn = game.changeToNextTurn();

        IPlayer player = getPlayer(nextTurn.getPlayerNum());

        if ((player != null) && (game.getEntitiesOwnedBy(player) == 0)) {
            endCurrentTurn(null);
            return;
        }

        send(createTurnIndexPacket());

        if ((null != player) && player.isGhost()) {
            sendGhostSkipMessage(player);
        } else if ((null == game.getFirstEntity())
                && (null != player)
                && ((game.getPhase() != IGame.Phase.PHASE_DEPLOY_MINEFIELDS)
                        && (game.getPhase()
                                != IGame.Phase.PHASE_SET_ARTYAUTOHITHEXES))) {
            sendTurnErrorSkipMessage(player);
        }
    }

    private void sendGhostSkipMessage(IPlayer ghost) {
        StringBuffer message = new StringBuffer();
        message.append("Player '")
               .append(ghost.getName())
               .append("' is disconnected.  You may skip his/her current turn with the /skip command.");
        sendServerChat(message.toString());
    }

    private void sendTurnErrorSkipMessage(IPlayer skip) {
        StringBuffer message = new StringBuffer();
        message.append("Player '")
               .append(skip.getName())
               .append("' has no units to move.  You should skip his/her/your current turn with the /skip command. " +
                       "You may want to report this error.  See the MegaMek homepage (http://megamek.sf.net/) for " +
                       "details.");
        sendServerChat(message.toString());
    }

    public void skipCurrentTurn() {
        
        Entity toSkip = game.getFirstEntity();

        switch (game.getPhase()) {
            case PHASE_DEPLOYMENT:
                
                
                endCurrentTurn(null);
                break;
            case PHASE_MOVEMENT:
                if (toSkip != null) {
                    processMovement(toSkip, new MovePath(game, toSkip), null);
                }
                endCurrentTurn(toSkip);
                break;
            case PHASE_FIRING:
            case PHASE_PHYSICAL:
            case PHASE_TARGETING:
            case PHASE_OFFBOARD:
                if (toSkip != null) {
                    processAttack(toSkip, new Vector<EntityAction>(0));
                }
                endCurrentTurn(toSkip);
                break;
            default:
        }
    }

    public boolean isTurnSkippable() {
        GameTurn turn = game.getTurn();
        if (null == turn) {
            return false;
        }
        IPlayer player = getPlayer(turn.getPlayerNum());
        return (null == player) || player.isGhost()
               || (game.getFirstEntity() == null);
    }

    public boolean victory() {
        Victory.Result vr = game.getVictory().victory(game,
                                                      game.getVictoryContext());
        for (Report r : vr.getReports()) {
            addReport(r);
        }

        if (vr.victory()) {
            boolean draw = vr.isDraw();
            int wonPlayer = vr.getWinningPlayer();
            int wonTeam = vr.getWinningTeam();

            if (wonPlayer != IPlayer.PLAYER_NONE) {
                Report r = new Report(7200, Report.PUBLIC);
                r.add(Server.getColorForPlayer(game.getPlayer(wonPlayer)));
                addReport(r);
            }
            if (wonTeam != IPlayer.TEAM_NONE) {
                Report r = new Report(7200, Report.PUBLIC);
                r.add("Team " + wonTeam);
                addReport(r);
            }
            if (draw) {
                
                game.setVictoryPlayerId(IPlayer.PLAYER_NONE);
                game.setVictoryTeam(IPlayer.TEAM_NONE);
            } else {
                
                
                
                game.setVictoryPlayerId(wonPlayer);
                game.setVictoryTeam(wonTeam);
            }
        } else {
            game.setVictoryPlayerId(IPlayer.PLAYER_NONE);
            game.setVictoryTeam(IPlayer.TEAM_NONE);
            if (game.isForceVictory()) {
                cancelVictory();
            }
        }
        return vr.victory();
    }

    private boolean isPlayerForcedVictory() {
        
        if (!game.getOptions().booleanOption("skip_forced_victory")) {
            return false;
        }

        if (!game.isForceVictory()) {
            return false;
        }

        for (IPlayer player : game.getPlayersVector()) {

            if ((player.getId() == game.getVictoryPlayerId())
                || ((player.getTeam() == game.getVictoryTeam()) && (game
                                                                            .getVictoryTeam() != IPlayer.TEAM_NONE))) {
                continue;
            }

            if (!player.admitsDefeat()) {
                return false;
            }
        }

        return true;
    }

    public void applyBoardSettings() {
        mapSettings.replaceBoardWithRandom(MapSettings.BOARD_RANDOM);
        mapSettings.replaceBoardWithRandom(MapSettings.BOARD_SURPRISE);
        IBoard[] sheetBoards = new IBoard[mapSettings.getMapWidth()
                                          * mapSettings.getMapHeight()];
        for (int i = 0; i < (mapSettings.getMapWidth() * mapSettings
                .getMapHeight()); i++) {
            sheetBoards[i] = new Board();
            String name = mapSettings.getBoardsSelectedVector().get(i);
            boolean isRotated = false;
            if (name.startsWith(Board.BOARD_REQUEST_ROTATION)) {
                
                if ((mapSettings.getBoardWidth() % 2) == 0) {
                    isRotated = true;
                }
                name = name.substring(Board.BOARD_REQUEST_ROTATION.length());
            }
            if (name.startsWith(MapSettings.BOARD_GENERATED)
                || (mapSettings.getMedium() == MapSettings.MEDIUM_SPACE)) {
                sheetBoards[i] = BoardUtilities.generateRandom(mapSettings);
            } else {
                sheetBoards[i].load(new File(Configuration.boardsDir(), name
                                                                        + ".board"));
                BoardUtilities.flip(sheetBoards[i], isRotated, isRotated);
            }
        }
        IBoard newBoard = BoardUtilities.combine(mapSettings.getBoardWidth(),
                                                 mapSettings.getBoardHeight(), mapSettings.getMapWidth(),
                                                 mapSettings.getMapHeight(), sheetBoards,
                                                 mapSettings.getMedium());
        if (game.getOptions().getOption("bridgeCF").intValue() > 0) {
            newBoard.setBridgeCF(game.getOptions().getOption("bridgeCF")
                                     .intValue());
        }
        if (!game.getOptions().booleanOption("random_basements")) {
            newBoard.setRandomBasementsOff();
        }
        BoardUtilities.addWeatherConditions(newBoard, game
                .getPlanetaryConditions().getWeather(), game
                                                    .getPlanetaryConditions().getWindStrength());
        game.setBoard(newBoard);
    }

    private void rollInitiative() {
        if (game.getOptions().booleanOption("individual_initiative")) {
            TurnOrdered.rollInitiative(game.getEntitiesVector(), false);
        } else {
            
            TurnOrdered.rollInitiative(
                    game.getTeamsVector(),
                    game.getOptions().booleanOption(
                            "initiative_streak_compensation")
                    && !game.shouldDeployThisRound());
        }

        transmitAllPlayerUpdates();
    }

    private void determineTurnOrderIUI(IGame.Phase phase) {
        for (Iterator<Entity> loop = game.getEntities(); loop.hasNext();) {
            final Entity entity = loop.next();
            entity.resetOtherTurns();
            if (entity.isSelectableThisTurn()) {
                entity.incrementOtherTurns();
            }
        }

        List<Entity> entities;
        
        boolean protosMoveMulti = game.getOptions().booleanOption(
                "protos_move_multi");
        if (!protosMoveMulti) {
            entities = new ArrayList<>(game.getEntitiesVector().size());
            Set<Character> movedUnits = new HashSet<Character>();
            for (Entity e : game.getEntitiesVector()) {
                
                if (!(e instanceof Protomech)) {
                    entities.add(e);
                    continue;
                }
                char unitNumber = e.getUnitNumber();
                if ((unitNumber == Entity.NONE)
                        || !movedUnits.contains(unitNumber)) {
                    entities.add(e);
                    if (unitNumber != Entity.NONE) {
                        movedUnits.add(unitNumber);
                    }
                }
            }
        } else {
            entities = game.getEntitiesVector();
        }
        
        TurnVectors team_order = TurnOrdered.generateTurnOrder(entities, game);

        
        Iterator<Entity> strandedUnits = game
                .getSelectedEntities(new EntitySelector() {
                    public boolean accept(Entity entity) {
                        if (game.isEntityStranded(entity)) {
                            return true;
                        }
                        return false;
                    }
                });

        Vector<GameTurn> turns;

        if (strandedUnits.hasNext()
                && (game.getPhase() == IGame.Phase.PHASE_MOVEMENT)) {
           
            turns = new Vector<GameTurn>(team_order.getTotalTurns()
                    + team_order.getEvenTurns() + 1);
            turns.addElement(new GameTurn.UnloadStrandedTurn(strandedUnits));
        } else {
            
            turns = new Vector<GameTurn>(team_order.getTotalTurns()
                    + team_order.getEvenTurns());
        }

        while (team_order.hasMoreElements()) {
            Entity e = (Entity) team_order.nextElement();
            if (e.isSelectableThisTurn()) {
                if (!protosMoveMulti && (e instanceof Protomech)
                        && (e.getUnitNumber() != Entity.NONE)) {
                    turns.addElement(new GameTurn.UnitNumberTurn(
                            e.getOwnerId(), e.getUnitNumber()));
                } else {
                    turns.addElement(new GameTurn.SpecificEntityTurn(e
                            .getOwnerId(), e.getId()));
                }

            }
        }

        game.setTurnVector(turns);
        game.resetTurnIndex();

        send(createTurnVectorPacket());
    }

    private void determineTurnOrder(IGame.Phase phase) {

        if (game.getOptions().booleanOption("individual_initiative")) {
            determineTurnOrderIUI(phase);
            return;
        }
        
        boolean infMoveEven = (game.getOptions().booleanOption("inf_move_even") && ((game
                .getPhase() == IGame.Phase.PHASE_INITIATIVE) || (game
                .getPhase() == IGame.Phase.PHASE_MOVEMENT)))
                || (game.getOptions().booleanOption("inf_deploy_even") && (game
                        .getPhase() == IGame.Phase.PHASE_DEPLOYMENT));
        boolean infMoveMulti = game.getOptions()
                .booleanOption("inf_move_multi")
                && ((game.getPhase() == IGame.Phase.PHASE_INITIATIVE) || ((game
                        .getPhase() == IGame.Phase.PHASE_MOVEMENT) || (game
                        .getPhase() == IGame.Phase.PHASE_DEPLOYMENT)));
        boolean protosMoveEven = (game.getOptions().booleanOption(
                "protos_move_even") && ((game.getPhase() == IGame.Phase.PHASE_INITIATIVE) || ((game
                .getPhase() == IGame.Phase.PHASE_MOVEMENT) || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))))
                || (game.getOptions().booleanOption("protos_deploy_even") && (game
                        .getPhase() == IGame.Phase.PHASE_DEPLOYMENT));
        boolean protosMoveMulti = game.getOptions().booleanOption(
                "protos_move_multi");
        boolean protosMoveByPoint = !protosMoveMulti;
        boolean tankMoveByLance = game.getOptions().booleanOption(
                "vehicle_lance_movement")
                && ((game.getPhase() == IGame.Phase.PHASE_INITIATIVE) || ((game
                        .getPhase() == IGame.Phase.PHASE_MOVEMENT) || (game
                        .getPhase() == IGame.Phase.PHASE_DEPLOYMENT)));
        boolean mekMoveByLance = game.getOptions().booleanOption(
                "mek_lance_movement")
                && ((game.getPhase() == IGame.Phase.PHASE_INITIATIVE) || ((game
                        .getPhase() == IGame.Phase.PHASE_MOVEMENT) || (game
                        .getPhase() == IGame.Phase.PHASE_DEPLOYMENT)));

        int evenMask = 0;
        if (infMoveEven) {
            evenMask += GameTurn.CLASS_INFANTRY;
        }
        if (protosMoveEven) {
            evenMask += GameTurn.CLASS_PROTOMECH;
        }
        
        for (Enumeration<IPlayer> loop = game.getPlayers(); loop
                .hasMoreElements(); ) {
            final IPlayer player = loop.nextElement();
            player.resetEvenTurns();
            player.resetMultiTurns();
            player.resetOtherTurns();
            player.resetSpaceStationTurns();
            player.resetJumpshipTurns();
            player.resetWarshipTurns();
            player.resetDropshipTurns();
            player.resetSmallCraftTurns();
            player.resetAeroTurns();
            
            if (protosMoveByPoint) {
                
                Iterator<Entity> playerProtos = game
                        .getSelectedEntities(new EntitySelector() {
                            private final int ownerId = player.getId();

                            public boolean accept(Entity entity) {
                                if ((entity instanceof Protomech)
                                    && (ownerId == entity.getOwnerId())
                                    && entity.isSelectableThisTurn()) {
                                    return true;
                                }
                                return false;
                            }
                        });
                HashSet<Integer> points = new HashSet<Integer>();
                int numPlayerProtos = 0;
                for (; playerProtos.hasNext(); ) {
                    Entity proto = playerProtos.next();
                    numPlayerProtos++;
                    points.add(new Integer(proto.getUnitNumber()));
                }
                int numProtoUnits = (int) Math.ceil(numPlayerProtos / 5.0);
                if (!protosMoveEven) {
                    numProtoUnits = points.size();
                }
                for (int unit = 0; unit < numProtoUnits; unit++) {
                    if (protosMoveEven) {
                        player.incrementEvenTurns();
                    } else {
                        player.incrementOtherTurns();
                    }
                }

            }
        } 

        for (Iterator<Entity> loop = game.getEntities(); loop.hasNext();) {
            final Entity entity = loop.next();
            if (entity.isSelectableThisTurn()) {
                final IPlayer player = entity.getOwner();
                if ((entity instanceof SpaceStation)
                    && ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                        || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))) {
                    player.incrementSpaceStationTurns();
                } else if ((entity instanceof Warship)
                           && ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                               || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))) {
                    player.incrementWarshipTurns();
                } else if ((entity instanceof Jumpship)
                           && ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                               || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))) {
                    player.incrementJumpshipTurns();
                } else if ((entity instanceof Dropship) && entity.isAirborne()
                           && ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                               || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))) {
                    player.incrementDropshipTurns();
                } else if ((entity instanceof SmallCraft)
                           && entity.isAirborne()
                           && ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                               || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))) {
                    player.incrementSmallCraftTurns();
                } else if (entity.isAirborne()
                           && ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                               || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))) {
                    player.incrementAeroTurns();
                } else if ((entity instanceof Infantry)) {
                    if (infMoveEven) {
                        player.incrementEvenTurns();
                    } else if (infMoveMulti) {
                        player.incrementMultiTurns(GameTurn.CLASS_INFANTRY);
                    } else {
                        player.incrementOtherTurns();
                    }
                } else if (entity instanceof Protomech) {
                    if (!protosMoveByPoint) {
                        if (protosMoveEven) {
                            player.incrementEvenTurns();
                        } else if (protosMoveMulti) {
                            player.incrementMultiTurns(GameTurn.CLASS_PROTOMECH);
                        } else {
                            player.incrementOtherTurns();
                        }
                    }
                } else if ((entity instanceof Tank) && tankMoveByLance) {
                    player.incrementMultiTurns(GameTurn.CLASS_TANK);
                } else if ((entity instanceof Mech) && mekMoveByLance) {
                    player.incrementMultiTurns(GameTurn.CLASS_MECH);
                } else {
                    player.incrementOtherTurns();
                }
            }
        }

        int nTeams = game.getNoOfTeams();
        Hashtable<Team, TurnVectors> allTeamTurns = new Hashtable<Team, TurnVectors>(
                nTeams);
        Hashtable<Team, int[]> evenTrackers = new Hashtable<Team, int[]>(nTeams);
        int numTeamsMoving = 0;
        for (Enumeration<Team> loop = game.getTeams(); loop.hasMoreElements(); ) {
            final Team team = loop.nextElement();
            allTeamTurns.put(team, team.determineTeamOrder(game));

            int[] evenTracker = new int[2];
            evenTracker[0] = 0;
            evenTracker[1] = 0;
            evenTrackers.put(team, evenTracker);

            if (team.getNormalTurns(game) > 0) {
                numTeamsMoving++;
            }
        }

        TurnVectors team_order = TurnOrdered.generateTurnOrder(
                game.getTeamsVector(), game);
        
        Iterator<Entity> strandedUnits = game
                .getSelectedEntities(new EntitySelector() {
                    public boolean accept(Entity entity) {
                        if (game.isEntityStranded(entity)) {
                            return true;
                        }
                        return false;
                    }
                });
       
        Vector<GameTurn> turns;

        if (strandedUnits.hasNext()
            && ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT))) {
            
            
            turns = new Vector<GameTurn>(team_order.getTotalTurns()
                                         + team_order.getEvenTurns() + 1);
            turns.addElement(new GameTurn.UnloadStrandedTurn(strandedUnits));
        } else {
            
            turns = new Vector<GameTurn>(team_order.getTotalTurns()
                                         + team_order.getEvenTurns());
        }
  
        Team prevTeam = null;
        int min = team_order.getMin();
        for (int numTurn = 0; team_order.hasMoreElements(); numTurn++) {
            Team team = (Team) team_order.nextElement();
            TurnVectors withinTeamTurns = allTeamTurns.get(team);

            int[] evenTracker = evenTrackers.get(team);
            float teamEvenTurns = team.getEvenTurns();
      
            int numEven = 0;
            if (1 == numTeamsMoving) {
               
                numEven += (teamEvenTurns / min) + 0.5;
            } else if (prevTeam == null) {
                
                evenTracker[0]++;
                
                numEven += teamEvenTurns / min;
            } else if (!team.equals(prevTeam)) {
                
                evenTracker[0]++;
    
                numEven += Math
                                   .ceil(((evenTracker[0] * (teamEvenTurns % min)) / min) - 0.5)
                           - evenTracker[1];

                
                evenTracker[1] += numEven;

                
                numEven += teamEvenTurns / min;
            }
   
            prevTeam = team;

            int aeroMask = GameTurn.CLASS_AERO + GameTurn.CLASS_SMALL_CRAFT
                           + GameTurn.CLASS_DROPSHIP + GameTurn.CLASS_JUMPSHIP
                           + GameTurn.CLASS_WARSHIP + GameTurn.CLASS_SPACE_STATION;
            if (withinTeamTurns.hasMoreNormalElements()) {
   
                IPlayer player = (IPlayer) withinTeamTurns.nextNormalElement();
 
                GameTurn turn = null;
                if (numTurn >= team_order.getTotalTurns()) {
                    turn = new GameTurn.EntityClassTurn(player.getId(),
                                                        evenMask);
                }
                
                else if (infMoveEven || protosMoveEven) {
                    int newMask = evenMask;
                    
                    if ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                        || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT)) {
                        newMask += aeroMask;
                    }
                    turn = new GameTurn.EntityClassTurn(player.getId(),
                                                        ~newMask);
                }
  
                else {
                    
                    
                    if ((game.getPhase() == IGame.Phase.PHASE_MOVEMENT)
                        || (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT)) {
                        turn = new GameTurn.EntityClassTurn(player.getId(),
                                                            ~aeroMask);
                    } else {
                        turn = new GameTurn(player.getId());
                    }
                }
                turns.addElement(turn);

            } 
            else if (withinTeamTurns.hasMoreSpaceStationElements()) {
                IPlayer player = (IPlayer) withinTeamTurns
                        .nextSpaceStationElement();
                GameTurn turn = null;
                turn = new GameTurn.EntityClassTurn(player.getId(),
                                                    GameTurn.CLASS_SPACE_STATION);
                turns.addElement(turn);
            } else if (withinTeamTurns.hasMoreJumpshipElements()) {
                IPlayer player = (IPlayer) withinTeamTurns
                        .nextJumpshipElement();
                GameTurn turn = null;
                turn = new GameTurn.EntityClassTurn(player.getId(),
                                                    GameTurn.CLASS_JUMPSHIP);
                turns.addElement(turn);
            } else if (withinTeamTurns.hasMoreWarshipElements()) {
                IPlayer player = (IPlayer) withinTeamTurns.nextWarshipElement();
                GameTurn turn = null;
                turn = new GameTurn.EntityClassTurn(player.getId(),
                                                    GameTurn.CLASS_WARSHIP);
                turns.addElement(turn);
            } else if (withinTeamTurns.hasMoreDropshipElements()) {
                IPlayer player = (IPlayer) withinTeamTurns
                        .nextDropshipElement();
                GameTurn turn = null;
                turn = new GameTurn.EntityClassTurn(player.getId(),
                                                    GameTurn.CLASS_DROPSHIP);
                turns.addElement(turn);
            } else if (withinTeamTurns.hasMoreSmallCraftElements()) {
                IPlayer player = (IPlayer) withinTeamTurns
                        .nextSmallCraftElement();
                GameTurn turn = null;
                turn = new GameTurn.EntityClassTurn(player.getId(),
                                                    GameTurn.CLASS_SMALL_CRAFT);
                turns.addElement(turn);
            } else if (withinTeamTurns.hasMoreAeroElements()) {
                IPlayer player = (IPlayer) withinTeamTurns.nextAeroElement();
                GameTurn turn = null;
                turn = new GameTurn.EntityClassTurn(player.getId(),
                                                    GameTurn.CLASS_AERO);
                turns.addElement(turn);
            }
    
            while ((numEven > 0) && withinTeamTurns.hasMoreEvenElements()) {
                IPlayer evenPlayer = (IPlayer) withinTeamTurns
                        .nextEvenElement();
                turns.addElement(new GameTurn.EntityClassTurn(evenPlayer
                                                                      .getId(), evenMask));
                numEven--;
            }
        }

        game.setTurnVector(turns);
        game.resetTurnIndex();

        send(createTurnVectorPacket());

    }

    private static String getColorForPlayer(IPlayer p) {
        String colorcode = Integer.toHexString(PlayerColors.getColor(
                p.getColorIndex()).getRGB() & 0x00f0f0f0);
        return "<B><font color='" + colorcode + "'>" + p.getName()
               + "</font></B>";

    }

    private void writeInitiativeReport(boolean abbreviatedReport) {
        
        Report r;
        boolean deployment = false;
        if (!abbreviatedReport) {
            r = new Report(1210);
            r.type = Report.PUBLIC;
            if ((game.getLastPhase() == IGame.Phase.PHASE_DEPLOYMENT)
                || game.isDeploymentComplete()
                || !game.shouldDeployThisRound()) {
                r.messageId = 1000;
                r.add(game.getRoundCount());
            } else {
                deployment = true;
                if (game.getRoundCount() == 0) {
                    r.messageId = 1005;
                } else {
                    r.messageId = 1010;
                    r.add(game.getRoundCount());
                }
            }
            addReport(r);
            
            addReport(new Report(1200, Report.PUBLIC));
        } else {
            addReport(new Report(1210, Report.PUBLIC));
        }

        if (game.getOptions().booleanOption("individual_initiative")) {
            r = new Report(1040, Report.PUBLIC);
            addReport(r);
            for (Enumeration<GameTurn> e = game.getTurns(); e.hasMoreElements(); ) {
                GameTurn t = e.nextElement();
                if (t instanceof GameTurn.SpecificEntityTurn) {
                    Entity entity = game
                            .getEntity(((GameTurn.SpecificEntityTurn) t)
                                               .getEntityNum());
                    r = new Report(1045);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(entity.getInitiative().toString());
                    addReport(r);
                } else {
                    IPlayer player = getPlayer(t.getPlayerNum());
                    if (null != player) {
                        r = new Report(1050, Report.PUBLIC);
                        r.add(Server.getColorForPlayer(player));
                        addReport(r);
                    }
                }
            }
        } else {
            for (Enumeration<Team> i = game.getTeams(); i.hasMoreElements(); ) {
                final Team team = i.nextElement();

                if (team.getSize() == 1) {
                    final IPlayer player = team.getPlayers().nextElement();
                    r = new Report(1015, Report.PUBLIC);
                    r.add(Server.getColorForPlayer(player));
                    r.add(team.getInitiative().toString());
                    addReport(r);
                } else {
                    
                    r = new Report(1015, Report.PUBLIC);
                    r.add(IPlayer.teamNames[team.getId()]);
                    r.add(team.getInitiative().toString());
                    addReport(r);
                    for (Enumeration<IPlayer> j = team.getPlayers(); j
                            .hasMoreElements(); ) {
                        final IPlayer player = j.nextElement();
                        r = new Report(1015, Report.PUBLIC);
                        r.indent();
                        r.add(player.getName());
                        r.add(player.getInitiative().toString());
                        addReport(r);
                    }
                }
            }

            if (!doBlind()) {

                r = new Report(1020, Report.PUBLIC);

                boolean hasEven = false;
                for (Enumeration<GameTurn> i = game.getTurns(); i
                        .hasMoreElements(); ) {
                    GameTurn turn = i.nextElement();
                    IPlayer player = getPlayer(turn.getPlayerNum());
                    if (null != player) {
                        r.add(player.getName());
                        if (player.getEvenTurns() > 0) {
                            hasEven = true;
                        }
                    }
                }
                r.newlines = 2;
                addReport(r);
                if (hasEven) {
                    r = new Report(1021, Report.PUBLIC);
                    if ((game.getOptions().booleanOption("inf_deploy_even") || game
                            .getOptions().booleanOption("protos_deploy_even"))
                        && !(game.getLastPhase() == IGame.Phase.PHASE_END_REPORT)) {
                        r.choose(true);
                    } else {
                        r.choose(false);
                    }
                    r.indent();
                    r.newlines = 2;
                    addReport(r);
                }
            }

        }
        if (!abbreviatedReport) {
            
            Report rWindDir = new Report(1025, Report.PUBLIC);
            rWindDir.add(game.getPlanetaryConditions().getWindDirDisplayableName());
            rWindDir.newlines = 0;
            Report rWindStr = new Report(1030, Report.PUBLIC);
            rWindStr.add(game.getPlanetaryConditions().getWindDisplayableName());
            rWindStr.newlines = 0;
            Report rWeather = new Report(1031, Report.PUBLIC);
            rWeather.add(game.getPlanetaryConditions().getWeatherDisplayableName());
            rWeather.newlines = 0;
            Report rLight = new Report(1032, Report.PUBLIC);
            rLight.add(game.getPlanetaryConditions().getLightDisplayableName());
            Report rVis = new Report(1033, Report.PUBLIC);
            rVis.add(game.getPlanetaryConditions().getFogDisplayableName());
            addReport(rWindDir);
            addReport(rWindStr);
            addReport(rWeather);
            addReport(rLight);
            addReport(rVis);

            if (deployment) {
                addNewLines();
            }
        }
    }

    private void applyDropshipLandingDamage(Coords centralPos, Entity killer) {

        
        IHex centralHex = game.getBoard().getHex(centralPos);
        if (null == centralHex) {
            
            return;
        }
        int finalElev = centralHex.getLevel();
        if (!centralHex.containsTerrain(Terrains.PAVEMENT)
            && !centralHex.containsTerrain(Terrains.ROAD)) {
            finalElev--;
        }
        Vector<Coords> positions = new Vector<Coords>();
        positions.add(centralPos);
        for (int i = 0; i < 6; i++) {
            Coords pos = centralPos.translated(i);
            IHex hex = game.getBoard().getHex(pos);
            if (null == hex) {
                continue;
            }
            if (hex.getLevel() < finalElev) {
                finalElev = hex.getLevel();
            }
            positions.add(pos);
        }
        
        for (Coords pos : positions) {
            IHex hex = game.getBoard().getHex(pos);
            hex.setLevel(finalElev);
            
            if (hex.containsTerrain(Terrains.WOODS)
                || hex.containsTerrain(Terrains.JUNGLE)) {
                hex.removeTerrain(Terrains.WOODS);
                hex.removeTerrain(Terrains.JUNGLE);
                hex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.ROUGH, 1));
            }
            sendChangedHex(pos);
        }

        applyDropshipProximityDamage(centralPos, killer);
    }

    private void applyDropshipProximityDamage(Coords centralPos, Entity killer) {
        applyDropshipProximityDamage(centralPos, false, 0, killer);
    }

    
    private void applyDropshipProximityDamage(Coords centralPos,
                                              boolean rearArc, int facing, Entity killer) {

        Vector<Integer> alreadyHit = new Vector<Integer>();

        
        Hashtable<Coords, Vector<Entity>> positionMap = game.getPositionMap();
        for (Entity en : game.getEntitiesVector(centralPos)) {
            if (!en.isAirborne()) {
                addReport(destroyEntity(en, "dropship proximity damage", false,
                                        false));
                alreadyHit.add(en.getId());
            }
        }
        Building bldg = game.getBoard().getBuildingAt(centralPos);
        if (null != bldg) {
            collapseBuilding(bldg, positionMap, centralPos, vPhaseReport);
        }
        for (int i = 0; i < 6; i++) {
            Coords pos = centralPos.translated(i);
            for (Entity en : game.getEntitiesVector(pos)) {
                if (!en.isAirborne()) {
                    addReport(destroyEntity(en, "dropship proximity damage",
                                            false, false));
                }
                alreadyHit.add(en.getId());
            }
            bldg = game.getBoard().getBuildingAt(pos);
            if (null != bldg) {
                collapseBuilding(bldg, positionMap, pos, vPhaseReport);
            }
        }

        for (int i = 2; i < 8; i++) {
            int damageDice = (8 - i) * 2;
            ArrayList<Coords> ring = Compute.coordsAtRange(centralPos, i);
            for (Coords pos : ring) {
                if (rearArc
                    && !Compute.isInArc(centralPos, facing, pos,
                                        Compute.ARC_AFT)) {
                    continue;
                }

                alreadyHit = artilleryDamageHex(pos, centralPos, damageDice,
                                                null, killer.getId(), killer, null, false, 0,
                                                vPhaseReport, false, alreadyHit, true);
     
            }
        }
        destroyDoomedEntities(alreadyHit);
    }

    private void setIneligible(IGame.Phase phase) {
        Vector<Entity> assistants = new Vector<Entity>();
        boolean assistable = false;

        if (isPlayerForcedVictory()) {
            assistants.addAll(game.getEntitiesVector());
        } else {
            for (Entity entity : game.getEntitiesVector()) {
                if (entity.isEligibleFor(phase)) {
                    assistable = true;
                } else {
                    assistants.addElement(entity);
                }
            }
        }
        for (Entity assistant : assistants) {
            if (!assistable || !assistant.canAssist(phase)) {
                assistant.setDone(true);
            }
        }
    }

    private void loadUnit(Entity loader, Entity unit, int bayNumber) {

        if ((game.getPhase() != IGame.Phase.PHASE_LOUNGE) && !unit.isDone()) {
           
            game.removeTurnFor(unit);
            send(createTurnVectorPacket());
        }
        if ((game.getPhase() == Phase.PHASE_LOUNGE)
            && (loader instanceof FighterSquadron)) {
            ((Aero) unit).setBombChoices(((FighterSquadron) loader)
                                                 .getBombChoices());
        }

        boolean checkElevation = (game.getPhase() != Phase.PHASE_DEPLOYMENT)
                                 && (game.getPhase() != Phase.PHASE_LOUNGE);
        if (bayNumber == -1) {
            loader.load(unit, checkElevation);
        } else {
            loader.load(unit, checkElevation, bayNumber);
        }

        unit.setTransportId(loader.getId());

        unit.setPosition(null);

        unit.setDeployRound(loader.getDeployRound());

        entityUpdate(unit.getId());
 
        entityUpdate(loader.getId());
    }

    private boolean unloadUnit(Entity unloader, Targetable unloaded,
                               Coords pos, int facing, int elevation) {
        return unloadUnit(unloader, unloaded, pos, facing, elevation, false);
    }

    private boolean unloadUnit(Entity unloader, Targetable unloaded,
            Coords pos, int facing, int elevation, boolean evacuation) {

        Entity unit = null;
        if (unloaded instanceof Entity) {
            unit = (Entity) unloaded;
        } else {
            return false;
        }

        if (!unloader.unload(unit)) {
            return false;
        }
 
        unit.setTransportId(Entity.NONE);

        unit.setPosition(pos);

        if (pos != null) {
            unit.setDeployed(true);
        }
  
        unit.setFacing(facing);
        unit.setSecondaryFacing(facing);

        IHex hex = game.getBoard().getHex(pos);
        boolean isBridge = hex.containsTerrain(Terrains.PAVEMENT);

        if (unloader.getMovementMode() == EntityMovementMode.VTOL) {
            if (unit.getMovementMode() == EntityMovementMode.VTOL) {
                
                
                unit.setElevation(elevation);
            } else if (game.getBoard().getBuildingAt(pos) != null) {
                
                
                unit.setElevation(hex.terrainLevel(Terrains.BLDG_ELEV));
            } else {
                while (elevation >= -hex.depth()) {
                    if (unit.isElevationValid(elevation, hex)) {
                        unit.setElevation(elevation);
                        break;
                    }
                    elevation--;
                 
                    if (unit.getJumpMP() > 0) {
                        unit.moved = EntityMovementType.MOVE_JUMP;
                    } else { 
                        unit.moved = EntityMovementType.MOVE_WALK;
                    }
                }
                if (!unit.isElevationValid(elevation, hex)) {
                    return false;
                }
            }
        } else if (game.getBoard().getBuildingAt(pos) != null) {
            
            
            unit.setElevation(elevation);
        } else if (hex.terrainLevel(Terrains.WATER) > 0) {
            if ((unit.getMovementMode() == EntityMovementMode.HOVER)
                || (unit.getMovementMode() == EntityMovementMode.WIGE)
                || (unit.getMovementMode() == EntityMovementMode.HYDROFOIL)
                || (unit.getMovementMode() == EntityMovementMode.NAVAL)
                || (unit.getMovementMode() == EntityMovementMode.SUBMARINE)
                || (unit.getMovementMode() == EntityMovementMode.INF_UMU)
                || hex.containsTerrain(Terrains.ICE) || isBridge) {
                
                unit.setElevation(0);
            }
        } else {
           
            unit.setElevation(hex.floor() - hex.surface());
        }

        
        if (unit.moved == EntityMovementType.MOVE_WALK) {
            if (game.getOptions().booleanOption("tacops_ziplines")
                    && (unit instanceof Infantry)
                    && !((Infantry) unit).isMechanized()) {

                PilotingRollData psr = getEjectModifiers(game, unit, false,
                        unit.getPosition(), "Anti-mek skill");
                int roll = Compute.d6(2);

                Report r = new Report(9920);
                r.subject = unit.getId();
                r.addDesc(unit);
                r.newlines = 0;
                addReport(r);

                r = new Report(9921);
                r.subject = unit.getId();
                r.add(psr.getValue());
                r.add(psr.getDesc());
                r.add(roll);
                r.newlines = 0;
                addReport(r);

                if (roll < psr.getValue()) { 
                    r = new Report(9923);
                    r.subject = unit.getId();
                    r.add(psr.getValue());
                    r.add(roll);
                    addReport(r);

                    HitData hit = unit.rollHitLocation(ToHitData.HIT_NORMAL,
                            ToHitData.SIDE_FRONT);
                    hit.setIgnoreInfantryDoubleDamage(true);
                    addReport(damageEntity(unit, hit, 5));
                } else { 
                    r = new Report(9922);
                    r.subject = unit.getId();
                    r.add(psr.getValue());
                    r.add(roll);
                    addReport(r);
                }
                addNewLines();
            } else {
                return false;
            }
        }

        addReport(doSetLocationsExposure(unit, hex, false, unit.getElevation()));

        if (!evacuation && (unloader instanceof SmallCraft)
            && !(unit instanceof Infantry)) {
            unit.setUnloaded(false);
            unit.setDone(false);

            unit.mpUsed = unit.getOriginalWalkMP() / 2;
            unit.delta_distance = 1;
        }

        entityUpdate(unit.getId());

        return true;
    }

    private void doAttemptLanding(Entity entity, PilotingRollData roll) {

        if (roll.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            return;
        }

        Report r = new Report(9605);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(roll.getLastPlainDesc(), true);
        addReport(r);

        final int diceRoll = Compute.d6(2);
        r = new Report(9606);
        r.subject = entity.getId();
        r.add(roll.getValueAsString());
        r.add(roll.getDesc());
        r.add(diceRoll);
        
        if (diceRoll < roll.getValue()) {
            r.choose(false);
            addReport(r);
            int mof = roll.getValue() - diceRoll;
            int damage = 10 * (mof);
            
            r = new Report(9609);
            r.indent();
            r.addDesc(entity);
            r.add(damage);
            r.add(mof);
            addReport(r);

            int side = ToHitData.SIDE_FRONT;
            if ((entity instanceof Aero) && ((Aero) entity).isSpheroid()) {
                side = ToHitData.SIDE_REAR;
            }
            while (damage > 0) {
                HitData hit = entity
                        .rollHitLocation(ToHitData.HIT_NORMAL, side);
                addReport(damageEntity(entity, hit, 10));
                damage -= 10;
            }
            
        } else {
            r.choose(true);
            addReport(r);
            
        }
    }

    private boolean launchUnit(Entity unloader, Targetable unloaded,
                               Coords pos, int facing, int velocity, int altitude, int[] moveVec,
                               int bonus) {

        Entity unit = null;
        if (unloaded instanceof Entity) {
            unit = (Entity) unloaded;
        } else {
            return false;
        }

        Aero a = new Aero();
        if ((unit instanceof Aero) && !(unit instanceof Jumpship)) {
            a = (Aero) unit;
        } else {
            return false;
        }

        Report r;

        if (!unloader.unload(unit)) {
            return false;
        }

        unit.setTransportId(Entity.NONE);

        unit.setUnloaded(false);

        unit.setPosition(pos);
        
        if (pos != null) {
            unit.setDeployed(true);
        }

        unit.setFacing(facing);
        unit.setSecondaryFacing(facing);

        a.setCurrentVelocity(velocity);
        a.setNextVelocity(velocity);

        a.setVectors(moveVec);

        unit.setAltitude(altitude);

        
        unit.setDone(false);

        if (bonus > 0) {
            PilotingRollData psr = unit.getBasePilotingRoll();
            psr.addModifier(bonus, "safe launch rate exceeded");
            int ctrlroll = Compute.d6(2);
            r = new Report(9375);
            r.subject = unit.getId();
            r.add(unit.getDisplayName());
            r.add(psr.getValue());
            r.add(ctrlroll);
            r.indent(1);
            if (ctrlroll < psr.getValue()) {
                r.choose(false);
                addReport(r);
                
                int damage = 10 * (psr.getValue() - ctrlroll);
                HitData hit = a.rollHitLocation(ToHitData.HIT_NORMAL,
                                                ToHitData.SIDE_FRONT);
                Vector<Report> rep = damageEntity(unit, hit, damage);
                Report.indentAll(rep, 1);
                rep.lastElement().newlines++;
                addReport(rep);
                
                if (unit.isDoomed()) {
                    
                    unit.setDestroyed(true);
                    game.moveToGraveyard(unit.getId());
                    send(createRemoveEntityPacket(unit.getId()));
                }
            } else {
                
                r.choose(true);
                r.newlines++;
                addReport(r);
            }
        } else {
            r = new Report(9374);
            r.subject = unit.getId();
            r.add(unit.getDisplayName());
            r.indent(1);
            r.newlines++;
            addReport(r);
        }

        if ((((Aero) unloader).isOutControlTotal() && !unit.isDoomed())
            || ((((Aero) unloader).getCurrentVelocity() > 2) && !game
                .getBoard().inSpace())) {
            int damroll = Compute.d6(2);
            int damage = damroll * 10;
            r = new Report(9385);
            r.subject = unit.getId();
            r.add(unit.getDisplayName());
            r.add(damage);
            addReport(r);
            HitData hit = a.rollHitLocation(ToHitData.HIT_NORMAL,
                                            ToHitData.SIDE_FRONT);
            addReport(damageEntity(unit, hit, damage));
            
            if (unit.isDoomed()) {
                
                unit.setDestroyed(true);
                game.moveToGraveyard(unit.getId());
                send(createRemoveEntityPacket(unit.getId()));
            }
        }

        
        entityUpdate(unit.getId());

        int turnMask;
        List<GameTurn> turnVector = game.getTurnVector();
        if (unit instanceof Dropship) {
            turnMask = GameTurn.CLASS_DROPSHIP;
        } else if (unit instanceof SmallCraft) {
            turnMask = GameTurn.CLASS_SMALL_CRAFT;
        } else {
            turnMask = GameTurn.CLASS_AERO;
        }
        
        int turnInsertIdx = game.getTurnIndex() + 1;
        
        for (; turnInsertIdx < turnVector.size(); turnInsertIdx++) {
            GameTurn turn = turnVector.get(turnInsertIdx);
            if (turn.isValidEntity(unit, game)) {
                break;
            }
        }

        GameTurn newTurn = new GameTurn.EntityClassTurn(
                unit.getOwner().getId(), turnMask);
        game.insertTurnAfter(newTurn, turnInsertIdx);
        
        send(createTurnVectorPacket());

        return true;
    }

    public void dropUnit(Entity drop, Entity entity, Coords curPos, int altitude) {
        
        entity.unload(drop);
        
        drop.setTransportId(Entity.NONE);
        
        if (game.getBoard().onGround() && (null != curPos)) {
            boolean selected = false;
            int count = 0;
            int max = 0;
            int facing = entity.getFacing();
            if (entity.getMovementMode() == EntityMovementMode.AERODYNE) {
                
                facing = (facing + 3) % 6;
            }
            boolean checkDanger = true;
            while (!selected) {
                
                boolean allDanger = true;
                for (int i = 0; i < 6; i++) {
                    int dir = (facing + i) % 6;
                    Coords newPos = curPos.translated(dir, 2);
                    count = 0;
                    if (game.getBoard().contains(newPos)) {
                        IHex newHex = game.getBoard().getHex(newPos);
                        Building bldg = game.getBoard().getBuildingAt(newPos);
                        boolean danger = newHex.containsTerrain(Terrains.WATER)
                                         || newHex.containsTerrain(Terrains.MAGMA)
                                         || (null != bldg);
                        for (Entity unit : game.getEntitiesVector(newPos)) {
                            if ((unit.getAltitude() == altitude)
                                && !(unit instanceof Aero)) {
                                count++;
                            }
                        }
                        if ((count <= max) && (!danger || !checkDanger)) {
                            selected = true;
                            curPos = newPos;
                            break;
                        }
                        if (!danger) {
                            allDanger = false;
                        }
                    }
                    newPos = newPos.translated((dir + 2) % 6);
                    count = 0;
                    if (game.getBoard().contains(newPos)) {
                        IHex newHex = game.getBoard().getHex(newPos);
                        Building bldg = game.getBoard().getBuildingAt(newPos);
                        boolean danger = newHex.containsTerrain(Terrains.WATER)
                                         || newHex.containsTerrain(Terrains.MAGMA)
                                         || (null != bldg);
                        for (Entity unit : game.getEntitiesVector(newPos)) {
                            if ((unit.getAltitude() == altitude)
                                && !(unit instanceof Aero)) {
                                count++;
                            }
                        }
                        if ((count <= max) && (!danger || !checkDanger)) {
                            selected = true;
                            curPos = newPos;
                            break;
                        }
                        if (!danger) {
                            allDanger = false;
                        }
                    }
                }
                if (allDanger && checkDanger) {
                    checkDanger = false;
                } else {
                    max++;
                }
            }
        }

        drop.setPosition(curPos);

        
        if (curPos != null) {
            drop.setDeployed(true);
        }

        drop.setFacing(entity.getFacing());
        drop.setSecondaryFacing(entity.getFacing());

        drop.setAltitude(altitude);
        entityUpdate(drop.getId());
    }

    private void addAffectedBldg(Building bldg, boolean collapse) {

        if (collapse) {
            affectedBldgs.remove(bldg);
        } else { 
            affectedBldgs.put(bldg, Boolean.FALSE);
        }
    }

    private void applyAffectedBldgs() {

        Vector<Building> bldgUpdates = new Vector<Building>();

        boolean bTurnsChanged = false;

        Enumeration<Building> bldgs = affectedBldgs.keys();
        while (bldgs.hasMoreElements()) {
            final Building bldg = bldgs.nextElement();

            Enumeration<Coords> bldgCoords = bldg.getCoords();
            while (bldgCoords.hasMoreElements()) {
                final Coords coords = bldgCoords.nextElement();

                for (Entity entity : game.getEntitiesVector(coords)) {

                    if (entity instanceof Infantry) {

                        if (entity.isDoomed() || entity.isDestroyed()) {
 
                            if (!entity.isDone()) {
 
                                game.removeTurnFor(entity);
                                bTurnsChanged = true;
                            } 

                            entity.setDestroyed(true);
                            game.moveToGraveyard(entity.getId());
                            send(createRemoveEntityPacket(entity.getId()));
                        }

                        else {
                            entityUpdate(entity.getId());
                        }
                    } 
                } 
            } 

            bldgUpdates.addElement(bldg);
        } 

        if (bTurnsChanged) {
            send(createTurnVectorPacket());
        }

        if (!bldgUpdates.isEmpty()) {

            sendChangedBuildings(bldgUpdates);

            affectedBldgs.clear();
        }
    } 

    private void receiveMovement(Packet packet, int connId) {
        Map<EntityTargetPair, LosEffects> losCache = new HashMap<>();
        Entity entity = game.getEntity(packet.getIntValue(0));
        MovePath md = (MovePath) packet.getObject(1);

        if (game.getPhase() != IGame.Phase.PHASE_MOVEMENT) {
            System.err
                    .println("error: server got movement packet in wrong phase");
            return;
        }

        GameTurn turn = game.getTurn();
        if (game.isPhaseSimultaneous()) {
            turn = game.getTurnForPlayer(connId);
        }
        if ((turn == null) || !turn.isValid(connId, entity, game)) {
            System.err.println("error: server got invalid movement packet");
            return;
        }

        processMovement(entity, md, losCache);

        if ((entity.getGrappled() != Entity.NONE)
                && entity.isChainWhipGrappled() && entity.isGrappleAttacker()
                && (md.getMpUsed() > 0)) {

            Entity te = game.getEntity(entity.getGrappled());
            Report r = new Report(4316);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.addDesc(te);
            addReport(r);

            entity.setGrappled(Entity.NONE, false);
            te.setGrappled(Entity.NONE, false);

            entityUpdate(entity.getId());
            entityUpdate(te.getId());
        }

        for (int missileId : entity.getTMTracker().getMissiles()) {
            Entity tm = game.getEntity(missileId);
            if ((null != tm) && !tm.isDestroyed()
                && (tm instanceof TeleMissile)) {
                if (LosEffects.calculateLos(game, entity.getId(), tm).canSee()) {
                    ((TeleMissile) tm).setOutContact(false);
                } else {
                    ((TeleMissile) tm).setOutContact(true);
                }
                entityUpdate(tm.getId());
            }
        }

        applyAffectedBldgs();

        
        if (doBlind()) {
            updateVisibilityIndicator(losCache);
        }

        endCurrentTurn(entity);
    }

    private boolean processSkid(Entity entity, Coords start, int elevation,
                                int direction, int distance, MoveStep step) {
        Coords nextPos = start;
        Coords curPos = nextPos;
        IHex curHex = game.getBoard().getHex(start);
        Report r;
        int skidDistance = 0; 
        while (!entity.isDoomed() && (distance > 0)) {
            nextPos = curPos.translated(direction);
            
            if (!game.getBoard().contains(nextPos)) {

                
                if (game.getOptions().booleanOption("push_off_board")) {
                    
                    game.removeEntity(entity.getId(),
                                      IEntityRemovalConditions.REMOVE_PUSHED);
                    send(createRemoveEntityPacket(entity.getId(),
                                                  IEntityRemovalConditions.REMOVE_PUSHED));
                    r = new Report(2030, Report.PUBLIC);
                    r.addDesc(entity);
                    addReport(r);

                    for (Entity e : entity.getLoadedUnits()) {
                        game.removeEntity(e.getId(),
                                          IEntityRemovalConditions.REMOVE_PUSHED);
                        send(createRemoveEntityPacket(e.getId(),
                                                      IEntityRemovalConditions.REMOVE_PUSHED));
                    }
                    Entity swarmer = game
                            .getEntity(entity.getSwarmAttackerId());
                    if (swarmer != null) {
                        if (!swarmer.isDone()) {
                            game.removeTurnFor(swarmer);
                            swarmer.setDone(true);
                            send(createTurnVectorPacket());
                        }
                        game.removeEntity(swarmer.getId(),
                                          IEntityRemovalConditions.REMOVE_PUSHED);
                        send(createRemoveEntityPacket(swarmer.getId(),
                                                      IEntityRemovalConditions.REMOVE_PUSHED));
                    }
                    
                    return true;

                }
                
                r = new Report(2035);
                r.subject = entity.getId();
                r.indent();
                addReport(r);
                
                break;
            }

            IHex nextHex = game.getBoard().getHex(nextPos);
            distance -= nextHex.movementCost(entity) + 1;
            
            
            int curAltitude = elevation + curHex.getLevel();
            int nextAltitude = nextHex.floor();

            
            if (entity.getMovementMode() == EntityMovementMode.VTOL) {
                nextAltitude = Math.max(nextAltitude, curAltitude);
            } else {
                
                if (nextHex.containsTerrain(Terrains.BLDG_ELEV)) {
               
                    if (curAltitude >= nextHex.floor()) {
                        nextAltitude = Math
                                .min(curAltitude,
                                     nextHex.getLevel()
                                     + nextHex
                                             .terrainLevel(Terrains.BLDG_ELEV));
                    }
                }
                
                if (nextHex.containsTerrain(Terrains.BRIDGE)) {
                    
                    
                    
                    int exitDir = (direction + 3) % 6;
                    exitDir = 1 << exitDir;
                    if ((nextHex.getTerrain(Terrains.BRIDGE).getExits() & exitDir) == exitDir) {
                        nextAltitude = Math
                                .min(curAltitude,
                                     Math.max(
                                             nextAltitude,
                                             nextHex.getLevel()
                                             + nextHex
                                                     .terrainLevel(Terrains.BRIDGE_ELEV)));
                    }
                }
                if ((nextAltitude <= nextHex.surface())
                    && (curAltitude >= curHex.surface())) {
                    
                    
                    if ((entity instanceof Tank)
                        && ((entity.getMovementMode() == EntityMovementMode.HOVER) || (entity
                                                                                               .getMovementMode() ==
                                                                                       EntityMovementMode.WIGE))) {
                        if (nextHex.containsTerrain(Terrains.WATER)) {
                            nextAltitude = nextHex.surface();
                        }
                    } else {
                        if (nextHex.containsTerrain(Terrains.ICE)) {
                            nextAltitude = nextHex.surface();
                        }
                    }
                }
            }

            int nextElevation = nextAltitude - nextHex.surface();

            boolean crashedIntoTerrain = curAltitude < nextAltitude;
            if (entity.getMovementMode() == EntityMovementMode.VTOL) {
                if ((nextElevation == 0)
                    || ((nextElevation == 1) && (nextHex
                                                         .containsTerrain(Terrains.WOODS) || nextHex
                                                         .containsTerrain(Terrains.JUNGLE)))) {
                    crashedIntoTerrain = true;
                }
            }

            if (nextHex.containsTerrain(Terrains.BLDG_ELEV)) {
                Building bldg = game.getBoard().getBuildingAt(nextPos);

                if (bldg.getType() == Building.WALL) {
                    crashedIntoTerrain = true;
                }

                if (bldg.getBldgClass() == Building.GUN_EMPLACEMENT) {
                    crashedIntoTerrain = true;
                }
            }

            if (entity.getMovementMode() == EntityMovementMode.WIGE) {
                if ((nextElevation == 0)
                    && !(nextHex.containsTerrain(Terrains.WOODS) || nextHex
                        .containsTerrain(Terrains.JUNGLE))) {
                    nextElevation = 1;
                    crashedIntoTerrain = false;
                } else if ((nextElevation == 1)
                           && (nextHex.containsTerrain(Terrains.WOODS) || nextHex
                        .containsTerrain(Terrains.JUNGLE))) {
                    nextElevation = 2;
                    crashedIntoTerrain = false;
                }
            }

            Entity crashDropship = null;
            for (Entity en : game.getEntitiesVector(nextPos)) {
                if ((en instanceof Dropship) && !en.isAirborne()
                    && (nextAltitude <= (en.relHeight()))) {
                    crashDropship = en;
                }
            }

            if (crashedIntoTerrain) {

                if (nextHex.containsTerrain(Terrains.BLDG_ELEV)) {
                    Building bldg = game.getBoard().getBuildingAt(nextPos);

                    if (bldg.getType() == Building.WALL) {
                        r = new Report(2047);
                    } else if (bldg.getBldgClass() == Building.GUN_EMPLACEMENT) {
                        r = new Report(2049);
                    } else {
                        r = new Report(2045);
                    }

                } else {
                    r = new Report(2045);
                }

                r.subject = entity.getId();
                r.indent();
                r.add(nextPos.getBoardNum(), true);
                addReport(r);

                if ((entity.getMovementMode() == EntityMovementMode.WIGE)
                    || (entity.getMovementMode() == EntityMovementMode.VTOL)) {
                    int hitSide = (step.getFacing() - direction) + 6;
                    hitSide %= 6;
                    int table = 0;
                    switch (hitSide) {
                        
                        case 0:
                            table = ToHitData.SIDE_FRONT;
                            break;
                        case 1:
                        case 2:
                            table = ToHitData.SIDE_LEFT;
                            break;
                        case 3:
                            table = ToHitData.SIDE_REAR;
                            break;
                        case 4:
                        case 5:
                            table = ToHitData.SIDE_RIGHT;
                            break;
                    }
                    elevation = nextElevation;
                    addReport(crashVTOLorWiGE((VTOL) entity, false, true,
                                              distance, curPos, elevation, table));

                    if ((nextHex.containsTerrain(Terrains.WATER) && !nextHex
                            .containsTerrain(Terrains.ICE))
                        || nextHex.containsTerrain(Terrains.WOODS)
                        || nextHex.containsTerrain(Terrains.JUNGLE)) {
                        addReport(destroyEntity(entity,
                                                "could not land in crash site"));
                    } else if (elevation < nextHex
                            .terrainLevel(Terrains.BLDG_ELEV)) {
                        Building bldg = game.getBoard().getBuildingAt(nextPos);

                        
                        
                        
                        if (bldg.getType() == Building.WALL) {
                            addReport(destroyEntity(entity,
                                                    "crashed into a wall"));
                            break;
                        }
                        if (bldg.getBldgClass() == Building.GUN_EMPLACEMENT) {
                            addReport(destroyEntity(entity,
                                                    "crashed into a gun emplacement"));
                            break;
                        }

                        addReport(destroyEntity(entity, "crashed into building"));
                    } else {
                        entity.setPosition(nextPos);
                        entity.setElevation(0);
                        addReport(doEntityDisplacementMinefieldCheck(entity,
                                                                     curPos, nextPos, nextElevation));
                    }
                    curPos = nextPos;
                    break;

                }
                
                
                int damage = ((int) entity.getWeight() + 19) / 20;
                while (damage > 0) {
                    int table = ToHitData.HIT_NORMAL;
                    int side = entity.sideTable(nextPos);
                    if (entity instanceof Protomech) {
                        table = ToHitData.HIT_SPECIAL_PROTO;
                    }
                    addReport(damageEntity(entity,
                                           entity.rollHitLocation(table, side),
                                           Math.min(5, damage)));
                    damage -= 5;
                }
                
                break;
            }
            else if (null != crashDropship) {
                r = new Report(2050);
                r.subject = entity.getId();
                r.indent();
                r.add(crashDropship.getShortName(), true);
                r.add(nextPos.getBoardNum(), true);
                addReport(r);
                ChargeAttackAction caa = new ChargeAttackAction(entity.getId(),
                                                                crashDropship.getTargetType(),
                                                                crashDropship.getTargetId(),
                                                                crashDropship.getPosition());
                ToHitData toHit = caa.toHit(game, true);
                resolveChargeDamage(entity, crashDropship, toHit, direction);
                if ((entity.getMovementMode() == EntityMovementMode.WIGE)
                    || (entity.getMovementMode() == EntityMovementMode.VTOL)) {
                    int hitSide = (step.getFacing() - direction) + 6;
                    hitSide %= 6;
                    int table = 0;
                    switch (hitSide) {
                        
                        case 0:
                            table = ToHitData.SIDE_FRONT;
                            break;
                        case 1:
                        case 2:
                            table = ToHitData.SIDE_LEFT;
                            break;
                        case 3:
                            table = ToHitData.SIDE_REAR;
                            break;
                        case 4:
                        case 5:
                            table = ToHitData.SIDE_RIGHT;
                            break;
                    }
                    elevation = nextElevation;
                    addReport(crashVTOLorWiGE((VTOL) entity, false, true,
                                              distance, curPos, elevation, table));
                    break;
                }
                if (!crashDropship.isDoomed() && !crashDropship.isDestroyed()
                    && !game.isOutOfGame(crashDropship)) {
                    break;
                }
            }
 
            else if (curAltitude > (nextAltitude + entity
                    .getMaxElevationChange())) {
                
                if ((entity.getMovementMode() == EntityMovementMode.WIGE)
                    && ((entity.getRunMP() - 2) >= entity.mpUsed)) {
                    entity.mpUsed += 2;
                    nextAltitude = curAltitude;
                } else {
                    addReport(doEntityFallsInto(entity, entity.getElevation(),
                                                curPos, nextPos,
                                                entity.getBasePilotingRoll(step.getMovementType()),
                                                true));
                    addReport(doEntityDisplacementMinefieldCheck(entity,
                                                                 curPos, nextPos, nextElevation));
                    
                    break;
                }
            }
            
            Building bldg = null;
            if (nextElevation < nextHex.terrainLevel(Terrains.BLDG_ELEV)) {
                
                
                bldg = game.getBoard().getBuildingAt(nextPos);
            }
            boolean bldgSuffered = false;
            boolean stopTheSkid = false;
            
            
            Iterator<Entity> targets = game.getEntities(nextPos);
            if (targets.hasNext()) {
                ArrayList<Entity> avoidedChargeUnits = new ArrayList<Entity>();
                boolean skidChargeHit = false;
                while (targets.hasNext()) {
                    Entity target = targets.next();

                    if ((target.getElevation() > (nextElevation + entity
                            .getHeight()))
                        || (target.relHeight() < nextElevation)) {
                        
                        continue;
                    }
  
                    if (!target.isDone()) {
                        if (target instanceof Infantry) {
                            r = new Report(2420);
                            r.subject = target.getId();
                            r.addDesc(target);
                            addReport(r);
                            continue;
                        } else if (target instanceof Protomech) {
                            if (target != Compute.stackingViolation(game,
                                    entity, nextPos, null)) {
                                r = new Report(2420);
                                r.subject = target.getId();
                                r.addDesc(target);
                                addReport(r);
                                continue;
                            }
                        } else {
                            PilotingRollData psr = target.getBasePilotingRoll();
                            psr.addModifier(0, "avoiding collision");
                            int roll = Compute.d6(2);
                            r = new Report(2425);
                            r.subject = target.getId();
                            r.addDesc(target);
                            r.add(psr.getValue());
                            r.add(psr.getDesc());
                            r.add(roll);
                            addReport(r);
                            if (roll >= psr.getValue()) {
                                game.removeTurnFor(target);
                                avoidedChargeUnits.add(target);
                                continue;
                                
                                
                            }
                        }
                    }

                    if ((target instanceof Mech) || (target instanceof Tank)
                        || (target instanceof Aero)) {
                        ChargeAttackAction caa = new ChargeAttackAction(
                                entity.getId(), target.getTargetType(),
                                target.getTargetId(), target.getPosition());
                        ToHitData toHit = caa.toHit(game, true);

                        
                        int roll = Compute.d6(2);
                        
                        r = new Report(2050);
                        r.subject = entity.getId();
                        r.indent();
                        r.add(target.getShortName(), true);
                        r.add(nextPos.getBoardNum(), true);
                        r.newlines = 0;
                        addReport(r);
                        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
                            roll = -12;
                            r = new Report(2055);
                            r.subject = entity.getId();
                            r.add(toHit.getDesc());
                            r.newlines = 0;
                            addReport(r);
                        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
                            r = new Report(2060);
                            r.subject = entity.getId();
                            r.add(toHit.getDesc());
                            r.newlines = 0;
                            addReport(r);
                            roll = Integer.MAX_VALUE;
                        } else {
                            
                            r = new Report(2065);
                            r.subject = entity.getId();
                            r.add(toHit.getValue());
                            r.add(roll);
                            r.newlines = 0;
                            addReport(r);
                        }
  
                        if (roll < toHit.getValue()) {
                            r = new Report(2070);
                            r.subject = entity.getId();
                            addReport(r);
                        } else {
                            
                            resolveChargeDamage(entity, target, toHit,
                                                direction);
                            
                            
                            if (targets.hasNext()) {
                                entity.setPosition(curPos);
                            }
                            bldgSuffered = true;
                            skidChargeHit = true;
                            
                            if (!target.isDoomed() && !target.isDestroyed()
                                && !game.isOutOfGame(target)) {
                                stopTheSkid = true;
                            }
                        }

          
                        if (!target.isDone()) {
                            addReport(resolvePilotingRolls(target));
                            game.resetPSRs(target);
                            target.applyDamage();
                            addNewLines();
                        }

                    }

                    else if ((target instanceof Infantry) && (bldg != null)) {
                        
                        r = new Report(2075);
                        r.subject = entity.getId();
                        r.indent();
                        r.add(target.getShortName(), true);
                        r.add(nextPos.getBoardNum(), true);
                        r.newlines = 0;
                        addReport(r);

                        HitData hit = target.rollHitLocation(
                                ToHitData.HIT_NORMAL,
                                Compute.targetSideTable(entity, target));
                        hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                        
                        
                        addReport(damageEntity(target, hit,
                                               Math.round(entity.getWeight() / 5)));
                        addNewLines();
                    }

                    if (target.isDoomed()) {

                        if (!target.isDone()) {
 
                            game.removeTurnFor(target);
                            send(createTurnVectorPacket());
                        } 

                        target.setDestroyed(true);
                        game.moveToGraveyard(target.getId());
                        send(createRemoveEntityPacket(target.getId()));
                    }

                    
                    if (!game.isOutOfGame(target)) {
                        entityUpdate(target.getId());
                    }

                } 

                if (skidChargeHit) {
                    
                    entity.setPosition(nextPos);
                }

                for (Entity e : avoidedChargeUnits) {
                    GameTurn newTurn = new GameTurn.SpecificEntityTurn(e
                            .getOwner().getId(), e.getId());
                    game.insertNextTurn(newTurn);
                    send(createTurnVectorPacket());
                }
            }
            if (bldg != null) {

                
                r = new Report(2080);
                r.subject = entity.getId();
                r.indent();
                r.add(bldg.getName());
                r.add(nextPos.getBoardNum(), true);
                addReport(r);

                
                int chargeDamage = ChargeAttackAction.getDamageFor(entity, game
                        .getOptions().booleanOption("tacops_charge_damage"),
                        entity.delta_distance);
                if (!bldgSuffered) {
                    Vector<Report> reports = damageBuilding(bldg, chargeDamage,
                                                            nextPos);
                    for (Report report : reports) {
                        report.subject = entity.getId();
                    }
                    addReport(reports);

                    
                    int toAttacker = ChargeAttackAction.getDamageTakenBy(
                            entity, bldg, nextPos);
                    HitData hit = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                         entity.sideTable(nextPos));
                    hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                    addReport(damageEntity(entity, hit, toAttacker));
                    addNewLines();

                    entity.setPosition(nextPos);
                    entity.setElevation(nextElevation);
                    addReport(doEntityDisplacementMinefieldCheck(entity,
                                                                 curPos, nextPos, nextElevation));
                    curPos = nextPos;
                } 
                
                addReport(damageInfantryIn(bldg, chargeDamage, nextPos));

                
                if (bldg.getCurrentCF(nextPos) > 0) {
                    stopTheSkid = true;
                    if (bldg.rollBasement(nextPos, game.getBoard(),
                                          vPhaseReport)) {
                        sendChangedHex(nextPos);
                        Vector<Building> buildings = new Vector<Building>();
                        buildings.add(bldg);
                        sendChangedBuildings(buildings);
                    }
                    addAffectedBldg(
                            bldg,
                            checkBuildingCollapseWhileMoving(bldg, entity,
                                                             nextPos));
                } else {
                    
                    checkForCollapse(bldg, game.getPositionMap(), nextPos,
                                     true, vPhaseReport);
                }
            } 
            
            if (stopTheSkid) {
                break;
            }

            entity.setPosition(nextPos);
            entity.setElevation(nextElevation);
            addReport(doEntityDisplacementMinefieldCheck(entity, curPos,
                                                         nextPos, nextElevation));
            skidDistance++;

            Building roof = game.getBoard().getBuildingAt(nextPos);
            if (roof != null) {
                if (checkForCollapse(roof, game.getPositionMap(), nextPos,
                                     true, vPhaseReport)) {
                    break; 
                }
            }

            
            if ((entity.isLocationProhibited(start) || entity
                    .isLocationProhibited(nextPos))
                && !Compute.canMoveOnPavement(game, curPos, nextPos,
                                              step.getParentUpToThisStep())) {
                
                r = new Report(2040);
                r.subject = entity.getId();
                r.indent();
                r.add(nextPos.getBoardNum(), true);
                addReport(r);

                if ((nextHex.terrainLevel(Terrains.WATER) > 0)
                        && (entity instanceof Tank)
                        && (entity.getMovementMode() != EntityMovementMode.HOVER)
                        && (entity.getMovementMode() != EntityMovementMode.WIGE)) {
                    addReport(destroyEntity(entity,
                            "skidded into a watery grave", false, true));
                }

                int damage = ((int) entity.getWeight() + 4) / 5;
                while (damage > 0) {
                    addReport(damageEntity(entity, entity.rollHitLocation(
                            ToHitData.HIT_NORMAL, ToHitData.SIDE_FRONT),
                            Math.min(5, damage)));
                    damage -= 5;
                }
                
                if (entity instanceof Tank) {
                    ((Tank) entity).immobilize();
                }

                
                break;
            }

            if ((nextHex.terrainLevel(Terrains.WATER) > 0)
                && (entity.getMovementMode() != EntityMovementMode.HOVER)
                && (entity.getMovementMode() != EntityMovementMode.WIGE)) {
                
                break;
            }

            if ((nextHex.terrainLevel(Terrains.MAGMA) == 1)
                && (nextElevation == 0)) {
                int roll = Compute.d6(1);
                r = new Report(2395);
                r.addDesc(entity);
                r.add(roll);
                r.subject = entity.getId();
                addReport(r);
                if (roll == 6) {
                    nextHex.removeTerrain(Terrains.MAGMA);
                    nextHex.addTerrain(Terrains.getTerrainFactory()
                                               .createTerrain(Terrains.MAGMA, 2));
                    sendChangedHex(curPos);
                    for (Entity en : game.getEntitiesVector(curPos)) {
                        if (en != entity) {
                            doMagmaDamage(en, false);
                        }
                    }
                }
            }

            if ((nextHex.terrainLevel(Terrains.MAGMA) == 2)
                && (nextElevation == 0)) {
                doMagmaDamage(entity, false);
            }

            
            PilotingRollData rollTarget = entity.checkBogDown(
                    step,
                    nextHex,
                    curPos,
                    nextPos,
                    step.getElevation(),
                    Compute.canMoveOnPavement(game, curPos, nextPos,
                                              step.getParentUpToThisStep()));
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                
                entity.setStuck(true);
                r = new Report(2081);
                r.subject = entity.getId();
                r.add(entity.getDisplayName(), true);
                addReport(r);
                
                addReport(checkQuickSand(nextPos));
                
                Entity violation = Compute.stackingViolation(game,
                                                             entity.getId(), curPos);
                if (violation != null) {
                    
                    Coords targetDest = Compute.getValidDisplacement(game,
                                                                     entity.getId(), curPos, direction);
                    addReport(doEntityDisplacement(violation, curPos,
                                                   targetDest, new PilotingRollData(violation.getId(),
                                                                                    0, "domino effect")));
                    
                    entityUpdate(violation.getId());
                }
                
                break;
                
            }

            curPos = nextPos;
            curHex = nextHex;
            r = new Report(2085);
            r.subject = entity.getId();
            r.indent();
            r.add(curPos.getBoardNum(), true);
            addReport(r);

        } 

        curPos = entity.getPosition();
        Entity target = Compute.stackingViolation(game, entity.getId(), curPos);
        while (target != null) {
            nextPos = Compute.getValidDisplacement(game, target.getId(),
                                                   target.getPosition(), direction);
            
            if (null == nextPos) {
                
                System.err.println("The skid of " + entity.getShortName()
                                   + " should displace " + target.getShortName()
                                   + " in hex " + curPos.getBoardNum()
                                   + " but there is nowhere to go.");
                break;
            }
            
            r = new Report(1210, Report.PUBLIC);
            r.indent();
            r.newlines = 0;
            addReport(r);
            addReport(doEntityDisplacement(target, curPos, nextPos, null));
            addReport(doEntityDisplacementMinefieldCheck(entity, curPos,
                                                         nextPos, entity.getElevation()));
            target = Compute.stackingViolation(game, entity.getId(), curPos);
        }

        if (entity instanceof Mech) {
            
            int damage = skidDistance
                         * (int) Math
                    .ceil(Math.round(entity.getWeight() / 10.0) / 2.0);

            r = new Report(2090);
            r.subject = entity.getId();
            r.indent();
            r.addDesc(entity);
            r.add(damage);
            addReport(r);

            while (damage > 0) {
                int cluster = Math.min(5, damage);
                HitData hit = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                     ToHitData.SIDE_FRONT);
                hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                addReport(damageEntity(entity, hit, cluster));
                damage -= cluster;
            }
            addNewLines();
        }

        if (entity.isDoomed()) {
            entity.setDestroyed(true);
            game.moveToGraveyard(entity.getId());
            send(createRemoveEntityPacket(entity.getId()));

            return true;
        }

        r = new Report(2095);
        r.subject = entity.getId();
        r.indent();
        addReport(r);

        return false;
    }

    private boolean processCollision(Entity entity, Entity target, Coords src) {
        Report r;

        r = new Report(9035);
        r.subject = entity.getId();
        r.add(entity.getDisplayName());
        r.add(target.getDisplayName());
        addReport(r);
        int partialroll = Compute.d6(1);
        boolean partial = (partialroll == 6);
        
        if ((target instanceof Aero)
            && (target.mpUsed < target.getRunMPwithoutMASC())
            && !((Aero) target).isOutControlTotal() && !target.isImmobile()) {
            
            
            Aero ta = (Aero) target;
            PilotingRollData psr = target.getBasePilotingRoll();
            psr.addModifier(0, "avoiding collision");
            int ctrlroll = Compute.d6(2);
            r = new Report(9045);
            r.subject = target.getId();
            r.add(target.getDisplayName());
            r.add(psr.getValue());
            r.add(ctrlroll);
            r.newlines = 0;
            r.indent(2);
            if (ctrlroll < psr.getValue()) {
                r.choose(false);
                addReport(r);
            } else {
                
                r.choose(true);
                addReport(r);
                
                if (!target.isDone()) {
                    int vel = ta.getCurrentVelocity();
                    MovePath md = new MovePath(game, target);
                    while (vel > 0) {
                        md.addStep(MoveStepType.FORWARDS);
                        vel--;
                    }
                    game.removeTurnFor(target);
                    send(createTurnVectorPacket());
                    processMovement(target, md, null);
                    
                } else {
                    
                    target.moved = EntityMovementType.MOVE_OVER_THRUST;
                    
                    
                    if (target.mpUsed <= ta.getSI()) {
                        PilotingRollData rollTarget = ta.checkThrustSITotal(
                                target.getRunMPwithoutMASC(), ta.moved);
                        if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                            game.addControlRoll(new PilotingRollData(
                                    ta.getId(), 0,
                                    "Thrust spent during turn exceeds SI"));
                        }
                    }
                    target.mpUsed = target.getRunMPwithoutMASC();
                }
                return false;
            }
        } else {
            
            r = new Report(9040);
            r.subject = entity.getId();
            r.add(entity.getDisplayName());
            r.indent(2);
            addReport(r);
        }

        ToHitData toHit = new ToHitData(TargetRoll.AUTOMATIC_SUCCESS,
                                        "Its a collision");
        toHit.setSideTable(target.sideTable(src));
        resolveRamDamage((Aero) entity, target, toHit, partial, false);

        if (target.isDoomed()) {
            
            if (!target.isDone()) {
                
                game.removeTurnFor(target);
                send(createTurnVectorPacket());
            } 
            
            target.setDestroyed(true);
            game.moveToGraveyard(target.getId());
            send(createRemoveEntityPacket(target.getId()));
        }
        
        
        if (!game.isOutOfGame(target)) {
            entityUpdate(target.getId());
        }

        return true;

    }

    private boolean checkCrash(Entity entity, Coords pos, int altitude) {

        
        if (!(entity instanceof Aero)) {
            return false;
        }
        
        if (game.getBoard().inSpace()) {
            return false;
        }
        
        else if (game.getBoard().onGround()) {
            if (altitude <= 0) {
                return true;
            }
            return false;
        }
        
        
        int ceiling = 0;
        if (game.getBoard().getHex(pos) != null) {
            ceiling = game.getBoard().getHex(pos).ceiling(true);
        }
        return ceiling >= altitude;
    }

    private Vector<Report> processCrash(Entity entity, int vel, Coords c) {
        Vector<Report> vReport = new Vector<Report>();
        Report r;
        if (c == null) {
            r = new Report(9701);
            r.subject = entity.getId();
            vReport.add(r);
            vReport.addAll(destroyEntity(entity, "crashed off the map", true,
                                         true));
            return vReport;
        }

        if (game.getBoard().inAtmosphere()) {
            r = new Report(9393, Report.PUBLIC);
            r.indent();
            r.addDesc(entity);
            vReport.add(r);
            entity.setDoomed(true);
        } else {
            ((Aero) entity).land();
        }

        ArrayList<Coords> coords = new ArrayList<Coords>();
        coords.add(c);
        IHex h = game.getBoard().getHex(c);
        int crateredElevation;
        boolean containsWater = false;
        if (h.containsTerrain(Terrains.WATER)) {
            crateredElevation = Math.min(2, h.depth() + 1);
            containsWater = true;
        } else {
            crateredElevation = h.getLevel() - 2;
        }
        if (entity instanceof Dropship) {
            for (int i = 0; i < 6; i++) {
                Coords adjCoords = c.translated(i);
                if (!game.getBoard().contains(c)) {
                    continue;
                }
                IHex adjHex = game.getBoard().getHex(adjCoords);
                coords.add(adjCoords);
                if (adjHex.containsTerrain(Terrains.WATER)) {
                    if (containsWater) {
                        int newDepth = Math.min(2, adjHex.depth() + 1);
                        if (newDepth > crateredElevation) {
                            crateredElevation = newDepth;
                        }
                    } else {
                        crateredElevation = Math.min(2, adjHex.depth() + 1);
                        containsWater = true;
                    }
                } else if (!containsWater
                           && (adjHex.getLevel() < crateredElevation)) {
                    crateredElevation = adjHex.getLevel();
                }
            }
        }
        if (vel < 1) {
            vel = 1;
        }

        boolean damageDealt = false;
        for (Coords hitCoords : coords) {
            int orig_crash_damage = Compute.d6(2) * 10 * vel;
            int crash_damage = orig_crash_damage;
            int direction = entity.getFacing();
            
            Building bldg = game.getBoard().getBuildingAt(hitCoords);
            if ((null != bldg) && (bldg.getType() == Building.HARDENED)) {
                crash_damage *= 2;
            }
            if (null != bldg) {
                collapseBuilding(bldg, game.getPositionMap(), hitCoords, true,
                                 vReport);
            }
            if (!damageDealt) {
                r = new Report(9700, Report.PUBLIC);
                r.indent();
                r.addDesc(entity);
                r.add(crash_damage);
                vReport.add(r);
                while (crash_damage > 0) {
                    HitData hit;
                    if ((entity instanceof SmallCraft)
                        && ((SmallCraft) entity).isSpheroid()) {
                        hit = entity.rollHitLocation(
                                ToHitData.HIT_SPHEROID_CRASH,
                                ToHitData.SIDE_REAR);
                    } else {
                        hit = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                     ToHitData.SIDE_FRONT);
                    }

                    if (crash_damage > 10) {
                        vReport.addAll(damageEntity(entity, hit, 10));
                    } else {
                        vReport.addAll(damageEntity(entity, hit, crash_damage));
                    }
                    crash_damage -= 10;
                }
                damageDealt = true;
            }

            for (Entity victim : game.getEntitiesVector(hitCoords)) {
                if (victim.getId() == entity.getId()) {
                    continue;
                }
                if (((victim.getElevation() > 0) && victim
                        .isAirborneVTOLorWIGE()) || (victim.getAltitude() > 0)) {
                    continue;
                }
                
                if ((entity instanceof Dropship) && !(victim instanceof Mech)) {
                    vReport.addAll(destroyEntity(victim,
                                                 "hit by crashing dropship"));
                } else {
                    crash_damage = orig_crash_damage / 2;
                    
                    int target = 2;
                    if (victim instanceof Infantry) {
                        target = 3;
                    }
                    int roll = Compute.d6();
                    r = new Report(9705, Report.PUBLIC);
                    r.indent();
                    r.addDesc(victim);
                    r.add(target);
                    r.add(crash_damage);
                    r.add(roll);
                    if (roll > target) {
                        r.choose(true);
                        vReport.add(r);
                      
                        while (crash_damage > 0) {
                            HitData hit = victim.rollHitLocation(
                                    ToHitData.HIT_NORMAL, ToHitData.SIDE_FRONT);
                            if (victim instanceof Mech) {
                                hit = victim.rollHitLocation(
                                        ToHitData.HIT_PUNCH,
                                        ToHitData.SIDE_FRONT);
                            }
                            if (victim instanceof Protomech) {
                                hit = victim.rollHitLocation(
                                        ToHitData.HIT_SPECIAL_PROTO,
                                        ToHitData.SIDE_FRONT);
                            }
                            if (crash_damage > 5) {
                                vReport.addAll(damageEntity(victim, hit, 5));
                            } else {
                                vReport.addAll(damageEntity(victim, hit,
                                                            crash_damage));
                            }
                            crash_damage -= 5;
                        }

                    } else {
                        r.choose(false);
                        vReport.add(r);
                    }
                }

                if (!victim.isDoomed() && !victim.isDestroyed()) {
                    
                    Coords dest = Compute.getValidDisplacement(game,
                                                               victim.getId(), hitCoords, direction);
                    if (null != dest) {
                        doEntityDisplacement(
                                victim,
                                hitCoords,
                                dest,
                                new PilotingRollData(victim.getId(), 0, "crash"));
                    } else if (!(victim instanceof Dropship)) {
                        
                        
                        addReport(destroyEntity(victim,
                                                "impossible displacement",
                                                victim instanceof Mech, victim instanceof Mech));
                    }
                }

            }

            h = game.getBoard().getHex(hitCoords);
            if (h.containsTerrain(Terrains.WOODS)) {
                if (entity instanceof Dropship) {
                    h.removeTerrain(Terrains.WOODS);
                    h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                            Terrains.ROUGH, 1));
                } else {
                    int level = h.terrainLevel(Terrains.WOODS) - 1;
                    if (level > 0) {
                        h.removeTerrain(Terrains.WOODS);
                        h.addTerrain(Terrains.getTerrainFactory()
                                             .createTerrain(Terrains.WOODS, level));
                    } else {
                        h.removeTerrain(Terrains.WOODS);
                        h.addTerrain(Terrains.getTerrainFactory()
                                             .createTerrain(Terrains.ROUGH, 1));
                    }
                }
            }
            
            if (h.containsTerrain(Terrains.JUNGLE)) {
                if (entity instanceof Dropship) {
                    h.removeTerrain(Terrains.JUNGLE);
                    h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                            Terrains.ROUGH, 1));
                } else {
                    int level = h.terrainLevel(Terrains.JUNGLE) - 1;
                    if (level > 0) {
                        h.removeTerrain(Terrains.JUNGLE);
                        h.addTerrain(Terrains.getTerrainFactory()
                                             .createTerrain(Terrains.JUNGLE, level));
                    } else {
                        h.removeTerrain(Terrains.JUNGLE);
                        h.addTerrain(Terrains.getTerrainFactory()
                                             .createTerrain(Terrains.ROUGH, 1));
                    }
                }
            }
            if (entity instanceof Dropship) {
                if (!containsWater) {
                    h.setLevel(crateredElevation);
                } else {
                    if (!h.containsTerrain(Terrains.WATER)) {
                        h.removeAllTerrains();
                    }
                    h.addTerrain(new Terrain(Terrains.WATER, crateredElevation,
                                             false, 0));
                }
            }
            sendChangedHex(hitCoords);
        }

        
        if (null != Compute.stackingViolation(game, entity.getId(), c)) {
            Coords dest = Compute.getValidDisplacement(game, entity.getId(), c,
                                                       Compute.d6() - 1);
            if (null != dest) {
                doEntityDisplacement(entity, c, dest, null);
            } else {
                
                vPhaseReport.addAll(destroyEntity(entity,
                                                  "impossible displacement", entity instanceof Mech,
                                                  entity instanceof Mech));
            }
        }

        h = game.getBoard().getHex(c);
        if (h.containsTerrain(Terrains.WATER) && !entity.isDestroyed()
            && !entity.isDoomed()) {
            int lethalDepth;
            if (entity instanceof Dropship) {
                lethalDepth = 2;
            } else {
                lethalDepth = 1;
            }

            if (h.depth() >= lethalDepth) {
                
                vReport.addAll(destroyEntity(entity,
                                             "crashing into deep water", true, true));
            }
        }

        return vReport;
    }

    private Vector<Report> processLeaveMap(Entity entity, int facing,
                                           boolean flewOff, int returnable) {
        Vector<Report> vReport = new Vector<Report>();
        Report r;
        
        r = new Report(2005, Report.PUBLIC);
        if (flewOff) {
            r = new Report(9370, Report.PUBLIC);
        }
        r.addDesc(entity);
        addReport(r);
        OffBoardDirection fleeDirection;
        if (facing == 0) {
            fleeDirection = OffBoardDirection.NORTH;
        } else if (facing == 3) {
            fleeDirection = OffBoardDirection.SOUTH;
        } else if (facing > 3) {
            fleeDirection = OffBoardDirection.WEST;
        } else {
            fleeDirection = OffBoardDirection.EAST;
        }

        if (returnable > -1) {

            entity.setDeployed(false);
            entity.setDeployRound(1 + game.getRoundCount() + returnable);
            entity.setPosition(null);
            entity.setDone(true);
            if (entity instanceof Aero) {
                
                ((Aero) entity).setOutControl(false);
            }
            switch (fleeDirection) {
                case WEST:
                    entity.setStartingPos(Board.START_W);
                    break;
                case NORTH:
                    entity.setStartingPos(Board.START_N);
                    break;
                case EAST:
                    entity.setStartingPos(Board.START_E);
                    break;
                case SOUTH:
                    entity.setStartingPos(Board.START_S);
                    break;
                default:
                    entity.setStartingPos(Board.START_EDGE);
            }
            entityUpdate(entity.getId());
            return vReport;
        }

        final List<Entity> passengers = entity.getLoadedUnits();
        if (!passengers.isEmpty()) {
            for (Entity passenger : passengers) {
                
                r = new Report(2010, Report.PUBLIC);
                r.indent();
                r.addDesc(passenger);
                addReport(r);
                passenger.setRetreatedDirection(fleeDirection);
                game.removeEntity(passenger.getId(),
                                  IEntityRemovalConditions.REMOVE_IN_RETREAT);
                send(createRemoveEntityPacket(passenger.getId(),
                                              IEntityRemovalConditions.REMOVE_IN_RETREAT));
            }
        }

        for (Integer mechWarriorId : entity.getPickedUpMechWarriors()) {
            Entity mw = game.getEntity(mechWarriorId.intValue());

            
            int condition = IEntityRemovalConditions.REMOVE_IN_RETREAT;
            r = new Report(2010);
            if (mw.isCaptured()) {
                r = new Report(2015);
                condition = IEntityRemovalConditions.REMOVE_CAPTURED;
            } else {
                mw.setRetreatedDirection(fleeDirection);
            }
            game.removeEntity(mw.getId(), condition);
            send(createRemoveEntityPacket(mw.getId(), condition));
            r.addDesc(mw);
            r.indent();
            addReport(r);
        }
        
        final int swarmerId = entity.getSwarmAttackerId();
        if (Entity.NONE != swarmerId) {
            final Entity swarmer = game.getEntity(swarmerId);

            if (!swarmer.isDone()) {

                game.removeTurnFor(swarmer);
                send(createTurnVectorPacket());

            } 

            swarmer.setSwarmTargetId(Entity.NONE);
            entity.setSwarmAttackerId(Entity.NONE);
            r = new Report(2015, Report.PUBLIC);
            r.indent();
            r.addDesc(swarmer);
            addReport(r);
            game.removeEntity(swarmerId,
                              IEntityRemovalConditions.REMOVE_CAPTURED);
            send(createRemoveEntityPacket(swarmerId,
                                          IEntityRemovalConditions.REMOVE_CAPTURED));
        }
        entity.setRetreatedDirection(fleeDirection);
        game.removeEntity(entity.getId(),
                          IEntityRemovalConditions.REMOVE_IN_RETREAT);
        send(createRemoveEntityPacket(entity.getId(),
                                      IEntityRemovalConditions.REMOVE_IN_RETREAT));
        return vReport;
    }

    private void processMovement(Entity entity, MovePath md,
            Map<EntityTargetPair, LosEffects> losCache) {
        
        if (losCache == null) {
            losCache = new HashMap<>();
        }
        Report r;
        boolean sideslipped = false; 
        PilotingRollData rollTarget;

        
        if (md.contains(MoveStepType.FLEE)) {
            addReport(processLeaveMap(entity, entity.getFacing(), false, -1));
            return;
        }

        if (md.contains(MoveStepType.EJECT)) {
            if ((entity instanceof Mech) || (entity instanceof Aero)) {
                r = new Report(2020);
                r.subject = entity.getId();
                r.add(entity.getCrew().getName());
                r.addDesc(entity);
                addReport(r);
            } else if ((entity instanceof Tank) && !entity.isCarcass()) {
                r = new Report(2025);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);
            }
            addReport(ejectEntity(entity, false));

            return;
        }

        if (md.contains(MoveStepType.CAREFUL_STAND)) {
            entity.setCarefulStand(true);
        }
        if (md.contains(MoveStepType.BACKWARDS)) {
            entity.setMovedBackwards(true);
            if (md.getMpUsed() > entity.getWalkMP()) {
                entity.setPowerReverse(true);
            }
        }

        if (md.contains(MoveStepType.TAKEOFF) && (entity instanceof Aero)) {
            Aero a = (Aero) entity;
            a.setCurrentVelocity(1);
            a.liftOff(1);
            if (entity instanceof Dropship) {
                applyDropshipProximityDamage(md.getFinalCoords(), true,
                                             md.getFinalFacing(), entity);
            }
            checkForTakeoffDamage(a);
            a.setPosition(a.getPosition().translated(a.getFacing(),
                                                     a.getTakeOffLength()));
            entity.setDone(true);
            entityUpdate(entity.getId());
            return;
        }

        if (md.contains(MoveStepType.VTAKEOFF) && (entity instanceof Aero)) {
            Aero a = (Aero) entity;
            rollTarget = a.checkVerticalTakeOff();
            if (doVerticalTakeOffCheck(entity, rollTarget)) {
                a.setCurrentVelocity(0);
                a.liftOff(1);
                if (entity instanceof Dropship) {
                    applyDropshipProximityDamage(md.getFinalCoords(), a);
                }
                checkForTakeoffDamage(a);
            }
            entity.setDone(true);
            entityUpdate(entity.getId());
            return;
        }

        if (md.contains(MoveStepType.LAND) && (entity instanceof Aero)) {
            Aero a = (Aero) entity;
            rollTarget = a.checkLanding(md.getLastStepMovementType(),
                    md.getFinalVelocity(), md.getFinalCoords(),
                    md.getFinalFacing(), false);
            doAttemptLanding(entity, rollTarget);
            a.land();
            entity.setPosition(md.getFinalCoords().translated(
                    md.getFinalFacing(), a.getLandingLength()));
            entity.setDone(true);
            entityUpdate(entity.getId());
            return;
        }

        if (md.contains(MoveStepType.VLAND) && (entity instanceof Aero)) {
            Aero a = (Aero) entity;
            rollTarget = a.checkLanding(md.getLastStepMovementType(),
                    md.getFinalVelocity(), md.getFinalCoords(),
                    md.getFinalFacing(), true);
            doAttemptLanding(entity, rollTarget);
            if (entity instanceof Dropship) {
                applyDropshipLandingDamage(md.getFinalCoords(), a);
            }
            a.land();
            entity.setPosition(md.getFinalCoords());
            entity.setDone(true);
            entityUpdate(entity.getId());
            return;
        }

        
        Coords lastPos = entity.getPosition();
        Coords curPos = entity.getPosition();
        int curFacing = entity.getFacing();
        int curVTOLElevation = entity.getElevation();
        int curElevation = entity.getElevation();
        int lastElevation = entity.getElevation();
        int curAltitude = entity.getAltitude();
        
        int distance = entity.delta_distance;
        int mpUsed = entity.mpUsed;
        EntityMovementType moveType = entity.moved;
        EntityMovementType overallMoveType = entity.moved;
        boolean firstStep;
        boolean wasProne = entity.isProne();
        boolean fellDuringMovement = false;
        boolean crashedDuringMovement = false;
        boolean dropshipStillUnloading = false;
        boolean turnOver;
        boolean wigeStartedLanded = false;
        int prevFacing = curFacing;
        IHex prevHex = null;
        final boolean isInfantry = entity instanceof Infantry;
        AttackAction charge = null;
        RamAttackAction ram = null;
        
        
        int cachedGravityLimit = -1;
        int thrustUsed = 0;
        int j = 0;
        boolean didMove = false;
        boolean recovered = false;
        Entity loader = null;

        Vector<Coords> passedThrough = entity.getPassedThrough();
        passedThrough.add(curPos);
        List<Integer> passedThroughFacing = entity.getPassedThroughFacing();
        passedThroughFacing.add(curFacing);

        
        md.compile(game, entity);

        entity.setVectors(md.getFinalVectors());

        overallMoveType = md.getLastStepMovementType();

        
        if ((game.getBoard().getHex(entity.getPosition())
                 .terrainLevel(Terrains.MAGMA) == 2)
            && (entity.getElevation() == 0)) {
            doMagmaDamage(entity, false);
        }

        if (entity instanceof Aero) {
            Aero a = (Aero) entity;
            a.setAccLast(false);
        }

        if (entity.isDropping()) {
            entity.setAltitude(entity.getAltitude()
                               - game.getPlanetaryConditions().getDropRate());
            
            if (md.length() > 0) {
                entity.setFacing(md.getFinalFacing());
            }
            passedThrough.add(entity.getPosition());
            entity.setPassedThrough(passedThrough);
            passedThroughFacing.add(entity.getFacing());
            entity.setPassedThroughFacing(passedThroughFacing);
            entity.setDone(true);
            entityUpdate(entity.getId());
            return;
        }

        
        firstStep = true;
        turnOver = false;
        
        MoveStep prevStep = null;

        Vector<UnitLocation> movePath = new Vector<UnitLocation>();

        for (final Enumeration<MoveStep> i = md.getSteps(); i.hasMoreElements(); ) {
            final MoveStep step = i.nextElement();
            wasProne = entity.isProne();
            boolean isPavementStep = step.isPavementStep();
            entity.inReverse = step.isThisStepBackwards();
            boolean entityFellWhileAttemptingToStand = false;
            boolean isOnGround = !i.hasMoreElements();
            isOnGround |= step.getMovementType() != EntityMovementType.MOVE_JUMP;
            isOnGround &= step.getElevation() < 1;

            
            if (step.getMovementType() == EntityMovementType.MOVE_ILLEGAL) {
                break;
            }

            
            if (entity.isDestroyed() || entity.isDoomed()) {
                break;
            }

            if (firstStep
                && (entity.getMovementMode() == EntityMovementMode.WIGE)
                && (entity.getElevation() == 0)) {
                wigeStartedLanded = true;
            }

            if (firstStep
                && ((entity instanceof Mech) || (entity instanceof Tank))) {
                HashMap<Integer, List<CriticalSlot>> crits = new HashMap<Integer, List<CriticalSlot>>();
                Vector<Report> vReport = new Vector<Report>();
                if (entity.checkForMASCFailure(md, vReport, crits)) {
                    boolean mascFailure = true;
                    
                    if (entity.getCrew().hasEdgeRemaining()
                        && entity.getCrew().getOptions()
                                 .booleanOption("edge_when_masc_fails")) {
                        entity.getCrew().decreaseEdge();
                        
                        entity.setMASCUsed(false);
                        
                        Report masc_report = new Report(6501);
                        masc_report.subject = entity.getId();
                        masc_report.indent(2);
                        masc_report.addDesc(entity);
                        vReport.add(masc_report);
                        
                        masc_report = new Report(6510);
                        masc_report.subject = entity.getId();
                        masc_report.indent(2);
                        masc_report.addDesc(entity);
                        masc_report.add(entity.getCrew().getOptions()
                                              .intOption("edge"));
                        vReport.addElement(masc_report);
                        
                        if (!entity.checkForMASCFailure(md, vReport, crits)) { 
                            
                        
                            mascFailure = false;
                            addReport(vReport);
                        }
                    }
                    
                    if (mascFailure) {
                        addReport(vReport);
                        for (Integer loc : crits.keySet()) {
                            List<CriticalSlot> lcs = crits.get(loc);
                            for (CriticalSlot cs : lcs) {
                              
                                if (loc == -1) {
                                    addReport(vehicleMotiveDamage(
                                            (Tank) entity, 0, true,
                                            cs.getIndex()));
                                } else {
                                    addReport(applyCriticalHit(entity, loc, cs,
                                                               true, 0, false));
                                }
                            }
                        }
                        
                        addReport(resolvePilotingRolls(entity));
                        game.resetPSRs(entity);
                        
                        
                        md.clear();
                        fellDuringMovement = true; 
                    }
                } else {
                    addReport(vReport);
                }
            }

            didMove = step.getDistance() > distance;

            
            if (entity.isAirborne() && (entity instanceof Aero)) {
                Aero a = (Aero) entity;
                j++;

                a.setStraightMoves(step.getNStraight());

                if (!didMove && (md.length() != j)) {
                    thrustUsed += step.getMp();
                } else {
                    
                    
                    if (!didMove && (md.length() == j)) {
                        thrustUsed += step.getMp();
                    }
                    
                    
                    rollTarget = a.checkThrustSI(thrustUsed, overallMoveType);
                    if ((rollTarget.getValue() != TargetRoll.CHECK_FALSE)
                        && !(entity instanceof FighterSquadron)
                        && !game.useVectorMove()) {
                        if (!doSkillCheckInSpace(entity, rollTarget)) {
                            a.setSI(a.getSI() - 1);
                            
                            if (a.getSI() == 0) {
                                addReport(destroyEntity(entity,
                                        "Structural Integrity Collapse", false));
                            }
                        }
                    }

                    int hits = entity.getCrew().getHits();
                    int health = 6 - hits;

                    if ((thrustUsed > (2 * health)) && !game.useVectorMove()
                        && !(entity instanceof TeleMissile)) {
                        int targetroll = 2 + (thrustUsed - (2 * health))
                                         + (2 * hits);
                        resistGForce(entity, targetroll);
                    }

                    thrustUsed = 0;
                }

                if (step.getType() == MoveStepType.RETURN) {
                    a.setCurrentVelocity(md.getFinalVelocity());
                    a.setAltitude(curAltitude);
                    processLeaveMap(entity, curFacing, true,
                                    Compute.roundsUntilReturn(game, entity));
                    return;
                }

                if (step.getType() == MoveStepType.OFF) {
                    a.setCurrentVelocity(md.getFinalVelocity());
                    a.setAltitude(curAltitude);
                    processLeaveMap(entity, curFacing, true, -1);
                    return;
                }

                rollTarget = a.checkRolls(step, overallMoveType);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    game.addControlRoll(new PilotingRollData(a.getId(), 0,
                                                             "excess roll"));
                }

                rollTarget = a.checkManeuver(step, overallMoveType);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    if (!doSkillCheckManeuver(entity, rollTarget)) {
                        a.setFailedManeuver(true);
                        int forward = Math.max(step.getVelocityLeft() / 2, 1);
                        if (forward < step.getVelocityLeft()) {
                            fellDuringMovement = true;
                        }
                        
                        if (game.getBoard().onGround()) {
                            forward *= 16;
                        }
                        while (forward > 0) {
                            curPos = curPos.translated(step.getFacing());
                            forward--;
                            distance++;
                            a.setStraightMoves(a.getStraightMoves() + 1);
                            
                            if (!game.getBoard().contains(curPos)) {
                                a.setCurrentVelocity(md.getFinalVelocity());
                                processLeaveMap(entity, curFacing, true,
                                                Compute.roundsUntilReturn(game, entity));
                                return;
                                
                            } else if (checkCrash(entity, curPos,
                                                  step.getAltitude())) {
                                addReport(processCrash(entity,
                                                       step.getVelocity(), curPos));
                                forward = 0;
                                fellDuringMovement = false;
                                crashedDuringMovement = true;
                            }
                            lastPos = curPos;
                        }
                        break;
                    }
                }

                if (didMove && a.isOutControlTotal()) {
                    Iterator<Entity> targets = game.getEntities(step
                            .getPosition());
                    if (targets.hasNext()) {
                        
                        int checkroll = Compute.d6(2);
                        
                        int targetroll = 11;
                        if ((a instanceof Dropship)
                            || (entity instanceof Jumpship)) {
                            targetroll = 10;
                        }
                        if (checkroll >= targetroll) {
                          
                            Vector<Integer> potentialSpaceStation;
                            Vector<Integer> potentialWarship;
                            Vector<Integer> potentialJumpship;
                            Vector<Integer> potentialDropship;
                            Vector<Integer> potentialSmallCraft;
                            Vector<Integer> potentialASF;
                            potentialSpaceStation = new Vector<Integer>();
                            potentialWarship = new Vector<Integer>();
                            potentialJumpship = new Vector<Integer>();
                            potentialDropship = new Vector<Integer>();
                            potentialSmallCraft = new Vector<Integer>();
                            potentialASF = new Vector<Integer>();
                            while (targets.hasNext()) {
                                int id = targets.next().getId();
                                Entity ce = game.getEntity(id);
                                
                                
                                if (!game.getBoard().inSpace()
                                    && (ce.getAltitude() != curAltitude)) {
                                    continue;
                                }
                                
                                if (ce.equals(a)) {
                                    continue;
                                }
                                if (ce instanceof SpaceStation) {
                                    potentialSpaceStation.addElement(id);
                                } else if (ce instanceof Warship) {
                                    potentialWarship.addElement(id);
                                } else if (ce instanceof Jumpship) {
                                    potentialJumpship.addElement(id);
                                } else if (ce instanceof Dropship) {
                                    potentialDropship.addElement(id);
                                } else if (ce instanceof SmallCraft) {
                                    potentialSmallCraft.addElement(id);
                                } else {
                                    
                                    potentialASF.addElement(id);
                                }
                            }

                            if (potentialSpaceStation.size() > 0) {
                                int chosen = Compute
                                        .randomInt(potentialSpaceStation.size());
                                Entity target = game
                                        .getEntity(potentialSpaceStation
                                                           .elementAt(chosen));
                                Coords dest = target.getPosition();
                                if (processCollision(entity, target, lastPos)) {
                                    curPos = dest;
                                    break;
                                }
                            } else if (potentialWarship.size() > 0) {
                                int chosen = Compute.randomInt(potentialWarship
                                                                       .size());
                                Entity target = game.getEntity(potentialWarship
                                                                       .elementAt(chosen));
                                Coords dest = target.getPosition();
                                if (processCollision(entity, target, lastPos)) {
                                    curPos = dest;
                                    break;
                                }
                            } else if (potentialJumpship.size() > 0) {
                                int chosen = Compute
                                        .randomInt(potentialJumpship.size());
                                Entity target = game
                                        .getEntity(potentialJumpship
                                                           .elementAt(chosen));
                                Coords dest = target.getPosition();
                                if (processCollision(entity, target, lastPos)) {
                                    curPos = dest;
                                    break;
                                }
                            } else if (potentialDropship.size() > 0) {
                                int chosen = Compute
                                        .randomInt(potentialDropship.size());
                                Entity target = game
                                        .getEntity(potentialDropship
                                                           .elementAt(chosen));
                                Coords dest = target.getPosition();
                                if (processCollision(entity, target, lastPos)) {
                                    curPos = dest;
                                    break;
                                }
                            } else if (potentialSmallCraft.size() > 0) {
                                int chosen = Compute
                                        .randomInt(potentialSmallCraft.size());
                                Entity target = game
                                        .getEntity(potentialSmallCraft
                                                           .elementAt(chosen));
                                Coords dest = target.getPosition();
                                if (processCollision(entity, target, lastPos)) {
                                    curPos = dest;
                                    break;
                                }
                            } else if (potentialASF.size() > 0) {
                                int chosen = Compute.randomInt(potentialASF
                                                                       .size());
                                Entity target = game.getEntity(potentialASF
                                                                       .elementAt(chosen));
                                Coords dest = target.getPosition();
                                if (processCollision(entity, target, lastPos)) {
                                    curPos = dest;
                                    break;
                                }
                            }

                        }
                    }
                }

                if (checkCrash(entity, step.getPosition(), step.getAltitude())) {
                    addReport(processCrash(entity, md.getFinalVelocity(),
                                           curPos));
                    crashedDuringMovement = true;
                    
                    break;
                }

                if (step.getType() == MoveStepType.LAUNCH) {
                    TreeMap<Integer, Vector<Integer>> launched = step
                            .getLaunched();
                    Set<Integer> bays = launched.keySet();
                    Iterator<Integer> bayIter = bays.iterator();
                    Bay currentBay;
                    while (bayIter.hasNext()) {
                        int bayId = bayIter.next();
                        currentBay = entity.getFighterBays().elementAt(bayId);
                        Vector<Integer> launches = launched.get(bayId);
                        int nLaunched = launches.size();
                       
                        int doors = currentBay.getDoors();
                        int[] distribution = new int[doors];
                        for (int l = 0; l < nLaunched; l++) {
                            distribution[l % doors] = distribution[l % doors] + 1;
                        }
                        
                        r = new Report(9380);
                        r.add(entity.getDisplayName());
                        r.subject = entity.getId();
                        r.add(nLaunched);
                        r.add("bay " + currentBay.getBayNumber() + " (" + doors
                              + " doors)");
                        addReport(r);
                        int currentDoor = 0;
                        int fighterCount = 0;
                        boolean doorDamage = false;
                        for (int fighterId : launches) {
                            
                            fighterCount++;

                            Report doorReport = null;
                            if (!doorDamage && (distribution[currentDoor] > 2)
                                && (fighterCount > 2)) {
                                doorReport = new Report(9378);
                                doorReport.subject = entity.getId();
                                doorReport.indent(2);
                                int roll = Compute.d6(2);
                                doorReport.add(roll);
                                if (roll == 2) {
                                    doorDamage = true;
                                    doorReport.choose(true);
                                    currentBay.destroyDoorNext();
                                } else {
                                    doorReport.choose(false);
                                }
                                doorReport.newlines++;
                            }

                            if (fighterCount > distribution[currentDoor]) {
                                
                                currentDoor++;
                                fighterCount = 0;
                                doorDamage = false;
                            }
                            int bonus = Math.max(0,
                                                 distribution[currentDoor] - 2);

                            Entity fighter = game.getEntity(fighterId);
                            if (!launchUnit(entity, fighter, curPos, curFacing,
                                            step.getVelocity(), step.getAltitude(),
                                            step.getVectors(), bonus)) {
                                System.err
                                        .println("Error! Server was told to unload "
                                                 + fighter.getDisplayName()
                                                 + " from "
                                                 + entity.getDisplayName()
                                                 + " into "
                                                 + curPos.getBoardNum());
                            }
                            if (doorReport != null) {
                                addReport(doorReport);
                            }
                        }
                    }
                    
                    entity.resetBayDoors();
                }

                if (step.getType() == MoveStepType.UNDOCK) {
                    TreeMap<Integer, Vector<Integer>> launched = step
                            .getLaunched();
                    Set<Integer> collars = launched.keySet();
                    Iterator<Integer> collarIter = collars.iterator();
                    while (collarIter.hasNext()) {
                        int collarId = collarIter.next();
                        Vector<Integer> launches = launched.get(collarId);
                        int nLaunched = launches.size();
                        
                        r = new Report(9380);
                        r.add(entity.getDisplayName());
                        r.subject = entity.getId();
                        r.add(nLaunched);
                        r.add("collar " + collarId);
                        addReport(r);
                        for (int dropshipId : launches) {
                            
                            Entity ds = game.getEntity(dropshipId);
                            if (!launchUnit(entity, ds, curPos, curFacing,
                                            step.getVelocity(), step.getAltitude(),
                                            step.getVectors(), 0)) {
                                System.err
                                        .println("Error! Server was told to unload "
                                                 + ds.getDisplayName()
                                                 + " from "
                                                 + entity.getDisplayName()
                                                 + " into "
                                                 + curPos.getBoardNum());
                            }
                        }
                    }
                }

                if (step.getType() == MoveStepType.DROP) {
                    TreeMap<Integer, Vector<Integer>> dropped = step
                            .getLaunched();
                    Set<Integer> bays = dropped.keySet();
                    Iterator<Integer> bayIter = bays.iterator();
                    Bay currentBay;
                    while (bayIter.hasNext()) {
                        int bayId = bayIter.next();
                        currentBay = entity.getTransportBays().elementAt(bayId);
                        Vector<Integer> drops = dropped.get(bayId);
                        int nDropped = drops.size();
                        
                        r = new Report(9386);
                        r.add(entity.getDisplayName());
                        r.subject = entity.getId();
                        r.add(nDropped);
                        addReport(r);
                        for (int unitId : drops) {
                            if (Compute.d6(2) == 2) {
                                r = new Report(9390);
                                r.subject = entity.getId();
                                r.indent(1);
                                r.add(currentBay.getType());
                                addReport(r);
                                currentBay.destroyDoorNext();
                            }
                            Entity drop = game.getEntity(unitId);
                            dropUnit(drop, entity, curPos, step.getAltitude());
                        }
                    }
                    
                    entity.resetBayDoors();
                }
            }

            rollTarget = entity.checkGetUp(step);

            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                
                
                if ((entity instanceof Mech)
                    && !(((Mech) entity).isIndustrial()
                        && ((entity.getEngine().getEngineType() == Engine.COMBUSTION_ENGINE)
                            || (entity.getEngine().getEngineType() == Engine.FUEL_CELL)))) {
                    entity.heatBuildup += 1;
                }
                entity.setProne(false);
                
                wasProne = false;
                game.resetPSRs(entity);
                entityFellWhileAttemptingToStand = !doSkillCheckInPlace(entity,
                                                                        rollTarget);
            }
            
            if (entityFellWhileAttemptingToStand) {
                moveType = step.getMovementType();
                curFacing = entity.getFacing();
                curPos = entity.getPosition();
                mpUsed = step.getMpUsed();
                fellDuringMovement = true;
                if (!entity.isCarefulStand()) {
                    break;
                }
            } else if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                entity.setHullDown(false);
            }

            if (step.getType() == MoveStepType.UNJAM_RAC) {
                entity.setUnjammingRAC(true);
                game.addAction(new UnjamAction(entity.getId()));

                if (!(entity.isAirborne())) {
                    break;
                }
            }

            if (step.getType() == MoveStepType.LAY_MINE) {
                layMine(entity, step.getMineToLay(), step.getPosition());
                continue;
            }

            if (step.getType() == MoveStepType.CLEAR_MINEFIELD) {
                ClearMinefieldAction cma = new ClearMinefieldAction(
                        entity.getId(), step.getMinefield());
                entity.setClearingMinefield(true);
                game.addAction(cma);
                break;
            }

            if ((step.getType() == MoveStepType.SEARCHLIGHT)
                && entity.hasSpotlight()) {
                final boolean SearchOn = !entity.isUsingSpotlight();
                entity.setSpotlightState(SearchOn);
                if (doBlind()) { 
                    
                    Vector<IPlayer> playersVector = game.getPlayersVector();
                    Vector<IPlayer> vCanSee = whoCanSee(entity);
                    for (IPlayer p : playersVector) {
                        if (vCanSee.contains(p)) { 
                            sendServerChat(p.getId(), entity.getDisplayName()
                                                      + " switched searchlight "
                                                      + (SearchOn ? "on" : "off") + '.');
                        } else {
                            sendServerChat(p.getId(), "An unseen unit"
                                                      + " switched searchlight "
                                                      + (SearchOn ? "on" : "off") + '.');
                        }
                    }
                } else { 
                    sendServerChat(entity.getDisplayName()
                                   + " switched searchlight "
                                   + (SearchOn ? "on" : "off") + '.');
                }
            }

            moveType = step.getMovementType();
            distance = step.getDistance();
            mpUsed = step.getMpUsed();

            if (cachedGravityLimit < 0) {
                cachedGravityLimit = EntityMovementType.MOVE_JUMP == moveType ? entity
                        .getJumpMP(false) : entity.getRunningGravityLimit();
            }
            
            if (step.getType() == MoveStepType.CHARGE) {
                if (entity.canCharge()) {
                    checkExtremeGravityMovement(entity, step, curPos,
                                                cachedGravityLimit);
                    Targetable target = step.getTarget(game);
                    if (target != null) {
                        ChargeAttackAction caa = new ChargeAttackAction(
                                entity.getId(), target.getTargetType(),
                                target.getTargetId(), target.getPosition());
                        entity.setDisplacementAttack(caa);
                        game.addCharge(caa);
                        charge = caa;
                    } else {
                        System.out.println("Illegal charge!! "
                                + entity.getDisplayName() + " is attempting to "
                                + "charge a null target!");
                        sendServerChat("Illegal charge!! "
                                + entity.getDisplayName() + " is attempting to "
                                + "charge a null target!");
                        return;
                    }
                } else {
                    sendServerChat("Illegal charge!! I don't think "
                                   + entity.getDisplayName()
                                   + " should be allowed to charge,"
                                   + " but the client of "
                                   + entity.getOwner().getName() + " disagrees.");
                    sendServerChat("Please make sure "
                                   + entity.getOwner().getName()
                                   + " is running MegaMek "
                                   + MegaMek.VERSION
                                   + ", or if that is already the case, submit a bug report at http://megamek.sf.net/");
                    return;
                }
                break;
            }

            if (step.getType() == MoveStepType.DFA) {
                if (entity.canDFA()) {
                    checkExtremeGravityMovement(entity, step, curPos,
                                                cachedGravityLimit);
                    Targetable target = step.getTarget(game);
                    DfaAttackAction daa = new DfaAttackAction(entity.getId(),
                                                              target.getTargetType(), target.getTargetId(),
                                                              target.getPosition());
                    entity.setDisplacementAttack(daa);
                    entity.setElevation(step.getElevation());
                    game.addCharge(daa);
                    charge = daa;
                } else {
                    sendServerChat("Illegal DFA!! I don't think "
                                   + entity.getDisplayName()
                                   + " should be allowed to DFA,"
                                   + " but the client of "
                                   + entity.getOwner().getName() + " disagrees.");
                    sendServerChat("Please make sure "
                                   + entity.getOwner().getName()
                                   + " is running MegaMek "
                                   + MegaMek.VERSION
                                   + ", or if that is already the case, submit a bug report at http://megamek.sf.net/");
                    return;
                }
                break;
            }

            if (step.getType() == MoveStepType.RAM) {
                if (entity.canRam()) {
                    Targetable target = step.getTarget(game);
                    RamAttackAction raa = new RamAttackAction(entity.getId(),
                                                              target.getTargetType(), target.getTargetId(),
                                                              target.getPosition());
                    entity.setRamming(true);
                    game.addRam(raa);
                    ram = raa;
                } else {
                    sendServerChat("Illegal ram!! I don't think "
                                   + entity.getDisplayName()
                                   + " should be allowed to charge,"
                                   + " but the client of "
                                   + entity.getOwner().getName() + " disagrees.");
                    sendServerChat("Please make sure "
                                   + entity.getOwner().getName()
                                   + " is running MegaMek "
                                   + MegaMek.VERSION
                                   + ", or if that is already the case, submit a bug report at http://megamek.sf.net/");
                    return;
                }
                break;
            }

            if ((step.getType() == MoveStepType.ACC)
                || (step.getType() == MoveStepType.ACCN)) {
                if (entity instanceof Aero) {
                    Aero a = (Aero) entity;
                    if (step.getType() == MoveStepType.ACCN) {
                        a.setAccLast(true);
                        a.setNextVelocity(a.getNextVelocity() + 1);
                    } else {
                        a.setAccDecNow(true);
                        a.setCurrentVelocity(a.getCurrentVelocity() + 1);
                        a.setNextVelocity(a.getNextVelocity() + 1);
                    }
                }
            }

            if ((step.getType() == MoveStepType.DEC)
                || (step.getType() == MoveStepType.DECN)) {
                if (entity instanceof Aero) {
                    Aero a = (Aero) entity;
                    if (step.getType() == MoveStepType.DECN) {
                        a.setAccLast(true);
                        a.setNextVelocity(a.getNextVelocity() - 1);
                    } else {
                        a.setAccDecNow(true);
                        a.setCurrentVelocity(a.getCurrentVelocity() - 1);
                        a.setNextVelocity(a.getNextVelocity() - 1);
                    }
                }
            }

            if (step.getType() == MoveStepType.EVADE) {
                entity.setEvading(true);
            }

            if (step.getType() == MoveStepType.SHUTDOWN) {
                entity.performManualShutdown();
                sendServerChat(entity.getDisplayName() + " has shutdown.");
            }

            if (step.getType() == MoveStepType.STARTUP) {
                entity.performManualStartup();
                sendServerChat(entity.getDisplayName() + " has started up.");
            }

            if (step.getType() == MoveStepType.SELF_DESTRUCT) {
                entity.setSelfDestructing(true);
            }

            if (step.getType() == MoveStepType.ROLL) {
                if (entity instanceof Aero) {
                    Aero a = (Aero) entity;
                    if (a.isRolled()) {
                        a.setRolled(false);
                    } else {
                        a.setRolled(true);
                    }

                }
            }

            if (entity instanceof Infantry) {
                Infantry inf = (Infantry) entity;
                if (step.getType() == MoveStepType.DIG_IN) {
                    inf.setDugIn(Infantry.DUG_IN_WORKING);
                    continue;
                } else if (step.getType() == MoveStepType.FORTIFY) {
                    if (!entity.hasWorkingMisc(MiscType.F_TOOLS,
                                               MiscType.S_VIBROSHOVEL)) {
                        sendServerChat(entity.getDisplayName()
                                       + " failed to fortify because it is missing suitable equipment");
                    }
                    inf.setDugIn(Infantry.DUG_IN_FORTIFYING1);
                    continue;
                } else if ((step.getType() != MoveStepType.TURN_LEFT)
                           && (step.getType() != MoveStepType.TURN_RIGHT)) {
                    
                    inf.setDugIn(Infantry.DUG_IN_NONE);
                }

                if (step.getType() == MoveStepType.TAKE_COVER) {
                    if (Infantry.hasValidCover(game, step.getPosition(),
                            step.getElevation())) {
                        inf.setTakingCover(true);
                    } else {
                        sendServerChat(entity.getDisplayName()
                                + " failed to take cover: "
                                + "no valid unit found in "
                                + step.getPosition());
                    }
                }
            }

            curPos = step.getPosition();
            if (!((entity.getJumpType() == Mech.JUMP_BOOSTER) && step
                    .getParent().isJumping())) {
                curFacing = step.getFacing();
            }
            
            int buildingMove = entity.checkMovementInBuilding(step, prevStep,
                                                              curPos, lastPos);
            curVTOLElevation = step.getElevation();
            curAltitude = step.getAltitude();
            curElevation = step.getElevation();
            
            entity.setElevation(step.getElevation());

            IHex curHex = game.getBoard().getHex(curPos);

            if (isOnGround && curHex.containsTerrain(Terrains.BUILDING)) {
                Building bldg = game.getBoard().getBuildingAt(curPos);
                if (bldg.rollBasement(curPos, game.getBoard(), vPhaseReport)) {
                    sendChangedHex(curPos);
                    Vector<Building> buildings = new Vector<Building>();
                    buildings.add(bldg);
                    sendChangedBuildings(buildings);
                }
            }

            if (entity.canUnstickByJumping() && entity.isStuck()
                && (moveType == EntityMovementType.MOVE_JUMP)) {
                entity.setStuck(false);
                entity.setCanUnstickByJumping(false);
            }

            if (!lastPos.equals(curPos)
                && (step.getMovementType() != EntityMovementType.MOVE_JUMP)
                && (entity instanceof Mech)
                && game.getOptions().booleanOption("tacops_leaping")) {
                int leapDistance = (lastElevation + game.getBoard()
                                                        .getHex(lastPos).getLevel())
                                   - (curElevation + curHex.getLevel());
                if (leapDistance > 2) {
                    
                    PilotingRollData roll = entity.getBasePilotingRoll(step
                                                                               .getMovementType());
                    entity.addPilotingModifierForTerrain(roll, curPos);
                    roll.append(new PilotingRollData(entity.getId(),
                                                     2 * leapDistance, "leaping (leg damage)"));
                    if (0 < doSkillCheckWhileMoving(entity, lastElevation,
                                                    lastPos, curPos, roll, false)) {
                        
                        addReport(damageEntity(entity, new HitData(
                                Mech.LOC_LLEG), leapDistance));
                        addReport(damageEntity(entity, new HitData(
                                Mech.LOC_RLEG), leapDistance));
                        addNewLines();
                        addReport(criticalEntity(entity, Mech.LOC_LLEG, false,
                                                 0, 0));
                        addNewLines();
                        addReport(criticalEntity(entity, Mech.LOC_RLEG, false,
                                                 0, 0));
                        if (entity instanceof QuadMech) {
                            addReport(damageEntity(entity, new HitData(
                                    Mech.LOC_LARM), leapDistance));
                            addReport(damageEntity(entity, new HitData(
                                    Mech.LOC_RARM), leapDistance));
                            addNewLines();
                            addReport(criticalEntity(entity, Mech.LOC_LARM,
                                                     false, 0, 0));
                            addNewLines();
                            addReport(criticalEntity(entity, Mech.LOC_RARM,
                                                     false, 0, 0));
                        }
                    }
                    
                    roll = entity.getBasePilotingRoll(step.getMovementType());
                    entity.addPilotingModifierForTerrain(roll, curPos);
                    roll.append(new PilotingRollData(entity.getId(),
                                                     leapDistance, "leaping (fall)"));
                    if (0 < doSkillCheckWhileMoving(entity, lastElevation,
                                                    lastPos, curPos, roll, false)) {
                        entity.setElevation(lastElevation);
                        addReport(doEntityFallsInto(entity, lastElevation,
                                                    lastPos, curPos,
                                                    entity.getBasePilotingRoll(overallMoveType),
                                                    false));
                    }
                }
            }

            
            rollTarget = entity.checkSkid(moveType, prevHex, overallMoveType,
                                          prevStep, prevFacing, curFacing, lastPos, curPos,
                                          isInfantry, distance - 1);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                
                boolean psrFailed = true;
                int startingfacing = entity.getFacing();
                if (entity instanceof Mech) {
                    
                    
                    entity.setFacing(curFacing);
                    psrFailed = (0 < doSkillCheckWhileMoving(entity,
                                                             lastElevation, lastPos, lastPos, rollTarget, true));
                } else {
                    psrFailed = (0 < doSkillCheckWhileMoving(entity,
                                                             lastElevation, lastPos, lastPos, rollTarget, false));
                }
                
                if (psrFailed) {

                    if (entity instanceof Tank) {
                        addReport(vehicleMotiveDamage((Tank) entity, 0));
                    }

                    curPos = lastPos;
                    int skidDistance = (int) Math
                            .round((double) (distance - 1) / 2);
                    int skidDirection = prevFacing;

                    
                    entity.delta_distance = distance - 1;

                    
                    moveType = EntityMovementType.MOVE_SKID;

                    
                    if (step.isThisStepBackwards()) {
                        skidDirection = (skidDirection + 3) % 6;
                    }

                    if (processSkid(entity, curPos, prevStep.getElevation(),
                                    skidDirection, skidDistance, prevStep)) {
                        return;
                    }

                    
                    curFacing = entity.getFacing();
                    curPos = entity.getPosition();
                    entity.setSecondaryFacing(curFacing);

                    
                    if (md.hasActiveMASC()) {
                        mpUsed = entity.getRunMP();
                    } else {
                        mpUsed = entity.getRunMPwithoutMASC();
                    }

                    entity.moved = moveType;
                    fellDuringMovement = true;
                    turnOver = true;
                    distance = entity.delta_distance;
                    break;

                } else { 
                    
                    
                    entity.setFacing(startingfacing);
                }

            } 

            
            if ((entity instanceof VTOL)
                || (entity.getMovementMode() == EntityMovementMode.HOVER)
                || (entity.getMovementMode() == EntityMovementMode.WIGE)) {
                rollTarget = entity.checkSideSlip(moveType, prevHex,
                                                  overallMoveType, prevStep, prevFacing, curFacing,
                                                  lastPos, curPos, distance);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    int moF = doSkillCheckWhileMoving(entity, lastElevation,
                                                      lastPos, curPos, rollTarget, false);
                    if (moF > 0) {
                        
                        int sideslipDistance = Math.min(moF, distance - 1);
                        if (sideslipDistance > 0) {
                            int skidDirection = prevFacing;
                            
                            sideslipped = true;
                            r = new Report(2100);
                            r.subject = entity.getId();
                            r.addDesc(entity);
                            r.add(sideslipDistance);
                            addReport(r);

                            if (processSkid(entity, lastPos,
                                            prevStep.getElevation(), skidDirection,
                                            sideslipDistance, prevStep)) {
                                return;
                            }

                            if (!entity.isDestroyed() && !entity.isDoomed()
                                && (mpUsed < entity.getRunMP())) {
                                fellDuringMovement = true; 
                                
                            }

                            if ((entity.getElevation() == 0)
                                && ((entity.getMovementMode() == EntityMovementMode.VTOL) || (entity
                                                                                                      .getMovementMode() == EntityMovementMode.WIGE))) {
                                turnOver = true;
                            }
                            
                            curFacing = step.getFacing();
                            curPos = entity.getPosition();
                            entity.setSecondaryFacing(curFacing);
                            break;
                        }
                    }
                }
            }

            
            boolean isLastStep = step.equals(md.getLastStep());
            rollTarget = entity.checkRubbleMove(step, curHex, lastPos, curPos,
                                                isLastStep);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                doSkillCheckWhileMoving(entity, lastElevation, lastPos, curPos,
                                        rollTarget, true);
            }

            
            rollTarget = entity.checkRecklessMove(step, curHex, lastPos,
                                                  curPos, prevHex);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                if (entity instanceof Mech) {
                    doSkillCheckWhileMoving(entity, lastElevation, lastPos,
                                            curPos, rollTarget, true);
                } else if (entity instanceof Tank) {
                    if (0 < doSkillCheckWhileMoving(entity, lastElevation,
                                                    lastPos, curPos, rollTarget, false)) {
                        
                        if ((0 == curHex.terrainsPresent())
                            || (step.getElevation() > 0)) {
                            if (entity instanceof VTOL) {
                                r = new Report(2208);
                            } else {
                                r = new Report(2206);
                            }
                            r.addDesc(entity);
                            r.subject = entity.getId();
                            addReport(r);
                            mpUsed = step.getMpUsed() + 1;
                            fellDuringMovement = true;
                            break;
                        }
                        r = new Report(2207);
                        r.addDesc(entity);
                        r.subject = entity.getId();
                        addReport(r);
                        
                        
                        
                        HitData hit = entity.rollHitLocation(
                                ToHitData.HIT_NORMAL, ToHitData.SIDE_FRONT);
                        addReport(damageEntity(entity, hit,
                                               ChargeAttackAction.getDamageTakenBy(entity,
                                                                                   entity)));
                        turnOver = true;
                        break;
                    }
                }
            }

            
            if ((curHex.terrainLevel(Terrains.MAGMA) == 1)
                && (step.getElevation() == 0)
                && (step.getMovementType() != EntityMovementType.MOVE_JUMP)) {
                int roll = Compute.d6(1);
                r = new Report(2395);
                r.addDesc(entity);
                r.add(roll);
                r.subject = entity.getId();
                addReport(r);
                if (roll == 6) {
                    curHex.removeTerrain(Terrains.MAGMA);
                    curHex.addTerrain(Terrains.getTerrainFactory()
                                              .createTerrain(Terrains.MAGMA, 2));
                    sendChangedHex(curPos);
                    for (Entity en : game.getEntitiesVector(curPos)) {
                        if (en != entity) {
                            doMagmaDamage(en, false);
                        }
                    }
                }
            }

            
            if ((curHex.terrainLevel(Terrains.MAGMA) == 2)
                && (step.getElevation() == 0)
                && (step.getMovementType() != EntityMovementType.MOVE_JUMP)) {
                doMagmaDamage(entity, false);
            }

            
            rollTarget = entity.checkBogDown(step, curHex, lastPos, curPos,
                                             lastElevation, isPavementStep);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                if (0 < doSkillCheckWhileMoving(entity, lastElevation, lastPos,
                                                curPos, rollTarget, false)) {
                    entity.setStuck(true);
                    entity.setCanUnstickByJumping(true);
                    r = new Report(2081);
                    r.add(entity.getDisplayName());
                    r.subject = entity.getId();
                    addReport(r);
                    
                    addReport(checkQuickSand(curPos));
                    
                    Entity violation = Compute.stackingViolation(game,
                                                                 entity.getId(), curPos);
                    if (violation != null) {
                        
                        int direction = lastPos.direction(curPos);
                        Coords targetDest = Compute.getValidDisplacement(game,
                                                                         entity.getId(), curPos, direction);
                        addReport(doEntityDisplacement(violation, curPos,
                                                       targetDest,
                                                       new PilotingRollData(violation.getId(), 0,
                                                                            "domino effect")));
                        
                        entityUpdate(violation.getId());
                    }
                    break;
                }
            }

            
            IHex lastHex = game.getBoard().getHex(lastPos);
            if (entity instanceof Mech) {
                if (!lastPos.equals(curPos)
                    && (prevStep != null)
                        && ((lastHex.containsTerrain(Terrains.FIRE)
                                && (prevStep.getElevation() <= 1))
                                || (lastHex.containsTerrain(Terrains.MAGMA)
                                        && (prevStep.getElevation() == 0)))
                        && ((step.getMovementType() != EntityMovementType.MOVE_JUMP)
                        
                        
                        
                        
                        
                        
                        
                        || (0 == (step.getMpUsed() - step.getMp())))) {
                    int heat = 0;
                    if (lastHex.containsTerrain(Terrains.FIRE)) {
                        heat += 2;
                    }
                    if (lastHex.terrainLevel(Terrains.MAGMA) == 1) {
                        heat += 2;
                    } else if (lastHex.terrainLevel(Terrains.MAGMA) == 2) {
                        heat += 5;
                    }
                    if (((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        heat /= 2;
                    }
                    entity.heatFromExternal += heat;
                    r = new Report(2115);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(heat);
                    addReport(r);
                    if (((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        r = new Report(5550);
                        addReport(r);
                    }
                }
            }

            
            if (!(entity instanceof Mech)) {
                boolean underwater = game.getBoard().getHex(curPos)
                                         .containsTerrain(Terrains.WATER)
                                     && (game.getBoard().getHex(curPos).depth() > 0)
                                     && (step.getElevation() < game.getBoard()
                                                                   .getHex(curPos).surface());
                if (game.getBoard().getHex(curPos)
                        .containsTerrain(Terrains.FIRE)
                    && !lastPos.equals(curPos)
                    && (step.getMovementType() != EntityMovementType.MOVE_JUMP)
                    && (step.getElevation() <= 1) && !underwater) {
                    doFlamingDamage(entity);
                }
            }
            
            if (!i.hasMoreElements() && !firstStep) {
                checkExtremeGravityMovement(entity, step, curPos,
                                            cachedGravityLimit);
            }
            
            
            
            if (!lastPos.equals(curPos) || (lastElevation != curElevation)) {
                boolean boom = false;
                if (isOnGround) {
                    boom = checkVibrabombs(entity, curPos, false, lastPos,
                                           curPos, vPhaseReport);
                }
                if (game.containsMinefield(curPos)) {
                    
                    
                    
                    entity.setPosition(curPos);
                    if (enterMinefield(entity, curPos, step.getElevation(),
                                       isOnGround, vPhaseReport)) {
                        
                        
                        if (step.getMovementType() != EntityMovementType.MOVE_JUMP) {
                            addReport(resolvePilotingRolls(entity));
                            game.resetPSRs(entity);
                        }
                        boom = true;
                    }
                    if (wasProne || !entity.isProne()) {
                        entity.setPosition(lastPos);
                    }
                }
                
                if (boom) {
                    
                    
                    
                    
                    
                    if (i.hasMoreElements()
                        && (step.getMovementType() != EntityMovementType.MOVE_JUMP)) {
                        md.clear();
                        fellDuringMovement = true;
                    }
                    
                    resetMines();
                }
            }

            
            
            if (!lastPos.equals(curPos) && !i.hasMoreElements() && isInfantry) {
                if (game.containsMinefield(curPos)) {
                    IPlayer owner = entity.getOwner();
                    for (Minefield mf : game.getMinefields(curPos)) {
                        if (!owner.containsMinefield(mf)) {
                            r = new Report(2120);
                            r.subject = entity.getId();
                            r.add(entity.getShortName(), true);
                            addReport(r);
                            revealMinefield(game.getTeamForPlayer(owner), mf);
                        }
                    }
                }
            }

            
            rollTarget = entity.checkWaterMove(step, curHex, lastPos, curPos,
                                               isPavementStep);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                
                final int swarmerId = entity.getSwarmAttackerId();
                boolean swarmerDone = true;
                Entity swarmer = null;
                if (Entity.NONE != swarmerId) {
                    swarmer = game.getEntity(swarmerId);
                    swarmerDone = swarmer.isDone();
                }

                
                entity.setFacing(curFacing);
                doSkillCheckWhileMoving(entity, lastElevation, lastPos, curPos,
                                        rollTarget, true);

                
                if (curHex.terrainLevel(Terrains.WATER) > 1) {
                    drownSwarmer(entity, curPos);
                }

                
                if (!swarmerDone && (swarmer != null)
                    && (swarmer.isDoomed() || swarmer.isDestroyed())) {
                    
                    
                    swarmer.setDone(false);
                    swarmer.setUnloaded(false);

                    
                    game.removeTurnFor(swarmer);
                    send(createTurnVectorPacket());

                    
                    swarmer.setDone(true);
                    swarmer.setUnloaded(true);
                }

                
                checkForWashedInfernos(entity, curPos);
            }

            
            
            
            
            addReport(doSetLocationsExposure(entity, curHex,
                                             step.getMovementType() == EntityMovementType.MOVE_JUMP,
                                             step.getElevation()));

            
            if ((lastElevation < 0) && (step.getElevation() == 0)
                && lastHex.containsTerrain(Terrains.ICE)
                && lastHex.containsTerrain(Terrains.WATER)
                && (step.getMovementType() != EntityMovementType.MOVE_JUMP)
                && !lastPos.equals(curPos)) {
                
                
                entity.setPosition(curPos);
                r = new Report(2410);
                r.addDesc(entity);
                addReport(r);
                addReport(resolveIceBroken(lastPos));
                
                entity.setPosition(lastPos);
            }
            
            if (curHex.containsTerrain(Terrains.ICE)
                && curHex.containsTerrain(Terrains.WATER)
                && (step.getMovementType() != EntityMovementType.MOVE_JUMP)
                && !lastPos.equals(curPos)
                && !(entity instanceof Infantry)
                && !(step.isPavementStep() && curHex
                    .containsTerrain(Terrains.BRIDGE))) {
                if (step.getElevation() == 0) {
                    int roll = Compute.d6(1);
                    r = new Report(2118);
                    r.addDesc(entity);
                    r.add(roll);
                    r.subject = entity.getId();
                    addReport(r);
                    if (roll == 6) {
                        entity.setPosition(curPos);
                        addReport(resolveIceBroken(curPos));
                        curPos = entity.getPosition();
                    }
                }
                
                else if ((step.getElevation() + entity.height()) == 0) {
                    r = new Report(2410);
                    r.addDesc(entity);
                    addReport(r);
                    addReport(resolveIceBroken(curPos));
                }
            }

            
            if (step.getType() == MoveStepType.LOAD) {

                
                Entity loaded = null;
                Iterator<Entity> entities = game.getEntities(curPos);
                while (entities.hasNext()) {

                    
                    loaded = entities.next();

                    
                    if (loaded.equals(null)) {
                        continue;
                    }

                    if (!entity.isEnemyOf(loaded) && !entity.equals(loaded)) {

                        
                        
                        if (!entity.canLoad(loaded)
                            || !loaded.isLoadableThisTurn()) {
                            
                            System.err.println(entity.getShortName()
                                               + " can not load " + loaded.getShortName());
                            loaded = null;
                        } else {
                            
                            loadUnit(entity, loaded, loaded.getTargetBay());

                            
                            break;
                        }

                    } else {
                        
                        loaded = null;
                    }

                } 

                
                if (loaded == null) {
                    System.err.println("Could not find unit for "
                                       + entity.getShortName() + " to load in " + curPos);
                }

            } 

            
            if (step.getType() == MoveStepType.MOUNT) {
                Targetable mountee = step.getTarget(game);
                if ((null != mountee) && (mountee instanceof Entity)) {
                    Entity dropship = (Entity) mountee;
                    if (!dropship.canLoad(entity)) {
                        
                        System.err.println(dropship.getShortName()
                                           + " can not load " + entity.getShortName());
                    } else {
                        
                        entity.setDone(true);
                        loadUnit(dropship, entity, entity.getTargetBay());
                        Bay currentBay = dropship.getBay(entity);
                        if ((null != currentBay) && (Compute.d6(2) == 2)) {
                            r = new Report(9390);
                            r.subject = entity.getId();
                            r.indent(1);
                            r.add(currentBay.getType());
                            addReport(r);
                            currentBay.destroyDoorNext();
                        }
                        
                        entityUpdate(dropship.getId());
                        return;
                    }
                }
            } 

            
            if (step.getType() == MoveStepType.RECOVER) {

                loader = game.getEntity(step.getRecoveryUnit());
                boolean isDS = (entity instanceof Dropship);

                PilotingRollData psr = entity
                        .getBasePilotingRoll(overallMoveType);
                if (loader.mpUsed > 0) {
                    psr.addModifier(5, "carrier used thrust");
                }
                int ctrlroll = Compute.d6(2);
                if (isDS) {
                    r = new Report(9388);
                } else {
                    r = new Report(9381);
                }
                r.subject = entity.getId();
                r.add(entity.getDisplayName());
                r.add(loader.getDisplayName());
                r.add(psr.getValue());
                r.add(ctrlroll);
                r.newlines = 0;
                r.indent(1);
                if (ctrlroll < psr.getValue()) {
                    r.choose(false);
                    addReport(r);
                    
                    Aero a = (Aero) entity;
                    HitData hit = a.rollHitLocation(ToHitData.HIT_NORMAL,
                                                    ToHitData.SIDE_FRONT);
                    addReport(damageEntity(entity, hit,
                                           2 * (psr.getValue() - ctrlroll)));
                } else {
                    r.choose(true);
                    addReport(r);
                    recovered = true;
                }
                
                if (ctrlroll == 2) {
                    loader.damageDoorRecovery(entity);
                    r = new Report(9384);
                    r.subject = entity.getId();
                    r.indent(0);
                    r.add(loader.getDisplayName());
                    addReport(r);
                }
            }

            
            if (step.getType() == MoveStepType.JOIN) {
                loader = game.getEntity(step.getRecoveryUnit());
                recovered = true;
            }

            
            if (step.getType() == MoveStepType.UNLOAD) {
                Targetable unloaded = step.getTarget(game);
                Coords unloadPos = curPos;
                int unloadFacing = curFacing;
                if (null != step.getTargetPosition()) {
                    unloadPos = step.getTargetPosition();
                    unloadFacing = curPos.direction(unloadPos);
                }
                if (!unloadUnit(entity, unloaded, unloadPos, unloadFacing,
                                step.getElevation())) {
                    System.err.println("Error! Server was told to unload "
                                       + unloaded.getDisplayName() + " from "
                                       + entity.getDisplayName() + " into "
                                       + curPos.getBoardNum());
                }
                
                
                if ((entity instanceof SmallCraft)
                    && (unloaded instanceof Entity)) {
                    Bay currentBay = entity.getBay((Entity) unloaded);
                    if ((null != currentBay) && (Compute.d6(2) == 2)) {
                        r = new Report(9390);
                        r.subject = entity.getId();
                        r.indent(1);
                        r.add(currentBay.getType());
                        addReport(r);
                        currentBay.destroyDoorNext();
                    }
                    
                    entity.resetBayDoors();
                    entityUpdate(entity.getId());
                    
                    
                    if (entity.getUnitsUnloadableFromBays().size() > 0) {
                        dropshipStillUnloading = true;
                        GameTurn newTurn = new GameTurn.SpecificEntityTurn(
                                entity.getOwner().getId(), entity.getId());
                        game.insertNextTurn(newTurn);
                    }
                    
                    
                    if (!(unloaded instanceof Infantry)) {
                        GameTurn newTurn = new GameTurn.SpecificEntityTurn(
                                ((Entity) unloaded).getOwner().getId(),
                                ((Entity) unloaded).getId());
                        game.insertNextTurn(newTurn);
                    }
                    
                    send(createTurnVectorPacket());
                }
            }
            
            if (((step.getType() == MoveStepType.BACKWARDS)
                    || (step.getType() == MoveStepType.LATERAL_LEFT_BACKWARDS) 
                    || (step.getType() == MoveStepType.LATERAL_RIGHT_BACKWARDS))
                    && !(md.isJumping() && (entity.getJumpType() == Mech.JUMP_BOOSTER))
                    && ((lastHex.getLevel() + entity.calcElevation(curHex,
                            lastHex, step.getElevation(),
                            md.getFinalClimbMode(), false)) != (curHex
                            .getLevel() + entity.getElevation()))
                    && !(entity instanceof VTOL)
                    && !(md.getFinalClimbMode()
                            && curHex.containsTerrain(Terrains.BRIDGE) && ((curHex
                            .terrainLevel(Terrains.BRIDGE_ELEV) + curHex
                            .getLevel()) == (prevHex.getLevel() + (prevHex
                            .containsTerrain(Terrains.BRIDGE) ? prevHex
                            .terrainLevel(Terrains.BRIDGE_ELEV) : 0))))) {

                if ((entity instanceof Mech)
                    && (curHex.getLevel() < game.getBoard()
                                                    .getHex(lastPos).getLevel())) {
                    PilotingRollData psr = entity
                            .getBasePilotingRoll(overallMoveType);
                    psr.addModifier(0,
                                    "moving backwards over an elevation change");
                    doSkillCheckWhileMoving(entity, entity.getElevation(),
                                            curPos, curPos, psr, true);
                } else if (entity instanceof Mech) {
                    PilotingRollData psr = entity
                            .getBasePilotingRoll(overallMoveType);
                    psr.addModifier(0,
                                    "moving backwards over an elevation change");
                    doSkillCheckWhileMoving(entity, lastElevation, lastPos,
                                            lastPos, psr, true);
                } else if (entity instanceof Tank) {
                    PilotingRollData psr = entity
                            .getBasePilotingRoll(overallMoveType);
                    psr.addModifier(0,
                                    "moving backwards over an elevation change");
                    if (doSkillCheckWhileMoving(entity, entity.getElevation(),
                                                curPos, lastPos, psr, false) < 0) {
                        curPos = lastPos;
                    }
                }

            }

            
            if (buildingMove > 0) {

                
                Building bldgExited = null;
                if ((buildingMove & 1) == 1) {
                    bldgExited = game.getBoard().getBuildingAt(lastPos);
                }

                
                Building bldgEntered = null;
                if ((buildingMove & 2) == 2) {
                    bldgEntered = game.getBoard().getBuildingAt(curPos);
                }

                
                if (((buildingMove & 8) == 8) && (entity instanceof Protomech)) {
                    Building bldg = game.getBoard().getBuildingAt(curPos);
                    Vector<Report> vBuildingReport = damageBuilding(bldg, 1,
                                                                    curPos);
                    for (Report report : vBuildingReport) {
                        report.subject = entity.getId();
                    }
                    addReport(vBuildingReport);
                }

                boolean collapsed = false;
                if ((bldgEntered != null)) {
                    
                    
                    String reason;
                    if (bldgExited == null) {
                        reason = "entering";
                    }
                    
                    
                    else if (bldgExited.equals(bldgEntered)
                             && !(entity instanceof Protomech)
                             && !(entity instanceof Infantry)) {
                        reason = "moving in";
                    }
                    
                    else {
                        reason = "entering";
                    }
                    passBuildingWall(entity, bldgEntered, lastPos, curPos,
                                     distance, reason, step.isThisStepBackwards(), step
                                    .getParent().getLastStepMovementType(),
                                     true);
                    addAffectedBldg(bldgEntered, collapsed);
                }

                
                if (entity.isDoomed()) {
                    entity.setDestroyed(true);
                    game.moveToGraveyard(entity.getId());
                    send(createRemoveEntityPacket(entity.getId()));

                    
                    return;
                }

                
            }

            if ((isOnGround || ((((step.getMovementType() != EntityMovementType.MOVE_JUMP) && (curElevation <= curHex
                    .terrainLevel(Terrains.BLDG_ELEV))) || (curElevation == curHex
                    .terrainLevel(Terrains.BRIDGE_ELEV))) && curHex
                                        .containsTerrain(Terrains.BLDG_ELEV)))) {
                Building bldg = game.getBoard().getBuildingAt(curPos);
                if ((bldg != null) && (entity.getElevation() >= 0)) {
                    addAffectedBldg(
                            bldg,
                            checkBuildingCollapseWhileMoving(bldg, entity,
                                                             curPos));
                }

            }

            
            if (!wasProne && entity.isProne()) {
                curFacing = entity.getFacing();
                curPos = entity.getPosition();
                mpUsed = step.getMpUsed();
                fellDuringMovement = true;
                break;
            }

            
            if (step.getType() == MoveStepType.GO_PRONE) {
                mpUsed = step.getMpUsed();
                rollTarget = entity.checkDislodgeSwarmers(step);
                if (rollTarget.getValue() == TargetRoll.CHECK_FALSE) {
                    
                    entity.setProne(true);
                    
                    checkForWashedInfernos(entity, curPos);
                } else {
                    
                    entity.setPosition(curPos);
                    if (doDislodgeSwarmerSkillCheck(entity, rollTarget, curPos)) {
                        
                        curFacing = entity.getFacing();
                        curPos = entity.getPosition();
                        fellDuringMovement = true;
                        break;
                    }
                    
                    entity.setProne(true);
                    
                    checkForWashedInfernos(entity, curPos);
                    break;
                }
            }

            
            if (step.getType() == MoveStepType.HULL_DOWN) {
                mpUsed = step.getMpUsed();
                entity.setHullDown(true);
            }

            
            for (Coords pos : step.getCrushedBuildingLocs()) {
                Building bldg = game.getBoard().getBuildingAt(pos);
                IHex hex = game.getBoard().getHex(pos);

                r = new Report(3443);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(bldg.getName());
                vPhaseReport.add(r);

                final int cf = bldg.getCurrentCF(pos);
                final int numFloors = Math.max(0,
                                               hex.terrainLevel(Terrains.BLDG_ELEV));
                vPhaseReport.addAll(damageBuilding(bldg, 150,
                                                   " is crushed for ", pos));
                int damage = (int) Math.round((cf / 10.0) * numFloors);
                HitData hit = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                     ToHitData.SIDE_FRONT);
                vPhaseReport.addAll(damageEntity(entity, hit, damage));
            }

            
            movePath.addElement(new UnitLocation(entity.getId(), curPos,
                                                 curFacing, step.getElevation()));

            
            
            
            if (!curPos.equals(lastPos)) {
                passedThrough.add(curPos);
                passedThroughFacing.add(curFacing);
            }

            
            if (!curPos.equals(lastPos)) {
                prevFacing = curFacing;
            }
            lastPos = new Coords(curPos);
            lastElevation = curElevation;
            prevStep = step;
            prevHex = curHex;

            firstStep = false;
        }

        
        entity.setPosition(curPos);
        entity.setFacing(curFacing);
        entity.setSecondaryFacing(curFacing);
        entity.delta_distance = distance;
        entity.moved = moveType;
        entity.mpUsed = mpUsed;
        if (!sideslipped && !fellDuringMovement && !crashedDuringMovement
            && (entity.getMovementMode() == EntityMovementMode.VTOL)) {
            entity.setElevation(curVTOLElevation);
        }
        entity.setAltitude(curAltitude);
        entity.setClimbMode(md.getFinalClimbMode());

        
        entity.setPassedThrough(passedThrough);
        entity.setPassedThroughFacing(passedThroughFacing);

        
        rollTarget = entity.checkRunningWithDamage(overallMoveType);
        if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
            doSkillCheckInPlace(entity, rollTarget);
        }

        
        rollTarget = entity.checkSprintingWithMASC(overallMoveType,
                                                   entity.mpUsed);
        if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
            doSkillCheckInPlace(entity, rollTarget);
        }

        
        
        if ((entity instanceof Protomech)
            && ((Protomech) entity).hasMyomerBooster()
            && (md.getMpUsed() > ((Protomech) entity)
                .getRunMPwithoutMyomerBooster(true, false, false))) {
            r = new Report(2373);
            r.addDesc(entity);
            r.subject = entity.getId();
            int roll = Compute.d6(2);
            r.add(roll);
            if (roll > 2) {
                r.choose(true);
                addReport(r);
            } else {
                r.choose(false);
                addReport(r);
                addReport(damageCrew(entity, 1));
            }
        }

        rollTarget = entity.checkSprintingWithSupercharger(overallMoveType,
                                                           entity.mpUsed);
        if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
            doSkillCheckInPlace(entity, rollTarget);
        }
        if ((md.getLastStepMovementType() == EntityMovementType.MOVE_SPRINT)
            && md.hasActiveMASC()) {
            doSkillCheckInPlace(entity,
                                entity.getBasePilotingRoll(EntityMovementType.MOVE_SPRINT));
        }

        if (entity.isAirborne() && (entity instanceof Aero)) {

            Aero a = (Aero) entity;
            int thrust = md.getMpUsed();

            
            if (((entity instanceof Aero) && game.getOptions().booleanOption(
                    "fuel_consumption"))
                || (entity instanceof TeleMissile)) {
                int fuelUsed = ((Aero) entity).getFuelUsed(thrust);
                a.useFuel(fuelUsed);
            }

            
            
            if (entity instanceof Jumpship) {
                Jumpship js = (Jumpship) entity;
                double penalty = 0.0;
                
                
                if (md.contains(MoveStepType.TURN_LEFT)
                    || md.contains(MoveStepType.TURN_RIGHT)) {
                    
                    
                    
                    penalty = js.getStationKeepingThrust();
                }
                if (thrust > 0) {
                    penalty = thrust;
                }
                if (penalty > 0.0) {
                    js.setAccumulatedThrust(Math.max(0,
                                                     js.getAccumulatedThrust() - penalty));
                }
            }

            

            rollTarget = a.checkThrustSITotal(thrust, overallMoveType);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                game.addControlRoll(new PilotingRollData(a.getId(), 0,
                                                         "Thrust spent during turn exceeds SI"));
            }

            if (!game.getBoard().inSpace()) {
                rollTarget = a.checkVelocityDouble(md.getFinalVelocity(),
                                                   overallMoveType);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    game.addControlRoll(new PilotingRollData(a.getId(), 0,
                                                             "Velocity greater than 2x safe thrust"));
                }

                rollTarget = a.checkDown(md.getFinalNDown(), overallMoveType);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    game.addControlRoll(new PilotingRollData(a.getId(), md
                            .getFinalNDown(),
                                                             "descended more than two altitudes"));
                }

                
                rollTarget = a.checkHover(md);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    game.addControlRoll(new PilotingRollData(a.getId(), 0,
                                                             "hovering"));
                }

                
                rollTarget = a.checkStall(md);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    r = new Report(9391);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    addReport(r);
                    game.addControlRoll(new PilotingRollData(entity.getId(), 0,
                                                             "stalled out"));
                    a.setAltitude(a.getAltitude() - 1);
                    
                    if (checkCrash(entity, entity.getPosition(),
                                   entity.getAltitude())) {
                        addReport(processCrash(entity, 0, entity.getPosition()));
                    }
                }

                
                if (a.isSpheroid() && !game.getBoard().inSpace()
                    && a.isAirborne() && (md.getFinalNDown() == 0)
                    && (md.getMpUsed() == 0)) {
                    r = new Report(9392);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    addReport(r);
                    a.setAltitude(a.getAltitude() - 1);
                    
                    if (checkCrash(entity, entity.getPosition(),
                                   entity.getAltitude())) {
                        addReport(processCrash(entity, 0, entity.getPosition()));
                    }
                }
            }
        }

        
        
        
        if ((entity instanceof Tank) && entity.isHullDown()
            && !md.getFinalHullDown()) {
            entity.setHullDown(false);
        }

        
        
        final int swarmerId = entity.getSwarmAttackerId();
        if ((Entity.NONE != swarmerId)
            && md.contains(MoveStepType.SHAKE_OFF_SWARMERS)) {
            final Entity swarmer = game.getEntity(swarmerId);
            final PilotingRollData roll = entity
                    .getBasePilotingRoll(overallMoveType);

            entity.addPilotingModifierForTerrain(roll);

            
            if (md.getLastStepMovementType() == EntityMovementType.MOVE_VTOL_RUN) {
                roll.addModifier(2,
                                 "dislodge swarming infantry with VTOL movement");
            } else {
                roll.addModifier(4, "dislodge swarming infantry");
            }

            
            
            for (Mounted mount : swarmer.getMisc()) {
                EquipmentType equip = mount.getType();
                if (equip.hasFlag(MiscType.F_MAGNET_CLAW)) {
                    roll.addModifier(1, "swarmer has magnetic claws");
                    break;
                }
            }

            
            r = new Report(2125);
            r.subject = entity.getId();
            r.addDesc(entity);
            addReport(r);

            
            final int diceRoll = Compute.d6(2);
            r = new Report(2130);
            r.subject = entity.getId();
            r.add(roll.getValueAsString());
            r.add(roll.getDesc());
            r.add(diceRoll);
            if (diceRoll < roll.getValue()) {
                r.choose(false);
                addReport(r);
            } else {
                
                game.removeTurnFor(swarmer);
                send(createTurnVectorPacket());

                
                r.choose(true);
                addReport(r);
                entity.setSwarmAttackerId(Entity.NONE);
                swarmer.setSwarmTargetId(Entity.NONE);

                IHex curHex = game.getBoard().getHex(curPos);

                
                if (curHex.terrainLevel(Terrains.WATER) > 0) {
                    
                    swarmer.setPosition(curPos);
                    r = new Report(2135);
                    r.subject = entity.getId();
                    r.indent();
                    r.addDesc(swarmer);
                    addReport(r);
                    addReport(destroyEntity(swarmer, "a watery grave", false));
                } else {
                    
                    
                    r = new Report(2140);
                    r.subject = entity.getId();
                    r.indent();
                    r.addDesc(swarmer);
                    r.add("3d6");
                    addReport(r);
                    addReport(damageEntity(swarmer, swarmer.rollHitLocation(
                                                   ToHitData.HIT_NORMAL, ToHitData.SIDE_FRONT),
                                           Compute.d6(3)));
                    addNewLines();
                    swarmer.setPosition(curPos);
                }
                entityUpdate(swarmerId);
            } 

        } 

        
        if ((overallMoveType == EntityMovementType.MOVE_JUMP)
            && !entity.isMakingDfa()) {
            final IHex curHex = game.getBoard().getHex(curPos);
            
            rollTarget = entity.checkLandingWithDamage(overallMoveType);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                doSkillCheckInPlace(entity, rollTarget);
            }
            
            rollTarget = entity.checkLandingWithPrototypeJJ(overallMoveType);
            if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                doSkillCheckInPlace(entity, rollTarget);
            }
            
            if (game.getOptions().booleanOption("psr_jump_heavy_woods")) {
                rollTarget = entity.checkLandingInHeavyWoods(overallMoveType,
                                                             curHex);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    doSkillCheckInPlace(entity, rollTarget);
                }
            }
            
            if (md.shouldMechanicalJumpCauseFallDamage()) {
                vPhaseReport.addAll(doEntityFallsInto(entity,
                                                      entity.getElevation(), md.getJumpPathHighestPoint(),
                                                      curPos, entity.getBasePilotingRoll(overallMoveType),
                                                      false, entity.getJumpMP()));
            }
            
            int waterLevel = curHex.terrainLevel(Terrains.WATER);
            if (curHex.containsTerrain(Terrains.ICE) && (waterLevel > 0)) {
                if (!(entity instanceof Infantry)) {
                    waterLevel = 0;
                    
                    int roll = Compute.d6(1);
                    r = new Report(2122);
                    r.add(entity.getDisplayName(), true);
                    r.add(roll);
                    r.subject = entity.getId();
                    addReport(r);
                    if (roll >= 4) {
                        
                        entity.setPosition(curPos);
                        addReport(resolveIceBroken(curPos));
                        curPos = entity.getPosition();
                    } else {
                        
                        
                        rollTarget = entity.checkLandingOnIce(overallMoveType,
                                                              curHex);
                        if (!doSkillCheckInPlace(entity, rollTarget)) {
                            
                            
                            entity.applyDamage();
                            roll = Compute.d6(1);
                            r = new Report(2118);
                            r.addDesc(entity);
                            r.add(roll);
                            r.subject = entity.getId();
                            addReport(r);
                            if (roll == 6) {
                                entity.setPosition(curPos);
                                addReport(resolveIceBroken(curPos));
                                curPos = entity.getPosition();
                            }
                        }
                    }
                }
            } else if (!(prevStep.climbMode() && curHex
                    .containsTerrain(Terrains.BRIDGE))
                       && !(entity.getMovementMode() == EntityMovementMode.HOVER)) {
                rollTarget = entity.checkWaterMove(waterLevel, overallMoveType);
                if (rollTarget.getValue() != TargetRoll.CHECK_FALSE) {
                    doSkillCheckInPlace(entity, rollTarget);
                }
                if (waterLevel > 1) {
                    
                    drownSwarmer(entity, curPos);
                }
            }

            
            Building bldg = game.getBoard().getBuildingAt(curPos);
            if (bldg != null) {
                checkForCollapse(bldg, game.getPositionMap(), curPos, true,
                                 vPhaseReport);
            }

            
            if (curHex.terrainLevel(Terrains.MAGMA) == 1) {
                int roll = Compute.d6(1);
                r = new Report(2395);
                r.addDesc(entity);
                r.add(roll);
                r.subject = entity.getId();
                addReport(r);
                if (roll == 6) {
                    curHex.removeTerrain(Terrains.MAGMA);
                    curHex.addTerrain(Terrains.getTerrainFactory()
                                              .createTerrain(Terrains.MAGMA, 2));
                    sendChangedHex(curPos);
                    for (Entity en : game.getEntitiesVector(curPos)) {
                        if (en != entity) {
                            doMagmaDamage(en, false);
                        }
                    }
                }
            }

            
            if (curHex.terrainLevel(Terrains.MAGMA) == 2) {
                doMagmaDamage(entity, false);
            }

            
            if (curHex.getBogDownModifier(entity.getMovementMode(),
                                          entity instanceof LargeSupportTank) != TargetRoll.AUTOMATIC_SUCCESS) {
                if (entity instanceof Mech) {
                    entity.setStuck(true);
                    r = new Report(2121);
                    r.add(entity.getDisplayName(), true);
                    r.subject = entity.getId();
                    addReport(r);
                    
                    addReport(checkQuickSand(curPos));
                } else {
                    PilotingRollData roll = new PilotingRollData(
                            entity.getId(), 5, "entering boggy terrain");
                    roll.append(new PilotingRollData(entity.getId(), curHex
                            .getBogDownModifier(entity.getMovementMode(),
                                                entity instanceof LargeSupportTank),
                                                     "avoid bogging down"));
                    if (0 < doSkillCheckWhileMoving(entity,
                                                    entity.getElevation(), curPos, curPos, roll, false)) {
                        entity.setStuck(true);
                        r = new Report(2081);
                        r.add(entity.getDisplayName());
                        r.subject = entity.getId();
                        addReport(r);
                        
                        addReport(checkQuickSand(curPos));
                    }
                }
            }

            
            if (Entity.NONE != swarmerId) {
                final Entity swarmer = game.getEntity(swarmerId);
                final PilotingRollData roll = entity
                        .getBasePilotingRoll(overallMoveType);

                entity.addPilotingModifierForTerrain(roll);

                
                roll.addModifier(4, "dislodge swarming infantry");

                
                
                if (swarmer.hasWorkingMisc(MiscType.F_MAGNET_CLAW, -1)) {
                    roll.addModifier(1, "swarmer has magnetic claws");
                }

                
                r = new Report(2125);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);

                
                final int diceRoll = Compute.d6(2);
                r = new Report(2130);
                r.subject = entity.getId();
                r.add(roll.getValueAsString());
                r.add(roll.getDesc());
                r.add(diceRoll);
                if (diceRoll < roll.getValue()) {
                    r.choose(false);
                    addReport(r);
                } else {
                    
                    game.removeTurnFor(swarmer);
                    send(createTurnVectorPacket());

                    
                    r.choose(true);
                    addReport(r);
                    entity.setSwarmAttackerId(Entity.NONE);
                    swarmer.setSwarmTargetId(Entity.NONE);

                    
                    if (curHex.terrainLevel(Terrains.WATER) > 0) {
                        
                        swarmer.setPosition(curPos);
                        r = new Report(2135);
                        r.subject = entity.getId();
                        r.indent();
                        r.addDesc(swarmer);
                        addReport(r);
                        addReport(destroyEntity(swarmer, "a watery grave",
                                                false));
                    } else {
                        
                        
                        r = new Report(2140);
                        r.subject = entity.getId();
                        r.indent();
                        r.addDesc(swarmer);
                        r.add("3d6");
                        addReport(r);
                        addReport(damageEntity(swarmer,
                                               swarmer.rollHitLocation(ToHitData.HIT_NORMAL,
                                                                       ToHitData.SIDE_FRONT), Compute.d6(3)));
                        addNewLines();
                        swarmer.setPosition(curPos);
                    }
                    entityUpdate(swarmerId);
                } 

            } 

            
            checkForWashedInfernos(entity, curPos);

            
            if (entity instanceof Tank) {
                int modifier = 0;
                if (curHex.containsTerrain(Terrains.ROUGH)
                    || curHex.containsTerrain(Terrains.WOODS)
                    || curHex.containsTerrain(Terrains.JUNGLE)) {
                    modifier = 1;
                }
                r = new Report(2126);
                r.subject = entity.getId();
                r.addDesc(entity);
                vPhaseReport.add(r);
                vPhaseReport.addAll(vehicleMotiveDamage((Tank) entity,
                                                        modifier, false, -1, true));
                Report.addNewline(vPhaseReport);
            }

        } 
        
        vPhaseReport.addAll(doSetLocationsExposure(entity, game.getBoard()
                                                               .getHex(curPos), false, entity.getElevation()));

        
        
        
        
        
        if (!(game.getOptions().booleanOption("falls_end_movement")
              && (entity instanceof Mech) && !wasProne && entity.isProne())
            && fellDuringMovement
            && !turnOver
            && (entity.mpUsed < entity.getRunMP())
            && entity.isSelectableThisTurn() && !entity.isDoomed()) {
            entity.applyDamage();
            entity.setDone(false);
            GameTurn currentTurn  = game.getTurn();
            if (game.isPhaseSimultaneous()) {
                currentTurn = game.getTurnForPlayer(entity.getOwnerId());
            }
            GameTurn newTurn = new GameTurn.SpecificEntityTurn(entity
                    .getOwner().getId(), entity.getId());
            
            newTurn.setMultiTurn(currentTurn.isMultiTurn());
            game.insertNextTurn(newTurn);
            
            send(createTurnVectorPacket());
            
            send(entity.getOwner().getId(), createSpecialReportPacket());
        } else {
            if ((entity.getMovementMode() == EntityMovementMode.WIGE)
                && (entity.getElevation() > 0)) {
                if (!wigeStartedLanded && (entity.delta_distance < 5)) {
                    
                    r = new Report(2123);
                    r.addDesc(entity);
                    r.subject = entity.getId();
                    vPhaseReport.add(r);
                    
                    IHex hex = game.getBoard().getHex(curPos);
                    if (!hex.hasPavement() && (hex.terrainsPresent() > 0)) {
                        
                        r = new Report(2124);
                        r.addDesc(entity);
                        r.subject = entity.getId();
                        vPhaseReport.add(r);
                        vPhaseReport.addAll(crashVTOLorWiGE((Tank) entity));
                    } else {
                        entity.setElevation(0);
                    }
                } else {
                    
                    
                    
                    
                    
                    IHex hex = game.getBoard().getHex(curPos);
                    entity.setElevation(1 + hex.maxTerrainFeatureElevation(game
                            .getBoard().inAtmosphere()));
                }
            }
            entity.setDone(true);
        }

        if (dropshipStillUnloading) {
            
            
            entity.setDone(false);
        }

        
        if (Entity.NONE != swarmerId) {
            final Entity swarmer = game.getEntity(swarmerId);
            swarmer.setPosition(curPos);
            
            
            if (!(swarmer instanceof BattleArmor)
                && game.getBoard().getHex(curPos)
                       .containsTerrain(Terrains.FIRE)) {
                swarmer.setSwarmTargetId(Entity.NONE);
                entity.setSwarmAttackerId(Entity.NONE);
                r = new Report(2145);
                r.subject = entity.getId();
                r.indent();
                r.add(swarmer.getShortName(), true);
                addReport(r);
            }
            entityUpdate(swarmerId);
        }

        
        
        if (!game.isOutOfGame(entity)) {
            entityUpdate(entity.getId(), movePath, true, losCache);
            if (entity.isDoomed()) {
                send(createRemoveEntityPacket(entity.getId(),
                                              entity.getRemovalCondition()));
            }
        }

        
        if ((entity instanceof Aero) && recovered && (loader != null)) {

            if (loader.isCapitalFighter()) {
                if (!(loader instanceof FighterSquadron)) {
                    
                    
                    FighterSquadron fs = new FighterSquadron();
                    fs.setDeployed(true);
                    fs.setId(getFreeEntityId());
                    fs.setCurrentVelocity(((Aero) loader).getCurrentVelocity());
                    fs.setNextVelocity(((Aero) loader).getNextVelocity());
                    fs.setVectors(loader.getVectors());
                    fs.setFacing(loader.getFacing());
                    fs.setOwner(entity.getOwner());
                    
                    game.addEntity(fs);
                    send(createAddEntityPacket(fs.getId()));
                    
                    fs.setDone(true);
                    
                    fs.setPosition(loader.getPosition());
                    loadUnit(fs, loader, -1);
                    loader = fs;
                    entityUpdate(fs.getId());
                }
                loader.load(entity);
            } else {
                loader.recover(entity);
                entity.setRecoveryTurn(5);
            }

            
            entity.setTransportId(loader.getId());

            
            entity.setPosition(null);

            
            entityUpdate(entity.getId());
        }

        
        if (null != loader) {
            entityUpdate(loader.getId());
        }

        
        if (doBlind()) {
            send(entity.getOwner().getId(),
                 createFilteredEntitiesPacket(entity.getOwner(), losCache));
        }

        
        if (charge != null) {
            send(createAttackPacket(charge, 1));
        }

        
        if (ram != null) {
            send(createAttackPacket(ram, 1));
        }
        if ((entity instanceof Mech)
            && ((Mech) entity).isIndustrial()
            && !((Mech) entity).hasEnvironmentalSealing()
            && (entity.getEngine().getEngineType() == Engine.COMBUSTION_ENGINE)) {
            if ((!entity.isProne() && (game.getBoard()
                                           .getHex(entity.getPosition()).terrainLevel(Terrains.WATER) >= 2))
                || (entity.isProne() && (game.getBoard()
                                             .getHex(entity.getPosition())
                                             .terrainLevel(Terrains.WATER) == 1))) {
                ((Mech) entity).setJustMovedIntoIndustrialKillingWater(true);

            } else {
                ((Mech) entity).setJustMovedIntoIndustrialKillingWater(false);
            }
        }
    }

    
    private void checkForTakeoffDamage(Aero a) {

        boolean unsecured = false;
        for (Entity loaded : a.getLoadedUnits()) {
            if (loaded.wasLoadedThisTurn() && !(loaded instanceof Infantry)) {
                unsecured = true;
                
                Report r = new Report(6800);
                r.subject = loaded.getId();
                r.addDesc(loaded);
                addReport(r);
                int damage = 25;
                ToHitData toHit = new ToHitData();
                while (damage > 0) {
                    HitData hit = loaded.rollHitLocation(toHit.getHitTable(),
                                                         ToHitData.SIDE_FRONT);
                    addReport(damageEntity(loaded, hit, 5, false,
                                           DamageType.NONE, false, true, false));
                    damage -= 5;
                }
            }
        }
        if (unsecured) {
            
            HitData hit = a.rollHitLocation(ToHitData.HIT_ABOVE,
                                            ToHitData.SIDE_FRONT);
            addReport(applyCriticalHit(a, hit.getLocation(), new CriticalSlot(
                    0, a.getPotCrit()), true, 1, false));
        }

    }

    
    public void deliverThunderAugMinefield(Coords coords, int playerId,
                                           int damage, int entityId) {
        Coords mfCoord = null;
        for (int dir = 0; dir < 7; dir++) {
            
            int hexDamage = damage;
            switch (dir) {
                case 6:
                    
                    mfCoord = new Coords(coords);
                    break;
                default:
                    
                    mfCoord = coords.translated(dir);
                    break;
            }

            
            if (game.getBoard().contains(mfCoord)) {
                Minefield minefield = null;
                Enumeration<Minefield> minefields = game.getMinefields(mfCoord)
                                                        .elements();
                
                while (minefields.hasMoreElements()) {
                    Minefield mf = minefields.nextElement();
                    if (mf.getType() == Minefield.TYPE_CONVENTIONAL) {
                        minefield = mf;
                        break;
                    }
                }

                
                
                
                if (minefield == null) {
                    
                    minefield = Minefield.createMinefield(mfCoord, playerId,
                                                          Minefield.TYPE_CONVENTIONAL, hexDamage);
                    game.addMinefield(minefield);
                    checkForRevealMinefield(minefield, game.getEntity(entityId));
                } else if (minefield.getDensity() < Minefield.MAX_DAMAGE) {
                    
                    removeMinefield(minefield);
                    hexDamage += minefield.getDensity();

                    
                    if (hexDamage > Minefield.MAX_DAMAGE) {
                        hexDamage = Minefield.MAX_DAMAGE;
                    }
                    minefield.setDensity(hexDamage);
                    game.addMinefield(minefield);
                    checkForRevealMinefield(minefield, game.getEntity(entityId));
                }
            } 

        } 

    }

    
    public void deliverThunderMinefield(Coords coords, int playerId,
                                        int damage, int entityId) {
        Minefield minefield = null;
        Enumeration<Minefield> minefields = game.getMinefields(coords)
                                                .elements();
        
        while (minefields.hasMoreElements()) {
            Minefield mf = minefields.nextElement();
            if (mf.getType() == Minefield.TYPE_CONVENTIONAL) {
                minefield = mf;
                break;
            }
        }

        
        if (minefield == null) {
            minefield = Minefield.createMinefield(coords, playerId,
                                                  Minefield.TYPE_CONVENTIONAL, damage);
            game.addMinefield(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        } else if (minefield.getDensity() < Minefield.MAX_DAMAGE) {
            
            removeMinefield(minefield);
            int oldDamage = minefield.getDensity();
            damage += oldDamage;
            damage = damage > Minefield.MAX_DAMAGE ? Minefield.MAX_DAMAGE
                                                   : damage;
            minefield.setDensity(damage);
            game.addMinefield(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        }
    }

    
    public void deliverThunderInfernoMinefield(Coords coords, int playerId,
                                               int damage, int entityId) {
        Minefield minefield = null;
        Enumeration<Minefield> minefields = game.getMinefields(coords)
                                                .elements();
        
        while (minefields.hasMoreElements()) {
            Minefield mf = minefields.nextElement();
            if (mf.getType() == Minefield.TYPE_INFERNO) {
                minefield = mf;
                break;
            }
        }

        
        if (minefield == null) {
            minefield = Minefield.createMinefield(coords, playerId,
                                                  Minefield.TYPE_INFERNO, damage);
            game.addMinefield(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        } else if (minefield.getDensity() < Minefield.MAX_DAMAGE) {
            
            removeMinefield(minefield);
            int oldDamage = minefield.getDensity();
            damage += oldDamage;
            damage = damage > Minefield.MAX_DAMAGE ? Minefield.MAX_DAMAGE
                                                   : damage;
            minefield.setDensity(damage);
            game.addMinefield(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        }
    }

    
    public void deliverFASCAMMinefield(Coords coords, int playerId, int damage,
                                       int entityId) {
        
        if (game.getBoard().contains(coords)) {
            Minefield minefield = null;
            Enumeration<Minefield> minefields = game.getMinefields(coords)
                                                    .elements();
            
            while (minefields.hasMoreElements()) {
                Minefield mf = minefields.nextElement();
                if (mf.getType() == Minefield.TYPE_CONVENTIONAL) {
                    minefield = mf;
                    break;
                }
            }
            
            if (minefield == null) {
                minefield = Minefield.createMinefield(coords, playerId,
                                                      Minefield.TYPE_CONVENTIONAL, damage);
                game.addMinefield(minefield);
                checkForRevealMinefield(minefield, game.getEntity(entityId));
            } else if (minefield.getDensity() < Minefield.MAX_DAMAGE) {
                
                removeMinefield(minefield);
                int oldDamage = minefield.getDensity();
                damage += oldDamage;
                damage = damage > Minefield.MAX_DAMAGE ? Minefield.MAX_DAMAGE
                                                       : damage;
                minefield.setDensity(damage);
                game.addMinefield(minefield);
                checkForRevealMinefield(minefield, game.getEntity(entityId));
            }
        } 
    }

    
    public void deliverThunderActiveMinefield(Coords coords, int playerId,
                                              int damage, int entityId) {
        Minefield minefield = null;
        Enumeration<Minefield> minefields = game.getMinefields(coords)
                                                .elements();
        
        while (minefields.hasMoreElements()) {
            Minefield mf = minefields.nextElement();
            if (mf.getType() == Minefield.TYPE_ACTIVE) {
                minefield = mf;
                break;
            }
        }

        
        if (minefield == null) {
            minefield = Minefield.createMinefield(coords, playerId,
                                                  Minefield.TYPE_ACTIVE, damage);
            game.addMinefield(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        } else if (minefield.getDensity() < Minefield.MAX_DAMAGE) {
            
            removeMinefield(minefield);
            int oldDamage = minefield.getDensity();
            damage += oldDamage;
            damage = damage > Minefield.MAX_DAMAGE ? Minefield.MAX_DAMAGE
                                                   : damage;
            minefield.setDensity(damage);
            game.addMinefield(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        }
    }

    
    public void deliverThunderVibraMinefield(Coords coords, int playerId,
                                             int damage, int sensitivity, int entityId) {
        Minefield minefield = null;
        Enumeration<Minefield> minefields = game.getMinefields(coords)
                                                .elements();
        
        while (minefields.hasMoreElements()) {
            Minefield mf = minefields.nextElement();
            if (mf.getType() == Minefield.TYPE_VIBRABOMB) {
                minefield = mf;
                break;
            }
        }

        
        if (minefield == null) {
            minefield = Minefield.createMinefield(coords, playerId,
                                                  Minefield.TYPE_VIBRABOMB, damage, sensitivity);
            game.addVibrabomb(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        } else if (minefield.getDensity() < Minefield.MAX_DAMAGE) {
            
            removeMinefield(minefield);
            int oldDamage = minefield.getDensity();
            damage += oldDamage;
            damage = damage > Minefield.MAX_DAMAGE ? Minefield.MAX_DAMAGE
                                                   : damage;
            minefield.setDensity(damage);
            game.addVibrabomb(minefield);
            checkForRevealMinefield(minefield, game.getEntity(entityId));
        }
    }

    
    public void deliverArtilleryFlare(Coords coords, int radius) {
        Flare flare = new Flare(coords, 5, radius, Flare.F_DRIFTING);
        game.addFlare(flare);
    }

    public void deliverMortarFlare(Coords coords, int duration) {
        Flare flare = new Flare(coords, duration, 1, Flare.F_IGNITED);
        game.addFlare(flare);
    }

    public void deliverMissileSmoke(Coords coords, Vector<Report> vPhaseReport) {
        Report r = new Report(5185, Report.PUBLIC);
        r.indent(2);
        r.add(coords.getBoardNum());
        vPhaseReport.add(r);
        createSmoke(coords, SmokeCloud.SMOKE_HEAVY, 3);
        sendChangedHex(coords);
    }

    public void deliverSmokeGrenade(Coords coords, Vector<Report> vPhaseReport) {
        Report r = new Report(5185, Report.PUBLIC);
        r.indent(2);
        r.add(coords.getBoardNum());
        vPhaseReport.add(r);
        createSmoke(coords, SmokeCloud.SMOKE_LIGHT, 3);
        sendChangedHex(coords);
    }

    public void deliverSmokeMortar(Coords coords, Vector<Report> vPhaseReport,
            int duration) {
        Report r = new Report(5185, Report.PUBLIC);
        r.indent(2);
        r.add(coords.getBoardNum());
        vPhaseReport.add(r);
        createSmoke(coords, SmokeCloud.SMOKE_HEAVY, duration);
        sendChangedHex(coords);
    }

    public void deliverChaffGrenade(Coords coords, Vector<Report> vPhaseReport) {
        Report r = new Report(5186, Report.PUBLIC);
        r.indent(2);
        r.add(coords.getBoardNum());
        vPhaseReport.add(r);
        createSmoke(coords, SmokeCloud.SMOKE_CHAFF_LIGHT, 1);
        sendChangedHex(coords);
    }

    
    public void deliverArtillerySmoke(Coords coords, Vector<Report> vPhaseReport) {
        Report r = new Report(5185, Report.PUBLIC);
        r.indent(2);
        r.add(coords.getBoardNum());
        vPhaseReport.add(r);
        createSmoke(coords, SmokeCloud.SMOKE_HEAVY, 3);
        sendChangedHex(coords);
        for (int dir = 0; dir <= 5; dir++) {
            Coords tempcoords = coords.translated(dir);
            if (!game.getBoard().contains(tempcoords)) {
                continue;
            }
            if (coords.equals(tempcoords)) {
                continue;
            }
            r = new Report(5185, Report.PUBLIC);
            r.indent(2);
            r.add(tempcoords.getBoardNum());
            vPhaseReport.add(r);
            createSmoke(tempcoords, 2, 3);
            sendChangedHex(tempcoords);
        }
    }

    public void deliverLIsmoke(Coords coords, Vector<Report> vPhaseReport) {
        Report r = new Report(5186, Report.PUBLIC);
        r.indent(2);
        r.add(coords.getBoardNum());
        vPhaseReport.add(r);
        createSmoke(coords, SmokeCloud.SMOKE_LI_HEAVY, 2);
        sendChangedHex(coords);
        for (int dir = 0; dir <= 5; dir++) {
            Coords tempcoords = coords.translated(dir);
            if (!game.getBoard().contains(tempcoords)) {
                continue;
            }
            if (coords.equals(tempcoords)) {
                continue;
            }
            r = new Report(5186, Report.PUBLIC);
            r.indent(2);
            r.add(tempcoords.getBoardNum());
            vPhaseReport.add(r);
            createSmoke(tempcoords, SmokeCloud.SMOKE_LI_HEAVY, 2);
            sendChangedHex(tempcoords);
        }
    }

    public void deliverArtilleryInferno(Coords coords, Entity ae,
                                        int subjectId, Vector<Report> vPhaseReport) {
        IHex h = game.getBoard().getHex(coords);
        Report r;
        
        if (h.terrainLevel(Terrains.FIRE) < Terrains.FIRE_LVL_INFERNO_IV) {
            ignite(coords, Terrains.FIRE_LVL_INFERNO_IV, vPhaseReport);
        }
        
        if (h.containsTerrain(Terrains.ICE) || h.containsTerrain(Terrains.SNOW)) {
            vPhaseReport.addAll(meltIceAndSnow(coords, subjectId));
        }
        for (Entity entity : game.getEntitiesVector(coords)) {
            
            r = new Report(6695);
            r.indent(3);
            r.add(entity.getDisplayName());
            r.subject = entity.getId();
            r.newlines = 0;
            vPhaseReport.add(r);
            if (entity instanceof Tank) {
                Report.addNewline(vPhaseReport);
            }
            Vector<Report> vDamageReport = deliverInfernoMissiles(ae, entity, 5);
            Report.indentAll(vDamageReport, 2);
            vPhaseReport.addAll(vDamageReport);
        }
        for (int dir = 0; dir <= 5; dir++) {
            Coords tempcoords = coords.translated(dir);
            if (!game.getBoard().contains(tempcoords)) {
                continue;
            }
            if (coords.equals(tempcoords)) {
                continue;
            }
            h = game.getBoard().getHex(tempcoords);
            
            if (h.terrainLevel(Terrains.FIRE) < Terrains.FIRE_LVL_INFERNO_IV) {
                ignite(tempcoords, Terrains.FIRE_LVL_INFERNO_IV, vPhaseReport);
            }
            
            if (h.containsTerrain(Terrains.ICE)
                || h.containsTerrain(Terrains.SNOW)) {
                vPhaseReport.addAll(meltIceAndSnow(tempcoords, subjectId));
            }
            for (Entity entity : game.getEntitiesVector(tempcoords)) {
                r = new Report(6695);
                r.indent(3);
                r.add(entity.getDisplayName());
                r.newlines = 0;
                r.subject = entity.getId();
                vPhaseReport.add(r);
                if (entity instanceof Tank) {
                    Report.addNewline(vPhaseReport);
                }
                Vector<Report> vDamageReport = deliverInfernoMissiles(ae,
                        entity, 5);
                Report.indentAll(vDamageReport, 2);
                vPhaseReport.addAll(vDamageReport);
            }
        }
    }

    public void deliverScreen(Coords coords, Vector<Report> vPhaseReport) {
        IHex h = game.getBoard().getHex(coords);
        Report r;
        Report.addNewline(vPhaseReport);
        r = new Report(9070, Report.PUBLIC);
        r.indent(2);
        r.add(coords.getBoardNum());
        vPhaseReport.add(r);
        
        
        int nscreens = h.terrainLevel(Terrains.SCREEN);
        if (nscreens > 0) {
            h.removeTerrain(Terrains.SCREEN);
            h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                    Terrains.SCREEN, nscreens + 1));
        } else {
            h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                    Terrains.SCREEN, 1));
        }
        sendChangedHex(coords);
    }

    
    public void deployTeleMissile(Entity ae, AmmoType atype, int wId,
                                  int capMisMod, Vector<Report> vPhaseReport) {
        Report r = new Report(9080);
        r.subject = ae.getId();
        r.addDesc(ae);
        r.indent(2);
        r.newlines = 0;
        r.add(atype.getName());
        vPhaseReport.add(r);
        TeleMissile tele = new TeleMissile(ae, atype.getDamagePerShot(),
                                           atype.getTonnage(ae), atype.getAmmoType(), capMisMod);
        tele.setDeployed(true);
        tele.setId(getFreeEntityId());
        if (ae instanceof Aero) {
            Aero a = (Aero) ae;
            tele.setCurrentVelocity(a.getCurrentVelocity());
            tele.setNextVelocity(a.getNextVelocity());
            tele.setVectors(a.getVectors());
            tele.setFacing(a.getFacing());
        }
        
        game.addEntity(tele);
        send(createAddEntityPacket(tele.getId()));
        
        tele.setDone(true);
        
        tele.setPosition(ae.getPosition());
        
        entityUpdate(tele.getId());
        
        
        if (ae.getTMTracker().containsLauncher(wId)) {
            Entity priorMissile = game.getEntity(ae.getTMTracker().getMissile(
                    wId));
            if ((null != priorMissile) && (priorMissile instanceof TeleMissile)) {
                ((TeleMissile) priorMissile).setOutContact(true);
                
                ae.getTMTracker().removeMissile(wId);
            }
        }
        
        ae.getTMTracker().addMissile(wId, tele.getId());
    }

    
    public Vector<Report> deliverInfernoMissiles(Entity ae, Targetable t,
                                                 int missiles) {
        return deliverInfernoMissiles(ae, t, missiles, CalledShot.CALLED_NONE);
    }

    
    public Vector<Report> deliverInfernoMissiles(Entity ae, Targetable t,
                                                 int missiles, int called) {
        IHex hex = game.getBoard().getHex(t.getPosition());
        Report r;
        Vector<Report> vPhaseReport = new Vector<Report>();
        switch (t.getTargetType()) {
            case Targetable.TYPE_HEX_ARTILLERY:
                
                for (Entity e : game.getEntitiesVector(t.getPosition())) {
                    if (e.getElevation() > hex.terrainLevel(Terrains.BLDG_ELEV)) {
                        r = new Report(6685);
                        r.subject = e.getId();
                        r.addDesc(e);
                        vPhaseReport.add(r);
                        vPhaseReport.addAll(deliverInfernoMissiles(ae, e, missiles,
                                                                   called));
                    } else {
                        int roll = Compute.d6();
                        r = new Report(3570);
                        r.subject = e.getId();
                        r.addDesc(e);
                        r.add(roll);
                        vPhaseReport.add(r);
                        if (roll >= 5) {
                            vPhaseReport.addAll(deliverInfernoMissiles(ae, e,
                                                                       missiles, called));
                        }
                    }
                }
                if (game.getBoard().getBuildingAt(t.getPosition()) != null) {
                    Vector<Report> vBuildingReport = damageBuilding(game.getBoard()
                                                                        .getBuildingAt(t.getPosition()), 2 * missiles,
                                                                    t.getPosition());
                    for (Report report : vBuildingReport) {
                        report.subject = ae.getId();
                    }
                    vPhaseReport.addAll(vBuildingReport);
                }
                
            case Targetable.TYPE_HEX_CLEAR:
            case Targetable.TYPE_HEX_IGNITE:
                
                IHex h = game.getBoard().getHex(t.getPosition());
                if ((h != null) && h.hasTerrainfactor()) {
                    r = new Report(3384);
                    r.indent(2);
                    r.subject = ae.getId();
                    r.add(t.getPosition().getBoardNum());
                    r.add(missiles * 4);
                    vPhaseReport.addElement(r);
                }
                vPhaseReport.addAll(tryClearHex(t.getPosition(), missiles * 4,
                                                ae.getId()));
                tryIgniteHex(t.getPosition(), ae.getId(), false, true,
                             new TargetRoll(0, "inferno"), -1, vPhaseReport);
                break;
            case Targetable.TYPE_BLDG_IGNITE:
            case Targetable.TYPE_BUILDING:
                Vector<Report> vBuildingReport = damageBuilding(game.getBoard()
                                                                    .getBuildingAt(t.getPosition()), 2 * missiles,
                                                                t.getPosition());
                for (Report report : vBuildingReport) {
                    report.subject = ae.getId();
                }
                vPhaseReport.addAll(vBuildingReport);

                
                for (Entity e : game.getEntitiesVector(t.getPosition())) {
                    if (e.getElevation() > hex.terrainLevel(Terrains.BLDG_ELEV)) {
                        continue;
                    }
                    for (int m = 0; m < missiles; m++) {
                        int roll = Compute.d6();
                        r = new Report(3570);
                        r.subject = e.getId();
                        r.indent(3);
                        r.addDesc(e);
                        r.add(roll);
                        vPhaseReport.add(r);
                        if (roll >= 5) {
                            Vector<Report> dmgReports = deliverInfernoMissiles(ae,
                                    e, 1, called);
                            for (Report rep : dmgReports) {
                                rep.indent(4);
                            }
                            vPhaseReport.addAll(dmgReports);
                        }
                    }
                }

                break;
            case Targetable.TYPE_ENTITY:
                Entity te = (Entity) t;
                if (te instanceof Mech) {
                    
                    int m = missiles;
                    LosEffects le = LosEffects.calculateLos(game, ae.getId(), t);
                    int cover = le.getTargetCover();
                    Vector<Report> coverDamageReports = new Vector<Report>();
                    int heatDamage = 0;
                    boolean heatReduced = false;
                    String reductionCause = "";
                    for (int i = 0; i < m; i++) {
                        int side = Compute.targetSideTable(ae, t, called);
                        HitData hit = te
                                .rollHitLocation(ToHitData.HIT_NORMAL, side);
                        if (te.removePartialCoverHits(hit.getLocation(), cover,
                                                      side)) {
                            missiles--;
                            
                            int damagableCoverType = LosEffects.DAMAGABLE_COVER_NONE;
                            Entity coverDropship = null;
                            Coords coverLoc = null;

                            
                            
                            
                            if (((cover == LosEffects.COVER_75RIGHT) || (cover == LosEffects.COVER_75LEFT))
                                ||
                                
                                ((cover == LosEffects.COVER_HORIZONTAL) && (le
                                                                                    .getDamagableCoverTypeSecondary()
                                                                            != LosEffects.DAMAGABLE_COVER_NONE))) {
                                
                                
                                int hitLoc = hit.getLocation();
                                
                                
                                if ((hitLoc == Mech.LOC_RLEG)
                                    || (hitLoc == Mech.LOC_RT)
                                    || (hitLoc == Mech.LOC_RARM)) {
                                    
                                    damagableCoverType = le
                                            .getDamagableCoverTypePrimary();
                                    coverDropship = le.getCoverDropshipPrimary();
                                    coverLoc = le.getCoverLocPrimary();
                                } else {
                                    
                                    
                                    damagableCoverType = le
                                            .getDamagableCoverTypeSecondary();
                                    coverDropship = le.getCoverDropshipSecondary();
                                    coverLoc = le.getCoverLocSecondary();
                                }
                            } else { 
                                damagableCoverType = le
                                        .getDamagableCoverTypePrimary();
                                coverDropship = le.getCoverDropshipPrimary();
                                coverLoc = le.getCoverLocPrimary();
                            }

                            
                            
                            
                            Vector<Report> coverDamageReport = new Vector<Report>();
                            if (damagableCoverType == LosEffects.DAMAGABLE_COVER_DROPSHIP) {
                                r = new Report(3465);
                                r.addDesc(coverDropship);
                                r.indent(1);
                                coverDamageReport = deliverInfernoMissiles(ae,
                                                                           coverDropship, 1, CalledShot.CALLED_NONE);
                                coverDamageReport.insertElementAt(r, 0);
                                for (Report report : coverDamageReport) {
                                    report.indent(1);
                                }
                            } else if (damagableCoverType == LosEffects.DAMAGABLE_COVER_BUILDING) {
                                BuildingTarget bldgTrgt = new BuildingTarget(
                                        coverLoc, game.getBoard(), false);
                                coverDamageReport = deliverInfernoMissiles(ae,
                                                                           bldgTrgt, 1, CalledShot.CALLED_NONE);
                            }
                            for (Report report : coverDamageReport) {
                                report.indent(1);
                            }
                            coverDamageReports.addAll(coverDamageReport);
                        } else { 
                            if ((te.getArmor(hit) > 0)
                                && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_HEAT_DISSIPATING)) {
                                heatDamage += 1;
                                heatReduced = true;
                                reductionCause = EquipmentType.armorNames[te
                                        .getArmorType(hit.getLocation())];
                            } else {
                                heatDamage += 2;
                            }
                        }
                    }
                    if (heatReduced) {
                        r = new Report(3406);
                        r.add(heatDamage);
                        r.subject = te.getId();
                        r.indent(2);
                        r.choose(true);
                        r.add(missiles * 2);
                        r.add(reductionCause);
                    } else {
                        r = new Report(3400);
                        r.add(heatDamage);
                        r.subject = te.getId();
                        r.indent(2);
                        r.choose(true);
                    }
                    vPhaseReport.add(r);
                    Report.addNewline(vPhaseReport);
                    te.heatFromExternal += heatDamage;

                    if (missiles != m) {
                        r = new Report(3403);
                        r.add(m - missiles);
                        r.indent(2);
                        r.subject = te.getId();
                        vPhaseReport.add(r);
                    }
                    vPhaseReport.addAll(coverDamageReports);
                    Report.addNewline(vPhaseReport);
                } else if (te instanceof Aero) {
                    r = new Report(3400);
                    r.add(2 * missiles);
                    r.subject = te.getId();
                    r.indent(2);
                    r.choose(true);
                    vPhaseReport.add(r);
                    te.heatFromExternal += 2 * missiles;
                    Report.addNewline(vPhaseReport);
                } else if (te instanceof Tank) {
                    boolean targetIsSupportVee = (te instanceof SupportTank)
                                                 || (te instanceof LargeSupportTank)
                                                 || (te instanceof SupportVTOL);
                    int direction = Compute.targetSideTable(ae, te, called);
                    while (missiles-- > 0) {
                        HitData hit = te.rollHitLocation(ToHitData.HIT_NORMAL,
                                                         direction);
                        int critRollMod = 0;
                        if (!targetIsSupportVee
                            || (te.hasArmoredChassis() && (te.getBARRating(hit
                                                                                   .getLocation()) > 9))) {
                            critRollMod -= 2;
                        }
                        if ((te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_HARDENED)
                            && (te.getArmor(hit.getLocation()) > 0)) {
                            critRollMod -= 2;
                        }
                        vPhaseReport.addAll(criticalEntity(te, hit.getLocation(),
                                                           hit.isRear(), critRollMod, 0, true));
                    }
                } else if (te instanceof Protomech) {
                    te.heatFromExternal += missiles;
                    while (te.heatFromExternal >= 3) {
                        te.heatFromExternal -= 3;
                        HitData hit = te.rollHitLocation(ToHitData.HIT_NORMAL,
                                                         ToHitData.SIDE_FRONT);
                        if (hit.getLocation() == Protomech.LOC_NMISS) {
                            r = new Report(6035);
                            r.subject = te.getId();
                            r.indent(2);
                            vPhaseReport.add(r);
                        } else {
                            r = new Report(6690);
                            r.subject = te.getId();
                            r.indent(2);
                            r.add(te.getLocationName(hit));
                            vPhaseReport.add(r);
                            te.destroyLocation(hit.getLocation());
                            
                            
                            int hits = Protomech.POSSIBLE_PILOT_DAMAGE[hit
                                    .getLocation()]
                                       - ((Protomech) te).getPilotDamageTaken(hit
                                                                                      .getLocation());
                            if (hits > 0) {
                                vPhaseReport.addAll(damageCrew(te, hits));
                                ((Protomech) te).setPilotDamageTaken(hit
                                                                             .getLocation(),
                                                                     Protomech.POSSIBLE_PILOT_DAMAGE[hit
                                                                             .getLocation()]);
                            }
                            if (te.getTransferLocation(hit).getLocation() == Entity.LOC_DESTROYED) {
                                vPhaseReport.addAll(destroyEntity(te,
                                                                  "flaming inferno death", false, true));
                                Report.addNewline(vPhaseReport);
                            }
                        }
                    }
                } else if (te instanceof BattleArmor) {
                    if (((BattleArmor) te).isFireResistant()) {
                        r = new Report(3395);
                        r.indent(2);
                        r.subject = te.getId();
                        r.addDesc(te);
                        vPhaseReport.add(r);
                        return vPhaseReport;
                    }
                    te.heatFromExternal += missiles;
                    while (te.heatFromExternal >= 3) {
                        te.heatFromExternal -= 3;
                        HitData hit = te.rollHitLocation(ToHitData.HIT_NORMAL,
                                                         ToHitData.SIDE_FRONT);
                        hit.setEffect(HitData.EFFECT_CRITICAL);
                        vPhaseReport.addAll(damageEntity(te, hit, 1));
                        Report.addNewline(vPhaseReport);
                    }
                } else if (te instanceof Infantry) {
                    HitData hit = new HitData(Infantry.LOC_INFANTRY);
                    if (te.getInternal(hit) > (3 * missiles)) {
                        
                        te.setInternal(te.getInternal(hit) - (3 * missiles), hit);
                        r = new Report(6065);
                        r.addDesc(te);
                        r.add(3 * missiles);
                        r.indent(2);
                        r.add(te.getLocationAbbr(hit));
                        r.newlines = 0;
                        r.subject = te.getId();
                        vPhaseReport.add(r);
                        Report.addNewline(vPhaseReport);
                        r = new Report(6095);
                        r.add(te.getInternal(hit));
                        r.subject = te.getId();
                        r.indent(3);
                        vPhaseReport.add(r);
                    } else {
                        vPhaseReport.addAll(destroyEntity(te, "damage", false));
                        Report.addNewline(vPhaseReport);
                    }
                } else {
                    
                    int direction = Compute.targetSideTable(ae, te, called);
                    while (missiles-- > 0) {
                        HitData hit = te.rollHitLocation(ToHitData.HIT_NORMAL,
                                                         direction);
                        vPhaseReport.addAll(damageEntity(te, hit, 2));
                    }
                }
        }
        return vPhaseReport;
    }

    
    private boolean enterMinefield(Entity entity, Coords c, int curElev,
            boolean isOnGround, Vector<Report> vMineReport) {
        return enterMinefield(entity, c, curElev, isOnGround, vMineReport, -1);
    }

    
    private boolean enterMinefield(Entity entity, Coords c, int curElev,
            boolean isOnGround, Vector<Report> vMineReport, int target) {
        Report r;
        boolean trippedMine = false;
        
        if (curElev > 0) {
            return false;
        }

        
        Mounted minesweeper = null;
        for (Mounted m : entity.getMisc()) {
            if (m.getType().hasFlag(MiscType.F_MINESWEEPER) && m.isReady()
                    && (m.getArmorValue() > 0)) {
                minesweeper = m;
                break; 
            }
        }

        Vector<Minefield> fieldsToRemove = new Vector<>();
        
        for (Minefield mf : game.getMinefields(c)) {

            
            if (mf.getType() == Minefield.TYPE_VIBRABOMB) {
                continue;
            }

            
            
            if (game.getBoard().getHex(mf.getCoords())
                    .containsTerrain(Terrains.WATER)) {
                if ((Math.abs(curElev) != mf.getDepth())
                    && (Math.abs(curElev + entity.getHeight()) != mf
                        .getDepth())) {
                    continue;
                }
            }

            
            if ((minesweeper != null)
                    && ((mf.getType() == Minefield.TYPE_CONVENTIONAL)
                            || (mf.getType() == Minefield.TYPE_ACTIVE)
                            || (mf.getType() == Minefield.TYPE_INFERNO))) {
                
                int roll = Compute.d6(2);

                
                if (doBlind()) { 
                r = new Report(2152);
                r.player = mf.getPlayerId();
                r.add(Minefield.getDisplayableName(mf.getType()));
                r.add(mf.getCoords().getBoardNum());
                r.add(roll);
                r.newlines = 0;
                vMineReport.add(r);
                }

                if (roll >= 6) {
                    
                    if (doBlind()) {
                        r = new Report(5543);
                        r.player = mf.getPlayerId();
                        vMineReport.add(r);
                    }

                    
                    r = new Report(2158);
                    r.subject = entity.getId();
                    r.add(entity.getShortName(), true);
                    r.add(Minefield.getDisplayableName(mf.getType()), true);
                    r.add(mf.getCoords().getBoardNum(), true);
                    r.indent();
                    vMineReport.add(r);
                    fieldsToRemove.add(mf);

                    
                    int remainingAV = minesweeper.getArmorValue() - 6;
                    minesweeper.setArmorValue(Math.max(remainingAV, 0));

                    r = new Report(2161);
                    r.indent(2);
                    r.subject = entity.getId();
                    r.add(entity.getShortName(), true);
                    r.add(6);
                    r.add(Math.max(remainingAV, 0));
                    vMineReport.add(r);

                    if (remainingAV <= 0) {
                        minesweeper.setDestroyed(true);
                    }
                    
                    if (remainingAV < 0) {
                        int damage = Math.abs(remainingAV);
                        r = new Report(2162);
                        r.indent(2);
                        r.subject = entity.getId();
                        r.add(damage, true);
                        vMineReport.add(r);

                        
                        HitData hit = new HitData(minesweeper.getLocation());
                        Vector<Report> damageReports = damageEntity(entity,
                                hit, damage);
                        for (Report r1 : damageReports) {
                            r1.indent(1);
                        }
                        vMineReport.addAll(damageReports);
                    }
                    Report.addNewline(vMineReport);
                    
                    continue;
                } else {
                    
                    if (doBlind()) {
                        r = new Report(5542);
                        r.player = mf.getPlayerId();
                        vMineReport.add(r);
                    }
                }
            }

            
            if ((mf.getType() == Minefield.TYPE_ACTIVE) && isOnGround) {
                continue;
            }
            if ((mf.getType() != Minefield.TYPE_ACTIVE) && !isOnGround) {
                continue;
            }

            
            if (target == -1) {
                target = mf.getTrigger();
                if (mf.getType() == Minefield.TYPE_ACTIVE) {
                    target = 9;
                }
                if (entity instanceof Infantry) {
                    target += 1;
                }
                if ((entity.getMovementMode() == EntityMovementMode.HOVER)
                    || (entity.getMovementMode() == EntityMovementMode.WIGE)) {
                    target = 12;
                }
            }

            int roll = Compute.d6(2);

            
            if (doBlind()) { 
                r = new Report(2151);
                r.player = mf.getPlayerId();
                r.add(Minefield.getDisplayableName(mf.getType()));
                r.add(mf.getCoords().getBoardNum());
                r.add(target);
                r.add(roll);
                r.newlines = 0;
                vMineReport.add(r);
            }

            if (roll < target) {
                
                if (doBlind()) {
                    r = new Report(2217);
                    r.player = mf.getPlayerId();
                    vMineReport.add(r);
                }
                continue;
            }

            
            if (doBlind()) {
                r = new Report(2270);
                r.player = mf.getPlayerId();
                vMineReport.add(r);
            }

            
            trippedMine = true;
            
            mf.setDetonated(true);
            if (mf.getType() == Minefield.TYPE_INFERNO) {
                
                r = new Report(2155);
                r.subject = entity.getId();
                r.add(entity.getShortName(), true);
                r.add(mf.getCoords().getBoardNum(), true);
                r.indent();
                vMineReport.add(r);
                vMineReport.addAll(deliverInfernoMissiles(entity, entity,
                        mf.getDensity() / 2));
            } else {
                r = new Report(2150);
                r.subject = entity.getId();
                r.add(entity.getShortName(), true);
                r.add(mf.getCoords().getBoardNum(), true);
                r.indent();
                vMineReport.add(r);
                int damage = mf.getDensity();
                while (damage > 0) {
                    int cur_damage = Math.min(5, damage);
                    damage = damage - cur_damage;
                    HitData hit;
                    if (minesweeper == null) {
                        hit = entity.rollHitLocation(Minefield.TO_HIT_TABLE,
                                Minefield.TO_HIT_SIDE);
                    } else { 
                        hit = new HitData(minesweeper.getLocation());
                    }
                    vMineReport.addAll(damageEntity(entity, hit, cur_damage));
                }
                if (entity instanceof Tank) {
                    
                    
                    
                    Tank tank = (Tank) entity;
                    vMineReport.addAll(vehicleMotiveDamage(tank,
                            tank.getMotiveSideMod(ToHitData.SIDE_LEFT)));
                }
                Report.addNewline(vMineReport);
            }

            
            mf.checkReduction(0, true);
            revealMinefield(mf);
        }

        for (Minefield mf : fieldsToRemove) {
            removeMinefield(mf);
        }

        return trippedMine;
    }

    
    private void resetMines() {

        Enumeration<Coords> mineLoc = game.getMinedCoords();
        while (mineLoc.hasMoreElements()) {
            Coords c = mineLoc.nextElement();
            Enumeration<Minefield> minefields = game.getMinefields(c)
                                                    .elements();
            while (minefields.hasMoreElements()) {
                Minefield minefield = minefields.nextElement();
                if (minefield.hasDetonated()) {
                    minefield.setDetonated(false);
                    Enumeration<Minefield> otherMines = game.getMinefields(c)
                                                            .elements();
                    while (otherMines.hasMoreElements()) {
                        Minefield otherMine = otherMines.nextElement();
                        if (otherMine.equals(minefield)) {
                            continue;
                        }
                        int bonus = 0;
                        if (otherMine.getDensity() > minefield.getDensity()) {
                            bonus = 1;
                        }
                        if (otherMine.getDensity() < minefield.getDensity()) {
                            bonus = -1;
                        }
                        otherMine.checkReduction(bonus, false);
                    }
                }
            }
            
            
            ArrayList<Minefield> mfRemoved = new ArrayList<Minefield>();
            Enumeration<Minefield> mines = game.getMinefields(c).elements();
            while (mines.hasMoreElements()) {
                Minefield mine = mines.nextElement();
                if (mine.getDensity() < 5) {
                    mfRemoved.add(mine);
                }
            }
            
            for (Minefield mf : mfRemoved) {
                removeMinefield(mf);
            }
            
            sendChangedMines(c);
        }
    }

    
    public boolean clearMinefield(Minefield mf, Entity en, int target,
                                  Vector<Report> vClearReport) {
        return clearMinefield(mf, en, target, -1, vClearReport, 2);
    }

    public boolean clearMinefield(Minefield mf, Entity en, int target,
                                  int botch, Vector<Report> vClearReport) {
        return clearMinefield(mf, en, target, botch, vClearReport, 1);
    }

    
    public boolean clearMinefield(Minefield mf, Entity en, int target,
            int botch, Vector<Report> vClearReport, int indent) {
        Report r;
        int roll = Compute.d6(2);
        if (roll >= target) {
            r = new Report(2250);
            r.subject = en.getId();
            r.add(Minefield.getDisplayableName(mf.getType()));
            r.add(target);
            r.add(roll);
            r.indent(indent);
            vClearReport.add(r);
            return true;
        } else if (roll <= botch) {
            
            r = new Report(2255);
            r.subject = en.getId();
            r.indent(indent);
            r.add(Minefield.getDisplayableName(mf.getType()));
            r.add(target);
            r.add(roll);
            vClearReport.add(r);
            
            
            for (Entity victim : game.getEntitiesVector(mf.getCoords())) {
                Report rVictim;
                if (victim.isClearingMinefield()) {
                    rVictim = new Report(2265);
                    rVictim.subject = victim.getId();
                    rVictim.add(victim.getShortName(), true);
                    rVictim.indent(indent + 1);
                    vClearReport.add(rVictim);
                    int damage = mf.getDensity();
                    while (damage > 0) {
                        int cur_damage = Math.min(5, damage);
                        damage = damage - cur_damage;
                        HitData hit = victim.rollHitLocation(
                                Minefield.TO_HIT_TABLE, Minefield.TO_HIT_SIDE);
                        vClearReport.addAll(damageEntity(victim, hit,
                                                         cur_damage));
                    }
                }
            }
            
            if (mf.getType() == Minefield.TYPE_CONVENTIONAL) {
                mf.setDensity(Math.max(5, mf.getDensity() - 5));
            } else {
                
                return true;
            }
        } else {
            
            r = new Report(2260);
            r.subject = en.getId();
            r.indent(indent);
            r.add(Minefield.getDisplayableName(mf.getType()));
            r.add(target);
            r.add(roll);
            vClearReport.add(r);
        }
        return false;
    }

    
    private void clearDetonatedMines(Coords c, int target) {

        Enumeration<Minefield> minefields = game.getMinefields(c).elements();
        ArrayList<Minefield> mfRemoved = new ArrayList<Minefield>();
        while (minefields.hasMoreElements()) {
            Minefield minefield = minefields.nextElement();
            if (minefield.hasDetonated() && (Compute.d6(2) >= target)) {
                mfRemoved.add(minefield);
            }
        }
        
        for (Minefield mf : mfRemoved) {
            removeMinefield(mf);
        }
    }

    
    private boolean checkVibrabombs(Entity entity, Coords coords,
                                    boolean displaced, Vector<Report> vMineReport) {
        return checkVibrabombs(entity, coords, displaced, null, null,
                               vMineReport);
    }

    
    private boolean checkVibrabombs(Entity entity, Coords coords,
            boolean displaced, Coords lastPos, Coords curPos,
            Vector<Report> vMineReport) {
        int mass = (int) entity.getWeight();

        
        Mounted minesweeper = null;
        for (Mounted m : entity.getMisc()) {
            if (m.getType().hasFlag(MiscType.F_MINESWEEPER) && m.isReady()
                    && (m.getArmorValue() > 0)) {
                minesweeper = m;
                break; 
            }
        }

        
        if (minesweeper != null) {
            Vector<Minefield> fieldsToRemove = new Vector<>();
            for (Minefield mf : game.getVibrabombs()) {
                
                if (!mf.getCoords().equals(coords)) {
                    continue;
                }

                
                
                if (Math.abs(mass - mf.getSetting()) < 10) {
                    
                    Report r;
                    r = new Report(2158);
                    r.subject = entity.getId();
                    r.add(entity.getShortName(), true);
                    r.add(Minefield.getDisplayableName(mf.getType()), true);
                    r.add(mf.getCoords().getBoardNum(), true);
                    r.indent();
                    vMineReport.add(r);
                    fieldsToRemove.add(mf);

                    
                    int remainingAV = minesweeper.getArmorValue() - 10;
                    minesweeper.setArmorValue(Math.max(remainingAV, 0));

                    r = new Report(2161);
                    r.indent(2);
                    r.subject = entity.getId();
                    r.add(entity.getShortName(), true);
                    r.add(10);
                    r.add(Math.max(remainingAV, 0));
                    vMineReport.add(r);

                    if (remainingAV <= 0) {
                        minesweeper.setDestroyed(true);
                    }
                    
                    if (remainingAV < 0) {
                        int damage = Math.abs(remainingAV);
                        r = new Report(2162);
                        r.indent(2);
                        r.subject = entity.getId();
                        r.add(damage, true);
                        vMineReport.add(r);

                        
                        HitData hit = new HitData(minesweeper.getLocation());
                        Vector<Report> damageReports = damageEntity(entity,
                                hit, damage);
                        for (Report r1 : damageReports) {
                            r1.indent(1);
                        }
                        vMineReport.addAll(damageReports);
                        entity.applyDamage();
                    }
                    Report.addNewline(vMineReport);
                }
            }
            for (Minefield mf : fieldsToRemove) {
                removeMinefield(mf);
            }
        }


        boolean boom = false;
        
        if (!(entity instanceof Mech)) {
            return boom;
        }

        Enumeration<Minefield> e = game.getVibrabombs().elements();

        while (e.hasMoreElements()) {
            Minefield mf = e.nextElement();

            if (game.getBoard().getHex(mf.getCoords())
                    .containsTerrain(Terrains.WATER)
                && !game.getBoard().getHex(mf.getCoords())
                        .containsTerrain(Terrains.PAVEMENT)
                && !game.getBoard().getHex(mf.getCoords())
                        .containsTerrain(Terrains.ICE)) {
                continue;
            }

            if (mass <= (mf.getSetting() - 10)) {
                continue;
            }

            int effectiveDistance = (mass - mf.getSetting()) / 10;
            int actualDistance = coords.distance(mf.getCoords());

            if (actualDistance <= effectiveDistance) {
                Report r = new Report(2156);
                r.subject = entity.getId();
                r.add(entity.getShortName(), true);
                r.add(mf.getCoords().getBoardNum(), true);
                vMineReport.add(r);
                explodeVibrabomb(mf, vMineReport, false);
            }

            if (!displaced && (actualDistance == 0)) {
                
                Report r = new Report(2160);
                r.subject = entity.getId();
                r.add(entity.getShortName(), true);
                vMineReport.add(r);
                int damage = mf.getDensity();
                while (damage > 0) {
                    int cur_damage = Math.min(5, damage);
                    damage = damage - cur_damage;
                    HitData hit = entity.rollHitLocation(
                            Minefield.TO_HIT_TABLE, Minefield.TO_HIT_SIDE);
                    vMineReport.addAll(damageEntity(entity, hit, cur_damage));
                }
                vMineReport.addAll(resolvePilotingRolls(entity, true, lastPos,
                                                        curPos));
                
                entity.applyDamage();
            }

            
            if (mf.hasDetonated()) {
                boom = true;
                mf.checkReduction(0, true);
            }

        }
        return boom;
    }

    
    public void removeMinefield(Minefield mf) {
        if (game.containsVibrabomb(mf)) {
            game.removeVibrabomb(mf);
        }
        game.removeMinefield(mf);

        Enumeration<IPlayer> players = game.getPlayers();
        while (players.hasMoreElements()) {
            IPlayer player = players.nextElement();
            removeMinefield(player, mf);
        }
    }

    
    private void removeMinefield(IPlayer player, Minefield mf) {
        if (player.containsMinefield(mf)) {
            player.removeMinefield(mf);
            send(player.getId(),
                 new Packet(Packet.COMMAND_REMOVE_MINEFIELD, mf));
        }
    }

    
    private void revealMinefield(Minefield mf) {
        Enumeration<Team> teams = game.getTeams();
        while (teams.hasMoreElements()) {
            Team team = teams.nextElement();
            revealMinefield(team, mf);
        }
    }

    
    private void revealMinefield(Team team, Minefield mf) {
        Enumeration<IPlayer> players = team.getPlayers();
        while (players.hasMoreElements()) {
            IPlayer player = players.nextElement();
            if (!player.containsMinefield(mf)) {
                player.addMinefield(mf);
                send(player.getId(), new Packet(
                        Packet.COMMAND_REVEAL_MINEFIELD, mf));
            }
        }
    }

    
    private void checkForRevealMinefield(Minefield mf, Entity layer) {
        Enumeration<Team> teams = game.getTeams();
        
        
        
        while (teams.hasMoreElements()) {
            Team team = teams.nextElement();
            boolean canSee = false;

            
            if (team.equals(game.getTeamForPlayer(game.getPlayer(mf
                                                                         .getPlayerId())))) {
                canSee = true;
            } else {

                int target = Integer.MAX_VALUE;
                Iterator<Entity> entities = game.getEntities();
                while (entities.hasNext()) {
                    Entity en = entities.next();
                    
                    if (!team.equals(game.getTeamForPlayer(en.getOwner()))) {
                        continue;
                    }
                    if (LosEffects.calculateLos(
                            game,
                            en.getId(),
                            new HexTarget(mf.getCoords(), game.getBoard(),
                                          Targetable.TYPE_HEX_CLEAR)).canSee()) {
                        target = 0;
                        break;
                    }
                    LosEffects los = LosEffects.calculateLos(game, en.getId(),
                                                             layer);
                    if (los.canSee()) {
                        
                        ToHitData current = new ToHitData(4, "base");
                        current.append(Compute.getAttackerMovementModifier(
                                game, en.getId()));
                        current.append(Compute.getTargetMovementModifier(game,
                                                                         layer.getId()));
                        current.append(los.losModifiers(game));
                        if (current.getValue() < target) {
                            target = current.getValue();
                        }
                    }
                }

                if (Compute.d6(2) >= target) {
                    canSee = true;
                }
            }
            if (canSee) {
                revealMinefield(team, mf);
            }
        }
    }

    private void explodeVibrabomb(Minefield mf, Vector<Report> vBoomReport,
                                  boolean reduce) {
        Iterator<Entity> targets = game.getEntities(mf.getCoords());
        Report r;

        while (targets.hasNext()) {
            Entity entity = targets.next();

            
            if (entity.isAirborne()) {
                continue;
            }

            if (!entity.isDone()
                && game.getOptions().booleanOption("no_premove_vibra")) {
                r = new Report(2157);
                r.subject = entity.getId();
                r.add(entity.getShortName(), true);
                vBoomReport.add(r);
                continue;
            }
            
            r = new Report(2160);
            r.subject = entity.getId();
            r.add(entity.getShortName(), true);
            vBoomReport.add(r);

            int damage = mf.getDensity();
            while (damage > 0) {
                int cur_damage = Math.min(5, damage);
                damage = damage - cur_damage;
                HitData hit = entity.rollHitLocation(Minefield.TO_HIT_TABLE,
                                                     Minefield.TO_HIT_SIDE);
                vBoomReport.addAll(damageEntity(entity, hit, cur_damage));
            }
            Report.addNewline(vBoomReport);

            if (entity instanceof Tank) {
                vBoomReport.addAll(vehicleMotiveDamage((Tank) entity, 2));
            }
            vBoomReport.addAll(resolvePilotingRolls(entity, true,
                                                    entity.getPosition(), entity.getPosition()));
            
            
            game.resetPSRs(entity);
            entity.applyDamage();
            Report.addNewline(vBoomReport);
            entityUpdate(entity.getId());
        }

        
        if (reduce) {
            mf.checkReduction(0, true);
        }
        mf.setDetonated(true);
    }
    
    private void drownSwarmer(Entity entity, Coords pos) {
        
        final int swarmerId = entity.getSwarmAttackerId();
        if (Entity.NONE != swarmerId) {
            final Entity swarmer = game.getEntity(swarmerId);
            
            if (!(swarmer instanceof BattleArmor)) {
                swarmer.setSwarmTargetId(Entity.NONE);
                entity.setSwarmAttackerId(Entity.NONE);
                swarmer.setPosition(pos);
                Report r = new Report(2165);
                r.subject = entity.getId();
                r.indent();
                r.add(entity.getShortName(), true);
                addReport(r);
                addReport(destroyEntity(swarmer, "a watery grave", false));
                entityUpdate(swarmerId);
            }
        }
    }

    
    void checkForWashedInfernos(Entity entity, Coords coords) {
        IHex hex = game.getBoard().getHex(coords);
        int waterLevel = hex.terrainLevel(Terrains.WATER);
        
        if (!(entity instanceof Mech) || !entity.infernos.isStillBurning()) {
            return;
        }
        
        if ((waterLevel > 0) && (entity.relHeight() < 0)) {
            washInferno(entity, coords);
        }
    }

    
    void washInferno(Entity entity, Coords coords) {
        game.getBoard().addInfernoTo(coords, InfernoTracker.STANDARD_ROUND, 1);
        entity.infernos.clear();

        
        IHex hex = game.getBoard().getHex(coords);
        Report r = new Report(2170);
        r.subject = entity.getId();
        r.addDesc(entity);
        if (!hex.containsTerrain(Terrains.FIRE)) {
            r.messageId = 2175;
            ignite(coords, Terrains.FIRE_LVL_INFERNO, null);
        }
        addReport(r);
    }

    
    public void addMovementHeat() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            Entity entity = i.next();

            if ((entity.getMovementMode() == EntityMovementMode.BIPED_SWIM)
                || (entity.getMovementMode() == EntityMovementMode.QUAD_SWIM)) {
                return;
            }

            
            if (entity.isEvading() && !(entity instanceof Aero)) {
                entity.heatBuildup += entity.getRunHeat() + 2;
            } else if (entity.moved == EntityMovementType.MOVE_NONE) {
                entity.heatBuildup += entity.getStandingHeat();
            } else if ((entity.moved == EntityMovementType.MOVE_WALK)
                       || (entity.moved == EntityMovementType.MOVE_VTOL_WALK)
                       || (entity.moved == EntityMovementType.MOVE_CAREFUL_STAND)) {
                entity.heatBuildup += entity.getWalkHeat();
            } else if ((entity.moved == EntityMovementType.MOVE_RUN)
                       || (entity.moved == EntityMovementType.MOVE_VTOL_RUN)
                       || (entity.moved == EntityMovementType.MOVE_SKID)) {
                entity.heatBuildup += entity.getRunHeat();
            } else if (entity.moved == EntityMovementType.MOVE_JUMP) {
                entity.heatBuildup += entity.getJumpHeat(entity.delta_distance);
            } else if (entity.moved == EntityMovementType.MOVE_SPRINT) {
                entity.heatBuildup += entity.getSprintHeat();
            }
            if (entity.hasDamagedRHS()) {
                entity.heatBuildup += 1;
            }
        }
    }

    
    public Vector<Report> doSetLocationsExposure(Entity entity, IHex hex,
                                                 boolean isJump, int elevation) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        if ((hex.terrainLevel(Terrains.WATER) > 0) && !isJump
            && (elevation < 0)) {
            if ((entity instanceof Mech) && !entity.isProne()
                && (hex.terrainLevel(Terrains.WATER) == 1)) {
                for (int loop = 0; loop < entity.locations(); loop++) {
                    if (game.getPlanetaryConditions().isVacuum()) {
                        entity.setLocationStatus(loop,
                                                 ILocationExposureStatus.VACUUM);
                    } else {
                        entity.setLocationStatus(loop,
                                                 ILocationExposureStatus.NORMAL);
                    }
                }
                entity.setLocationStatus(Mech.LOC_RLEG,
                                         ILocationExposureStatus.WET);
                entity.setLocationStatus(Mech.LOC_LLEG,
                                         ILocationExposureStatus.WET);
                vPhaseReport.addAll(breachCheck(entity, Mech.LOC_RLEG, hex));
                vPhaseReport.addAll(breachCheck(entity, Mech.LOC_LLEG, hex));
                if (entity instanceof QuadMech) {
                    entity.setLocationStatus(Mech.LOC_RARM,
                                             ILocationExposureStatus.WET);
                    entity.setLocationStatus(Mech.LOC_LARM,
                                             ILocationExposureStatus.WET);
                    vPhaseReport
                            .addAll(breachCheck(entity, Mech.LOC_RARM, hex));
                    vPhaseReport
                            .addAll(breachCheck(entity, Mech.LOC_LARM, hex));
                }
                if (entity instanceof TripodMech) {
                    entity.setLocationStatus(Mech.LOC_CLEG,
                            ILocationExposureStatus.WET);
                    vPhaseReport.addAll(breachCheck(entity, Mech.LOC_CLEG, hex));
                }
            } else {
                for (int loop = 0; loop < entity.locations(); loop++) {
                    entity.setLocationStatus(loop, ILocationExposureStatus.WET);
                    vPhaseReport.addAll(breachCheck(entity, loop, hex));
                }
            }
        } else {
            for (int loop = 0; loop < entity.locations(); loop++) {
                if (game.getPlanetaryConditions().isVacuum()) {
                    entity.setLocationStatus(loop,
                                             ILocationExposureStatus.VACUUM);
                } else {
                    entity.setLocationStatus(loop,
                                             ILocationExposureStatus.NORMAL);
                }
            }
        }
        return vPhaseReport;
    }

    
    private void resistGForce(Entity entity, int targetroll) {

        
        Report r = new Report(9330);
        r.subject = entity.getId();
        r.addDesc(entity);
        addReport(r);

        
        final int diceRoll = Compute.d6(2);
        r = new Report(9335);
        r.subject = entity.getId();
        r.add(Integer.toString(targetroll));
        r.add(diceRoll);
        if (diceRoll < targetroll) {
            r.choose(false);
            addReport(r);
            addReport(damageCrew(entity, 1));
        } else {
            r.choose(true);
            addReport(r);
        }
    }

    
    private boolean doSkillCheckInSpace(Entity entity, PilotingRollData roll) {

        if (roll.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            return true;
        }

        
        Report r = new Report(9320);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(roll.getLastPlainDesc(), true);
        addReport(r);

        
        final int diceRoll = Compute.d6(2);
        r = new Report(9325);
        r.subject = entity.getId();
        r.add(roll.getValueAsString());
        r.add(roll.getDesc());
        r.add(diceRoll);
        boolean suc;
        if (diceRoll < roll.getValue()) {
            r.choose(false);
            addReport(r);
            suc = false;
        } else {
            r.choose(true);
            addReport(r);
            suc = true;
        }

        return suc;
    }

    
    private boolean doVerticalTakeOffCheck(Entity entity, PilotingRollData roll) {

        if (!(entity instanceof Aero)) {
            return false;
        }

        Aero a = (Aero) entity;

        if (roll.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            return true;
        }

        
        Report r = new Report(9320);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(roll.getLastPlainDesc(), true);
        addReport(r);

        
        final int diceRoll = Compute.d6(2);
        r = new Report(9321);
        r.subject = entity.getId();
        r.add(roll.getValueAsString());
        r.add(roll.getDesc());
        r.add(diceRoll);
        r.newlines = 0;
        addReport(r);
        boolean suc = false;
        if (diceRoll < roll.getValue()) {
            int mof = roll.getValue() - diceRoll;
            if (mof < 3) {
                r = new Report(9322);
                r.subject = entity.getId();
                addReport(r);
                suc = true;
            } else if (mof < 5) {
                PilotingRollData newRoll = entity.getBasePilotingRoll();
                if (Compute.d6(2) >= newRoll.getValue()) {
                    r = new Report(9322);
                    r.subject = entity.getId();
                    addReport(r);
                    suc = true;
                } else {
                    r = new Report(9323);
                    r.subject = entity.getId();
                    addReport(r);
                    int damage = 20;
                    while (damage > 0) {
                        addReport(damageEntity(entity, entity.rollHitLocation(
                                                       ToHitData.HIT_NORMAL, ToHitData.SIDE_REAR),
                                               Math.min(5, damage)));
                        damage -= 5;
                    }
                }
            } else if (mof < 6) {
                r = new Report(9323);
                r.subject = entity.getId();
                addReport(r);
                int damage = 50;
                while (damage > 0) {
                    addReport(damageEntity(entity, entity.rollHitLocation(
                                                   ToHitData.HIT_NORMAL, ToHitData.SIDE_REAR),
                                           Math.min(5, damage)));
                    damage -= 5;
                }
            } else {
                r = new Report(9323);
                r.subject = entity.getId();
                addReport(r);
                int damage = 100;
                while (damage > 0) {
                    addReport(damageEntity(entity, entity.rollHitLocation(
                                                   ToHitData.HIT_NORMAL, ToHitData.SIDE_REAR),
                                           Math.min(5, damage)));
                    damage -= 5;
                }
            }
            a.setGearHit(true);
            r = new Report(9125);
            r.subject = entity.getId();
            addReport(r);
        } else {
            r = new Report(9322);
            r.subject = entity.getId();
            addReport(r);
            suc = true;
        }

        return suc;
    }

    
    private boolean doSkillCheckManeuver(Entity entity, PilotingRollData roll) {

        if (roll.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            return true;
        }

        
        Report r = new Report(9600);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(roll.getLastPlainDesc(), true);
        addReport(r);

        
        final int diceRoll = Compute.d6(2);
        r = new Report(9601);
        r.subject = entity.getId();
        r.add(roll.getValueAsString());
        r.add(roll.getDesc());
        r.add(diceRoll);
        boolean suc;
        if (diceRoll < roll.getValue()) {
            r.choose(false);
            addReport(r);
            suc = false;
        } else {
            r.choose(true);
            addReport(r);
            suc = true;
        }

        return suc;
    }

    
    private boolean doSkillCheckInPlace(Entity entity, PilotingRollData roll) {
        if (roll.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            return true;
        }

        if (entity.isProne()) {
            return true;
        }

        
        Report r = new Report(2180);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(roll.getLastPlainDesc(), true);
        addReport(r);

        
        final int diceRoll = entity.getCrew().rollPilotingSkill();
        r = new Report(2185);
        r.subject = entity.getId();
        r.add(roll.getValueAsString());
        r.add(roll.getDesc());
        r.add(diceRoll);
        boolean suc;
        if (diceRoll < roll.getValue()) {
            r.choose(false);
            addReport(r);
            if ((entity instanceof Mech)
                && game.getOptions().booleanOption(
                    "tacops_falling_expanded")
                && (entity.getCrew().getPiloting() < 6)
                && !entity.isHullDown() && entity.canGoHullDown()) {
                if ((entity.getCrew().getPiloting() > 1)
                    && ((roll.getValue() - diceRoll) < 2)) {
                    entity.setHullDown(true);
                } else if ((entity.getCrew().getPiloting() <= 1)
                           && ((roll.getValue() - diceRoll) < 3)) {
                    entity.setHullDown(true);
                }
            }
            if (!entity.isHullDown()
                || (entity.isHullDown() && !entity.canGoHullDown())) {
                addReport(doEntityFall(entity, roll));
            } else {
                r = new Report(2317);
                r.subject = entity.getId();
                r.add(entity.getDisplayName());
                addReport(r);
            }

            suc = false;
            
            entity.doCheckEngineStallRoll(vPhaseReport);
        } else {
            r.choose(true);
            addReport(r);
            suc = true;
        }

        return suc;
    }

    
    private boolean doDislodgeSwarmerSkillCheck(Entity entity,
                                                PilotingRollData roll, Coords curPos) {
        
        Report r = new Report(2180);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(roll.getLastPlainDesc(), true);
        addReport(r);

        
        final int diceRoll = Compute.d6(2);
        r = new Report(2190);
        r.subject = entity.getId();
        r.add(roll.getValueAsString());
        r.add(roll.getDesc());
        r.add(diceRoll);
        if (diceRoll < roll.getValue()) {
            r.choose(false);
            addReport(r);
            
            entity.doCheckEngineStallRoll(vPhaseReport);
            return false;
        }
        
        int swarmerId = entity.getSwarmAttackerId();
        final Entity swarmer = game.getEntity(swarmerId);
        if (!swarmer.isDone()) {
            game.removeTurnFor(swarmer);
            swarmer.setDone(true);
            send(createTurnVectorPacket());
        }

        
        r.choose(true);
        addReport(r);
        entity.setPosition(curPos);
        addReport(doEntityFallsInto(entity, curPos, roll, false));
        return true;
    }

    
    private int doSkillCheckWhileMoving(Entity entity, int entityElevation,
                                        Coords src, Coords dest, PilotingRollData roll, boolean isFallRoll) {
        boolean fallsInPlace;

        
        Report r = new Report(1210);
        r.subject = entity.getId();
        r.addDesc(entity);

        
        if (src.equals(dest)) {
            fallsInPlace = true;
            r.messageId = 2195;
            r.add(src.getBoardNum(), true);
        } else {
            fallsInPlace = false;
            r.messageId = 2200;
            r.add(src.getBoardNum(), true);
            r.add(dest.getBoardNum(), true);
        }

        
        r.add(roll.getLastPlainDesc(), true);
        addReport(r);

        
        final int diceRoll = Compute.d6(2);
        r = new Report(2185);
        r.subject = entity.getId();
        r.add(roll.getValueAsString());
        r.add(roll.getDesc());
        r.add(diceRoll);
        if (diceRoll < roll.getValue()) {
            
            if (isFallRoll) {
                r.choose(false);
                addReport(r);
                addReport(doEntityFallsInto(entity, entityElevation,
                                            fallsInPlace ? dest : src, fallsInPlace ? src : dest,
                                            roll, true));
            } else {
                r.messageId = 2190;
                r.choose(false);
                addReport(r);
                entity.setPosition(fallsInPlace ? src : dest);
            }
            
            entity.doCheckEngineStallRoll(vPhaseReport);
            return roll.getValue() - diceRoll;
        }
        r.choose(true);
        r.newlines = 2;
        addReport(r);
        return 0;
    }

    
    private Vector<Report> doEntityFallsInto(Entity entity, Coords src,
                                             PilotingRollData roll, boolean causeAffa) {
        return doEntityFallsInto(entity, entity.getElevation(), src, src, roll,
                                 causeAffa);
    }

    
    private Vector<Report> doEntityFallsInto(Entity entity,
                                             int entitySrcElevation, Coords src, Coords dest,
                                             PilotingRollData roll, boolean causeAffa) {
        return doEntityFallsInto(entity, entitySrcElevation, src, dest, roll,
                                 causeAffa, 0);
    }

    
    private Vector<Report> doEntityFallsInto(Entity entity,
                                             int entitySrcElevation, Coords origSrc, Coords origDest,
                                             PilotingRollData roll, boolean causeAffa, int fallReduction) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        IHex srcHex = game.getBoard().getHex(origSrc);
        IHex destHex = game.getBoard().getHex(origDest);
        Coords src, dest;
        
        if (srcHex.getLevel() < destHex.getLevel()) {
            IHex swapHex = destHex;
            destHex = srcHex;
            srcHex = swapHex;
            src = origDest;
            dest = origSrc;
            
            entitySrcElevation -= (destHex.getLevel() - srcHex.getLevel());
        } else {
            src = origSrc;
            dest = origDest;
        }
        final int srcHeightAboveFloor = entitySrcElevation + srcHex.depth(true);
        int fallElevation = Math.abs((srcHex.floor() + srcHeightAboveFloor)
                - (destHex.containsTerrain(Terrains.ICE) ? destHex.surface()
                        : destHex.floor()))
                - fallReduction;
        if (destHex.containsTerrain(Terrains.BLDG_ELEV)) {
            fallElevation -= destHex.terrainLevel(Terrains.BLDG_ELEV);
        }
        if (destHex.containsTerrain(Terrains.BLDG_BASEMENT_TYPE)) {
            if (entity.getElevation() == 0) { 
                fallElevation = destHex.depth(true);
            }
        }

        int direction;
        if (src.equals(dest)) {
            direction = Compute.d6() - 1;
        } else {
            direction = src.direction(dest);
        }
        Report r;
        
        Entity affaTarget = game.getAffaTarget(dest, entity);
        
        r = new Report(2205);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(fallElevation);
        r.add(dest.getBoardNum(), true);
        r.newlines = 0;

        
        if (affaTarget == null) {
            r.newlines = 1;
            vPhaseReport.add(r);
            
            if (src.equals(dest)) {
                vPhaseReport.addAll(doEntityFall(entity, dest, fallElevation,
                                                 direction, roll, false));
            } else {
                
                vPhaseReport.addAll(doEntityFall(entity, dest, fallElevation,
                                                 roll));
            }

            return vPhaseReport;
        }
        vPhaseReport.add(r);

        
        if ((fallElevation >= 2) && causeAffa) {
            
            r = new Report(2210);
            r.subject = entity.getId();
            r.addDesc(affaTarget);
            vPhaseReport.add(r);

            
            ToHitData toHit = new ToHitData(7, "base");
            if ((affaTarget instanceof Tank)
                || (affaTarget instanceof Dropship)) {
                toHit = new ToHitData(TargetRoll.AUTOMATIC_SUCCESS,
                                      "Target is a Tank");
            } else {
                toHit.append(Compute.getTargetMovementModifier(game,
                                                               affaTarget.getId()));
                toHit.append(Compute.getTargetTerrainModifier(game, affaTarget));
            }

            if (toHit.getValue() != TargetRoll.AUTOMATIC_FAIL) {
                
                final int diceRoll = Compute.d6(2);
                if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
                    r = new Report(2212);
                    r.add(toHit.getValue());
                    r.indent();
                } else {
                    r = new Report(2215);
                    r.subject = entity.getId();
                    r.add(toHit.getValue());
                    r.add(diceRoll);
                    r.newlines = 0;
                    r.indent();
                }
                vPhaseReport.add(r);
                if (diceRoll >= toHit.getValue()) {
                    
                    int damage = Compute.getAffaDamageFor(entity);
                    r = new Report(2220);
                    r.subject = affaTarget.getId();
                    r.addDesc(affaTarget);
                    r.add(damage);
                    vPhaseReport.add(r);
                    while (damage > 0) {
                        int cluster = Math.min(5, damage);
                        HitData hit = affaTarget.rollHitLocation(
                                ToHitData.HIT_PUNCH, ToHitData.SIDE_FRONT);
                        hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                        vPhaseReport.addAll(damageEntity(affaTarget, hit,
                                                         cluster));
                        damage -= cluster;
                    }

                    
                    
                    
                    PilotingRollData pilotRoll = entity.getBasePilotingRoll();
                    pilotRoll.append(roll);
                    vPhaseReport.addAll(doEntityFall(entity, dest,
                                                     fallElevation, 3, pilotRoll, false));
                    vPhaseReport.addAll(doEntityDisplacementMinefieldCheck(
                            entity, src, dest, entity.getElevation()));

                    
                    
                    Entity violation = Compute.stackingViolation(game,
                                                                 entity.getId(), dest);
                    if (violation != null) {
                        PilotingRollData prd = new PilotingRollData(
                                violation.getId(), 2, "fallen on");
                        if (violation instanceof Dropship) {
                            violation = entity;
                            prd = null;
                        }
                        Coords targetDest = Compute.getValidDisplacement(game,
                                                                         violation.getId(), dest, direction);
                        if (targetDest != null) {
                            vPhaseReport.addAll(doEntityDisplacement(violation,
                                                                     dest, targetDest, prd));
                            
                            
                            entityUpdate(violation.getId());
                        } else {
                            
                            
                            
                            vPhaseReport.addAll(destroyEntity(violation,
                                                              "impossible displacement",
                                                              violation instanceof Mech,
                                                              violation instanceof Mech));
                        }
                    }
                    return vPhaseReport;
                }
            } else {
                
                r = new Report(2213);
                r.add(toHit.getDesc());
                vPhaseReport.add(r);
            }
            
            
            Coords targetDest = Compute.getValidDisplacement(game,
                                                             entity.getId(), dest, direction);
            if (targetDest != null) {
                vPhaseReport.addAll(doEntityFallsInto(entity,
                                                      entitySrcElevation, src, targetDest,
                                                      new PilotingRollData(entity.getId(),
                                                                           TargetRoll.IMPOSSIBLE, "pushed off a cliff"),
                                                      false));
                
                entityUpdate(entity.getId());
            } else {
                
                
                
                vPhaseReport.addAll(destroyEntity(entity,
                                                  "impossible displacement", entity instanceof Mech,
                                                  entity instanceof Mech));
            }
        } else {
            
            vPhaseReport
                    .addAll(doEntityFall(entity, dest, fallElevation, roll));
            Entity violation = Compute.stackingViolation(game, entity.getId(),
                                                         dest);
            if (violation != null) {
                PilotingRollData prd = new PilotingRollData(violation.getId(),
                                                            0, "domino effect");
                if (violation instanceof Dropship) {
                    violation = entity;
                    prd = null;
                }
                
                Coords targetDest = Compute.getValidDisplacement(game,
                                                                 violation.getId(), dest, direction);
                vPhaseReport.addAll(doEntityDisplacement(violation, dest,
                                                         targetDest, prd));
                
                if (!game.getOutOfGameEntitiesVector().contains(violation)) {
                    entityUpdate(violation.getId());
                }
            }
        }
        return vPhaseReport;
    }

    
    private Vector<Report> doEntityDisplacement(Entity entity, Coords src,
            Coords dest, PilotingRollData roll) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        Report r;
        if (!game.getBoard().contains(dest)) {
            
            entity.setPosition(dest);
            if (!entity.isDoomed()) {
                
                int turnsRemoved = game.removeSpecificEntityTurnsFor(entity);
                
                if ((game.getPhase() == Phase.PHASE_MOVEMENT)
                        && !entity.isDone() && (turnsRemoved == 0)) {
                    game.removeTurnFor(entity);
                    send(createTurnVectorPacket());
                } else if (turnsRemoved > 0) {
                    send(createTurnVectorPacket());
                }
                game.removeEntity(entity.getId(),
                        IEntityRemovalConditions.REMOVE_PUSHED);
                send(createRemoveEntityPacket(entity.getId(),
                        IEntityRemovalConditions.REMOVE_PUSHED));
                
                r = new Report(2230);
                r.subject = entity.getId();
                r.addDesc(entity);
                vPhaseReport.add(r);
                
            }
            return vPhaseReport;
        }
        final IHex srcHex = game.getBoard().getHex(src);
        final IHex destHex = game.getBoard().getHex(dest);
        final int direction = src.direction(dest);

        
        if ((srcHex == null) || (destHex == null)) {
            System.err.println("Can not displace " + entity.getShortName()
                               + " from " + src + " to " + dest + '.');
            return vPhaseReport;
        }
        int bldgElev = destHex.containsTerrain(Terrains.BLDG_ELEV) ? destHex
                .terrainLevel(Terrains.BLDG_ELEV) : 0;
        int fallElevation = entity.elevationOccupied(srcHex)
                            - (entity.elevationOccupied(destHex) + bldgElev);
        if (fallElevation > 1) {
            if (roll == null) {
                roll = entity.getBasePilotingRoll();
            }
            if (!(entity.isAirborneVTOLorWIGE())) {
                vPhaseReport.addAll(doEntityFallsInto(entity,
                        entity.getElevation(), src, dest, roll, true));
            } else {
                entity.setPosition(dest);
            }
            return vPhaseReport;
        }
        
        boolean wasStuck = entity.isStuck();
        entity.setStuck(false);
        int oldElev = entity.getElevation();
        
        entity.setPosition(dest);
        entity.setElevation(entity.calcElevation(srcHex, destHex));
        Building bldg = game.getBoard().getBuildingAt(dest);
        if (bldg != null) {
            if (destHex.terrainLevel(Terrains.BLDG_ELEV) > oldElev) {
                
                passBuildingWall(
                        entity,
                        game.getBoard().getBuildingAt(dest),
                        src,
                        dest,
                        1,
                        "displaced into",
                        Math.abs(entity.getFacing() - src.direction(dest)) == 3,
                        entity.moved, true);
            }
            checkBuildingCollapseWhileMoving(bldg, entity, dest);
        }
        if (destHex.containsTerrain(Terrains.MAGMA, 1)) {
            int d6 = Compute.d6(1);
            r = new Report(2395);
            r.addDesc(entity);
            r.add(d6);
            r.subject = entity.getId();
            addReport(r);
            if (d6 == 6) {
                destHex.removeTerrain(Terrains.MAGMA);
                destHex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.MAGMA, 2));
                sendChangedHex(dest);
                for (Entity en : game.getEntitiesVector(dest)) {
                    if (en != entity) {
                        doMagmaDamage(en, false);
                    }
                }
            }
        }
        Entity violation = Compute
                .stackingViolation(game, entity.getId(), dest);
        if (violation == null) {
            
            r = new Report(2235);
            r.indent();
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(dest.getBoardNum(), true);
            vPhaseReport.add(r);
        } else {
            
            r = new Report(2240);
            r.indent();
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(dest.getBoardNum(), true);
            r.addDesc(violation);
            vPhaseReport.add(r);
        }
        
        vPhaseReport.addAll(doEntityDisplacementMinefieldCheck(entity, src,
                                                               dest, entity.getElevation()));
        vPhaseReport.addAll(doSetLocationsExposure(entity, destHex, false,
                                                   entity.getElevation()));
        if (destHex.containsTerrain(Terrains.BLDG_ELEV)
            && (entity.getElevation() == 0)) {
            bldg = game.getBoard().getBuildingAt(dest);
            if (bldg.rollBasement(dest, game.getBoard(), vPhaseReport)) {
                sendChangedHex(dest);
                Vector<Building> buildings = new Vector<Building>();
                buildings.add(bldg);
                sendChangedBuildings(buildings);
            }
        }
        int waterDepth = destHex.terrainLevel(Terrains.WATER);
        
        if ((waterDepth > 0)
            && !(entity instanceof Mech)
            && !(entity instanceof Protomech)
            && !((entity.getRunMP() > 0) && (entity.getMovementMode() == EntityMovementMode.HOVER))
            && (entity.getMovementMode() != EntityMovementMode.HYDROFOIL)
            && (entity.getMovementMode() != EntityMovementMode.NAVAL)
            && (entity.getMovementMode() != EntityMovementMode.SUBMARINE)
            && (entity.getMovementMode() != EntityMovementMode.INF_UMU)) {
            vPhaseReport.addAll(destroyEntity(entity, "a watery grave", false));
        }
        
        if (wasStuck && (entity instanceof Mech) && !entity.isProne()) {
            if (roll == null) {
                roll = entity.getBasePilotingRoll();
            }
            vPhaseReport.addAll(doEntityFall(entity, dest, 0, roll));
        }
        
        vPhaseReport.addAll(doEntityDisplacementBogDownCheck(entity, dest,
                                                             entity.getElevation()));

        if (roll != null) {
            game.addPSR(roll);
        }

        if ((waterDepth > 0)
                && !(entity.getMovementMode() == EntityMovementMode.HOVER)) {
            PilotingRollData waterRoll = entity.checkWaterMove(waterDepth,
                    entity.moved);
            if (waterRoll.getValue() != TargetRoll.CHECK_FALSE) {
                doSkillCheckInPlace(entity, waterRoll);
            }
        }
        
        entityUpdate(entity.getId());

        if (violation != null) {
            
            
            
            MovePath stepForward = new MovePath(game, violation);
            MovePath stepBackwards = new MovePath(game, violation);
            stepForward.addStep(MoveStepType.FORWARDS);
            stepBackwards.addStep(MoveStepType.BACKWARDS);
            if ((direction != violation.getFacing())
                && (direction != ((violation.getFacing() + 3) % 6))
                && !entity.getIsJumpingNow()
                && (stepForward.isMoveLegal() || stepBackwards
                    .isMoveLegal())) {
                
                int result = Compute.d6(2);
                roll = entity.getBasePilotingRoll();

                r = new Report(2351);
                r.indent(2);
                r.subject = violation.getId();
                r.addDesc(violation);
                r.add(roll.getValue());
                r.add(result);
                vPhaseReport.add(r);
                if (result < roll.getValue()) {
                    r.choose(false);
                    Vector<Report> newReports = doEntityDisplacement(violation,
                            dest, dest.translated(direction),
                            new PilotingRollData(violation.getId(),
                                    TargetRoll.AUTOMATIC_FAIL,
                                    "failed to step out of a "
                                            + "domino effect"));
                    for (Report newReport : newReports) {
                        newReport.indent(3);
                    }
                    vPhaseReport.addAll(newReports);
                } else {
                    r.choose(true);
                    sendDominoEffectCFR(violation);
                    synchronized (cfrPacketQueue) {
                        try {
                            cfrPacketQueue.wait();
                        } catch (InterruptedException e) {
                            
                        }
                        if (cfrPacketQueue.size() > 0) {
                            ReceivedPacket rp = cfrPacketQueue.poll();
                            int cfrType = (int) rp.packet.getData()[0];
                            
                            if (cfrType != Packet.COMMAND_CFR_DOMINO_EFFECT) {
                                System.err
                                        .println("Excepted a "
                                                + "COMMAND_CFR_DOMINO_EFFECT CFR packet, "
                                                + "received: " + cfrType);
                                throw new IllegalStateException();
                            }
                            MovePath mp = (MovePath) rp.packet.getData()[1];
                            
                            if (mp != null) {
                                
                                r = new Report(2352);
                                r.indent(3);
                                r.subject = violation.getId();
                                r.addDesc(violation);
                                if (mp.getLastStep().getType() == MoveStepType.FORWARDS) {
                                    r.choose(false);
                                } else {
                                    r.choose(true);
                                }
                                r.add(mp.getLastStep().getPosition()
                                        .getBoardNum());
                                vPhaseReport.add(r);
                                
                                violation.setPosition(mp.getFinalCoords());
                                violation.mpUsed += mp.getMpUsed();
                                violation.moved = mp.getLastStepMovementType();
                            } else { 
                                r = new Report(2358);
                                r.indent(3);
                                r.subject = violation.getId();
                                r.addDesc(violation);
                                vPhaseReport.add(r);
                                vPhaseReport.addAll(doEntityDisplacement(
                                        violation, dest,
                                        dest.translated(direction), null));
                            }
                        } else { 
                            vPhaseReport.addAll(doEntityDisplacement(violation,
                                    dest, dest.translated(direction),
                                    new PilotingRollData(violation.getId(), 0,
                                            "domino effect")));
                        }
                    }
                }
            } else { 
                r = new Report(2359);
                r.indent(2);
                r.subject = violation.getId();
                r.addDesc(violation);
                vPhaseReport.add(r);
                vPhaseReport.addAll(doEntityDisplacement(violation, dest, dest
                        .translated(direction),
                        new PilotingRollData(violation.getId(), 0,
                                "domino effect")));

            }
            
            
            if (!game.isOutOfGame(violation)) {
                entityUpdate(violation.getId());
            }
        }
        return vPhaseReport;
    }

    private void sendDominoEffectCFR(Entity e) {
        send(e.getOwnerId(), new Packet(Packet.COMMAND_CLIENT_FEEDBACK_REQUEST,
                new Object[] { Packet.COMMAND_CFR_DOMINO_EFFECT, e.getId() }));
    }

    private void sendAMSAssignCFR(Entity e, Mounted ams,
            ArrayList<WeaponAttackAction> waas) {
        send(e.getOwnerId(),
                new Packet(Packet.COMMAND_CLIENT_FEEDBACK_REQUEST,
                        new Object[] { Packet.COMMAND_CFR_AMS_ASSIGN,
                                e.getId(), e.getEquipmentNum(ams), waas }));
    }

    private Vector<Report> doEntityDisplacementMinefieldCheck(Entity entity,
            Coords src, Coords dest, int elev) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        boolean boom = checkVibrabombs(entity, dest, true, vPhaseReport);
        if (game.containsMinefield(dest)) {
            boom = enterMinefield(entity, dest, elev, true, vPhaseReport)
                    || boom;
        }
        if (boom) {
            resetMines();
        }

        return vPhaseReport;
    }

    private Vector<Report> doEntityDisplacementBogDownCheck(Entity entity,
            Coords c, int elev) {
        Vector<Report> vReport = new Vector<Report>();
        Report r;
        IHex destHex = game.getBoard().getHex(c);
        int bgMod = destHex.getBogDownModifier(entity.getMovementMode(),
                entity instanceof LargeSupportTank);
        if ((bgMod != TargetRoll.AUTOMATIC_SUCCESS)
                && (entity.getMovementMode() != EntityMovementMode.HOVER)
                && (entity.getMovementMode() != EntityMovementMode.WIGE)
                && (elev == 0)) {
            PilotingRollData roll = entity.getBasePilotingRoll();
            roll.append(new PilotingRollData(entity.getId(), bgMod,
                    "avoid bogging down"));
            int stuckroll = Compute.d6(2);
            if (stuckroll < roll.getValue()) {
                entity.setStuck(true);
                r = new Report(2081);
                r.subject = entity.getId();
                r.add(entity.getDisplayName(), true);
                vReport.add(r);
                
                vReport.addAll(checkQuickSand(c));
            }

        }
        return vReport;
    }

    
    private void receiveDeployment(Packet packet, int connId) {
        Entity entity = game.getEntity(packet.getIntValue(0));
        Coords coords = (Coords) packet.getObject(1);
        int nFacing = packet.getIntValue(2);
        int elevation = packet.getIntValue(3);

        
        int loadedCount = packet.getIntValue(4);
        Vector<Entity> loadVector = new Vector<Entity>();
        for (int i = 0; i < loadedCount; i++) {
            int loadedId = packet.getIntValue(6 + i);
            loadVector.addElement(game.getEntity(loadedId));
        }

        
        if (game.getPhase() != IGame.Phase.PHASE_DEPLOYMENT) {
            System.err
                    .println("error: server got deployment packet in wrong phase");
            return;
        }

        
        final boolean assaultDrop = packet.getBooleanValue(5);
        
        GameTurn turn = game.getTurn();
        if (game.isPhaseSimultaneous()) {
            turn = game.getTurnForPlayer(connId);
        }
        if ((turn == null) || !turn.isValid(connId, entity, game)
            || !(game.getBoard().isLegalDeployment(coords,
                                                   entity.getStartingPos()) || (assaultDrop
                                                                                && game.getOptions().booleanOption
                ("assault_drop") && entity
                                                                                        .canAssaultDrop()))) {
            String msg = "error: server got invalid deployment packet from "
                         + "connection " + connId;
            if (entity != null) {
                msg += ", Entity: " + entity.getShortName();
            } else {
                msg += ", Entity was null!";
            }
            System.err.println(msg);
            send(connId, createTurnVectorPacket());
            send(connId, createTurnIndexPacket());
            return;
        }

        
        processDeployment(entity, coords, nFacing, elevation, loadVector,
                          assaultDrop);

        
        if (doBlind()) {
            updateVisibilityIndicator(null);
        }

        endCurrentTurn(entity);
    }

    
    private void processDeployment(Entity entity, Coords coords, int nFacing,
                                   int elevation, Vector<Entity> loadVector, boolean assaultDrop) {

        for (Entity loaded : loadVector) {
            if (loaded.getTransportId() != Entity.NONE) {
                
                continue;
            }
            if (loaded.getPosition() != null) {
                
                System.err.println("error: " + entity
                                   + " can not load entity #" + loaded);
                break;
            }
            
            loadUnit(entity, loaded, loaded.getTargetBay());
        }

        
        if ((entity instanceof Aero) && game.useVectorMove()) {
            Aero a = (Aero) entity;
            if (a.getCurrentVelocityActual() > 0) {
                int[] v = {0, 0, 0, 0, 0, 0};
                v[nFacing] = a.getCurrentVelocityActual();
                entity.setVectors(v);
            }
        }

        entity.setPosition(coords);
        entity.setFacing(nFacing);
        entity.setSecondaryFacing(nFacing);
        IHex hex = game.getBoard().getHex(coords);
        if (assaultDrop) {
            entity.setAltitude(1);
            
            
            
            entity.setAssaultDropInProgress(true);
        } else if ((entity instanceof VTOL)
                && (entity.getExternalUnits().size() <= 0)) {
            
            
            
            entity.setElevation((hex.ceiling() - hex.surface()) + 1);
            while ((Compute.stackingViolation(game, entity, coords, null) != null)
                   && (entity.getElevation() <= 50)) {
                entity.setElevation(entity.getElevation() + 1);
            }
            if (entity.getElevation() > 50) {
                throw new IllegalStateException(
                        "Entity #"
                        + entity.getId()
                        + " appears to be in an infinite loop trying to get a legal elevation.");
            }
        } else if (entity instanceof Aero) {
            
            
            
            if (entity.isAirborne()) {
                entity.setElevation(0);
                elevation = 0;
            }
            if (!game.getBoard().inSpace()) {
                
                
                Aero a = (Aero) entity;
                if (a.isSpheroid() || game.getPlanetaryConditions().isVacuum()) {
                    a.setCurrentVelocity(0);
                    a.setNextVelocity(0);
                }
                
                
                if (game.getBoard().inAtmosphere()
                    && (a.getAltitude() <= hex.ceiling(true))) {
                    
                    a.setAltitude(hex.ceiling(true) + 1);
                }
            }
        } else if (entity.getMovementMode() == EntityMovementMode.SUBMARINE) {
            
            
            entity.setElevation(0);
        } else if ((entity.getMovementMode() == EntityMovementMode.HOVER)
                   || (entity.getMovementMode() == EntityMovementMode.WIGE)
                   || (entity.getMovementMode() == EntityMovementMode.NAVAL)
                   || (entity.getMovementMode() == EntityMovementMode.HYDROFOIL)) {
            
            
            entity.setElevation(0);
        } else if (hex.containsTerrain(Terrains.ICE)) {
            entity.setElevation(0);
        } else {
            Building bld = game.getBoard().getBuildingAt(entity.getPosition());
            if ((bld != null) && (bld.getType() == Building.WALL)) {
                entity.setElevation(hex.terrainLevel(Terrains.BLDG_ELEV));
            }

        }
        
        
        
        
        entity.setElevation(entity.getElevation() + elevation);

        
        
        Building bldg = game.getBoard().getBuildingAt(entity.getPosition());
        if ((bldg != null)) {
            if (bldg.rollBasement(entity.getPosition(), game.getBoard(),
                                  vPhaseReport)) {
                sendChangedHex(entity.getPosition());
                Vector<Building> buildings = new Vector<Building>();
                buildings.add(bldg);
                sendChangedBuildings(buildings);
            }
            boolean collapse = checkBuildingCollapseWhileMoving(bldg, entity,
                                                                entity.getPosition());
            if (collapse) {
                addAffectedBldg(bldg, collapse);
            }
        }

        entity.setDone(true);
        entity.setDeployed(true);
        entityUpdate(entity.getId());
    }

    
    @SuppressWarnings("unchecked")
    private void receiveArtyAutoHitHexes(Packet packet, int connId) {
        PlayerIDandList<Coords> artyAutoHitHexes = (PlayerIDandList<Coords>) packet
                .getObject(0);

        int playerId = artyAutoHitHexes.getPlayerID();

        
        if (game.getPhase() != IGame.Phase.PHASE_SET_ARTYAUTOHITHEXES) {
            System.err
                    .println("error: server got set artyautohithexespacket in wrong phase");
            return;
        }
        game.getPlayer(playerId).setArtyAutoHitHexes(artyAutoHitHexes);

        for (Coords coord : artyAutoHitHexes) {
            game.getBoard().addSpecialHexDisplay(
                    coord,
                    new SpecialHexDisplay(
                            SpecialHexDisplay.Type.ARTILLERY_AUTOHIT,
                            SpecialHexDisplay.NO_ROUND, getPlayer(playerId),
                            "Artillery auto hit hex, for "
                            + getPlayer(playerId).getName(),
                            SpecialHexDisplay.SHD_OBSCURED_TEAM));
        }
        endCurrentTurn(null);
    }

    
    @SuppressWarnings("unchecked")
    private void receiveDeployMinefields(Packet packet, int connId) {
        Vector<Minefield> minefields = (Vector<Minefield>) packet.getObject(0);

        
        if (game.getPhase() != IGame.Phase.PHASE_DEPLOY_MINEFIELDS) {
            System.err
                    .println("error: server got deploy minefields packet in wrong phase");
            return;
        }

        
        processDeployMinefields(minefields);
        endCurrentTurn(null);
    }

    
    private void processDeployMinefields(Vector<Minefield> minefields) {
        int playerId = IPlayer.PLAYER_NONE;
        for (int i = 0; i < minefields.size(); i++) {
            Minefield mf = minefields.elementAt(i);
            playerId = mf.getPlayerId();

            game.addMinefield(mf);
            if (mf.getType() == Minefield.TYPE_VIBRABOMB) {
                game.addVibrabomb(mf);
            }
        }

        IPlayer player = game.getPlayer(playerId);
        if (null != player) {
            int teamId = player.getTeam();

            if (teamId != IPlayer.TEAM_NONE) {
                Enumeration<Team> teams = game.getTeams();
                while (teams.hasMoreElements()) {
                    Team team = teams.nextElement();
                    if (team.getId() == teamId) {
                        Enumeration<IPlayer> players = team.getPlayers();
                        while (players.hasMoreElements()) {
                            IPlayer teamPlayer = players.nextElement();
                            if (teamPlayer.getId() != player.getId()) {
                                send(teamPlayer.getId(), new Packet(
                                        Packet.COMMAND_DEPLOY_MINEFIELDS,
                                        minefields));
                            }
                            teamPlayer.addMinefields(minefields);
                        }
                        break;
                    }
                }
            } else {
                player.addMinefields(minefields);
            }
        }
    }

    
    private void receiveGroundToAirHexSelectPacket(Packet packet, int connId) {
        Integer targetId = (Integer)packet.getObject(0);
        Integer attackerId = (Integer)packet.getObject(1);
        Coords pos = (Coords)packet.getObject(2);
        game.getEntity(targetId).setPlayerPickedPassThrough(attackerId, pos);
    }

    
    @SuppressWarnings("unchecked")
    private void receiveAttack(Packet packet, int connId) {
        Entity entity = game.getEntity(packet.getIntValue(0));
        Vector<EntityAction> vector = (Vector<EntityAction>) packet
                .getObject(1);

        
        if ((game.getPhase() != IGame.Phase.PHASE_FIRING)
            && (game.getPhase() != IGame.Phase.PHASE_PHYSICAL)
            && (game.getPhase() != IGame.Phase.PHASE_TARGETING)
            && (game.getPhase() != IGame.Phase.PHASE_OFFBOARD)) {
            System.err
                    .println("error: server got attack packet in wrong phase");
            return;
        }

        
        GameTurn turn = game.getTurn();
        if (game.isPhaseSimultaneous()) {
            turn = game.getTurnForPlayer(connId);
        }
        if ((turn == null) || !turn.isValid(connId, entity, game)) {
            String msg = "error: server got invalid attack packet from "
                         + "connection " + connId;
            if (entity != null) {
                msg += ", Entity: " + entity.getShortName();
            } else {
                msg += ", Entity was null!";
            }
            System.err.println(msg);
            send(connId, createTurnVectorPacket());
            send(connId, createTurnIndexPacket());
            return;
        }

        
        processAttack(entity, vector);

        
        if (doBlind()) {
            updateVisibilityIndicator(null);
        }

        endCurrentTurn(entity);
    }

    
    private void processAttack(Entity entity, Vector<EntityAction> vector) {

        
        if (vector == null) {
            vector = new Vector<EntityAction>(0);
        }

        
        boolean setDone = !((game.getTurn() instanceof GameTurn.TriggerAPPodTurn) || (game
                .getTurn() instanceof GameTurn.TriggerBPodTurn));
        for (EntityAction ea : vector) {
            
            if (ea.getEntityId() != entity.getId()) {
                System.err.println("error: attack packet has wrong attacker");
                continue;
            }
            if (ea instanceof PushAttackAction) {
                
                PushAttackAction paa = (PushAttackAction) ea;
                entity.setDisplacementAttack(paa);
                game.addCharge(paa);
            } else if (ea instanceof DodgeAction) {
                entity.dodging = true;
            } else if (ea instanceof SpotAction) {
                entity.setSpotting(true);
                entity.setSpotTargetId(((SpotAction) ea).getTargetId());
            } else {
                
                game.addAction(ea);
            }

            
            
            if (ea instanceof WeaponAttackAction) {
                final WeaponAttackAction waa = (WeaponAttackAction) ea;
                final String weaponName = entity
                        .getEquipment(waa.getWeaponId()).getType()
                        .getInternalName();

                if (Infantry.SWARM_MEK.equals(weaponName)
                    || Infantry.LEG_ATTACK.equals(weaponName)) {

                    
                    final Entity target = game.getEntity(waa.getTargetId());
                    for (Mounted equip : target.getMisc()) {
                        if (equip.getType().hasFlag(MiscType.F_AP_POD)
                            && equip.canFire()) {

                            
                            
                            
                            game.insertNextTurn(new GameTurn.TriggerAPPodTurn(
                                    target.getOwnerId(), target.getId()));
                            send(createTurnVectorPacket());

                            
                            break;

                        } 

                    } 

                    for (Mounted weapon : target.getWeaponList()) {
                        if (weapon.getType().hasFlag(WeaponType.F_B_POD)
                            && weapon.canFire()) {

                            
                            
                            
                            game.insertNextTurn(new GameTurn.TriggerBPodTurn(
                                    target.getOwnerId(), target.getId(),
                                    weaponName));
                            send(createTurnVectorPacket());

                            
                            break;

                        } 
                    } 
                } 
            }

            
            if (ea instanceof BreakGrappleAttackAction) {
                final BreakGrappleAttackAction bgaa = (BreakGrappleAttackAction) ea;
                final Entity att = (game.getEntity(bgaa.getEntityId()));
                if (att.isGrappleAttacker()) {
                    final Entity def = (game.getEntity(bgaa.getTargetId()));
                    
                    if (def.isDone()) {
                        game.removeActionsFor(def.getId());
                    } else {
                        game.removeTurnFor(def);
                        def.setDone(true);
                    }
                    
                    game.insertNextTurn(new GameTurn.CounterGrappleTurn(def
                                                                                .getOwnerId(), def.getId()));
                    send(createTurnVectorPacket());
                }
            }
            if (ea instanceof ArtilleryAttackAction) {
                boolean firingAtNewHex = false;
                final ArtilleryAttackAction aaa = (ArtilleryAttackAction) ea;
                final Entity firingEntity = game.getEntity(aaa.getEntityId());
                for (Enumeration<AttackHandler> j = game.getAttacks(); !firingAtNewHex
                                                                       && j.hasMoreElements(); ) {
                    WeaponHandler wh = (WeaponHandler) j.nextElement();
                    if (wh.waa instanceof ArtilleryAttackAction) {
                        ArtilleryAttackAction oaaa = (ArtilleryAttackAction) wh.waa;
                        if ((oaaa.getEntityId() == aaa.getEntityId())
                            && !oaaa.getTarget(game)
                                    .getPosition()
                                    .equals(aaa.getTarget(game)
                                               .getPosition())) {
                            firingAtNewHex = true;
                        }
                    }
                }
                if (firingAtNewHex) {
                    clearArtillerySpotters(firingEntity.getId(),
                                           aaa.getWeaponId());
                }
                Iterator<Entity> spotters = game
                        .getSelectedEntities(new EntitySelector() {
                            public int player = firingEntity.getOwnerId();
                            public Targetable target = aaa.getTarget(game);

                            public boolean accept(Entity entity) {
                                LosEffects los = LosEffects.calculateLos(game,
                                                                         entity.getId(), target);
                                return ((player == entity.getOwnerId())
                                        && !(los.isBlocked()) && entity
                                        .isActive());
                            }
                        });
                Vector<Integer> spotterIds = new Vector<Integer>();
                while (spotters.hasNext()) {
                    Integer id = new Integer(spotters.next().getId());
                    spotterIds.addElement(id);
                }
                aaa.setSpotterIds(spotterIds);
            }

            
            if (ea instanceof ClubAttackAction) {
                ClubAttackAction caa = (ClubAttackAction) ea;
                Mounted club = caa.getClub();
                club.restore();
            }

            
            if (ea instanceof TriggerAPPodAction) {
                TriggerAPPodAction tapa = (TriggerAPPodAction) ea;
                Mounted pod = entity.getEquipment(tapa.getPodId());
                pod.setUsedThisRound(true);
            }
            
            if (ea instanceof TriggerBPodAction) {
                TriggerBPodAction tba = (TriggerBPodAction) ea;
                Mounted pod = entity.getEquipment(tba.getPodId());
                pod.setUsedThisRound(true);
            }

            
            if (ea instanceof SearchlightAttackAction) {
                boolean hexesAdded =
                        ((SearchlightAttackAction) ea).setHexesIlluminated(game);
                
                
                if (hexesAdded) {
                    send(createIlluminatedHexesPacket());
                }
            }
        }

        
        
        if (setDone) {
            entity.setDone(true);
        }
        entityUpdate(entity.getId());

        Packet p = createAttackPacket(vector, 0);
        if (game.isPhaseSimultaneous()) {
            
            for (IPlayer player : game.getPlayersVector()) {
                if (player.canSeeAll() || player.isObserver()
                    || (entity.getOwnerId() == player.getId())) {
                    send(player.getId(), p);
                }
            }
        } else {
            
            
            
            send(p);
        }
    }

    
    private void assignAMS() {

        
        Hashtable<Entity, Vector<WeaponHandler>> htAttacks = new Hashtable<Entity, Vector<WeaponHandler>>();
        for (AttackHandler ah : game.getAttacksVector()) {
            WeaponHandler wh = (WeaponHandler) ah;
            WeaponAttackAction waa = wh.waa;
            
            
            if (waa instanceof ArtilleryAttackAction) {
                continue;
            }
            Mounted weapon = game.getEntity(waa.getEntityId()).getEquipment(
                    waa.getWeaponId());

            
            if (Targetable.TYPE_ENTITY != waa.getTargetType()) {
                continue;
            }

            
            if (wh.roll < wh.toHit.getValue()) {
                continue;
            }

            
            if (weapon.getType().hasFlag(WeaponType.F_MISSILE)) {
                Entity target = game.getEntity(waa.getTargetId());
                Vector<WeaponHandler> v = htAttacks.get(target);
                if (v == null) {
                    v = new Vector<WeaponHandler>();
                    htAttacks.put(target, v);
                }
                v.addElement(wh);
            }
        }

        
        for (Entity e : htAttacks.keySet()) {
            Vector<WeaponHandler> vAttacks = htAttacks.get(e);
            
            if (game.getOptions().booleanOption("auto_ams")) {
                e.assignAMS(vAttacks);
            } else { 
                
                HashSet<WeaponAttackAction> amsTargets =
                        new HashSet<WeaponAttackAction>();
                
                for (Mounted ams : e.getActiveAMS()) {
                    
                    ArrayList<WeaponAttackAction> vAttacksInArc =
                            new ArrayList<WeaponAttackAction>(vAttacks.size());
                    for (WeaponHandler wr : vAttacks) {
                        if (!amsTargets.contains(wr.waa)
                                && Compute.isInArc(game, e.getId(),
                                        e.getEquipmentNum(ams),
                                        game.getEntity(wr.waa.getEntityId()))) {
                            vAttacksInArc.add(wr.waa);
                        }
                    }

                    
                    if (vAttacksInArc.size() < 1) {
                        continue;
                    }

                    WeaponAttackAction targetedWAA = null;

                    if (ams.curMode().equals("Automatic")) {
                        targetedWAA = Compute.getHighestExpectedDamage(game,
                                vAttacksInArc, true);
                    } else {
                        
                        sendAMSAssignCFR(e, ams, vAttacksInArc);
                        synchronized (cfrPacketQueue) {
                            try {
                                cfrPacketQueue.wait();
                            } catch (InterruptedException ex) {
                                
                            }
                            if (cfrPacketQueue.size() > 0) {
                                ReceivedPacket rp = cfrPacketQueue.poll();
                                int cfrType = (int) rp.packet.getData()[0];
                                
                                if (cfrType != Packet.COMMAND_CFR_AMS_ASSIGN) {
                                    System.err.println("Excepted a "
                                            + "COMMAND_CFR_AMS_ASSIGN CFR "
                                            + "packet, received: " + cfrType);
                                    throw new IllegalStateException();
                                }
                                Integer waaIndex =
                                        (Integer)rp.packet.getData()[1];
                                if (waaIndex != null) {
                                    targetedWAA = vAttacksInArc.get(waaIndex);
                                }
                            }
                        }
                    }

                    if (targetedWAA != null) {
                        targetedWAA.addCounterEquipment(ams);
                        amsTargets.add(targetedWAA);
                    }
                }
            }
        }
    }

    
    private void resolveWhatPlayersCanSeeWhatUnits() {
        List<ECMInfo> allECMInfo = null;
        if (game.getOptions().booleanOption("tacops_sensors")) {
            allECMInfo = ComputeECM.computeAllEntitiesECMInfo(game
                    .getEntitiesVector());
        }
        Map<EntityTargetPair, LosEffects> losCache = new HashMap<>();
        for (Entity entity : game.getEntitiesVector()) {
            
            entity.clearSeenBy();
            entity.clearDetectedBy();
            
            for (IPlayer p : whoCanSee(entity, false, losCache)) {
                entity.addBeenSeenBy(p);
            }
            
            for (IPlayer p : whoCanDetect(entity, allECMInfo, losCache)) {
                    entity.addBeenDetectedBy(p);
            }
        }
    }

    
    private void resolveAllButWeaponAttacks() {

        Vector<EntityAction> triggerPodActions = new Vector<EntityAction>();
        
        for (Enumeration<EntityAction> i = game.getActions(); i
                .hasMoreElements(); ) {
            EntityAction ea = i.nextElement();
            Entity entity = game.getEntity(ea.getEntityId());
            if (ea instanceof TorsoTwistAction) {
                TorsoTwistAction tta = (TorsoTwistAction) ea;
                if (entity.canChangeSecondaryFacing()) {
                    entity.setSecondaryFacing(tta.getFacing());
                }
            } else if (ea instanceof FlipArmsAction) {
                FlipArmsAction faa = (FlipArmsAction) ea;
                entity.setArmsFlipped(faa.getIsFlipped());
            } else if (ea instanceof FindClubAction) {
                resolveFindClub(entity);
            } else if (ea instanceof UnjamAction) {
                resolveUnjam(entity);
            } else if (ea instanceof ClearMinefieldAction) {
                resolveClearMinefield(entity,
                                      ((ClearMinefieldAction) ea).getMinefield());
            } else if (ea instanceof TriggerAPPodAction) {
                TriggerAPPodAction tapa = (TriggerAPPodAction) ea;

                
                if (!triggerPodActions.contains(tapa)) {
                    triggerAPPod(entity, tapa.getPodId());
                    triggerPodActions.addElement(tapa);
                } else {
                    System.err.print("AP Pod #");
                    System.err.print(tapa.getPodId());
                    System.err.print(" on ");
                    System.err.print(entity.getDisplayName());
                    System.err.println(" was already triggered this round!!");
                }
            } else if (ea instanceof TriggerBPodAction) {
                TriggerBPodAction tba = (TriggerBPodAction) ea;

                
                if (!triggerPodActions.contains(tba)) {
                    triggerBPod(entity, tba.getPodId(),
                                game.getEntity(tba.getTargetId()));
                    triggerPodActions.addElement(tba);
                } else {
                    System.err.print("B Pod #");
                    System.err.print(tba.getPodId());
                    System.err.print(" on ");
                    System.err.print(entity.getDisplayName());
                    System.err.println(" was already triggered this round!!");
                }
            } else if (ea instanceof SearchlightAttackAction) {
                SearchlightAttackAction saa = (SearchlightAttackAction) ea;
                addReport(saa.resolveAction(game));
            } else if (ea instanceof UnjamTurretAction) {
                if (entity instanceof Tank) {
                    ((Tank) entity).unjamTurret(((Tank) entity).getLocTurret());
                    ((Tank) entity)
                            .unjamTurret(((Tank) entity).getLocTurret2());
                    Report r = new Report(3033);
                    r.addDesc(entity);
                    addReport(r);
                } else {
                    System.err.println("Non-Tank tried to unjam turret");
                }
            } else if (ea instanceof RepairWeaponMalfunctionAction) {
                if (entity instanceof Tank) {
                    Mounted m = entity
                            .getEquipment(((RepairWeaponMalfunctionAction) ea)
                                                  .getWeaponId());
                    m.setJammed(false);
                    ((Tank) entity).getJammedWeapons().remove(m);
                    Report r = new Report(3034);
                    r.addDesc(entity);
                    r.add(m.getName());
                    addReport(r);
                } else {
                    System.err
                            .println("Non-Tank tried to repair weapon malfunction");
                }
            }
        }
    }

    
    private void resolveSelfDestructions() {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;
        for (Entity e : game.getEntitiesVector()) {
            if (e.getSelfDestructInitiated()) {
                r = new Report(6166, Report.PUBLIC);
                int target = e.getCrew().getPiloting();
                int roll = e.getCrew().rollPilotingSkill();
                r.subject = e.getId();
                r.addDesc(e);
                r.indent();
                r.add(target);
                r.add(roll);
                if (roll >= target) {
                    r.choose(true);
                } else {
                    r.choose(false);
                }
                vDesc.add(r);

                
                if (roll >= target) {
                    int engineRating = e.getEngine().getRating();
                    r = new Report(5400, Report.PUBLIC);
                    r.subject = e.getId();
                    r.indent(2);
                    vDesc.add(r);

                    if (e instanceof Mech) {
                        Mech mech = (Mech) e;
                        if (mech.isAutoEject()
                            && (!game.getOptions().booleanOption(
                                "conditional_ejection") || (game
                                                                    .getOptions().booleanOption(
                                        "conditional_ejection") && mech
                                                                    .isCondEjectEngine()))) {
                            vDesc.addAll(ejectEntity(e, true));
                        }
                    }

                    doFusionEngineExplosion(engineRating, e.getPosition(),
                                            vDesc, null);
                    Report.addNewline(vDesc);
                    r = new Report(5410, Report.PUBLIC);
                    r.subject = e.getId();
                    r.indent(2);
                    Report.addNewline(vDesc);
                    vDesc.add(r);
                }
                e.setSelfDestructInitiated(false);
            }
        }
        addReport(vDesc);
    }

    private void reportGhostTargetRolls() {
        
        
        
        Report r;
        for (Iterator<Entity> e = game.getEntities(); e.hasNext(); ) {
            Entity ent = e.next();
            if (ent.isDeployed() && ent.hasGhostTargets(false)) {
                r = new Report(3630);
                r.subject = ent.getId();
                r.addDesc(ent);
                int target = ent.getCrew().getPiloting() + 2;
                int roll = ent.getGhostTargetRoll();
                r.add(target);
                r.add(roll);
                if (roll >= target) {
                    r.choose(true);
                } else {
                    r.choose(false);
                }
                addReport(r);
            }
        }
    }

    private void reportLargeCraftECCMRolls() {
        
        
        
        if (!game.getBoard().inSpace()
            || !game.getOptions().booleanOption("stratops_ecm")) {
            return;
        }
        Report r;
        for (Iterator<Entity> e = game.getEntities(); e.hasNext(); ) {
            Entity ent = e.next();
            if (ent.isDeployed() && ent.isLargeCraft()) {
                r = new Report(3635);
                r.addDesc(ent);
                int target = ((Aero) ent).getECCMTarget();
                int roll = ((Aero) ent).getECCMRoll();
                r.add(roll);
                r.add(target);
                int mod = ((Aero) ent).getECCMBonus();
                r.add(mod);
                addReport(r);
            }
        }
    }

    private void resolveClearMinefield(Entity ent, Minefield mf) {

        if ((null == mf) || (null == ent) || ent.isDoomed()
            || ent.isDestroyed()) {
            return;
        }

        Coords pos = mf.getCoords();
        int clear = Minefield.CLEAR_NUMBER_INFANTRY;
        int boom = Minefield.CLEAR_NUMBER_INFANTRY_ACCIDENT;

        Report r;
        
        if ((ent instanceof BattleArmor)) {
            BattleArmor ba = (BattleArmor)ent;
            String mcmName = BattleArmor.MANIPULATOR_TYPE_STRINGS
                    [BattleArmor.MANIPULATOR_BASIC_MINE_CLEARANCE];
            if (ba.getLeftManipulatorName().equals(mcmName)) {
                clear = Minefield.CLEAR_NUMBER_BA_SWEEPER;
                boom = Minefield.CLEAR_NUMBER_BA_SWEEPER_ACCIDENT;
            }
            r = new Report(2246);
        } else {
            r = new Report(2245);
        }
        
        r.subject = ent.getId();
        r.add(ent.getShortName(), true);
        r.add(Minefield.getDisplayableName(mf.getType()));
        r.add(pos.getBoardNum(), true);
        addReport(r);

        if (clearMinefield(mf, ent, clear, boom, vPhaseReport)) {
            removeMinefield(mf);
        }
        
        resetMines();
    }

    
    private void resolveOnlyWeaponAttacks() {
        
        for (Enumeration<EntityAction> i = game.getActions(); i
                .hasMoreElements(); ) {
            EntityAction ea = i.nextElement();
            if (ea instanceof WeaponAttackAction) {
                WeaponAttackAction waa = (WeaponAttackAction) ea;
                Entity ae = game.getEntity(waa.getEntityId());
                Mounted m = ae.getEquipment(waa.getWeaponId());
                Weapon w = (Weapon) m.getType();
                
                waa.setOriginalTargetId(waa.getTargetId());
                waa.setOriginalTargetType(waa.getTargetType());
                AttackHandler ah = w.fire(waa, game, this);
                if (ah != null) {
                    ah.setStrafing(waa.isStrafing());
                    ah.setStrafingFirstShot(waa.isStrafingFirstShot());
                    game.addAttack(ah);
                    
                    if ((ae instanceof Aero)
                        && (waa.getAltitudeLoss(game) > ((Aero) ae)
                            .getAltLoss())) {
                        ((Aero) ae).setAltLoss(waa.getAltitudeLoss(game));
                    }
                }
            }
        }
        
        game.resetActions();
    }

    
    private void triggerAPPod(Entity entity, int podId) {

        
        Mounted mount = entity.getEquipment(podId);

        
        if (null == mount) {
            System.err.print("Expecting to find an AP Pod at ");
            System.err.print(podId);
            System.err.print(" on the unit, ");
            System.err.print(entity.getDisplayName());
            System.err.println(" but found NO equipment at all!!!");
            return;
        }
        EquipmentType equip = mount.getType();
        if (!(equip instanceof MiscType) || !equip.hasFlag(MiscType.F_AP_POD)) {
            System.err.print("Expecting to find an AP Pod at ");
            System.err.print(podId);
            System.err.print(" on the unit, ");
            System.err.print(entity.getDisplayName());
            System.err.print(" but found ");
            System.err.print(equip.getName());
            System.err.println(" instead!!!");
            return;
        }

        
        
        boolean oldFired = mount.isUsedThisRound();
        mount.setUsedThisRound(false);
        boolean canFire = mount.canFire();
        mount.setUsedThisRound(oldFired);
        if (!canFire) {
            System.err.print("Can not trigger the AP Pod at ");
            System.err.print(podId);
            System.err.print(" on the unit, ");
            System.err.print(entity.getDisplayName());
            System.err.println("!!!");
            return;
        }

        Report r;

        
        mount.setFired(true);
        r = new Report(3010);
        r.newlines = 0;
        r.subject = entity.getId();
        r.addDesc(entity);
        addReport(r);

        
        for (Entity target : game.getEntitiesVector(entity.getPosition())) {

            
            if ((target instanceof Infantry)
                && !(target instanceof BattleArmor)) {

                
                final int damage = Math.max(1, Compute.d6() - 1);

                
                addReport(damageEntity(target, new HitData(
                        Infantry.LOC_INFANTRY), damage));

                
                target.applyDamage();

            } 

            
            
            else if (!entity.equals(target)) {
                r = new Report(3020);
                r.indent(2);
                r.subject = target.getId();
                r.addDesc(target);
                addReport(r);
            }

        } 
    }

    
    private void triggerBPod(Entity entity, int podId, Entity target) {

        
        Mounted mount = entity.getEquipment(podId);

        
        if (null == mount) {
            System.err.print("Expecting to find an B Pod at ");
            System.err.print(podId);
            System.err.print(" on the unit, ");
            System.err.print(entity.getDisplayName());
            System.err.println(" but found NO equipment at all!!!");
            return;
        }
        EquipmentType equip = mount.getType();
        if (!(equip instanceof WeaponType)
            || !equip.hasFlag(WeaponType.F_B_POD)) {
            System.err.print("Expecting to find an B Pod at ");
            System.err.print(podId);
            System.err.print(" on the unit, ");
            System.err.print(entity.getDisplayName());
            System.err.print(" but found ");
            System.err.print(equip.getName());
            System.err.println(" instead!!!");
            return;
        }

        
        
        boolean oldFired = mount.isUsedThisRound();
        mount.setUsedThisRound(false);
        boolean canFire = mount.canFire();
        mount.setUsedThisRound(oldFired);
        if (!canFire) {
            System.err.print("Can not trigger the B Pod at ");
            System.err.print(podId);
            System.err.print(" on the unit, ");
            System.err.print(entity.getDisplayName());
            System.err.println("!!!");
            return;
        }

        Report r;

        
        mount.setFired(true);
        r = new Report(3011);
        r.newlines = 0;
        r.subject = entity.getId();
        r.addDesc(entity);
        addReport(r);

        
        if ((target instanceof Infantry) && !(target instanceof BattleArmor)) {

            
            final int damage = Compute.d6();

            
            addReport(damageEntity(target, new HitData(Infantry.LOC_INFANTRY),
                                   damage));

            
            target.applyDamage();

            
        } else if (target instanceof BattleArmor) {
            
            final int damage = 5;

            
            addReport(damageEntity(target, target.rollHitLocation(0, 0), damage));
            addReport(damageEntity(target, target.rollHitLocation(0, 0), damage));
            addReport(damageEntity(target, target.rollHitLocation(0, 0), damage));
            addReport(damageEntity(target, target.rollHitLocation(0, 0), damage));

            
            target.applyDamage();
        }

        
        
        else if (!entity.equals(target)) {
            r = new Report(3020);
            r.indent(2);
            r.subject = target.getId();
            r.addDesc(target);
            addReport(r);
        }
    }

    
    private void resolveUnjam(Entity entity) {
        Report r;
        final int TN = entity.getCrew().getGunnery() + 3;
        if (game.getOptions().booleanOption("unjam_uac")) {
            r = new Report(3026);
        } else {
            r = new Report(3025);
        }
        r.subject = entity.getId();
        r.addDesc(entity);
        addReport(r);
        for (Mounted mounted : entity.getTotalWeaponList()) {
            if (mounted.isJammed() && !mounted.isDestroyed()) {
                WeaponType wtype = (WeaponType) mounted.getType();
                if (wtype.getAmmoType() == AmmoType.T_AC_ROTARY) {
                    int roll = Compute.d6(2);
                    r = new Report(3030);
                    r.indent();
                    r.subject = entity.getId();
                    r.add(wtype.getName());
                    r.add(TN);
                    r.add(roll);
                    if (roll >= TN) {
                        r.choose(true);
                        mounted.setJammed(false);
                    } else {
                        r.choose(false);
                    }
                    addReport(r);
                }
                
                
                if (((wtype.getAmmoType() == AmmoType.T_AC_ULTRA)
                     || (wtype.getAmmoType() == AmmoType.T_AC_ULTRA_THB)
                     || (wtype.getAmmoType() == AmmoType.T_AC) || (wtype
                                                                           .getAmmoType() == AmmoType.T_LAC))
                    && game.getOptions().booleanOption("unjam_uac")) {
                    int roll = Compute.d6(2);
                    r = new Report(3030);
                    r.indent();
                    r.subject = entity.getId();
                    r.add(wtype.getName());
                    r.add(TN);
                    r.add(roll);
                    if (roll >= TN) {
                        r.choose(true);
                        mounted.setJammed(false);
                    } else {
                        r.choose(false);
                    }
                    addReport(r);
                }
            }
        }
    }

    private void resolveFindClub(Entity entity) {
        EquipmentType clubType = null;

        entity.setFindingClub(true);

        
        Coords coords = entity.getPosition();
        IHex curHex = game.getBoard().getHex(coords);

        Report r;

        
        if (curHex.terrainLevel(Terrains.ARMS) > 0) {
            clubType = EquipmentType.get("Limb Club");
            curHex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                    Terrains.ARMS, curHex.terrainLevel(Terrains.ARMS) - 1));
            sendChangedHex(entity.getPosition());
            r = new Report(3035);
            r.subject = entity.getId();
            r.addDesc(entity);
            addReport(r);
        }
        
        else if (curHex.terrainLevel(Terrains.LEGS) > 0) {
            clubType = EquipmentType.get("Limb Club");
            curHex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                    Terrains.LEGS, curHex.terrainLevel(Terrains.LEGS) - 1));
            sendChangedHex(entity.getPosition());
            r = new Report(3040);
            r.subject = entity.getId();
            r.addDesc(entity);
            addReport(r);
        }

        
        
        else if (Building.LIGHT < curHex.terrainLevel(Terrains.RUBBLE)) {

            
            
            
            boolean found = false;
            int roll = Compute.d6(2);
            switch (curHex.terrainLevel(Terrains.RUBBLE)) {
                case Building.MEDIUM:
                    if (roll >= 7) {
                        found = true;
                    }
                    break;
                case Building.HEAVY:
                    if (roll >= 6) {
                        found = true;
                    }
                    break;
                case Building.HARDENED:
                    if (roll >= 5) {
                        found = true;
                    }
                    break;
                case Building.WALL:
                    if (roll >= 13) {
                        found = true;
                    }
                    break;
                default:
                    
                    if (roll >= 4) {
                        found = true;
                    }
            }

            
            if (found) {
                clubType = EquipmentType.get("Girder Club");
                r = new Report(3045);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);
            } else {
                
                r = new Report(3050);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);
            }
        }

        
        else if (curHex.containsTerrain(Terrains.WOODS)
                 || curHex.containsTerrain(Terrains.JUNGLE)) {
            clubType = EquipmentType.get("Tree Club");
            r = new Report(3055);
            r.subject = entity.getId();
            r.addDesc(entity);
            addReport(r);
        }

        
        try {
            if (clubType != null) {
                entity.addEquipment(clubType, Entity.LOC_NONE);
            }
        } catch (LocationFullException ex) {
            
            r = new Report(3060);
            r.subject = entity.getId();
            r.addDesc(entity);
            addReport(r);
        }
    }

    
    public boolean tryIgniteHex(Coords c, int entityId, boolean bHotGun,
                                boolean bInferno, TargetRoll nTargetRoll, boolean bReportAttempt,
                                int accidentTarget, Vector<Report> vPhaseReport) {

        IHex hex = game.getBoard().getHex(c);
        Report r;

        
        if (hex == null) {
            return false;
        }

        
        if (!game.getOptions().booleanOption("tacops_start_fire")) {
            return false;
        }

        
        if (!hex.isIgnitable()) {
            return false;
        }

        
        if (accidentTarget > -1) {
            
            
            if (hex.containsTerrain(Terrains.SNOW)) {
                return false;
            }
            nTargetRoll.addModifier(2, "accidental");
            int accidentRoll = Compute.d6(2);
            r = new Report(3066);
            r.subject = entityId;
            r.add(accidentTarget);
            r.add(accidentRoll);
            r.indent(2);
            if (accidentRoll > accidentTarget) {
                r.choose(false);
                vPhaseReport.add(r);
                return false;
            }
            r.choose(true);
            vPhaseReport.add(r);
        }

        int terrainMod = hex.getIgnitionModifier();
        if (terrainMod != 0) {
            nTargetRoll.addModifier(terrainMod, "terrain");
        }

        
        Building bldg = game.getBoard().getBuildingAt(c);
        if (null != bldg) {
            nTargetRoll.addModifier(bldg.getType() - 1, "building");
        }

        
        int weatherMod = game.getPlanetaryConditions().getIgniteModifiers();
        if (weatherMod != 0) {
            nTargetRoll.addModifier(weatherMod, "conditions");
        }

        
        
        if ((hex.containsTerrain(Terrains.SNOW) || hex
                .containsTerrain(Terrains.ICE)) && (bHotGun || bInferno)) {
            boolean melted = false;
            int meltCheck = Compute.d6(2);
            if ((hex.terrainLevel(Terrains.SNOW) > 1) && (meltCheck == 12)) {
                melted = true;
            } else if (hex.containsTerrain(Terrains.ICE) && (meltCheck > 9)) {
                melted = true;
            } else if (hex.containsTerrain(Terrains.SNOW) && (meltCheck > 7)) {
                melted = true;
            }
            if (bInferno) {
                melted = true;
            }
            if (melted) {
                vPhaseReport.addAll(meltIceAndSnow(c, entityId));
                return false;
            }

        }

        
        
        if (bInferno && !game.getOptions().booleanOption("no_ignite_clear")) {
            nTargetRoll = new TargetRoll(0, "inferno");
        }

        
        if (game.getPlanetaryConditions().getWindStrength() > PlanetaryConditions.WI_STORM) {
            nTargetRoll = new TargetRoll(TargetRoll.AUTOMATIC_FAIL, "tornado");
        }

        
        if (hex.containsTerrain(Terrains.FIRE)) {
            if (bReportAttempt) {
                r = new Report(3065);
                r.indent(2);
                r.subject = entityId;
                vPhaseReport.add(r);
            }
        } else if (checkIgnition(c, nTargetRoll, bInferno, entityId,
                                 vPhaseReport)) {
            return true;
        }
        return false;
    }

    
    public boolean tryIgniteHex(Coords c, int entityId, boolean bHotGun,
                                boolean bInferno, TargetRoll nTargetRoll, int accidentTarget,
                                Vector<Report> vPhaseReport) {
        return tryIgniteHex(c, entityId, bHotGun, bInferno, nTargetRoll, false,
                            accidentTarget, vPhaseReport);
    }

    public Vector<Report> tryClearHex(Coords c, int nDamage, int entityId) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        IHex h = game.getBoard().getHex(c);
        if (h == null) {
            return vPhaseReport;
        }
        ITerrain woods = h.getTerrain(Terrains.WOODS);
        ITerrain jungle = h.getTerrain(Terrains.JUNGLE);
        ITerrain ice = h.getTerrain(Terrains.ICE);
        ITerrain magma = h.getTerrain(Terrains.MAGMA);
        Report r;
        int reportType = Report.HIDDEN;
        if (entityId == Entity.NONE) {
            reportType = Report.PUBLIC;
        }
        if (woods != null) {
            int tf = woods.getTerrainFactor() - nDamage;
            int level = woods.getLevel();
            if (tf <= 0) {
                h.removeTerrain(Terrains.WOODS);
                h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.ROUGH, 1));
                
                r = new Report(3090, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
            } else if ((tf <= 50) && (level > 1)) {
                h.removeTerrain(Terrains.WOODS);
                h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.WOODS, 1));
                woods = h.getTerrain(Terrains.WOODS);
                
                r = new Report(3085, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
            } else if ((tf <= 90) && (level > 2)) {
                h.removeTerrain(Terrains.WOODS);
                h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.WOODS, 2));
                woods = h.getTerrain(Terrains.WOODS);
                
                r = new Report(3082, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
            }
            woods.setTerrainFactor(tf);
        }
        if (jungle != null) {
            int tf = jungle.getTerrainFactor() - nDamage;
            int level = jungle.getLevel();
            if (tf < 0) {
                h.removeTerrain(Terrains.JUNGLE);
                h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.ROUGH, 1));
                
                r = new Report(3091, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
            } else if ((tf <= 50) && (level > 1)) {
                h.removeTerrain(Terrains.JUNGLE);
                h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.JUNGLE, 1));
                jungle = h.getTerrain(Terrains.JUNGLE);
                
                r = new Report(3086, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
            } else if ((tf <= 90) && (level > 2)) {
                h.removeTerrain(Terrains.JUNGLE);
                h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.JUNGLE, 2));
                jungle = h.getTerrain(Terrains.JUNGLE);
                
                r = new Report(3083, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
            }
            jungle.setTerrainFactor(tf);
        }
        if (ice != null) {
            int tf = ice.getTerrainFactor() - nDamage;
            if (tf <= 0) {
                
                r = new Report(3092, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
                vPhaseReport.addAll(resolveIceBroken(c));
            } else {
                ice.setTerrainFactor(tf);
            }
        }
        if ((magma != null) && (magma.getLevel() == 1)) {
            int tf = magma.getTerrainFactor() - nDamage;
            if (tf <= 0) {
                
                r = new Report(3093, reportType);
                r.subject = entityId;
                vPhaseReport.add(r);
                h.removeTerrain(Terrains.MAGMA);
                h.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.MAGMA, 2));
                for (Entity en : game.getEntitiesVector(c)) {
                    doMagmaDamage(en, false);
                }
            } else {
                magma.setTerrainFactor(tf);
            }
        }
        sendChangedHex(c);

        
        checkExplodeIndustrialZone(c, vPhaseReport);

        return vPhaseReport;
    }

    
    private void resolvePhysicalAttacks() {
        
        addReport(new Report(4000, Report.PUBLIC));

        
        for (Enumeration<AttackAction> i = game.getCharges(); i
                .hasMoreElements(); ) {
            game.addAction(i.nextElement());
        }
        game.resetCharges();

        
        for (Enumeration<AttackAction> i = game.getRams(); i.hasMoreElements(); ) {
            game.addAction(i.nextElement());
        }
        game.resetRams();

        
        for (Enumeration<AttackAction> i = game.getTeleMissileAttacks(); i
                .hasMoreElements(); ) {
            game.addAction(i.nextElement());
        }
        game.resetTeleMissileAttacks();

        
        cleanupPhysicalAttacks();

        
        for (Enumeration<EntityAction> i = game.getActions(); i
                .hasMoreElements(); ) {
            Object o = i.nextElement();
            
            AttackAction aa = (AttackAction) o;
            if (!game.getEntity(aa.getEntityId()).isActive()
                && !(o instanceof DfaAttackAction)) {
                continue;
            }
            AbstractAttackAction aaa = (AbstractAttackAction) o;
            
            if (aaa instanceof SearchlightAttackAction) {
                SearchlightAttackAction saa = (SearchlightAttackAction) aaa;
                addReport(saa.resolveAction(game));
            } else {
                physicalResults.addElement(preTreatPhysicalAttack(aaa));
            }
        }
        int cen = Entity.NONE;
        for (PhysicalResult pr : physicalResults) {
            resolvePhysicalAttack(pr, cen);
            cen = pr.aaa.getEntityId();
        }
        physicalResults.removeAllElements();
    }

    
    private void cleanupPhysicalAttacks() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            Entity entity = i.next();
            removeDuplicateAttacks(entity.getId());
        }
        removeDeadAttacks();
    }

    
    private void removeDuplicateAttacks(int entityId) {
        int allowed = 1;
        Entity en = game.getEntity(entityId);
        if (null != en) {
            allowed = en.getAllowedPhysicalAttacks();
        }
        Vector<EntityAction> toKeep = new Vector<EntityAction>();

        for (Enumeration<EntityAction> i = game.getActions(); i
                .hasMoreElements(); ) {
            EntityAction action = i.nextElement();
            if (action.getEntityId() != entityId) {
                toKeep.addElement(action);
            } else if (allowed > 0) {
                toKeep.addElement(action);
                if (!(action instanceof SearchlightAttackAction)) {
                    allowed--;
                }
            } else {
                System.err
                        .println("server: removing duplicate phys attack for id#"
                                 + entityId);
                System.err.println("        action was " + action.toString());
            }
        }

        
        game.resetActions();
        for (EntityAction entityAction : toKeep) {
            game.addAction(entityAction);
        }
    }

    
    private void removeDeadAttacks() {
        Vector<EntityAction> toKeep = new Vector<EntityAction>(
                game.actionsSize());

        for (Enumeration<EntityAction> i = game.getActions(); i
                .hasMoreElements(); ) {
            EntityAction action = i.nextElement();
            Entity entity = game.getEntity(action.getEntityId());
            if ((entity != null)
                && !entity.isDestroyed()
                && (entity.isActive() || (action instanceof DfaAttackAction))) {
                toKeep.addElement(action);
            }
        }

        
        game.resetActions();
        for (EntityAction entityAction : toKeep) {
            game.addAction(entityAction);
        }
    }

    
    private void resolvePunchAttack(PhysicalResult pr, int lastEntityId) {
        final PunchAttackAction paa = (PunchAttackAction) pr.aaa;
        final Entity ae = game.getEntity(paa.getEntityId());
        final Targetable target = game.getTarget(paa.getTargetType(),
                                                 paa.getTargetId());
        Entity te = null;
        if (target.getTargetType() == Targetable.TYPE_ENTITY) {
            te = (Entity) target;
        }
        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }
        final String armName = paa.getArm() == PunchAttackAction.LEFT ? "Left Arm"
                                                                      : "Right Arm";
        
        int damage = paa.getArm() == PunchAttackAction.LEFT ? pr.damage
                                                            : pr.damageRight;
        final ToHitData toHit = paa.getArm() == PunchAttackAction.LEFT ? pr.toHit
                                                                       : pr.toHitRight;
        int roll = paa.getArm() == PunchAttackAction.LEFT ? pr.roll
                                                          : pr.rollRight;
        final boolean targetInBuilding = Compute.isInBuilding(game, te);
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());

        Report r;
        
        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        
        Building bldg = game.getBoard().getBuildingAt(target.getPosition());

        if (lastEntityId != paa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4010);
        r.subject = ae.getId();
        r.indent();
        r.add(armName);
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4015);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4020);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            r.newlines = 0;
            addReport(r);
            roll = Integer.MAX_VALUE;
        } else {
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
            if (glancing) {
                r = new Report(4030);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
            if (directBlow) {
                r = new Report(4032);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
        }

        
        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);

            
            if (targetInBuilding && (bldg != null)) {

                
                if (damage > 0) {
                    Vector<Report> buildingReport = damageBuilding(bldg,
                                                                   damage, target.getPosition());
                    for (Report report : buildingReport) {
                        report.subject = ae.getId();
                    }
                    addReport(buildingReport);
                }

            }
            return;
        }

        
        if ((target.getTargetType() == Targetable.TYPE_BUILDING)
            || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) {
            
            r = new Report(4040);
            r.subject = ae.getId();
            addReport(r);
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            
            addReport(damageInfantryIn(bldg, damage, target.getPosition()));

            
            return;
        }

        HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                         toHit.getSideTable());
        hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
        r = new Report(4045);
        r.subject = ae.getId();
        r.add(toHit.getTableDesc());
        r.add(te.getLocationAbbr(hit));
        addReport(r);

        
        
        if (targetInBuilding && (bldg != null)) {
            int bldgAbsorbs = bldg.getAbsorbtion(target.getPosition());
            int toBldg = Math.min(bldgAbsorbs, damage);
            damage -= toBldg;
            addNewLines();
            Vector<Report> buildingReport = damageBuilding(bldg, toBldg,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            
            
            damage = (int) Math.floor(bldg.getDamageToScale() * damage);
        }

        
        if (damage == 0) {
            r = new Report(4050);
            r.subject = ae.getId();
            r.add(te.getShortName());
            r.add(te.getOwner().getName());
            r.indent();
            addReport(r);
        } else {
            if (glancing) {
                
                if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                    damage = (int) Math.ceil(damage / 2.0);
                } else {
                    damage = (int) Math.floor(damage / 2.0);
                }
            }
            if (directBlow) {
                damage += toHit.getMoS() / 3;
                hit.makeDirectBlow(toHit.getMoS() / 3);
            }

            if ((damage >= 1)
                && te.hasWorkingMisc(MiscType.F_SPIKES, -1,
                                     hit.getLocation())) {
                r = new Report(4330);
                r.indent(2);
                r.newlines = 0;
                r.subject = ae.getId();
                addReport(r);
                checkBreakSpikes(te, hit.getLocation());
                damage = Math.max(1, damage - 4);
                HitData ahit;
                if (paa.getArm() == PunchAttackAction.LEFT) {
                    ahit = new HitData(Mech.LOC_LARM);
                } else {
                    ahit = new HitData(Mech.LOC_RARM);
                }
                addReport(damageEntity(ae, ahit, 2, false, DamageType.NONE,
                                       false, false, false));
            }
            DamageType damageType = DamageType.NONE;
            addReport(damageEntity(te, hit, damage, false, damageType, false,
                                   false, throughFront));
            if (target instanceof VTOL) {
                
                addReport(applyCriticalHit(te, VTOL.LOC_ROTOR,
                                           new CriticalSlot(CriticalSlot.TYPE_SYSTEM,
                                                            VTOL.CRIT_ROTOR_DESTROYED), false, 0, false));
            }
            
            if (paa.isBladeExtended(paa.getArm())) {
                addNewLines();
                r = new Report(4455);
                r.indent(2);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
                
                
                if (!(target instanceof Infantry)) {
                    addNewLines();
                    addReport(criticalEntity(te, hit.getLocation(),
                                             hit.isRear(), 0, true, false, damage));
                }
                if ((target instanceof BattleArmor)
                    && (hit.getLocation() < te.locations())
                    && (te.getInternal(hit.getLocation()) > 0)) {
                    
                    
                    
                    HitData bahit = new HitData(hit.getLocation(), false,
                                                HitData.EFFECT_CRITICAL);
                    addReport(damageEntity(te, bahit, 0));
                }
                
                
                
                
                
                
                if (Compute.d6(2) > 9) {
                    int armLoc = (paa.getArm() == PunchAttackAction.RIGHT) ? Mech.LOC_RARM
                                                                           : Mech.LOC_LARM;
                    addNewLines();
                    r = new Report(4456);
                    r.indent(2);
                    r.subject = ae.getId();
                    r.newlines = 0;
                    addReport(r);
                    ae.destroyRetractableBlade(armLoc);
                }
            }
        }
        addNewLines();

        
        
        if ((target instanceof Mech) && ((Mech) target).isIndustrial()) {
            ((Mech) target).setCheckForCrit(true);
        }
    }

    
    private void resolveKickAttack(PhysicalResult pr, int lastEntityId) {
        KickAttackAction kaa = (KickAttackAction) pr.aaa;
        final Entity ae = game.getEntity(kaa.getEntityId());
        final Targetable target = game.getTarget(kaa.getTargetType(),
                                                 kaa.getTargetId());
        Entity te = null;
        if (target.getTargetType() == Targetable.TYPE_ENTITY) {
            te = (Entity) target;
        }
        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }
        String legName = (kaa.getLeg() == KickAttackAction.LEFT)
                         || (kaa.getLeg() == KickAttackAction.LEFTMULE) ? "Left "
                                                                        : "Right ";
        if ((kaa.getLeg() == KickAttackAction.LEFTMULE)
            || (kaa.getLeg() == KickAttackAction.RIGHTMULE)) {
            legName = legName.concat("rear ");
        } else if (ae instanceof QuadMech) {
            legName = legName.concat("front ");
        }
        legName = legName.concat("leg");
        Report r;

        
        int damage = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        final boolean targetInBuilding = Compute.isInBuilding(game, te);
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());

        
        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        
        Building bldg = game.getBoard().getBuildingAt(target.getPosition());

        if (lastEntityId != ae.getId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4055);
        r.subject = ae.getId();
        r.indent();
        r.add(legName);
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4060);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            game.addPSR(new PilotingRollData(ae.getId(), 0, "missed a kick"));
            return;
        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4065);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            r.newlines = 0;
            addReport(r);
            roll = Integer.MAX_VALUE;
        } else {
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
            if (glancing) {
                r = new Report(4030);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
            if (directBlow) {
                r = new Report(4032);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }

        }

        
        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            game.addPSR(new PilotingRollData(ae.getId(), 0, "missed a kick"));

            
            if (targetInBuilding && (bldg != null)) {

                
                if (damage > 0) {
                    Vector<Report> buildingReport = damageBuilding(bldg,
                                                                   damage, target.getPosition());
                    for (Report report : buildingReport) {
                        report.subject = ae.getId();
                    }
                    addReport(buildingReport);
                }

            }
            return;
        }

        
        if ((target.getTargetType() == Targetable.TYPE_BUILDING)
            || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) {
            
            r = new Report(4040);
            r.subject = ae.getId();
            addReport(r);
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            
            addReport(damageInfantryIn(bldg, damage, target.getPosition()));

            
            return;
        }

        HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                         toHit.getSideTable());
        hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
        r = new Report(4045);
        r.subject = ae.getId();
        r.add(toHit.getTableDesc());
        r.add(te.getLocationAbbr(hit));
        addReport(r);

        
        
        if (targetInBuilding && (bldg != null)) {
            int bldgAbsorbs = bldg.getAbsorbtion(target.getPosition());
            int toBldg = Math.min(bldgAbsorbs, damage);
            damage -= toBldg;
            addNewLines();
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            
            
            damage = (int) Math.floor(bldg.getDamageToScale() * damage);
        }

        
        if (damage == 0) {
            r = new Report(4050);
            r.subject = ae.getId();
            r.add(te.getShortName());
            r.add(te.getOwner().getName());
            r.newlines = 0;
            addReport(r);
        } else {
            if (glancing) {
                
                if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                    damage = (int) Math.ceil(damage / 2.0);
                } else {
                    damage = (int) Math.floor(damage / 2.0);
                }
            }
            if (directBlow) {
                damage += toHit.getMoS() / 3;
                hit.makeDirectBlow(toHit.getMoS() / 3);
            }

            if ((damage >= 1)
                && te.hasWorkingMisc(MiscType.F_SPIKES, -1,
                                     hit.getLocation())) {
                r = new Report(4330);
                r.indent(2);
                r.newlines = 0;
                r.subject = ae.getId();
                addReport(r);
                checkBreakSpikes(te, hit.getLocation());
                damage = Math.max(1, damage - 4);
                HitData ahit;
                switch (kaa.getLeg()) {
                    case KickAttackAction.LEFT:
                        if (ae instanceof QuadMech) {
                            ahit = new HitData(Mech.LOC_LARM);
                        } else {
                            ahit = new HitData(Mech.LOC_LLEG);
                        }
                        break;
                    case KickAttackAction.RIGHT:
                        if (ae instanceof QuadMech) {
                            ahit = new HitData(Mech.LOC_RARM);
                        } else {
                            ahit = new HitData(Mech.LOC_RLEG);
                        }
                        break;
                    case KickAttackAction.LEFTMULE:
                        ahit = new HitData(Mech.LOC_LLEG);
                        break;
                    case KickAttackAction.RIGHTMULE:
                    default:
                        ahit = new HitData(Mech.LOC_RLEG);
                        break;
                }
                addReport(damageEntity(ae, ahit, 2, false, DamageType.NONE,
                                       false, false, false));
            }
            DamageType damageType = DamageType.NONE;
            addReport(damageEntity(te, hit, damage, false, damageType, false,
                                   false, throughFront));
            if (target instanceof VTOL) {
                
                addReport(applyCriticalHit(te, VTOL.LOC_ROTOR,
                                           new CriticalSlot(CriticalSlot.TYPE_SYSTEM,
                                                            VTOL.CRIT_ROTOR_DESTROYED), false, 0, false));
            }
            if (te.hasQuirk(OptionsConstants.QUIRK_NEG_WEAK_LEGS)) {
                addNewLines();
                addReport(criticalEntity(te, hit.getLocation(), hit.isRear(),
                                         0, 0));
            }
        }

        if ((te.getMovementMode() == EntityMovementMode.BIPED)
            || (te.getMovementMode() == EntityMovementMode.QUAD)) {
            PilotingRollData kickPRD = getKickPushPSR(te, ae, te, "was kicked");
            game.addPSR(kickPRD);
        }

        
        
        if ((te instanceof Mech) && ((Mech) te).isIndustrial()) {
            ((Mech) te).setCheckForCrit(true);
        }

        addNewLines();
    }

    
    private void resolveJumpJetAttack(PhysicalResult pr, int lastEntityId) {
        JumpJetAttackAction kaa = (JumpJetAttackAction) pr.aaa;
        final Entity ae = game.getEntity(kaa.getEntityId());
        final Targetable target = game.getTarget(kaa.getTargetType(),
                                                 kaa.getTargetId());
        Entity te = null;
        if (target.getTargetType() == Targetable.TYPE_ENTITY) {
            te = (Entity) target;
        }
        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }
        String legName = null;
        switch (kaa.getLeg()) {
            case JumpJetAttackAction.LEFT:
                legName = "Left leg";
                break;
            case JumpJetAttackAction.RIGHT:
                legName = "Right leg";
                break;
            default:
                legName = "Both legs";
                break;
        }

        Report r;

        
        int damage = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        final boolean targetInBuilding = Compute.isInBuilding(game, te);
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());

        
        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        
        Building bldg = game.getBoard().getBuildingAt(target.getPosition());

        if (lastEntityId != ae.getId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4290);
        r.subject = ae.getId();
        r.indent();
        r.add(legName);
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4075);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4080);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            r.newlines = 0;
            addReport(r);
            roll = Integer.MAX_VALUE;
        } else {
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
            if (glancing) {
                r = new Report(4030);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
            if (directBlow) {
                r = new Report(4032);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }

        }

        
        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);

            
            if (targetInBuilding && (bldg != null)) {

                damage += pr.damageRight;
                
                if (damage > 0) {
                    Vector<Report> buildingReport = damageBuilding(bldg,
                                                                   damage, target.getPosition());
                    for (Report report : buildingReport) {
                        report.subject = ae.getId();
                    }
                    addReport(buildingReport);
                }

            }
            return;
        }

        
        if ((target.getTargetType() == Targetable.TYPE_BUILDING)
            || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) {
            damage += pr.damageRight;
            
            r = new Report(4040);
            r.subject = ae.getId();
            addReport(r);
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            
            addReport(damageInfantryIn(bldg, damage, target.getPosition()));

            
            return;
        }

        r = new Report(4040);
        r.subject = ae.getId();
        r.newlines = 0;
        addReport(r);

        for (int leg = 0; leg < 2; leg++) {
            if (leg == 1) {
                damage = pr.damageRight;
                if (damage == 0) {
                    break;
                }
            }
            HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                             toHit.getSideTable());
            hit.setGeneralDamageType(HitData.DAMAGE_ENERGY);

            
            
            if (targetInBuilding && (bldg != null)) {
                int bldgAbsorbs = bldg.getAbsorbtion(target.getPosition());
                int toBldg = Math.min(bldgAbsorbs, damage);
                damage -= toBldg;
                addNewLines();
                Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                               target.getPosition());
                for (Report report : buildingReport) {
                    report.subject = ae.getId();
                }
                addReport(buildingReport);

                
                
                damage = (int) Math.floor(bldg.getDamageToScale() * damage);
            }

            
            if (damage == 0) {
                r = new Report(4050);
                r.subject = ae.getId();
                r.add(te.getShortName());
                r.add(te.getOwner().getName());
                r.newlines = 0;
                addReport(r);
            } else {
                if (glancing) {
                    
                    if ((te instanceof Infantry)
                        && !(te instanceof BattleArmor)) {
                        damage = (int) Math.ceil(damage / 2.0);
                    } else {
                        damage = (int) Math.floor(damage / 2.0);
                    }
                }
                if (directBlow) {
                    damage += toHit.getMoS() / 3;
                    hit.makeDirectBlow(toHit.getMoS() / 3);
                }
                addReport(damageEntity(te, hit, damage, false, DamageType.NONE,
                                       false, false, throughFront));
            }
        }

        addNewLines();

        
        
        if ((target instanceof Mech) && ((Mech) target).isIndustrial()) {
            ((Mech) target).setCheckForCrit(true);
        }
    }

    

    private void resolveProtoAttack(PhysicalResult pr, int lastEntityId) {
        final ProtomechPhysicalAttackAction ppaa = (ProtomechPhysicalAttackAction) pr.aaa;
        final Entity ae = game.getEntity(ppaa.getEntityId());
        
        int damage = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        final Targetable target = game.getTarget(ppaa.getTargetType(),
                                                 ppaa.getTargetId());
        Entity te = null;
        if (target.getTargetType() == Targetable.TYPE_ENTITY) {
            te = (Entity) target;
        }
        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }
        final boolean targetInBuilding = Compute.isInBuilding(game, te);
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());
        
        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        Report r;

        
        Building bldg = game.getBoard().getBuildingAt(target.getPosition());

        if (lastEntityId != ae.getId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4070);
        r.subject = ae.getId();
        r.indent();
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4075);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4080);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            r.newlines = 0;
            addReport(r);
            roll = Integer.MAX_VALUE;
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
            if (glancing) {
                r = new Report(4030);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
            if (directBlow) {
                r = new Report(4032);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }

        }

        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
 
            if (targetInBuilding && (bldg != null)) {
   
                if (damage > 0) {
                    Vector<Report> buildingReport = damageBuilding(bldg,
                                                                   damage, target.getPosition());
                    for (Report report : buildingReport) {
                        report.subject = ae.getId();
                    }
                    addReport(buildingReport);
                }

            }
            return;
        }

        if ((target.getTargetType() == Targetable.TYPE_BUILDING)
            || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) {
            
            r = new Report(4040);
            r.subject = ae.getId();
            addReport(r);
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            addReport(damageInfantryIn(bldg, damage, target.getPosition()));

            return;
        }

        HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                         toHit.getSideTable());
        hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);

        r = new Report(4045);
        r.subject = ae.getId();
        r.add(toHit.getTableDesc());
        r.add(te.getLocationAbbr(hit));
        addReport(r);

        if (targetInBuilding && (bldg != null)) {
            int bldgAbsorbs = bldg.getAbsorbtion(target.getPosition());
            int toBldg = Math.min(bldgAbsorbs, damage);
            damage -= toBldg;
            addNewLines();
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            damage = (int) Math.floor(bldg.getDamageToScale() * damage);
        }

        if (damage == 0) {
            r = new Report(4050);
            r.subject = ae.getId();
            r.add(te.getShortName());
            r.add(te.getOwner().getName());
            r.newlines = 0;
            addReport(r);
        } else {
            if (glancing) {
                
                if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                    damage = (int) Math.ceil(damage / 2.0);
                } else {
                    damage = (int) Math.floor(damage / 2.0);
                }
            }
            if (directBlow) {
                damage += toHit.getMoS() / 3;
                hit.makeDirectBlow(toHit.getMoS() / 3);
            }
            addReport(damageEntity(te, hit, damage, false, DamageType.NONE,
                                   false, false, throughFront));
            if (((Protomech) ae).isEDPCharged()) {
                r = new Report(3701);
                int taserRoll = Compute.d6(2) - 2;
                r.add(taserRoll);
                r.newlines = 0;
                vPhaseReport.add(r);

                if (te instanceof BattleArmor) {
                    r = new Report(3706);
                    r.addDesc(te);
                    
                    HitData targetTrooper = te.rollHitLocation(
                            ToHitData.HIT_NORMAL, ToHitData.SIDE_FRONT);
                    r.add(te.getLocationAbbr(targetTrooper));
                    vPhaseReport.add(r);
                    vPhaseReport.addAll(criticalEntity(ae,
                                                       targetTrooper.getLocation(),
                                                       targetTrooper.isRear(), 0, false, false, 0));
                } else if (te instanceof Mech) {
                    if (((Mech) te).isIndustrial()) {
                        if (taserRoll >= 8) {
                            r = new Report(3705);
                            r.addDesc(te);
                            r.add(4);
                            te.taserShutdown(4, false);
                        } else {
                            
                            r = new Report(3710);
                            r.addDesc(te);
                            r.add(2);
                            r.add(4);
                            te.setTaserInterference(2, 4, true);
                        }
                    } else {
                        if (taserRoll >= 11) {
                            r = new Report(3705);
                            r.addDesc(te);
                            r.add(3);
                            vPhaseReport.add(r);
                            te.taserShutdown(3, false);
                        } else {
                            r = new Report(3710);
                            r.addDesc(te);
                            r.add(2);
                            r.add(3);
                            vPhaseReport.add(r);
                            te.setTaserInterference(2, 3, true);
                        }
                    }
                } else if ((te instanceof Protomech) || (te instanceof Tank)
                           || (te instanceof Aero)) {
                    if (taserRoll >= 8) {
                        r = new Report(3705);
                        r.addDesc(te);
                        r.add(4);
                        vPhaseReport.add(r);
                        te.taserShutdown(4, false);
                    } else {
                        r = new Report(3710);
                        r.addDesc(te);
                        r.add(2);
                        r.add(4);
                        vPhaseReport.add(r);
                        te.setTaserInterference(2, 4, false);
                    }
                }

            }
        }
        addNewLines();

        if ((target instanceof Mech) && ((Mech) target).isIndustrial()) {
            ((Mech) target).setCheckForCrit(true);
        }
    }

    
    private void resolveBrushOffAttack(PhysicalResult pr, int lastEntityId) {
        final BrushOffAttackAction baa = (BrushOffAttackAction) pr.aaa;
        final Entity ae = game.getEntity(baa.getEntityId());
        
        
        Targetable target = game.getTarget(baa.getTargetType(),
                                           baa.getTargetId());
        Entity te = null;
        final String armName = baa.getArm() == BrushOffAttackAction.LEFT ? "Left Arm"
                                                                         : "Right Arm";
        Report r;

        if (target.getTargetType() == Targetable.TYPE_ENTITY) {
            te = game.getEntity(baa.getTargetId());
        }

        int damage = baa.getArm() == BrushOffAttackAction.LEFT ? pr.damage
                                                               : pr.damageRight;
        final ToHitData toHit = baa.getArm() == BrushOffAttackAction.LEFT ? pr.toHit
                                                                          : pr.toHitRight;
        int roll = baa.getArm() == BrushOffAttackAction.LEFT ? pr.roll
                                                             : pr.rollRight;

        if (lastEntityId != baa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4085);
        r.subject = ae.getId();
        r.indent();
        r.add(target.getDisplayName());
        r.add(armName);
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4090);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        }

        
        r = new Report(4025);
        r.subject = ae.getId();
        r.add(toHit.getValue());
        r.add(roll);
        r.newlines = 0;
        addReport(r);

        
        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);

            
            toHit.setHitTable(ToHitData.HIT_PUNCH);
            toHit.setSideTable(ToHitData.SIDE_FRONT);
            HitData hit = ae.rollHitLocation(toHit.getHitTable(),
                                             toHit.getSideTable());
            hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
            r = new Report(4095);
            r.subject = ae.getId();
            r.addDesc(ae);
            r.add(ae.getLocationAbbr(hit));
            r.newlines = 0;
            addReport(r);
            addReport(damageEntity(ae, hit, damage));
            addNewLines();
            
            
            if ((ae instanceof Mech) && ((Mech) ae).isIndustrial()) {
                ((Mech) ae).setCheckForCrit(true);
            }
            return;
        }

        switch (target.getTargetType()) {
            case Targetable.TYPE_ENTITY:
                
                HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                                 toHit.getSideTable());
                hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                r = new Report(4045);
                r.subject = ae.getId();
                r.add(toHit.getTableDesc());
                r.add(te.getLocationAbbr(hit));
                addReport(r);
                addReport(damageEntity(te, hit, damage));
                addNewLines();

                
                ae.setSwarmAttackerId(Entity.NONE);
                te.setSwarmTargetId(Entity.NONE);
                r = new Report(4100);
                r.subject = ae.getId();
                r.add(te.getDisplayName());
                addReport(r);
                break;
            case Targetable.TYPE_INARC_POD:
                
                
                ae.removeINarcPod((INarcPod) target);
                
                r = new Report(4105);
                r.subject = ae.getId();
                r.add(target.getDisplayName());
                addReport(r);
                break;
            
        }
    }

    
    private void resolveThrashAttack(PhysicalResult pr, int lastEntityId) {
        final ThrashAttackAction taa = (ThrashAttackAction) pr.aaa;
        final Entity ae = game.getEntity(taa.getEntityId());

        int hits = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());

        
        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        final Entity te = game.getEntity(taa.getTargetId());
        Report r;

        if (lastEntityId != taa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4110);
        r.subject = ae.getId();
        r.indent();
        r.addDesc(te);
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4115);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        }

        
        if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4120);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);

            
            if (roll < toHit.getValue()) {
                
                r = new Report(4035);
                r.subject = ae.getId();
                addReport(r);
                return;
            }
            r = new Report(4125);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
        }

        
        if (glancing) {
            hits = (int) Math.floor(hits / 2.0);
        }
        if (directBlow) {
            hits += toHit.getMoS() / 3;
        }

        r = new Report(4130);
        r.subject = ae.getId();
        r.add(hits);
        r.newlines = 0;
        addReport(r);
        if (glancing) {
            r = new Report(4030);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
        }
        if (directBlow) {
            r = new Report(4032);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
        }

        while (hits > 0) {
            int damage = Math.min(5, hits);
            hits -= damage;
            HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                             toHit.getSideTable());
            hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
            r = new Report(4135);
            r.subject = ae.getId();
            r.add(te.getLocationAbbr(hit));
            r.newlines = 0;
            addReport(r);
            addReport(damageEntity(te, hit, damage));
        }

        addNewLines();

        PilotingRollData rollData = ae.getBasePilotingRoll();
        ae.addPilotingModifierForTerrain(rollData);
        rollData.addModifier(0, "thrashing at infantry");
        r = new Report(4140);
        r.subject = ae.getId();
        r.addDesc(ae);
        addReport(r);
        final int diceRoll = Compute.d6(2);
        r = new Report(2190);
        r.subject = ae.getId();
        r.add(rollData.getValueAsString());
        r.add(rollData.getDesc());
        r.add(diceRoll);
        if (diceRoll < rollData.getValue()) {
            r.choose(false);
            addReport(r);
            addReport(doEntityFall(ae, rollData));
        } else {
            r.choose(true);
            addReport(r);
        }
    }

    private void resolveBAVibroClawAttack(PhysicalResult pr, int lastEntityId) {
        final BAVibroClawAttackAction bvaa = (BAVibroClawAttackAction) pr.aaa;
        final Entity ae = game.getEntity(bvaa.getEntityId());

        
        int hits = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());

        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        final Entity te = game.getEntity(bvaa.getTargetId());
        Report r;

        if (lastEntityId != bvaa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4146);
        r.subject = ae.getId();
        r.indent();
        r.addDesc(te);
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4147);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        }

        if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4120);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);

            
            if (roll < toHit.getValue()) {
                
                r = new Report(4035);
                r.subject = ae.getId();
                addReport(r);
                return;
            }
        }

        if (glancing) {
            hits = (int) Math.floor(hits / 2.0);
        }
        if (directBlow) {
            hits += toHit.getMoS() / 3;
        }
        if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
            r = new Report(4149);
            r.subject = ae.getId();
            r.add(hits);
        } else {
            r = new Report(4148);
            r.subject = ae.getId();
            r.add(hits);
            r.add(ae.getVibroClaws());
        }
        r.newlines = 0;
        addReport(r);
        if (glancing) {
            r = new Report(4030);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
        }
        if (directBlow) {
            r = new Report(4032);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
        }
        while (hits > 0) {
            
            int damage = Math.min(ae.getVibroClaws(), hits);
            
            if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                damage = hits;
            }
            hits -= damage;
            HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                             toHit.getSideTable());
            hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
            r = new Report(4135);
            r.subject = ae.getId();
            r.add(te.getLocationAbbr(hit));
            r.newlines = 0;
            addReport(r);
            addReport(damageEntity(te, hit, damage));
        }
        addNewLines();
    }

    private void resolveClubAttack(PhysicalResult pr, int lastEntityId) {
        final ClubAttackAction caa = (ClubAttackAction) pr.aaa;
        final Entity ae = game.getEntity(caa.getEntityId());
        
        int damage = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        final Targetable target = game.getTarget(caa.getTargetType(),
                                                 caa.getTargetId());
        Entity te = null;
        if (target.getTargetType() == Targetable.TYPE_ENTITY) {
            te = (Entity) target;
        }
        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }
        final boolean targetInBuilding = Compute.isInBuilding(game, te);
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());

        toHit.setMoS((roll == Integer.MAX_VALUE) ? 0 : roll
                                                       - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        Report r;

        Building bldg = game.getBoard().getBuildingAt(target.getPosition());

        caa.getClub().restore();

        if (((MiscType) caa.getClub().getType()).isShield()) {
            ((Mech) ae).shieldAbsorptionDamage(1, caa.getClub().getLocation(),
                                               false);
        }

        if (lastEntityId != caa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4145);
        r.subject = ae.getId();
        r.indent();
        r.add(caa.getClub().getName());
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        
        if ((((MiscType) caa.getClub().getType()).hasSubType(MiscType.S_FLAIL) || ((MiscType) caa
                .getClub().getType()).hasSubType(MiscType.S_WRECKING_BALL))
            && (roll == 2)) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            ToHitData newToHit = new ToHitData(TargetRoll.AUTOMATIC_SUCCESS,
                                               "hit with own flail/wrecking ball");
            pr.damage = ClubAttackAction.getDamageFor(ae, caa.getClub(), false);
            pr.damage = (pr.damage / 2) + (pr.damage % 2);
            newToHit.setHitTable(ToHitData.HIT_NORMAL);
            newToHit.setSideTable(ToHitData.SIDE_FRONT);
            pr.toHit = newToHit;
            pr.aaa.setTargetId(ae.getId());
            pr.aaa.setTargetType(Targetable.TYPE_ENTITY);
            pr.roll = Integer.MAX_VALUE;
            resolveClubAttack(pr, ae.getId());
            game.addPSR(new PilotingRollData(ae.getId(), 0,
                                             "missed a flail/wrecking ball attack"));
            return;
        }

        
        if (((MiscType) (caa.getClub().getType()))
                .hasSubType(MiscType.S_BUZZSAW)) {

            damage = Compute.d6(2);
            ae.heatBuildup += 3;

            if (roll == 2) {

                Mounted club = caa.getClub();

                for (Mounted eq : ae.getWeaponList()) {
                    if ((eq.getLocation() == club.getLocation())
                        && (eq.getType() instanceof MiscType)
                        && ((MiscType) eq.getType())
                            .hasFlag(MiscType.F_CLUB)
                        && ((MiscType) eq.getType())
                            .hasSubType(MiscType.S_BUZZSAW)) {
                        eq.setHit(true);
                        break;
                    }
                }
                r = new Report(4037);
                r.subject = ae.getId();
                addReport(r);
                damage = 0;
                return;
            }
        }

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4075);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            if (((MiscType) caa.getClub().getType())
                    .hasSubType(MiscType.S_MACE_THB)) {
                game.addPSR(new PilotingRollData(ae.getId(), 0,
                                                 "missed a mace attack"));
            }
            if (((MiscType) caa.getClub().getType())
                    .hasSubType(MiscType.S_MACE)) {
                game.addPSR(new PilotingRollData(ae.getId(), 2,
                                                 "missed a mace attack"));
            }
            return;
        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4080);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            r.newlines = 0;
            addReport(r);
            roll = Integer.MAX_VALUE;
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
            if (glancing) {
                r = new Report(4030);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
            if (directBlow) {
                r = new Report(4032);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }

        }

        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            if (((MiscType) caa.getClub().getType())
                    .hasSubType(MiscType.S_MACE_THB)) {
                game.addPSR(new PilotingRollData(ae.getId(), 0,
                                                 "missed a mace attack"));
            }
            if (((MiscType) caa.getClub().getType())
                    .hasSubType(MiscType.S_MACE)) {
                game.addPSR(new PilotingRollData(ae.getId(), 2,
                                                 "missed a mace attack"));
            }

            if (targetInBuilding && (bldg != null)) {

                
                if (damage > 0) {
                    Vector<Report> buildingReport = damageBuilding(bldg,
                                                                   damage, target.getPosition());
                    for (Report report : buildingReport) {
                        report.subject = ae.getId();
                    }
                    addReport(buildingReport);
                }

            }
            return;
        }

        if ((target.getTargetType() == Targetable.TYPE_BUILDING)
            || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) {
            
            r = new Report(4040);
            r.subject = ae.getId();
            addReport(r);
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            addReport(damageInfantryIn(bldg, damage, target.getPosition()));

            return;
        }

        HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                         toHit.getSideTable());
        hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
        r = new Report(4045);
        r.subject = ae.getId();
        r.add(toHit.getTableDesc());
        r.add(te.getLocationAbbr(hit));
        addReport(r);

        if (targetInBuilding && (bldg != null)) {
            int bldgAbsorbs = bldg.getAbsorbtion(target.getPosition());
            int toBldg = Math.min(bldgAbsorbs, damage);
            damage -= toBldg;
            addNewLines();
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            
            damage = (int) Math.floor(bldg.getDamageToScale() * damage);
        }

        if (damage == 0) {
            r = new Report(4050);
            r.subject = ae.getId();
            r.add(te.getShortName());
            r.add(te.getOwner().getName());
            r.newlines = 0;
            addReport(r);
        } else {
            if (glancing) {
                
                if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                    damage = (int) Math.ceil(damage / 2.0);
                } else {
                    damage = (int) Math.floor(damage / 2.0);
                }
            }
            if (directBlow) {
                damage += toHit.getMoS() / 3;
                hit.makeDirectBlow(toHit.getMoS() / 3);
            }

            if ((damage >= 1)
                && te.hasWorkingMisc(MiscType.F_SPIKES, -1,
                                     hit.getLocation())) {
                r = new Report(4331);
                r.indent(2);
                r.newlines = 0;
                r.subject = ae.getId();
                addReport(r);
                checkBreakSpikes(te, hit.getLocation());
                damage = Math.max(1, damage - 4);
            }

            DamageType damageType = DamageType.NONE;
            addReport(damageEntity(te, hit, damage, false, damageType, false,
                                   false, throughFront));
            if (target instanceof VTOL) {
                
                addReport(applyCriticalHit(te, VTOL.LOC_ROTOR,
                                           new CriticalSlot(CriticalSlot.TYPE_SYSTEM,
                                                            VTOL.CRIT_ROTOR_DESTROYED), false, 0, false));
            }
        }

        if (((MiscType) caa.getClub().getType()).hasSubType(MiscType.S_LANCE)
            && (te.getArmor(hit) > 0)
            && (te.getArmorType(hit.getLocation()) != EquipmentType.T_ARMOR_HARDENED)
            && (te.getArmorType(hit.getLocation()) != EquipmentType.T_ARMOR_FERRO_LAMELLOR)) {
            roll = Compute.d6(2);
            
            r = new Report(4021);
            r.indent(2);
            r.subject = ae.getId();
            r.add(te.getLocationAbbr(hit));
            r.add(roll);
            addReport(r);
            if (roll >= 10) {
                hit.makeGlancingBlow();
                addReport(damageEntity(te, hit, 1, false, DamageType.NONE,
                                       true, false, throughFront));
            }
        }

        if (((MiscType) caa.getClub().getType())
                    .hasSubType(MiscType.S_WRECKING_BALL)
            && (ae instanceof SupportTank) && (te instanceof Mech)) {
            
            game.addPSR(new PilotingRollData(te.getId(), 2,
                                             "was hit by wrecking ball"));
        }

        
        if (((MiscType) caa.getClub().getType())
                .hasSubType(MiscType.S_CHAIN_WHIP)
                && (te instanceof Mech || te instanceof Protomech)) {
            addNewLines();

            int loc = hit.getLocation();
            int toHitNumber = toHit.getValue();

            boolean mightTrip = (te instanceof Mech)
                    && ((Mech) te).locationIsLeg(loc)
                    && !te.isLocationBad(loc)
                    && !te.isLocationDoomed(loc)
                    && !te.hasActiveShield(loc)
                    && !te.hasPassiveShield(loc);

            boolean mightGrapple = ((te instanceof Mech)
                    && (loc == Mech.LOC_LARM || loc == Mech.LOC_RARM)
                    && !te.isLocationBad(loc)
                    && !te.isLocationDoomed(loc)
                    && !te.hasActiveShield(loc)
                    && !te.hasPassiveShield(loc)
                    && !te.hasNoDefenseShield(loc))
                    || ((te instanceof Protomech)
                        && (loc == Protomech.LOC_LARM || loc == Protomech.LOC_RARM
                            || loc == Protomech.LOC_LEG)
                        && !te.isLocationBad(loc)
                        && !te.isLocationDoomed(loc));

            if (mightTrip) {

                roll = Compute.d6(2);

                if ((((Mech) ae).hasTSM() && (ae.heat >= 9))
                        && (!((Mech) te).hasTSM() || ((((Mech) te).hasTSM()) && (te.heat < 9)))) {
                    toHitNumber -= 2;
                }

                r = new Report(4450);
                r.subject = ae.getId();
                r.add(ae.getShortName());
                r.add(te.getShortName());
                r.add(toHitNumber);
                r.add(roll);
                r.indent(2);
                r.newlines = 0;
                addReport(r);

                if (roll >= toHit.getValue()) {
                    r = new Report(2270);
                    r.subject = ae.getId();
                    r.newlines = 0;
                    addReport(r);

                    game.addPSR(new PilotingRollData(te.getId(), 3,
                            "Snared by chain whip"));
                } else {
                    r = new Report(2357);
                    r.subject = ae.getId();
                    r.newlines = 0;
                    addReport(r);
                }
            } else if (mightGrapple) {
                GrappleAttackAction gaa = new GrappleAttackAction(ae.getId(),
                        te.getId());
                int grappleSide;
                if (caa.getClub().getLocation() == Mech.LOC_RARM) {
                    grappleSide = Entity.GRAPPLE_RIGHT;
                } else {
                    grappleSide = Entity.GRAPPLE_LEFT;
                }
                ToHitData grappleHit = GrappleAttackAction.toHit(game,
                        ae.getId(), target, grappleSide, true);
                PhysicalResult grappleResult = new PhysicalResult();
                grappleResult.aaa = gaa;
                grappleResult.toHit = grappleHit;
                grappleResult.roll = Compute.d6(2);
                resolveGrappleAttack(
                        grappleResult,
                        lastEntityId,
                        grappleSide,
                        hit.getLocation() == Mech.LOC_RARM ? Entity.GRAPPLE_RIGHT
                                : Entity.GRAPPLE_LEFT);
            }
        }

        addNewLines();

        if (((MiscType) caa.getClub().getType())
                .hasSubType(MiscType.S_TREE_CLUB)) {
            
            r = new Report(4150);
            r.subject = ae.getId();
            r.add(caa.getClub().getName());
            addReport(r);
            ae.removeMisc(caa.getClub().getName());
        }

        if ((target instanceof Mech) && ((Mech) target).isIndustrial()) {
            ((Mech) target).setCheckForCrit(true);
        }
    }

    private void resolvePushAttack(PhysicalResult pr, int lastEntityId) {
        final PushAttackAction paa = (PushAttackAction) pr.aaa;
        final Entity ae = game.getEntity(paa.getEntityId());
        
        final Entity te = game.getEntity(paa.getTargetId());
        
        int roll = pr.roll;
        final ToHitData toHit = pr.toHit;
        Report r;

        
        if (pr.pushBackResolved) {
            return;
        }
        
        pr.pushBackResolved = true;

        if (lastEntityId != paa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4155);
        r.subject = ae.getId();
        r.indent();
        r.addDesc(te);
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4160);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        }

        
        r = new Report(4025);
        r.subject = ae.getId();
        r.add(toHit.getValue());
        r.add(roll);
        r.newlines = 0;
        addReport(r);

        
        PhysicalResult targetPushResult = null;
        for (PhysicalResult tpr : physicalResults) {
            if ((tpr.aaa.getEntityId() == te.getId())
                && (tpr.aaa instanceof PushAttackAction)
                && (tpr.aaa.getTargetId() == ae.getId())) {
                targetPushResult = tpr;
            }
        }
        
        
        if ((targetPushResult != null) && !targetPushResult.pushBackResolved
            && (roll >= toHit.getValue())) {
            targetPushResult.pushBackResolved = true;
            
            if (targetPushResult.roll >= targetPushResult.toHit.getValue()) {
                r = new Report(4165);
                r.subject = ae.getId();
                r.addDesc(te);
                r.addDesc(te);
                r.addDesc(ae);
                r.add(targetPushResult.toHit.getValue());
                r.add(targetPushResult.roll);
                r.addDesc(ae);
                addReport(r);
                PilotingRollData targetPushPRD = getKickPushPSR(te, ae, te,
                                                                "was pushed");
                PilotingRollData pushPRD = getKickPushPSR(ae, ae, te,
                                                          "was pushed");
                game.addPSR(pushPRD);
                game.addPSR(targetPushPRD);
                return;
            }
            
            r = new Report(4166);
            r.subject = ae.getId();
            r.addDesc(te);
            r.addDesc(ae);
            r.add(targetPushResult.toHit.getValue());
            r.add(targetPushResult.roll);
            addReport(r);
        }

        
        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            return;
        }

        int direction = ae.getFacing();

        Coords src = te.getPosition();
        Coords dest = src.translated(direction);

        PilotingRollData pushPRD = getKickPushPSR(te, ae, te, "was pushed");

        if (Compute.isValidDisplacement(game, te.getId(), te.getPosition(),
                                        direction)) {
            r = new Report(4170);
            r.subject = ae.getId();
            r.newlines = 0;
            addReport(r);
            if (game.getBoard().contains(dest)) {
                r = new Report(4175);
                r.subject = ae.getId();
                r.add(dest.getBoardNum(), true);
                addReport(r);
            } else {
                
                r = new Report(4180);
                r.subject = ae.getId();
                addReport(r);
            }

            addReport(doEntityDisplacement(te, src, dest, pushPRD));

            
            if (!te.getPosition().equals(src)) {
                ae.setPosition(src);
            }
        } else {
            
            r = new Report(4185);
            r.subject = ae.getId();
            addReport(r);
            game.addPSR(pushPRD);
        }

        if ((te instanceof Mech) && ((Mech) te).isIndustrial()) {
            ((Mech) te).setCheckForCrit(true);
        }

        addNewLines();
    }

    private void resolveTripAttack(PhysicalResult pr, int lastEntityId) {
        final TripAttackAction paa = (TripAttackAction) pr.aaa;
        final Entity ae = game.getEntity(paa.getEntityId());
        
        final Entity te = game.getEntity(paa.getTargetId());
        
        int roll = pr.roll;
        final ToHitData toHit = pr.toHit;
        Report r;

        if (lastEntityId != paa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4280);
        r.subject = ae.getId();
        r.indent();
        r.addDesc(te);
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4285);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        }

        
        r = new Report(4025);
        r.subject = ae.getId();
        r.add(toHit.getValue());
        r.add(roll);
        r.newlines = 0;
        addReport(r);

        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            return;
        }

        PilotingRollData pushPRD = getKickPushPSR(te, ae, te, "was tripped");

        game.addPSR(pushPRD);

        r = new Report(4040);
        r.subject = ae.getId();
        addReport(r);
        addNewLines();
        
        if ((te instanceof Mech) && ((Mech) te).isIndustrial()) {
            ((Mech) te).setCheckForCrit(true);
        }
    }

    private void resolveGrappleAttack(PhysicalResult pr, int lastEntityId) {
        resolveGrappleAttack(pr, lastEntityId, Entity.GRAPPLE_BOTH,
                Entity.GRAPPLE_BOTH);
    }

    private void resolveGrappleAttack(PhysicalResult pr, int lastEntityId,
                                      int aeGrappleSide, int teGrappleSide) {
        final GrappleAttackAction paa = (GrappleAttackAction) pr.aaa;
        final Entity ae = game.getEntity(paa.getEntityId());
        
        final Entity te = game.getEntity(paa.getTargetId());
        
        int roll = pr.roll;
        final ToHitData toHit = pr.toHit;
        Report r;

        if (pr.pushBackResolved) {
            return;
        }

        if ((te.getGrappled() != Entity.NONE)
            || (ae.getGrappled() != Entity.NONE)) {
            toHit.addModifier(TargetRoll.IMPOSSIBLE, "Already Grappled");
        }

        if (lastEntityId != paa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4295);
        r.subject = ae.getId();
        r.indent();
        r.addDesc(te);
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4300);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        }

        
        r = new Report(4025);
        r.subject = ae.getId();
        r.add(toHit.getValue());
        r.add(roll);
        r.newlines = 0;
        addReport(r);

        
        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            return;
        }

        ae.setGrappled(te.getId(), true);
        te.setGrappled(ae.getId(), false);
        ae.setGrappledThisRound(true);
        te.setGrappledThisRound(true);
        
        if (aeGrappleSide == Entity.GRAPPLE_BOTH) {
            Coords pos = te.getPosition();
            ae.setPosition(pos);
            ae.setElevation(te.getElevation());
            te.setFacing((ae.getFacing() + 3) % 6);
            addReport(doSetLocationsExposure(ae, game.getBoard().getHex(pos),
                    false, ae.getElevation()));
        }

        ae.setGrappleSide(aeGrappleSide);
        te.setGrappleSide(teGrappleSide);

        r = new Report(4040);
        r.subject = ae.getId();
        addReport(r);
        addNewLines();

        if ((te instanceof Mech) && ((Mech) te).isIndustrial()) {
            ((Mech) te).setCheckForCrit(true);
        }
    }

    private void resolveBreakGrappleAttack(PhysicalResult pr, int lastEntityId) {
        final BreakGrappleAttackAction paa = (BreakGrappleAttackAction) pr.aaa;
        final Entity ae = game.getEntity(paa.getEntityId());
        
        final Entity te = game.getEntity(paa.getTargetId());
        
        int roll = pr.roll;
        final ToHitData toHit = pr.toHit;
        Report r;

        if (lastEntityId != paa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        r = new Report(4305);
        r.subject = ae.getId();
        r.indent();
        r.addDesc(te);
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            r = new Report(4310);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            return;
        }

        if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4320);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);

            
            if (roll < toHit.getValue()) {
                
                r = new Report(4035);
                r.subject = ae.getId();
                addReport(r);
                return;
            }

            r = new Report(4040);
            r.subject = ae.getId();
            addReport(r);
        }

        PhysicalResult targetGrappleResult = null;
        for (PhysicalResult tpr : physicalResults) {
            if ((tpr.aaa.getEntityId() == te.getId())
                && (tpr.aaa instanceof GrappleAttackAction)
                && (tpr.aaa.getTargetId() == ae.getId())) {
                targetGrappleResult = tpr;
                break;
            }
        }

        if (targetGrappleResult != null) {
            targetGrappleResult.pushBackResolved = true;
            
            r = new Report(4315);
            r.subject = te.getId();
            r.newlines = 0;
            r.addDesc(te);
            addReport(r);

            
            r = new Report(4025);
            r.subject = te.getId();
            r.add(targetGrappleResult.toHit.getValue());
            r.add(targetGrappleResult.roll);
            r.newlines = 0;
            addReport(r);

            if (roll < toHit.getValue()) {
                
                r = new Report(4035);
                r.subject = ae.getId();
                addReport(r);
            } else {
                
                r = new Report(4040);
                r.subject = ae.getId();
                addReport(r);

                
                ae.setGrappled(te.getId(), false);
                te.setGrappled(ae.getId(), true);

                return;
            }
        }

        Coords hexes[] = new Coords[6];
        int scores[] = new int[6];

        IHex curHex = game.getBoard().getHex(ae.getPosition());
        for (int i = 0; i < 6; i++) {
            hexes[i] = ae.getPosition().translated(i);
            scores[i] = 0;
            IHex hex = game.getBoard().getHex(hexes[i]);
            if (hex.containsTerrain(Terrains.MAGMA)) {
                scores[i] += 10;
            }
            if (hex.containsTerrain(Terrains.WATER)) {
                scores[i] += hex.terrainLevel(Terrains.WATER);
            }
            if ((curHex.surface() - hex.surface()) >= 2) {
                scores[i] += 2 * (curHex.surface() - hex.surface());
            }
        }

        int bestScore = 99999;
        int best = 0;
        int worstScore = -99999;
        int worst = 0;

        for (int i = 0; i < 6; i++) {
            if (bestScore > scores[i]) {
                best = i;
                bestScore = scores[i];
            }
            if (worstScore < scores[i]) {
                worst = i;
                worstScore = scores[i];
            }
        }

        if (ae.isGrappleAttacker()) {
            
            PilotingRollData psr = ae.getBasePilotingRoll();
            psr.addModifier(TargetRoll.AUTOMATIC_SUCCESS, "break grapple");
            addReport(doEntityDisplacement(ae, ae.getPosition(), hexes[best],
                                           psr));
            ae.setFacing(hexes[best].direction(te.getPosition()));
        } else {
            
            PilotingRollData psr = te.getBasePilotingRoll();
            psr.addModifier(TargetRoll.AUTOMATIC_SUCCESS, "break grapple");
            addReport(doEntityDisplacement(te, te.getPosition(), hexes[worst],
                                           psr));
            te.setFacing(hexes[worst].direction(ae.getPosition()));
        }

        ae.setGrappled(Entity.NONE, false);
        te.setGrappled(Entity.NONE, false);

        addNewLines();
    }

    private void resolveChargeAttack(PhysicalResult pr, int lastEntityId) {
        final ChargeAttackAction caa = (ChargeAttackAction) pr.aaa;
        final Entity ae = game.getEntity(caa.getEntityId());
        final Targetable target = game.getTarget(caa.getTargetType(),
                                                 caa.getTargetId());
        
        int damage = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;

        Entity te = null;
        if ((target != null)
            && (target.getTargetType() == Targetable.TYPE_ENTITY)) {
            te = (Entity) target;
        }
        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());

        
        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        Report r;

        
        Building bldg = game.getBoard().getBuildingAt(caa.getTargetPos());

        if (ae == null) {
            return;
        }

        final int direction = ae.getFacing();

        ae.setDisplacementAttack(null);

        if (lastEntityId != caa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }
        if ((target == null)
            || ((target.getTargetType() == Targetable.TYPE_ENTITY) && (te
                                                                               .isDestroyed() || te.isDoomed() || te
                                                                               .getCrew()
                                                                               .isDead()))) {
            r = new Report(4190);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
           
            return;
        }

        if (ae.isProne()) {
            r = new Report(4195);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
            return;
        }

        if (ae.isImmobile()) {
            r = new Report(4200);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
            return;
        }

        if ((te != null) && (te.isProne()) && (ae instanceof Mech)) {
            r = new Report(4205);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
            return;
        }

        r = new Report(4210);
        r.subject = ae.getId();
        r.indent();
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (!target.getPosition().equals(caa.getTargetPos())) {
            r = new Report(4215);
            r.subject = ae.getId();
            addReport(r);
            addReport(doEntityDisplacement(ae, ae.getPosition(),
                                           caa.getTargetPos(), null));
            return;
        }

        if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            roll = -12;
            r = new Report(4220);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            roll = Integer.MAX_VALUE;
            r = new Report(4225);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            addReport(r);
            if (glancing) {
                r = new Report(4030);
                r.subject = ae.getId();
                addReport(r);
            }
            if (directBlow) {
                r = new Report(4032);
                r.subject = ae.getId();
                addReport(r);
            }

        }

        if (roll < toHit.getValue()) {
            Coords src = ae.getPosition();
            Coords dest = Compute.getMissedChargeDisplacement(game, ae.getId(),
                                                              src, direction);

            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            
            addReport(doEntityDisplacement(ae, src, dest, null));
        } else if ((target.getTargetType() == Targetable.TYPE_BUILDING)
                   || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) { 
            
            r = new Report(4040);
            r.subject = ae.getId();
            addReport(r);
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            addReport(damageInfantryIn(bldg, damage, target.getPosition()));

            int toAttacker = ChargeAttackAction.getDamageTakenBy(ae, bldg,
                                                                 target.getPosition());
            HitData hit = ae.rollHitLocation(ToHitData.HIT_NORMAL,
                                             ae.sideTable(target.getPosition()));
            hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
            addReport(damageEntity(ae, hit, toAttacker, false, DamageType.NONE,
                                   false, false, throughFront));
            addNewLines();
            entityUpdate(ae.getId());

            
        } else {
            
            resolveChargeDamage(ae, te, toHit, direction, glancing,
                                throughFront);
        }
    }

    private void resolveTeleMissileAttack(PhysicalResult pr, int lastEntityId) {
        final TeleMissileAttackAction taa = (TeleMissileAttackAction) pr.aaa;
        final Entity ae = game.getEntity(taa.getEntityId());
        if (!(ae instanceof TeleMissile)) {
            return;
        }
        TeleMissile tm = (TeleMissile) ae;
        final Targetable target = game.getTarget(taa.getTargetType(),
                                                 taa.getTargetId());
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        Entity te = null;
        if ((target != null)
            && (target.getTargetType() == Targetable.TYPE_ENTITY)) {
            te = (Entity) target;
        }

        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }

        Report r;

        if (lastEntityId != taa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        if ((target == null)
            || ((target.getTargetType() == Targetable.TYPE_ENTITY) && (te
                                                                               .isDestroyed() || te.isDoomed() || te
                                                                               .getCrew()
                                                                               .isDead()))) {
            r = new Report(4190);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
            return;
        }

        r = new Report(9031);
        r.subject = ae.getId();
        r.indent();
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (ae.damageThisRound > 10) {
            toHit.addModifier((int) (Math.floor(ae.damageThisRound / 10.0)),
                              "damage taken");
        }

        Entity ride = game.getEntity(tm.getOriginalRideId());
        if ((null != ride) && (ride instanceof Aero)) {
            Aero aride = (Aero) ride;
            int cic = aride.getCICHits();
            if (cic > 0) {
                toHit.addModifier(cic * 2, "CIC damage");
            }

            int sensors = aride.getSensorHits();
            if ((sensors > 0) && (sensors < 3)) {
                toHit.addModifier(sensors, "sensor damage");
            }
            if (sensors > 2) {
                toHit.addModifier(+5, "sensors destroyed");
            }

            
            int fcs = aride.getFCSHits();
            if (fcs > 0) {
                toHit.addModifier(fcs * 2, "fcs damage");
            }
        }

        if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            roll = Integer.MAX_VALUE;
            r = new Report(4225);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
        } else {
            
            r = new Report(9032);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(toHit.getDesc());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
        }

        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
        } else {
            
            HitData hit = te.rollHitLocation(ToHitData.HIT_NORMAL,
                                             te.sideTable(ae.getPosition(), true));
            hit.setCapital(true);
            hit.setCapMisCritMod(tm.getCritMod());
            addReport(damageEntity(te, hit,
                                   TeleMissileAttackAction.getDamageFor(ae), false,
                                   DamageType.NONE, false, false, throughFront));
            destroyEntity(ae, "successful attack");
        }

    }

    private void resolveRamAttack(PhysicalResult pr, int lastEntityId) {
        final RamAttackAction raa = (RamAttackAction) pr.aaa;
        final Entity ae = game.getEntity(raa.getEntityId());
        final Targetable target = game.getTarget(raa.getTargetType(),
                                                 raa.getTargetId());
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        Entity te = null;
        if ((target != null)
            && (target.getTargetType() == Targetable.TYPE_ENTITY)) {
            te = (Entity) target;
        }

        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }

        Report r;

        boolean glancing = Compute.d6(1) == 6;

        
        ae.setRamming(false);

        if (lastEntityId != raa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        
        if ((target == null)
            || ((target.getTargetType() == Targetable.TYPE_ENTITY) && (te
                                                                               .isDestroyed() || te.isDoomed() || te
                                                                               .getCrew()
                                                                               .isDead()))) {
            r = new Report(4190);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
            return;
        }

        
        int steelroll = Compute.d6(2);
        r = new Report(9020);
        r.subject = ae.getId();
        r.add(steelroll);

        if (steelroll >= 11) {
            r.choose(true);
            addReport(r);
        } else {
            r.choose(false);
            addReport(r);
            return;
        }

        if (ae.isImmobile()) {
            r = new Report(4200);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
            return;
        }

        r = new Report(9030);
        r.subject = ae.getId();
        r.indent();
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            roll = Integer.MAX_VALUE;
            r = new Report(4225);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
        }

        if (roll < toHit.getValue()) {
            
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
        } else {
            
            resolveRamDamage((Aero) ae, (Aero) te, toHit, glancing,
                             throughFront);
        }

    }

    private void resolveRamDamage(Aero ae, Entity te, ToHitData toHit,
                                  boolean glancing, boolean throughFront) {

        int damage = RamAttackAction.getDamageFor(ae, te);
        int damageTaken = RamAttackAction.getDamageTakenBy(ae, te);
        if (glancing) {
            
            if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                damage = (int) Math.ceil(damage / 2.0);
            } else {
                damage = (int) Math.floor(damage / 2.0);
            }
            damageTaken = (int) Math.floor(damageTaken / 2.0);
        }

        
        if (te.isCapitalScale()
            && !game.getOptions().booleanOption("aero_sanity")) {
            damage = (int) Math.floor(damage / 10.0);
        }
        if (ae.isCapitalScale()
            && !game.getOptions().booleanOption("aero_sanity")) {
            damageTaken = (int) Math.floor(damageTaken / 10.0);
        }

        Report r;

        if (glancing) {
            r = new Report(9015);
            r.subject = ae.getId();
            r.indent(1);
            addReport(r);
        }

        r = new Report(4240);
        r.subject = ae.getId();
        r.add(damageTaken);
        r.indent();
        addReport(r);

        HitData hit = ae.rollHitLocation(ToHitData.HIT_NORMAL,
                                         ae.sideTable(te.getPosition(), true));
        
        if ((2 * ae.getOArmor(hit)) < damageTaken) {
            addReport(destroyEntity(ae, "by massive ramming damage", false));
        } else {
            addReport(damageEntity(ae, hit, damageTaken, false,
                                   DamageType.NONE, false, false, throughFront));
        }

        r = new Report(4230);
        r.subject = ae.getId();
        r.add(damage);
        r.add(toHit.getTableDesc());
        r.indent();
        addReport(r);

        hit = te.rollHitLocation(toHit.getHitTable(), toHit.getSideTable());
        if ((2 * te.getOArmor(hit)) < damage) {
            addReport(destroyEntity(te, "by massive ramming damage", false));
        } else {
            addReport(damageEntity(te, hit, damage, false, DamageType.NONE,
                                   false, false, throughFront));
        }
    }

    private void resolveChargeDamage(Entity ae, Entity te, ToHitData toHit,
                                     int direction) {
        resolveChargeDamage(ae, te, toHit, direction, false, true);
    }

    private void resolveChargeDamage(Entity ae, Entity te, ToHitData toHit,
                                     int direction, boolean glancing, boolean throughFront) {

        PilotingRollData chargePSR = null;
        
        if (!ae.isProne()) {
            chargePSR = new PilotingRollData(ae.getId(), 2, "charging");
        }

        int damage = ChargeAttackAction.getDamageFor(ae, te, game.getOptions()
                                                                 .booleanOption("tacops_charge_damage"),
                                                     toHit.getMoS());

        int damageTaken = ChargeAttackAction.getDamageTakenBy(ae, te, game
                .getOptions().booleanOption("tacops_charge_damage"));
        if (glancing) {
            
            if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                damage = (int) Math.ceil(damage / 2.0);
            } else {
                damage = (int) Math.floor(damage / 2.0);
            }
            damageTaken = (int) Math.floor(damageTaken / 2.0);
        }
        boolean bDirect = false;
        int directBlowCritMod = toHit.getMoS() / 3;
        if (game.getOptions().booleanOption("tacops_direct_blow")
            && ((toHit.getMoS() / 3) >= 1)) {
            damage += toHit.getMoS() / 3;
            bDirect = false;
        }

        final boolean targetInBuilding = Compute.isInBuilding(game, te);

        Building bldg = game.getBoard().getBuildingAt(te.getPosition());

        int bldgAbsorbs = 0;
        if (targetInBuilding && (bldg != null)) {
            bldgAbsorbs = bldg.getAbsorbtion(te.getPosition());
        }

        Report r;

        r = new Report(4240);
        r.subject = ae.getId();
        r.add(damageTaken);
        r.indent();
        addReport(r);

        if (ae instanceof Tank) {
            r = new Report(4241);
            r.indent();
            addReport(r);
            int side = Compute.targetSideTable(te, ae);
            int mod = ((Tank) ae).getMotiveSideMod(side);
            addReport(vehicleMotiveDamage((Tank) ae, mod));
        }

        int[] spikes = new int[ae.locations()];
        for (int i = 0; i < ae.locations(); i++) {
            spikes[i] = 0;
        }
        for (Mounted m : ae.getMisc()) {
            if ((m.getLocation() != Entity.LOC_NONE)
                && m.getType().hasFlag(MiscType.F_SPIKES)) {
                spikes[m.getLocation()] = 1;
            }
        }
        int spikeDamage = 0;

        while (damageTaken > 0) {
            int cluster = Math.min(5, damageTaken);
            HitData hit = ae.rollHitLocation(toHit.getHitTable(),
                                             ae.sideTable(te.getPosition()));
            hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
            if (spikes[hit.getLocation()] == 1) {
                r = new Report(4335);
                r.indent(2);
                r.subject = ae.getId();
                addReport(r);
                spikes[hit.getLocation()] = 0;
                checkBreakSpikes(ae, hit.getLocation());
                spikeDamage += 2;
            }
            addReport(damageEntity(ae, hit, cluster, false, DamageType.NONE,
                                   false, false, throughFront));
            damageTaken -= cluster;
        }

        damage += spikeDamage;
        r = new Report(4230);
        r.subject = ae.getId();
        r.add(damage);
        r.add(toHit.getTableDesc());
        r.indent();
        addReport(r);

        if ((te instanceof Tank) && !(te instanceof VTOL)) {
            r = new Report(4242);
            r.indent();
            addReport(r);

            int side = Compute.targetSideTable(ae, te);
            int mod = ((Tank) te).getMotiveSideMod(side);
            addReport(vehicleMotiveDamage((Tank) te, mod));
        }

        spikes = new int[te.locations()];
        for (int i = 0; i < te.locations(); i++) {
            spikes[i] = 0;
        }
        for (Mounted m : te.getMisc()) {
            if ((m.getLocation() != Entity.LOC_NONE)
                && m.getType().hasFlag(MiscType.F_SPIKES)) {
                spikes[m.getLocation()] = 1;
            }
        }
        spikeDamage = 0;

        while (damage > 0) {
            int cluster = Math.min(5, damage);
            damage -= cluster;
            if (bldgAbsorbs > 0) {
                int toBldg = Math.min(bldgAbsorbs, cluster);
                cluster -= toBldg;
                addNewLines();
                Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                               te.getPosition());
                for (Report report : buildingReport) {
                    report.subject = ae.getId();
                }
                addReport(buildingReport);

                damage = (int) Math.floor(bldg.getDamageToScale() * damage);
            }

            if (cluster == 0) {
                r = new Report(4235);
                r.subject = ae.getId();
                r.addDesc(te);
                r.indent();
                addReport(r);
            } else {
                HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                                 toHit.getSideTable());
                hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                if (bDirect) {
                    hit.makeDirectBlow(directBlowCritMod);
                }
                if (spikes[hit.getLocation()] == 1) {
                    r = new Report(4330);
                    r.indent(2);
                    r.subject = ae.getId();
                    addReport(r);
                    spikes[hit.getLocation()] = 0;
                    checkBreakSpikes(te, hit.getLocation());
                    cluster = 1;
                    spikeDamage += 2;
                }
                addReport(damageEntity(te, hit, cluster, false,
                                       DamageType.NONE, false, false, throughFront));
            }
        }
        
        if (ae instanceof Mech) {
            addReport(damageEntity(ae, new HitData(Mech.LOC_CT), spikeDamage,
                                   false, DamageType.NONE, false, false, throughFront));
        } else if (ae instanceof Tank) {
            addReport(damageEntity(ae, new HitData(Tank.LOC_FRONT),
                                   spikeDamage, false, DamageType.NONE, false, false,
                                   throughFront));
        }

        Coords src = te.getPosition();
        Coords dest = src.translated(direction);

        if (Compute.isValidDisplacement(game, te.getId(), te.getPosition(),
                                        direction)) {
            addNewLines();
            addReport(doEntityDisplacement(te, src, dest, new PilotingRollData(
                    te.getId(), 2, "was charged")));
            addReport(doEntityDisplacement(ae, ae.getPosition(), src, chargePSR));
        }

        addNewLines();

        if ((te instanceof Mech) && ((Mech) te).isIndustrial()) {
            ((Mech) te).setCheckForCrit(true);
        }

    } 

    private void resolveLayExplosivesAttack(PhysicalResult pr, int lastEntityId) {
        final LayExplosivesAttackAction laa = (LayExplosivesAttackAction) pr.aaa;
        final Entity ae = game.getEntity(laa.getEntityId());
        if (ae instanceof Infantry) {
            Infantry inf = (Infantry) ae;
            if (inf.turnsLayingExplosives < 0) {
                inf.turnsLayingExplosives = 0;
                Report r = new Report(4270);
                r.subject = inf.getId();
                r.addDesc(inf);
                addReport(r);
            } else {
                Building building = game.getBoard().getBuildingAt(
                        ae.getPosition());
                if (building != null) {
                    building.addDemolitionCharge(ae.getOwner().getId(),
                                                 pr.damage);
                    Report r = new Report(4275);
                    r.subject = inf.getId();
                    r.addDesc(inf);
                    r.add(pr.damage);
                    addReport(r);
                }
                inf.turnsLayingExplosives = -1;
            }
        }
    }

    private void resolveDfaAttack(PhysicalResult pr, int lastEntityId) {
        final DfaAttackAction daa = (DfaAttackAction) pr.aaa;
        final Entity ae = game.getEntity(daa.getEntityId());

        
        if (ae == null) {
            return;
        }

        final IHex aeHex = game.getBoard().getHex(ae.getPosition());
        final IHex teHex = game.getBoard().getHex(daa.getTargetPos());
        final Targetable target = game.getTarget(daa.getTargetType(),
                                                 daa.getTargetId());
        
        int damage = pr.damage;
        final ToHitData toHit = pr.toHit;
        int roll = pr.roll;
        Entity te = null;
        if ((target != null)
            && (target.getTargetType() == Targetable.TYPE_ENTITY)) {
           
            te = (Entity) target;
            IHex hex = game.getBoard().getHex(te.getPosition());
            if (hex.containsTerrain(Terrains.WATER)) {
                if (te.relHeight() < 0) {
                    damage = (int) Math.ceil(damage * 0.5f);
                }
            }
        }
        boolean throughFront = true;
        if (te != null) {
            throughFront = Compute
                    .isThroughFrontHex(game, ae.getPosition(), te);
        }
        final boolean glancing = game.getOptions().booleanOption(
                "tacops_glancing_blows")
                                 && (roll == toHit.getValue());
        
        toHit.setMoS(roll - Math.max(2, toHit.getValue()));
        final boolean directBlow = game.getOptions().booleanOption(
                "tacops_direct_blow")
                                   && ((toHit.getMoS() / 3) >= 1);

        Report r;

        Building bldg = game.getBoard().getBuildingAt(daa.getTargetPos());

        final int direction = ae.getFacing();

        if (lastEntityId != daa.getEntityId()) {
            
            r = new Report(4005);
            r.subject = ae.getId();
            r.addDesc(ae);
            addReport(r);
        }

        if ((target == null)
                || ((target.getTargetType() == Targetable.TYPE_ENTITY) && (te
                        .isDestroyed() || te.isDoomed() || te.getCrew()
                        .isDead()))) {
            r = new Report(4245);
            r.subject = ae.getId();
            r.indent();
            addReport(r);
            
            ae.setDisplacementAttack(null);
            if (ae.isProne()) {
                
                addReport(doEntityFall(ae, daa.getTargetPos(), 2, 3,
                                       ae.getBasePilotingRoll(), false));

            } else {
                
                ae.setElevation(ae.calcElevation(aeHex, teHex, 0, false, false));
                addReport(doEntityDisplacement(ae, ae.getPosition(),
                                               daa.getTargetPos(), new PilotingRollData(ae.getId(), 4,
                                                                                        "executed death from above")));
            }
            return;
        }

        r = new Report(4246);
        r.subject = ae.getId();
        r.indent();
        r.add(target.getDisplayName());
        r.newlines = 0;
        addReport(r);

        if (!target.getPosition().equals(daa.getTargetPos())) {
            r = new Report(4215);
            r.subject = ae.getId();
            addReport(r);
            
            ae.setDisplacementAttack(null);
            addReport(doEntityFallsInto(ae, ae.getElevation(),
                                        ae.getPosition(), daa.getTargetPos(),
                                        ae.getBasePilotingRoll(), true));
            return;
        }

        if (ae.isProne() || !ae.isActive()) {
            roll = -12;
            r = new Report(4250);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            r.newlines--;
            addReport(r);
        } else if (toHit.getValue() == TargetRoll.IMPOSSIBLE) {
            roll = -12;
            r = new Report(4255);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
        } else if (toHit.getValue() == TargetRoll.AUTOMATIC_SUCCESS) {
            r = new Report(4260);
            r.subject = ae.getId();
            r.add(toHit.getDesc());
            addReport(r);
            roll = Integer.MAX_VALUE;
        } else {
            
            r = new Report(4025);
            r.subject = ae.getId();
            r.add(toHit.getValue());
            r.add(roll);
            r.newlines = 0;
            addReport(r);
            if (glancing) {
                r = new Report(4030);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
            if (directBlow) {
                r = new Report(4032);
                r.subject = ae.getId();
                r.newlines = 0;
                addReport(r);
            }
        }

        if (roll < toHit.getValue()) {
            Coords dest = te.getPosition();
            Coords targetDest = Compute.getPreferredDisplacement(game,
                                                                 te.getId(), dest, direction);
            r = new Report(4035);
            r.subject = ae.getId();
            addReport(r);
            if (targetDest != null) {
                
                addReport(doEntityDisplacement(te, dest, targetDest, null));
                
                r = new Report(4265);
                r.subject = ae.getId();
                r.addDesc(ae);
                r.add(dest.getBoardNum(), true);
                r.indent();
                addReport(r);
                
                ae.setDisplacementAttack(null);
                addReport(doEntityFall(ae, dest, 2, 3,
                                       ae.getBasePilotingRoll(), false), 1);
                Entity violation = Compute.stackingViolation(game, ae.getId(),
                                                             dest);
                if (violation != null) {
                    
                    targetDest = Compute.getValidDisplacement(game,
                                                              violation.getId(), dest, direction);
                    vPhaseReport.addAll(doEntityDisplacement(violation, dest,
                                                             targetDest, new PilotingRollData(violation.getId(),
                                                                                              0, "domino effect")));  
                    if (!game.getOutOfGameEntitiesVector().contains(violation)) {
                        entityUpdate(violation.getId());
                    }
                }
            } else {
               
                addReport(destroyEntity(ae, "impossible displacement",
                                        ae instanceof Mech, ae instanceof Mech));
            }
            return;
        }

        r = new Report(4040);
        r.subject = ae.getId();
        addReport(r);

        Coords dest = target.getPosition();

        int damageTaken = DfaAttackAction.getDamageTakenBy(ae);

        if ((target.getTargetType() == Targetable.TYPE_BUILDING)
            || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) {

            
            Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                           target.getPosition());
            for (Report report : buildingReport) {
                report.subject = ae.getId();
            }
            addReport(buildingReport);

            addReport(damageInfantryIn(bldg, damage, target.getPosition()));

            ae.setDisplacementAttack(null);
        } else { 

            if (glancing) {
                if ((te instanceof Infantry) && !(te instanceof BattleArmor)) {
                    damage = (int) Math.ceil(damage / 2.0);
                } else {
                    damage = (int) Math.floor(damage / 2.0);
                }
            }
            if (directBlow) {
                damage += toHit.getMoS() / 3;
            }
            
            r = new Report(4230);
            r.subject = ae.getId();
            r.add(damage);
            r.add(toHit.getTableDesc());
            r.indent(2);
            addReport(r);

            int[] spikes = new int[te.locations()];
            for (int i = 0; i < te.locations(); i++) {
                spikes[i] = 0;
            }
            for (Mounted m : te.getMisc()) {
                if ((m.getLocation() != Entity.LOC_NONE)
                    && m.getType().hasFlag(MiscType.F_SPIKES)) {
                    spikes[m.getLocation()] = 1;
                }
            }
            int spikeDamage = 0;

            while (damage > 0) {
                int cluster = Math.min(5, damage);
                HitData hit = te.rollHitLocation(toHit.getHitTable(),
                                                 toHit.getSideTable());
                hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                if (directBlow) {
                    hit.makeDirectBlow(toHit.getMoS() / 3);
                }

                if (spikes[hit.getLocation()] == 1) {
                    r = new Report(4330);
                    r.indent(2);
                    r.newlines = 0;
                    r.subject = ae.getId();
                    addReport(r);
                    cluster = 1;
                    spikes[hit.getLocation()] = 0;
                    checkBreakSpikes(te, hit.getLocation());
                    spikeDamage += 2;
                }
                addReport(damageEntity(te, hit, cluster, false,
                                       DamageType.NONE, false, false, throughFront));
                damage -= 5;
            }

            if (spikeDamage > 0) {
                if (ae instanceof QuadMech) {
                    addReport(damageEntity(ae, new HitData(Mech.LOC_LARM),
                                           (spikeDamage + 2) / 4, false, DamageType.NONE,
                                           false, false, false));
                    addReport(damageEntity(ae, new HitData(Mech.LOC_RARM),
                                           (spikeDamage + 2) / 4, false, DamageType.NONE,
                                           false, false, false));
                    if (spikeDamage > 2) {
                        addReport(damageEntity(ae, new HitData(Mech.LOC_LLEG),
                                               spikeDamage / 4, false, DamageType.NONE, false,
                                               false, false));
                        addReport(damageEntity(ae, new HitData(Mech.LOC_RLEG),
                                               spikeDamage / 4, false, DamageType.NONE, false,
                                               false, false));
                    }
                } else {
                    addReport(damageEntity(ae, new HitData(Mech.LOC_LLEG),
                                           spikeDamage / 2, false, DamageType.NONE, false,
                                           false, false));
                    addReport(damageEntity(ae, new HitData(Mech.LOC_RLEG),
                                           spikeDamage / 2, false, DamageType.NONE, false,
                                           false, false));
                }
            }
            if (target instanceof VTOL) {
                
                addReport(applyCriticalHit(te, VTOL.LOC_ROTOR,
                                           new CriticalSlot(CriticalSlot.TYPE_SYSTEM,
                                                            VTOL.CRIT_ROTOR_DESTROYED), false, 0, false));
            }
            
            Coords targetDest = Compute.getValidDisplacement(game, te.getId(),
                                                             dest, direction);
            if (targetDest != null) {
                addReport(doEntityDisplacement(te, dest, targetDest,
                                               new PilotingRollData(te.getId(), 2,
                                                                    "hit by death from above")));
            } else {
                
                
                
                addReport(destroyEntity(te, "impossible displacement",
                                        te instanceof Mech, te instanceof Mech));
            }

            
            ae.setDisplacementAttack(null);
        }

        if (glancing) {
           
            damageTaken = (int) Math.floor(damageTaken / 2.0);
        }

        if (ae.hasQuirk(OptionsConstants.QUIRK_POS_REINFORCED_LEGS)) {
            damageTaken = (int) Math.floor(damageTaken / 2.0);
        }

        r = new Report(4240);
        r.subject = ae.getId();
        r.add(damageTaken);
        r.indent(2);
        addReport(r);
        while (damageTaken > 0) {
            int cluster = Math.min(5, damageTaken);
            HitData hit = ae.rollHitLocation(ToHitData.HIT_KICK,
                                             ToHitData.SIDE_FRONT);
            hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
            addReport(damageEntity(ae, hit, cluster));
            damageTaken -= cluster;
        }

        if (ae.hasQuirk(OptionsConstants.QUIRK_NEG_WEAK_LEGS)) {
            addNewLines();
            addReport(criticalEntity(ae, Mech.LOC_LLEG, false, 0, 0));
            addNewLines();
            addReport(criticalEntity(ae, Mech.LOC_RLEG, false, 0, 0));
            if (ae instanceof QuadMech) {
                addNewLines();
                addReport(criticalEntity(ae, Mech.LOC_LARM, false, 0, 0));
                addNewLines();
                addReport(criticalEntity(ae, Mech.LOC_RARM, false, 0, 0));
            }
        }

        addNewLines();

        if ((target.getTargetType() == Targetable.TYPE_BUILDING)
            || (target.getTargetType() == Targetable.TYPE_FUEL_TANK)) {
            return;
        }
        ae.setElevation(ae.calcElevation(aeHex, teHex, 0, false, false));
        
        addReport(doEntityDisplacement(ae, dest, dest, new PilotingRollData(
                ae.getId(), 4, "executed death from above")));

        if ((target instanceof Mech) && ((Mech) target).isIndustrial()) {
            ((Mech) target).setCheckForCrit(true);
        }
    }

    private PilotingRollData getKickPushPSR(Entity psrEntity, Entity attacker,
                                            Entity target, String reason) {
        int mod = 0;
        PilotingRollData psr = new PilotingRollData(psrEntity.getId(), mod,
                                                    reason);
        if (psrEntity.hasQuirk(OptionsConstants.QUIRK_POS_STABLE)) {
            psr.addModifier(-1, "stable", false);
        }
        if (game.getOptions().booleanOption("tacops_physical_psr")) {

            switch (target.getWeightClass()) {
                case EntityWeightClass.WEIGHT_LIGHT:
                    mod = 1;
                    break;
                case EntityWeightClass.WEIGHT_MEDIUM:
                    mod = 0;
                    break;
                case EntityWeightClass.WEIGHT_HEAVY:
                    mod = -1;
                    break;
                case EntityWeightClass.WEIGHT_ASSAULT:
                    mod = -2;
                    break;
            }
            String reportStr;
            if (mod > 0) {
                reportStr = ("weight class modifier +") + mod;
            } else {
                reportStr = ("weight class modifier ") + mod;
            }
            psr.addModifier(mod, reportStr, false);
        }
        return psr;
    }

    private void resolveHeat() {
        Report r;
        
        addReport(new Report(5000, Report.PUBLIC));
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            Entity entity = i.next();
            if ((null == entity.getPosition()) && !(entity instanceof Aero)) {
                continue;
            }
            IHex entityHex = game.getBoard().getHex(entity.getPosition());

            int hotDogMod = 0;
            if (entity.getCrew().getOptions().booleanOption("hot_dog")) {
                hotDogMod = 1;
            }
            if (entity.getTaserInterferenceHeat()) {
                entity.heatBuildup += 5;
            }
            if (entity.hasDamagedRHS() && entity.weaponFired()) {
                entity.heatBuildup += 1;
            }

            int radicalHSBonus = 0;
            Vector<Report> rhsReports = new Vector<Report>();
            if (entity.hasActivatedRadicalHS()) {
                entity.setConsecutiveRHSUses(entity.getConsecutiveRHSUses() + 1);
                if (entity instanceof Mech) {
                    radicalHSBonus = ((Mech) entity).getActiveSinks();
                } else if (entity instanceof Aero) {
                    radicalHSBonus = ((Aero) entity).getHeatSinks();
                } else {
                    System.out.println("Server.resolveHeat() Error: "
                                       + "Radical heatsinks mounted on non-mech "
                                       + "non-aero Entity!");
                }
                int rhsRoll = Compute.d6(2);
                int targetNumber = 2;
                switch (entity.getConsecutiveRHSUses()) {
                    case 1:
                        targetNumber = 3;
                        break;
                    case 2:
                        targetNumber = 5;
                        break;
                    case 3:
                        targetNumber = 7;
                        break;
                    case 4:
                        targetNumber = 10;
                        break;
                    case 5:
                        targetNumber = 11;
                        break;
                    case 6:
                    default:
                        targetNumber = 13; 
                }
                
                r = new Report(5540);
                r.subject = entity.getId();
                r.indent();
                r.addDesc(entity);
                r.add(radicalHSBonus);
                rhsReports.add(r);

                boolean rhsFailure = rhsRoll < targetNumber;
                r = new Report(5541);
                r.indent(2);
                r.subject = entity.getId();
                r.add(targetNumber);
                r.add(rhsRoll);
                r.choose(rhsFailure);
                rhsReports.add(r);

                if (rhsFailure) {
                    entity.setHasDamagedRHS(true);
                    int loc = Entity.LOC_NONE;
                    for (Mounted m : entity.getEquipment()) {
                        if (m.getType().hasFlag(MiscType.F_RADICAL_HEATSINK)) {
                            loc = m.getLocation();
                            m.setDestroyed(true);
                            break;
                        }
                    }
                    if (loc == Entity.LOC_NONE) {
                        throw new IllegalStateException("Server."
                                                        + "resolveHeat(): Could not find Radical "
                                                        + "Heatsink mount on unit that used RHS!");
                    }
                    for (int s = 0; s < entity.getNumberOfCriticals(loc); s++) {
                        CriticalSlot slot = entity.getCritical(loc, s);
                        if ((slot.getType() == CriticalSlot.TYPE_EQUIPMENT)
                            && slot.getMount().getType()
                                   .hasFlag(MiscType.F_RADICAL_HEATSINK)) {
                            slot.setHit(true);
                            break;
                        }
                    }
                }
            }

            if ((entity instanceof Aero) && !(entity instanceof ConvFighter)) {
                
                
                if ((game.getEntity(entity.getTransportId()) instanceof FighterSquadron)) {
                    continue;
                }
                Aero a = (Aero) entity;

                if (entity.isDestroyed() || entity.isDoomed()
                    || entity.getCrew().isDoomed()
                    || entity.getCrew().isDead()) {
                    continue;
                }

                entity.heatBuildup += entity.getEngineCritHeat();

                if (entity.isStealthOn()) {
                    entity.heatBuildup += 10;
                    r = new Report(5015);
                    r.subject = entity.getId();
                    addReport(r);
                }

                if (entity.hasQuirk(OptionsConstants.QUIRK_POS_COMBAT_COMPUTER)) {
                    int reduce = Math.min(entity.heatBuildup, 4);
                    r = new Report(5026);
                    r.subject = entity.getId();
                    r.add(reduce);
                    addReport(r);
                    entity.heatBuildup -= reduce;
                }

                int max_ext_heat = game.getOptions().intOption(
                        "max_external_heat"); 
                if (max_ext_heat < 0) {
                    max_ext_heat = 15; 
                }
                entity.heatBuildup += Math.min(max_ext_heat,
                                               entity.heatFromExternal);
                entity.heatFromExternal = 0;
                
                entity.heatBuildup -= Math.min(9, entity.coolFromExternal);
                entity.coolFromExternal = 0;

                entity.heat += entity.heatBuildup;

                int tosink = entity.getHeatCapacityWithWater() + radicalHSBonus;

                int safeHeat = entity.hasInfernoAmmo() ? 9 : 13;
                int possibleSinkage = ((Aero) entity).getHeatSinks()
                                      - entity.getCoolantFailureAmount();
                for (Mounted m : entity.getEquipment()) {
                    if (m.getType() instanceof AmmoType) {
                        AmmoType at = (AmmoType) m.getType();
                        if ((at.getAmmoType() == AmmoType.T_COOLANT_POD)
                            && m.isAmmoUsable()) {
                            EquipmentMode mode = m.curMode();
                            if (mode.equals("dump")) {
                                r = new Report(5260);
                                r.subject = entity.getId();
                                addReport(r);
                                m.setShotsLeft(0);
                                tosink += possibleSinkage;
                                break;
                            }
                            if (mode.equals("safe")
                                && ((entity.heat - tosink) > safeHeat)) {
                                r = new Report(5265);
                                r.subject = entity.getId();
                                addReport(r);
                                m.setShotsLeft(0);
                                tosink += possibleSinkage;
                                break;
                            }
                            if (mode.equals("efficient")
                                && ((entity.heat - tosink) >= possibleSinkage)) {
                                r = new Report(5270);
                                r.subject = entity.getId();
                                addReport(r);
                                m.setShotsLeft(0);
                                tosink += possibleSinkage;
                                break;
                            }
                        }
                    }
                }

                tosink = Math.min(tosink, entity.heat);
                entity.heat -= tosink;
                r = new Report(5035);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(entity.heatBuildup);
                r.add(tosink);
                r.add(entity.heat);
                addReport(r);
                entity.heatBuildup = 0;
                vPhaseReport.addAll(rhsReports);

                if ((entity instanceof Dropship)
                    || (entity instanceof Jumpship)) {
                    
                    if (entity.heat > 0) {
                        int bonus = (int) Math.ceil(entity.heat / 100.0);
                        game.addControlRoll(new PilotingRollData(
                                entity.getId(), bonus, "used too much heat"));
                        entity.heat = 0;
                    }
                    continue;
                }

                if (entity.isCapitalFighter() && (entity.heat > 0)) {
                    int penalty = (int) Math.ceil(entity.heat / 15.0);
                    game.addControlRoll(new PilotingRollData(entity.getId(),
                                                             penalty, "used too much heat"));
                }

                if (entity.isCapitalFighter()) {
                    continue;
                }

                int autoShutDownHeat = 30;
                boolean mtHeat = game.getOptions().booleanOption("tacops_heat");
                if (mtHeat) {
                    autoShutDownHeat = 50;
                }

                if ((entity.heat < autoShutDownHeat) && entity.isShutDown()) {
                    
                    if ((entity.getTaserShutdownRounds() == 0)
                        && (entity.getTsempEffect() != TSEMPWeapon.TSEMP_EFFECT_SHUTDOWN)) {
                        if ((entity.heat < 14) && !(entity.isManualShutdown())) {
                            
                            entity.setShutDown(false);
                            r = new Report(5045);
                            r.subject = entity.getId();
                            r.addDesc(entity);
                            addReport(r);
                        } else if (!(entity.isManualShutdown())) {
                            
                            if (!entity.getCrew().isActive()) {
                                r = new Report(5049);
                                r.subject = entity.getId();
                                r.addDesc(entity);
                                addReport(r);
                            } else {
                                
                                int startup = (4 + (((entity.heat - 14) / 4) * 2))
                                              - hotDogMod;
                                if (mtHeat) {
                                    startup -= 5;
                                    switch (entity.getCrew().getPiloting()) {
                                        case 0:
                                        case 1:
                                            startup -= 2;
                                            break;
                                        case 2:
                                        case 3:
                                            startup -= 1;
                                            break;
                                        case 6:
                                        case 7:
                                            startup += 1;
                                            break;
                                    }
                                }
                                int suroll = entity.getCrew().rollPilotingSkill();
                                r = new Report(5050);
                                r.subject = entity.getId();
                                r.addDesc(entity);
                                r.add(startup);
                                r.add(suroll);
                                if (suroll >= startup) {
                                    
                                    entity.setShutDown(false);
                                    r.choose(true);
                                } else {
                                    r.choose(false);
                                }
                                addReport(r);
                            }
                        }
                    } else {
                        
                        
                        if (entity.isBATaserShutdown()) {
                            int roll = Compute.d6(2);
                            if (roll >= 8) {
                                entity.setTaserShutdownRounds(0);
                                if (!(game.getOptions().booleanOption(
                                        "manual_shutdown") && entity
                                              .isManualShutdown())) {
                                    entity.setShutDown(false);
                                }
                                entity.setBATaserShutdown(false);
                            }
                        }
                    }
                }

                else if ((entity.heat >= 14) && !entity.isShutDown()) {
                    if (entity.heat >= autoShutDownHeat) {
                        r = new Report(5055);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        addReport(r);
                        
                        entity.setShutDown(true);
                    } else if (entity.heat >= 14) {
                        
                        if (!entity.getCrew().isActive()) {
                            r = new Report(5056);
                            r.subject = entity.getId();
                            r.addDesc(entity);
                            addReport(r);
                            entity.setShutDown(true);
                        } else {
                            int shutdown = (4 + (((entity.heat - 14) / 4) * 2))
                                           - hotDogMod;
                            if (mtHeat) {
                                shutdown -= 5;
                                switch (entity.getCrew().getPiloting()) {
                                    case 0:
                                    case 1:
                                        shutdown -= 2;
                                        break;
                                    case 2:
                                    case 3:
                                        shutdown -= 1;
                                        break;
                                    case 6:
                                    case 7:
                                        shutdown += 1;
                                        break;
                                }
                            }
                            int sdroll = Compute.d6(2);
                            r = new Report(5060);
                            r.subject = entity.getId();
                            r.addDesc(entity);
                            r.add(shutdown);
                            r.add(sdroll);
                            if (sdroll >= shutdown) {
                                
                                r.choose(true);
                                addReport(r);
                            } else {
                                
                                r.choose(false);
                                addReport(r);
                                
                                entity.setShutDown(true);
                            }
                        }
                    }
                }

                if ((entity.heat >= 5) && !a.isRandomMove()) {
                    int controlavoid = (5 + (entity.heat >= 10 ? 1 : 0)
                                        + (entity.heat >= 15 ? 1 : 0)
                                        + (entity.heat >= 20 ? 1 : 0) + (entity.heat >= 25 ? 2
                                                                                           : 0))
                                       - hotDogMod;
                    int controlroll = Compute.d6(2);
                    r = new Report(9210);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(controlavoid);
                    r.add(controlroll);
                    if (controlroll >= controlavoid) {
                        
                        r.choose(true);
                        addReport(r);
                    } else {
                        
                        r.choose(false);
                        addReport(r);
                        
                        
                        if (!a.isOutControl() && !game.getBoard().inSpace()
                            && a.isAirborne()) {
                            int loss = Compute.d6(1);
                            r = new Report(9366);
                            r.newlines = 0;
                            r.subject = entity.getId();
                            r.addDesc(entity);
                            r.add(loss);
                            addReport(r);
                            a.setAltitude(a.getAltitude() - loss);
                            
                            if (checkCrash(a, a.getPosition(), a.getAltitude())) {
                                addReport(processCrash(entity,
                                                       a.getCurrentVelocity(), a.getPosition()));
                            }
                        }
                        
                        a.setOutCtrlHeat(true);
                        a.setRandomMove(true);
                    }
                } 

                if (entity.heat >= 19) {
                    int boom = (4 + (entity.heat >= 23 ? 2 : 0) + (entity.heat >= 28 ? 2
                                                                                     : 0))
                               - hotDogMod;
                    if (mtHeat) {
                        boom += (entity.heat >= 35 ? 2 : 0)
                                + (entity.heat >= 40 ? 2 : 0)
                                + (entity.heat >= 45 ? 2 : 0);
                        
                    }
                    int boomroll = Compute.d6(2);
                    r = new Report(5065);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(boom);
                    r.add(boomroll);
                    if (boomroll >= boom) {
                        
                        r.choose(true);
                        addReport(r);
                    } else {
                        
                        r.choose(false);
                        addReport(r);
                        addReport(explodeAmmoFromHeat(entity));
                    }
                }

                if (entity.heat >= 21) {
                    int ouch = (6 + (entity.heat >= 27 ? 3 : 0)) - hotDogMod;
                    int ouchroll = Compute.d6(2);
                    r = new Report(5075);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(ouch);
                    r.add(ouchroll);
                    if (ouchroll >= ouch) {
                        
                        r.choose(true);
                        addReport(r);
                    } else {
                        
                        r.choose(false);
                        addReport(r);
                        addReport(damageCrew(entity, 1));
                    }
                }

                if ((entity.getCrew().isDead() || entity.getCrew().isDoomed())
                    && !entity.getCrew().isEjected()) {
                    r = new Report(5080);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    addReport(r);
                    addReport(destroyEntity(entity, "pilot death", true));
                }

                continue;
            }

            if (!(entity instanceof Mech)) {
                entity.heatBuildup = 0;
                entity.heatFromExternal = 0;
                entity.coolFromExternal = 0;

                if (entity.infernos.isStillBurning()) {
                    doFlamingDamage(entity);
                }
                if (entity.getTaserShutdownRounds() == 0) {
                    entity.setBATaserShutdown(false);
                    if (entity.isShutDown()
                        && !entity.isManualShutdown()
                        && (entity.getTsempEffect() != TSEMPWeapon.TSEMP_EFFECT_SHUTDOWN)) {
                        entity.setShutDown(false);
                        r = new Report(5045);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        addReport(r);
                    }
                } else if (entity.isBATaserShutdown()) {
                    
                    
                    int roll = Compute.d6(2);
                    if (roll >= 8) {
                        entity.setTaserShutdownRounds(0);
                        if (!(entity.isManualShutdown())) {
                            entity.setShutDown(false);
                        }
                        entity.setBATaserShutdown(false);
                    }
                }

                continue;
            }
            
            if (entity.infernos.isStillBurning()) {
                int infernoHeat = entity.infernos.getHeat();
                entity.heatFromExternal += infernoHeat;
                r = new Report(5010);
                r.subject = entity.getId();
                r.add(infernoHeat);
                addReport(r);
            }

            if (entity.isDestroyed() || entity.isDoomed()
                || entity.getCrew().isDoomed() || entity.getCrew().isDead()) {
                continue;
            }

            entity.heatBuildup += entity.getEngineCritHeat();

            if (((entity instanceof Mech)) && entity.isStealthOn()) {
                entity.heatBuildup += 10;
                r = new Report(5015);
                r.subject = entity.getId();
                addReport(r);
            }

            if (((entity instanceof Mech) || (entity instanceof Aero))
                && entity.hasActiveNovaCEWS()) {
                entity.heatBuildup += 2;
                r = new Report(5013);
                r.subject = entity.getId();
                addReport(r);
            }

            if ((entity instanceof Mech) && entity.isVoidSigOn()) {
                entity.heatBuildup += 10;
                r = new Report(5016);
                r.subject = entity.getId();
                addReport(r);
            }

            if ((entity instanceof Mech) && entity.isNullSigOn()) {
                entity.heatBuildup += 10;
                r = new Report(5017);
                r.subject = entity.getId();
                addReport(r);
            }

            if ((entity instanceof Mech) && entity.isChameleonShieldOn()) {
                entity.heatBuildup += 6;
                r = new Report(5014);
                r.subject = entity.getId();
                addReport(r);
            }

            if ((entity instanceof Mech)
                && (game.getPlanetaryConditions().getTemperatureDifference(
                    50, -30) != 0)
                && !((Mech) entity).hasLaserHeatSinks()) {
                if (game.getPlanetaryConditions().getTemperature() > 50) {
                    int heatToAdd = game.getPlanetaryConditions()
                                        .getTemperatureDifference(50, -30);
                    if (((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        heatToAdd /= 2;
                    }
                    entity.heatFromExternal += heatToAdd;
                    r = new Report(5020);
                    r.subject = entity.getId();
                    r.add(heatToAdd);
                    addReport(r);
                    if (((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        r = new Report(5550);
                        addReport(r);
                    }
                } else {
                    entity.heatFromExternal -= game.getPlanetaryConditions()
                                                   .getTemperatureDifference(50, -30);
                    r = new Report(5025);
                    r.subject = entity.getId();
                    r.add(game.getPlanetaryConditions()
                              .getTemperatureDifference(50, -30));
                    addReport(r);
                }
            }

            if (entityHex != null) {
                if (entityHex.containsTerrain(Terrains.FIRE)
                    && (entityHex.getFireTurn() > 0)
                    && (entity.getElevation() <= 1)) {
                    int heatToAdd = 5;
                    if ((entity instanceof Mech)
                        && ((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        heatToAdd /= 2;
                    }
                    entity.heatFromExternal += heatToAdd;
                    r = new Report(5030);
                    r.add(heatToAdd);
                    r.subject = entity.getId();
                    addReport(r);
                    if ((entity instanceof Mech)
                        && ((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        r = new Report(5550);
                        addReport(r);
                    }
                }
                int magma = entityHex.terrainLevel(Terrains.MAGMA);
                if ((magma > 0) && (entity.getElevation() == 0)) {
                    int heatToAdd = 5 * magma;
                    if ((entity instanceof Mech)
                        && ((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        heatToAdd /= 2;
                    }
                    entity.heatFromExternal += heatToAdd;
                    r = new Report(5032);
                    r.subject = entity.getId();
                    r.add(heatToAdd);
                    addReport(r);
                    if ((entity instanceof Mech)
                        && ((Mech) entity).hasIntactHeatDissipatingArmor()) {
                        r = new Report(5550);
                        addReport(r);
                    }
                }
            }

            if (entity.hasVibroblades()) {
                int vibroHeat = 0;

                vibroHeat = entity.getActiveVibrobladeHeat(Mech.LOC_RARM);
                vibroHeat += entity.getActiveVibrobladeHeat(Mech.LOC_LARM);

                if (vibroHeat > 0) {
                    r = new Report(5018);
                    r.subject = entity.getId();
                    r.add(vibroHeat);
                    addReport(r);
                    entity.heatBuildup += vibroHeat;
                }
            }

            int capHeat = 0;
            for (Mounted m : entity.getEquipment()) {
                if ((m.hasChargedCapacitor() == 1) && !m.isUsedThisRound()) {
                    capHeat += 5;
                }
                if ((m.hasChargedCapacitor() == 2) && !m.isUsedThisRound()) {
                    capHeat += 10;
                }
            }
            if (capHeat > 0) {
                r = new Report(5019);
                r.subject = entity.getId();
                r.add(capHeat);
                addReport(r);
                entity.heatBuildup += capHeat;
            }

            int max_ext_heat = game.getOptions().intOption("max_external_heat"); 
            
            if (max_ext_heat < 0) {
                max_ext_heat = 15; 
            }
            entity.heatBuildup += Math.min(max_ext_heat,
                                           entity.heatFromExternal);
            entity.heatFromExternal = 0;
            
            entity.heatBuildup -= Math.min(9, entity.coolFromExternal);
            entity.coolFromExternal = 0;

            
            if (entity.hasQuirk(OptionsConstants.QUIRK_POS_COMBAT_COMPUTER)) {
                int reduce = Math.min(entity.heatBuildup, 4);
                r = new Report(5026);
                r.subject = entity.getId();
                r.add(reduce);
                addReport(r);
                entity.heatBuildup -= reduce;
            }

            if (entity.heatBuildup < 0) {
                entity.heatBuildup = 0;
            }

            entity.heat += entity.heatBuildup;

            
            int tosink = entity.getHeatCapacityWithWater() + radicalHSBonus;

            if (entity.getCoolantFailureAmount() > 0) {
                int failureAmount = entity.getCoolantFailureAmount();
                r = new Report(5520);
                r.subject = entity.getId();
                r.add(failureAmount);
                tosink -= failureAmount;
            }

            int safeHeat = entity.hasInfernoAmmo() ? 9 : 13;
            int possibleSinkage = ((Mech) entity).getNumberOfSinks()
                                  - entity.getCoolantFailureAmount();
            for (Mounted m : entity.getEquipment()) {
                if (m.getType() instanceof AmmoType) {
                    AmmoType at = (AmmoType) m.getType();
                    if ((at.getAmmoType() == AmmoType.T_COOLANT_POD)
                        && m.isAmmoUsable()) {
                        EquipmentMode mode = m.curMode();
                        if (mode.equals("dump")) {
                            r = new Report(5260);
                            r.subject = entity.getId();
                            addReport(r);
                            m.setShotsLeft(0);
                            tosink += possibleSinkage;
                            break;
                        }
                        if (mode.equals("safe")
                            && ((entity.heat - tosink) > safeHeat)) {
                            r = new Report(5265);
                            r.subject = entity.getId();
                            addReport(r);
                            m.setShotsLeft(0);
                            tosink += possibleSinkage;
                            break;
                        }
                        if (mode.equals("efficient")
                            && ((entity.heat - tosink) >= possibleSinkage)) {
                            r = new Report(5270);
                            r.subject = entity.getId();
                            addReport(r);
                            m.setShotsLeft(0);
                            tosink += possibleSinkage;
                            break;
                        }
                    }
                }
            }

            tosink = Math.min(tosink, entity.heat);
            entity.heat -= tosink;
            r = new Report(5035);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(entity.heatBuildup);
            r.add(tosink);
            r.add(entity.heat);
            addReport(r);
            entity.heatBuildup = 0;
            vPhaseReport.addAll(rhsReports);

            if (entity.hasInfernoAmmo()) {

                if (entity.heat >= 10) {
                    int boom = (4 + (entity.heat >= 14 ? 2 : 0)
                                + (entity.heat >= 19 ? 2 : 0)
                                + (entity.heat >= 23 ? 2 : 0) + (entity.heat >= 28 ? 2
                                                                                   : 0))
                               - hotDogMod;
                    int boomroll = Compute.d6(2);
                    r = new Report(5040);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(boom);
                    r.add(boomroll);

                    if (boomroll >= boom) {
                        
                        r.choose(true);
                        addReport(r);
                    } else {
                        r.choose(false);
                        addReport(r);
                        addReport(explodeInfernoAmmoFromHeat(entity));
                    }
                }
            } 
            int autoShutDownHeat;
            boolean mtHeat;

            if (game.getOptions().booleanOption("tacops_heat")) {
                autoShutDownHeat = 50;
                mtHeat = true;
            } else {
                autoShutDownHeat = 30;
                mtHeat = false;
            }
            
            if ((entity.heat < autoShutDownHeat) && entity.isShutDown()
                && !entity.isStalled()) {
                if ((entity.getTaserShutdownRounds() == 0)
                    && (entity.getTsempEffect() != TSEMPWeapon.TSEMP_EFFECT_SHUTDOWN)) {
                    if ((entity.heat < 14) && !(entity.isManualShutdown())) {
                        
                        entity.setShutDown(false);
                        r = new Report(5045);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        addReport(r);
                    } else if (!(entity.isManualShutdown())) {
                        
                        if (!entity.getCrew().isActive()) {
                            r = new Report(5049);
                            r.subject = entity.getId();
                            r.addDesc(entity);
                            addReport(r);
                        } else {
                            
                            int startup = (4 + (((entity.heat - 14) / 4) * 2))
                                          - hotDogMod;
                            if (mtHeat) {
                                startup -= 5;
                                switch (entity.getCrew().getPiloting()) {
                                    case 0:
                                    case 1:
                                        startup -= 2;
                                        break;
                                    case 2:
                                    case 3:
                                        startup -= 1;
                                        break;
                                    case 6:
                                    case 7:
                                        startup += 1;
                                }
                            }
                            int suroll = Compute.d6(2);
                            r = new Report(5050);
                            r.subject = entity.getId();
                            r.addDesc(entity);
                            r.add(startup);
                            r.add(suroll);
                            if (suroll >= startup) {
                                
                                entity.setShutDown(false);
                                r.choose(true);
                            } else {
                                r.choose(false);
                            }
                            addReport(r);
                        }
                    }
                } else {
                   
                    if (entity.isBATaserShutdown()) {
                        int roll = Compute.d6(2);
                        if (roll >= 7) {
                            entity.setTaserShutdownRounds(0);
                            if (!(entity.isManualShutdown())) {
                                entity.setShutDown(false);
                            }
                            entity.setBATaserShutdown(false);
                        }
                    }
                }
            }

            else if ((entity.heat >= 14) && !entity.isShutDown()) {
                if (entity.heat >= autoShutDownHeat) {
                    r = new Report(5055);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    addReport(r);
                    
                    game.addPSR(new PilotingRollData(entity.getId(), 3,
                                                     "reactor shutdown"));
                    addReport(resolvePilotingRolls());
                    
                    entity.setShutDown(true);
                } else if (entity.heat >= 14) {
                    
                    if (!entity.getCrew().isActive()) {
                        r = new Report(5056);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        addReport(r);
                        entity.setShutDown(true);
                    } else {
                        int shutdown = (4 + (((entity.heat - 14) / 4) * 2))
                                       - hotDogMod;
                        if (mtHeat) {
                            shutdown -= 5;
                            switch (entity.getCrew().getPiloting()) {
                                case 0:
                                case 1:
                                    shutdown -= 2;
                                    break;
                                case 2:
                                case 3:
                                    shutdown -= 1;
                                    break;
                                case 6:
                                case 7:
                                    shutdown += 1;
                            }
                        }
                        int sdroll = Compute.d6(2);
                        r = new Report(5060);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        r.add(shutdown);
                        r.add(sdroll);
                        if (sdroll >= shutdown) {
                            
                            r.choose(true);
                            addReport(r);
                        } else {
                            
                            r.choose(false);
                            addReport(r);
                            
                            game.addPSR(new PilotingRollData(entity.getId(), 3,
                                                             "reactor shutdown"));
                            addReport(resolvePilotingRolls());
                            
                            entity.setShutDown(true);
                        }
                    }
                }
            }

            if (entity.heat >= 19) {
                int boom = (4 + (entity.heat >= 23 ? 2 : 0) + (entity.heat >= 28 ? 2
                                                                                 : 0))
                           - hotDogMod;
                if (mtHeat) {
                    boom += (entity.heat >= 35 ? 2 : 0)
                            + (entity.heat >= 40 ? 2 : 0)
                            + (entity.heat >= 45 ? 2 : 0);
                  
                }
                if ((entity instanceof Mech)
                    && ((Mech) entity).hasLaserHeatSinks()) {
                    boom--;
                }
                int boomroll = Compute.d6(2);
                r = new Report(5065);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(boom);
                r.add(boomroll);
                if (boomroll >= boom) {
                    
                    r.choose(true);
                    addReport(r);
                } else {
                    
                    r.choose(false);
                    addReport(r);
                    addReport(explodeAmmoFromHeat(entity));
                }
            }

            int lifeSupportCritCount = 0;
            boolean torsoMountedCockpit = ((Mech) entity).getCockpitType() == Mech.COCKPIT_TORSO_MOUNTED;
            if ((entity instanceof Mech) && torsoMountedCockpit) {
                lifeSupportCritCount = entity.getHitCriticals(
                        CriticalSlot.TYPE_SYSTEM, Mech.SYSTEM_LIFE_SUPPORT,
                        Mech.LOC_RT);
                lifeSupportCritCount += entity.getHitCriticals(
                        CriticalSlot.TYPE_SYSTEM, Mech.SYSTEM_LIFE_SUPPORT,
                        Mech.LOC_LT);
            } else {
                lifeSupportCritCount = entity.getHitCriticals(
                        CriticalSlot.TYPE_SYSTEM, Mech.SYSTEM_LIFE_SUPPORT,
                        Mech.LOC_HEAD);
            }
            int damageHeat = entity.heat;
            if (entity.hasQuirk(OptionsConstants.QUIRK_POS_IMP_LIFE_SUPPORT)) {
                damageHeat -= 5;
            }
            if (entity.hasQuirk(OptionsConstants.QUIRK_NEG_POOR_LIFE_SUPPORT)) {
                damageHeat += 5;
            }
            if ((lifeSupportCritCount > 0)
                && ((damageHeat >= 15) || (torsoMountedCockpit && (damageHeat > 0)))
                && !entity.getCrew().isDead()
                && !entity.getCrew().isDoomed()
                && !entity.getCrew().isEjected()) {
                int heatLimitDesc = 1;
                int damageToCrew = 0;
                if ((damageHeat >= 47) && mtHeat) {
                    
                    heatLimitDesc = 47;
                    damageToCrew = 5;
                } else if ((damageHeat >= 39) && mtHeat) {
                    
                    heatLimitDesc = 39;
                    damageToCrew = 4;
                } else if ((damageHeat >= 32) && mtHeat) {
                    
                    heatLimitDesc = 32;
                    damageToCrew = 3;
                } else if (damageHeat >= 25) {
                    
                    heatLimitDesc = 25;
                    damageToCrew = 2;
                } else if (damageHeat >= 15) {
                    
                    heatLimitDesc = 15;
                    damageToCrew = 1;
                }
                if ((damageHeat > 0)
                    && (entity instanceof Mech)
                    && (((Mech) entity).getCockpitType() == Mech.COCKPIT_TORSO_MOUNTED)
                    && !entity.getCrew().getOptions()
                              .booleanOption("pain_shunt")) {
                    damageToCrew += 1;
                }
                r = new Report(5070);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(heatLimitDesc);
                r.add(damageToCrew);
                addReport(r);
                addReport(damageCrew(entity, damageToCrew));
            } else if (mtHeat
                       && (entity.heat >= 32)
                       && !entity.getCrew().isDead()
                       && !entity.getCrew().isDoomed()
                       && !entity.getCrew().getOptions()
                                 .booleanOption("pain_shunt")) {
                
                int heatroll = Compute.d6(2);
                int avoidNumber = -1;
                if (entity.heat >= 47) {
                    avoidNumber = 12;
                } else if (entity.heat >= 39) {
                    avoidNumber = 10;
                } else if (entity.heat >= 32) {
                    avoidNumber = 8;
                }
                avoidNumber -= hotDogMod;
                r = new Report(5075);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(avoidNumber);
                r.add(heatroll);
                if (heatroll >= avoidNumber) {
                    
                    r.choose(true);
                    addReport(r);
                } else {
                    r.choose(false);
                    addReport(r);
                    addReport(damageCrew(entity, 1));
                }
            }

            if ((entity.getCrew().isDead() || entity.getCrew().isDoomed())
                && !entity.getCrew().isEjected()) {
                r = new Report(5080);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);
                addReport(destroyEntity(entity, "crew death", true));
            }

            if (mtHeat) {
                if (entity.heat >= 36) {
                    int damageroll = Compute.d6(2);
                    int damageNumber = -1;
                    if (entity.heat >= 44) {
                        damageNumber = 10;
                    } else if (entity.heat >= 36) {
                        damageNumber = 8;
                    }
                    damageNumber -= hotDogMod;
                    r = new Report(5085);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(damageNumber);
                    r.add(damageroll);
                    r.newlines = 0;
                    if (damageroll >= damageNumber) {
                        r.choose(true);
                        addReport(r);
                    } else {
                        r.choose(false);
                        addReport(r);
                        addReport(oneCriticalEntity(entity,
                                                    Compute.randomInt(8), false, 0));
                        
                        r = new Report(1210, Report.PUBLIC);
                        addReport(r);
                    }
                }
            }

            if (game.getOptions().booleanOption("tacops_coolant_failure")
                && (entity.getHeatCapacity() > entity
                    .getCoolantFailureAmount()) && (entity.heat >= 5)) {
                int roll = Compute.d6(2);
                int hitNumber = 10;

                hitNumber -= Math
                        .max(0, (int) Math.ceil(entity.heat / 5.0) - 2);

                r = new Report(5525);
                r.subject = entity.getId();
                r.add(entity.getShortName());
                r.add(hitNumber);
                r.add(roll);
                r.newlines = 0;
                addReport(r);
                if (roll >= hitNumber) {
                    r = new Report(5052);
                    r.subject = entity.getId();
                    addReport(r);
                    r = new Report(5526);
                    r.subject = entity.getId();
                    r.add(entity.getShortNameRaw());
                    addReport(r);
                    entity.addCoolantFailureAmount(1);
                } else {
                    r = new Report(5041);
                    r.subject = entity.getId();
                    addReport(r);
                }
            }
        }

        if (vPhaseReport.size() == 1) {
            
            addReport(new Report(1205, Report.PUBLIC));
        }
    }

    private void resolveHarjelRepairs()
    {
        Report r;
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            Entity entity = i.next();
            if (!(entity instanceof Mech)) {
                continue;
            }

            Mech me = (Mech) entity;
            for (int loc = 0; loc < me.locations(); ++loc) {
                boolean harjelII = me.hasHarJelIIIn(loc); 
                if ((harjelII || me.hasHarJelIIIIn(loc))
                    && me.isArmorDamagedThisTurn(loc)) {
                    if (me.hasRearArmor(loc)) {
                        
                        if (!((me.getArmor(loc) > 0)
                            || (me.getArmor(loc, true) > 0))) {
                            continue;
                        }

                        int toRepair = harjelII ? 2 : 4;
                        int frontRepair, rearRepair;
                        int desiredFrontRepair, desiredRearRepair;

                        Mounted harjel = null;
                        
                        for (Mounted m: me.getMisc()) {
                            if ((m.getLocation() == loc)
                                && (m.getType().hasFlag(MiscType.F_HARJEL_II)
                                    || m.getType().hasFlag(MiscType.F_HARJEL_III))) {
                                harjel = m;
                            }
                        }

                        if (harjelII) {
                            if (harjel.curMode().equals(MiscType.S_HARJEL_II_1F1R)) {
                                desiredFrontRepair = 1;
                            } else if (harjel.curMode().equals(MiscType.S_HARJEL_II_2F0R)) {
                                desiredFrontRepair = 2;
                            } else { 
                                desiredFrontRepair = 0;
                            }
                        } else { 
                            if (harjel.curMode().equals(MiscType.S_HARJEL_III_2F2R)) {
                                desiredFrontRepair = 2;
                            } else if (harjel.curMode().equals(MiscType.S_HARJEL_III_4F0R)) {
                                desiredFrontRepair = 4;
                            } else if (harjel.curMode().equals(MiscType.S_HARJEL_III_3F1R)) {
                                desiredFrontRepair = 3;
                            } else if (harjel.curMode().equals(MiscType.S_HARJEL_III_1F3R)) {
                                desiredFrontRepair = 1;
                            } else { 
                                desiredFrontRepair = 0;
                            }
                        }
                        desiredRearRepair = toRepair - desiredFrontRepair;

                        int availableFrontRepair = me.getOArmor(loc) - me.getArmor(loc);
                        int availableRearRepair = me.getOArmor(loc, true)
                                - me.getArmor(loc, true);
                        frontRepair = Math.min(availableFrontRepair, desiredFrontRepair);
                        rearRepair = Math.min(availableRearRepair, desiredRearRepair);
                        int surplus = desiredFrontRepair - frontRepair;
                        if (surplus > 0) { 
                            rearRepair = Math.min(availableRearRepair,
                                    rearRepair + surplus);
                        } else {
                            surplus = desiredRearRepair - rearRepair;
                            
                            frontRepair = Math.min(availableFrontRepair,
                                    frontRepair + surplus);
                        }

                        if (frontRepair > 0) {
                            me.setArmor(me.getArmor(loc) + frontRepair, loc);
                            r = new Report(harjelII ? 9850 : 9851);
                            r.subject = me.getId();
                            r.addDesc(entity);
                            r.add(frontRepair);
                            r.add(me.getLocationAbbr(loc));
                            addReport(r);
                        }
                        if (rearRepair > 0) {
                            me.setArmor(me.getArmor(loc, true) + rearRepair,
                                    loc, true);
                            r = new Report(harjelII ? 9850 : 9851);
                            r.subject = me.getId();
                            r.addDesc(entity);
                            r.add(rearRepair);
                            r.add(me.getLocationAbbr(loc) + " (R)");
                            addReport(r);
                        }
                    } else {
                        
                        if (!(me.getArmor(loc) > 0)) {
                            continue;
                        }
                        int toRepair = harjelII ? 2 : 4;
                        toRepair = Math.min(toRepair,
                                me.getOArmor(loc) - me.getArmor(loc));
                        me.setArmor(me.getArmor(loc) + toRepair, loc);
                        r = new Report(harjelII ? 9850 : 9851);
                        r.subject = me.getId();
                        r.addDesc(entity);
                        r.add(toRepair);
                        r.add(me.getLocationAbbr(loc));
                        addReport(r);
                    }
                }
            }
        }
    }

    private void doFlamingDamage(Entity entity) {
        Report r;
        int boomroll = Compute.d6(2);
        
        if (Entity.NONE != entity.getSwarmTargetId()) {
            return;
        }
        if ((entity.getMovementMode() == EntityMovementMode.VTOL)
            && !entity.infernos.isStillBurning()) {
           
            if (entity.getElevation() > 1) {
                return;
            }
        }
        
        if ((entity instanceof BattleArmor)
            && ((BattleArmor) entity).isFireResistant()) {
            r = new Report(5095);
            r.subject = entity.getId();
            r.indent(1);
            r.addDesc(entity);
            addReport(r);
            return;
        }

        if (entity instanceof Mech) {
            return;
        }

        if (entity instanceof Dropship) {
            return;
        }

        r = new Report(5100);
        r.subject = entity.getId();
        r.newlines = 0;
        r.addDesc(entity);
        r.add(boomroll);
        if (boomroll >= 8) {
            
            r.choose(true);
            addReport(r);
        } else {
            
            r.choose(false);
            r.newlines = 1;
            addReport(r);
            
            if (entity instanceof Tank) {
                int bonus = -2;
                if ((entity instanceof SupportTank)
                    || (entity instanceof SupportVTOL)) {
                    bonus = 0;
                }
                
                Report.addNewline(vPhaseReport);
                addReport(criticalTank((Tank) entity, Tank.LOC_FRONT, bonus, 0, true));
            } else if (entity instanceof Protomech) {
                
                HitData hit = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                     ToHitData.SIDE_FRONT);
                if (hit.getLocation() == Protomech.LOC_NMISS) {
                    r = new Report(6035);
                    r.subject = entity.getId();
                    r.indent(1);
                    addReport(r);
                } else {
                    r = new Report(6690);
                    r.subject = entity.getId();
                    r.indent(1);
                    r.add(entity.getLocationName(hit));
                    addReport(r);
                    entity.destroyLocation(hit.getLocation());
                    
                    
                    int hits = Protomech.POSSIBLE_PILOT_DAMAGE[hit
                            .getLocation()]
                               - ((Protomech) entity).getPilotDamageTaken(hit
                                                                                  .getLocation());
                    if (hits > 0) {
                        addReport(damageCrew(entity, hits));
                        ((Protomech) entity).setPilotDamageTaken(hit
                                                                         .getLocation(),
                                                                 Protomech.POSSIBLE_PILOT_DAMAGE[hit
                                                                         .getLocation()]);
                    }
                    if (entity.getTransferLocation(hit).getLocation() == Entity.LOC_DESTROYED) {
                        addReport(destroyEntity(entity, "flaming death", false,
                                                true));
                        Report.addNewline(vPhaseReport);
                    }
                }
            } else {
                
                addReport(destroyEntity(entity, "fire", false, false));
                Report.addNewline(vPhaseReport);
            }
        }
    }

    private void clearFlawedCoolingFlags(Entity entity) {
        
        if (!game.getOptions().booleanOption("stratops_quirks")) {
            return;
        }
        
        if (!((entity instanceof Mech) || ((entity instanceof Aero) && !((entity instanceof SmallCraft) || (entity
                instanceof Jumpship))))) {
            return;
        }

        
        if (!entity.hasQuirk(OptionsConstants.QUIRK_NEG_FLAWED_COOLING)) {
            return;
        }
        entity.setFallen(false);
        entity.setStruck(false);
    }

    private void checkForFlawedCooling() {

        
        if (!game.getOptions().booleanOption("stratops_quirks")) {
            return;
        }

        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {

            final Entity entity = i.next();

            
            if (!((entity instanceof Mech)
                    || ((entity instanceof Aero)
                            && !((entity instanceof SmallCraft)
                                    || (entity instanceof Jumpship))))) {
                continue;
            }

            
            if (!entity.hasQuirk(OptionsConstants.QUIRK_NEG_FLAWED_COOLING)) {
                continue;
            }

            
            if (entity.damageThisPhase >= 20) {
                addReport(doFlawedCoolingCheck("20+ damage", entity));
            }
            if (entity.hasFallen()) {
                addReport(doFlawedCoolingCheck("fall", entity));
            }
            if (entity.wasStruck()) {
                addReport(doFlawedCoolingCheck("being struck", entity));
            }
            clearFlawedCoolingFlags(entity);
        }
    }

    private Vector<Report> doFlawedCoolingCheck(String reason, Entity entity) {
        Vector<Report> out = new Vector<Report>();
        Report r = new Report(9800);
        r.addDesc(entity);
        r.add(reason);
        int roll = Compute.d6(2);
        r.add(roll);
        out.add(r);
        if (roll >= 10) {
            Report s = new Report(9805);
            entity.heatBuildup += 5;
            out.add(s);
        }

        return out;
    }

    private void checkForChainWhipGrappleChecks() {
        for (Entity ae : game.getEntitiesVector()) {
            if ((ae.getGrappled() != Entity.NONE) && ae.isChainWhipGrappled()
                    && ae.isGrappleAttacker() && !ae.isGrappledThisRound()) {
                Entity te = game.getEntity(ae.getGrappled());
                ToHitData grappleHit = GrappleAttackAction.toHit(game,
                        ae.getId(), te, ae.getGrappleSide(), true);
                int roll = Compute.d6(2);

                Report r = new Report(4317);
                r.subject = ae.getId();
                r.indent();
                r.addDesc(ae);
                r.addDesc(te);
                r.newlines = 0;
                addReport(r);

                if (grappleHit.getValue() == TargetRoll.IMPOSSIBLE) {
                    r = new Report(4300);
                    r.subject = ae.getId();
                    r.add(grappleHit.getDesc());
                    addReport(r);
                    return;
                }

                r = new Report(4025);
                r.subject = ae.getId();
                r.add(grappleHit.getValue());
                r.add(roll);
                r.newlines = 0;
                addReport(r);

                if (roll >= grappleHit.getValue()) {
                    
                    r = new Report(4040);
                    r.subject = ae.getId();
                    addReport(r);
                    
                    return;
                }
                r = new Report(4035);
                r.subject = ae.getId();
                addReport(r);

                ae.setGrappled(Entity.NONE, false);
                te.setGrappled(Entity.NONE, false);
            }
        }
    }

    private void checkForPSRFromDamage() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            final Entity entity = i.next();
            if (entity instanceof Mech) {
                if (entity.isAirborne()) {
                    
                    continue;
                }
                
                if ((entity.damageThisPhase >= 20) && !entity.isHullDown()) {
                    if (game.getOptions().booleanOption("tacops_taking_damage")) {
                        PilotingRollData damPRD = new PilotingRollData(
                                entity.getId());
                        int damMod = entity.damageThisPhase / 20;
                        damPRD.addModifier(damMod, (damMod * 20) + "+ damage");
                        int weightMod = 0;
                        if (game.getOptions().booleanOption(
                                "tacops_physical_psr")) {
                            switch (entity.getWeightClass()) {
                                case EntityWeightClass.WEIGHT_LIGHT:
                                    weightMod = 1;
                                    break;
                                case EntityWeightClass.WEIGHT_MEDIUM:
                                    weightMod = 0;
                                    break;
                                case EntityWeightClass.WEIGHT_HEAVY:
                                    weightMod = -1;
                                    break;
                                case EntityWeightClass.WEIGHT_ASSAULT:
                                    weightMod = -2;
                                    break;
                            }
                            if ((entity instanceof Mech)
                                && ((Mech) entity).isSuperHeavy()) {
                                weightMod = -4;
                            }
                            
                            damPRD.addModifier(weightMod,
                                               "weight class modifier", false);
                        }
                        if (entity.hasQuirk(OptionsConstants.QUIRK_POS_EASY_PILOT)
                            && (entity.getCrew().getPiloting() > 3)) {
                            damPRD.addModifier(-1, "easy to pilot");
                        }
                        game.addPSR(damPRD);
                    } else {
                        PilotingRollData damPRD = new PilotingRollData(
                                entity.getId(), 1, "20+ damage");
                        if (entity.hasQuirk(OptionsConstants.QUIRK_POS_EASY_PILOT)
                            && (entity.getCrew().getPiloting() > 3)) {
                            damPRD.addModifier(-1, "easy to pilot");
                        }
                        game.addPSR(damPRD);
                    }
                }
            }
            if ((entity instanceof Aero) && entity.isAirborne()
                && !game.getBoard().inSpace()) {
                
                if (entity.damageThisPhase > 0) {
                    if (!game.getOptions().booleanOption("atmospheric_control")) {
                        int damMod = entity.damageThisPhase / 20;
                        StringBuffer reportStr = new StringBuffer();
                        reportStr.append(entity.damageThisPhase)
                                 .append(" damage +").append(damMod);
                        PilotingRollData damPRD = new PilotingRollData(
                                entity.getId(), damMod, reportStr.toString());
                        if (entity.hasQuirk(OptionsConstants.QUIRK_POS_EASY_PILOT)
                            && (entity.getCrew().getPiloting() > 3)) {
                            damPRD.addModifier(-1, "easy to pilot");
                        }
                        game.addControlRoll(damPRD);
                    } else {
                        
                        if (((Aero) entity).wasCritThresh()) {
                            PilotingRollData damThresh = new PilotingRollData(
                                    entity.getId(), 0,
                                    "damage threshold exceeded");
                            if (entity.hasQuirk(OptionsConstants.QUIRK_POS_EASY_PILOT)
                                && (entity.getCrew().getPiloting() > 3)) {
                                damThresh.addModifier(-1, "easy to pilot");
                            }
                            game.addControlRoll(damThresh);
                        }
                    }
                }
            }
        }
    }

    public void checkForFlamingDamage() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext();) {
            final Entity entity = i.next();
            if ((null == entity.getPosition()) || (entity instanceof Mech)
                    || entity.isDoomed() || entity.isDestroyed()
                    || entity.isOffBoard()) {
                continue;
            }
            final IHex curHex = game.getBoard().getHex(entity.getPosition());
            final boolean underwater = curHex.containsTerrain(Terrains.WATER)
                    && (curHex.depth() > 0)
                    && (entity.getElevation() < curHex.surface());
            if (curHex.containsTerrain(Terrains.FIRE)
                    && (entity.getElevation() <= 1) && !underwater) {
                doFlamingDamage(entity);
            }
        }
    }

    private void checkForTeleMissileAttacks() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext();) {
            final Entity entity = i.next();
            if (entity instanceof TeleMissile) {
                
                Vector<Integer> potTargets = new Vector<Integer>();
                for (Entity te : game.getEntitiesVector(entity.getPosition())) {
                    if (te.isEnemyOf(entity)) {
                        
                        potTargets.add(te.getId());
                    }
                }
                if (potTargets.size() > 0) {
                    
                    Entity target = game.getEntity(potTargets.get(Compute
                            .randomInt(potTargets.size())));
                    
                    Report r = new Report(9085);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.addDesc(target);
                    addReport(r);
                    game.addTeleMissileAttack(new TeleMissileAttackAction(
                            entity, target));
                }
            }
        }
    }

    private void checkForBlueShieldDamage() {
        Report r;
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            final Entity entity = i.next();
            if (!(entity instanceof Aero) && entity.hasActiveBlueShield()
                && (entity.getBlueShieldRounds() >= 6)) {
                int roll = Compute.d6(2);
                int target = (3 + entity.getBlueShieldRounds()) - 6;
                r = new Report(1240);
                r.addDesc(entity);
                r.add(target);
                r.add(roll);
                if (roll < target) {
                    for (Mounted m : entity.getMisc()) {
                        if (m.getType().hasFlag(MiscType.F_BLUE_SHIELD)) {
                            m.setBreached(true);
                        }
                    }
                    r.choose(true);
                } else {
                    r.choose(false);
                }
                vPhaseReport.add(r);
            }
        }
    }

    private void checkForConditionDeath() {
        Report r;
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            final Entity entity = i.next();
            if ((null == entity.getPosition()) || entity.isOffBoard()) {
                
                continue;
            }
            String reason = game.getPlanetaryConditions().whyDoomed(entity);
            if (null != reason) {
                r = new Report(6015);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(reason);
                addReport(r);
                addReport(destroyEntity(entity, reason, true, true));
            }
        }
    }

    private void checkForAtmosphereDeath() {
        Report r;
        for (Iterator<Entity> i = game.getEntities(); i.hasNext();) {
            final Entity entity = i.next();
            if ((null == entity.getPosition()) || entity.isOffBoard()) {
                
                continue;
            }
            if (entity.doomedInAtmosphere() && (entity.getAltitude() == 0)) {
                r = new Report(6016);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);
                addReport(destroyEntity(entity,
                        "being in atmosphere where it can't survive", true,
                        true));
            }
        }
    }

    private void checkForIndustrialWaterDeath() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext();) {
            final Entity entity = i.next();
            if ((null == entity.getPosition()) || entity.isOffBoard()) {
                
                continue;
            }
            if ((entity instanceof Mech) && ((Mech) entity).isIndustrial()
                    && ((Mech) entity).shouldDieAtEndOfTurnBecauseOfWater()) {
                addReport(destroyEntity(entity,
                        "being in water without environmental shielding", true,
                        true));
            }

        }
    }

    private void checkForIndustrialEndOfTurn() {
        checkForIndustrialWaterDeath();
        checkForIndustrialUnstall();
        checkForIndustrialCrit(); 
        addReport(resolvePilotingRolls());
    }

    private void checkForIndustrialUnstall() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            final Entity entity = i.next();
            entity.checkUnstall(vPhaseReport);
        }
    }

    private void checkForIndustrialCrit() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext();) {
            final Entity entity = i.next();
            if ((entity instanceof Mech) && ((Mech) entity).isIndustrial()) {
                Mech mech = (Mech) entity;
                
                if (mech.isCheckForCrit()) {
                    Report r = new Report(5530);
                    r.addDesc(mech);
                    r.subject = mech.getId();
                    r.newlines = 0;
                    vPhaseReport.add(r);
                    
                    if (mech.getLevelsFallen() == 0) {
                        r = new Report(5531);
                        r.subject = mech.getId();
                        
                    } else {
                        r = new Report(5532);
                        r.subject = mech.getId();
                        r.add(mech.getLevelsFallen());
                    }
                    vPhaseReport.add(r);
                    HitData newHit = mech.rollHitLocation(ToHitData.HIT_NORMAL,
                            ToHitData.SIDE_FRONT);
                    vPhaseReport.addAll(criticalEntity(mech,
                            newHit.getLocation(), newHit.isRear(),
                            mech.getLevelsFallen(), 0));
                }
            }
        }
    }

    private void checkForSpaceDeath() {
        Report r;
        for (Iterator<Entity> i = game.getEntities(); i.hasNext();) {
            final Entity entity = i.next();
            if ((null == entity.getPosition()) || entity.isOffBoard()) {
                
                continue;
            }
            if (entity.doomedInSpace()) {
                r = new Report(6017);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);
                addReport(destroyEntity(entity,
                        "being in space where it can't survive", true, true));
            }
        }
    }

    private void checkForSuffocation() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext();) {
            final Entity entity = i.next();
            if ((null == entity.getPosition()) || entity.isOffBoard()) {
                continue;
            }
            final IHex curHex = game.getBoard().getHex(entity.getPosition());
            if ((((entity.getElevation() < 0) && ((curHex
                    .terrainLevel(Terrains.WATER) > 1) || ((curHex
                    .terrainLevel(Terrains.WATER) == 1) && entity.isProne()))) || game
                    .getPlanetaryConditions().isVacuum())
                    && (entity.getHitCriticals(CriticalSlot.TYPE_SYSTEM,
                            Mech.SYSTEM_LIFE_SUPPORT, Mech.LOC_HEAD) > 0)) {
                Report r = new Report(6020);
                r.subject = entity.getId();
                r.addDesc(entity);
                addReport(r);
                addReport(damageCrew(entity, 1));

            }
        }
    }

    private void cleanupDestroyedNarcPods() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            i.next().clearDestroyedNarcPods();
        }
    }

    private Vector<Report> resolvePilotingRolls() {
        Vector<Report> vPhaseReport = new Vector<Report>();
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            vPhaseReport.addAll(resolvePilotingRolls(i.next()));
        }
        game.resetPSRs();
        return vPhaseReport;
    }

    private Vector<Report> resolvePilotingRolls(Entity entity) {
        return resolvePilotingRolls(entity, false, entity.getPosition(),
                                    entity.getPosition());
    }

    private Vector<Report> resolvePilotingRolls(Entity entity, boolean moving,
                                                Coords src, Coords dest) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        
        if (entity.isDoomed() || entity.isDestroyed() || entity.isOffBoard()
            || !entity.isDeployed()
            || (entity.getTransportId() != Entity.NONE)) {
            return vPhaseReport;
        }

        if (entity.isAirborne()) {
            return vPhaseReport;
        }

        Report r;

        PilotingRollData rollTarget = null;
        for (Enumeration<PilotingRollData> i = game.getExtremeGravityPSRs(); i
                .hasMoreElements(); ) {
            final PilotingRollData roll = i.nextElement();
            if (roll.getEntityId() != entity.getId()) {
                continue;
            }
            
            rollTarget = roll;
            game.resetExtremeGravityPSRs(entity);
        }
        if ((rollTarget != null)
            && (rollTarget.getValue() != TargetRoll.CHECK_FALSE)) {
            
            r = new Report(2180);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(rollTarget.getLastPlainDesc());
            vPhaseReport.add(r);
            
            final int diceRoll = Compute.d6(2);
            r = new Report(2190);
            r.subject = entity.getId();
            r.add(rollTarget.getValueAsString());
            r.add(rollTarget.getDesc());
            r.add(diceRoll);
            if ((diceRoll < rollTarget.getValue())
                || (game.getOptions().booleanOption("tacops_fumbles") && (diceRoll == 2))) {
                r.choose(false);
                
                if (game.getOptions().booleanOption("tacops_fumbles")
                    && (diceRoll == 2)) {
                    r.messageId = 2306;
                }
                vPhaseReport.add(r);
                
                if ((entity.moved == EntityMovementType.MOVE_WALK)
                    || (entity.moved == EntityMovementType.MOVE_VTOL_WALK)
                    || (entity.moved == EntityMovementType.MOVE_RUN)
                    || (entity.moved == EntityMovementType.MOVE_SPRINT)
                    || (entity.moved == EntityMovementType.MOVE_VTOL_RUN)) {
                    if (entity instanceof Mech) {
                        int j = entity.mpUsed;
                        int damage = 0;
                        while (j > entity.getRunningGravityLimit()) {
                            j--;
                            damage++;
                        }
                        
                        vPhaseReport.addAll(doExtremeGravityDamage(entity,
                                                                   damage));
                    } else if (entity instanceof Tank) {
                        
                        int k = entity.gotPavementBonus ? 1 : 0;
                        if (!entity.gotPavementBonus) {
                            int j = entity.mpUsed;
                            int damage = 0;
                            while (j > (entity.getRunMP(false, false, false) + k)) {
                                j--;
                                damage++;
                            }
                            vPhaseReport.addAll(doExtremeGravityDamage(entity,
                                                                       damage));
                        }
                    }
                }
                
                if ((entity.moved == EntityMovementType.MOVE_JUMP)
                    && (entity instanceof Mech)) {
                    
                    if (game.getPlanetaryConditions().getGravity() < 1) {
                        int j = entity.mpUsed;
                        int damage = 0;
                        while (j > entity.getJumpMP(false)) {
                            j--;
                            damage++;
                        }
                        
                        vPhaseReport.addAll(doExtremeGravityDamage(entity,
                                                                   damage));
                    }
                    else if (game.getPlanetaryConditions().getGravity() > 1) {
                        int damage = entity.getWalkMP(false, false)
                                     - entity.getWalkMP();
                        
                        vPhaseReport.addAll(doExtremeGravityDamage(entity,
                                                                   damage));
                    }
                }
                
                entity.doCheckEngineStallRoll(vPhaseReport);
            } else {
                r.choose(true);
                vPhaseReport.add(r);
            }
        }
        
        if (!(entity instanceof Mech) || entity.isProne()
            || (entity.isHullDown() && entity.canGoHullDown())) {
            return vPhaseReport;
        }

        
        if (entity instanceof Mech) {
            IHex hex = game.getBoard().getHex(dest);
            int water = hex.terrainLevel(Terrains.WATER);
            if ((water > 0)
                && (entity.getElevation() != -hex.depth(true))
                && ((entity.getElevation() < 0) || ((entity.getElevation() == 0)
                                                    && (hex.terrainLevel(Terrains.BRIDGE_ELEV) != 0) && !hex
                    .containsTerrain(Terrains.ICE)))
                && !entity.isMakingDfa() && !entity.isDropping()) {
                
                if (entity.hasUMU()) {
                    return vPhaseReport;
                }
                
            }
        }
        
        Vector<PilotingRollData> rolls = new Vector<PilotingRollData>();
        
        StringBuffer reasons = new StringBuffer();
        PilotingRollData base = entity.getBasePilotingRoll();
        entity.addPilotingModifierForTerrain(base);
        for (Enumeration<PilotingRollData> i = game.getPSRs(); i
                .hasMoreElements(); ) {
            PilotingRollData psr = i.nextElement();
            if (psr.getEntityId() != entity.getId()) {
                continue;
            }
            
            if (reasons.length() > 0) {
                reasons.append("; ");
            }
            reasons.append(psr.getPlainDesc());
            PilotingRollData toUse = entity.getBasePilotingRoll();
            entity.addPilotingModifierForTerrain(toUse);
            toUse.append(psr);
            
            for (Enumeration<PilotingRollData> j = game.getPSRs(); j
                    .hasMoreElements(); ) {
                final PilotingRollData other = j.nextElement();
                if ((other.getEntityId() != entity.getId())
                    || other.equals(psr)) {
                    continue;
                }
                toUse.append(other, false);
            }
            rolls.add(toUse);
        }
        
        if (rolls.size() == 0) {
            return vPhaseReport;
        }
        
        if ((base.getValue() == TargetRoll.AUTOMATIC_FAIL)
            || (base.getValue() == TargetRoll.IMPOSSIBLE)) {
            r = new Report(2275);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(rolls.size());
            r.add(base.getDesc()); 
            vPhaseReport.add(r);
            if (moving) {
                vPhaseReport.addAll(doEntityFallsInto(entity,
                                                      entity.getElevation(), src, dest, base, true));
            } else if ((entity instanceof Mech)
                       && game.getOptions().booleanOption(
                    "tacops_falling_expanded")
                       && (entity.getCrew().getPiloting() < 6)
                       && !entity.isHullDown() && entity.canGoHullDown()) {
                if (entity.isHullDown() && entity.canGoHullDown()) {
                    r = new Report(2317);
                    r.subject = entity.getId();
                    r.add(entity.getDisplayName());
                    vPhaseReport.add(r);
                } else {
                    vPhaseReport.addAll(doEntityFall(entity, base));
                }
            } else {
                vPhaseReport.addAll(doEntityFall(entity, base));
            }
            
            entity.doCheckEngineStallRoll(vPhaseReport);
            return vPhaseReport;
        }
        
        r = new Report(2280);
        r.subject = entity.getId();
        r.addDesc(entity);
        r.add(rolls.size());
        r.add(reasons.toString()); 
        vPhaseReport.add(r);
        r = new Report(2285);
        r.subject = entity.getId();
        r.add(base.getValueAsString());
        r.add(base.getDesc()); 
        vPhaseReport.add(r);
        for (int i = 0; i < rolls.size(); i++) {
            PilotingRollData roll = rolls.elementAt(i);
            r = new Report(2290);
            r.subject = entity.getId();
            r.indent();
            r.newlines = 0;
            r.add(i + 1);
            r.add(roll.getDesc()); 
            vPhaseReport.add(r);
            if ((roll.getValue() == TargetRoll.AUTOMATIC_FAIL)
                || (roll.getValue() == TargetRoll.IMPOSSIBLE)) {
                r = new Report(2295);
                r.subject = entity.getId();
                vPhaseReport.add(r);
                if (moving) {
                    vPhaseReport.addAll(doEntityFallsInto(entity,
                                                          entity.getElevation(), src, dest, roll, true));
                } else {
                    if ((entity instanceof Mech)
                        && game.getOptions().booleanOption(
                            "tacops_falling_expanded")
                        && (entity.getCrew().getPiloting() < 6)
                        && !entity.isHullDown() && entity.canGoHullDown()) {
                        if (entity.isHullDown() && entity.canGoHullDown()) {
                            r = new Report(2317);
                            r.subject = entity.getId();
                            r.add(entity.getDisplayName());
                            vPhaseReport.add(r);
                        } else {
                            vPhaseReport.addAll(doEntityFall(entity, roll));
                        }
                    } else {
                        vPhaseReport.addAll(doEntityFall(entity, roll));
                    }
                }
                
                entity.doCheckEngineStallRoll(vPhaseReport);
                return vPhaseReport;
            }
            int diceRoll = entity.getCrew().rollPilotingSkill();
            r = new Report(2300);
            r.add(roll.getValueAsString());
            r.add(diceRoll);
            r.subject = entity.getId();
            if ((diceRoll < roll.getValue())
                || (game.getOptions().booleanOption("tacops_fumbles") && (diceRoll == 2))) {
                r.choose(false);
                
                if (game.getOptions().booleanOption("tacops_fumbles")
                    && (diceRoll == 2)) {
                    r.messageId = 2306;
                }
                vPhaseReport.add(r);
                if (moving) {
                    vPhaseReport.addAll(doEntityFallsInto(entity,
                                                          entity.getElevation(), src, dest, roll, true));
                } else {
                    if ((entity instanceof Mech)
                        && game.getOptions().booleanOption(
                            "tacops_falling_expanded")
                        && (entity.getCrew().getPiloting() < 6)
                        && !entity.isHullDown() && entity.canGoHullDown()) {
                        if ((entity.getCrew().getPiloting() > 1)
                            && ((roll.getValue() - diceRoll) < 2)) {
                            entity.setHullDown(true);
                        } else if ((entity.getCrew().getPiloting() <= 1)
                                   && ((roll.getValue() - diceRoll) < 3)) {
                            entity.setHullDown(true);
                        }
                        if (entity.isHullDown() && entity.canGoHullDown()) {
                            r = new Report(2317);
                            r.subject = entity.getId();
                            r.add(entity.getDisplayName());
                            vPhaseReport.add(r);
                        } else {
                            vPhaseReport.addAll(doEntityFall(entity, roll));
                        }
                    } else {
                        vPhaseReport.addAll(doEntityFall(entity, roll));
                    }
                }
                
                entity.doCheckEngineStallRoll(vPhaseReport);
                return vPhaseReport;
            }
            r.choose(true);
            vPhaseReport.add(r);
        }
        return vPhaseReport;
    }

    private Vector<Report> checkForTraitors() {
        Vector<Report> vFullReport = new Vector<Report>();
        
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            Entity entity = i.next();
            if (entity.isDoomed() || entity.isDestroyed()
                || entity.isOffBoard() || !entity.isDeployed()) {
                continue;
            }
            if ((entity.getTraitorId() != -1)
                && (entity.getOwnerId() != entity.getTraitorId())) {
                IPlayer p = game.getPlayer(entity.getTraitorId());
                if (null != p) {
                    Report r = new Report(7305);
                    r.subject = entity.getId();
                    r.add(entity.getDisplayName());
                    r.add(p.getName());
                    entity.setOwner(p);
                    entityUpdate(entity.getId());
                    vFullReport.add(r);
                }
                entity.setTraitorId(-1);
            }
        }
        if (!vFullReport.isEmpty()) {
            vFullReport.add(0, new Report(7300));
        }
        return vFullReport;
    }

    private Vector<Report> resolveControlRolls() {
        Vector<Report> vFullReport = new Vector<Report>();
        vFullReport.add(new Report(5001, Report.PUBLIC));
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            vFullReport.addAll(resolveControl(i.next()));
        }
        game.resetControlRolls();
        return vFullReport;
    }

    private Vector<Report> resolveControl(Entity e) {
        Vector<Report> vReport = new Vector<Report>();
        if (e.isDoomed() || e.isDestroyed() || e.isOffBoard()
            || !e.isDeployed()) {
            return vReport;
        }
        Report r;

        if ((e instanceof Aero) && (e.isAirborne() || e.isSpaceborne())) {
            Aero a = (Aero) e;

            boolean canRecover = a.isOutControl();

            if (!a.isRandomMove()) {
  
                Vector<PilotingRollData> rolls = new Vector<PilotingRollData>();
                StringBuffer reasons = new StringBuffer();
                PilotingRollData base = e.getBasePilotingRoll();
                
                if (a.isUsingManAce()) {
                    base.addModifier(-1, "maneuvering ace");
                }
                for (Enumeration<PilotingRollData> j = game.getControlRolls(); j
                        .hasMoreElements(); ) {
                    final PilotingRollData modifier = j.nextElement();
                    if (modifier.getEntityId() != e.getId()) {
                        continue;
                    }
                    
                    rolls.addElement(modifier);
                    if (reasons.length() > 0) {
                        reasons.append("; ");
                    }
                    reasons.append(modifier.getCumulativePlainDesc());
                    base.append(modifier);
                }
                
                if (rolls.size() > 0) {
                    
                    r = new Report(9310);
                    r.subject = e.getId();
                    r.addDesc(e);
                    r.add(rolls.size());
                    r.add(reasons.toString()); 
                    vReport.add(r);
                    r = new Report(2285);
                    r.subject = e.getId();
                    r.add(base.getValueAsString());
                    r.add(base.getDesc()); 
                    vReport.add(r);
                    for (int j = 0; j < rolls.size(); j++) {
                        PilotingRollData modifier = rolls.elementAt(j);
                        PilotingRollData target = base;
                        r = new Report(2290);
                        r.subject = e.getId();
                        r.indent();
                        r.newlines = 0;
                        r.add(j + 1);
                        r.add(modifier.getPlainDesc()); 
                        vReport.add(r);
                        int diceRoll = Compute.d6(2);
                        
                        if (a.isOutControl()) {
                            r = new Report(9360);
                            r.subject = e.getId();
                            r.add(target.getValueAsString());
                            r.add(diceRoll);
                            if (diceRoll < (target.getValue() - 5)) {
                                r.choose(false);
                                vReport.add(r);
                                a.setRandomMove(true);
                            } else {
                                r.choose(true);
                                vReport.add(r);
                            }
                        } else {
                            r = new Report(9315);
                            r.subject = e.getId();
                            r.add(target.getValueAsString());
                            r.add(diceRoll);
                            r.newlines = 1;
                            if (diceRoll < target.getValue()) {
                                r.choose(false);
                                vReport.add(r);
                                a.setOutControl(true);
                                
                                if ((target.getValue() - diceRoll) > 5) {
                                    r = new Report(9365);
                                    r.newlines = 0;
                                    r.subject = e.getId();
                                    vReport.add(r);
                                    a.setRandomMove(true);
                                }
                               
                                if (!game.getBoard().inSpace()
                                    && a.isAirborne()) {
                                    int loss = Compute.d6(1);
                                    r = new Report(9366);
                                    r.newlines = 0;
                                    r.subject = e.getId();
                                    r.addDesc(e);
                                    r.add(loss);
                                    vReport.add(r);
                                    a.setAltitude(a.getAltitude() - loss);
                                    
                                    if (checkCrash(a, a.getPosition(),
                                                   a.getAltitude())) {
                                        vReport.addAll(processCrash(e,
                                                                    a.getCurrentVelocity(),
                                                                    a.getPosition()));
                                        break;
                                    }
                                }
                            } else {
                                r.choose(true);
                                vReport.add(r);
                            }
                        }
                    }
                }
            }

            if (canRecover) {
                PilotingRollData base = e.getBasePilotingRoll();
                
                if ((base.getValue() == TargetRoll.AUTOMATIC_FAIL)
                    || (base.getValue() == TargetRoll.IMPOSSIBLE)) {
                    
                    r = new Report(9340);
                    r.subject = e.getId();
                    r.addDesc(e);
                    r.add(base.getDesc()); 
                    vReport.add(r);
                    return vReport;
                }
                r = new Report(9345);
                r.subject = e.getId();
                r.addDesc(e);
                r.add(base.getDesc()); 
                vReport.add(r);
                int diceRoll = Compute.d6(2);
                r = new Report(9350);
                r.subject = e.getId();
                r.add(base.getValueAsString());
                r.add(diceRoll);
                if (diceRoll < base.getValue()) {
                    r.choose(false);
                    vReport.add(r);
                } else {
                    r.choose(true);
                    vReport.add(r);
                    a.setOutControl(false);
                    a.setOutCtrlHeat(false);
                    a.setRandomMove(false);
                }
            }

        }
        return vReport;
    }

    public Vector<Report> damageCrew(Entity en, int damage) {
        Vector<Report> vDesc = new Vector<Report>();
        Crew crew = en.getCrew();
        Report r;
        if (!crew.isDead() && !crew.isEjected() && !crew.isDoomed()) {
            crew.setHits(crew.getHits() + damage);
            r = new Report(6025);
            r.subject = en.getId();
            r.indent(2);
            r.addDesc(en);
            r.add(crew.getName());
            r.add(damage);
            r.newlines = 0;
            vDesc.addElement(r);
            if (Crew.DEATH > crew.getHits()) {
                vDesc.addAll(resolveCrewDamage(en, damage));
            } else if (!crew.isDoomed()) {
                crew.setDoomed(true);
                vDesc.addAll(destroyEntity(en, "pilot death", true));
            }
        } else {
            boolean isPilot = (en instanceof Mech)
                              || (en instanceof ConvFighter)
                              || ((en instanceof Aero)
                                  && !(en instanceof Dropship)
                                  && !(en instanceof SmallCraft)
                                  && !(en instanceof Jumpship) && !(en instanceof Warship));
            if (crew.isDead() || crew.isDoomed()) {
                if (isPilot) {
                    r = new Report(6021);
                } else {
                    r = new Report(6022);
                }
            } else {
                if (isPilot) {
                    r = new Report(6023);
                } else {
                    r = new Report(6024);
                }
            }
            r.addDesc(en);
            r.add(crew.getName());
            r.indent(2);
            vDesc.add(r);
        }
        return vDesc;
    }

    private Vector<Report> resolveCrewDamage(Entity e, int damage) {
        Vector<Report> vDesc = new Vector<Report>();
        final int totalHits = e.getCrew().getHits();
        if ((e instanceof MechWarrior) || !e.isTargetable()
            || !e.getCrew().isActive() || (damage == 0)) {
            return vDesc;
        }

        if (e.getCrew().getOptions().booleanOption("pain_shunt")) {
            return vDesc;
        }

        if (e.isCapitalFighter()) {
            return vDesc;
        }

        for (int hit = (totalHits - damage) + 1; hit <= totalHits; hit++) {
            int rollTarget = Compute.getConsciousnessNumber(hit);
            if (game.getOptions().booleanOption("toughness")) {
                rollTarget -= e.getCrew().getToughness();
            }
            boolean edgeUsed = false;
            do {
                if (edgeUsed) {
                    e.getCrew().decreaseEdge();
                }
                int roll = Compute.d6(2);
                if (e.getCrew().getOptions().booleanOption("pain_resistance")) {
                    roll = Math.min(12, roll + 1);
                }
                Report r = new Report(6030);
                r.indent(2);
                r.subject = e.getId();
                r.addDesc(e);
                r.add(e.getCrew().getName());
                r.add(rollTarget);
                r.add(roll);
                if (roll >= rollTarget) {
                    e.getCrew().setKoThisRound(false);
                    r.choose(true);
                } else {
                    e.getCrew().setKoThisRound(true);
                    r.choose(false);
                    if (e.getCrew().hasEdgeRemaining()
                        && e.getCrew().getOptions()
                            .booleanOption("edge_when_ko")) {
                        edgeUsed = true;
                        vDesc.add(r);
                        r = new Report(6520);
                        r.subject = e.getId();
                        r.addDesc(e);
                        r.add(e.getCrew().getName());
                        r.add(e.getCrew().getOptions().intOption("edge"));
                    } 
                    
                } 
                vDesc.add(r);
            } while (e.getCrew().hasEdgeRemaining()
                     && e.getCrew().isKoThisRound()
                     && e.getCrew().getOptions().booleanOption("edge_when_ko"));
            
            if (e.getCrew().isKoThisRound()) {
                e.getCrew().setUnconscious(true);
                return vDesc;
            }
        }
        return vDesc;
    }

    private void resolveCrewWakeUp() {
        for (Iterator<Entity> i = game.getEntities(); i.hasNext(); ) {
            final Entity e = i.next();

            if (!e.isTargetable()
                || !e.getCrew().isUnconscious()
                || e.getCrew().isKoThisRound()
                || !((e instanceof Mech) || (e instanceof Protomech)
                     || (e instanceof MechWarrior) || ((e instanceof Aero) && !(e instanceof Jumpship)))) {
                continue;
            }
            int roll = Compute.d6(2);

            if (e.getCrew().getOptions().booleanOption("pain_resistance")) {
                roll = Math.min(12, roll + 1);
            }

            int rollTarget = Compute.getConsciousnessNumber(e.getCrew()
                                                             .getHits());
            Report r = new Report(6029);
            r.subject = e.getId();
            r.addDesc(e);
            r.add(e.getCrew().getName());
            r.add(rollTarget);
            r.add(roll);
            if (roll >= rollTarget) {
                r.choose(true);
                e.getCrew().setUnconscious(false);
            } else {
                r.choose(false);
            }
            addReport(r);
        }
    }

    private void resolveSelfDestruct() {
        for (Entity e : game.getEntitiesVector()) {
            if (e.getSelfDestructing()) {
                e.setSelfDestructing(false);
                e.setSelfDestructInitiated(true);
                Report r = new Report(5535, Report.PUBLIC);
                r.subject = e.getId();
                r.addDesc(e);
                addReport(r);
            }
        }
    }

    private void resolveShutdownCrashes() {
        for (Entity e : game.getEntitiesVector()) {
            if (e.isShutDown() && e.isAirborneVTOLorWIGE()
                && !(e.isDestroyed() || e.isDoomed())) {
                Tank t = (Tank) e;
                t.immobilize();
                addReport(forceLandVTOLorWiGE(t));
            }
        }
    }

    private Vector<Report> checkFatalThresholds(int nextAE, int prevAE) {
        Vector<Report> vDesc = new Vector<Report>();
        for (Iterator<Entity> e = game.getEntities(); e.hasNext();) {
            Entity en = e.next();
            if (!en.isCapitalFighter() || (nextAE == Entity.NONE)) {
                continue;
            }
            Aero ship = (Aero) en;
            int damage = ship.getCurrentDamage();
            double divisor = 2.0;
            if (game.getOptions().booleanOption("aero_sanity")) {
                divisor = 20.0;
            }
            if (damage >= ship.getFatalThresh()) {
                int roll = Compute.d6(2)
                        + (int) Math.floor((damage - ship.getFatalThresh())
                                / divisor);
                if (roll > 9) {
                    vDesc.addAll(destroyEntity(ship, "fatal damage threshold"));
                    if (prevAE != Entity.NONE) {
                        creditKill(ship, game.getEntity(prevAE));
                    }
                }
            }
            ship.setCurrentDamage(0);
        }
        return vDesc;
    }

    private Vector<Report> damageEntity(Entity te, HitData hit, int damage,
                                        boolean ammoExplosion) {
        return damageEntity(te, hit, damage, ammoExplosion, DamageType.NONE,
                            false, false);
    }

    public Vector<Report> damageEntity(Entity te, HitData hit, int damage) {
        return damageEntity(te, hit, damage, false, DamageType.NONE, false,
                            false);
    }

    public Vector<Report> damageEntity(Entity te, HitData hit, int damage,
                                       boolean ammoExplosion, DamageType bFrag, boolean damageIS) {
        return damageEntity(te, hit, damage, ammoExplosion, bFrag, damageIS,
                            false);
    }

    private Vector<Report> damageEntity(Entity te, HitData hit, int damage,
                                        boolean ammoExplosion, DamageType bFrag, boolean damageIS,
                                        boolean areaSatArty) {
        return damageEntity(te, hit, damage, ammoExplosion, bFrag, damageIS,
                            areaSatArty, true);
    }

    public Vector<Report> damageEntity(Entity te, HitData hit, int damage,
                                       boolean ammoExplosion, DamageType bFrag, boolean damageIS,
                                       boolean areaSatArty, boolean throughFront) {
        return damageEntity(te, hit, damage, ammoExplosion, bFrag, damageIS,
                            areaSatArty, throughFront, false, false);
    }

    public Vector<Report> damageEntity(Entity te, HitData hit, int damage,
                                       boolean ammoExplosion, DamageType bFrag, boolean damageIS,
                                       boolean areaSatArty, boolean throughFront, boolean underWater) {
        return damageEntity(te, hit, damage, ammoExplosion, bFrag, damageIS,
                            areaSatArty, throughFront, underWater, false);
    }

    public Vector<Report> damageEntity(Entity te, HitData hit, int damage,
            boolean ammoExplosion, DamageType bFrag, boolean damageIS,
            boolean areaSatArty, boolean throughFront, boolean underWater,
            boolean nukeS2S) {

        Vector<Report> vDesc = new Vector<Report>();
        Report r;
        int te_n = te.getId();

        if (te instanceof FighterSquadron) {
            FighterSquadron fs = (FighterSquadron) te;
            if (fs.getFighters().size() < 1) {
                return vDesc;
            }
            Aero fighter = fs.getFighter(hit.getLocation());
            HitData new_hit = fighter.rollHitLocation(ToHitData.HIT_NORMAL,
                                                      ToHitData.SIDE_FRONT);
            new_hit.setBoxCars(hit.rolledBoxCars());
            new_hit.setGeneralDamageType(hit.getGeneralDamageType());
            new_hit.setCapital(hit.isCapital());
            new_hit.setCapMisCritMod(hit.getCapMisCritMod());
            new_hit.setSingleAV(hit.getSingleAV());
            new_hit.setAttackerId(hit.getAttackerId());
            return damageEntity(fighter, new_hit, damage, ammoExplosion, bFrag,
                                damageIS, areaSatArty, throughFront, underWater, nukeS2S);
        }

        boolean isHeadHit = (te instanceof Mech)
                            && (((Mech) te).getCockpitType() != Mech.COCKPIT_TORSO_MOUNTED)
                            && (hit.getLocation() == Mech.LOC_HEAD)
                            && ((hit.getEffect() & HitData.EFFECT_NO_CRITICALS) != HitData.EFFECT_NO_CRITICALS);

        boolean critSI = false;
        boolean critThresh = false;

        int threshDamage = damage;
        
        if ((hit.getSingleAV() > -1)
            && !game.getOptions().booleanOption("aero_sanity")) {
            threshDamage = hit.getSingleAV();
        }

        boolean isCapital = hit.isCapital();

        if (isCapital
            && (!te.isCapitalScale() || game.getOptions().booleanOption(
                "aero_sanity"))) {
            damage = 10 * damage;
            threshDamage = 10 * threshDamage;
        }
        if (!isCapital && te.isCapitalScale()
            && !game.getOptions().booleanOption("aero_sanity")) {
            damage = (int) Math.round(damage / 10.0);
            threshDamage = (int) Math.round(threshDamage / 10.0);
        }

        int damage_orig = damage;

        HitData undoneLocation = hit.getUndoneLocation();
        while (undoneLocation != null) {
            r = new Report(6500);
            r.subject = te_n;
            r.indent(2);
            r.addDesc(te);
            r.add(te.getLocationAbbr(undoneLocation));
            vDesc.addElement(r);
            undoneLocation = undoneLocation.getUndoneLocation();
        } 
        
        if (hit.getUndoneLocation() != null) {
            r = new Report(6510);
            r.subject = te_n;
            r.indent(2);
            r.addDesc(te);
            r.add(te.getCrew().getOptions().intOption("edge"));
            vDesc.addElement(r);
        } 

        boolean autoEject = false;
        if (ammoExplosion) {
            if (te instanceof Mech) {
                Mech mech = (Mech) te;
                if (mech.isAutoEject()
                    && (!game.getOptions().booleanOption(
                        "conditional_ejection") || (game.getOptions()
                                                        .booleanOption("conditional_ejection") && mech
                                                            .isCondEjectAmmo()))) {
                    autoEject = true;
                    vDesc.addAll(ejectEntity(te, true));
                }
            }
        }
        boolean isBattleArmor = te instanceof BattleArmor;
        boolean isPlatoon = !isBattleArmor && (te instanceof Infantry);
        boolean isFerroFibrousTarget = false;
        boolean wasDamageIS = false;
        boolean tookInternalDamage = damageIS;
        IHex te_hex = null;

        boolean hardenedArmor = false;
        boolean ferroLamellorArmor = false;
        boolean reflectiveArmor = false;
        boolean reactiveArmor = false;
        boolean ballisticArmor = false;
        boolean impactArmor = false;
        boolean bar5 = te.getBARRating(hit.getLocation()) <= 5;

        if (((te instanceof Mech) || (te instanceof Tank))
            && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_HARDENED)) {
            hardenedArmor = true;
        }

        if (((te instanceof Mech) || (te instanceof Tank) || (te instanceof Aero))
            && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_FERRO_LAMELLOR)) {
            ferroLamellorArmor = true;
        }

        if ((((te instanceof Mech) || (te instanceof Tank) || (te instanceof Aero))
                    && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_REFLECTIVE))
            || (isBattleArmor && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_BA_REFLECTIVE))) {
            reflectiveArmor = true; 
            
        }

        if ((((te instanceof Mech) || (te instanceof Tank) || (te instanceof Aero))
                    && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_REACTIVE))
            || (isBattleArmor && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_BA_REACTIVE))) {
            reactiveArmor = true; 
            
        }

        if (((te instanceof Mech) || (te instanceof Tank) || (te instanceof Aero))
            && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_BALLISTIC_REINFORCED)) {
            ballisticArmor = true;
        }

        if ((te instanceof Mech)
            && (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_IMPACT_RESISTANT)) {
            impactArmor = true;
        }

        int crits;
        if ((hit.getEffect() & HitData.EFFECT_CRITICAL) == HitData.EFFECT_CRITICAL) {
            crits = 1;
        } else {
            crits = 0;
        }

        int specCrits = 0;

        int critBonus = 0;
        if (game.getOptions().booleanOption("tacops_crit_roll")
            && (damage_orig > 0)
            && ((te instanceof Mech) || (te instanceof Protomech))) {
            critBonus = Math.min((damage_orig - 1) / 5, 4);
        }

        HitData nextHit = null;

        if ((te instanceof Protomech)
            && (hit.getLocation() == Protomech.LOC_NMISS)) {
            r = new Report(6035);
            r.subject = te_n;
            r.indent(2);
            vDesc.addElement(r);
            return vDesc;
        }

        if ((crits > 0) && (te instanceof BattleArmor)) {
            
            if ((hit.getLocation() >= te.locations())
                || (te.getInternal(hit.getLocation()) <= 0)) {
                r = new Report(6037);
                r.add(hit.getLocation());
                r.subject = te_n;
                r.indent(2);
                vDesc.addElement(r);
                return vDesc;
            }
            
            r = new Report(6225);
            r.add(te.getLocationAbbr(hit));
            r.subject = te_n;
            r.indent(2);
            vDesc.addElement(r);

            crits = 0;
            damage = Math.max(
                    te.getInternal(hit.getLocation())
                    + te.getArmor(hit.getLocation()), damage);
        }

        if ((te.getArmor(hit) > 0)
            && ((te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_FERRO_FIBROUS)
                || (te.getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_LIGHT_FERRO) || (te
                                                                                                         .getArmorType(hit.getLocation()) == EquipmentType.T_ARMOR_HEAVY_FERRO))) {
            isFerroFibrousTarget = true;
        }

        if (isPlatoon && areaSatArty) {
            
            damage *= 2;
            r = new Report(6039);
            r.subject = te_n;
            r.indent(2);
            vDesc.addElement(r);
        }

        if (isPlatoon && !te.isDestroyed() && !te.isDoomed()
                && !hit.isIgnoreInfantryDoubleDamage()
                && (((Infantry) te).getDugIn() != Infantry.DUG_IN_COMPLETE)
                && !te.getCrew().getOptions().booleanOption("dermal_armor")) {
            te_hex = game.getBoard().getHex(te.getPosition());
            if ((te_hex != null) && !te_hex.containsTerrain(Terrains.WOODS)
                && !te_hex.containsTerrain(Terrains.JUNGLE)
                && !te_hex.containsTerrain(Terrains.ROUGH)
                && !te_hex.containsTerrain(Terrains.RUBBLE)
                && !te_hex.containsTerrain(Terrains.SWAMP)
                && !te_hex.containsTerrain(Terrains.BUILDING)
                && !te_hex.containsTerrain(Terrains.FUEL_TANK)
                && !te_hex.containsTerrain(Terrains.FORTIFIED)
                && !ammoExplosion) {
                
                damage *= 2;
                r = new Report(6040);
                r.subject = te_n;
                r.indent(2);
                vDesc.addElement(r);
            }
        }
        
        if ((isPlatoon || isBattleArmor) && !te.isDestroyed() && !te.isDoomed()
            && game.getPlanetaryConditions().isVacuum()) {
            
            damage *= 2;
            r = new Report(6041);
            r.subject = te_n;
            r.indent(2);
            vDesc.addElement(r);
        }
      
        switch (bFrag) {
            case ANTI_INFANTRY:
                if (isPlatoon
                    && te.getCrew().getOptions().booleanOption("dermal_armor")) {
                    int reduce = Math.min(damage - 1, Compute.d6());
                    damage -= reduce;
                    r = new Report(6042);
                    r.subject = te_n;
                    r.add(reduce);
                    r.add(damage);
                    r.indent(2);
                    vDesc.addElement(r);
                }
                break;
            case FRAGMENTATION:
            
                if (!isPlatoon) {
                    damage = 0;
                    r = new Report(6050); 
                    
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                } else {
                    r = new Report(6045); 
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                }
                break;
            case NONPENETRATING:
                if (!isPlatoon) {
                    damage = 0;
                    r = new Report(6051);
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                }
                break;
            case FLECHETTE:
                
                
                if (!isPlatoon) {
                    damage /= 2;
                    r = new Report(6060);
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                } else {
                    r = new Report(6055);
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                }
                break;
            case ACID:
                if (isFerroFibrousTarget || reactiveArmor || reflectiveArmor
                    || ferroLamellorArmor || bar5) {
                    if (te.getArmor(hit) <= 0) {
                        break; 
                    }
                    damage = te.getArmor(hit) >= 3 ? 3 : te.getArmor(hit);
                    r = new Report(6061);
                    r.subject = te_n;
                    r.indent(2);
                    r.add(damage);
                    vDesc.addElement(r);
                } else if (isPlatoon) {
                    damage *= 1.5;
                    r = new Report(6062);
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                }
                break;
            case INCENDIARY:
                
                if (isPlatoon) {
                    damage += 2;
                    r = new Report(6064);
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                }
                break;
            case ANTI_TSM:
                te.hitThisRoundByAntiTSM = true;
                break;
            case NAIL_RIVET:
                
                if ((te.getBARRating(hit.getLocation()) >= 5)
                    && (te.getArmor(hit.getLocation()) > 0)) {
                    damage = 0;
                    r = new Report(6063);
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.add(r);
                }
                break;
            default:
                
                break;
        }

        if ((te instanceof VTOL) && (hit.getLocation() == VTOL.LOC_ROTOR)
            && (hit.getGeneralDamageType() != HitData.DAMAGE_PHYSICAL)
            && !game.getOptions().booleanOption("full_rotor_hits")) {
            damage = (damage + 9) / 10;
        }

        final boolean eiStatus = te.hasActiveEiCockpit();
        
        if (!(te instanceof Mech) && !(te instanceof Protomech) && eiStatus) {
            damage += 1;
        }

        if (te instanceof Aero) {
            Aero a = (Aero) te;
            if (ammoExplosion && a.hasCase()) {
                
                
                damage /= 2;
                r = new Report(9010);
                r.subject = te_n;
                r.add(damage);
                r.indent(3);
                vDesc.addElement(r);
            }
        }

        if (isPlatoon && (((Infantry) te).getDamageDivisor() != 1.0)) {
            r = new Report(6074);
            r.subject = te_n;
            r.indent(2);
            r.add(damage);
            damage = (int) Math.ceil((damage)
                                     / ((Infantry) te).getDamageDivisor());
            r.add(damage);
            vDesc.addElement(r);
        }

        while (damage > 0) {

            if ((te instanceof Aero) && (hit.getLocation() == Aero.LOC_AFT)
                && !damageIS) {
                for (Mounted mAmmo : te.getAmmo()) {
                    if (mAmmo.isDumping() && !mAmmo.isDestroyed()
                        && !mAmmo.isHit()
                        && !(mAmmo.getType() instanceof BombType)) {
                        
                        vDesc.addAll(explodeEquipment(te, mAmmo.getLocation(),
                                                      mAmmo));
                        mAmmo.setHit(true);
                    }
                }
            }

            if (te instanceof Aero) {
                
                Aero a = (Aero) te;
                if ((threshDamage > a.getThresh(hit.getLocation()))) {
                    critThresh = true;
                    a.setCritThresh(true);
                }
            }

            if (te.isCapitalFighter()) {
                Aero a = (Aero) te;
                a.setCurrentDamage(a.getCurrentDamage() + damage);
                a.setCapArmor(a.getCapArmor() - damage);
                r = new Report(9065);
                r.subject = te_n;
                r.indent(2);
                r.newlines = 0;
                r.addDesc(te);
                r.add(damage);
                vDesc.addElement(r);
                r = new Report(6085);
                r.subject = te_n;
                r.add(Math.max(a.getCapArmor(), 0));
                vDesc.addElement(r);
                
                if (a.getCapArmor() <= 0) {
                    vDesc.addAll(destroyEntity(te,
                            "Structural Integrity Collapse"));
                    a.setCapArmor(0);
                    if (hit.getAttackerId() != Entity.NONE) {
                        creditKill(a, game.getEntity(hit.getAttackerId()));
                    }
                }
                damage = 0;
                
                checkAeroCrits(vDesc, (Aero) te, hit, damage_orig, critThresh,
                               critSI, ammoExplosion, nukeS2S);
                return vDesc;
            }

            if (!((te instanceof Aero) && ammoExplosion)) {
                
                r = new Report(6065);
                r.subject = te_n;
                r.indent(2);
                r.addDesc(te);
                r.add(damage);
                if (damageIS) {
                    r.messageId = 6070;
                }
                r.add(te.getLocationAbbr(hit));
                vDesc.addElement(r);
            }

            if (te.getInternal(hit) == IArmorState.ARMOR_DOOMED) {
                
                crits = 0;
            }

            if (!ammoExplosion
                && !hit.isFallDamage()
                && !damageIS
                && te.hasShield()
                && ((hit.getEffect() & HitData.EFFECT_NO_CRITICALS) != HitData.EFFECT_NO_CRITICALS)) {
                Mech me = (Mech) te;
                int damageNew = me.shieldAbsorptionDamage(damage,
                                                          hit.getLocation(), hit.isRear());
                
                if (damageNew != damage) {
                    int absorb = damage - damageNew;
                    te.damageThisPhase += absorb;
                    damage = damageNew;

                    r = new Report(3530);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(absorb);
                    vDesc.addElement(r);

                    if (damage <= 0) {
                        crits = 0;
                        specCrits = 0;
                        isHeadHit = false;
                    }
                }
            }

            if (te instanceof Mech) {
                Mech me = (Mech) te;
                if (me.hasCowl() && (hit.getLocation() == Mech.LOC_HEAD)
                    && !throughFront) {
                    int damageNew = me.damageCowl(damage);
                    int damageDiff = damage - damageNew;
                    me.damageThisPhase += damageDiff;
                    damage = damageNew;

                    r = new Report(3520);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damageDiff);
                    vDesc.addElement(r);
                }
            }

            if (!ammoExplosion
                && !damageIS
                && ((hit.getEffect() & HitData.EFFECT_NO_CRITICALS) != HitData.EFFECT_NO_CRITICALS)) {
                int damageNew = te.getDamageReductionFromModularArmor(hit,
                        damage, vDesc);
                int damageDiff = damage - damageNew;
                te.damageThisPhase += damageDiff;
                damage = damageNew;
            }

            if (te.hasSpotlight()) {
                boolean spotlightHittable = true;
                int loc = hit.getLocation();
                if (te instanceof Mech) {
                    if ((loc != Mech.LOC_CT) && (loc != Mech.LOC_LT)
                        && (loc != Mech.LOC_RT)) {
                        spotlightHittable = false;
                    }
                } else if (te instanceof Tank) {
                    if (te instanceof SuperHeavyTank) {
                        if ((loc != Tank.LOC_FRONT)
                            && (loc != SuperHeavyTank.LOC_FRONTRIGHT)
                            && (loc != SuperHeavyTank.LOC_FRONTLEFT)
                            && (loc != SuperHeavyTank.LOC_REARRIGHT)
                            && (loc != SuperHeavyTank.LOC_REARLEFT)) {
                            spotlightHittable = false;
                        }
                    } else if (te instanceof LargeSupportTank) {
                        if ((loc != Tank.LOC_FRONT)
                            && (loc != LargeSupportTank.LOC_FRONTRIGHT)
                            && (loc != LargeSupportTank.LOC_FRONTLEFT)
                            && (loc != LargeSupportTank.LOC_REARRIGHT)
                            && (loc != LargeSupportTank.LOC_REARLEFT)) {
                            spotlightHittable = false;
                        }
                    } else {
                        if ((loc != Tank.LOC_FRONT) && (loc != Tank.LOC_RIGHT)
                            && (loc != Tank.LOC_LEFT)) {
                            spotlightHittable = false;
                        }
                    }

                }
                if (spotlightHittable) {
                    int spotroll = Compute.d6(2);
                    r = new Report(6072);
                    r.indent(2);
                    r.subject = te_n;
                    r.add("7+");
                    r.add("Searchlight");
                    r.add(spotroll);
                    vDesc.addElement(r);
                    if (spotroll >= 7) {
                        r = new Report(6071);
                        r.subject = te_n;
                        r.indent(2);
                        r.add("Searchlight");
                        vDesc.addElement(r);
                        te.destroyOneSpotlight();
                    }
                }
            }

            if (!damageIS) {
                int nLoc = hit.getLocation();
                Entity passenger = te.getExteriorUnitAt(nLoc, hit.isRear());
                
                int avoidRoll = Compute.d6();
                if (!ammoExplosion && (null != passenger) && (avoidRoll >= 5)
                    && !passenger.isDoomed()
                    && (bFrag != DamageType.IGNORE_PASSENGER)) {
                    
                    r = new Report(6075);
                    r.subject = passenger.getId();
                    r.indent(3);
                    r.addDesc(passenger);
                    vDesc.addElement(r);

                    HitData passHit = passenger.getTrooperAtLocation(hit, te);
                    passHit.setGeneralDamageType(hit.getGeneralDamageType());

                    int absorb = 0;
                    HitData nextPassHit = passHit;
                    do {
                        int armorType = passenger.getArmorType(nextPassHit
                                                                       .getLocation());
                        boolean armorDamageReduction = false;
                        if (((armorType == EquipmentType.T_ARMOR_BA_REACTIVE)
                             && ((hit.getGeneralDamageType() ==
                                  HitData.DAMAGE_MISSILE)))
                            || (hit.getGeneralDamageType() ==
                                HitData.DAMAGE_ARMOR_PIERCING_MISSILE)) {
                            armorDamageReduction = true;
                        }
                        
                        if ((armorType == EquipmentType.T_ARMOR_BA_REFLECTIVE)
                            && (hit.getGeneralDamageType() ==
                                HitData.DAMAGE_ENERGY)) {
                            armorDamageReduction = true;
                        }
                        if (0 < passenger.getArmor(nextPassHit)) {
                            absorb += passenger.getArmor(nextPassHit);
                            if (armorDamageReduction) {
                                absorb *= 2;
                            }
                        }
                        if (0 < passenger.getInternal(nextPassHit)) {
                            absorb += passenger.getInternal(nextPassHit);
                            
                            if (((absorb % 2) != 0) && armorDamageReduction) {
                                absorb++;
                            }
                        }
                        nextPassHit = passenger
                                .getTransferLocation(nextPassHit);
                    } while ((damage > absorb)
                             && (nextPassHit.getLocation() >= 0));

                    int absorbedDamage = Math.min(damage, absorb);
                    Vector<Report> newReports = damageEntity(passenger,
                                                             passHit, absorbedDamage);
                    for (Report newReport : newReports) {
                        newReport.indent(2);
                    }
                    vDesc.addAll(newReports);

                    if (damage > absorb) {
                        
                        damage -= absorb;
                        r = new Report(6080);
                        r.subject = te_n;
                        r.indent(2);
                        r.add(damage);
                        r.addDesc(te);
                        vDesc.addElement(r);
                    } else {   
                        return vDesc;
                    }

                } else if (!ammoExplosion && (null != passenger)
                           && (avoidRoll < 5) && !passenger.isDoomed()
                           && (bFrag != DamageType.IGNORE_PASSENGER)) {
                 
                    r = new Report(6084);
                    r.subject = passenger.getId();
                    r.indent(3);
                    r.addDesc(passenger);
                    vDesc.addElement(r);
                } 

                boolean bTorso = (nLoc == Mech.LOC_CT) || (nLoc == Mech.LOC_RT)
                                 || (nLoc == Mech.LOC_LT);

                int swarmer = te.getSwarmAttackerId();
                if ((!(te instanceof Mech) || bTorso)
                    && (swarmer != Entity.NONE)
                    && ((hit.getEffect() & HitData.EFFECT_CRITICAL) == 0)
                    && (Compute.d6() >= 5)
                    && (bFrag != DamageType.IGNORE_PASSENGER)
                    && !ammoExplosion) {
                    Entity swarm = game.getEntity(swarmer);
                    
                    r = new Report(6076);
                    r.subject = swarmer;
                    r.indent(3);
                    r.addDesc(swarm);
                    vDesc.addElement(r);

                    HitData passHit = swarm.rollHitLocation(
                            ToHitData.HIT_NORMAL, ToHitData.SIDE_FRONT);

                    int absorb = 0;
                    HitData nextPassHit = passHit;
                    do {
                        if (0 < swarm.getArmor(nextPassHit)) {
                            absorb += swarm.getArmor(nextPassHit);
                        }
                        if (0 < swarm.getInternal(nextPassHit)) {
                            absorb += swarm.getInternal(nextPassHit);
                        }
                        nextPassHit = swarm.getTransferLocation(nextPassHit);
                    } while ((damage > absorb)
                             && (nextPassHit.getLocation() >= 0));

                    int absorbedDamage = Math.min(damage, absorb);
                    Vector<Report> newReports = damageEntity(swarm, passHit,
                                                             absorbedDamage);
                    for (Report newReport : newReports) {
                        newReport.indent(2);
                    }
                    vDesc.addAll(newReports);

                    if (damage > absorb) {
                        
                        damage -= absorb;
                        r = new Report(6080);
                        r.subject = te_n;
                        r.indent(2);
                        r.add(damage);
                        r.addDesc(te);
                        vDesc.addElement(r);
                    } else {
                        
                        return vDesc;
                    }
                }

                if (((te instanceof Mech) && hit.isRear() && bTorso)
                    || ((te instanceof Tank) && (hit.getLocation() == (te instanceof SuperHeavyTank ? SuperHeavyTank
                        .LOC_REAR
                                                                                                    : Tank.LOC_REAR))
                )) {
                    for (Mounted mAmmo : te.getAmmo()) {
                        if (mAmmo.isDumping() && !mAmmo.isDestroyed()
                            && !mAmmo.isHit()) {
                            
                            vDesc.addAll(explodeEquipment(te,
                                                          mAmmo.getLocation(), mAmmo));
                            mAmmo.setHit(true);
                        }
                    }
                }
            }
            
            if (!ammoExplosion && (te.getArmor(hit) > 0) && !damageIS) {
                int tmpDamageHold = -1;
                int origDamage = damage;

                if (isPlatoon) {
                    
                    int armor = te.getArmor(hit);
                    int men = te.getInternal(hit);
                    tmpDamageHold = damage % 2;
                    damage /= 2;
                    if ((tmpDamageHold == 1) && (armor >= men)) {
                        
                        tmpDamageHold = damage;
                        damage++;
                    } else {
                        
                        tmpDamageHold += damage;
                    }
                  
                } else if (ferroLamellorArmor
                           && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING)
                           && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING_MISSILE)
                           && (hit.getGeneralDamageType() != HitData.DAMAGE_IGNORES_DMG_REDUCTION)) {
                    tmpDamageHold = damage;
                    damage = (int) Math.floor((((double) damage) * 4) / 5);
                    if (damage <= 0) {
                        isHeadHit = false;
                        crits = 0;
                    }
                    r = new Report(6073);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damage);
                    vDesc.addElement(r);
                } else if (ballisticArmor
                           && ((hit.getGeneralDamageType() == HitData.DAMAGE_ARMOR_PIERCING_MISSILE)
                               || (hit.getGeneralDamageType() == HitData.DAMAGE_ARMOR_PIERCING)
                               || (hit.getGeneralDamageType() == HitData.DAMAGE_BALLISTIC) || (hit
                                                                                                       .getGeneralDamageType() == HitData.DAMAGE_MISSILE))) {
                    tmpDamageHold = damage;
                    damage = Math.max(1, damage / 2);
                    r = new Report(6088);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damage);
                    vDesc.addElement(r);
                } else if (impactArmor
                           && (hit.getGeneralDamageType() == HitData.DAMAGE_PHYSICAL)) {
                    tmpDamageHold = damage;
                    damage -= (int) Math.ceil((double) damage / 3);
                    damage = Math.max(1, damage);
                    r = new Report(6089);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damage);
                    vDesc.addElement(r);
                } else if (reflectiveArmor
                           && (hit.getGeneralDamageType() == HitData.DAMAGE_PHYSICAL)
                           && !isBattleArmor) { 
                    
                    tmpDamageHold = damage;
                    damage *= 2;
                    r = new Report(6066);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damage);
                    vDesc.addElement(r);
                } else if (reflectiveArmor && areaSatArty && !isBattleArmor) {
                    tmpDamageHold = damage; 
                    
                    damage *= 2;
                    r = new Report(6087);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damage);
                    vDesc.addElement(r);
                } else if (reflectiveArmor
                           && (hit.getGeneralDamageType() == HitData.DAMAGE_ENERGY)) {
                    tmpDamageHold = damage;
                    damage = (int) Math.floor(((double) damage) / 2);
                    if (tmpDamageHold == 1) {
                        damage = 1;
                    }
                    r = new Report(6067);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damage);
                    vDesc.addElement(r);
                } else if (reactiveArmor
                           && ((hit.getGeneralDamageType() == HitData.DAMAGE_MISSILE)
                               || (hit.getGeneralDamageType() == HitData.DAMAGE_ARMOR_PIERCING_MISSILE) ||
                               areaSatArty)) {
                    tmpDamageHold = damage;
                    damage = (int) Math.floor(((double) damage) / 2);
                    if (tmpDamageHold == 1) {
                        damage = 1;
                    }
                    r = new Report(6068);
                    r.subject = te_n;
                    r.indent(3);
                    r.add(damage);
                    vDesc.addElement(r);
                }

                if ((te instanceof Tank)
                        && game.getOptions()
                                .booleanOption("vehicles_threshold")
                        && !((te instanceof VTOL) || (te instanceof GunEmplacement))) {
                    int thresh = (int) Math.ceil((game.getOptions()
                            .booleanOption("vehicles_threshold_variable") ? te
                            .getArmor(hit) : te.getOArmor(hit))
                            / game.getOptions().intOption(
                                    "vehicles_threshold_divisor"));

                    if (hardenedArmor
                            && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING)
                            && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING_MISSILE)
                            && (hit.getGeneralDamageType() != HitData.DAMAGE_IGNORES_DMG_REDUCTION)) {
                        thresh *= 2;
                    }

                    if ((damage > thresh) || (te.getArmor(hit) < damage)) {
                        hit.setEffect(((Tank) te).getPotCrit());
                        ((Tank) te).setOverThresh(true);
                        
                        crits = ((hit.getEffect() & HitData.EFFECT_CRITICAL)
                                == HitData.EFFECT_CRITICAL) ? 1 : 0;
                    } else {
                        ((Tank) te).setOverThresh(false);
                        crits = 0;
                    }
                }

                if ((te instanceof VTOL)
                    && (hit.getLocation() == VTOL.LOC_ROTOR)
                    && te.hasWorkingMisc(MiscType.F_MAST_MOUNT, -1,
                                         VTOL.LOC_ROTOR)) {
                    r = new Report(6081);
                    r.subject = te_n;
                    r.indent(2);
                    vDesc.addElement(r);
                    for (Mounted mount : te.getMisc()) {
                        if (mount.getLocation() == VTOL.LOC_ROTOR) {
                            mount.setHit(true);
                        }
                    }
                }
                
                int armorThreshold = te.getArmor(hit);
                if (hardenedArmor
                    && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING)
                    && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING_MISSILE)
                    && (hit.getGeneralDamageType() != HitData.DAMAGE_IGNORES_DMG_REDUCTION)) {
                    armorThreshold *= 2;
                    armorThreshold -= (te.isHardenedArmorDamaged(hit)) ? 1 : 0;
                    vDesc.lastElement().newlines = 0;
                    r = new Report(6069);
                    r.subject = te_n;
                    r.indent(3);
                    int reportedDamage = damage / 2;
                    if ((damage % 2) > 0) {
                        r.add(String.valueOf(reportedDamage) + ".5");
                    } else {
                        r.add(reportedDamage);
                    }

                    vDesc.addElement(r);
                }
                if (armorThreshold >= damage) {

                    if (hardenedArmor
                        && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING)
                        && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING_MISSILE)
                        && (hit.getGeneralDamageType() != HitData.DAMAGE_IGNORES_DMG_REDUCTION)) {
                        armorThreshold -= damage;
                        te.setHardenedArmorDamaged(hit,
                                                   (armorThreshold % 2) > 0);
                        te.setArmor(
                                (armorThreshold / 2) + (armorThreshold % 2),
                                hit);
                    } else {
                        te.setArmor(te.getArmor(hit) - damage, hit);
                    }

                    if ((te instanceof Mech) && (damage > 0)) {
                        ((Mech) te).setArmorDamagedThisTurn(hit.getLocation(), true);
                    }

                    if (hardenedArmor) {
                        critBonus -= 2;
                    }

                    if (tmpDamageHold >= 0) {
                        te.damageThisPhase += tmpDamageHold;
                    } else {
                        te.damageThisPhase += damage;
                    }
                    damage = 0;
                    if (!te.isHardenedArmorDamaged(hit)) {
                        r = new Report(6085);
                    } else {
                        r = new Report(6086);
                    }

                    r.subject = te_n;
                    r.indent(3);
                    r.add(te.getArmor(hit));
                    vDesc.addElement(r);

                    if ((te instanceof TeleMissile)
                        && (te.getArmor(hit) == damage)) {
                        vDesc.addAll(destroyEntity(te, "damage", false));
                    }

                } else {
                    
                    int absorbed = Math.max(te.getArmor(hit), 0);
                    if (hardenedArmor
                        && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING)
                        && (hit.getGeneralDamageType() != HitData.DAMAGE_ARMOR_PIERCING_MISSILE)) {
                        absorbed = (absorbed * 2)
                                   - ((te.isHardenedArmorDamaged(hit)) ? 1 : 0);
                    }
                    if (reflectiveArmor
                        && (hit.getGeneralDamageType() == HitData.DAMAGE_PHYSICAL)
                        && !isBattleArmor) {
                        absorbed = (int) Math.round(Math.ceil(absorbed / 2));
                        damage = tmpDamageHold;
                        tmpDamageHold = 0;
                    }
                    te.setArmor(IArmorState.ARMOR_DESTROYED, hit);
                    if (tmpDamageHold >= 0) {
                        te.damageThisPhase += 2 * absorbed;
                    } else {
                        te.damageThisPhase += absorbed;
                    }
                    damage -= absorbed;
                    r = new Report(6090);
                    r.subject = te_n;
                    r.indent(3);
                    vDesc.addElement(r);
                    if (te instanceof GunEmplacement) {
                        
                        te.destroyLocation(hit.getLocation());
                        r = new Report(6115);
                        r.subject = te_n;
                        vDesc.addElement(r);

                        if (te.getTransferLocation(hit).getLocation() == Entity.LOC_DESTROYED) {
                            vDesc.addAll(destroyEntity(te, "damage", false));
                        }
                    }
                }

                if (te.hasBARArmor(hit.getLocation())) {
                    if (origDamage > te.getBARRating(hit.getLocation())) {
                        if (te.hasArmoredChassis()) {
                            
                            vDesc.addAll(criticalEntity(te, hit.getLocation(),
                                                        hit.isRear(), -1 + critBonus, damage_orig));
                        } else {
                            vDesc.addAll(criticalEntity(te, hit.getLocation(),
                                                        hit.isRear(), critBonus, damage_orig));
                        }
                    }
                }

                if ((tmpDamageHold > 0) && isPlatoon) {
                    damage = tmpDamageHold;
                }
            }

            if ((te instanceof Tank)
                    && ((te.getArmor(hit) < 1) || damageIS)
                    && game.getOptions().booleanOption("vehicles_threshold")
                    && !((te instanceof VTOL)
                            || (te instanceof GunEmplacement))) {
                ((Tank) te).setOverThresh(true);
            }

            if (damage > 0) {

                if (te instanceof Aero) {
                    Aero a = (Aero) te;

                    if (game.getOptions().booleanOption(
                            "stratops_over_penetrate")) {
                        int opRoll = Compute.d6(1);
                        if ((((te instanceof Jumpship) || (te instanceof SpaceStation))
                             && !(te instanceof Warship) && (opRoll > 3))
                            || ((te instanceof Dropship) && (opRoll > 4))
                            || ((te instanceof Warship)
                                && (a.get0SI() <= 30) && (opRoll > 5))) {
                            
                            r = new Report(9090);
                            r.subject = te_n;
                            r.newlines = 0;
                            vDesc.addElement(r);
                            int new_loc = a.getOppositeLocation(hit
                                                                        .getLocation());
                            damage = Math.min(damage, te.getArmor(new_loc));
                            
                            damage = Math.max(damage, 0);
                            r = new Report(6065);
                            r.subject = te_n;
                            r.indent(2);
                            r.newlines = 0;
                            r.addDesc(te);
                            r.add(damage);
                            r.add(te.getLocationAbbr(new_loc));
                            vDesc.addElement(r);
                            te.setArmor(te.getArmor(new_loc) - damage, new_loc);
                            if ((te instanceof Warship)
                                || (te instanceof Dropship)) {
                                damage = 2;
                            } else {
                                damage = 0;
                            }
                        }
                    }

                    if (!ammoExplosion && !nukeS2S
                        && !game.getOptions().booleanOption("aero_sanity")) {
                        damage /= 2;
                    }

                    if (damage > 0) {
                        critSI = true;
                    }

                    a.setSI(a.getSI() - damage);
                    te.damageThisPhase += damage;
                    
                    r = new Report(1210);
                    r.subject = te_n;
                    r.newlines = 1;
                    if (!ammoExplosion) {
                        r.messageId = 9005;
                    } else {
                        r.messageId = 9006;
                    }
                    r.add(damage);
                    r.add(Math.max(a.getSI(), 0));
                    vDesc.addElement(r);
                    damage = 0;
                    
                    if (a.getSI() <= 0) {
                        vDesc.addAll(destroyEntity(te,
                                "Structural Integrity Collapse"));
                        a.setSI(0);
                        if (hit.getAttackerId() != Entity.NONE) {
                            creditKill(a, game.getEntity(hit.getAttackerId()));
                        }
                    }
                    checkAeroCrits(vDesc, a, hit, damage_orig, critThresh,
                                   critSI, ammoExplosion, nukeS2S);
                    return vDesc;
                }

                
                if (ammoExplosion && te.hasCASEII(hit.getLocation())) {
                    
                    damage--;
                    
                    r = new Report(6126);
                    r.subject = te_n;
                    r.add(damage);
                    r.indent(3);
                    vDesc.addElement(r);
                    int loc = hit.getLocation();
                    if ((te instanceof Mech)
                        && ((loc == Mech.LOC_HEAD)
                            || ((Mech) te).isArm(loc) || ((Mech) te)
                            .locationIsLeg(loc))) {
                        int half = (int) Math.ceil(((Mech) te).getOArmor(loc,
                                                                         false) / 2);
                        if (damage > half) {
                            damage = half;
                        }
                        if (damage >= te.getArmor(loc, false)) {
                            te.setArmor(IArmorState.ARMOR_DESTROYED, loc, false);
                        } else {
                            te.setArmor(te.getArmor(loc, false) - damage, loc,
                                        false);
                        }
                    } else {
                        if (damage >= te.getArmor(loc, true)) {
                            te.setArmor(IArmorState.ARMOR_DESTROYED, loc, true);
                        } else {
                            te.setArmor(te.getArmor(loc, true) - damage, loc,
                                        true);
                        }
                    }

                    if (te.getInternal(hit) > 0) {
                        
                        damage = 1;
                    } else {
                        damage = 0;
                    }

                    te.damageThisPhase += damage;

                    int roll = Compute.d6(2);
                    r = new Report(6127);
                    r.subject = te.getId();
                    r.add(roll);
                    vDesc.add(r);
                    if (roll >= 8) {
                        hit.setEffect(HitData.EFFECT_NO_CRITICALS);
                    }
                }
               
                if (ammoExplosion && (te instanceof Tank)
                    && te.locationHasCase(Tank.LOC_BODY)) {
                    te.damageThisPhase += damage;
                    r = new Report(6124);
                    r.subject = te_n;
                    r.indent(2);
                    r.add(damage);
                    vDesc.add(r);
                    if (damage > te
                            .getArmor(te instanceof SuperHeavyTank ? SuperHeavyTank.LOC_REAR
                                                                   : te instanceof LargeSupportTank ?
                                                                     LargeSupportTank.LOC_REAR
                                                                                                    : Tank.LOC_REAR)) {
                        te.setArmor(
                                IArmorState.ARMOR_DESTROYED,
                                te instanceof SuperHeavyTank ? SuperHeavyTank.LOC_REAR
                                                             : te instanceof LargeSupportTank ? LargeSupportTank
                                        .LOC_REAR
                                                                                              : Tank.LOC_REAR);
                        r = new Report(6090);
                    } else {
                        te.setArmor(
                                te.getArmor(te instanceof SuperHeavyTank ? SuperHeavyTank.LOC_REAR
                                                                         : te instanceof LargeSupportTank ?
                                                                           LargeSupportTank.LOC_REAR
                                                                                                          : Tank
                                                                                   .LOC_REAR)
                                - damage,
                                te instanceof SuperHeavyTank ? SuperHeavyTank.LOC_REAR
                                                             : te instanceof LargeSupportTank ? LargeSupportTank
                                        .LOC_REAR
                                                                                              : Tank.LOC_REAR);
                        r = new Report(6085);
                        r.add(te.getArmor(te instanceof SuperHeavyTank ? SuperHeavyTank.LOC_REAR
                                                                       : te instanceof LargeSupportTank ?
                                                                         LargeSupportTank.LOC_REAR
                                                                                                        : Tank
                                                                                 .LOC_REAR));
                    }
                    r.subject = te_n;
                    r.indent(3);
                    vDesc.add(r);
                    damage = 0;
                    int critIndex;
                    if (((Tank) te).isCommanderHit()
                        && ((Tank) te).isDriverHit()) {
                        critIndex = Tank.CRIT_CREW_KILLED;
                    } else {
                        critIndex = Tank.CRIT_CREW_STUNNED;
                    }
                    vDesc.addAll(applyCriticalHit(te, Entity.NONE,
                                                  new CriticalSlot(0, critIndex), true, 0, false));
                }

                if (te.getInternal(hit) > 0) {

                    int tmpDamageHold = -1;
                    if ((te instanceof Mech)
                        && ((Mech) te).hasCompositeStructure()) {
                        tmpDamageHold = damage;
                        damage *= 2;
                        r = new Report(6091);
                        r.subject = te_n;
                        r.indent(3);
                        vDesc.add(r);
                    }
                    if ((te instanceof Mech)
                        && ((Mech) te).hasReinforcedStructure()) {
                        tmpDamageHold = damage;
                        damage /= 2;
                        damage += tmpDamageHold % 2;
                        r = new Report(6092);
                        r.subject = te_n;
                        r.indent(3);
                        vDesc.add(r);
                    }
                    if ((te.getInternal(hit) > damage) && (damage > 0)) {
                        
                        te.setInternal(te.getInternal(hit) - damage, hit);
                        
                        if (!isPlatoon && !isBattleArmor) {
                            crits++;
                        }
                        tookInternalDamage = true;
                        
                        te.damageThisPhase += (tmpDamageHold > -1) ?
                                tmpDamageHold : damage;
                        damage = 0;
                        r = new Report(6100);
                        r.subject = te_n;
                        r.indent(3);
                        
                        if (isPlatoon) {
                            r.messageId = 6095;
                        }
                        r.add(te.getInternal(hit));
                        vDesc.addElement(r);
                    } else if (damage > 0) {
                        
                        if (!isPlatoon && !isBattleArmor) {
                            crits++;
                        }
                        
                        int absorbed = Math.max(te.getInternal(hit), 0);

                        if (te instanceof Protomech) {
                            int hits = Protomech.POSSIBLE_PILOT_DAMAGE[hit
                                    .getLocation()]
                                       - ((Protomech) te).getPilotDamageTaken(hit
                                                                                      .getLocation());
                            if (hits > 0) {
                                vDesc.addAll(damageCrew(te, hits));
                                ((Protomech) te).setPilotDamageTaken(hit
                                                                             .getLocation(),
                                                                     Protomech.POSSIBLE_PILOT_DAMAGE[hit
                                                                             .getLocation()]);
                            }
                        }

                        r = new Report(1210);
                        r.subject = te_n;
                        if (isPlatoon) {
                           
                            if (((Infantry) te).isSquad()) {
                                r.messageId = 6106; 
                            } else {
                                r.messageId = 6105; 
                            }
                        } else if (isBattleArmor) {
                            r.messageId = 6110;
                        } else {
                            r.messageId = 6115;
                        }
                        r.indent(3);
                        vDesc.addElement(r);

                        if ((te instanceof Mech)
                            && (((hit.getLocation() == Mech.LOC_RT) && (te
                                                                                .getInternal(Mech.LOC_RARM) > 0)) ||
                                ((hit
                                          .getLocation() == Mech.LOC_LT) && (te
                                                                                     .getInternal(Mech.LOC_LARM) > 0)
                                ))) {
                            int blownOffLocation = -1; 
                            if (hit.getLocation() == Mech.LOC_RT) {
                                blownOffLocation = Mech.LOC_RARM;
                            } else {
                                blownOffLocation = Mech.LOC_LARM;
                            }
                            r = new Report(6120);
                            r.subject = te_n;
                            r.add(te.getLocationName(blownOffLocation));
                            vDesc.addElement(r);
                            IHex h = game.getBoard().getHex(te.getPosition());
                            if (te instanceof BipedMech) {
                                if (!h.containsTerrain(Terrains.ARMS)) {
                                    h.addTerrain(Terrains.getTerrainFactory()
                                                         .createTerrain(Terrains.ARMS, 1));
                                } else {
                                    h.addTerrain(Terrains
                                                         .getTerrainFactory()
                                                         .createTerrain(
                                                                 Terrains.ARMS,
                                                                 h.terrainLevel(Terrains.ARMS) + 1));
                                }
                            } else if (!h.containsTerrain(Terrains.LEGS)) {
                                h.addTerrain(Terrains.getTerrainFactory()
                                                     .createTerrain(Terrains.LEGS, 1));
                            } else {
                                h.addTerrain(Terrains
                                                     .getTerrainFactory()
                                                     .createTerrain(
                                                             Terrains.LEGS,
                                                             h.terrainLevel(Terrains.LEGS) + 1));
                            }
                            sendChangedHex(te.getPosition());
                        }

                        if ((te instanceof Mech) || (te instanceof Tank)) {
                            Entity passenger = te.getExteriorUnitAt(
                                    hit.getLocation(), hit.isRear());
                            if ((null != passenger) && !passenger.isDoomed()) {
                                HitData passHit = passenger
                                        .getTrooperAtLocation(hit, te);
                                
                                passHit.setEffect(HitData.EFFECT_CRITICAL);
                                if (passenger.getInternal(passHit) > 0) {
                                    vDesc.addAll(damageEntity(passenger,
                                                              passHit, damage));
                                }
                                passHit = new HitData(hit.getLocation(),
                                                      !hit.isRear());
                                passHit = passenger.getTrooperAtLocation(
                                        passHit, te);
                                
                                passHit.setEffect(HitData.EFFECT_CRITICAL);
                                if (passenger.getInternal(passHit) > 0) {
                                    vDesc.addAll(damageEntity(passenger,
                                                              passHit, damage));
                                }
                            }
                        }

                        if (te instanceof BattleArmor) {
                            int infernos = 0;
                            for (Mounted m : te.getEquipment()) {
                                if (m.getType() instanceof AmmoType) {
                                    AmmoType at = (AmmoType) m.getType();
                                    if (((at.getAmmoType() == AmmoType.T_SRM) || (at
                                                                                          .getAmmoType() == AmmoType
                                                                                          .T_MML))
                                        && (at.getMunitionType() == AmmoType.M_INFERNO)) {
                                        infernos += at.getRackSize()
                                                    * m.getHittableShotsLeft();
                                    }
                                } else if (m.getType().hasFlag(
                                        MiscType.F_FIRE_RESISTANT)) {
                                    
                                    infernos = 0;
                                    break;
                                }
                            }
                            if (infernos > 0) {
                                int roll = Compute.d6(2);
                                r = new Report(6680);
                                r.add(roll);
                                vDesc.add(r);
                                if (roll >= 8) {
                                    Coords c = te.getPosition();
                                    if (c == null) {
                                        Entity transport = game.getEntity(te
                                                                                  .getTransportId());
                                        if (transport != null) {
                                            c = transport.getPosition();
                                        }
                                        vPhaseReport
                                                .addAll(deliverInfernoMissiles(
                                                        te, te, infernos));
                                    }
                                    if (c != null) {
                                        vPhaseReport
                                                .addAll(deliverInfernoMissiles(
                                                        te,
                                                        new HexTarget(
                                                                c,
                                                                game.getBoard(),
                                                                Targetable.TYPE_HEX_ARTILLERY),
                                                        infernos));
                                    }
                                }
                            }
                        }

                        te.setInternal(0, hit);
                        te.damageThisPhase += absorbed;
                        damage -= absorbed;

                        if (tmpDamageHold > 0) {
                            if (((Mech) te).hasCompositeStructure()) {
                                
                                
                                damage /= 2;
                            } else if (((Mech) te).hasReinforcedStructure()) {
                                damage *= 2;
                                damage -= tmpDamageHold % 2;
                            }
                        }
                    }
                }
                if (te.getInternal(hit) <= 0) {
                    
                    
                    nextHit = te.getTransferLocation(hit);
                    if (nextHit.getLocation() == Entity.LOC_DESTROYED) {
                        if (te instanceof Mech) {
                            
                            
                            te.engineHitsThisPhase += te.getNumberOfCriticals(
                                    CriticalSlot.TYPE_SYSTEM,
                                    Mech.SYSTEM_ENGINE, hit.getLocation());
                            
                            
                            te.engineHitsThisPhase -= te.getHitCriticals(
                                    CriticalSlot.TYPE_SYSTEM,
                                    Mech.SYSTEM_ENGINE, hit.getLocation());
                        }

                        boolean engineExploded = checkEngineExplosion(te,
                                                                      vDesc, te.engineHitsThisPhase);

                        if (!engineExploded) {
                            
                            vDesc.addAll(destroyEntity(
                                    te,
                                    "damage",
                                    !ammoExplosion,
                                    !((ammoExplosion || areaSatArty) && ((te instanceof Tank) || ((te instanceof
                                            Mech) && (hit
                                                              .getLocation() == Mech.LOC_CT))))));
                            
                            if ((te instanceof Mech)
                                && (hit.getLocation() == Mech.LOC_HEAD)
                                && !te.getCrew().isDead()
                                && !te.getCrew().isDoomed()
                                && game.getOptions()
                                       .booleanOption(
                                               "tacops_skin_of_the_teeth_ejection")) {
                                Mech mech = (Mech) te;
                                if (mech.isAutoEject()
                                    && (!game.getOptions().booleanOption(
                                        "conditional_ejection") || (game
                                                                            .getOptions().booleanOption(
                                                "conditional_ejection") && mech
                                                                            .isCondEjectHeadshot()))) {
                                    autoEject = true;
                                    vDesc.addAll(ejectEntity(te, true, true));
                                }
                            }

                            if ((te instanceof Mech)
                                && (hit.getLocation() == Mech.LOC_CT)
                                && !te.getCrew().isDead()
                                && !te.getCrew().isDoomed()) {
                                Mech mech = (Mech) te;
                                if (mech.isAutoEject()
                                    && game.getOptions().booleanOption(
                                        "conditional_ejection")
                                    && mech.isCondEjectCTDest()) {
                                    if (mech.getCrew().getHits() < 5) {
                                        Report.addNewline(vDesc);
                                        mech.setDoomed(false);
                                        
                                        mech.setDoomed(true);
                                    }
                                    autoEject = true;
                                    vDesc.addAll(ejectEntity(te, true));
                                }
                            }

                            if ((hit.getLocation() == Mech.LOC_HEAD)
                                || ((hit.getLocation() == Mech.LOC_CT) && ((ammoExplosion && !autoEject) ||
                                                                           areaSatArty))) {
                                te.getCrew().setDoomed(true);
                            }
                            if (game.getOptions().booleanOption(
                                    "auto_abandon_unit")) {
                                vDesc.addAll(abandonEntity(te));
                            }
                        }

                        damage = 0;
                    } else if (nextHit.getLocation() == Entity.LOC_NONE) {
                        
                        damage = 0;
                    } else if (ammoExplosion
                               && te.locationHasCase(hit.getLocation())) {
                        
                        r = new Report(6125);
                        r.subject = te_n;
                        r.add(damage);
                        r.indent(3);
                        vDesc.addElement(r);

                        damage = 0;
                    } else if (damage > 0) {
                        
                        r = new Report(6130);
                        r.subject = te_n;
                        r.indent(2);
                        r.add(damage);
                        r.add(te.getLocationAbbr(nextHit));
                        vDesc.addElement(r);

                        for (Mounted m : te.getWeaponList()) {
                            if (m.isSplit()) {
                                if ((m.getLocation() == hit.getLocation())
                                    || (m.getLocation() == nextHit
                                        .getLocation())) {
                                    te.setWeaponHit(m);
                                }
                            }
                        }
                        
                        if ((bFrag == DamageType.NAIL_RIVET)
                            && (te.getArmor(nextHit.getLocation()) > 0)
                            && (te.getBARRating(nextHit.getLocation()) >= 5)) {
                            damage = 0;
                            r = new Report(6065);
                            r.subject = te_n;
                            r.indent(2);
                            vDesc.add(r);
                        }
                    }
                }
            } else if (hit.getSpecCritMod() < 0) {
                
                if (!hardenedArmor && !ferroLamellorArmor && !reactiveArmor) {
                    specCrits++;
                }
            }
            
            vDesc.addAll(breachCheck(te, hit.getLocation(), null, underWater));

            
            if ((hit.getEffect() & HitData.EFFECT_VEHICLE_MOVE_DAMAGED) == HitData.EFFECT_VEHICLE_MOVE_DAMAGED) {
                vDesc.addAll(vehicleMotiveDamage((Tank) te, hit.getMotiveMod()));
            }
            
            if ((te.getInternal(hit) != IArmorState.ARMOR_DESTROYED)
                && ((hit.getEffect() & HitData.EFFECT_NO_CRITICALS) != HitData.EFFECT_NO_CRITICALS)) {
                for (int i = 0; i < crits; i++) {
                    vDesc.addAll(criticalEntity(te, hit.getLocation(),
                                                hit.isRear(), hit.glancingMod() + critBonus,
                                                damage_orig));
                }
                crits = 0;

                for (int i = 0; i < specCrits; i++) {
                    
                    int critMod = te.hasBARArmor(hit.getLocation()) ? 2 : 0;
                    critMod += (reflectiveArmor && !isBattleArmor) ? 2 : 0; 
                    
                    critMod += impactArmor ? 1 : 0;
                    
                    if (!hardenedArmor) {
                        
                        critMod += hit.getSpecCritMod();
                        critMod += hit.glancingMod();
                    }
                    vDesc.addAll(criticalEntity(te, hit.getLocation(),
                                                hit.isRear(), critMod + critBonus, damage_orig));
                }
                specCrits = 0;
            }

            if (te instanceof Aero) {
                checkAeroCrits(vDesc, (Aero) te, hit, damage_orig, critThresh,
                               critSI, ammoExplosion, nukeS2S);
            }

            if (isHeadHit
                && !te.getCrew().getOptions().booleanOption("dermal_armor")) {
                Report.addNewline(vDesc);
                vDesc.addAll(damageCrew(te, 1));
            }

            if (!(te instanceof Aero) && (te.getInternal(hit) <= 0)) {
                te.destroyLocation(hit.getLocation());

                if ((te instanceof Mech)
                    && ((hit.getLocation() == Mech.LOC_RT) || (hit
                                                                       .getLocation() == Mech.LOC_LT))) {

                    int numEngineHits = te.getEngineHits();
                    boolean engineExploded = checkEngineExplosion(te, vDesc,
                                                                  numEngineHits);

                    int hitsToDestroy = 3;
                    if ((te instanceof Mech)
                        && ((Mech) te).isSuperHeavy()
                        && (te.getEngine().getEngineType() == Engine.COMPACT_ENGINE)) {
                        hitsToDestroy = 2;
                    }

                    if (!engineExploded && (numEngineHits >= hitsToDestroy)) {
                        
                        vDesc.addAll(destroyEntity(te, "engine destruction"));
                        if (game.getOptions()
                                .booleanOption("auto_abandon_unit")) {
                            vDesc.addAll(abandonEntity(te));
                        }
                        te.setSelfDestructing(false);
                        te.setSelfDestructInitiated(false);
                    }
                }
            }

            if (damage > 0) {
                hit = nextHit;
            }
            if (damageIS) {
                wasDamageIS = true;
                damageIS = false;
            }
        }
        
        if (tookInternalDamage
            && ((te instanceof Mech) || (te instanceof Protomech))
            && te.hasActiveEiCockpit()) {
            Report.addNewline(vDesc);
            int roll = Compute.d6(2);
            r = new Report(5075);
            r.subject = te.getId();
            r.addDesc(te);
            r.add(7);
            r.add(roll);
            r.choose(roll >= 7);
            r.indent(2);
            vDesc.add(r);
            if (roll < 7) {
                vDesc.addAll(damageCrew(te, 1));
            }
        }

        if (tookInternalDamage
            && te.getCrew().getOptions().booleanOption("vdni")
            && !te.getCrew().getOptions().booleanOption("bvdni")
            && !te.getCrew().getOptions().booleanOption("pain_shunt")) {
            Report.addNewline(vDesc);
            int roll = Compute.d6(2);
            r = new Report(3580);
            r.subject = te.getId();
            r.addDesc(te);
            r.add(7);
            r.add(roll);
            r.choose(roll >= 8);
            r.indent(2);
            vDesc.add(r);
            if (roll >= 8) {
                vDesc.addAll(damageCrew(te, 1));
            }
        }

       
        if (ammoExplosion
            && game.getOptions().booleanOption("tacops_ammunition")
            
            && !(te.locationHasCase(hit.getLocation()) || te.hasCASEII(hit
                                                                               .getLocation()))
            
            && !(((te instanceof Tank) || (te instanceof Aero)) && te
                .hasCase()) && (te.isDestroyed() || te.isDoomed())
            && (damage_orig > 0) && ((damage_orig / 10) > 0)) {
            Report.addNewline(vDesc);
            r = new Report(5068, Report.PUBLIC);
            r.subject = te.getId();
            r.addDesc(te);
            r.indent(2);
            vDesc.add(r);
            Report.addNewline(vDesc);
            r = new Report(5400, Report.PUBLIC);
            r.subject = te.getId();
            r.indent(2);
            vDesc.add(r);
            int[] damages = {(int) Math.floor(damage_orig / 10),
                             (int) Math.floor(damage_orig / 20)};
            doExplosion(damages, false, te.getPosition(), true, vDesc, null, 5,
                        te.getId());
            Report.addNewline(vDesc);
            r = new Report(5410, Report.PUBLIC);
            r.subject = te.getId();
            r.indent(2);
            vDesc.add(r);
        }

        if (wasDamageIS) {
            Report.addNewline(vDesc);
        }
        return vDesc;
    }

    private boolean checkEngineExplosion(Entity en, Vector<Report> vDesc,
                                         int hits) {
        if (!(en instanceof Mech) && !(en instanceof QuadMech)
            && !(en instanceof BipedMech) && !(en instanceof Aero)
            && !(en instanceof Tank)) {
            return false;
        }
        
        if (en.isDestroyed() || (en.engineHitsThisPhase <= 0)) {
            return false;
        }
        int explosionBTH = 10;
        int hitsPerRound = 4;
        Engine engine = null;

        if (en instanceof Mech) {
            engine = ((Mech) en).getEngine();
        } else if (en instanceof Tank) {
            explosionBTH = 12;
            hitsPerRound = 1;
            engine = ((Tank) en).getEngine();
        } else {
            engine = ((Aero) en).getEngine();
            explosionBTH = 12;
            hitsPerRound = 1;
        }

        if (en.rolledForEngineExplosion || !(en instanceof Mech)) {
            return false;
        }
        
        if (engine.isFusion()
            && (!game.getOptions()
                     .booleanOption("tacops_engine_explosions") || (en.engineHitsThisPhase < hitsPerRound))) {
            return false;
        }
        if (!engine.isFusion()) {
            switch (hits) {
                case 0:
                    return false;
                case 1:
                    explosionBTH = 10;
                    break;
                case 2:
                    explosionBTH = 7;
                    break;
                case 3:
                default:
                    explosionBTH = 4;
                    break;
            }
        }
        int explosionRoll = Compute.d6(2);
        boolean didExplode = explosionRoll >= explosionBTH;

        Report r;
        r = new Report(6150);
        r.subject = en.getId();
        r.indent(2);
        r.addDesc(en);
        r.add(en.engineHitsThisPhase);
        vDesc.addElement(r);
        r = new Report(6155);
        r.subject = en.getId();
        r.indent(2);
        r.add(explosionBTH);
        r.add(explosionRoll);
        vDesc.addElement(r);

        if (!didExplode) {
            
            if (engine.isFusion()) {
                en.rolledForEngineExplosion = true;
            }
            
            r = new Report(6160);
            r.subject = en.getId();
            r.indent(2);
            vDesc.addElement(r);
        } else {
            en.rolledForEngineExplosion = true;
            r = new Report(6165, Report.PUBLIC);
            r.subject = en.getId();
            r.indent(2);
            vDesc.addElement(r);
            vDesc.addAll(destroyEntity(en, "engine explosion", false, false));
            
            en.getCrew().setDoomed(true);

            if (en instanceof Mech) {
                en.destroyLocation(Mech.LOC_CT);
            }

            if (engine.isFusion()) {
                int engineRating = en.getEngine().getRating();
                Report.addNewline(vDesc);
                r = new Report(5400, Report.PUBLIC);
                r.subject = en.getId();
                r.indent(2);
                vDesc.add(r);

                if (en instanceof Mech) {
                    Mech mech = (Mech) en;
                    if (mech.isAutoEject()
                        && (!game.getOptions().booleanOption(
                            "conditional_ejection") || (game
                                                                .getOptions().booleanOption(
                                    "conditional_ejection") && mech
                                                                .isCondEjectEngine()))) {
                        vDesc.addAll(ejectEntity(en, true));
                    }
                }

                doFusionEngineExplosion(engineRating, en.getPosition(), vDesc,
                                        null);
                Report.addNewline(vDesc);
                r = new Report(5410, Report.PUBLIC);
                r.subject = en.getId();
                r.indent(2);
                vDesc.add(r);
            }
        }

        return didExplode;
    }

    public void doFusionEngineExplosion(int engineRating, Coords position,
                                        Vector<Report> vDesc, Vector<Integer> vUnits) {
        int[] myDamages = {engineRating, (engineRating / 10),
                           (engineRating / 20), (engineRating / 40)};
        doExplosion(myDamages, true, position, false, vDesc, vUnits, 5, -1);
    }

    public void doExplosion(int damage, int degredation,
                            boolean autoDestroyInSameHex, Coords position,
                            boolean allowShelter, Vector<Report> vDesc, Vector<Integer> vUnits,
                            int excludedUnitId) {

        if (degredation < 1) {
            return;
        }

        int[] myDamages = new int[damage / degredation];

        if (myDamages.length < 1) {
            return;
        }

        myDamages[0] = damage;
        for (int x = 1; x < myDamages.length; x++) {
            myDamages[x] = myDamages[x - 1] - degredation;
        }
        doExplosion(myDamages, autoDestroyInSameHex, position, allowShelter,
                    vDesc, vUnits, 5, excludedUnitId);
    }

    public void doExplosion(int[] damages, boolean autoDestroyInSameHex,
                            Coords position, boolean allowShelter, Vector<Report> vDesc,
                            Vector<Integer> vUnits, int clusterAmt, int excludedUnitId) {
        if (vDesc == null) {
            vDesc = new Vector<Report>();
        }

        if (vUnits == null) {
            vUnits = new Vector<Integer>();
        }

        Report r;
        HashSet<Entity> entitiesHit = new HashSet<Entity>();

        Enumeration<Building> bldgs = game.getBoard().getBuildings();
        while (bldgs.hasMoreElements()) {
            final Building bldg = bldgs.nextElement();

            Enumeration<Coords> hexes = bldg.getCoords();

            while (hexes.hasMoreElements()) {
                final Coords coords = hexes.nextElement();
                int dist = position.distance(coords);
                if (dist < damages.length) {
                    Vector<Report> buildingReport = damageBuilding(bldg,
                                                                   damages[dist], coords);
                    for (Report report : buildingReport) {
                        report.type = Report.PUBLIC;
                    }
                    vDesc.addAll(buildingReport);
                }
            }
        }

        ArrayList<Entity> loaded = new ArrayList<Entity>();
        for (Iterator<Entity> ents = game.getEntities(); ents.hasNext();) {
            Entity entity = ents.next();

            if (entitiesHit.contains(entity)) {
                continue;
            }

            if (entity.getId() == excludedUnitId) {
                continue;
            }

            if (entity.isDestroyed() || !entity.isDeployed()) {
              
                continue;
            }
            if (entity.isAirborne()) {
                continue;
            }

            if ((entity instanceof MechWarrior)
                && !((MechWarrior) entity).hasLanded()) {
             
                continue;
            }

            Coords entityPos = entity.getPosition();
            if (entityPos == null) {
                
                Entity transport = game.getEntity(entity.getTransportId());
                if ((transport != null) && !transport.isAirborne()) {
                    loaded.add(entity);
                }
                continue;
            }
            int range = position.distance(entityPos);

            if (range >= damages.length) {
                
                continue;
            }

            if ((range == 0) && autoDestroyInSameHex) {
                
                vDesc.addAll(destroyEntity(entity, "explosion proximity",
                                           false, false));

                vUnits.add(entity.getId());

                entity.getCrew().setDoomed(true);

                entitiesHit.add(entity);

                continue;
            }

            int damage = damages[range];

            if (allowShelter
                && canShelter(entityPos, position, entity.relHeight())) {
                if (isSheltered()) {
                    r = new Report(6545);
                    r.addDesc(entity);
                    r.subject = entity.getId();
                    vDesc.addElement(r);
                    continue;
                }
                
                r = new Report(6546);
                r.subject = entity.getId();
                r.addDesc(entity);
                vDesc.addElement(r);
            }

            vUnits.add(entity.getId());

            r = new Report(6175);
            r.subject = entity.getId();
            r.indent(2);
            r.addDesc(entity);
            r.add(damage);
            vDesc.addElement(r);

            while (damage > 0) {
                int cluster = Math.min(clusterAmt, damage);
                int table = ToHitData.HIT_NORMAL;
                if (entity instanceof Protomech) {
                    table = ToHitData.HIT_SPECIAL_PROTO;
                }
                HitData hit = entity.rollHitLocation(table,
                                                     Compute.targetSideTable(position, entity));
                vDesc.addAll(damageEntity(entity, hit, cluster, false,
                                          DamageType.IGNORE_PASSENGER, false, true));
                damage -= cluster;
            }
            Report.addNewline(vDesc);
        }

        for (Entity e : loaded) {
            
            final Entity transporter = game.getEntity(e.getTransportId());
            if ((transporter == null)
                || transporter.getExternalUnits().contains(e)) {
                
                final Coords entityPos = (transporter == null ? e.getPosition()
                                                              : transporter.getPosition());
                final int range = position.distance(entityPos);

                if (range >= damages.length) {
                    
                    continue;
                }

                int damage = damages[range];
                if (allowShelter) {
                    final int absHeight = (transporter == null ? e.relHeight()
                                                               : transporter.relHeight());
                    if ((entityPos != null)
                        && canShelter(entityPos, position, absHeight)) {
                        if (isSheltered()) {
                            r = new Report(6545);
                            r.addDesc(e);
                            r.subject = e.getId();
                            vDesc.addElement(r);
                            continue;
                        }
                        
                        r = new Report(6546);
                        r.subject = e.getId();
                        r.addDesc(e);
                        vDesc.addElement(r);
                    }
                }
                
                vUnits.add(e.getId());

                r = new Report(6175);
                r.subject = e.getId();
                r.indent(2);
                r.addDesc(e);
                r.add(damage);
                vDesc.addElement(r);

                while (damage > 0) {
                    int cluster = Math.min(5, damage);
                    int table = ToHitData.HIT_NORMAL;
                    if (e instanceof Protomech) {
                        table = ToHitData.HIT_SPECIAL_PROTO;
                    }
                    HitData hit = e
                            .rollHitLocation(table, ToHitData.SIDE_FRONT);
                    vDesc.addAll(damageEntity(e, hit, cluster, false,
                                              DamageType.IGNORE_PASSENGER, false, true));
                    damage -= cluster;
                }
                Report.addNewline(vDesc);
            }
        }
    }

    public boolean canShelter(Coords entityPosition, Coords position,
                              int entityAbsHeight) {
        
        Coords shelteringCoords = Coords.nextHex(entityPosition, position);
        IHex shelteringHex = game.getBoard().getHex(shelteringCoords);

        
        if (shelteringHex == null) {
            return false;
        }

        int shelterLevel = shelteringHex.floor();
        if (shelteringHex.containsTerrain(Terrains.BUILDING)) {
            shelterLevel = shelteringHex.ceiling();
        }

        entityAbsHeight += game.getBoard().getHex(entityPosition).surface();

        
        if (entityAbsHeight < shelterLevel) {
            return true;
        }

        return false;
    }

    private boolean isSheltered() {
        if (Compute.d6(2) >= 9) {
            return true;
        }
        return false;
    }

    public void addScheduledNuke(int[] nuke) {
        scheduledNukes.add(nuke);
    }

    
    private void resolveScheduledNukes() {
        for (int[] nuke : scheduledNukes) {
            if (nuke.length == 3) {
                doNuclearExplosion(new Coords(nuke[0] - 1, nuke[1] - 1),
                                   nuke[2], vPhaseReport);
            }
            if (nuke.length == 6) {
                doNuclearExplosion(new Coords(nuke[0] - 1, nuke[1] - 1),
                                   nuke[2], nuke[3], nuke[4], nuke[5], vPhaseReport);
            }
        }
        scheduledNukes.clear();
    }

    public void doNuclearExplosion(Coords position, int nukeType,
                                   Vector<Report> vDesc) {
        
        switch (nukeType) {
            case 0:
            case 1:
                doNuclearExplosion(position, 100, 5, 40, 0, vDesc);
                break;
            case 2:
                doNuclearExplosion(position, 1000, 23, 86, 1, vDesc);
                break;
            case 3:
                doNuclearExplosion(position, 10000, 109, 184, 3, vDesc);
                break;
            case 4:
                doNuclearExplosion(position, 100000, 505, 396, 5, vDesc);
                break;
            default:
                
                
                return;
        }
    }

    public void doNuclearExplosion(Coords position, int baseDamage,
                                   int degredation, int secondaryRadius, int craterDepth,
                                   Vector<Report> vDesc) {
        
        if (vDesc == null) {
            vDesc = new Vector<Report>();
        }

        Report r = new Report(1215, Report.PUBLIC);

        r.indent();
        r.add(position.getBoardNum(), true);
        vDesc.add(r);

        int curDepth = craterDepth;
        int range = 0;
        while (range < (2 * craterDepth)) {
            
            Enumeration<Coords> hexSet = game.getBoard().getHexesAtDistance(
                    position, range);

            
            while (hexSet.hasMoreElements()) {
                Coords myHexCoords = hexSet.nextElement();
                IHex myHex = game.getBoard().getHex(myHexCoords);
                
                myHex.setLevel((myHex.getLevel() - curDepth));

                if (myHex.containsTerrain(Terrains.WATER)) {
                    myHex.setLevel(myHex.floor());
                }
                myHex.removeAllTerrains();
                myHex.clearExits();

                sendChangedHex(myHexCoords);
            }

            if ((range > 0) && ((range % 2) == 0)) {
                curDepth--;
            }

            
            range++;
        }

        for (Entity entity : game.getEntitiesVector()) {
            
            if ((entity.getTransportId() != Entity.NONE)
                || (entity.getPosition() == null)) {
                continue;
            }

            if (position.distance(entity.getPosition()) >= range) {
                continue;
            }
            if (entity.isDestroyed()) {
                continue;
            }

            vDesc.addAll(destroyEntity(entity, "nuclear explosion proximity",
                                       false, false));
            
            entity.getCrew().setDoomed(true);
        }

        Vector<Report> tmpV = new Vector<Report>();
        Vector<Integer> blastedUnitsVec = new Vector<Integer>();
        doExplosion(baseDamage, degredation, true, position, true, tmpV,
                    blastedUnitsVec, -1);
        Report.indentAll(tmpV, 2);
        vDesc.addAll(tmpV);

        for (int i : blastedUnitsVec) {
            Entity o = game.getEntity(i);
            if (o instanceof Mech) {
                Mech bm = (Mech) o;
                
                if (!bm.isProne()) {
                    game.addPSR(new PilotingRollData(bm.getId(), 6,
                                                     "hit by nuclear blast"));
                }
            } else if (o instanceof VTOL) {
               
                VTOL vt = (VTOL) o;

                if (vt.getElevation() > 0) {
                    game.addPSR(new PilotingRollData(vt.getId(), 6,
                                                     "hit by nuclear blast"));
                }
            } else if (o instanceof Tank) {
            
            }
        }

        tmpV = null;

        Building tmpB = game.getBoard().getBuildingAt(position);
        if (tmpB != null) {
            r = new Report(2415);
            r.add(tmpB.getName());
            addReport(r);
            tmpB.setCurrentCF(0, position);
        }
        IHex gzHex = game.getBoard().getHex(position);
        if (gzHex.containsTerrain(Terrains.WATER)) {
            gzHex.setLevel(gzHex.floor());
        }
        gzHex.removeAllTerrains();

        boolean damageFlag = true;
        int damageAtRange = baseDamage - (degredation * range);
        if (damageAtRange > 0) {
            for (int x = range; damageFlag; x++) {
                
                Enumeration<Coords> hexSet = game.getBoard()
                                                 .getHexesAtDistance(position, x);

                while (hexSet.hasMoreElements()) {
                    Coords myHexCoords = hexSet.nextElement();
                    IHex myHex = game.getBoard().getHex(myHexCoords);

                    if ((damageAtRange >= 3000)
                        && (myHex.containsTerrain(Terrains.WATER))) {
                        int numCleared = damageAtRange / 3000;
                        int oldLevel = myHex.terrainLevel(Terrains.WATER);
                        myHex.removeTerrain(Terrains.WATER);
                        if (oldLevel > numCleared) {
                            myHex.setLevel(myHex.getLevel()
                                               - numCleared);
                            myHex.addTerrain(new Terrain(Terrains.WATER,
                                                         oldLevel - numCleared));
                        } else {
                            myHex.setLevel(myHex.getLevel() - oldLevel);
                        }
                    }

                    if ((damageAtRange >= 200)
                        && (!myHex.containsTerrain(Terrains.WATER))) {
                        myHex.removeAllTerrains();
                        myHex.clearExits();
                        myHex.addTerrain(new Terrain(Terrains.ROUGH, 1));
                    } else if ((damageAtRange >= 20)
                               && ((myHex.containsTerrain(Terrains.WOODS)) || (myHex
                            .containsTerrain(Terrains.JUNGLE)))) {
                        
                        int numCleared = damageAtRange / 20;
                        int terrainType = (myHex
                                                   .containsTerrain(Terrains.WOODS) ? Terrains.WOODS
                                                                                    : Terrains.JUNGLE);
                        int oldLevel = myHex.terrainLevel(terrainType);
                        myHex.removeTerrain(terrainType);
                        if (oldLevel > numCleared) {
                            myHex.addTerrain(new Terrain(terrainType, oldLevel
                                                                      - numCleared));
                        }
                    }

                    sendChangedHex(myHexCoords);
                }

                damageAtRange = baseDamage - ((degredation * x) + 1);

                if (damageAtRange < 20) {
                    damageFlag = false;
                }
            }
        }

        for (Entity entity : game.getEntitiesVector()) {

            if ((entity.getTransportId() != Entity.NONE)
                || (entity.getPosition() == null)) {
                continue;
            }

            if ((entity.isDoomed()) || (entity.isDestroyed())) {
                continue;
            }

            if (position.distance(entity.getPosition()) > secondaryRadius) {
                continue;
            }

            applySecondaryNuclearEffects(entity, position, vDesc);
        }

        r = new Report(1216, Report.PUBLIC);
        r.indent();
        r.newlines = 2;
        vDesc.add(r);
    }

    public void applySecondaryNuclearEffects(Entity entity, Coords position,
                                             Vector<Report> vDesc) {
        
        if (entity.isDestroyed()) {
            return;
        }

        boolean inHardenedBuilding = (Compute.isInBuilding(game, entity) && (game
                                                                                     .getBoard().getHex(entity.getPosition())
                                                                                     .terrainLevel(Terrains.BUILDING)
                                                                             == 4));

        int roll = Compute.d6(2);

        Report r = new Report(6555);
        r.subject = entity.getId();
        r.add(entity.getDisplayName());
        r.add(roll);

        if (inHardenedBuilding) {
            roll += 2;
            r.add(" + 2 (unit is in hardened building)");
        } else {
            r.add("");
        }

        if (entity.isNuclearHardened()) {
            roll += 2;
            r.add(" + 2 (unit is hardened against EMI)");
        } else {
            r.add("");
        }

        r.indent(2);
        vDesc.add(r);

        if (roll <= 4) {
          
            vDesc.addAll(destroyEntity(entity,
                                       "nuclear explosion secondary effects", false, false));
            
            entity.getCrew().setDoomed(true);
        } else if (roll <= 6) {
            if (entity instanceof BattleArmor) {
                
                BattleArmor myBA = (BattleArmor) entity;
                int numDeaths = (int) (Math.ceil((myBA
                        .getNumberActiverTroopers())) / 2.0);
                for (int x = 0; x < numDeaths; x++) {
                    vDesc.addAll(applyCriticalHit(entity, 0, null, false, 0,
                                                  false));
                }
            } else if (entity instanceof Infantry) {
                
                if (game.getBoard().getHex(entity.getPosition())
                        .containsTerrain(Terrains.BUILDING)) {
                    
                    int damage = (int) (Math.ceil((((Infantry) entity)
                            .getInternal(Infantry.LOC_INFANTRY)) / 2.0));
                    vDesc.addAll(damageEntity(entity, new HitData(
                            Infantry.LOC_INFANTRY), damage, true));
                } else {
                    vDesc.addAll(destroyEntity(entity,
                                               "nuclear explosion secondary effects", false, false));
                    entity.getCrew().setDoomed(true);
                }
            } else if (entity instanceof Tank) {
                
                HitData hd = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                    entity.sideTable(position));
                vDesc.addAll(oneCriticalEntity(entity, hd.getLocation(),
                                               hd.isRear(), 0));
                hd = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                            entity.sideTable(position));
                vDesc.addAll(oneCriticalEntity(entity, hd.getLocation(),
                                               hd.isRear(), 0));

                vDesc.addAll(applyCriticalHit(entity, 0, new CriticalSlot(0,
                                                                          Tank.CRIT_CREW_KILLED), false, 0, false));
            } else if ((entity instanceof Mech)
                       || (entity instanceof Protomech)) {
                
                HitData hd = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                    entity.sideTable(position));
                vDesc.addAll(oneCriticalEntity(entity, hd.getLocation(),
                                               hd.isRear(), 0));
                hd = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                            entity.sideTable(position));
                vDesc.addAll(oneCriticalEntity(entity, hd.getLocation(),
                                               hd.isRear(), 0));

                vDesc.addAll(damageCrew(entity, 4));
            } else {
                
            }
        } else if (roll <= 10) {
            if (entity instanceof BattleArmor) {
                
                BattleArmor myBA = (BattleArmor) entity;
                int numDeaths = (int) (Math.ceil(((myBA
                        .getNumberActiverTroopers())) / 4.0));
                for (int x = 0; x < numDeaths; x++) {
                    vDesc.addAll(applyCriticalHit(entity, 0, null, false, 0,
                                                  false));
                }
            } else if (entity instanceof Infantry) {
                if (game.getBoard().getHex(entity.getPosition())
                        .containsTerrain(Terrains.BUILDING)) {
                    
                    int damage = (int) (Math.ceil((((Infantry) entity)
                            .getInternal(Infantry.LOC_INFANTRY)) / 4.0));
                    vDesc.addAll(damageEntity(entity, new HitData(
                            Infantry.LOC_INFANTRY), damage, true));
                } else {
                    
                    int damage = (int) (Math.ceil((((Infantry) entity)
                            .getInternal(Infantry.LOC_INFANTRY)) / 2.0));
                    vDesc.addAll(damageEntity(entity, new HitData(
                            Infantry.LOC_INFANTRY), damage, true));
                }
            } else if (entity instanceof Tank) {
                
                HitData hd = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                    entity.sideTable(position));
                vDesc.addAll(oneCriticalEntity(entity, hd.getLocation(),
                                               hd.isRear(), 0));

                vDesc.addAll(applyCriticalHit(entity, 0, new CriticalSlot(0,
                                                                          Tank.CRIT_CREW_STUNNED), false, 0, false));
            } else if ((entity instanceof Mech)
                       || (entity instanceof Protomech)) {
                
                HitData hd = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                    entity.sideTable(position));
                vDesc.addAll(oneCriticalEntity(entity, hd.getLocation(),
                                               hd.isRear(), 0));

                vDesc.addAll(damageCrew(entity, 2));
            } else {
                
              
            }
        }
        
        entity.setEMI(true);
    }

    
    public Vector<Report> applyCriticalHit(Entity en, int loc, CriticalSlot cs,
                                           boolean secondaryEffects, int damageCaused, boolean isCapital) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if (en instanceof Tank) {
            Tank t = (Tank) en;
            HitData hit;
            switch (cs.getIndex()) {
                case Tank.CRIT_NONE:
                    
                    r = new Report(6005);
                    r.subject = t.getId();
                    vDesc.add(r);
                    break;
                case Tank.CRIT_AMMO:
                    
                    r = new Report(6610);
                    r.subject = t.getId();
                    vDesc.add(r);
                    int damage = 0;
                    for (Mounted m : t.getAmmo()) {
                        
                        if (m.getLocation() == Entity.LOC_NONE) {
                            continue;
                        }
                        m.setHit(true);
                        int tmp = m.getHittableShotsLeft()
                                  * ((AmmoType) m.getType()).getDamagePerShot()
                                  * ((AmmoType) m.getType()).getRackSize();
                        m.setShotsLeft(0);
                        
                        if (!m.getType().isExplosive(m)) {
                            continue;
                        }
                        damage += tmp;
                        r = new Report(6390);
                        r.subject = t.getId();
                        r.add(m.getName());
                        r.add(tmp);
                        vDesc.add(r);
                    }
                    hit = new HitData(loc);
                    vDesc.addAll(damageEntity(t, hit, damage, true));
                    break;
                case Tank.CRIT_CARGO:
                    
                    r = new Report(6615);
                    r.subject = t.getId();
                    vDesc.add(r);
                    List<Entity> passengers = t.getLoadedUnits();
                    if (passengers.size() > 0) {
                        Entity target = passengers.get(Compute.randomInt(passengers
                                                                                 .size()));
                        hit = target.rollHitLocation(ToHitData.HIT_NORMAL,
                                                     ToHitData.SIDE_FRONT);
                        vDesc.addAll(damageEntity(target, hit, damageCaused));
                    }
                    break;
                case Tank.CRIT_COMMANDER:
                    if (en.getCrew().getOptions().booleanOption("vdni")
                        || en.getCrew().getOptions().booleanOption("bvdni")) {
                        r = new Report(6191);
                        r.subject = t.getId();
                        vDesc.add(r);
                        vDesc.addAll(damageCrew(en, 1));
                    } else {
                        if (en.getCrew().getOptions().booleanOption("pain_shunt")
                            && !t.isCommanderHitPS()) {
                            r = new Report(6606);
                            r.subject = t.getId();
                            vDesc.add(r);
                            t.setCommanderHitPS(true);
                        } else {
                            r = new Report(6605);
                            r.subject = t.getId();
                            vDesc.add(r);
                            t.setCommanderHit(true);
                        }
                    }
                    
                    
                case Tank.CRIT_CREW_STUNNED:
                    if (en.getCrew().getOptions().booleanOption("vdni")
                        || en.getCrew().getOptions().booleanOption("bvdni")) {
                        r = new Report(6191);
                        r.subject = t.getId();
                        vDesc.add(r);
                        vDesc.addAll(damageCrew(en, 1));
                    } else {
                        if (en.getCrew().getOptions().booleanOption("pain_shunt")
                            || en.getCrew().getOptions()
                                 .booleanOption("dermal_armor")) {
                            r = new Report(6186);
                            r.subject = t.getId();
                            vDesc.add(r);
                        } else {
                            t.stunCrew();
                            r = new Report(6185);
                            r.add(t.getStunnedTurns() - 1);
                            r.subject = t.getId();
                            vDesc.add(r);
                        }
                    }
                    break;
                case Tank.CRIT_DRIVER:
                    if (en.getCrew().getOptions().booleanOption("vdni")
                        || en.getCrew().getOptions().booleanOption("bvdni")) {
                        r = new Report(6191);
                        r.subject = t.getId();
                        vDesc.add(r);
                        vDesc.addAll(damageCrew(en, 1));
                    } else {
                        if (en.getCrew().getOptions().booleanOption("pain_shunt")
                            && !t.isDriverHitPS()) {
                            r = new Report(6601);
                            r.subject = t.getId();
                            vDesc.add(r);
                            t.setDriverHitPS(true);
                        } else {
                            r = new Report(6600);
                            r.subject = t.getId();
                            vDesc.add(r);
                            t.setDriverHit(true);
                        }
                    }
                    break;
                case Tank.CRIT_CREW_KILLED:
                    if (en.getCrew().getOptions().booleanOption("vdni")
                        || en.getCrew().getOptions().booleanOption("bvdni")) {
                        r = new Report(6191);
                        r.subject = t.getId();
                        vDesc.add(r);
                        vDesc.addAll(damageCrew(en, 1));
                    } else {
                        if (en.getCrew().getOptions().booleanOption("pain_shunt")
                            && !t.isCrewHitPS()) {
                            r = new Report(6191);
                            r.subject = t.getId();
                            vDesc.add(r);
                            t.setCrewHitPS(true);
                        } else {
                            r = new Report(6190);
                            r.subject = t.getId();
                            vDesc.add(r);
                            t.getCrew().setDoomed(true);
                            if (en.isAirborneVTOLorWIGE()) {
                                vDesc.addAll(crashVTOLorWiGE((Tank)en));
                            }
                        }
                    }
                    break;
                case Tank.CRIT_ENGINE:
                    r = new Report(6210);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.engineHit();
                    t.engineHitsThisPhase++;
                    boolean engineExploded = checkEngineExplosion(t, vDesc, 1);
                    if (engineExploded) {
                        vDesc.addAll(destroyEntity(en, "engine destruction", true,
                                                   true));
                        t.setSelfDestructing(false);
                        t.setSelfDestructInitiated(false);
                    }
                    if (t.isAirborneVTOLorWIGE()
                        && !(t.isDestroyed() || t.isDoomed())) {
                        t.immobilize();
                        vDesc.addAll(forceLandVTOLorWiGE(t));
                    }
                    break;
                case Tank.CRIT_FUEL_TANK:
                    r = new Report(6215);
                    r.subject = t.getId();
                    vDesc.add(r);
                    vDesc.addAll(destroyEntity(t, "fuel explosion", false, false));
                    break;
                case Tank.CRIT_SENSOR:
                    r = new Report(6620);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.setSensorHits(t.getSensorHits() + 1);
                    break;
                case Tank.CRIT_STABILIZER:
                    r = new Report(6625);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.setStabiliserHit(loc);
                    break;
                case Tank.CRIT_TURRET_DESTROYED:
                    r = new Report(6630);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.destroyLocation(t.getLocTurret());
                    vDesc.addAll(destroyEntity(t, "turret blown off", true, true));
                    break;
                case Tank.CRIT_TURRET_JAM:
                    if (t.isTurretEverJammed(loc)) {
                        r = new Report(6640);
                        r.subject = t.getId();
                        vDesc.add(r);
                        t.lockTurret(loc);
                        break;
                    }
                    r = new Report(6635);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.jamTurret(loc);
                    break;
                case Tank.CRIT_TURRET_LOCK:
                    r = new Report(6640);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.lockTurret(loc);
                    break;
                case Tank.CRIT_WEAPON_DESTROYED: {
                    r = new Report(6305);
                    r.subject = t.getId();
                    ArrayList<Mounted> weapons = new ArrayList<Mounted>();
                    for (Mounted weap : t.getWeaponList()) {
                        if ((weap.getLocation() == loc) && !weap.isHit()
                            && !weap.isDestroyed()) {
                            weapons.add(weap);
                        }
                    }
                    
                    Collections.sort(weapons, new WeaponComparatorBV());
                    int roll = Compute.d6();
                    Mounted weapon;
                    if (roll < 4) {
                        
                        
                        weapon = weapons.get(weapons.size() - 1);
                    } else {
                        
                        weapon = weapons.get(0);
                    }
                    r.add(weapon.getName());
                    vDesc.add(r);
                    
                    if (weapon.getType().isExplosive(weapon) && !weapon.isHit()
                        && !weapon.isDestroyed()) {
                        vDesc.addAll(explodeEquipment(t, loc, weapon));
                    }
                    weapon.setHit(true);
                    break;
                }
                case Tank.CRIT_WEAPON_JAM: {
                    r = new Report(6645);
                    r.subject = t.getId();
                    ArrayList<Mounted> weapons = new ArrayList<Mounted>();
                    for (Mounted weap : t.getWeaponList()) {
                        if ((weap.getLocation() == loc) && !weap.isJammed()
                            && !weap.jammedThisPhase() && !weap.isHit()
                            && !weap.isDestroyed()) {
                            weapons.add(weap);
                        }
                    }
                    if (weapons.size() > 0) {
                        Mounted weapon = weapons.get(Compute.randomInt(weapons
                                                                               .size()));
                        weapon.setJammed(true);
                        t.addJammedWeapon(weapon);
                        r.add(weapon.getName());
                        vDesc.add(r);
                    }
                    break;
                }
                case VTOL.CRIT_PILOT:
                    r = new Report(6650);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.setDriverHit(true);
                    PilotingRollData psr = t.getBasePilotingRoll();
                    psr.addModifier(0, "pilot injury");
                    if (!doSkillCheckInPlace(t, psr)) {
                        r = new Report(6675);
                        r.subject = t.getId();
                        r.addDesc(t);
                        vDesc.add(r);
                        boolean crash = true;
                        if (t.canGoDown()) {
                            t.setElevation(t.getElevation() - 1);
                            crash = !t.canGoDown();
                        }
                        if (crash) {
                            vDesc.addAll(crashVTOLorWiGE(t));
                        }
                    }
                    break;
                case VTOL.CRIT_COPILOT:
                    r = new Report(6655);
                    r.subject = t.getId();
                    vDesc.add(r);
                    t.setCommanderHit(true);
                    break;
                case VTOL.CRIT_ROTOR_DAMAGE: {
                    
                    
                    if (!(t.isLocationBad(VTOL.LOC_ROTOR) || t
                            .isLocationDoomed(VTOL.LOC_ROTOR))) {
                        r = new Report(6660);
                        r.subject = t.getId();
                        vDesc.add(r);
                        t.setMotiveDamage(t.getMotiveDamage() + 1);
                        if (t.getMotiveDamage() >= t.getOriginalWalkMP()) {
                            t.immobilize();
                            if (t.isAirborneVTOLorWIGE()
                                
                                
                                && !(t.isDestroyed() || t.isDoomed())) {
                                vDesc.addAll(forceLandVTOLorWiGE(t));
                            }
                        }
                    }
                    break;
                }
                case VTOL.CRIT_ROTOR_DESTROYED:
                    
                    if (!(t.isLocationBad(VTOL.LOC_ROTOR) || t
                            .isLocationDoomed(VTOL.LOC_ROTOR))) {
                        r = new Report(6670);
                        r.subject = t.getId();
                        vDesc.add(r);
                        t.immobilize();
                        vDesc.addAll(crashVTOLorWiGE(t, true));
                    }
                    break;
                case VTOL.CRIT_FLIGHT_STABILIZER:
                    
                    
                    if (!(t.isLocationBad(VTOL.LOC_ROTOR) || t
                            .isLocationDoomed(VTOL.LOC_ROTOR))) {
                        r = new Report(6665);
                        r.subject = t.getId();
                        vDesc.add(r);
                        t.setStabiliserHit(VTOL.LOC_ROTOR);
                    }
                    break;
            }
        } else if (en instanceof Aero) {
            Aero a = (Aero) en;

            Jumpship js = new Jumpship();
            if (en instanceof Jumpship) {
                js = (Jumpship) en;
            } else {
                js = null;
            }

            switch (cs.getIndex()) {
                case Aero.CRIT_NONE:
                    
                    r = new Report(6005);
                    r.subject = a.getId();
                    vDesc.add(r);
                    break;
                case Aero.CRIT_FCS:
                    
                    r = new Report(9105);
                    r.subject = a.getId();
                    vDesc.add(r);
                    a.setFCSHits(a.getFCSHits() + 1);
                    break;
                case Aero.CRIT_SENSOR:
                    
                    r = new Report(6620);
                    r.subject = a.getId();
                    vDesc.add(r);
                    a.setSensorHits(a.getSensorHits() + 1);
                    break;
                case Aero.CRIT_AVIONICS:
                    
                    r = new Report(9110);
                    r.subject = a.getId();
                    vDesc.add(r);
                    a.setAvionicsHits(a.getAvionicsHits() + 1);
                    if (a.isPartOfFighterSquadron()) {
                        game.addControlRoll(new PilotingRollData(
                                a.getTransportId(), 1, "avionics hit"));
                    } else if (a.isCapitalFighter()) {
                        game.addControlRoll(new PilotingRollData(a.getId(), 1,
                                                                 "avionics hit"));
                    } else {
                        game.addControlRoll(new PilotingRollData(a.getId(), 0,
                                                                 "avionics hit"));
                    }
                    break;
                case Aero.CRIT_CONTROL:
                    
                    r = new Report(9115);
                    r.subject = a.getId();
                    vDesc.add(r);
                    if (a.isPartOfFighterSquadron()) {
                        game.addControlRoll(new PilotingRollData(
                                a.getTransportId(), 1, "critical hit"));
                    } else if (a.isCapitalFighter()) {
                        game.addControlRoll(new PilotingRollData(a.getId(), 1,
                                                                 "critical hit"));
                    } else {
                        game.addControlRoll(new PilotingRollData(a.getId(), 0,
                                                                 "critical hit"));
                    }
                    break;
                case Aero.CRIT_FUEL_TANK:
                    
                    r = new Report(9120);
                    r.subject = a.getId();
                    int boomTarget = 9;
                    if (a.hasQuirk("fragile_fuel")) {
                        boomTarget = 7;
                    }
                    if (a.isLargeCraft() && a.isClan()
                        && game.getOptions().booleanOption("stratops_harjel")) {
                        boomTarget = 11;
                    }
                    
                    int fuelroll = Compute.d6(2);
                    if (fuelroll > boomTarget) {
                        r.choose(true);
                        vDesc.add(r);
                        vDesc.addAll(destroyEntity(a, "fuel explosion", false,
                                                   false));
                    } else {
                        r.choose(false);
                        vDesc.add(r);
                    }
                    break;
                case Aero.CRIT_CREW:
                    
                    r = new Report(6650);
                    if (en.getCrew().getOptions().booleanOption("dermal_armor")) {
                        r = new Report(6651);
                        r.subject = a.getId();
                        vDesc.add(r);
                        break;
                    } else if (en.getCrew().getOptions()
                                 .booleanOption("tsm_implant")) {
                        r = new Report(6652);
                        r.subject = a.getId();
                        vDesc.add(r);
                        break;
                    }
                    if ((a instanceof SmallCraft) || (a instanceof Jumpship)) {
                        r = new Report(9197);
                    }
                    if (a.isLargeCraft() && a.isClan()
                        && game.getOptions().booleanOption("stratops_harjel")
                        && (a.getIgnoredCrewHits() < 2)) {
                        a.setIgnoredCrewHits(a.getIgnoredCrewHits() + 1);
                        r = new Report(9198);
                        r.subject = a.getId();
                        vDesc.add(r);
                        break;
                    }
                    r.subject = a.getId();
                    vDesc.add(r);
                    vDesc.addAll(damageCrew(a, 1));
                    
                    if ((a.getCrew().isDead() || a.getCrew().isDoomed())
                        && !a.getCrew().isEjected()) {
                        vDesc.addAll(destroyEntity(a, "pilot death", true, true));
                    }
                    break;
                case Aero.CRIT_GEAR:
                    
                    r = new Report(9125);
                    r.subject = a.getId();
                    vDesc.add(r);
                    a.setGearHit(true);
                    break;
                case Aero.CRIT_BOMB:
                    
                    
                    ArrayList<Mounted> bombs = new ArrayList<Mounted>();
                    for (Mounted bomb : a.getBombs()) {
                        if (bomb.getType().isHittable()
                            && (bomb.getHittableShotsLeft() > 0)) {
                            bombs.add(bomb);
                        }
                    }
                    if (bombs.size() > 0) {
                        Mounted hitbomb = bombs
                                .get(Compute.randomInt(bombs.size()));
                        hitbomb.setShotsLeft(0);
                        hitbomb.setDestroyed(true);
                        r = new Report(9130);
                        r.subject = a.getId();
                        r.add(hitbomb.getDesc());
                        vDesc.add(r);
                        
                        
                        if (a.getTransportId() != Entity.NONE) {
                            Entity e = game.getEntity(a.getTransportId());
                            if (e instanceof FighterSquadron) {
                                ((FighterSquadron) e).computeSquadronBombLoadout();
                            }
                        }
                    } else {
                        r = new Report(9131);
                        r.subject = a.getId();
                        vDesc.add(r);
                    }
                    break;
                case Aero.CRIT_HEATSINK:
                    
                    int sinksLost = 1;
                    if (isCapital) {
                        sinksLost = 10;
                    }
                    r = new Report(9135);
                    r.subject = a.getId();
                    r.add(sinksLost);
                    vDesc.add(r);
                    a.setHeatSinks(Math.max(0, a.getHeatSinks() - sinksLost));
                    break;
                case Aero.CRIT_WEAPON_BROAD:
                    if (a instanceof Warship) {
                        if ((loc == Jumpship.LOC_ALS) || (loc == Jumpship.LOC_FLS)) {
                            loc = Warship.LOC_LBS;
                        } else if ((loc == Jumpship.LOC_ARS)
                                   || (loc == Jumpship.LOC_FRS)) {
                            loc = Warship.LOC_RBS;
                        }
                    }
                case Aero.CRIT_WEAPON:
                    if (a.isCapitalFighter()) {
                        boolean destroyAll = false;
                        if ((loc == Aero.LOC_NOSE) || (loc == Aero.LOC_AFT)) {
                            destroyAll = true;
                        }
                        if (loc == Aero.LOC_WINGS) {
                            if (a.areWingsHit()) {
                                destroyAll = true;
                            } else {
                                a.setWingsHit(true);
                            }
                        }
                        for (Mounted weap : a.getWeaponList()) {
                            if (weap.getLocation() == loc) {
                                if (destroyAll) {
                                    weap.setHit(true);
                                } else {
                                    weap.setNWeapons(weap.getNWeapons() / 2);
                                }
                            }
                        }
                        
                        for (Mounted misc : a.getMisc()) {
                            if (misc.getType().hasFlag(MiscType.F_ECM)
                                || misc.getType().hasFlag(MiscType.F_ANGEL_ECM)
                                || misc.getType().hasFlag(MiscType.F_BAP)) {
                                misc.setHit(true);
                            }
                        }
                        r = new Report(9152);
                        r.subject = a.getId();
                        r.add(a.getLocationName(loc));
                        vDesc.add(r);
                        break;
                    }
                    r = new Report(9150);
                    r.subject = a.getId();
                    ArrayList<Mounted> weapons = new ArrayList<Mounted>();
                    for (Mounted weap : a.getWeaponList()) {
                        if ((weap.getLocation() == loc) && !weap.isDestroyed()
                            && weap.getType().isHittable()) {
                            weapons.add(weap);
                        }
                    }
                    
                    for (Mounted misc : a.getMisc()) {
                        if (misc.getType().isHittable()
                            && (misc.getLocation() == loc)
                            && !misc.isDestroyed()) {
                            weapons.add(misc);
                        }
                    }
                    if (weapons.size() > 0) {
                        Mounted weapon = weapons.get(Compute.randomInt(weapons
                                                                               .size()));
                        
                        if (game.getOptions().booleanOption("ammo_explosions")
                            && !(a instanceof FighterSquadron)
                            && (weapon.getType() instanceof WeaponType)) {
                            
                            WeaponType wtype = (WeaponType) weapon.getType();
                            if (wtype.getAmmoType() != AmmoType.T_NA) {
                                Mounted m = weapon.getLinked();
                                int ammoroll = Compute.d6(2);
                                if (ammoroll >= 10) {
                                    r = new Report(9151);
                                    r.subject = a.getId();
                                    r.add(m.getName());
                                    r.newlines = 0;
                                    vDesc.add(r);
                                    vDesc.addAll(explodeEquipment(a, loc, m));
                                    break;
                                }
                            }
                        }
                        r.add(weapon.getName());
                        vDesc.add(r);
                        
                        if (weapon.getType().isExplosive(weapon) && !weapon.isHit()
                            && !weapon.isDestroyed()) {
                            vDesc.addAll(explodeEquipment(a, loc, weapon));
                        }
                        weapon.setHit(true);
                    } else {
                        r = new Report(9155);
                        r.subject = a.getId();
                        vDesc.add(r);
                    }
                    break;
                case Aero.CRIT_ENGINE:
                    
                    r = new Report(9140);
                    r.subject = a.getId();
                    vDesc.add(r);
                    a.engineHitsThisPhase++;
                    boolean engineExploded = checkEngineExplosion(a, vDesc, 1);
                    a.setEngineHits(a.getEngineHits() + 1);
                    if ((a.getEngineHits() >= a.getMaxEngineHits())
                        || engineExploded) {
                        
                        vDesc.addAll(destroyEntity(a, "engine destruction", true,
                                                   true));
                        a.setSelfDestructing(false);
                        a.setSelfDestructInitiated(false);
                    }
                    break;
                case Aero.CRIT_LEFT_THRUSTER:
                    
                    r = new Report(9160);
                    r.subject = a.getId();
                    vDesc.add(r);
                    a.setLeftThrustHits(a.getLeftThrustHits() + 1);
                    break;
                case Aero.CRIT_RIGHT_THRUSTER:
                    
                    r = new Report(9160);
                    r.subject = a.getId();
                    vDesc.add(r);
                    a.setRightThrustHits(a.getRightThrustHits() + 1);
                    break;
                case Aero.CRIT_CARGO:
                    
                    
                    double percentDestroyed = 0.0;
                    double mult = 2.0;
                    if (a.isLargeCraft() && a.isClan()
                        && game.getOptions().booleanOption("stratops_harjel")) {
                        mult = 4.0;
                    }
                    if (damageCaused > 0) {
                        percentDestroyed = Math.min(
                                damageCaused / (mult * a.getSI()), 1.0);
                    }
                    
                    int roll = Compute.d6(1);
                    if (roll < 4) {
                        
                        r = new Report(9165);
                        r.subject = a.getId();
                        r.add((int) (percentDestroyed * 100));
                        vDesc.add(r);
                    } else {
                        
                        Vector<Entity> passengers = en.getBayLoadedUnits();
                        int unitsDestroyed = (int) Math.ceil(percentDestroyed
                                                             * passengers.size());
                        r = new Report(9166);
                        r.subject = a.getId();
                        r.add(unitsDestroyed);
                        vDesc.add(r);
                        while (unitsDestroyed > 0) {
                            
                            List<Entity> units = en.getLoadedUnits();
                            if (units.size() > 0) {
                                Entity target = units.get(Compute.randomInt(units
                                                                                    .size()));
                                vDesc.addAll(destroyEntity(target, "cargo damage",
                                                           false, false));
                            }
                            unitsDestroyed--;
                        }
                    }
                    break;
                case Aero.CRIT_DOOR:
                    String bayType = en.damageBayDoor();
                    if (!bayType.equals("none")) {
                        r = new Report(9170);
                        r.subject = a.getId();
                        r.add(bayType);
                        vDesc.add(r);
                    } else {
                        r = new Report(9171);
                        r.subject = a.getId();
                        vDesc.add(r);
                    }
                    break;
                case Aero.CRIT_DOCK_COLLAR:
                    
                    if (en instanceof Dropship) {
                        Dropship ds = (Dropship) en;
                        ds.setDamageDockCollar(true);
                        r = new Report(9175);
                        r.subject = a.getId();
                        vDesc.add(r);
                    }
                    if (en instanceof Jumpship) {
                        
                        if (en.damageDockCollar()) {
                            r = new Report(9176);
                            r.subject = a.getId();
                            vDesc.add(r);
                        } else {
                            r = new Report(9177);
                            r.subject = a.getId();
                            vDesc.add(r);
                        }
                    }
                    break;
                case Aero.CRIT_KF_BOOM:
                    
                    r = new Report(9180);
                    r.subject = a.getId();
                    vDesc.add(r);
                    break;
                case Aero.CRIT_CIC:
                    if (js == null) {
                        break;
                    }
                    
                    r = new Report(9185);
                    r.subject = a.getId();
                    vDesc.add(r);
                    js.setCICHits(js.getCICHits() + 1);
                    break;
                case Aero.CRIT_KF_DRIVE:
                    if (js == null) {
                        break;
                    }
                    
                    r = new Report(9190);
                    r.subject = a.getId();
                    vDesc.add(r);
                    js.setKFIntegrity(js.getKFIntegrity() - 1);
                    break;
                case Aero.CRIT_GRAV_DECK:
                    if (js == null) {
                        break;
                    }
                    
                    r = new Report(9195);
                    r.subject = a.getId();
                    vDesc.add(r);
                    break;
                case Aero.CRIT_LIFE_SUPPORT:
                    
                    a.setLifeSupport(false);
                    r = new Report(9196);
                    r.subject = a.getId();
                    vDesc.add(r);
                    break;
            }
        } else if (en instanceof BattleArmor) {
            
            BattleArmor ba = (BattleArmor) en;
            r = new Report(6111);
            int randomTrooper = ba.getRandomTrooper();
            ba.destroyLocation(randomTrooper);
            r.add(randomTrooper);
            r.newlines = 1;
            vDesc.add(r);
        } else if (CriticalSlot.TYPE_SYSTEM == cs.getType()) {
            
            cs.setHit(true);
            if (en instanceof Protomech) {
                int numHit = ((Protomech) en).getCritsHit(loc);
                if ((cs.getIndex() != Protomech.SYSTEM_TORSO_WEAPON_A)
                    && (cs.getIndex() != Protomech.SYSTEM_TORSO_WEAPON_B)
                    && (cs.getIndex() != Protomech.SYSTEM_TORSO_WEAPON_C)
                    && (cs.getIndex() != Protomech.SYSTEM_TORSO_WEAPON_D)
                    && (cs.getIndex() != Protomech.SYSTEM_TORSO_WEAPON_E)
                    && (cs.getIndex() != Protomech.SYSTEM_TORSO_WEAPON_F)) {
                    r = new Report(6225);
                    r.subject = en.getId();
                    r.indent(3);
                    r.add(Protomech.systemNames[cs.getIndex()]);
                    vDesc.addElement(r);
                }
                switch (cs.getIndex()) {
                    case Protomech.SYSTEM_HEADCRIT:
                        if (2 == numHit) {
                            r = new Report(6230);
                            r.subject = en.getId();
                            vDesc.addElement(r);
                            en.destroyLocation(loc);
                        }
                        break;
                    case Protomech.SYSTEM_ARMCRIT:
                        if (2 == numHit) {
                            r = new Report(6235);
                            r.subject = en.getId();
                            vDesc.addElement(r);
                            en.destroyLocation(loc);
                        }
                        break;
                    case Protomech.SYSTEM_LEGCRIT:
                        if (3 == numHit) {
                            r = new Report(6240);
                            r.subject = en.getId();
                            r.newlines = 0;
                            vDesc.addElement(r);
                            en.destroyLocation(loc);
                        }
                        break;
                    case Protomech.SYSTEM_TORSOCRIT:
                        if (3 == numHit) {
                            vDesc.addAll(destroyEntity(en, "torso destruction"));
                        }
                        
                        else if (secondaryEffects) {
                            int tweapRoll = Compute.d6(1);
                            CriticalSlot newSlot = null;

                            switch (tweapRoll) {
                                case 1:
                                    if (((Protomech) en).isQuad()) {
                                        newSlot = new CriticalSlot(
                                                CriticalSlot.TYPE_SYSTEM,
                                                Protomech.SYSTEM_TORSO_WEAPON_A);
                                        vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                      newSlot, secondaryEffects,
                                                                      damageCaused, isCapital));
                                        break;
                                    }
                                case 2:
                                    if (((Protomech) en).isQuad()) {
                                        newSlot = new CriticalSlot(
                                                CriticalSlot.TYPE_SYSTEM,
                                                Protomech.SYSTEM_TORSO_WEAPON_B);
                                        vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                      newSlot, secondaryEffects,
                                                                      damageCaused, isCapital));
                                        break;
                                    }
                                    newSlot = new CriticalSlot(
                                            CriticalSlot.TYPE_SYSTEM,
                                            Protomech.SYSTEM_TORSO_WEAPON_A);
                                    vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                  newSlot, secondaryEffects, damageCaused,
                                                                  isCapital));
                                    break;
                                case 3:
                                    if (((Protomech) en).isQuad()) {
                                        newSlot = new CriticalSlot(
                                                CriticalSlot.TYPE_SYSTEM,
                                                Protomech.SYSTEM_TORSO_WEAPON_C);
                                        vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                      newSlot, secondaryEffects,
                                                                      damageCaused, isCapital));
                                        break;
                                    }
                                case 4:
                                    if (((Protomech) en).isQuad()) {
                                        newSlot = new CriticalSlot(
                                                CriticalSlot.TYPE_SYSTEM,
                                                Protomech.SYSTEM_TORSO_WEAPON_D);
                                        vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                      newSlot, secondaryEffects,
                                                                      damageCaused, isCapital));
                                        break;
                                    }
                                    newSlot = new CriticalSlot(
                                            CriticalSlot.TYPE_SYSTEM,
                                            Protomech.SYSTEM_TORSO_WEAPON_B);
                                    vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                  newSlot, secondaryEffects, damageCaused,
                                                                  isCapital));
                                    break;
                                case 5:
                                    if (en.getWeight() > 9) {
                                        if (((Protomech) en).isQuad()) {
                                            newSlot = new CriticalSlot(
                                                    CriticalSlot.TYPE_SYSTEM,
                                                    Protomech.SYSTEM_TORSO_WEAPON_E);
                                            vDesc.addAll(applyCriticalHit(en,
                                                                          Entity.NONE, newSlot,
                                                                          secondaryEffects, damageCaused,
                                                                          isCapital));
                                            break;
                                        }
                                        newSlot = new CriticalSlot(
                                                CriticalSlot.TYPE_SYSTEM,
                                                Protomech.SYSTEM_TORSO_WEAPON_C);
                                        vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                      newSlot, secondaryEffects,
                                                                      damageCaused, isCapital));
                                        break;
                                    }
                                case 6:
                                    if (en.getWeight() > 9) {
                                        if (((Protomech) en).isQuad()) {
                                            newSlot = new CriticalSlot(
                                                    CriticalSlot.TYPE_SYSTEM,
                                                    Protomech.SYSTEM_TORSO_WEAPON_F);
                                            vDesc.addAll(applyCriticalHit(en,
                                                                          Entity.NONE, newSlot,
                                                                          secondaryEffects, damageCaused,
                                                                          isCapital));
                                            break;
                                        }
                                        newSlot = new CriticalSlot(
                                                CriticalSlot.TYPE_SYSTEM,
                                                Protomech.SYSTEM_TORSO_WEAPON_C);
                                        vDesc.addAll(applyCriticalHit(en, Entity.NONE,
                                                                      newSlot, secondaryEffects,
                                                                      damageCaused, isCapital));
                                        break;
                                    }
                            }
                        }
                        break;
                    case Protomech.SYSTEM_TORSO_WEAPON_A:
                        Mounted weaponA = ((Protomech) en).getTorsoWeapon(cs
                                                                                  .getIndex());
                        if (null != weaponA) {
                            weaponA.setHit(true);
                            r = new Report(6245);
                            r.subject = en.getId();
                            r.newlines = 0;
                            vDesc.addElement(r);
                        }
                        break;
                    case Protomech.SYSTEM_TORSO_WEAPON_B:
                        Mounted weaponB = ((Protomech) en).getTorsoWeapon(cs
                                                                                  .getIndex());
                        if (null != weaponB) {
                            weaponB.setHit(true);
                            r = new Report(6250);
                            r.subject = en.getId();
                            r.newlines = 0;
                            vDesc.addElement(r);
                        }
                        break;
                    case Protomech.SYSTEM_TORSO_WEAPON_C:
                        Mounted weaponC = ((Protomech) en).getTorsoWeapon(cs
                                                                                  .getIndex());
                        if (null != weaponC) {
                            weaponC.setHit(true);
                            r = new Report(6245);
                            r.subject = en.getId();
                            r.newlines = 0;
                            vDesc.addElement(r);
                        }
                        break;
                    case Protomech.SYSTEM_TORSO_WEAPON_D:
                        Mounted weaponD = ((Protomech) en).getTorsoWeapon(cs
                                                                                  .getIndex());
                        if (null != weaponD) {
                            weaponD.setHit(true);
                            r = new Report(6250);
                            r.subject = en.getId();
                            r.newlines = 0;
                            vDesc.addElement(r);
                        }
                        break;
                    case Protomech.SYSTEM_TORSO_WEAPON_E:
                        Mounted weaponE = ((Protomech) en).getTorsoWeapon(cs
                                                                                  .getIndex());
                        if (null != weaponE) {
                            weaponE.setHit(true);
                            r = new Report(6245);
                            r.subject = en.getId();
                            r.newlines = 0;
                            vDesc.addElement(r);
                        }
                        break;
                    case Protomech.SYSTEM_TORSO_WEAPON_F:
                        Mounted weaponF = ((Protomech) en).getTorsoWeapon(cs
                                                                                  .getIndex());
                        if (null != weaponF) {
                            weaponF.setHit(true);
                            r = new Report(6250);
                            r.subject = en.getId();
                            r.newlines = 0;
                            vDesc.addElement(r);
                        }
                        break;
                } 

                if (((Protomech) en).shaded(loc, numHit)) {
                    
                    int pHits = Protomech.POSSIBLE_PILOT_DAMAGE[loc]
                                - ((Protomech) en).getPilotDamageTaken(loc);
                    if (Math.min(1, pHits) > 0) {
                        Report.addNewline(vDesc);
                        vDesc.addAll(damageCrew(en, 1));
                        pHits = 1 + ((Protomech) en).getPilotDamageTaken(loc);
                        ((Protomech) en).setPilotDamageTaken(loc, pHits);
                    }
                } 

            } 
            else {
                r = new Report(6225);
                r.subject = en.getId();
                r.indent(3);
                r.add(((Mech) en).getSystemName(cs.getIndex()));
                vDesc.addElement(r);
                switch (cs.getIndex()) {
                    case Mech.SYSTEM_COCKPIT:
                        
                        if (en instanceof Mech) {
                            Mech mech = (Mech) en;
                            if (game.getOptions().booleanOption(
                                    "tacops_skin_of_the_teeth_ejection")) {
                                if (mech.isAutoEject()
                                    && (!game.getOptions().booleanOption(
                                        "conditional_ejection") || (game
                                                                            .getOptions().booleanOption(
                                                "conditional_ejection") && mech
                                                                            .isCondEjectHeadshot()))) {
                                    vDesc.addAll(ejectEntity(en, true, true));
                                }
                            }
                        }

                        if (Crew.DEATH > en.getCrew().getHits()) {
                            
                            en.getCrew().setDoomed(true);
                            Report.addNewline(vDesc);
                            vDesc.addAll(destroyEntity(en, "pilot death", true));
                        }
                        break;
                    case Mech.SYSTEM_ENGINE:
                        
                        if (!cs.isMissing()) {
                            en.engineHitsThisPhase++;
                        }
                        int numEngineHits = en.getEngineHits();
                        boolean engineExploded = checkEngineExplosion(en, vDesc,
                                                                      numEngineHits);
                        int hitsToDestroy = 3;
                        if ((en instanceof Mech)
                            && ((Mech) en).isSuperHeavy()
                            && (en.getEngine().getEngineType() == Engine.COMPACT_ENGINE)) {
                            hitsToDestroy = 2;
                        }

                        if (!engineExploded && (numEngineHits >= hitsToDestroy)) {
                            
                            vDesc.addAll(destroyEntity(en, "engine destruction"));
                            if (game.getOptions()
                                    .booleanOption("auto_abandon_unit")) {
                                vDesc.addAll(abandonEntity(en));
                            }
                            en.setSelfDestructing(false);
                            en.setSelfDestructInitiated(false);
                        }
                        break;
                    case Mech.SYSTEM_GYRO:
                        int gyroHits = en.getHitCriticals(CriticalSlot.TYPE_SYSTEM,
                                                          Mech.SYSTEM_GYRO, loc);
                        if (en.getGyroType() != Mech.GYRO_HEAVY_DUTY) {
                            gyroHits++;
                        }
                        switch (gyroHits) {
                            case 3:
                                
                                game.addPSR(new PilotingRollData(en.getId(),
                                                                 TargetRoll.AUTOMATIC_FAIL, 1, "gyro destroyed"));
                                
                                en.setHullDown(false);
                                break;
                            case 2:
                                
                                game.addPSR(new PilotingRollData(en.getId(), 3,
                                                                 "gyro hit"));
                                break;
                            case 1:
                                
                                game.addPSR(new PilotingRollData(en.getId(), 2,
                                                                 "gyro hit"));
                                break;
                            default:
                              
                        }
                        break;
                    case Mech.ACTUATOR_UPPER_LEG:
                    case Mech.ACTUATOR_LOWER_LEG:
                    case Mech.ACTUATOR_FOOT:
                        
                        game.addPSR(new PilotingRollData(en.getId(), 1,
                                                         "leg/foot actuator hit"));
                        break;
                    case Mech.ACTUATOR_HIP:
                        
                        game.addPSR(new PilotingRollData(en.getId(), 2,
                                                         "hip actuator hit"));
                        break;
                }

            } 

        } 

        else if (CriticalSlot.TYPE_EQUIPMENT == cs.getType()) {
            cs.setHit(true);
            Mounted mounted = cs.getMount();
            EquipmentType eqType = mounted.getType();
            boolean hitBefore = mounted.isHit();

            r = new Report(6225);
            r.subject = en.getId();
            r.indent(3);
            r.add(mounted.getDesc());
            vDesc.addElement(r);

            if ((eqType instanceof MiscType) && ((MiscType) eqType).isShield()) {
                mounted.setHit(false);
            } else {
                mounted.setHit(true);
            }

            if ((eqType instanceof MiscType)
                && eqType.hasFlag(MiscType.F_HARJEL)) {
                vDesc.addAll(breachLocation(en, loc, null, true));
            }

            if (secondaryEffects && (eqType instanceof MiscType)
                && (eqType.hasFlag(MiscType.F_HARJEL_II)
                || eqType.hasFlag(MiscType.F_HARJEL_III))
                && !hitBefore) {
                r = new Report(9852);
                r.subject = en.getId();
                r.indent(2);
                vDesc.addElement(r);
                vDesc.addAll(criticalEntity(en, loc, false, 0, 0));
            }

            if (!hitBefore && (eqType instanceof MiscType)
                && eqType.hasFlag(MiscType.F_ECM)
                && (mounted.getLinkedBy() != null)) {
                Mounted stealth = mounted.getLinkedBy();
                r = new Report(6255);
                r.subject = en.getId();
                r.indent(2);
                r.add(stealth.getType().getName());
                vDesc.addElement(r);
                stealth.setMode("Off");
            }

            if (((secondaryEffects && eqType.isExplosive(mounted))
                 || mounted.isHotLoaded() || (mounted.hasChargedCapacitor() != 0))
                && !hitBefore) {
                vDesc.addAll(explodeEquipment(en, loc, mounted));
            }

            
            if (mounted.getBaseShotsLeft() > 0) {
                mounted.setShotsLeft(0);
            }
        } 
        
        if ((en instanceof Mech) && en.hitThisRoundByAntiTSM) {
            Mech mech = (Mech) en;
            if (mech.hasTSM()) {
                r = new Report(6430);
                r.subject = en.getId();
                r.indent(2);
                r.addDesc(en);
                r.newlines = 0;
                vDesc.addElement(r);
                vDesc.addAll(oneCriticalEntity(en, Compute.d6(2), false,
                                               damageCaused));
            }
            en.hitThisRoundByAntiTSM = false;
        }

        if (en.getCrew().getOptions().booleanOption("bvdni")
            && !en.getCrew().getOptions().booleanOption("pain_shunt")) {
            Report.addNewline(vDesc);
            int roll = Compute.d6(2);
            r = new Report(3580);
            r.subject = en.getId();
            r.addDesc(en);
            r.add(7);
            r.add(roll);
            r.choose(roll >= 8);
            r.indent(2);
            vDesc.add(r);
            if (roll >= 8) {
                vDesc.addAll(damageCrew(en, 1));
            }
        }

        return vDesc;
    }

    public Vector<Report> criticalEntity(Entity en, int loc, boolean isRear,
                                         int critMod, int damage) {
        return criticalEntity(en, loc, isRear, critMod, true, false, damage);
    }

    public Vector<Report> criticalEntity(Entity en, int loc, boolean isRear,
                                         int critMod, int damage, boolean damagedByFire) {
        return criticalEntity(en, loc, isRear, critMod, true, false, damage, damagedByFire);
    }

    public Vector<Report> oneCriticalEntity(Entity en, int loc,
                                             boolean isRear, int damage) {
        return criticalEntity(en, loc, isRear, 0, false, false, damage);
    }

    public Vector<Report> oneCriticalEntity(Entity en, int loc,
                                             boolean isRear, int damage, boolean damagedByFire) {
        return criticalEntity(en, loc, isRear, 0, false, false, damage, damagedByFire);
    }

    private Vector<Report> forceLandVTOLorWiGE(Tank en) {
        Vector<Report> vDesc = new Vector<Report>();
        PilotingRollData psr = en.getBasePilotingRoll();
        IHex hex = game.getBoard().getHex(en.getPosition());
        if (en instanceof VTOL) {
            psr.addModifier(4, "VTOL making forced landing");
        } else {
            psr.addModifier(0, "WiGE making forced landing");
        }
        int elevation = Math.max(hex.terrainLevel(Terrains.BLDG_ELEV),
                                 hex.terrainLevel(Terrains.BRIDGE_ELEV));
        elevation = Math.max(elevation, 0);
        elevation = Math.min(elevation, en.getElevation());
        if (en.getElevation() > elevation) {
            if (!hex.containsTerrain(Terrains.FUEL_TANK)
                && !hex.containsTerrain(Terrains.JUNGLE)
                && !hex.containsTerrain(Terrains.MAGMA)
                && !hex.containsTerrain(Terrains.MUD)
                && !hex.containsTerrain(Terrains.RUBBLE)
                && !hex.containsTerrain(Terrains.WATER)
                && !hex.containsTerrain(Terrains.WOODS)) {
                Report r = new Report(2180);
                r.subject = en.getId();
                r.addDesc(en);
                r.add(psr.getLastPlainDesc(), true);
                vDesc.add(r);

                
                final int diceRoll = Compute.d6(2);
                r = new Report(2185);
                r.subject = en.getId();
                r.add(psr.getValueAsString());
                r.add(psr.getDesc());
                r.add(diceRoll);
                if (diceRoll < psr.getValue()) {
                    r.choose(false);
                    vDesc.add(r);
                    vDesc.addAll(crashVTOLorWiGE(en, true));
                } else {
                    r.choose(true);
                    vDesc.add(r);
                    en.setElevation(elevation);
                }
            } else {
                vDesc.addAll(crashVTOLorWiGE(en, true));
            }
        }
        return vDesc;
    }

    private Vector<Report> crashVTOLorWiGE(Tank en) {
        return crashVTOLorWiGE(en, false, false, 0, en.getPosition(),
                               en.getElevation(), 0);
    }

    private Vector<Report> crashVTOLorWiGE(Tank en, boolean rerollRotorHits) {
        return crashVTOLorWiGE(en, rerollRotorHits, false, 0, en.getPosition(),
                               en.getElevation(), 0);
    }

    private Vector<Report> crashVTOLorWiGE(Tank en, boolean rerollRotorHits,
                                           boolean sideSlipCrash, int hexesMoved, Coords crashPos,
                                           int crashElevation, int impactSide) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if (!game.getBoard().contains(crashPos)) {
            return vDesc;
        }

        if (!sideSlipCrash) {
            
            if (en.getCrew().isDoomed()) {
                r = new Report(6260);
            } else {
                r = new Report(6260);
            }
            r.subject = en.getId();
            r.newlines = 0;
            r.addDesc(en);
            vDesc.addElement(r);
            int newElevation = 0;
            IHex fallHex = game.getBoard().getHex(crashPos);

            
            if (fallHex.containsTerrain(Terrains.BLDG_ELEV)) {
                newElevation = fallHex.terrainLevel(Terrains.BLDG_ELEV);
            } else if (fallHex.containsTerrain(Terrains.BRIDGE_ELEV)) {
                newElevation = fallHex.terrainLevel(Terrains.BRIDGE_ELEV);
                if (newElevation > crashElevation) {
                    newElevation = 0; 
                }
            }

            int fall = crashElevation - newElevation;
            if (fall == 0) {
                
                r = new Report(6265);
                r.subject = en.getId();
                vDesc.addElement(r);
                return vDesc;
            }
            
            en.setElevation(newElevation);

            
            r = new Report(6270);
            r.subject = en.getId();
            r.add(fall);
            vDesc.addElement(r);

            String side;
            int table;
            int facing = Compute.d6() - 1;
            switch (facing) {
                case 1:
                case 2:
                    side = "right side";
                    table = ToHitData.SIDE_RIGHT;
                    break;
                case 3:
                    side = "rear";
                    table = ToHitData.SIDE_REAR;
                    break;
                case 4:
                case 5:
                    side = "left side";
                    table = ToHitData.SIDE_LEFT;
                    break;
                case 0:
                default:
                    side = "front";
                    table = ToHitData.SIDE_FRONT;
            }

            if (newElevation <= 0) {
                boolean waterFall = fallHex.containsTerrain(Terrains.WATER);
                if (waterFall && fallHex.containsTerrain(Terrains.ICE)) {
                    int roll = Compute.d6(1);
                    r = new Report(2119);
                    r.subject = en.getId();
                    r.addDesc(en);
                    r.add(roll);
                    r.subject = en.getId();
                    vDesc.add(r);
                    if (roll > 3) {
                        vDesc.addAll(resolveIceBroken(crashPos));
                    } else {
                        waterFall = false; 
                    }
                }
                if (waterFall) {
                    
                    r = new Report(6275);
                    r.subject = en.getId();
                    vDesc.addElement(r);
                    vDesc.addAll(destroyEntity(en, "Fell into water", false,
                                               false));      
                }
            }

            int damage = (int) Math.round(en.getWeight() / 10.0) * (fall + 1);

            damage = Math.round(damage
                                * game.getPlanetaryConditions().getGravity());
            
            r = new Report(6280);
            r.subject = en.getId();
            r.indent();
            r.addDesc(en);
            r.add(side);
            r.add(damage);
            
            vDesc.addElement(r);

            en.setFacing((en.getFacing() + (facing)) % 6);

            boolean exploded = false;

            while (damage > 0) {
                int cluster = Math.min(5, damage);
                HitData hit = en.rollHitLocation(ToHitData.HIT_NORMAL, table);
                if ((en instanceof VTOL)
                    && (hit.getLocation() == VTOL.LOC_ROTOR)
                    && rerollRotorHits) {
                    continue;
                }
                hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                int isBefore[] = {en.getInternal(Tank.LOC_FRONT),
                                  en.getInternal(Tank.LOC_RIGHT),
                                  en.getInternal(Tank.LOC_LEFT),
                                  en.getInternal(Tank.LOC_REAR)};
                vDesc.addAll(damageEntity(en, hit, cluster));
                int ISAfter[] = {en.getInternal(Tank.LOC_FRONT),
                                 en.getInternal(Tank.LOC_RIGHT),
                                 en.getInternal(Tank.LOC_LEFT),
                                 en.getInternal(Tank.LOC_REAR)};
                for (int x = 0; x <= 3; x++) {
                    if (isBefore[x] != ISAfter[x]) {
                        exploded = true;
                    }
                }
                damage -= cluster;
            }
            if (exploded) {
                r = new Report(6285);
                r.subject = en.getId();
                r.addDesc(en);
                vDesc.addElement(r);
                vDesc.addAll(explodeVTOLorWiGE(en));
            }

            vDesc.addAll(doSetLocationsExposure(en, fallHex, false,
                                                newElevation));

        } else {
            en.setElevation(0);
            
            r = new Report(6290);
            r.subject = en.getId();
            r.addDesc(en);
            vDesc.addElement(r);
            int damage = (int) Math.round(en.getWeight() / 10.0)
                         * (hexesMoved + 1);
            boolean exploded = false;

            while (damage > 0) {
                int cluster = Math.min(5, damage);
                HitData hit = en.rollHitLocation(ToHitData.HIT_NORMAL,
                                                 impactSide);
                hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                int ISBefore[] = {en.getInternal(Tank.LOC_FRONT),
                                  en.getInternal(Tank.LOC_RIGHT),
                                  en.getInternal(Tank.LOC_LEFT),
                                  en.getInternal(Tank.LOC_REAR)};
                vDesc.addAll(damageEntity(en, hit, cluster));
                int ISAfter[] = {en.getInternal(Tank.LOC_FRONT),
                                 en.getInternal(Tank.LOC_RIGHT),
                                 en.getInternal(Tank.LOC_LEFT),
                                 en.getInternal(Tank.LOC_REAR)};
                for (int x = 0; x <= 3; x++) {
                    if (ISBefore[x] != ISAfter[x]) {
                        exploded = true;
                    }
                }
                damage -= cluster;
            }
            if (exploded) {
                r = new Report(6295);
                r.subject = en.getId();
                r.addDesc(en);
                vDesc.addElement(r);
                vDesc.addAll(explodeVTOLorWiGE(en));
            }
        }

        if (game.containsMinefield(crashPos)) {
            
            enterMinefield(en, crashPos, 0, true, vDesc, 7);
            
            clearDetonatedMines(crashPos, 5);
            resetMines();
        }

        return vDesc;

    }

    private Vector<Report> explodeVTOLorWiGE(Tank en) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if (en.getEngine().isFusion()) {
            
            r = new Report(6300);
            r.subject = en.getId();
            vDesc.addElement(r);
        } else {
            Coords pos = en.getPosition();
            IHex hex = game.getBoard().getHex(pos);
            if (hex.containsTerrain(Terrains.WOODS)
                || hex.containsTerrain(Terrains.JUNGLE)) {
                ignite(pos, Terrains.FIRE_LVL_NORMAL, vDesc);
            } else {
                ignite(pos, Terrains.FIRE_LVL_INFERNO, vDesc);
            }
            vDesc.addAll(destroyEntity(en, "crashed and burned", false, false));
        }

        return vDesc;
    }

    private Vector<Report> criticalTank(Tank t, int loc, int critMod, int damage, boolean damagedByFire) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        r = new Report(6305);
        r.subject = t.getId();
        r.indent(3);
        r.add(t.getLocationAbbr(loc));
        r.newlines = 0;
        vDesc.add(r);
        int roll = Compute.d6(2);
        r = new Report(6310);
        r.subject = t.getId();
        String rollString = "";
        if (critMod != 0) {
            rollString = "(" + roll;
            if (critMod > 0) {
                rollString += "+";
            }
            rollString += critMod + ") = ";
            roll += critMod;
        }
        rollString += roll;
        r.add(rollString);
        r.newlines = 0;
        vDesc.add(r);

        int critType = t.getCriticalEffect(roll, loc, damagedByFire);
        if ((critType == Tank.CRIT_NONE)
                && game.getOptions().booleanOption("vehicles_threshold")
                && !t.getOverThresh()) {
            r = new Report(6006);
            r.subject = t.getId();
            r.newlines = 0;
            vDesc.add(r);
        }
        vDesc.addAll(applyCriticalHit(t, loc, new CriticalSlot(0, critType),
                                      true, damage, false));
        if ((critType != Tank.CRIT_NONE) && !t.getEngine().isFusion()
            && t.hasQuirk("fragile_fuel") && (Compute.d6(2) > 9)) {
            
            vDesc.addAll(applyCriticalHit(t, loc, new CriticalSlot(0,
                    Tank.CRIT_FUEL_TANK), true, damage, false));
        }
        return vDesc;
    }

    private void checkAeroCrits(Vector<Report> vDesc, Aero a, HitData hit,
                                int damage_orig, boolean critThresh, boolean critSI,
                                boolean ammoExplosion, boolean nukeS2S) {
        Report r;

        boolean isCapital = hit.isCapital();
        
        int capitalMissile = hit.getCapMisCritMod();

        if (nukeS2S) {
            
            PilotingRollData nukePSR = new PilotingRollData(a.getId(), 4,
                                                            "Nuclear attack", false);
            game.addControlRoll(nukePSR);

            int nukeroll = Compute.d6(2);
            r = new Report(9145);
            r.subject = a.getId();
            r.indent(3);
            r.add(capitalMissile);
            r.add(nukeroll);
            vDesc.add(r);
            if (nukeroll >= capitalMissile) {
                int nukeDamage = damage_orig;
                a.setSI(a.getSI() - nukeDamage);
                a.damageThisPhase += nukeDamage;
                r = new Report(9146);
                r.subject = a.getId();
                r.add(nukeDamage);
                r.add(Math.max(a.getSI(), 0));
                vDesc.addElement(r);
                if (a.getSI() <= 0) {
                    vDesc.addAll(destroyEntity(a,
                            "Structural Integrity Collapse"));
                    a.setSI(0);
                    if (hit.getAttackerId() != Entity.NONE) {
                        creditKill(a, game.getEntity(hit.getAttackerId()));
                    }
                } else if (!critSI) {
                    critSI = true;
                }
            } else {
                r = new Report(9147);
                r.subject = a.getId();
                vDesc.addElement(r);
            }
        }

        if (hit.rolledBoxCars()) {
            if (hit.isFirstHit()) {
                vDesc.addAll(criticalAero(a, hit.getLocation(),
                        hit.glancingMod(), "12 to hit", 8, damage_orig,
                        isCapital));
            } else { 
                r = new Report(9102);
                r.subject = a.getId();
                r.indent(3);
                vDesc.addElement(r);
            }
        }
        
        if (critThresh && !ammoExplosion) {
            vDesc.addAll(criticalAero(a, hit.getLocation(), hit.glancingMod(),
                    "Damage threshold exceeded", 8, damage_orig, isCapital));
        }
        if (critSI && !ammoExplosion) {
            vDesc.addAll(criticalAero(a, hit.getLocation(), hit.glancingMod(),
                    "SI damaged", 8, damage_orig, isCapital));
        }
        if ((capitalMissile > 0) && !nukeS2S) {
            vDesc.addAll(criticalAero(a, hit.getLocation(), hit.glancingMod(),
                    "Capital Missile", capitalMissile, damage_orig, isCapital));
        }
    }

    private Vector<Report> criticalAero(Aero a, int loc, int critMod,
                                        String reason, int target, int damage, boolean isCapital) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        r = new Report(9100);
        r.subject = a.getId();
        r.add(reason);
        r.indent(3);
        r.newlines = 0;
        vDesc.add(r);
        int roll = Compute.d6(2);
        r = new Report(9101);
        r.subject = a.getId();
        r.add(target);
        String rollString = "";
        if (critMod != 0) {
            rollString = "(" + roll;
            if (critMod > 0) {
                rollString += "+";
            }
            rollString += critMod + ") = ";
            roll += critMod;
        }
        rollString += roll;
        r.add(rollString);
        r.newlines = 0;
        vDesc.add(r);

        int critType = a.getCriticalEffect(roll, target);
        vDesc.addAll(applyCriticalHit(a, loc, new CriticalSlot(0, critType),
                                      true, damage, isCapital));
        return vDesc;
    }

    public Vector<Report> criticalEntity(Entity en, int loc, boolean isRear,
                                         int critMod, boolean rollNumber, boolean isCapital, int damage) {
        return criticalEntity(en, loc, isRear, critMod, rollNumber, isCapital, damage, false);
    }

    public Vector<Report> criticalEntity(Entity en, int loc, boolean isRear,
                                         int critMod, boolean rollNumber, boolean isCapital, int damage, boolean damagedByFire) {

        if (en.hasQuirk("poor_work")) {
            critMod += 1;
        }
        if (en.hasQuirk(OptionsConstants.QUIRK_NEG_PROTOTYPE)) {
            critMod += 2;
        }

        if ((en.getArmor(loc, isRear) > 0)
            && (en.getArmorType(loc) == EquipmentType.T_ARMOR_ANTI_PENETRATIVE_ABLATION)) {
            critMod -= 2;
        }

        if (en instanceof Tank) {
            return criticalTank((Tank) en, loc, critMod, damage, damagedByFire);
        }
        if (en instanceof Aero) {
            return criticalAero((Aero) en, loc, critMod, "unknown", 8, damage,
                                isCapital);
        }
        CriticalSlot slot = null;
        Vector<Report> vDesc = new Vector<Report>();
        Report r;
        Coords coords = en.getPosition();
        IHex hex = null;
        int hits;
        if (rollNumber) {
            if (null != coords) {
                hex = game.getBoard().getHex(coords);
            }
            r = new Report(6305);
            r.subject = en.getId();
            r.indent(3);
            r.add(en.getLocationAbbr(loc));
            r.newlines = 0;
            vDesc.addElement(r);
            hits = 0;
            int roll = Compute.d6(2);
            r = new Report(6310);
            r.subject = en.getId();
            String rollString = "";
            
            if ((en instanceof Mech) && ((Mech) en).isIndustrial()) {
                critMod += 2;
            }
            
            if ((en instanceof Mech) && ((Mech) en).hasReinforcedStructure()) {
                critMod -= 1;
            }
            if (critMod != 0) {
                rollString = "(" + roll;
                if (critMod > 0) {
                    rollString += "+";
                }
                rollString += critMod + ") = ";
                roll += critMod;
            }
            rollString += roll;
            r.add(rollString);
            r.newlines = 0;
            vDesc.addElement(r);
            boolean advancedCrit = game.getOptions().booleanOption(
                    "tacops_crit_roll");
            if ((!advancedCrit && (roll <= 7)) || (advancedCrit && (roll <= 8))) {
                
                r = new Report(6005);
                r.subject = en.getId();
                vDesc.addElement(r);
                return vDesc;
            } else if ((!advancedCrit && (roll >= 8) && (roll <= 9))
                       || (advancedCrit && (roll >= 9) && (roll <= 10))) {
                hits = 1;
                r = new Report(6315);
                r.subject = en.getId();
                vDesc.addElement(r);
            } else if ((!advancedCrit && (roll >= 10) && (roll <= 11))
                       || (advancedCrit && (roll >= 11) && (roll <= 12))) {
                hits = 2;
                r = new Report(6320);
                r.subject = en.getId();
                vDesc.addElement(r);
            } else if (advancedCrit && (roll >= 13) && (roll <= 14)) {
                hits = 3;
                r = new Report(6325);
                r.subject = en.getId();
                vDesc.addElement(r);
            } else if ((!advancedCrit && (roll >= 12))
                       || (advancedCrit && (roll >= 15))) {
                if (en instanceof Protomech) {
                    hits = 3;
                    r = new Report(6325);
                    r.subject = en.getId();
                    vDesc.addElement(r);
                } else if (en.locationIsLeg(loc)) {

                    CriticalSlot cs = en.getCritical(loc, 0);
                    if ((cs != null) && cs.isArmored()) {
                        r = new Report(6700);
                        r.subject = en.getId();
                        r.add(en.getLocationName(loc));
                        r.newlines = 0;
                        vDesc.addElement(r);
                        cs.setArmored(false);
                        return vDesc;
                    }
                    
                    r = new Report(6120);
                    r.subject = en.getId();
                    r.add(en.getLocationName(loc));
                    vDesc.addElement(r);
                    if (en.getInternal(loc) > 0) {
                        en.destroyLocation(loc, true);
                    }
                    if (null != hex) {
                        if (!hex.containsTerrain(Terrains.LEGS)) {
                            hex.addTerrain(Terrains.getTerrainFactory()
                                                   .createTerrain(Terrains.LEGS, 1));
                        } else {
                            hex.addTerrain(Terrains
                                                   .getTerrainFactory()
                                                   .createTerrain(Terrains.LEGS,
                                                                  hex.terrainLevel(Terrains.LEGS) + 1));
                        }
                    }
                    sendChangedHex(en.getPosition());
                    return vDesc;
                } else if ((loc == Mech.LOC_RARM) || (loc == Mech.LOC_LARM)) {
                    CriticalSlot cs = en.getCritical(loc, 0);
                    if ((cs != null) && cs.isArmored()) {
                        r = new Report(6700);
                        r.subject = en.getId();
                        r.add(en.getLocationName(loc));
                        r.newlines = 0;
                        vDesc.addElement(r);
                        cs.setArmored(false);
                        return vDesc;
                    }

                    r = new Report(6120);
                    r.subject = en.getId();
                    r.add(en.getLocationName(loc));
                    vDesc.addElement(r);
                    en.destroyLocation(loc, true);
                    if (null != hex) {
                        if (!hex.containsTerrain(Terrains.ARMS)) {
                            hex.addTerrain(Terrains.getTerrainFactory()
                                                   .createTerrain(Terrains.ARMS, 1));
                        } else {
                            hex.addTerrain(Terrains
                                                   .getTerrainFactory()
                                                   .createTerrain(Terrains.ARMS,
                                                                  hex.terrainLevel(Terrains.ARMS) + 1));
                        }
                    }
                    sendChangedHex(en.getPosition());
                    return vDesc;
                } else if (loc == Mech.LOC_HEAD) {
                    
                    r = new Report(6330);
                    r.subject = en.getId();
                    r.add(en.getLocationName(loc));
                    vDesc.addElement(r);
                    en.destroyLocation(loc, true);
                    if (((Mech) en).getCockpitType() != Mech.COCKPIT_TORSO_MOUNTED) {
                        
                        if (Crew.DEATH > en.getCrew().getHits()) {
                            en.getCrew().setDoomed(true);
                            Report.addNewline(vDesc);
                            vDesc.addAll(destroyEntity(en, "pilot death", true));
                        }
                    }
                    return vDesc;
                } else {
                    
                    hits = 3;
                    
                    if ((roll >= 14) && (en instanceof Mech)
                        && ((Mech) en).isIndustrial()) {
                        hits = 4;
                    }
                    r = new Report(6325);
                    r.subject = en.getId();
                    vDesc.addElement(r);
                }
            }
        } else {
            hits = 1;
        }

        while ((hits > 0) && en.canTransferCriticals(loc)
               && (en.getTransferLocation(loc) != Entity.LOC_DESTROYED)
               && (en.getTransferLocation(loc) != Entity.LOC_NONE)) {
            loc = en.getTransferLocation(loc);
            r = new Report(6335);
            r.subject = en.getId();
            r.indent(3);
            r.add(en.getLocationAbbr(loc));
            vDesc.addElement(r);
        }

        while (hits > 0) {

            
            if (en.getHittableCriticals(loc) <= 0) {
                r = new Report(6340);
                r.subject = en.getId();
                r.indent(3);
                vDesc.addElement(r);
                break;
            }

            int slotIndex = Compute.randomInt(en.getNumberOfCriticals(loc));
            slot = en.getCritical(loc, slotIndex);

            if ((slot != null) && slot.isHittable()) {

                if (slot.isArmored()) {
                    r = new Report(6710);
                    r.subject = en.getId();
                    if (slot.getType() == CriticalSlot.TYPE_SYSTEM) {
                        
                        if (en instanceof Mech) {
                            r.add(((Mech) en).getSystemName(slot.getIndex()));
                        }
                    } else {
                        
                        if (slot.getMount() != null) {
                            r.add(slot.getMount().getName());
                        }
                    }
                    vDesc.addElement(r);
                    slot.setArmored(false);
                    hits--;
                    continue;
                }
                
                if ((en instanceof Mech)
                    && (en.getCrew().hasEdgeRemaining() && en.getCrew()
                                                             .getOptions()
                                                             .booleanOption("edge_when_explosion"))
                    && (slot.getType() == CriticalSlot.TYPE_EQUIPMENT)
                    && slot.getMount().getType()
                           .isExplosive(slot.getMount())) {
                    en.getCrew().decreaseEdge();
                    r = new Report(6530);
                    r.subject = en.getId();
                    r.indent(3);
                    r.add(en.getCrew().getOptions().intOption("edge"));
                    vDesc.addElement(r);
                    continue;
                }

                if (en.getArmorType(loc) == EquipmentType.T_ARMOR_REACTIVE) {
                    Mounted mount = en.getEquipment(slot.getIndex());
                    if ((mount != null)
                        && (mount.getType() instanceof MiscType)
                        && ((MiscType) mount.getType())
                            .hasFlag(MiscType.F_REACTIVE)) {
                        int roll = Compute.d6(2);
                        r = new Report(6082);
                        r.subject = en.getId();
                        r.indent(3);
                        r.newlines = 0;
                        r.add(roll);
                        vDesc.addElement(r);
                        
                        if (roll == 2) {
                            r = new Report(6083);
                            r.subject = en.getId();
                            r.indent(3);
                            vDesc.addElement(r);
                            vDesc.addElement(r);
                            vDesc.addAll(damageEntity(en, new HitData(loc),
                                                      en.getArmor(loc)));
                            if (en.hasRearArmor(loc)) {
                                vDesc.addAll(damageEntity(en, new HitData(loc,
                                                                          true), en.getArmor(loc, true)));
                            }
                            vDesc.addAll(damageEntity(en, new HitData(loc), 1));
                        } else {
                            continue;
                        }
                    }
                }
                vDesc.addAll(applyCriticalHit(en, loc, slot, true, damage,
                                              isCapital));
                hits--;
            }

        } 

        return vDesc;
    }

    private Vector<Report> breachCheck(Entity entity, int loc, IHex hex) {
        return breachCheck(entity, loc, hex, false);
    }

    private Vector<Report> breachCheck(Entity entity, int loc, IHex hex,
                                       boolean underWater) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if (entity instanceof Infantry) {
            return vDesc;
        }

        if (entity instanceof VTOL) {
            return vDesc;
        }

        boolean dumping = false;
        for (Mounted m : entity.getAmmo()) {
            if (m.isDumping()) {
                
                dumping = true;
                break;
            }
        }
        
        if ((entity.getLocationStatus(loc) > ILocationExposureStatus.NORMAL)
            || (entity.isSurfaceNaval() && (loc != ((Tank) entity)
                .getLocTurret()))) {
            
            int breachroll = 0;
            
            int target = 10;
            
            if ((entity.getLocationStatus(loc) == ILocationExposureStatus.VACUUM)
                && (game.getPlanetaryConditions().getAtmosphere() == PlanetaryConditions.ATMO_TRACE)) {
                target = 12;
            }
            
            if (entity.isSurfaceNaval() && !underWater) {
                target = 12;
            }
            if ((entity.getArmor(loc) > 0)
                && (entity instanceof Mech ? entity.getArmor(loc, true) > 0
               
                if (entity.hasHarJelIn(loc)) {
                    r = new Report(6342);
                    r.subject = entity.getId();
                    r.indent(3);
                    vDesc.addElement(r);
                    return vDesc;
                }
                if ((entity instanceof Mech)
                    && (((Mech) entity).hasHarJelIIIn(loc)
                        || ((Mech) entity).hasHarJelIIIIn(loc))) {
                    r = new Report(6343);
                    r.subject = entity.getId();
                    r.indent(3);
                    vDesc.addElement(r);
                    target -= 2;
                }
                
                if ((entity.getArmorType(loc) == EquipmentType.T_ARMOR_IMPACT_RESISTANT)) {
                    r = new Report(6344);
                    r.subject = entity.getId();
                    r.indent(3);
                    vDesc.addElement(r);
                    target += 1;
                }
                breachroll = Compute.d6(2);
                r = new Report(6345);
                r.subject = entity.getId();
                r.indent(3);
                r.add(entity.getLocationAbbr(loc));
                r.add(breachroll);
                r.newlines = 0;
                if (breachroll >= target) {
                    r.choose(false);
                } else {
                    r.choose(true);
                }
                vDesc.addElement(r);
            }
            
            if ((breachroll >= target)
                || !(entity.getArmor(loc) > 0)
                || (dumping && (!(entity instanceof Mech)
                                || (loc == Mech.LOC_CT) || (loc == Mech.LOC_RT) || (loc == Mech.LOC_LT)))
                || !(entity instanceof Mech ? entity.getArmor(loc, true) > 0
                                            : true)) {
                
                if ((entity.hasHarJelIn(loc))
                    && (entity.getArmor(loc) > 0)
                    && (entity instanceof Mech ? entity.getArmor(loc, true) > 0
                                               : true) && !dumping) {
                    r = new Report(6342);
                    r.subject = entity.getId();
                    r.indent(3);
                    vDesc.addElement(r);
                    return vDesc;
                }
                vDesc.addAll(breachLocation(entity, loc, hex, false));
            }
        }
        return vDesc;
    }

    private Vector<Report> breachLocation(Entity entity, int loc, IHex hex,
                                          boolean harJel) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if ((entity.getInternal(loc) < 0)
            || (entity.getLocationStatus(loc) < ILocationExposureStatus.NORMAL)) {
            
            return vDesc;
        }

        r = new Report(6350);
        if (harJel) {
            r.messageId = 6351;
        }
        r.subject = entity.getId();
        r.add(entity.getShortName());
        r.add(entity.getLocationAbbr(loc));
        vDesc.addElement(r);

        if (entity instanceof Tank) {
            vDesc.addAll(destroyEntity(entity, "hull breach", true, true));
            return vDesc;
        }
        if (entity instanceof Mech) {
            Mech mech = (Mech) entity;
            
            for (Mounted mounted : entity.getEquipment()) {
                if (mounted.getLocation() == loc) {
                    mounted.setBreached(true);
                }
            }
            
            for (int i = 0; i < entity.getNumberOfCriticals(loc); i++) {
                final CriticalSlot cs = entity.getCritical(loc, i);
                if (cs != null) {
                    
                    
                    if (entity.locationIsLeg(loc)) {
                        if (cs.isHittable()) {
                            switch (cs.getIndex()) {
                                case Mech.ACTUATOR_UPPER_LEG:
                                case Mech.ACTUATOR_LOWER_LEG:
                                case Mech.ACTUATOR_FOOT:
                                    
                                    game.addPSR(new PilotingRollData(
                                            entity.getId(), 1,
                                            "leg/foot actuator hit"));
                                    break;
                                case Mech.ACTUATOR_HIP:
                       
                                    game.addPSR(new PilotingRollData(
                                            entity.getId(), 0, "hip actuator hit"));
                                    break;
                            }
                        }
                    }
                    cs.setBreached(true);
                }
            }

            if (loc == Mech.LOC_CT) {
                vDesc.addAll(destroyEntity(entity, "hull breach"));
                if (game.getOptions().booleanOption("auto_abandon_unit")) {
                    vDesc.addAll(abandonEntity(entity));
                }
            }
            if (loc == Mech.LOC_HEAD) {
                entity.getCrew().setDoomed(true);
                vDesc.addAll(destroyEntity(entity, "hull breach"));
                if (entity.getLocationStatus(loc) == ILocationExposureStatus.WET) {
                    r = new Report(6355);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    vDesc.addElement(r);
                } else {
                    r = new Report(6360);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    vDesc.addElement(r);
                }
            }

            entity.setLocationStatus(loc, ILocationExposureStatus.BREACHED);

            int hitsToDestroy = 3;
            if (mech.isSuperHeavy()
                && (mech.getEngine().getEngineType() == Engine.COMPACT_ENGINE)) {
                hitsToDestroy = 2;
            }
            if ((entity.getHitCriticals(CriticalSlot.TYPE_SYSTEM,
                                        Mech.SYSTEM_ENGINE, Mech.LOC_LT)
                 + entity.getHitCriticals(CriticalSlot.TYPE_SYSTEM,
                                          Mech.SYSTEM_ENGINE, Mech.LOC_CT) + entity
                         .getHitCriticals(CriticalSlot.TYPE_SYSTEM,
                                          Mech.SYSTEM_ENGINE, Mech.LOC_RT)) >= hitsToDestroy) {
                vDesc.addAll(destroyEntity(entity, "engine destruction"));
                if (game.getOptions().booleanOption("auto_abandon_unit")) {
                    vDesc.addAll(abandonEntity(entity));
                }
            }

            if (loc == Mech.LOC_LT) {
                vDesc.addAll(breachLocation(entity, Mech.LOC_LARM, hex, false));
            }
            if (loc == Mech.LOC_RT) {
                vDesc.addAll(breachLocation(entity, Mech.LOC_RARM, hex, false));
            }

        }
        return vDesc;
    }

    private Vector<Report> destroyEntity(Entity entity, String reason) {
        return destroyEntity(entity, reason, true);
    }

    public Vector<Report> destroyEntity(Entity entity, String reason,
                                        boolean survivable) {
        
        return destroyEntity(entity, reason, survivable, true);
    }

    private Vector<Report> destroyEntity(Entity entity, String reason,
                                         boolean survivable, boolean canSalvage) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if (entity.isAirborne()) {
            survivable = false;
        } else if (entity instanceof Aero) {
            survivable = true;
        }

        int condition = IEntityRemovalConditions.REMOVE_SALVAGEABLE;
        if (!canSalvage) {
            entity.setSalvage(canSalvage);
            condition = IEntityRemovalConditions.REMOVE_DEVASTATED;
        }

        if ((entity instanceof Infantry) && !(entity instanceof BattleArmor)
            && (null != entity.getCrew())) {
        
            entity.getCrew().setEjected(false);
            entity.getCrew().setDead(true);
        }

        if (!entity.isDoomed() && !entity.isDestroyed()) {
            r = new Report(6365);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(reason);
            vDesc.addElement(r);

            entity.setDoomed(true);

            Enumeration<Integer> iter = entity.getPickedUpMechWarriors()
                                              .elements();
            while (iter.hasMoreElements()) {
                int mechWarriorId = iter.nextElement();
                Entity mw = game.getEntity(mechWarriorId);
                mw.setDestroyed(true);
                
                game.removeEntity(mw.getId(), condition);
                entityUpdate(mw.getId());
                send(createRemoveEntityPacket(mw.getId(), condition));
                r = new Report(6370);
                r.subject = mw.getId();
                r.addDesc(mw);
                vDesc.addElement(r);
            }

            for (int missileId : entity.getTMTracker().getMissiles()) {
                Entity tm = game.getEntity(missileId);
                if ((null != tm) && !tm.isDestroyed()
                    && (tm instanceof TeleMissile)) {
                    ((TeleMissile) tm).setOutContact(true);
                    entityUpdate(tm.getId());
                }
            }

            ArrayList<Entity> externalUnits = entity.getExternalUnits();

            if (entity.getLoadedUnits().size() > 0) {
                Coords curPos = entity.getPosition();
                int curFacing = entity.getFacing();
                for (Entity other : entity.getLoadedUnits()) {
                    
                    boolean survived = false;
                    if (entity instanceof Tank) {
                        if ((entity.getMovementMode() == EntityMovementMode.NAVAL)
                            || (entity.getMovementMode() == EntityMovementMode.HYDROFOIL)) {
                            if (other.getMovementMode() == EntityMovementMode.INF_UMU) {
                                survived = Compute.d6() <= 3;
                            } else if (other.getMovementMode() == EntityMovementMode.INF_JUMP) {
                                survived = Compute.d6() == 1;
                            } else if (other.getMovementMode() == EntityMovementMode.VTOL) {
                                survived = Compute.d6() <= 2;
                            }
                        } else if (entity.getMovementMode() == EntityMovementMode.SUBMARINE) {
                            if (other.getMovementMode() == EntityMovementMode.INF_UMU) {
                                survived = Compute.d6() == 1;
                            }
                        } else {
                            survived = Compute.d6() <= 4;
                        }
                    } else if (entity instanceof Mech) {
                        
                        if (externalUnits.contains(other)) {
                            survived = Compute.d6() < 3;
                        }
                    }
                    if (!survivable
                        || (externalUnits.contains(other) && !survived)) {
                        
                        other.setDestroyed(true);
                        
                        entity.unload(other);
                        
                        game.moveToGraveyard(other.getId());
                        send(createRemoveEntityPacket(other.getId(), condition));
                        r = new Report(6370);
                        r.subject = other.getId();
                        r.addDesc(other);
                        vDesc.addElement(r);
                    }
                   
                    else if ((null != Compute.stackingViolation(game,
                                                                other.getId(), curPos))
                             || other.isLocationProhibited(curPos)) {
                        
                        other.setDestroyed(true);
                        
                        entity.unload(other);
                        
                        game.moveToGraveyard(other.getId());
                        send(createRemoveEntityPacket(other.getId(), condition));
                        r = new Report(6375);
                        r.subject = other.getId();
                        r.addDesc(other);
                        vDesc.addElement(r);
                    } 
                    else {
                        
                        unloadUnit(entity, other, curPos, curFacing,
                                   entity.getElevation(), true);
                    }
                } 
            } 

            if (Entity.NONE != entity.getTransportId()) {
                final Entity transport = game
                        .getEntity(entity.getTransportId());
                Coords curPos = transport.getPosition();
                int curFacing = transport.getFacing();
                unloadUnit(transport, entity, curPos, curFacing,
                           transport.getElevation());
                entityUpdate(transport.getId());

                if ((transport instanceof FighterSquadron)
                    && (((FighterSquadron) transport).getFighters().size() < 1)) {
                    transport.setDestroyed(true);
                  
                    r = new Report(6365);
                    r.subject = transport.getId();
                    r.addDesc(transport);
                    r.add("fighter destruction");
                    vDesc.addElement(r);
                }
            } 

            final int swarmerId = entity.getSwarmAttackerId();
            if (Entity.NONE != swarmerId) {
                final Entity swarmer = game.getEntity(swarmerId);

                swarmer.setSwarmTargetId(Entity.NONE);
                
                swarmer.setUnloaded(false);
                swarmer.setDone(false);
                entity.setSwarmAttackerId(Entity.NONE);
                Report.addNewline(vDesc);
                r = new Report(6380);
                r.subject = swarmerId;
                r.addDesc(swarmer);
                vDesc.addElement(r);
                
                entityUpdate(swarmerId);
            }

            final int swarmedId = entity.getSwarmTargetId();
            if (Entity.NONE != swarmedId) {
                final Entity swarmed = game.getEntity(swarmedId);
                swarmed.setSwarmAttackerId(Entity.NONE);
                entity.setSwarmTargetId(Entity.NONE);
                r = new Report(6385);
                r.subject = swarmed.getId();
                r.addDesc(swarmed);
                vDesc.addElement(r);
                entityUpdate(swarmedId);
            }

            if (entity.getGrappled() != Entity.NONE) {
                int grappler = entity.getGrappled();
                entity.setGrappled(Entity.NONE, false);
                Entity e = game.getEntity(grappler);
                if (e != null) {
                    e.setGrappled(Entity.NONE, false);
                }
                entityUpdate(grappler);
            }
        } 

        Coords curPos = entity.getPosition();
        IHex entityHex = game.getBoard().getHex(curPos);
        if (game.getOptions().booleanOption("tacops_battle_wreck")
            && (entityHex != null)
            && game.getBoard().onGround()
            && !((entity instanceof Infantry) || (entity instanceof Protomech))) {
            
            if (entity instanceof LargeSupportTank) {
                if (entityHex.terrainLevel(Terrains.ROUGH) < 2) {
                    entityHex.addTerrain(Terrains.getTerrainFactory()
                                                 .createTerrain(Terrains.ROUGH, 2));
                    sendChangedHex(curPos);
                }
            } else if ((entity.getWeight() >= 40)
                       && !entityHex.containsTerrain(Terrains.ROUGH)) {
                entityHex.addTerrain(Terrains.getTerrainFactory()
                                             .createTerrain(Terrains.ROUGH, 1));
                sendChangedHex(curPos);
            }

        }

        entityUpdate(entity.getId());

        return vDesc;
    }

    private Vector<Report> explodeEquipment(Entity en, int loc, int slot) {
        CriticalSlot critSlot = en.getCritical(loc, slot);
        Vector<Report> reports = explodeEquipment(en, loc, critSlot.getMount());
        if (critSlot.getMount2() != null) {
            reports.addAll(explodeEquipment(en, loc, critSlot.getMount2()));
        }
        return reports;
    }

    public Vector<Report> explodeEquipment(Entity en, int loc, Mounted mounted) {
        Vector<Report> vDesc = new Vector<Report>();
        
        if (mounted.isDestroyed()) {
            System.err.println("server: explodeEquipment called on destroyed"
                               + " equipment (" + mounted.getName() + ')');
            return vDesc;
        }

        if (((mounted.getType() instanceof MPodWeapon) || (mounted.getType() instanceof BPodWeapon))
            && (mounted.getLinked().getHittableShotsLeft() == 0)) {
            return vDesc;
        }

        if ((mounted.getType() instanceof WeaponType)
            && (((WeaponType) mounted.getType()).getAmmoType() == AmmoType.T_AC_ROTARY)) {
            if (!mounted.isJammed()) {
                return vDesc;
            }
        }

        
        if ((mounted.getType() instanceof WeaponType)
            && !mounted.curMode().equals("Rapid")
            && ((((WeaponType) mounted.getType()).getAmmoType() == AmmoType.T_AC) || (((WeaponType) mounted
                .getType()).getAmmoType() == AmmoType.T_LAC))) {
            if (!mounted.isUsedThisRound()) {
                return vDesc;
            }
            Mounted ammo = mounted.getLinked();
            if ((ammo == null)
                || !(ammo.getType() instanceof AmmoType)
                || (((AmmoType) ammo.getType()).getMunitionType() != AmmoType.M_INCENDIARY_AC)) {
                return vDesc;
            }

            WeaponType wtype = (WeaponType) mounted.getType();
            if ((wtype.getAmmoType() == AmmoType.T_LRM)
                || (wtype.getAmmoType() == AmmoType.T_LRM_STREAK)
                || (wtype.getAmmoType() == AmmoType.T_LRM_TORPEDO)
                || (wtype.getAmmoType() == AmmoType.T_LRM_TORPEDO_COMBO)) {
                return vDesc;
            }
        }

        if ((mounted.getType() instanceof HVACWeapon)
            && !mounted.curMode().equals("Rapid")) {
            if (mounted.getEntity().getTotalAmmoOfType(
                    mounted.getLinked().getType()) == 0) {
                return vDesc;
            }
        }

        if ((mounted.getType() instanceof MiscType)
            && (mounted.getType().hasFlag(MiscType.F_BLUE_SHIELD) && mounted
                .curMode().equals("Off"))) {
            return vDesc;
        }

        if ((mounted.getType() instanceof MiscType)
            && mounted.getType().hasFlag(MiscType.F_PPC_CAPACITOR)
            && !mounted.curMode().equals("Charge")) {
            return vDesc;
        }
        if ((mounted.getType() instanceof PPCWeapon)
            && (mounted.hasChargedCapacitor() == 0)) {
            return vDesc;
        }

        if ((mounted.getType() instanceof AmmoType)
            && ((((AmmoType) mounted.getType()).getAmmoType() == AmmoType.T_SRM)
                || (((AmmoType) mounted.getType()).getAmmoType() == AmmoType.T_IATM) || (((AmmoType) mounted
                .getType()).getAmmoType() == AmmoType.T_MML))
            && (((AmmoType) mounted.getType()).getMunitionType() == AmmoType.M_INFERNO)
            && (mounted.getHittableShotsLeft() > 0)) {
            en.heatBuildup += Math.min(mounted.getExplosionDamage(), 30);
        }

        int damage = mounted.getExplosionDamage();

        if ((mounted.getType() instanceof AmmoType)
            && ((((AmmoType) mounted.getType()).getAmmoType() == AmmoType.T_SRM) || (((AmmoType) mounted
                .getType()).getAmmoType() == AmmoType.T_LRM))
            && (((AmmoType) mounted.getType()).getMunitionType() == AmmoType.M_SMOKE_WARHEAD)
            && (mounted.getHittableShotsLeft() > 0)) {
            damage = ((mounted.getExplosionDamage()) / 2);
        }
        
        if ((mounted.getType() instanceof AmmoType)
            && ((((AmmoType) mounted.getType()).getAmmoType() == AmmoType.T_VEHICLE_FLAMER) || (((AmmoType) mounted
                .getType()).getAmmoType() == AmmoType.T_HEAVY_FLAMER))
            && (((AmmoType) mounted.getType()).getMunitionType() == AmmoType.M_COOLANT)
            && (mounted.getHittableShotsLeft() > 0)) {
            damage = 2;
            en.coolFromExternal += 3;
        }

        if (en instanceof Aero) {
            int newdamage = (int) Math.floor(damage / 10.0);
            if ((newdamage == 0) && (damage > 0)) {
                damage = 1;
            } else {
                damage = newdamage;
            }
        }

        if (damage <= 0) {
            return vDesc;
        }

        Report r = new Report(6390);
        r.subject = en.getId();
        r.add(mounted.getName());
        r.add(damage);
        r.indent(3);
        vDesc.addElement(r);
        
        if ((mounted.getType() instanceof WeaponType) && mounted.isHotLoaded()) {
            int roll = Compute.d6(2);
            int ammoExploded = 0;
            r = new Report(6077);
            r.subject = en.getId();
            r.add(roll);
            r.indent(2);
            vDesc.addElement(r);

            if (roll < 6) {
                for (Mounted ammo : en.getAmmo()) {
                    if ((ammo.getLocation() == loc)
                        && (ammo.getExplosionDamage() > 0)
                     
                        && (((AmmoType) ammo.getType()).getMunitionType() != AmmoType.M_DEAD_FIRE)) {
                        ammoExploded++;
                        vDesc.addAll(this.explodeEquipment(en, loc, ammo));
                        break;
                    }
                }
                if (ammoExploded == 0) {
                    r = new Report(6078);
                    r.subject = en.getId();
                    r.indent(2);
                    vDesc.addElement(r);
                }
            } else {
                r = new Report(6079);
                r.subject = en.getId();
                r.indent(2);
                vDesc.addElement(r);
            }
        }

        HitData hit = new HitData(loc);
        
        
        if (mounted.getType() instanceof AmmoType) {
            if (((AmmoType) mounted.getType()).isCapital()) {
                hit.setCapital(true);
            }
        }
        
        if ((mounted.getType() instanceof MiscType)  && mounted.getType().hasFlag(MiscType.F_RISC_LASER_PULSE_MODULE)) {
            hit.setEffect(HitData.EFFECT_NO_CRITICALS);
            Mounted laser = mounted.getLinkedBy();
            if (en instanceof Mech) {
                for (int slot = 0; slot < en.getNumberOfCriticals(laser.getLocation()); slot++) {
                    CriticalSlot cs = en.getCritical(laser.getLocation(), slot);
                    if ((cs.getType() == CriticalSlot.TYPE_EQUIPMENT) && cs.getMount().equals(laser) && cs.isHittable()) {
                        cs.setHit(true);
                        cs.setRepairable(true);
                        break;
                    }
                }
            }
            laser.setHit(true);
        }

        mounted.setShotsLeft(0);
        Vector<Report> newReports = damageEntity(en, hit, damage, true);
        for (Report rep : newReports) {
            rep.indent(2);
        }
        vDesc.addAll(newReports);
        Report.addNewline(vDesc);

        int pilotDamage = 2;
        if (en instanceof Aero) {
            pilotDamage = 1;
        }
        if (game.getOptions().booleanOption("case_pilot_damage")
            && (en.locationHasCase(hit.getLocation()) || en.hasCASEII(hit
                                                                              .getLocation()))) {
            pilotDamage = 1;
        }
        if (en.getCrew().getOptions().booleanOption("pain_resistance")
            || en.getCrew().getOptions().booleanOption("iron_man")) {
            pilotDamage -= 1;
        }
        
        if ((en instanceof Tank)
            && !(en.getCrew().getOptions().booleanOption("vdni") || en
                .getCrew().getOptions().booleanOption("bvdni"))) {
            pilotDamage = 0;
        }
        if (!en.getCrew().getOptions().booleanOption("pain_shunt")) {
            vDesc.addAll(damageCrew(en, pilotDamage));
        }
        if (en.getCrew().isDoomed() || en.getCrew().isDead()) {
            vDesc.addAll(destroyEntity(en, "crew death", true));
        } else {
            Report.addNewline(vDesc);
        }

        return vDesc;
    }

    private Vector<Report> explodeAmmoFromHeat(Entity entity) {
        int damage = 0;
        int rack = 0;
        int boomloc = -1;
        int boomslot = -1;
        Vector<Report> vDesc = new Vector<Report>();

        for (int j = 0; j < entity.locations(); j++) {
            for (int k = 0; k < entity.getNumberOfCriticals(j); k++) {
                CriticalSlot cs = entity.getCritical(j, k);
                if ((cs == null) || cs.isDestroyed() || cs.isHit()
                    || (cs.getType() != CriticalSlot.TYPE_EQUIPMENT)) {
                    continue;
                }
                Mounted mounted = cs.getMount();
                if ((mounted == null)
                    || (!(mounted.getType() instanceof AmmoType))) {
                    continue;
                }
                AmmoType atype = (AmmoType) mounted.getType();
                if (!atype.isExplosive(mounted)) {
                    continue;
                }
                
                if ((atype.getAmmoType() == AmmoType.T_COOLANT_POD)
                        || (((atype.getAmmoType() == AmmoType.T_VEHICLE_FLAMER)
                                || (atype.getAmmoType() == AmmoType.T_HEAVY_FLAMER))
                                && (atype.getMunitionType() == AmmoType.M_COOLANT))) {
                    continue;
                }
                
                if (mounted.getHittableShotsLeft() == 0) {
                    continue;
                }
                
                int newRack = atype.getDamagePerShot() * atype.getRackSize();
                int newDamage = mounted.getExplosionDamage();
                Mounted mount2 = cs.getMount2();
                if ((mount2 != null) && (mount2.getType() instanceof AmmoType)
                    && (mount2.getHittableShotsLeft() > 0)) {
                    
                    atype = (AmmoType) mount2.getType();
                    newRack += atype.getDamagePerShot() * atype.getRackSize();
                    newDamage += mount2.getExplosionDamage();
                }
                if (!mounted.isHit()
                    && ((rack < newRack) || ((rack == newRack) && (damage < newDamage)))) {
                    rack = newRack;
                    damage = newDamage;
                    boomloc = j;
                    boomslot = k;
                }
            }
        }
        if ((boomloc != -1) && (boomslot != -1)) {
            CriticalSlot slot = entity.getCritical(boomloc, boomslot);
            slot.setHit(true);
            slot.getMount().setHit(true);
            if (slot.getMount2() != null) {
                slot.getMount2().setHit(true);
            }
            vDesc.addAll(explodeEquipment(entity, boomloc, boomslot));
        } else {
            
            Report r = new Report(5105);
            r.subject = entity.getId();
            r.indent();
            vDesc.addElement(r);
        }
        return vDesc;
    }

    private Vector<Report> doEntityFall(Entity entity, Coords fallPos,
            int fallHeight, int facing, PilotingRollData roll,
            boolean intoBasement) {
        entity.setFallen(true);

        Vector<Report> vPhaseReport = new Vector<Report>();
        Report r;

        IHex fallHex = game.getBoard().getHex(fallPos);

        boolean handlingBasement = false;
        int damageTable = ToHitData.HIT_NORMAL;

        
        if (entity.isMakingDfa()) {
            r = new Report(2305);
            r.subject = entity.getId();
            vPhaseReport.add(r);
            entity.setProne(true);
            return vPhaseReport;
        }

        String side;
        int table;
        switch (facing) {
            case 1:
            case 2:
                side = "right side";
                table = ToHitData.SIDE_RIGHT;
                break;
            case 3:
                side = "rear";
                table = ToHitData.SIDE_REAR;
                break;
            case 4:
            case 5:
                side = "left side";
                table = ToHitData.SIDE_LEFT;
                break;
            case 0:
            default:
                side = "front";
                table = ToHitData.SIDE_FRONT;
        }

        int waterDepth = 0;
        if (fallHex.containsTerrain(Terrains.WATER)) {
            
            waterDepth = fallHex.terrainLevel(Terrains.WATER);
        }
        boolean fallOntoBridge = false;
        
        if ((entity.climbMode() && (entity.getPosition() != fallPos)
             && fallHex.containsTerrain(Terrains.BRIDGE) && fallHex
                .containsTerrainExit(Terrains.BRIDGE,
                                     fallPos.direction(entity.getPosition())))
            || (entity.getElevation() == fallHex
                .terrainLevel(Terrains.BRIDGE_ELEV))) {
            fallOntoBridge = true;
        }
        int bridgeElev = fallHex.terrainLevel(Terrains.BRIDGE_ELEV);
        int buildingElev = fallHex.terrainLevel(Terrains.BLDG_ELEV);
        int damageHeight = fallHeight;
        int newElevation = 0;

        boolean checkCollapse = false;

        if ((entity.relHeight() >= buildingElev)
            && (buildingElev >= 0)) {
            
            newElevation = buildingElev;
            checkCollapse = true;
        } else if (fallOntoBridge && (entity.relHeight() >= bridgeElev)
                && (bridgeElev >= 0)) {
            
            waterDepth = 0;
            newElevation = fallHex.terrainLevel(Terrains.BRIDGE_ELEV);
            checkCollapse = true;
        } else if (fallHex.containsTerrain(Terrains.ICE)
                   && (entity.getElevation() == 0)) {
            waterDepth = 0;
            newElevation = 0;
            
        } else if ((entity.getMovementMode() != EntityMovementMode.VTOL)
                   && (game.getBoard().getBuildingAt(fallPos) != null)) {
            newElevation = entity.getElevation();
        }
        
        else if (waterDepth > 0) {
            damageHeight = fallHeight - waterDepth;
            newElevation = -waterDepth;
        }
        
        if (intoBasement) {
            Building bldg = game.getBoard().getBuildingAt(fallPos);
            if ((bldg.getBasement(fallPos) != BasementType.NONE)
                && (bldg.getBasement(fallPos) != BasementType.ONE_DEEP_NORMALINFONLY)
                && (entity.getElevation() == 0)
                && (bldg.getBasementCollapsed(fallPos) == true)) {

                if (fallHex.depth(true) == 0) {
                    System.err.println(" Entity " + entity.getDisplayName()
                                       + " is falling into a depth " + fallHex.depth(true)
                                       + " basement -- not allowed!!");
                    return vPhaseReport;
                }
                damageHeight = bldg.getBasement(fallPos).getDepth();

                newElevation = newElevation - damageHeight;

                handlingBasement = true;
                if ((bldg.getBasement(fallPos) == BasementType.TWO_DEEP_FEET)
                    || (bldg.getBasement(fallPos) == BasementType.ONE_DEEP_FEET)) {
                    damageTable = ToHitData.HIT_KICK;
                } else if ((bldg.getBasement(fallPos) == BasementType.TWO_DEEP_HEAD)
                           || (bldg.getBasement(fallPos) == BasementType.ONE_DEEP_HEAD)) {
                    damageTable = ToHitData.HIT_PUNCH;
                } else {
                    damageTable = ToHitData.HIT_NORMAL;
                }
            }
        }
        if (entity instanceof Protomech) {
            damageTable = ToHitData.HIT_SPECIAL_PROTO;
        }
        
        if ((waterDepth > 0)
            && !(entity instanceof Mech)
            && !(entity instanceof Protomech)
            && !((entity.getRunMP() > 0) && (entity.getMovementMode() == EntityMovementMode.HOVER))
            && (entity.getMovementMode() != EntityMovementMode.HYDROFOIL)
            && (entity.getMovementMode() != EntityMovementMode.NAVAL)
            && (entity.getMovementMode() != EntityMovementMode.SUBMARINE)
            && (entity.getMovementMode() != EntityMovementMode.INF_UMU)) {
            vPhaseReport.addAll(destroyEntity(entity, "a watery grave", false));
            return vPhaseReport;
        }

        if (entity instanceof Mech) {
            Mech mech = (Mech) entity;
            mech.setLevelsFallen(damageHeight + waterDepth + 1);
            
            if (mech.isIndustrial()) {
                mech.setCheckForCrit(true);
            }
        }

        int damage = (int) Math.round(entity.getWeight() / 10.0)
                     * (damageHeight + 1);
        
        if (entity instanceof Infantry) {
            damage = (int) Math.ceil(damageHeight / 2.0);
            
            if (damageHeight < 2) {
                damage = 0;
            }
            if (!(entity instanceof BattleArmor)) {
                int dice = 3;
                if (entity.getMovementMode() == EntityMovementMode.INF_MOTORIZED) {
                    dice = 2;
                } else if ((entity.getMovementMode() == EntityMovementMode.INF_JUMP)
                           || ((Infantry) entity).isMechanized()) {
                    dice = 1;
                }
                damage = damage * Compute.d6(dice);
            }
        }
        
        int waterDamage = 0;
        if (waterDepth > 0) {
            damage /= 2;
            waterDamage = ((int) Math.round(entity.getWeight() / 10.0) * (waterDepth + 1)) / 2;
        }

        if ((waterDepth >= fallHeight) && ((waterDepth != 0) || (fallHeight != 0))) {
            damage = 0;
            waterDamage = ((int) Math.round(entity.getWeight() / 10.0) * (fallHeight + 1)) / 2;
        }
        
        damage = Math
                .round(damage * game.getPlanetaryConditions().getGravity());
        waterDamage = Math.round(waterDamage
                                 * game.getPlanetaryConditions().getGravity());

        if (waterDamage == 0) {
            r = new Report(2310);
            r.subject = entity.getId();
            r.indent();
            r.addDesc(entity);
            r.add(side); 
            r.add(damage);
        } else if (damage > 0) {
            r = new Report(2315);
            r.subject = entity.getId();
            r.indent();
            r.addDesc(entity);
            r.add(side); 
            r.add(damage);
            r.add(waterDamage);
        } else {
            r = new Report(2310);
            r.subject = entity.getId();
            r.indent();
            r.addDesc(entity);
            r.add(side); 
            r.add(waterDamage);
        }
        vPhaseReport.add(r);

        final int swarmerId = entity.getSwarmAttackerId();

        if (entity instanceof Mech) {
            entity.setProne(true);
        }
        entity.setPosition(fallPos);
        entity.setFacing((entity.getFacing() + (facing)) % 6);
        entity.setSecondaryFacing(entity.getFacing());
        entity.setElevation(newElevation);

        if (fallHex.getBogDownModifier(entity.getMovementMode(),
                                       entity instanceof LargeSupportTank) != TargetRoll.AUTOMATIC_SUCCESS) {
            entity.setStuck(true);
            r = new Report(2081);
            r.subject = entity.getId();
            r.add(entity.getDisplayName(), true);
            vPhaseReport.add(r);
            
            vPhaseReport.addAll(checkQuickSand(fallPos));
        }

        if ((entity instanceof Infantry) && (damage > 0)) {
            if (entity instanceof BattleArmor) {
                for (int i = 1; i < entity.locations(); i++) {
                    HitData h = new HitData(i);
                    vPhaseReport.addAll(damageEntity(entity, h, damage));
                    addNewLines();
                }
            } else {
                HitData h = new HitData(Infantry.LOC_INFANTRY);
                vPhaseReport.addAll(damageEntity(entity, h, damage));
            }
        } else {
            while (damage > 0) {
                int cluster = Math.min(5, damage);
                HitData hit = entity.rollHitLocation(damageTable, table);
                hit.makeFallDamage(true);
                vPhaseReport.addAll(damageEntity(entity, hit, cluster));
                damage -= cluster;
            }
        }

        if (waterDepth > 0) {
            for (int loop = 0; loop < entity.locations(); loop++) {
                entity.setLocationStatus(loop, ILocationExposureStatus.WET);
            }
        }
        
        while (waterDamage > 0) {
            int cluster = Math.min(5, waterDamage);
            HitData hit = entity.rollHitLocation(damageTable, table);
            hit.makeFallDamage(true);
            vPhaseReport.addAll(damageEntity(entity, hit, cluster));
            waterDamage -= cluster;
        }

        vPhaseReport.addAll(doSetLocationsExposure(entity, fallHex, false,
                                                   -waterDepth));

        
        if ((entity instanceof Mech)
            && !entity.getCrew().getOptions().booleanOption("dermal_armor")
            && !entity.getCrew().getOptions().booleanOption("tsm_implant")) {
            
            roll.removeAutos();

            if (fallHeight > 1) {
                roll.addModifier(fallHeight - newElevation - 1, "height of fall");
            }

            if (roll.getValue() == TargetRoll.IMPOSSIBLE) {
                r = new Report(2320);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(entity.getCrew().getName());
                r.indent();
                vPhaseReport.add(r);
                vPhaseReport.addAll(damageCrew(entity, 1));
                Report.addNewline(vPhaseReport);
            } else {
                int diceRoll = Compute.d6(2);
                r = new Report(2325);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.add(entity.getCrew().getName());
                r.add(roll.getValueAsString());
                r.add(diceRoll);
                if (diceRoll >= roll.getValue()) {
                    r.choose(true);
                    vPhaseReport.add(r);
                } else {
                    r.choose(false);
                    vPhaseReport.add(r);
                    vPhaseReport.addAll(damageCrew(entity, 1));
                    Report.addNewline(vPhaseReport);
                }
            }
        }

        if (Entity.NONE != swarmerId) {
            final Entity swarmer = game.getEntity(swarmerId);
            entity.setSwarmAttackerId(Entity.NONE);
            swarmer.setSwarmTargetId(Entity.NONE);
            
            if ((waterDepth > 0)
                && (swarmer.getMovementMode() != EntityMovementMode.INF_UMU)) {
                
                swarmer.setPosition(fallPos);
                r = new Report(2330);
                r.newlines = 0;
                r.subject = swarmer.getId();
                r.addDesc(swarmer);
                vPhaseReport.add(r);
                vPhaseReport.addAll(destroyEntity(swarmer, "a watery grave",
                                                  false));
            } else {
               
                r = new Report(2335);
                r.newlines = 0;
                r.subject = swarmer.getId();
                r.addDesc(swarmer);
                vPhaseReport.add(r);
                vPhaseReport.addAll(damageEntity(swarmer, swarmer
                        .rollHitLocation(ToHitData.HIT_NORMAL,
                                         ToHitData.SIDE_FRONT), Compute.d6(2)));
                Report.addNewline(vPhaseReport);
            }
            swarmer.setPosition(fallPos);
            entityUpdate(swarmerId);
            if (!swarmer.isDone()) {
                game.removeTurnFor(swarmer);
                swarmer.setDone(true);
                send(createTurnVectorPacket());
            }
        } 

        game.resetPSRs(entity);

        if (game.containsMinefield(fallPos)
            && enterMinefield(entity, fallPos, newElevation, true,
                              vPhaseReport, 12)) {
            resetMines();
        }
        
        if (checkCollapse && !handlingBasement) {

            checkForCollapse(game.getBoard().getBuildingAt(fallPos),
                             game.getPositionMap(), fallPos, false, vPhaseReport);
        }

        return vPhaseReport;
    }

    private Vector<Report> doEntityFall(Entity entity, Coords fallPos,
                                        int height, PilotingRollData roll) {
        return doEntityFall(entity, fallPos, height, Compute.d6(1) - 1, roll,
                            false);
    }

    private Vector<Report> doEntityFall(Entity entity, PilotingRollData roll) {
        boolean fallToSurface = false;
        
        int toSubtract = 0;
        IHex currHex = game.getBoard().getHex(entity.getPosition());
        if (currHex.containsTerrain(Terrains.ICE)
            && (entity.getElevation() != -currHex.depth())) {
            fallToSurface = true;
            toSubtract = 0;
        }
        
        if (currHex.containsTerrain(Terrains.BRIDGE_ELEV)
            && (entity.getElevation() >= currHex
                .terrainLevel(Terrains.BRIDGE_ELEV))) {
            fallToSurface = true;
            toSubtract = currHex.terrainLevel(Terrains.BRIDGE_ELEV);
        }
        
        if (currHex.containsTerrain(Terrains.BLDG_ELEV)
            && (entity.getElevation() >= currHex
                .terrainLevel(Terrains.BLDG_ELEV))) {
            fallToSurface = true;
            toSubtract = currHex.terrainLevel(Terrains.BLDG_ELEV);
        }
        return doEntityFall(entity, entity.getPosition(), entity.getElevation()
                + (!fallToSurface ? currHex.depth(true) : -toSubtract), roll);
    }

    private void resolveAeroElevationLoss() {
        Report r;
        for (Entity entity : game.getEntitiesVector()) {
            if (entity instanceof Aero) {
                Aero a = (Aero) entity;
                if (a.getAltLoss() > 0) {
                    r = new Report(9095);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(a.getAltLoss());
                    addReport(r);
                    a.setAltitude(a.getAltitude() - a.getAltLoss());
                    a.resetAltLoss();
                    entityUpdate(entity.getId());
                }
            }
        }
    }

    private void resolveAmmoDumps() {
        Report r;
        for (Entity entity : game.getEntitiesVector()) {
            for (Mounted m : entity.getAmmo()) {
                if (m.isPendingDump()) {
                    
                    r = new Report(5110);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(m.getName());
                    addReport(r);
                    
                    m.setPendingDump(false);
                    m.setDumping(true);
                } else if (m.isDumping()) {
                    
                    r = new Report(5115);
                    r.subject = entity.getId();
                    r.addDesc(entity);
                    r.add(m.getName());
                    addReport(r);
                    
                    m.setDumping(false);
                    m.setShotsLeft(0);
                }
            }
            
            if (entity instanceof BattleArmor) {
                for (Mounted m : entity.getWeaponList()) {
                    if (m.isDWPMounted() && m.isPendingDump()) {
                        m.setMissing(true);
                        r = new Report(5116);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        r.add(m.getName());
                        addReport(r);
                        m.setPendingDump(false);
                        
                        for (Mounted ammo : entity.getAmmo()) {
                            if (m.equals(ammo.getLinkedBy())) {
                                ammo.setMissing(true);
                            }
                        }
                    
                    } else if (m.isBodyMounted() && m.isPendingDump()
                            && m.getType().hasFlag(WeaponType.F_MISSILE)
                            && (m.getLinked() != null)
                            && (m.getLinked().getUsableShotsLeft() > 0)) {
                        m.setMissing(true);
                        r = new Report(5116);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        r.add(m.getName());
                        addReport(r);
                        m.setPendingDump(false);
                       
                        while ((m.getLinked() != null)
                                && (m.getLinked().getUsableShotsLeft() > 0)) {
                            m.getLinked().setMissing(true);
                            entity.loadWeapon(m);
                        }
                    }
                }
            }
            entity.reloadEmptyWeapons();
        }
    }

    public boolean checkIgnition(Coords c, TargetRoll roll, boolean bInferno,
                                 int entityId, Vector<Report> vPhaseReport) {

        IHex hex = game.getBoard().getHex(c);

        if (null == hex) {
            return false;
        }

        if (hex.containsTerrain(Terrains.FIRE)) {
            return false;
        }

        if (!bInferno && !hex.isIgnitable()) {
            return false;
        }

        int fireRoll = Compute.d6(2);
        Report r = null;
        if (entityId != Entity.NONE) {
            r = new Report(3430);
            r.indent(2);
            r.subject = entityId;
            r.add(roll.getValueAsString());
            r.add(roll.getDesc());
            r.add(fireRoll);
            vPhaseReport.add(r);
        }
        if (fireRoll >= roll.getValue()) {
            ignite(c, Terrains.FIRE_LVL_NORMAL, vPhaseReport);
            return true;
        }
        return false;
    }

    public boolean checkIgnition(Coords c, TargetRoll roll, boolean bInferno) {
        return checkIgnition(c, roll, bInferno, Entity.NONE, null);
    }

    public boolean checkIgnition(Coords c, TargetRoll roll) {
        
        return checkIgnition(c, roll, false, Entity.NONE, null);
    }

    public void ignite(Coords c, int fireLevel, Vector<Report> vReport) {
        
        if (null != game.getPlanetaryConditions().cannotStartFire()) {
            if (null != vReport) {
                Report r = new Report(3007);
                r.indent(2);
                r.add(game.getPlanetaryConditions().cannotStartFire());
                r.type = Report.PUBLIC;
                vReport.add(r);
            }
            return;
        }

        if (!game.getOptions().booleanOption("tacops_start_fire")) {
            if (null != vReport) {
                Report r = new Report(3008);
                r.indent(2);
                r.type = Report.PUBLIC;
                vReport.add(r);
            }
            return;
        }

        IHex hex = game.getBoard().getHex(c);
        if (null == hex) {
            return;
        }

        Report r = new Report(3005);
        r.indent(2);
        r.add(c.getBoardNum());
        r.type = Report.PUBLIC;

        switch (fireLevel) {
            case Terrains.FIRE_LVL_INFERNO:
                r.messageId = 3006;
                break;
            case Terrains.FIRE_LVL_INFERNO_BOMB:
                r.messageId = 3003;
                break;
            case Terrains.FIRE_LVL_INFERNO_IV:
                r.messageId = 3004;
                break;
        }

        if (null != vReport) {
            vReport.add(r);
        }
        hex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                Terrains.FIRE, fireLevel));
        sendChangedHex(c);
    }

    
    public void removeFire(Coords fireCoords, String reason) {
        IHex hex = game.getBoard().getHex(fireCoords);
        if (null == hex) {
            return;
        }
        hex.removeTerrain(Terrains.FIRE);
        hex.resetFireTurn();
        sendChangedHex(fireCoords);
        
        Report r = new Report(5170, Report.PUBLIC);
        r.add(fireCoords.getBoardNum());
        r.add(reason);
        addReport(r);
    }

    
    public void addSmoke(ArrayList<Coords> coords, int windDir, boolean bInferno) {

        
        if (game.getPlanetaryConditions().getWindStrength() > PlanetaryConditions.WI_STORM) {
            return;
        }

        int smokeLevel = 0;
        for (Coords smokeCoords : coords) {
            IHex smokeHex = game.getBoard().getHex(smokeCoords);
            Report r;
            if (smokeHex == null) {
                continue;
            }
            
            if (bInferno
                || (Building.MEDIUM < smokeHex
                    .terrainLevel(Terrains.FUEL_TANK))
                || (Building.MEDIUM < smokeHex
                    .terrainLevel(Terrains.BUILDING))) {
                if (smokeHex.terrainLevel(Terrains.SMOKE)
                    == SmokeCloud.SMOKE_HEAVY) {
                    
                    r = new Report(5180, Report.PUBLIC);
                    r.add(smokeCoords.getBoardNum());
                    addReport(r);
                    smokeLevel = SmokeCloud.SMOKE_HEAVY;
                } else {
                    smokeLevel = Math.max(smokeLevel, SmokeCloud.SMOKE_HEAVY);
                    r = new Report(5185, Report.PUBLIC);
                    r.add(smokeCoords.getBoardNum());
                    addReport(r);
                    smokeLevel = SmokeCloud.SMOKE_HEAVY;
                }
            } else {
                if (smokeHex.terrainLevel(Terrains.SMOKE)
                    == SmokeCloud.SMOKE_HEAVY) {
                    
                    r = new Report(5190, Report.PUBLIC);
                    r.add(smokeCoords.getBoardNum());
                    smokeLevel = Math.max(smokeLevel, SmokeCloud.SMOKE_LIGHT);
                    addReport(r);
                } else if (smokeHex.terrainLevel(Terrains.SMOKE)
                           == SmokeCloud.SMOKE_LIGHT) {
                    
                    r = new Report(5195, Report.PUBLIC);
                    r.add(smokeCoords.getBoardNum());
                    addReport(r);
                    smokeLevel = Math.max(smokeLevel, SmokeCloud.SMOKE_LIGHT);
                } else {
                    smokeLevel = Math.max(smokeLevel, SmokeCloud.SMOKE_LIGHT);
                    
                    r = new Report(5200, Report.PUBLIC);
                    r.add(smokeCoords.getBoardNum());
                    addReport(r);
                }
            }
        }
        createSmoke(coords, smokeLevel, 0);
    }

    
    private ArrayList<String> scanForBoardsInDir(final File boardDir,
                                                 final String basePath, final BoardDimensions dimensions,
                                                 ArrayList<String> boards) {
        if (boardDir == null) {
            throw new IllegalArgumentException("must provide searchDir");
        }

        if (basePath == null) {
            throw new IllegalArgumentException("must provide basePath");
        }

        if (dimensions == null) {
            throw new IllegalArgumentException("must provide dimensions");
        }

        if (boards == null) {
            throw new IllegalArgumentException("must provide boards");
        }

        String fileList[] = boardDir.list();
        for (String filename : fileList) {
            File filepath = new File(boardDir, filename);
            if (filepath.isDirectory()) {
                scanForBoardsInDir(new File(boardDir, filename), basePath
                                           .concat(File.separator).concat(filename), dimensions,
                                   boards);
            } else {
                if (filename.endsWith(".board")) { 
                    if (Board.boardIsSize(filepath, dimensions)) {
                        boards.add(basePath.concat(File.separator)
                                           .concat(filename.substring(0,
                                                                      filename.lastIndexOf("."))));
                    }
                }
            }
        }
        return boards;
    }

    
    private void getBoardSizesInDir(final File searchDir,
                                    TreeSet<BoardDimensions> sizes) {
        if (searchDir == null) {
            throw new IllegalArgumentException("must provide searchDir");
        }

        if (sizes == null) {
            throw new IllegalArgumentException("must provide sizes");
        }

        String file_list[] = searchDir.list();

        for (String filename : file_list) {
            File query_file = new File(searchDir, filename);

            if (query_file.isDirectory()) {
                getBoardSizesInDir(query_file, sizes);
            } else {
                try {
                    if (filename.endsWith(".board")) { 
                        BoardDimensions size = Board.getSize(query_file);
                        if (size == null) {
                            throw new Exception();
                        }
                        sizes.add(Board.getSize(query_file));
                    }
                } catch (Exception e) {
                    System.out.println("Error parsing board: "
                                       + query_file.getAbsolutePath());
                }
            }
        }
    }

    private Set<BoardDimensions> getBoardSizes() {
        TreeSet<BoardDimensions> board_sizes = new TreeSet<BoardDimensions>();

        File boards_dir = Configuration.boardsDir();
        
        if (boards_dir.isDirectory()) {
            getBoardSizesInDir(boards_dir, board_sizes);
        }
        return board_sizes;
    }

    
    @Deprecated
    private ArrayList<String> scanForBoards(final int boardWidth,
                                            final int boardHeight) {
        return scanForBoards(new BoardDimensions(boardWidth, boardHeight));
    }

    
    private ArrayList<String> scanForBoards(final BoardDimensions dimensions) {
        ArrayList<String> boards = new ArrayList<String>();

        File boardDir = Configuration.boardsDir();
        boards.add(MapSettings.BOARD_GENERATED);
        
        if (!boardDir.isDirectory()) {
            return boards;
        }

        ArrayList<String> tempList = new ArrayList<String>();
        Comparator<String> sortComp = StringUtil.stringComparator();
        scanForBoardsInDir(boardDir, "", dimensions, tempList);
        
        if (tempList.size() > 0) {
            boards.add(MapSettings.BOARD_RANDOM);
            boards.add(MapSettings.BOARD_SURPRISE);
            Collections.sort(tempList, sortComp);
            for (int loop = 0; loop < tempList.size(); loop++) {
                boards.add(tempList.get(loop));
            }
        }

        return boards;
    }

    
    private boolean doBlind() {
        return game.getOptions().booleanOption("double_blind")
               && game.getPhase()
                      .isDuringOrAfter(IGame.Phase.PHASE_DEPLOYMENT);
    }

    
    public void entityUpdate(int nEntityID) {
        entityUpdate(nEntityID, new Vector<UnitLocation>(), true, null);
    }

    public void entityUpdate(int nEntityID, Vector<UnitLocation> movePath,
            boolean updateVisibility,
            Map<EntityTargetPair, LosEffects> losCache) {
        Entity eTarget = game.getEntity(nEntityID);
        if (eTarget == null) {
            if (game.getOutOfGameEntity(nEntityID) != null) {
                System.err
                        .print("S: attempted to send entity update for out of game entity, id was ");
                System.err.println(nEntityID);
            } else {
                System.err
                        .print("S: attempted to send entity update for null entity, id was ");
                System.err.println(nEntityID);
            }
            return; 
        }

        if (doBlind()) {
            Vector<IPlayer> playersVector = game.getPlayersVector();
            Vector<IPlayer> vCanSee;
            if (updateVisibility) {
                vCanSee = whoCanSee(eTarget, true, losCache);
            } else {
                vCanSee = eTarget.getWhoCanSee();
            }

            if ((eTarget.getECMRange() > 0) && (eTarget.getPosition() != null)) {
                int ecmRange = eTarget.getECMRange();
                Coords pos = eTarget.getPosition();
                for (Entity ent : game.getEntitiesVector()) {
                    if ((ent.getPosition() != null)
                        && (pos.distance(ent.getPosition()) <= ecmRange)) {
                        if (!vCanSee.contains(ent.getOwner())) {
                            vCanSee.add(ent.getOwner());
                        }
                    }
                }
            }

            Packet pack = createEntityPacket(nEntityID, movePath);
            for (int x = 0; x < vCanSee.size(); x++) {
                IPlayer p = vCanSee.elementAt(x);
                send(p.getId(), pack);
            }
            
            pack = createRemoveEntityPacket(nEntityID,
                                            eTarget.getRemovalCondition());
            for (int x = 0; x < playersVector.size(); x++) {
                if (!vCanSee.contains(playersVector.elementAt(x))) {
                    IPlayer p = playersVector.elementAt(x);
                    send(p.getId(), pack);
                }
            }

            entityUpdateLoadedUnits(eTarget, vCanSee, playersVector);
        } else {
            
            send(createEntityPacket(nEntityID, movePath));
        }
    }

    private void entityUpdateLoadedUnits(Entity loader,
                                         Vector<IPlayer> vCanSee, Vector<IPlayer> playersVector) {
        Packet pack;

        for (Entity eLoaded : loader.getLoadedUnits()) {
            
            pack = createEntityPacket(eLoaded.getId(), null);
            for (int x = 0; x < vCanSee.size(); x++) {
                IPlayer p = vCanSee.elementAt(x);
                send(p.getId(), pack);
            }
            
            pack = createRemoveEntityPacket(eLoaded.getId(),
                                            eLoaded.getRemovalCondition());
            for (int x = 0; x < playersVector.size(); x++) {
                if (!vCanSee.contains(playersVector.elementAt(x))) {
                    IPlayer p = playersVector.elementAt(x);
                    send(p.getId(), pack);
                }
            }
            entityUpdateLoadedUnits(eLoaded, vCanSee, playersVector);
        }
    }

   
    private Vector<IPlayer> whoCanSee(Entity entity) {
        return whoCanSee(entity, true, null);
    }

    private Vector<IPlayer> whoCanSee(Entity entity, boolean useSensors,
            Map<EntityTargetPair, LosEffects> losCache) {
        if (losCache == null) {
            losCache = new HashMap<>();
        }
        
        if (entity == null) {
            return new Vector<IPlayer>();
        }

        List<ECMInfo> allECMInfo = null;
        if (game.getOptions().booleanOption("tacops_sensors") && useSensors) {
            allECMInfo = ComputeECM.computeAllEntitiesECMInfo(game
                    .getEntitiesVector());
        }

        boolean bTeamVision = game.getOptions().booleanOption("team_vision");
        List<Entity> vEntities = game.getEntitiesVector();

        Vector<IPlayer> vCanSee = new Vector<IPlayer>();
        vCanSee.addElement(entity.getOwner());
        if (bTeamVision) {
            addTeammates(vCanSee, entity.getOwner());
        }

        
        for (Enumeration<IPlayer> p = game.getPlayers(); p.hasMoreElements();) {
            IPlayer player = p.nextElement();

            if (player.canSeeAll() && !vCanSee.contains(p)) {
                vCanSee.addElement(player);
            }
        }

        if (entity.isHidden()) {
            return vCanSee;
        }
        for (Entity spotter : vEntities) {
            
            if (!spotter.isActive() || spotter.isOffBoard()
                    || vCanSee.contains(spotter.getOwner())) {
                continue;
            }
            
            EntityTargetPair etp = new EntityTargetPair(spotter, entity);
            LosEffects los = losCache.get(etp);
            if (los == null) {
                los = LosEffects.calculateLos(game, spotter.getId(), entity);
                losCache.put(etp, los);
            }
            if (Compute.canSee(game, spotter, entity, useSensors, los,
                    allECMInfo)) {
                if (!vCanSee.contains(spotter.getOwner())) {
                    vCanSee.addElement(spotter.getOwner());
                }
                if (bTeamVision) {
                    addTeammates(vCanSee, spotter.getOwner());
                }
                addObservers(vCanSee);
            }
        }
        return vCanSee;
    }

    
    private Vector<IPlayer> whoCanDetect(Entity entity,
            List<ECMInfo> allECMInfo,
            Map<EntityTargetPair, LosEffects> losCache) {
        if (losCache == null) {
            losCache = new HashMap<>();
        }

        boolean bTeamVision = game.getOptions().booleanOption("team_vision");
        List<Entity> vEntities = game.getEntitiesVector();

        Vector<IPlayer> vCanDetect = new Vector<IPlayer>();

        if (entity.isHidden() || entity.isOffBoard()) {
            return vCanDetect;
        }

        for (Entity spotter : vEntities) {
            if (!spotter.isActive() || spotter.isOffBoard()
                    || vCanDetect.contains(spotter.getOwner())) {
                continue;
            }
            
            EntityTargetPair etp = new EntityTargetPair(spotter, entity);
            LosEffects los = losCache.get(etp);
            if (los == null) {
                los = LosEffects.calculateLos(game, spotter.getId(), entity);
                losCache.put(etp, los);
            }
            if (Compute.inSensorRange(game, los, spotter, entity, allECMInfo)) {
                if (!vCanDetect.contains(spotter.getOwner())) {
                    vCanDetect.addElement(spotter.getOwner());
                }
                if (bTeamVision) {
                    addTeammates(vCanDetect, spotter.getOwner());
                }
                addObservers(vCanDetect);
            }
            }

        return vCanDetect;
    }

    
    private void addTeammates(Vector<IPlayer> vector, IPlayer player) {
        Vector<IPlayer> playersVector = game.getPlayersVector();
        for (int j = 0; j < playersVector.size(); j++) {
            IPlayer p = playersVector.elementAt(j);
            if (!player.isEnemyOf(p) && !vector.contains(p)) {
                vector.addElement(p);
            }
        }
    }

    
    private void addObservers(Vector<IPlayer> vector) {
        Vector<IPlayer> playersVector = game.getPlayersVector();
        for (int j = 0; j < playersVector.size(); j++) {
            IPlayer p = playersVector.elementAt(j);
            if (p.isObserver() && !vector.contains(p)) {
                vector.addElement(p);
            }
        }
    }

    
    private void entityAllUpdate() {
        
        if (doBlind()) {
            Vector<IPlayer> playersVector = game.getPlayersVector();
            for (int x = 0; x < playersVector.size(); x++) {
                IPlayer p = playersVector.elementAt(x);
                send(p.getId(), createFilteredEntitiesPacket(p, null));
            }
            return;
        }

        send(createEntitiesPacket());
    }

    
    private List<Entity> filterEntities(IPlayer pViewer,
            List<Entity> vEntities,
            Map<EntityTargetPair, LosEffects> losCache) {
        if (losCache == null) {
            losCache = new HashMap<>();
        }
        Vector<Entity> vCanSee = new Vector<Entity>();
        Vector<Entity> vMyEntities = new Vector<Entity>();
        boolean bTeamVision = game.getOptions().booleanOption("team_vision");

        if (pViewer.canSeeAll()) {
            return vEntities;
        }

        List<ECMInfo> allECMInfo = null;
        if (game.getOptions().booleanOption("tacops_sensors")) {
            allECMInfo = ComputeECM.computeAllEntitiesECMInfo(game
                    .getEntitiesVector());
        }

        if (pViewer.isObserver()) {
            vMyEntities.addAll(vEntities);
            for (Entity a : vMyEntities) {
                for (Entity b : vMyEntities) {
                    if (a.isEnemyOf(b)
                        && Compute.canSee(game, b, a, true, null, allECMInfo)) {
                        addVisibleEntity(vCanSee, a);
                        break;
                    }
                }
            }
            return vCanSee;
        }

        for (Entity e : vEntities) {
            if ((e.getOwner() == pViewer)
                || (bTeamVision && !e.getOwner().isEnemyOf(pViewer))) {
                vMyEntities.addElement(e);
            }
        }

        for (Entity e : vEntities) {
            
            if (vMyEntities.contains(e)) {
                addVisibleEntity(vCanSee, e);
                continue;
            } else if (e.isHidden()) {
                
                continue;
            }
            for (Entity spotter : vMyEntities) {

                if (spotter.isOffBoard()) {
                    continue;
                }

                EntityTargetPair etp = new EntityTargetPair(spotter, e);
                LosEffects los = losCache.get(etp);
                if (los == null) {
                    los = LosEffects.calculateLos(game, spotter.getId(), e);
                    losCache.put(etp, los);
                }
                
                if (Compute.canSee(game, spotter, e, true, los, allECMInfo)) {
                    addVisibleEntity(vCanSee, e);
                    break;
                }

                if ((e.getECMRange() > 0) && (e.getPosition() != null) &&
                    (spotter.getPosition() != null)) {
                    int ecmRange = e.getECMRange();
                    Coords pos = e.getPosition();
                    if (pos.distance(spotter.getPosition()) <= ecmRange) {
                        addVisibleEntity(vCanSee, e);
                    }
                }
            }
        }

        return vCanSee;
    }

    
    private void addVisibleEntity(Vector<Entity> vCanSee, Entity e) {
        if (!vCanSee.contains(e)) {
            vCanSee.add(e);
        }
        for (Entity transported : e.getLoadedUnits()) {
            addVisibleEntity(vCanSee, transported);
        }
    }

    
    private Vector<Report> filterReportVector(
            Vector<Report> originalReportVector, IPlayer p) {
        
        if (!doBlind()) {
            return new Vector<Report>(originalReportVector);
        }
        
        Vector<Report> filteredReportVector = new Vector<Report>();
        for (Report r : originalReportVector) {
            Report filteredReport = filterReport(r, p, false);
            if (filteredReport != null) {
                filteredReportVector.addElement(filteredReport);
            }
        }
        return filteredReportVector;
    }

    
    private Report filterReport(Report r, IPlayer p, boolean omitCheck) {

        if ((r.subject == Entity.NONE) && (r.type != Report.PLAYER)
            && (r.type != Report.PUBLIC)) {
            
            System.err.println("Error: Attempting to filter a Report object "
                               + "that is not public yet has no subject.\n\t\tmessageId: "
                               + r.messageId);
            return r;
        }
        if ((r.type == Report.PUBLIC) || ((p == null) && !omitCheck)) {
            return r;
        }
        Entity entity = game.getEntity(r.subject);
        if (entity == null) {
            entity = game.getOutOfGameEntity(r.subject);
        }
        IPlayer owner = null;
        if (entity != null) {
            owner = entity.getOwner();
            
            if (entity.isOffBoard()) {
                return r;
            }
        }

        if ((r.type != Report.PLAYER) && !omitCheck
            && ((entity == null) || (owner == null))) {
            System.err.println("Error: Attempting to filter a Report object "
                               + "that is not public but has a subject (" + entity
                               + ") with owner (" + owner + ").\n\tmessageId: "
                               + r.messageId);
            return r;
        }

        boolean shouldObscure = omitCheck
                                || ((entity != null) && !entity.hasSeenEntity(p))
                                || ((r.type == Report.PLAYER) && (p.getId() != r.player));
        
        if (game.getOptions()
                .booleanOption("supress_all_double_blind_messages")
            && shouldObscure) {
            
            if (p != null) {
                r.addObscuredRecipient(p.getName());
            }
            return null;
        }
        Report copy = new Report(r);
        
        for (int j = 0; j < copy.dataCount(); j++) {
            if (shouldObscure) {
                
                if (r.isValueObscured(j)) {
                    copy.hideData(j);
                    
                    if (p != null) {
                        r.addObscuredRecipient(p.getName());
                    }
                }
            }
        }
        return copy;
    }

    
    private Vector<Vector<Report>> filterPastReports(
            Vector<Vector<Report>> pastReports, IPlayer p) {
        
        if (!doBlind()) {
            return pastReports;
        }
        
        Vector<Vector<Report>> filteredReports = new Vector<Vector<Report>>();
        for (Vector<Report> roundReports : pastReports) {
            Vector<Report> filteredRoundReports = new Vector<Report>();
            for (Report r : roundReports) {
                if (r.isObscuredRecipient(p.getName())) {
                    r = filterReport(r, null, true);
                }
                if (r != null) {
                    filteredRoundReports.addElement(r);
                }
            }
            filteredReports.addElement(filteredRoundReports);
        }
        return filteredReports;
    }

    
    private void updateVisibilityIndicator(
            Map<EntityTargetPair, LosEffects> losCache) {
        if (losCache == null) {
            losCache = new HashMap<>();
        }
        List<ECMInfo> allECMInfo = null;
        if (game.getOptions().booleanOption("tacops_sensors")) {
            allECMInfo = ComputeECM.computeAllEntitiesECMInfo(game
                    .getEntitiesVector());
        }

        List<Entity> vAllEntities = game.getEntitiesVector();
        for (int x = 0; x < vAllEntities.size(); x++) {
            Entity e = vAllEntities.get(x);
            boolean previousVisibleValue = e.isVisibleToEnemy();
            boolean previousSeenValue = e.isEverSeenByEnemy();
            boolean previousDetectedValue = e.isDetectedByEnemy();
            e.setVisibleToEnemy(false);
            e.setDetectedByEnemy(false);
            e.clearSeenBy();
            e.clearDetectedBy();
            Vector<IPlayer> vCanSee = whoCanSee(e, false, losCache);
            
            for (IPlayer p : vCanSee) {
                if (e.getOwner().isEnemyOf(p) && !p.isObserver()) {
                    e.setVisibleToEnemy(true);
                    e.setEverSeenByEnemy(true);
                    
                    e.setDetectedByEnemy(true);
                }
                e.addBeenSeenBy(p);
            }
            
            Vector<IPlayer> vCanDetect = whoCanDetect(e, allECMInfo, losCache);
            for (IPlayer p : vCanDetect) {
                if (e.getOwner().isEnemyOf(p) && !p.isObserver()) {
                    e.setDetectedByEnemy(true);
                }
                e.addBeenDetectedBy(p);
            }

            if ((!previousVisibleValue && e.isVisibleToEnemy())
                || (!previousDetectedValue && e.isDetectedByEnemy())) {
                entityUpdate(e.getId(), new Vector<UnitLocation>(), false, losCache);
            } else if ((previousVisibleValue != e.isVisibleToEnemy())
                       || (previousSeenValue != e.isEverSeenByEnemy())
                       || (previousDetectedValue != e.isDetectedByEnemy())) {
                sendVisibilityIndicator(e);
            }
        }
    }

    
    private void receiveEntityAdd(Packet c, int connIndex) {
        @SuppressWarnings("unchecked")
        final List<Entity> entities = (List<Entity>) c.getObject(0);
        ArrayList<Integer> entityIds = new ArrayList<Integer>(entities.size());

        for (final Entity entity : entities) {

            
            if (Server.entityVerifier == null) {
                Server.entityVerifier = new EntityVerifier(new File(
                        Configuration.unitsDir(),
                        EntityVerifier.CONFIG_FILENAME));
            }

            TestEntity testEntity = null;
            entity.restore();
            if (entity instanceof Mech) {
                testEntity = new TestMech((Mech) entity,
                        entityVerifier.mechOption, null);
            } else if ((entity.getEntityType() == Entity.ETYPE_TANK)
                       && (entity.getEntityType() != Entity.ETYPE_GUN_EMPLACEMENT)) {
                if (entity.isSupportVehicle()) {
                    testEntity = new TestSupportVehicle((Tank) entity,
                            entityVerifier.tankOption, null);
                } else {
                    testEntity = new TestTank((Tank) entity,
                            entityVerifier.tankOption, null);
                }
            } else if ((entity.getEntityType() == Entity.ETYPE_AERO)
                       && (entity.getEntityType() != Entity.ETYPE_DROPSHIP)
                       && (entity.getEntityType() != Entity.ETYPE_SMALL_CRAFT)
                       && (entity.getEntityType() != Entity.ETYPE_FIGHTER_SQUADRON)
                       && (entity.getEntityType() != Entity.ETYPE_JUMPSHIP)
                       && (entity.getEntityType() != Entity.ETYPE_SPACE_STATION)) {
                testEntity = new TestAero((Aero) entity,
                        entityVerifier.aeroOption, null);
            } else if (entity instanceof BattleArmor) {
                testEntity = new TestBattleArmor((BattleArmor) entity,
                        entityVerifier.baOption, null);
            }

            if (testEntity != null) {
                StringBuffer sb = new StringBuffer();
                if (testEntity.correctEntity(sb, !game.getOptions()
                        .booleanOption("is_eq_limits"))) {
                    entity.setDesignValid(true);
                } else {
                    System.err.println(sb);
                    if (game.getOptions().booleanOption("allow_illegal_units")) {
                        entity.setDesignValid(false);
                    } else {
                        IPlayer cheater = game.getPlayer(connIndex);
                        sendServerChat("Player " + cheater.getName()
                                       + " attempted to add an illegal unit design ("
                                       + entity.getShortNameRaw()
                                       + "), the unit was rejected.");
                        return;
                    }
                }
            }

            if (entity instanceof Protomech) {

                
                int numPlayerProtos = game
                        .getSelectedEntityCount(new EntitySelector() {
                            private final int ownerId = entity.getOwnerId();

                            public boolean accept(Entity entity) {
                                if ((entity instanceof Protomech)
                                    && (ownerId == entity.getOwnerId())) {
                                    return true;
                                }
                                return false;
                            }
                        });

                entity.setUnitNumber((char) (numPlayerProtos / 5));

            } 

            if (Entity.NONE == entity.getId()) {
                entity.setId(getFreeEntityId());
            }

            game.addEntity(entity);

            if (entity.hasC3() || entity.hasC3i()) {
                boolean C3iSet = false;

                for (Entity e : game.getEntitiesVector()) {

                    if (entity.hasC3()) {
                        if ((entity.getC3MasterIsUUIDAsString() != null)
                            && entity.getC3MasterIsUUIDAsString().equals(
                                e.getC3UUIDAsString())) {
                            entity.setC3Master(e, false);
                            entity.setC3MasterIsUUIDAsString(null);
                        } else if ((e.getC3MasterIsUUIDAsString() != null)
                                   && e.getC3MasterIsUUIDAsString().equals(
                                entity.getC3UUIDAsString())) {
                            e.setC3Master(entity, false);
                            e.setC3MasterIsUUIDAsString(null);
                         
                            if (!entities.contains(e)) {
                                entityUpdate(e.getId());
                            }
                        }
                    }

                    if (entity.hasC3i() && (C3iSet == false)) {
                        entity.setC3NetIdSelf();
                        int pos = 0;
                        while (pos < Entity.MAX_C3i_NODES) {
                            
                            if ((entity.getC3iNextUUIDAsString(pos) != null)
                                && (e.getC3UUIDAsString() != null)
                                && entity.getC3iNextUUIDAsString(pos)
                                         .equals(e.getC3UUIDAsString())) {
                                entity.setC3NetId(e);
                                C3iSet = true;
                                break;
                            }

                            pos++;
                        }
                    }
                }
            }
            
            entity.setExternalSpotlight(entity.hasExternaSpotlight()
                                        || entity.hasQuirk(OptionsConstants.QUIRK_POS_SEARCHLIGHT));
            entityIds.add(entity.getId());
        }

        send(createAddEntityPacket(entityIds));
    }

    @SuppressWarnings("unchecked")
    private void receiveSquadronAdd(Packet c, int connIndex) {

        final FighterSquadron fs = (FighterSquadron) c.getObject(0);
        final Vector<Integer> fighters = (Vector<Integer>) c.getObject(1);
        if (fighters.size() < 1) {
            return;
        }
        
        if (Entity.NONE == fs.getId()) {
            fs.setId(getFreeEntityId());
        }
        game.addEntity(fs);
        for (int id : fighters) {
            Aero fighter = (Aero) game.getEntity(id);
            if (null != fighter) {
                fs.load(fighter, false);
                fs.autoSetMaxBombPoints();
                fighter.setTransportId(fs.getId());
                
                if (game.getPhase() == Phase.PHASE_LOUNGE) {
                    fighter.setBombChoices(fs.getBombChoices());
                }
                entityUpdate(fighter.getId());
            }
        }
        send(createAddEntityPacket(fs.getId()));
    }

    private void receiveEntityUpdate(Packet c, int connIndex) {
        Entity entity = (Entity) c.getObject(0);
        Entity oldEntity = game.getEntity(entity.getId());
        if ((oldEntity != null)
            && ((oldEntity.getOwner() == getPlayer(connIndex)) || (oldEntity
                                                                           .getOwner().getTeam() == getPlayer
                                                                           (connIndex).getTeam()))) {
            game.setEntity(entity.getId(), entity);
            entityUpdate(entity.getId());
            
            if (game.getPhase() == IGame.Phase.PHASE_LOUNGE) {
                StringBuffer message = new StringBuffer();
                message.append("Unit ");
                if (game.getOptions().booleanOption("blind_drop")
                    || game.getOptions().booleanOption("real_blind_drop")) {
                    if (!entity.getExternalIdAsString().equals("-1")) {
                        message.append('[')
                               .append(entity.getExternalIdAsString())
                               .append("] ");
                    }
                    message.append(entity.getId()).append('(')
                           .append(entity.getOwner().getName()).append(')');
                } else {
                    message.append(entity.getDisplayName());
                }
                message.append(" has been customized.");
                sendServerChat(message.toString());
            }
        } else {
            
        }
    }

    private void receiveEntityLoad(Packet c, int connIndex) {
        int loadeeId = (Integer) c.getObject(0);
        int loaderId = (Integer) c.getObject(1);
        int bayNumber = (Integer) c.getObject(2);
        Entity loadee = game.getEntity(loadeeId);
        Entity loader = game.getEntity(loaderId);

        if ((loadee != null) && (loader != null)) {
            loadUnit(loader, loadee, bayNumber);
            
            if (game.getPhase() == IGame.Phase.PHASE_LOUNGE) {
                
                loadee.setLoadedThisTurn(false);
            }
        } else {
            
        }
    }

    private void receiveCustomInit(Packet c, int connIndex) {
        
        if (game.getPhase() == IGame.Phase.PHASE_LOUNGE) {
            IPlayer p = (IPlayer) c.getObject(0);
            sendServerChat("" + p.getName() + " has customized initiative.");
        }
    }

    
    private void receiveEntityModeChange(Packet c, int connIndex) {
        int entityId = c.getIntValue(0);
        int equipId = c.getIntValue(1);
        int mode = c.getIntValue(2);
        Entity e = game.getEntity(entityId);
        if (e.getOwner() != getPlayer(connIndex)) {
            return;
        }
        Mounted m = e.getEquipment(equipId);

        if (m == null) {
            return;
        }

        try {
            
            if ((e instanceof BattleArmor) && (!m.isMissing())
                    && m.isBodyMounted()
                    && m.getType().hasFlag(WeaponType.F_MISSILE)
                    && (m.getLinked() != null)
                    && (m.getLinked().getUsableShotsLeft() > 0)
                    && (mode <= 0)) {
                m.setPendingDump(mode == -1);
            
            } else if ((m.getType() instanceof AmmoType)
                && !m.getType().hasInstantModeSwitch() && (mode <= 0)) {
                m.setPendingDump(mode == -1);
            } else if ((m.getType() instanceof WeaponType) && m.isDWPMounted()
                       && (mode <= 0)) {
                m.setPendingDump(mode == -1);
            } else {
                if (!m.setMode(mode)) {
                    System.err.println(e.getShortName() + ": " + m.getName()
                                       + ": " + e.getLocationName(m.getLocation())
                                       + " trying to compensate");
                    sendServerChat(e.getShortName() + ": " + m.getName() + ": "
                                   + e.getLocationName(m.getLocation())
                                   + " trying to compensate");
                    e.setGameOptions();

                    if (!m.setMode(mode)) {
                        System.err.println(e.getShortName() + ": "
                                           + m.getName() + ": "
                                           + e.getLocationName(m.getLocation())
                                           + " unable to compensate");
                        sendServerChat(e.getShortName() + ": " + m.getName()
                                       + ": " + e.getLocationName(m.getLocation())
                                       + " unable to compensate");
                    }

                }
            }
        } catch (Exception ex) {
            ex.printStackTrace();
        }

    }

    private void receiveEntitySensorChange(Packet c, int connIndex) {
        int entityId = c.getIntValue(0);
        int sensorId = c.getIntValue(1);
        Entity e = game.getEntity(entityId);
        e.setNextSensor(e.getSensors().elementAt(sensorId));
    }

    
    private void receiveEntitySinksChange(Packet c, int connIndex) {
        int entityId = c.getIntValue(0);
        int numSinks = c.getIntValue(1);
        Entity e = game.getEntity(entityId);
        if (e instanceof Mech) {
            ((Mech)e).setActiveSinksNextRound(numSinks);
        }
    }


    
    private void receiveEntityNovaNetworkModeChange(Packet c, int connIndex) {

        try {
            int entityId = c.getIntValue(0);
            String networkID = c.getObject(1).toString();
            Entity e = game.getEntity(entityId);
            if (e.getOwner() != getPlayer(connIndex)) {
                return;
            }
            
            e.setNewRoundNovaNetworkString(networkID);
        } catch (Exception ex) {
            ex.printStackTrace();
        }

    }

    
    private void receiveEntityMountedFacingChange(Packet c, int connIndex) {
        int entityId = c.getIntValue(0);
        int equipId = c.getIntValue(1);
        int facing = c.getIntValue(2);
        Entity e = game.getEntity(entityId);
        if (e.getOwner() != getPlayer(connIndex)) {
            return;
        }
        Mounted m = e.getEquipment(equipId);

        if (m == null) {
            return;
        }
        m.setFacing(facing);
    }

    
    private void receiveEntityCalledShotChange(Packet c, int connIndex) {
        int entityId = c.getIntValue(0);
        int equipId = c.getIntValue(1);
        Entity e = game.getEntity(entityId);
        if (e.getOwner() != getPlayer(connIndex)) {
            return;
        }
        Mounted m = e.getEquipment(equipId);

        if (m == null) {
            return;
        }
        m.getCalledShot().switchCalledShot();
    }

    
    private void receiveEntitySystemModeChange(Packet c, int connIndex) {
        int entityId = c.getIntValue(0);
        int equipId = c.getIntValue(1);
        int mode = c.getIntValue(2);
        Entity e = game.getEntity(entityId);
        if (e.getOwner() != getPlayer(connIndex)) {
            return;
        }
        if ((e instanceof Mech) && (equipId == Mech.SYSTEM_COCKPIT)) {
            ((Mech) e).setCockpitStatus(mode);
        }
    }

    
    private void receiveEntityAmmoChange(Packet c, int connIndex) {
        int entityId = c.getIntValue(0);
        int weaponId = c.getIntValue(1);
        int ammoId = c.getIntValue(2);
        Entity e = game.getEntity(entityId);

        
        if (null == e) {
            System.err
                    .print("Server.receiveEntityAmmoChange: could not find entity #");
            System.err.println(entityId);
            return;
        }
        IPlayer player = getPlayer(connIndex);
        if ((null != player) && (e.getOwner() != player)) {
            System.err.print("Server.receiveEntityAmmoChange: player ");
            System.err.print(player.getName());
            System.err.print(" does not own the entity ");
            System.err.println(e.getDisplayName());
            return;
        }

        
        Mounted mWeap = e.getEquipment(weaponId);
        Mounted mAmmo = e.getEquipment(ammoId);
        if (null == mAmmo) {
            System.err.print("Server.receiveEntityAmmoChange: entity ");
            System.err.print(e.getDisplayName());
            System.err.print(" does not have ammo #");
            System.err.println(ammoId);
            return;
        }
        if (!(mAmmo.getType() instanceof AmmoType)) {
            System.err.print("Server.receiveEntityAmmoChange: item # ");
            System.err.print(ammoId);
            System.err.print(" of entity ");
            System.err.print(e.getDisplayName());
            System.err.print(" is a ");
            System.err.print(mAmmo.getName());
            System.err.println(" and not ammo.");
            return;
        }
        if (null == mWeap) {
            System.err.print("Server.receiveEntityAmmoChange: entity ");
            System.err.print(e.getDisplayName());
            System.err.print(" does not have weapon #");
            System.err.println(weaponId);
            return;
        }
        if (!(mWeap.getType() instanceof WeaponType)) {
            System.err.print("Server.receiveEntityAmmoChange: item # ");
            System.err.print(weaponId);
            System.err.print(" of entity ");
            System.err.print(e.getDisplayName());
            System.err.print(" is a ");
            System.err.print(mWeap.getName());
            System.err.println(" and not a weapon.");
            return;
        }
        if (((WeaponType) mWeap.getType()).getAmmoType() == AmmoType.T_NA) {
            System.err.print("Server.receiveEntityAmmoChange: item # ");
            System.err.print(weaponId);
            System.err.print(" of entity ");
            System.err.print(e.getDisplayName());
            System.err.print(" is a ");
            System.err.print(mWeap.getName());
            System.err.println(" and does not use ammo.");
            return;
        }
        if (((WeaponType) mWeap.getType()).hasFlag(WeaponType.F_ONESHOT)) {
            System.err.print("Server.receiveEntityAmmoChange: item # ");
            System.err.print(weaponId);
            System.err.print(" of entity ");
            System.err.print(e.getDisplayName());
            System.err.print(" is a ");
            System.err.print(mWeap.getName());
            System.err.println(" and cannot use external ammo.");
            return;
        }

        e.loadWeapon(mWeap, mAmmo);
    }

    
    private void receiveEntityDelete(Packet c, int connIndex) {
        @SuppressWarnings("unchecked")
        List<Integer> ids = (List<Integer>) c.getObject(0);
        for (Integer entityId : ids) {
            final Entity entity = game.getEntity(entityId);

            if ((entity != null) && (entity.getOwner() == getPlayer(connIndex))) {

                if (entity instanceof Protomech) {

                    int numPlayerProtos = game
                            .getSelectedEntityCount(new EntitySelector() {
                                private final int ownerId = entity.getOwnerId();

                                public boolean accept(Entity entity) {
                                    if ((entity instanceof Protomech)
                                        && (ownerId == entity.getOwnerId())) {
                                        return true;
                                    }
                                    return false;
                                }
                            });

                    final char oldMax = (char) (Math
                                                        .ceil(numPlayerProtos / 5.0) - 1);
                    char newMax = (char) (Math
                                                  .ceil((numPlayerProtos - 1) / 5.0) - 1);
                    char deletedUnitNum = entity.getUnitNumber();

                    if ((oldMax != deletedUnitNum) && (oldMax != newMax)) {

                        Iterator<Entity> lastUnit = game
                                .getSelectedEntities(new EntitySelector() {
                                    private final int ownerId = entity
                                            .getOwnerId();

                                    private final char lastUnitNum = oldMax;

                                    public boolean accept(Entity entity) {
                                        if ((entity instanceof Protomech)
                                            && (ownerId == entity
                                                .getOwnerId())
                                            && (lastUnitNum == entity
                                                .getUnitNumber())) {
                                            return true;
                                        }
                                        return false;
                                    }
                                });
                        Entity lastUnitMember = lastUnit.next();
                        lastUnitMember.setUnitNumber(deletedUnitNum);
                        entityUpdate(lastUnitMember.getId());

                    } 

                } 

                if (game.getPhase() != IGame.Phase.PHASE_DEPLOYMENT) {
                    
                    game.removeTurnFor(entity);
                }
                game.removeEntity(entityId,
                        IEntityRemovalConditions.REMOVE_NEVER_JOINED);

            }
        }
        send(createRemoveEntityPacket(ids,
                IEntityRemovalConditions.REMOVE_NEVER_JOINED));

        
        if (game.getPhase() == IGame.Phase.PHASE_DEPLOYMENT) {
            for (Integer entityId : ids) {
                final Entity entity = game.getEntity(entityId);
                endCurrentTurn(entity); 
            }
        }
    }

    
    private void receivePlayerDone(Packet pkt, int connIndex) {
        boolean ready = pkt.getBooleanValue(0);
        IPlayer player = getPlayer(connIndex);
        if (null != player) {
            player.setDone(ready);
        }
    }

    private void receiveInitiativeRerollRequest(Packet pkt, int connIndex) {
        IPlayer player = getPlayer(connIndex);
        if (IGame.Phase.PHASE_INITIATIVE_REPORT != game.getPhase()) {
            StringBuffer message = new StringBuffer();
            if (null == player) {
                message.append("Player #").append(connIndex);
            } else {
                message.append(player.getName());
            }
            message.append(" is not allowed to ask for a reroll at this time.");
            System.err.println(message.toString());
            sendServerChat(message.toString());
            return;
        }
        if (game.hasTacticalGenius(player)) {
            game.addInitiativeRerollRequest(game.getTeamForPlayer(player));
        }
        if (null != player) {
            player.setDone(true);
        }
        checkReady();
    }

    
    private boolean receiveGameOptions(Packet packet, int connId) {
        IPlayer player = game.getPlayer(connId);
        
        if (null == player) {
            System.err.print("Server does not recognize player at connection ");
            System.err.println(connId);
            return false;
        }

        
        if ((password != null) && (password.length() > 0)
            && !password.equals(packet.getObject(0))) {
            sendServerChat(connId,
                           "The password you specified to change game options is incorrect.");
            return false;
        }

        if (game.getPhase().isDuringOrAfter(Phase.PHASE_DEPLOYMENT)) {
            return false;
        }

        int changed = 0;

        for (Enumeration<?> i = ((Vector<?>) packet.getObject(1)).elements(); i
                .hasMoreElements(); ) {
            IBasicOption option = (IBasicOption) i.nextElement();
            IOption originalOption = game.getOptions().getOption(
                    option.getName());

            if (originalOption == null) {
                continue;
            }

            StringBuffer message = new StringBuffer();
            message.append("Player ").append(player.getName())
                   .append(" changed option \"")
                   .append(originalOption.getDisplayableName())
                   .append("\" to ").append(option.getValue().toString())
                   .append('.');
            sendServerChat(message.toString());
            originalOption.setValue(option.getValue());
            changed++;
        }

        
        Compute.setRNG(game.getOptions().intOption("rng_type"));

        if (changed > 0) {
            for (Entity en : game.getEntitiesVector()) {
                en.setGameOptions();
            }
            entityAllUpdate();
            return true;
        }
        return false;
    }

    private void receiveGameOptionsAux(Packet packet, int connId) {

        for (Enumeration<?> i = ((Vector<?>) packet.getObject(1)).elements(); i
                .hasMoreElements(); ) {
            IBasicOption option = (IBasicOption) i.nextElement();
            IOption originalOption = game.getOptions().getOption(
                    option.getName());
            if (originalOption != null) {
                if ("maps_include_subdir".equals(originalOption.getName())) {
                    mapSettings.setBoardsAvailableVector(scanForBoards(
                            mapSettings.getBoardWidth(),
                            mapSettings.getBoardHeight()));
                    mapSettings.removeUnavailable();
                    mapSettings.setNullBoards(DEFAULT_BOARD);
                    send(createMapSettingsPacket());
                }
            }
        }

    }

    
    private void transmitGameVictoryEventToAll() {
        for (IConnection conn : connections) {
            send(conn.getId(), new Packet(Packet.COMMAND_GAME_VICTORY_EVENT));
        }
    }

    
    private void transmitAllPlayerConnects(int connId) {
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer player = i.nextElement();

            send(connId, createPlayerConnectPacket(player.getId()));
        }
    }

    
    private Packet createPlayerConnectPacket(int playerId) {
        final Object[] data = new Object[2];
        data[0] = new Integer(playerId);
        data[1] = getPlayer(playerId);
        return new Packet(Packet.COMMAND_PLAYER_ADD, data);
    }

    
    private Packet createPlayerUpdatePacket(int playerId) {
        final Object[] data = new Object[2];
        data[0] = new Integer(playerId);
        data[1] = getPlayer(playerId);
        return new Packet(Packet.COMMAND_PLAYER_UPDATE, data);
    }

    
    private void transmitAllPlayerUpdates() {
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer player = i.nextElement();
            if (null != player) {
                send(createPlayerUpdatePacket(player.getId()));
            }
        }
    }

    
    private void transmitAllPlayerDones() {
        for (Enumeration<IPlayer> i = game.getPlayers(); i.hasMoreElements(); ) {
            final IPlayer player = i.nextElement();

            send(createPlayerDonePacket(player.getId()));
        }
    }

    
    private Packet createPlayerDonePacket(int playerId) {
        Object[] data = new Object[2];
        data[0] = new Integer(playerId);
        data[1] = Boolean.valueOf(getPlayer(playerId).isDone());
        return new Packet(Packet.COMMAND_PLAYER_READY, data);
    }

    
    private Packet createTurnVectorPacket() {
        return new Packet(Packet.COMMAND_SENDING_TURNS, game.getTurnVector());
    }

    
    private Packet createTurnIndexPacket() {
        return new Packet(Packet.COMMAND_TURN, new Integer(game.getTurnIndex()));
    }

    
    private Packet createMapSettingsPacket() {
        return new Packet(Packet.COMMAND_SENDING_MAP_SETTINGS, mapSettings);
    }

    private Packet createMapSizesPacket() {
        Set<BoardDimensions> sizes = getBoardSizes();
        return new Packet(Packet.COMMAND_SENDING_AVAILABLE_MAP_SIZES, sizes);
    }

    
    private Packet createPlanetaryConditionsPacket() {
        return new Packet(Packet.COMMAND_SENDING_PLANETARY_CONDITIONS,
                          game.getPlanetaryConditions());
    }

    
    private Packet createGameSettingsPacket() {
        return new Packet(Packet.COMMAND_SENDING_GAME_SETTINGS,
                          game.getOptions());
    }

    
    private Packet createBoardPacket() {
        return new Packet(Packet.COMMAND_SENDING_BOARD, game.getBoard());
    }

    
    private Packet createEntityPacket(int entityId,
                                      Vector<UnitLocation> movePath) {
        final Entity entity = game.getEntity(entityId);
        final Object[] data = new Object[3];
        data[0] = new Integer(entityId);
        data[1] = entity;
        data[2] = movePath;
        return new Packet(Packet.COMMAND_ENTITY_UPDATE, data);
    }

    
    private Packet createReportPacket(IPlayer p) {
        
        
        
        if ((p == null) || !doBlind()) {
            return new Packet(Packet.COMMAND_SENDING_REPORTS, vPhaseReport);
        }
        return new Packet(Packet.COMMAND_SENDING_REPORTS, filterReportVector(
                vPhaseReport, p));
    }

    
    private Packet createSpecialReportPacket() {
        return new Packet(Packet.COMMAND_SENDING_REPORTS_SPECIAL,
                          vPhaseReport.clone());
    }

    
    private Packet createTacticalGeniusReportPacket() {
        return new Packet(Packet.COMMAND_SENDING_REPORTS_TACTICAL_GENIUS,
                          vPhaseReport.clone());
    }

    
    private Packet createAllReportsPacket(IPlayer p) {
        return new Packet(Packet.COMMAND_SENDING_REPORTS_ALL,
                          filterPastReports(game.getAllReports(), p));
    }

    
    private Packet createEntitiesPacket() {
        return new Packet(Packet.COMMAND_SENDING_ENTITIES,
                          game.getEntitiesVector());
    }

    
    private Packet createFullEntitiesPacket() {
        final Object[] data = new Object[2];
        data[0] = game.getEntitiesVector();
        data[1] = game.getOutOfGameEntitiesVector();
        return new Packet(Packet.COMMAND_SENDING_ENTITIES, data);
    }

    
    private Packet createFilteredEntitiesPacket(IPlayer p,
            Map<EntityTargetPair, LosEffects> losCache) {
        return new Packet(Packet.COMMAND_SENDING_ENTITIES, filterEntities(p,
                game.getEntitiesVector(), losCache));
    }

    
    private Packet createFilteredFullEntitiesPacket(IPlayer p) {
        final Object[] data = new Object[2];
        data[0] = filterEntities(p, game.getEntitiesVector(), null);
        data[1] = game.getOutOfGameEntitiesVector();
        return new Packet(Packet.COMMAND_SENDING_ENTITIES, data);
    }

    private Packet createAddEntityPacket(int entityId) {
        ArrayList<Integer> entityIds = new ArrayList<Integer>(1);
        entityIds.add(entityId);
        return createAddEntityPacket(entityIds);
    }

    
    private Packet createAddEntityPacket(List<Integer> entityIds) {
        ArrayList<Entity> entities = new ArrayList<Entity>(entityIds.size());
        for (Integer id : entityIds) {
            entities.add(game.getEntity(id));
        }
        final Object[] data = new Object[2];
        data[0] = entityIds;
        data[1] = entities;
        return new Packet(Packet.COMMAND_ENTITY_ADD, data);
    }

    
    private Packet createRemoveEntityPacket(int entityId) {
        return createRemoveEntityPacket(entityId,
                                        IEntityRemovalConditions.REMOVE_SALVAGEABLE);
    }

    
    private Packet createRemoveEntityPacket(int entityId, int condition) {
        ArrayList<Integer> ids = new ArrayList<Integer>(1);
        ids.add(entityId);
        return createRemoveEntityPacket(ids, condition);
    }

    
    private Packet createRemoveEntityPacket(List<Integer> entityIds,
                                            int condition) {
        if ((condition != IEntityRemovalConditions.REMOVE_UNKNOWN)
            && (condition != IEntityRemovalConditions.REMOVE_IN_RETREAT)
            && (condition != IEntityRemovalConditions.REMOVE_PUSHED)
            && (condition != IEntityRemovalConditions.REMOVE_SALVAGEABLE)
            && (condition != IEntityRemovalConditions.REMOVE_EJECTED)
            && (condition != IEntityRemovalConditions.REMOVE_CAPTURED)
            && (condition != IEntityRemovalConditions.REMOVE_DEVASTATED)
            && (condition != IEntityRemovalConditions.REMOVE_NEVER_JOINED)) {
            throw new IllegalArgumentException("Unknown unit condition: "
                                               + condition);
        }
        Object[] array = new Object[2];
        array[0] = entityIds;
        array[1] = new Integer(condition);
        return new Packet(Packet.COMMAND_ENTITY_REMOVE, array);
    }

    
    private Packet createEndOfGamePacket() {
        Object[] array = new Object[3];
        array[0] = getDetailedVictoryReport();
        array[1] = new Integer(game.getVictoryPlayerId());
        array[2] = new Integer(game.getVictoryTeam());
        return new Packet(Packet.COMMAND_END_OF_GAME, array);
    }

    public static String ORIGIN = "***Server";

    public static String formatChatMessage(String origin, String message) {
        return origin + ": " + message;
    }

    
    public void sendChat(int connId, String origin, String message) {
        send(connId,
             new Packet(Packet.COMMAND_CHAT, formatChatMessage(origin,
                                                               message)));
    }

    
    private void sendChat(String origin, String message) {
        String chat = formatChatMessage(origin, message);
        send(new Packet(Packet.COMMAND_CHAT, chat));
    }

    public void sendServerChat(int connId, String message) {
        sendChat(connId, ORIGIN, message);
    }

    public void sendServerChat(String message) {
        sendChat(ORIGIN, message);
    }

    
    private Packet createHexChangePacket(Coords coords, IHex hex) {
        final Object[] data = new Object[2];
        data[0] = coords;
        data[1] = hex;
        return new Packet(Packet.COMMAND_CHANGE_HEX, data);
    }

    public void sendSmokeCloudAdded(SmokeCloud cloud) {
        final Object[] data = new Object[1];
        data[0] = cloud;
        send(new Packet(Packet.COMMAND_ADD_SMOKE_CLOUD, data));
    }

    
    public void sendChangedHex(Coords coords) {
        send(createHexChangePacket(coords, game.getBoard().getHex(coords)));
    }

    
    private Packet createHexesChangePacket(Set<Coords> coords, Set<IHex> hex) {
        final Object[] data = new Object[2];
        data[0] = coords;
        data[1] = hex;
        return new Packet(Packet.COMMAND_CHANGE_HEXES, data);
    }

    
    public void sendChangedHexes(Set<Coords> coords) {
        Set<IHex> hexes = new LinkedHashSet<IHex>();
        for (Coords coord : coords) {
            hexes.add(game.getBoard().getHex(coord));
        }
        send(createHexesChangePacket(coords, hexes));
    }

    
    private Packet createMineChangePacket(Coords coords) {
        return new Packet(Packet.COMMAND_UPDATE_MINEFIELDS,
                          game.getMinefields(coords));
    }

    
    public void sendChangedMines(Coords coords) {
        send(createMineChangePacket(coords));
    }

    public void sendVisibilityIndicator(Entity e) {
        final Object[] data = new Object[6];
        data[0] = new Integer(e.getId());
        data[1] = Boolean.valueOf(e.isEverSeenByEnemy());
        data[2] = Boolean.valueOf(e.isVisibleToEnemy());
        data[3] = Boolean.valueOf(e.isDetectedByEnemy());
        data[4] = e.getWhoCanSee();
        data[5] = e.getWhoCanDetect();
        send(new Packet(Packet.COMMAND_ENTITY_VISIBILITY_INDICATOR, data));
    }

    
    private Packet createAttackPacket(List<?> vector, int charges) {
        final Object[] data = new Object[2];
        data[0] = vector;
        data[1] = new Integer(charges);
        return new Packet(Packet.COMMAND_ENTITY_ATTACK, data);
    }

    
    private Packet createAttackPacket(EntityAction ea, int charge) {
        Vector<EntityAction> vector = new Vector<EntityAction>(1);
        vector.addElement(ea);
        Object[] data = new Object[2];
        data[0] = vector;
        data[1] = new Integer(charge);
        return new Packet(Packet.COMMAND_ENTITY_ATTACK, data);
    }

    
    private Packet createSpecialHexDisplayPacket(int toPlayer) {
        Hashtable<Coords, Collection<SpecialHexDisplay>> shdTable = game
                .getBoard().getSpecialHexDisplayTable();
        Hashtable<Coords, Collection<SpecialHexDisplay>> shdTable2 = new Hashtable<Coords,
                Collection<SpecialHexDisplay>>();
        LinkedList<SpecialHexDisplay> tempList = null;
        IPlayer player = getPlayer(toPlayer);
        if (player != null) {
            for (Coords coord : shdTable.keySet()) {
                tempList = new LinkedList<SpecialHexDisplay>();
                for (SpecialHexDisplay shd : shdTable.get(coord)) {
                    if (!shd.isObscured(player)) {
                        tempList.add(0, shd);
                    }
                }
                if (!tempList.isEmpty()) {
                    shdTable2.put(coord, tempList);
                }
            }
        }
        return new Packet(Packet.COMMAND_SENDING_SPECIAL_HEX_DISPLAY, shdTable2);
    }

    private Packet createTagInfoUpdatesPacket() {
        return new Packet(Packet.COMMAND_SENDING_TAGINFO, game.getTagInfo());
    }

    
    private Packet createArtilleryPacket(IPlayer p) {
        Vector<ArtilleryAttackAction> v = new Vector<ArtilleryAttackAction>();
        int team = p.getTeam();
        for (Enumeration<AttackHandler> i = game.getAttacks(); i
                .hasMoreElements(); ) {
            WeaponHandler wh = (WeaponHandler) i.nextElement();
            if (wh.waa instanceof ArtilleryAttackAction) {
                ArtilleryAttackAction aaa = (ArtilleryAttackAction) wh.waa;
                if ((aaa.getPlayerId() == p.getId())
                    || ((team != IPlayer.TEAM_NONE) && (team == game
                        .getPlayer(aaa.getPlayerId()).getTeam()))
                    || p.getSeeAll()) {
                    v.addElement(aaa);
                }
            }
        }
        return new Packet(Packet.COMMAND_SENDING_ARTILLERYATTACKS, v);
    }

    private Packet createIlluminatedHexesPacket() {
        HashSet<Coords> illumHexes = game.getIlluminatedPositions();
        return new Packet(Packet.COMMAND_SENDING_ILLUM_HEXES, illumHexes);
    }

    
    private Packet createFlarePacket() {

        return new Packet(Packet.COMMAND_SENDING_FLARES, game.getFlares());
    }

    
    private void send(Packet packet) {
        if (connections == null) {
            return;
        }
        for (Enumeration<IConnection> connEnum = connections.elements(); connEnum
                .hasMoreElements(); ) {
            IConnection conn = connEnum.nextElement();
            conn.send(packet);
        }
    }

    
    public void send_Nova_Change(int ID, String net) {
        Object[] data = {new Integer(ID), new String(net)};
        Packet packet = new Packet(Packet.COMMAND_ENTITY_NOVA_NETWORK_CHANGE,
                                   data);
        send(packet);
    }

    private void sendReport() {
        sendReport(false);
    }

    
    private void sendReport(boolean tacticalGeniusReport) {
        if (connections == null) {
            return;
        }

        for (Enumeration<IConnection> connEnum = connections.elements(); connEnum
                .hasMoreElements(); ) {
            IConnection conn = connEnum.nextElement();
            IPlayer p = game.getPlayer(conn.getId());
            Packet packet;
            if (tacticalGeniusReport) {
                packet = createTacticalGeniusReportPacket();
            } else {
                packet = createReportPacket(p);
            }
            conn.send(packet);
        }
    }

    
    public void send(int connId, Packet packet) {
        if (getClient(connId) != null) {
            getClient(connId).send(packet);
        } else {
            
            
        }
    }

    
    private void sendToPending(int connId, Packet packet) {
        IConnection pendingConn = getPendingConnection(connId);
        if (pendingConn != null) {
            pendingConn.send(packet);
        } else {
            
            
        }
    }

    
    private void processCommand(int connId, String commandString) {
        String[] args;
        String commandName;
        
        args = commandString.split("\\s+");

        commandName = args[0].substring(1);

        ServerCommand command = getCommand(commandName);
        if (command != null) {
            command.run(connId, args);
        } else {
            sendServerChat(connId,
                           "Command not recognized.  Type /help for a list of commands.");
        }
    }

    
    private static final String DUNE_CALL = "They tried and failed?";

    private static final String DUNE_RESPONSE = "They tried and died!";

    private static final String STAR_WARS_CALL = "I'd just as soon kiss a Wookiee.";

    private static final String STAR_WARS_RESPONSE = "I can arrange that!";

    private static final String INVADER_ZIM_CALL = "What does the G stand for?";

    private static final String INVADER_ZIM_RESPONSE = "I dont know.";

    protected synchronized void handle(int connId, Packet packet) {
        IPlayer player = game.getPlayer(connId);
        
        if ((null == player) && (null == getPendingConnection(connId))) {
            System.err.print("Server does not recognize player at connection ");
            System.err.println(connId);
            return;
        }

        if (packet == null) {
            System.out.println("server.connection.handle: got null packet");
            return;
        }
        
        switch (packet.getCommand()) {
            case Packet.COMMAND_CLIENT_VERSIONS:
                receivePlayerVersion(packet, connId);
                break;
            case Packet.COMMAND_CLOSE_CONNECTION:
                
                IConnection c = getConnection(connId);
                if (c != null) {
                    c.close();
                }
                break;
            case Packet.COMMAND_CLIENT_NAME:
                receivePlayerName(packet, connId);
                break;
            case Packet.COMMAND_PLAYER_UPDATE:
                receivePlayerInfo(packet, connId);
                validatePlayerInfo(connId);
                send(createPlayerUpdatePacket(connId));
                break;
            case Packet.COMMAND_PLAYER_READY:
                receivePlayerDone(packet, connId);
                send(createPlayerDonePacket(connId));
                checkReady();
                break;
            case Packet.COMMAND_REROLL_INITIATIVE:
                receiveInitiativeRerollRequest(packet, connId);
                
                break;
            case Packet.COMMAND_FORWARD_INITIATIVE:
                receiveForwardIni(packet, connId);
                break;
            case Packet.COMMAND_CHAT:
                String chat = (String) packet.getObject(0);
                if (chat.startsWith("/")) {
                    processCommand(connId, chat);
                } else if (packet.getData().length > 1) {
                    connId = (int) packet.getObject(1);
                    if (connId == IPlayer.PLAYER_NONE) {
                        sendServerChat(chat);
                    } else {
                        sendServerChat(connId, chat);
                    }
                } else {
                    sendChat(player.getName(), chat);
                }
                
                if (DUNE_CALL.equalsIgnoreCase(chat)) {
                    sendServerChat(DUNE_RESPONSE);
                } else if (STAR_WARS_CALL.equalsIgnoreCase(chat)) {
                    sendServerChat(STAR_WARS_RESPONSE);
                } else if (INVADER_ZIM_CALL.equalsIgnoreCase(chat)) {
                    sendServerChat(INVADER_ZIM_RESPONSE);
                }
                break;
            case Packet.COMMAND_ENTITY_MOVE:
                receiveMovement(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_DEPLOY:
                receiveDeployment(packet, connId);
                break;
            case Packet.COMMAND_DEPLOY_MINEFIELDS:
                receiveDeployMinefields(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_ATTACK:
                receiveAttack(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_GTA_HEX_SELECT:
                receiveGroundToAirHexSelectPacket(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_ADD:
                receiveEntityAdd(packet, connId);
                resetPlayersDone();
                break;
            case Packet.COMMAND_ENTITY_UPDATE:
                receiveEntityUpdate(packet, connId);
                resetPlayersDone();
                break;
            case Packet.COMMAND_ENTITY_LOAD:
                receiveEntityLoad(packet, connId);
                resetPlayersDone();
                transmitAllPlayerDones();
                break;
            case Packet.COMMAND_ENTITY_MODECHANGE:
                receiveEntityModeChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_SENSORCHANGE:
                receiveEntitySensorChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_SINKSCHANGE:
                receiveEntitySinksChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_NOVA_NETWORK_CHANGE:
                receiveEntityNovaNetworkModeChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_MOUNTED_FACINGCHANGE:
                receiveEntityMountedFacingChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_CALLEDSHOTCHANGE:
                receiveEntityCalledShotChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_SYSTEMMODECHANGE:
                receiveEntitySystemModeChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_AMMOCHANGE:
                receiveEntityAmmoChange(packet, connId);
                break;
            case Packet.COMMAND_ENTITY_REMOVE:
                receiveEntityDelete(packet, connId);
                resetPlayersDone();
                break;
            case Packet.COMMAND_ENTITY_WORDER_UPDATE:
                Object data[] = packet.getData();
                Entity ent = game.getEntity((Integer)data[0]);
                if (ent != null) {
                    Entity.WeaponSortOrder order =
                        (Entity.WeaponSortOrder)data[1];
                    ent.setWeaponSortOrder(order);
                    
                    ent.setWeapOrderChanged(false);
                    if (order == Entity.WeaponSortOrder.CUSTOM) {
                        @SuppressWarnings("unchecked") 
                        Map<Integer, Integer> customWeapOrder = (Map<Integer, Integer>)data[2];
                        ent.setCustomWeaponOrder(customWeapOrder);
                    }
                }
                break;
            case Packet.COMMAND_SENDING_GAME_SETTINGS:
                if (receiveGameOptions(packet, connId)) {
                    resetPlayersDone();
                    transmitAllPlayerDones();
                    send(createGameSettingsPacket());
                    receiveGameOptionsAux(packet, connId);
                }
                break;
            case Packet.COMMAND_SENDING_MAP_SETTINGS:
                if (game.getPhase().isBefore(Phase.PHASE_DEPLOYMENT)) {
                    MapSettings newSettings = (MapSettings) packet.getObject(0);
                    if (!mapSettings.equalMapGenParameters(newSettings)) {
                        sendServerChat("Player " + player.getName()
                                       + " changed map settings");
                    }
                    mapSettings = newSettings;
                    mapSettings.setBoardsAvailableVector(scanForBoards(
                            mapSettings.getBoardWidth(),
                            mapSettings.getBoardHeight()));
                    mapSettings.removeUnavailable();
                    mapSettings.setNullBoards(DEFAULT_BOARD);
                    mapSettings.replaceBoardWithRandom(MapSettings.BOARD_RANDOM);
                    mapSettings.removeUnavailable();
                    
                    if (mapSettings.getBoardsSelected().next() == null) {
                        mapSettings.setNullBoards((MapSettings.BOARD_GENERATED));
                    }
                    newSettings = null;
                    resetPlayersDone();
                    transmitAllPlayerDones();
                    send(createMapSettingsPacket());
                }
                break;
            case Packet.COMMAND_SENDING_MAP_DIMENSIONS:
                if (game.getPhase().isBefore(Phase.PHASE_DEPLOYMENT)) {
                    MapSettings newSettings = (MapSettings) packet.getObject(0);
                    if (!mapSettings.equalMapGenParameters(newSettings)) {
                        sendServerChat("Player " + player.getName()
                                       + " changed map dimensions");
                    }
                    mapSettings = newSettings;
                    newSettings = null;
                    mapSettings.setBoardsAvailableVector(scanForBoards(
                            mapSettings.getBoardWidth(),
                            mapSettings.getBoardHeight()));
                    mapSettings.removeUnavailable();
                    mapSettings.setNullBoards(DEFAULT_BOARD);
                    mapSettings.replaceBoardWithRandom(MapSettings.BOARD_RANDOM);
                    mapSettings.removeUnavailable();
                    
                    if (mapSettings.getBoardsSelected().next() == null) {
                        mapSettings.setNullBoards((MapSettings.BOARD_GENERATED));
                    }
                    resetPlayersDone();
                    transmitAllPlayerDones();
                    send(createMapSettingsPacket());
                }
                break;
            case Packet.COMMAND_SENDING_PLANETARY_CONDITIONS:
                
                if (game.getPhase().isBefore(Phase.PHASE_DEPLOYMENT)) {
                    PlanetaryConditions conditions = (PlanetaryConditions) packet
                            .getObject(0);
                    sendServerChat("Player " + player.getName()
                                   + " changed planetary conditions");
                    game.setPlanetaryConditions(conditions);
                    resetPlayersDone();
                    transmitAllPlayerDones();
                    send(createPlanetaryConditionsPacket());
                }
                break;
            case Packet.COMMAND_UNLOAD_STRANDED:
                receiveUnloadStranded(packet, connId);
                break;
            case Packet.COMMAND_SET_ARTYAUTOHITHEXES:
                receiveArtyAutoHitHexes(packet, connId);
                break;
            case Packet.COMMAND_CUSTOM_INITIATIVE:
                receiveCustomInit(packet, connId);
                resetPlayersDone();
                transmitAllPlayerDones();
                break;
            case Packet.COMMAND_LOAD_GAME:
                try {
                    sendServerChat(getPlayer(connId).getName()
                                   + " loaded a new game.");
                    setGame((IGame) packet.getObject(0));
                    for (IConnection conn : connections) {
                        sendCurrentInfo(conn.getId());
                    }
                } catch (Exception e) {
                    System.out.println("Error loading savegame sent from client");
                }
                break;
            case Packet.COMMAND_SQUADRON_ADD:
                receiveSquadronAdd(packet, connId);
                resetPlayersDone();
                transmitAllPlayerDones();
                break;
            case Packet.COMMAND_RESET_ROUND_DEPLOYMENT:
                game.setupRoundDeployment();
                break;
            case Packet.COMMAND_SPECIAL_HEX_DISPLAY_DELETE:
                game.getBoard().removeSpecialHexDisplay(
                        (Coords) packet.getObject(0),
                        (SpecialHexDisplay) packet.getObject(1));
                sendSpecialHexDisplayPackets();
                break;
            case Packet.COMMAND_SPECIAL_HEX_DISPLAY_APPEND:
                game.getBoard().addSpecialHexDisplay((Coords) packet.getObject(0),
                                                     (SpecialHexDisplay) packet.getObject(1));
                sendSpecialHexDisplayPackets();
                break;
        }
    }

    
    public void run() {
        Thread currentThread = Thread.currentThread();
        System.out.println("s: listening for clients...");
        
        while (connector == currentThread) {
            try {
                Socket s = serverSocket.accept();
                synchronized (serverLock) {
                    int id = getFreeConnectionId();
                    System.out.println("s: accepting player connection #" + id
                            + " ...");

                    IConnection c = ConnectionFactory.getInstance()
                            .createServerConnection(s, id);
                    c.addConnectionListener(connectionListener);
                    c.open();
                    connectionsPending.addElement(c);
                    ConnectionHandler ch = new ConnectionHandler(c);
                    Thread newConnThread = new Thread(ch, "Connection " + id);
                    newConnThread.start();
                    connectionHandlers.put(id, ch);

                    greeting(id);
                    ConnectionWatchdog w = new ConnectionWatchdog(this, id);
                    watchdogTimer.schedule(w, 1000, 500);
                }
            } catch (InterruptedIOException iioe) {
                
            } catch (IOException ex) {

            }
            
        }
    }

    
    private Vector<Report> explodeInfernoAmmoFromHeat(Entity entity) {
        int damage = 0;
        int rack = 0;
        int boomloc = -1;
        int boomslot = -1;
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        for (int j = 0; j < entity.locations(); j++) {
            for (int k = 0; k < entity.getNumberOfCriticals(j); k++) {
                CriticalSlot cs = entity.getCritical(j, k);
                
                if ((cs == null) || cs.isDestroyed() || cs.isHit()
                    || (cs.getType() != CriticalSlot.TYPE_EQUIPMENT)) {
                    continue;
                }
                
                Mounted mounted = cs.getMount();
                if (!(mounted.getType() instanceof AmmoType)) {
                    continue;
                }
                
                AmmoType atype = (AmmoType) mounted.getType();
                if (!atype.isExplosive(mounted)
                    || ((atype.getMunitionType() != AmmoType.M_INFERNO) && (atype
                                                                                    .M_IATM_IIW))) {
                    continue;
                }

                if (mounted.getHittableShotsLeft() == 0) {
                    continue;
                }

                int newRack = atype.getDamagePerShot() * atype.getRackSize();
                int newDamage = mounted.getExplosionDamage();
                Mounted mount2 = cs.getMount2();
                if ((mount2 != null) && (mount2.getType() instanceof AmmoType)
                    && (mount2.getHittableShotsLeft() > 0)) {
                    
                    atype = (AmmoType) mount2.getType();
                    newRack += atype.getDamagePerShot() * atype.getRackSize();
                    newDamage += mount2.getExplosionDamage();
                }
                if (!mounted.isHit()
                    && ((rack < newRack) || ((rack == newRack) && (damage < newDamage)))) {
                    rack = newRack;
                    damage = newDamage;
                    boomloc = j;
                    boomslot = k;
                }
            }
        }
        
        if ((boomloc != -1) && (boomslot != -1)) {
            CriticalSlot slot = entity.getCritical(boomloc, boomslot);
            slot.setHit(true);
            Mounted equip = slot.getMount();
            equip.setHit(true);
            
            entity.heat += Math.min(equip.getExplosionDamage(), 30);
            vDesc.addAll(explodeEquipment(entity, boomloc, boomslot));
            r = new Report(5155);
            r.indent();
            r.subject = entity.getId();
            r.add(entity.heat);
            vDesc.addElement(r);
            entity.heatBuildup = 0;
        } else { 
            r = new Report(5160);
            r.indent();
            r.subject = entity.getId();
            vDesc.addElement(r);
        }
        return vDesc;
    }
    public void checkExplodeIndustrialZone(Coords c, Vector<Report> vDesc) {
        Report r;
        IHex hex = game.getBoard().getHex(c);
        if (null == hex) {
            return;
        }

        if (!hex.containsTerrain(Terrains.INDUSTRIAL)) {
            return;
        }

        r = new Report(3590, Report.PUBLIC);
        r.add(c.getBoardNum());
        r.indent(2);
        int effect = Compute.d6(2);
        r.add(8);
        r.add(effect);
        if (effect > 7) {
            r.choose(true);
            r.newlines = 0;
            vDesc.add(r);
            boolean onFire = false;
            boolean powerLine = false;
            boolean minorExp = false;
            boolean elecExp = false;
            boolean majorExp = false;
            if (effect == 8) {
                onFire = true;
                r = new Report(3600, Report.PUBLIC);
                r.newlines = 0;
                vDesc.add(r);
            } else if (effect == 9) {
                powerLine = true;
                r = new Report(3605, Report.PUBLIC);
                r.newlines = 0;
                vDesc.add(r);
            } else if (effect == 10) {
                minorExp = true;
                onFire = true;
                r = new Report(3610, Report.PUBLIC);
                r.newlines = 0;
                vDesc.add(r);
            } else if (effect == 11) {
                elecExp = true;
                r = new Report(3615, Report.PUBLIC);
                r.newlines = 0;
                vDesc.add(r);
            } else {
                onFire = true;
                majorExp = true;
                r = new Report(3620, Report.PUBLIC);
                r.newlines = 0;
                vDesc.add(r);
            }
            
            if (powerLine || minorExp || elecExp || majorExp) {
                
                for (Entity en : game.getEntitiesVector(c)) {
                    int damage = 3;
                    if (minorExp) {
                        damage = 5;
                    }
                    if (elecExp) {
                        damage = Compute.d6(1) + 3;
                    }
                    if (majorExp) {
                        damage = Compute.d6(2);
                    }
                    HitData hit = en.rollHitLocation(ToHitData.HIT_NORMAL,
                                                     ToHitData.SIDE_FRONT);
                    if (en instanceof BattleArmor) {
                        
                        
                        for (int loc = 0; loc < en.locations(); loc++) {
                            if ((IArmorState.ARMOR_NA != en.getInternal(loc))
                                && (IArmorState.ARMOR_DESTROYED != en
                                    .getInternal(loc))
                                && (IArmorState.ARMOR_DOOMED != en
                                    .getInternal(loc))) {
                                vDesc.addAll(damageEntity(en, new HitData(loc),
                                                          damage));
                            }
                        }
                    } else {
                        vDesc.addAll(damageEntity(en, hit, damage));
                    }
                    if (majorExp) {
                        
                        
                        vDesc.addAll(deliverInfernoMissiles(en, en,
                                                            Compute.d6(2)));
                    }
                }
            }
            Report.addNewline(vDesc);
            if (onFire && !hex.containsTerrain(Terrains.FIRE)) {
                ignite(c, Terrains.FIRE_LVL_NORMAL, vDesc);
            }
        } else {
            
            r.choose(false);
            vDesc.add(r);
        }
    }

    private void passBuildingWall(Entity entity, Building bldg, Coords lastPos,
                                  Coords curPos, int distance, String why, boolean backwards,
                                  EntityMovementType overallMoveType, boolean entering) {

        Report r;

        if (entity instanceof Protomech) {
            Vector<Report> vBuildingReport = damageBuilding(bldg, 1, curPos);
            for (Report report : vBuildingReport) {
                report.subject = entity.getId();
            }
            addReport(vBuildingReport);
        } else {
            
            PilotingRollData psr = entity.rollMovementInBuilding(bldg,
                                                                 distance, why, overallMoveType);

            
            if (0 < doSkillCheckWhileMoving(entity, entity.getElevation(),
                                            lastPos, curPos, psr, false)) {

                
                int damage = (int) Math.floor(bldg.getDamageFromScale()
                                              * Math.ceil(bldg.getCurrentCF(entering ? curPos
                                                                                     : lastPos) / 10.0));

                
                if (entity instanceof Infantry) {
                    damage = bldg.getType() + 1;
                }
                
                if (damage == 0) {
                    r = new Report(6440);
                    r.add(entity.getDisplayName());
                    r.subject = entity.getId();
                    r.indent(2);
                    addReport(r);
                } else {
                    
                    int side = ToHitData.SIDE_FRONT;
                    if (backwards) {
                        side = ToHitData.SIDE_REAR;
                    }
                    HitData hit = entity.rollHitLocation(ToHitData.HIT_NORMAL,
                                                         side);
                    hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                    addReport(damageEntity(entity, hit, damage));
                }
            }

            if (entity instanceof Infantry) {
                return;
            }
            
            int toBldg = (int) Math.floor(bldg.getDamageToScale()
                                          * Math.ceil(entity.getWeight() / 10.0));
            int curCF = bldg.getCurrentCF(entering ? curPos : lastPos);
            curCF -= Math.min(curCF, toBldg);
            bldg.setCurrentCF(curCF, entering ? curPos : lastPos);

            addReport(damageInfantryIn(bldg, toBldg, entering ? curPos
                                                              : lastPos));
        }
    }

    private boolean checkBuildingCollapseWhileMoving(Building bldg,
                                                     Entity entity, Coords curPos) {
        Coords oldPos = entity.getPosition();
        
        entity.setPosition(curPos);

        Hashtable<Coords, Vector<Entity>> positionMap = game.getPositionMap();

        boolean rv = checkForCollapse(bldg, positionMap, curPos, true,
                                      vPhaseReport);

        
        if (curPos.equals(entity.getPosition()) && !entity.isProne()) {
            entity.setPosition(oldPos);
        }
        return rv;
    }

    public Vector<Report> damageInfantryIn(Building bldg, int damage,
                                           Coords hexCoords) {
        return damageInfantryIn(bldg, damage, hexCoords, WeaponType.WEAPON_NA);
    }

    
    public Vector<Report> damageInfantryIn(Building bldg, int damage,
                                           Coords hexCoords, int infDamageClass) {

        Vector<Report> vDesc = new Vector<Report>();

        if (bldg == null) {
            return vDesc;
        }
        
        float percent = 0.0f;
        Report r;
        switch (bldg.getType()) {
            case Building.LIGHT:
                percent = 0.75f;
                break;
            case Building.MEDIUM:
                percent = 0.5f;
                break;
            case Building.HEAVY:
                percent = 0.25f;
                break;
        }

        
        int toInf = Math.round(damage * percent);

        damage = (int) Math.floor(bldg.getDamageToScale() * toInf);

        boolean foundInfantry = false;

        for (Entity entity : game.getEntitiesVector()) {
            final Coords coords = entity.getPosition();

            
            if ((entity instanceof Infantry) && bldg.isIn(coords)
                && coords.equals(hexCoords)) {

                if (Compute.isInBuilding(game, entity, coords)) {

                    if (toInf == 0) {
                        r = new Report(6445);
                        r.subject = entity.getId();
                        vDesc.addElement(r);
                    } else {
                        
                        r = new Report(6450);
                        r.indent(2);
                        r.subject = entity.getId();
                        r.add(toInf);
                        r.add(entity.getDisplayName());
                        vDesc.addElement(r);
                        
                        if (!(entity instanceof BattleArmor)) {
                            toInf = Compute.directBlowInfantryDamage(toInf, 0,
                                                                     WeaponType.WEAPON_DIRECT_FIRE, false);
                        }
                        int remaining = toInf;
                        int cluster = toInf;
                        
                        if (entity instanceof BattleArmor) {
                            cluster = 5;
                        }
                        while (remaining > 0) {
                            int next = Math.min(cluster, remaining);
                            HitData hit = entity.rollHitLocation(
                                    ToHitData.HIT_NORMAL, ToHitData.SIDE_FRONT);
                            vDesc.addAll((damageEntity(entity, hit, next)));
                            remaining -= next;
                        }
                        Report.addNewline(vDesc);
                    }

                } 

            } 

        } 

        if (foundInfantry) {
            Report.addNewline(vDesc);
        }

        return vDesc;

    } 

    
    public boolean checkForCollapse(Building bldg,
                                    Hashtable<Coords, Vector<Entity>> positionMap, Coords coords,
                                    boolean checkBecauseOfDamage, Vector<Report> vPhaseReport) {

        if ((bldg == null) || (positionMap == null) || positionMap.isEmpty()
            || (coords == null) || !bldg.isIn(coords)
            || !bldg.hasCFIn(coords)) {
            return false;
        }

        int currentCF = bldg.getCurrentCF(coords);

        Hashtable<Coords, Vector<Entity>> basementMap = new Hashtable<Coords, Vector<Entity>>();

        boolean collapse = false;

        boolean basementCollapse = false;

        if (checkBecauseOfDamage && (currentCF <= 0)) {
            collapse = true;
        }

        final Vector<Entity> vector = positionMap.get(coords);

        if (vector != null) {

            
            final IHex curHex = game.getBoard().getHex(coords);
            final int numFloors = Math.max(0,
                                           curHex.terrainLevel(Terrains.BLDG_ELEV));
            final int bridgeEl = curHex.terrainLevel(Terrains.BRIDGE_ELEV);
            int numLoads = numFloors;
            if (bridgeEl != ITerrain.LEVEL_NONE) {
                numLoads++;
            }
            if (numLoads < 1) {
                System.err.println("Check for collapse: hex "
                                   + coords.toString() + " has no bridge or building");
                return false;
            }

            float[] loads = new float[numLoads + 1];
            
            Vector<Entity> basement = new Vector<Entity>();

            boolean recheckLoop = true;
            for (int i = 0; (i < 2) && recheckLoop; i++) {

                recheckLoop = false;
                for (int loop = 0; loop < numLoads; loop++) {
                    loads[loop] = 0;
                }

                Enumeration<Entity> entities = vector.elements();
                while (!collapse && entities.hasMoreElements()) {
                    final Entity entity = entities.nextElement();
                    final int entityElev = entity.getElevation();

                    if (entityElev != bridgeEl) {
                        
                        if (entityElev > numFloors) {
                            continue;
                        }
                    }

                    if ((entity.getMovementMode() == EntityMovementMode.HYDROFOIL)
                        || (entity.getMovementMode() == EntityMovementMode.NAVAL)
                        || (entity.getMovementMode() == EntityMovementMode.SUBMARINE)
                        || (entity.getMovementMode() == EntityMovementMode.INF_UMU)) {
                        continue; 
                    }

                    if (entityElev == 0) {
                        basement.add(entity);
                    }

                    
                    if (entityElev < 0) {
                        continue;
                    }

                    float load = entity.getWeight();
                    int floor = entityElev;
                    if (floor == bridgeEl) {
                        floor = numLoads;
                    }

                    loads[floor] += load;
                    if (loads[floor] > currentCF) {
                       
                        if (floor != 0) {
                            collapse = true;
                        } else if (!bldg.getBasementCollapsed(coords)) {
                            basementCollapse = true;
                        }
                    }
                   
                } 

                if (basementCollapse) {
                    basementMap.put(coords, basement);
                }

            } 
            if (!basementMap.isEmpty()
                && (bldg.getBasement(coords) != BasementType.NONE)
                && !collapse) {

                collapseBasement(bldg, basementMap, coords, vPhaseReport);
                if (currentCF == 0) {
                    collapse = true;
                    recheckLoop = false;
                } else {
                    recheckLoop = true; 
                    
                }
            } else {
                recheckLoop = false; 
                
            }
            if (collapse == true) {
                recheckLoop = false;
              
            }
        }
        
        if (collapse) {
            Report r = new Report(2375, Report.PUBLIC);
            r.add(bldg.getName());
            vPhaseReport.add(r);

            collapseBuilding(bldg, positionMap, coords, vPhaseReport);
        }

        return collapse;

    } 

    public void collapseBuilding(Building bldg,
                                 Hashtable<Coords, Vector<Entity>> positionMap, Coords coords,
                                 Vector<Report> vPhaseReport) {
        collapseBuilding(bldg, positionMap, coords, true, vPhaseReport);
    }

    
    public void collapseBasement(Building bldg,
                                 Hashtable<Coords, Vector<Entity>> positionMap, Coords coords,
                                 Vector<Report> vPhaseReport) {
        if (!bldg.hasCFIn(coords)) {
            return;
        }
        int runningCFTotal;
        runningCFTotal = bldg.getCurrentCF(coords);

        
        final Vector<Entity> entities = positionMap.get(coords);

        if (bldg.getBasement(coords) == BasementType.NONE) {
            return;
        } else {
            bldg.collapseBasement(coords, game.getBoard(), vPhaseReport);
        }

        if (entities != null) {

            Collections.sort(entities, new Comparator<Entity>() {
                public int compare(Entity a, Entity b) {
                    if (a.getElevation() > b.getElevation()) {
                        return -1;
                    } else if (a.getElevation() > b.getElevation()) {
                        return 1;
                    }
                    return 0;
                }
            });
            
            for (Entity entity : entities) {

                int cfDamage = (int) Math
                        .ceil(Math.round(entity.getWeight() / 10.0));

                PilotingRollData psr = entity.getBasePilotingRoll();
                entity.addPilotingModifierForTerrain(psr, coords);

                
                if ((bldg.getBasement(coords) == BasementType.TWO_DEEP_HEAD)
                    || (bldg.getBasement(coords) == BasementType.TWO_DEEP_FEET)) {
                    System.err.println(entity.getDisplayName()
                                       + " is falling 2 floors into " + coords.toString());
                    
                    vPhaseReport.addAll(doEntityFall(entity, coords, 0,
                                                     Compute.d6(), psr, true));
                    runningCFTotal -= cfDamage * 2;
                } else if ((bldg.getBasement(coords) != BasementType.NONE)
                           && (bldg.getBasement(coords) != BasementType.ONE_DEEP_NORMALINFONLY)) {
                    System.err.println(entity.getDisplayName()
                                       + " is falling 1 floor into " + coords.toString());
                    
                    vPhaseReport.addAll(doEntityFall(entity, coords, 0,
                                                     Compute.d6(), psr, true));
                    runningCFTotal -= cfDamage;
                } else {
                    System.err.println(entity.getDisplayName()
                                       + " is not falling into " + coords.toString());
                }

                entityUpdate(entity.getId());

            } 

        } 

        if (runningCFTotal < 0) {
            bldg.setCurrentCF(0, coords);
            bldg.setPhaseCF(0, coords);
        } else {
            bldg.setCurrentCF(runningCFTotal, coords);
            bldg.setPhaseCF(runningCFTotal, coords);
        }
        sendChangedHex(coords);
        Vector<Building> buildings = new Vector<Building>();
        buildings.add(bldg);
        sendChangedBuildings(buildings);
    }

    public void collapseBuilding(Building bldg,
                                 Hashtable<Coords, Vector<Entity>> positionMap, Coords coords,
                                 boolean collapseAll, Vector<Report> vPhaseReport) {
        if (!bldg.hasCFIn(coords)) {
            return;
        }
        
        Report r;
        final int phaseCF = bldg.getPhaseCF(coords);
        
        final Vector<Entity> vector = positionMap.get(coords);

        if (vector != null) {

            
            final IHex curHex = game.getBoard().getHex(coords);
            final int bridgeEl = curHex.terrainLevel(Terrains.BRIDGE_ELEV);
            final int numFloors = Math.max(bridgeEl,
                                           curHex.terrainLevel(Terrains.BLDG_ELEV));

            bldg.setCurrentCF(0, coords);
            bldg.setPhaseCF(0, coords);
            send(createCollapseBuildingPacket(coords));
            game.getBoard().collapseBuilding(coords);

            Collections.sort(vector, new Comparator<Entity>() {
                public int compare(Entity a, Entity b) {
                    if (a.getElevation() > b.getElevation()) {
                        return -1;
                    } else if (a.getElevation() > b.getElevation()) {
                        return 1;
                    }
                    return 0;
                }
            });
            
            Enumeration<Entity> entities = vector.elements();
            while (entities.hasMoreElements()) {
                final Entity entity = entities.nextElement();
                
                if (entity instanceof GunEmplacement) {
                    vPhaseReport.addAll(destroyEntity(entity,
                                                      "building collapse"));
                    addNewLines();
                    continue;
                }

                int floor = entity.getElevation();
                
                
                if (floor < 0) {
                    vPhaseReport.addAll(destroyEntity(entity,
                                                      "Crushed under building rubble", false, false));
                }

                if (floor > numFloors) {
                    continue;
                }

                if (floor == numFloors) {
                    floor--;
                }
                int damage = (int) Math.floor(bldg.getDamageFromScale()
                                              * Math.ceil((phaseCF * (numFloors - floor)) / 10.0));

                if (entity instanceof Infantry) {
                    if ((entity instanceof BattleArmor)
                        || ((Infantry) entity).isMechanized()) {
                        damage *= 2;
                    } else {
                        damage *= 3;
                    }
                }

                r = new Report(6455);
                r.indent();
                r.subject = entity.getId();
                r.add(entity.getDisplayName());
                r.add(damage);
                vPhaseReport.add(r);
                int remaining = damage;
                int cluster = damage;
                if ((entity instanceof BattleArmor) || (entity instanceof Mech)
                    || (entity instanceof Tank)) {
                    cluster = 5;
                }
                while (remaining > 0) {
                    int next = Math.min(cluster, remaining);
                    int table;
                    if (entity instanceof Protomech) {
                        table = ToHitData.HIT_SPECIAL_PROTO;
                    } else if (entity.getElevation() == numFloors) {
                        table = ToHitData.HIT_NORMAL;
                    } else {
                        table = ToHitData.HIT_PUNCH;
                    }
                    HitData hit = entity.rollHitLocation(table,
                                                         ToHitData.SIDE_FRONT);
                    hit.setGeneralDamageType(HitData.DAMAGE_PHYSICAL);
                    vPhaseReport.addAll(damageEntity(entity, hit, next));
                    remaining -= next;
                }
                vPhaseReport.add(new Report(1210, Report.PUBLIC));

                floor = entity.getElevation();
                if ((floor > 0) || (floor == bridgeEl)) {
                    
                    PilotingRollData psr = entity.getBasePilotingRoll();
                    entity.addPilotingModifierForTerrain(psr, coords);
                    if (damage >= 20) {
                        psr.addModifier(1, "20+ damage");
                    }
                    vPhaseReport.addAll(doEntityFallsInto(entity, coords, psr,
                                                          true));
                }
                
                entityUpdate(entity.getId());

            } 

        } 
        else {
            
            bldg.setCurrentCF(0, coords);
            bldg.setPhaseCF(0, coords);
            send(createCollapseBuildingPacket(coords));
            game.getBoard().collapseBuilding(coords);
        }
        
        if (bldg.getCollapsedHexCount() > (bldg.getOriginalHexCount() / 2)) {
            for (Enumeration<Coords> coordsEnum = bldg.getCoords(); coordsEnum
                    .hasMoreElements(); ) {
                coords = coordsEnum.nextElement();
                collapseBuilding(bldg, game.getPositionMap(), coords, false,
                                 vPhaseReport);
            }
        }

    } 

    private Packet createCollapseBuildingPacket(Coords coords) {
        Vector<Coords> coordsV = new Vector<Coords>();
        coordsV.addElement(coords);
        return createCollapseBuildingPacket(coordsV);
    }

    
    private Packet createCollapseBuildingPacket(Vector<Coords> coords) {
        return new Packet(Packet.COMMAND_BLDG_COLLAPSE, coords);
    }

    
    private Packet createUpdateBuildingPacket(Vector<Building> buildings) {
        return new Packet(Packet.COMMAND_BLDG_UPDATE, buildings);
    }

    
    private void applyBuildingDamage() {

        Map<Building, Vector<Coords>> collapse = new HashMap<Building, Vector<Coords>>();
        Map<Building, Vector<Coords>> update = new HashMap<Building, Vector<Coords>>();
        Enumeration<Building> buildings = game.getBoard().getBuildings();
        while (buildings.hasMoreElements()) {
            Building bldg = buildings.nextElement();
            Vector<Coords> collapseCoords = new Vector<Coords>();
            Vector<Coords> updateCoords = new Vector<Coords>();
            Enumeration<Coords> buildingCoords = bldg.getCoords();
            while (buildingCoords.hasMoreElements()) {
                Coords coords = buildingCoords.nextElement();
                
                if (bldg.getCurrentCF(coords) == 0) {
                    collapseCoords.addElement(coords);
                }
                
                else if (bldg.getPhaseCF(coords) != bldg.getCurrentCF(coords)) {
                    bldg.setPhaseCF(bldg.getCurrentCF(coords), coords);
                    updateCoords.addElement(coords);
                }
            }
            collapse.put(bldg, collapseCoords);
            update.put(bldg, updateCoords);
        } 

        if (!collapse.isEmpty()) {

            Hashtable<Coords, Vector<Entity>> positionMap = game
                    .getPositionMap();

            for (Building bldg : collapse.keySet()) {
                Vector<Coords> coordsVector = collapse.get(bldg);
                for (Coords coords : coordsVector) {
                    Report r = new Report(6460, Report.PUBLIC);
                    r.add(bldg.getName());
                    addReport(r);
                    collapseBuilding(bldg, positionMap, coords, vPhaseReport);
                }
            }
        }

        if (!update.isEmpty()) {
            Hashtable<Coords, Vector<Entity>> positionMap = game
                    .getPositionMap();
            for (Building bldg : update.keySet()) {
                Vector<Coords> updateCoords = update.get(bldg);
                Vector<Coords> coordsToRemove = new Vector<Coords>();
                for (Coords coords : updateCoords) {
                    if (checkForCollapse(bldg, positionMap, coords, false,
                                         vPhaseReport)) {
                        coordsToRemove.add(coords);
                    }
                }
                updateCoords.removeAll(coordsToRemove);
                update.put(bldg, updateCoords);
            }
        }

        if (!update.isEmpty()) {
            sendChangedBuildings(new Vector<Building>(update.keySet()));
        }
    }

    public Vector<Report> damageBuilding(Building bldg, int damage,
                                         Coords coords) {
        final String defaultWhy = " absorbs ";
        return damageBuilding(bldg, damage, defaultWhy, coords);
    }

    
    public Vector<Report> damageBuilding(Building bldg, int damage, String why,
                                         Coords coords) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        Report r = new Report(1210, Report.PUBLIC);

        
        if ((bldg != null) && (damage > 0)) {
            r.messageId = 3435;
            r.add(bldg.getName());
            r.add(why);
            r.add(damage);
            vPhaseReport.add(r);
            int curArmor = bldg.getArmor(coords);
            if (curArmor >= damage) {
                curArmor -= Math.min(curArmor, damage);
                bldg.setArmor(curArmor, coords);
                r = new Report(3436, Report.PUBLIC);
                r.indent(0);
                r.add(damage);
                r.add(curArmor);
                vPhaseReport.add(r);
            } else {
                r.add(damage);
                if (curArmor > 0) {
                    bldg.setArmor(0, coords);
                    damage = damage - curArmor;
                    r = new Report(3436, Report.PUBLIC);
                    r.indent(0);
                    r.add(curArmor);
                    r.add(0);
                    vPhaseReport.add(r);
                }
                damage = (int) Math.floor(bldg.getDamageToScale() * damage);
                if (bldg.getDamageToScale() < 1.0) {
                    r = new Report(3437, Report.PUBLIC);
                    r.indent(0);
                    r.add(damage);
                    vPhaseReport.add(r);
                }
                if (bldg.getDamageToScale() > 1.0) {
                    r = new Report(3438, Report.PUBLIC);
                    r.indent(0);
                    r.add(damage);
                    vPhaseReport.add(r);
                }
                int curCF = bldg.getCurrentCF(coords);
                final int startingCF = curCF;
                curCF -= Math.min(curCF, damage);
                bldg.setCurrentCF(curCF, coords);
                final int damageThresh = (int) Math.ceil(bldg
                                                                 .getPhaseCF(coords) / 10.0);

                
                if ((curCF == 0) && (startingCF != 0)) {
                    if (bldg instanceof FuelTank) {
                        
                        r = new Report(3441);
                        r.type = Report.PUBLIC;
                        r.indent(0);
                        vPhaseReport.add(r);
                        r = new Report(3560);
                        r.type = Report.PUBLIC;
                        r.newlines = 1;
                        vPhaseReport.add(r);
                        Vector<Report> vRep = new Vector<Report>();
                        doExplosion(((FuelTank) bldg).getMagnitude(), 10,
                                    false, bldg.getCoords().nextElement(), true,
                                    vRep, null, -1);
                        Report.indentAll(vRep, 2);
                        vPhaseReport.addAll(vRep);
                        return vPhaseReport;
                    }
                    if (bldg.getType() == Building.WALL) {
                        r = new Report(3442);
                        r.type = Report.PUBLIC;
                        r.indent(0);
                        vPhaseReport.add(r);
                    } else {
                        r = new Report(3440);
                        r.type = Report.PUBLIC;
                        r.indent(0);
                        vPhaseReport.add(r);
                    }
                } else if ((curCF < startingCF) && (damage > damageThresh)) {
                    
                    
                    Vector<GunEmplacement> guns = game
                            .getGunEmplacements(coords);
                    if (guns.size() > 0) {
                        vPhaseReport.addAll(criticalGunEmplacement(guns, bldg,
                                                                   coords));
                    }
                }
            }
        }
        Report.indentAll(vPhaseReport, 2);
        return vPhaseReport;
    }

    private Vector<Report> criticalGunEmplacement(Vector<GunEmplacement> guns,
                                                  Building bldg, Coords coords) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;
        r = new Report(3800);
        r.type = Report.PUBLIC;
        r.indent(0);
        vDesc.add(r);

        int critroll = Compute.d6(2);
        if (critroll < 6) {
            r = new Report(3805);
            r.type = Report.PUBLIC;
            r.indent(1);
            vDesc.add(r);
        } else if (critroll == 6) {
            
            Vector<Mounted> wpns = new Vector<Mounted>();
            for (GunEmplacement gun : guns) {
                for (Mounted wpn : gun.getWeaponList()) {
                    if (!wpn.isHit() && !wpn.isJammed()
                        && !wpn.jammedThisPhase()) {
                        wpns.add(wpn);
                    }
                }
            }
            if (wpns.size() > 0) {
                Mounted weapon = wpns.elementAt(Compute.randomInt(wpns.size()));
                weapon.setJammed(true);
                ((GunEmplacement) weapon.getEntity()).addJammedWeapon(weapon);
                r = new Report(3845);
                r.type = Report.PUBLIC;
                r.indent(1);
                r.add(weapon.getDesc());
                vDesc.add(r);
            } else {
                r = new Report(3846);
                r.type = Report.PUBLIC;
                r.indent(1);
                vDesc.add(r);
            }
        } else if (critroll == 7) {
            
            for (GunEmplacement gun : guns) {
                gun.stunCrew();
                r = new Report(3810);
                r.type = Report.PUBLIC;
                r.indent(1);
                vDesc.add(r);
            }
        } else if (critroll == 8) {
            
            Vector<Mounted> wpns = new Vector<Mounted>();
            for (GunEmplacement gun : guns) {
                for (Mounted wpn : gun.getWeaponList()) {
                    if (!wpn.isHit()) {
                        wpns.add(wpn);
                    }
                }
            }
            if (wpns.size() > 0) {
                Mounted weapon = wpns.elementAt(Compute.randomInt(wpns.size()));
                weapon.setHit(true);
                r = new Report(3840);
                r.type = Report.PUBLIC;
                r.indent(1);
                r.add(weapon.getDesc());
                vDesc.add(r);
            } else {
                r = new Report(3841);
                r.type = Report.PUBLIC;
                r.indent(1);
                vDesc.add(r);
            }
        } else if (critroll == 9) {
            
            r = new Report(3815);
            r.type = Report.PUBLIC;
            r.indent(1);
            vDesc.add(r);
            for (GunEmplacement gun : guns) {
                gun.getCrew().setDoomed(true);
            }
        } else if (critroll == 10) {
            if (Compute.d6() > 3) {
                
                r = new Report(3820);
                r.type = Report.PUBLIC;
                r.indent(1);
                vDesc.add(r);
                for (GunEmplacement gun : guns) {
                    gun.lockTurret(gun.getLocTurret());
                }
            } else {
                
                r = new Report(3825);
                r.type = Report.PUBLIC;
                r.indent(1);
                vDesc.add(r);
                for (GunEmplacement gun : guns) {
                    if (gun.isTurretEverJammed(gun.getLocTurret())) {
                        gun.lockTurret(gun.getLocTurret());
                    } else {
                        gun.jamTurret(gun.getLocTurret());
                    }
                }
            }
        } else if (critroll == 11) {
            r = new Report(3830);
            r.type = Report.PUBLIC;
            r.indent(1);
            r.add(bldg.getName());
            int boom = 0;
            for (GunEmplacement gun : guns) {
                for (Mounted ammo : gun.getAmmo()) {
                    ammo.setHit(true);
                    if (ammo.getType().isExplosive(ammo)) {
                        boom += ammo.getHittableShotsLeft()
                                * ((AmmoType) ammo.getType())
                                .getDamagePerShot()
                                * ((AmmoType) ammo.getType()).getRackSize();
                    }
                }
            }
            boom = (int) Math.floor(bldg.getDamageToScale() * boom);
            r.add(boom);
            int curCF = bldg.getCurrentCF(coords);
            curCF -= Math.min(curCF, boom);
            bldg.setCurrentCF(curCF, coords);
            r.add(bldg.getCurrentCF(coords));
            vDesc.add(r);
            
            if ((curCF == 0) && (bldg.getPhaseCF(coords) != 0)) {
                if (bldg instanceof FuelTank) {
                    
                    
                    r = new Report(3441);
                    r.type = Report.PUBLIC;
                    r.indent(0);
                    vDesc.add(r);
                    
                    
                    r = new Report(3560);
                    r.type = Report.PUBLIC;
                    r.newlines = 1;
                    vDesc.add(r);
                    Vector<Report> vRep = new Vector<Report>();
                    doExplosion(((FuelTank) bldg).getMagnitude(), 10, false,
                                bldg.getCoords().nextElement(), true, vRep, null,
                                -1);
                    Report.indentAll(vRep, 2);
                    vDesc.addAll(vRep);
                    return vPhaseReport;
                }
                if (bldg.getType() == Building.WALL) {
                    r = new Report(3442);
                    r.type = Report.PUBLIC;
                    r.indent(0);
                    vDesc.add(r);
                } else {
                    r = new Report(3440);
                    r.type = Report.PUBLIC;
                    r.indent(0);
                    vDesc.add(r);
                }
            }
        } else if (critroll == 12) {
            
            r = new Report(3835);
            r.type = Report.PUBLIC;
            r.indent(1);
            vDesc.add(r);
        }
        return vDesc;
    }

    public void sendChangedBuildings(Vector<Building> buildings) {
        send(createUpdateBuildingPacket(buildings));
    }

    
    private void receiveUnloadStranded(Packet packet, int connId) {
        GameTurn.UnloadStrandedTurn turn = null;
        final IPlayer player = game.getPlayer(connId);
        int[] entityIds = (int[]) packet.getObject(0);
        Vector<IPlayer> declared = null;
        IPlayer other = null;
        Enumeration<EntityAction> pending = null;
        UnloadStrandedAction action = null;
        Entity entity = null;

        if (game.getPhase() != IGame.Phase.PHASE_MOVEMENT) {
            System.err
                    .println("error: server got unload stranded packet in wrong phase");
            return;
        }

        if (game.getTurn() instanceof GameTurn.UnloadStrandedTurn) {
            turn = (GameTurn.UnloadStrandedTurn) game.getTurn();
        } else {
            System.err
                    .println("error: server got unload stranded packet out of sequence");
            StringBuffer message = new StringBuffer();
            message.append(player.getName())
                   .append(" should not be sending 'unload stranded entity' packets at this time.");
            sendServerChat(message.toString());
            return;
        }
        if (!turn.isValid(connId, game)) {
            System.err
                    .println("error: server got unload stranded packet from invalid player");
            StringBuffer message = new StringBuffer();
            message.append(player.getName()).append(
                    " should not be sending 'unload stranded entity' packets.");
            sendServerChat(message.toString());
            return;
        }

        declared = new Vector<IPlayer>();
        pending = game.getActions();
        while (pending.hasMoreElements()) {
            action = (UnloadStrandedAction) pending.nextElement();
            if (action.getPlayerId() == connId) {
                System.err
                        .println("error: server got multiple unload stranded packets from player");
                StringBuffer message = new StringBuffer();
                message.append(player.getName())
                       .append(" should not send multiple 'unload stranded entity' packets.");
                sendServerChat(message.toString());
                return;
            }
            
            other = game.getPlayer(action.getPlayerId());
            if (!declared.contains(other)) {
                declared.addElement(other);
            }
        } 

        boolean foundValid = false;
        for (int index = 0; (null != entityIds) && (index < entityIds.length); index++) {
            entity = game.getEntity(entityIds[index]);
            if (!game.getTurn().isValid(connId, entity, game)) {
                System.err
                        .println("error: server got unload stranded packet for invalid entity");
                StringBuffer message = new StringBuffer();
                message.append(player.getName()).append(
                        " can not unload stranded entity ");
                if (null == entity) {
                    message.append('#').append(entityIds[index]);
                } else {
                    message.append(entity.getDisplayName());
                }
                message.append(" at this time.");
                sendServerChat(message.toString());
            } else {
                foundValid = true;
                game.addAction(new UnloadStrandedAction(connId,
                                                        entityIds[index]));
            }
        }

        if (!foundValid) {
            game.addAction(new UnloadStrandedAction(connId, Entity.NONE));
        }

        declared.addElement(player);

        entityIds = turn.getEntityIds();
        for (int entityId : entityIds) {
            entity = game.getEntity(entityId);
            other = entity.getOwner();
            if (!declared.contains(other)) {
                
                return;
            }
        }

        pending = game.getActions();
        while (pending.hasMoreElements()) {
            action = (UnloadStrandedAction) pending.nextElement();

            if (Entity.NONE != action.getEntityId()) {
                entity = game.getEntity(action.getEntityId());
                if (null == entity) {
                    
                    System.err
                            .print("error: server could not find stranded entity #");
                    System.err.print(action.getEntityId());
                    System.err.println(" to unload!!!");
                } else {
                    
                    Entity transporter = game
                            .getEntity(entity.getTransportId());
                    unloadUnit(transporter, entity, transporter.getPosition(),
                               transporter.getFacing(), transporter.getElevation());
                }
            }

        } 

        game.resetActions();
        changeToNextTurn();
    }

    private void clearArtillerySpotters(int entityID, int weaponID) {
        for (Enumeration<AttackHandler> i = game.getAttacks(); i
                .hasMoreElements(); ) {
            WeaponHandler wh = (WeaponHandler) i.nextElement();
            if ((wh.waa instanceof ArtilleryAttackAction)
                && (wh.waa.getEntityId() == entityID)
                && (wh.waa.getWeaponId() == weaponID)) {
                ArtilleryAttackAction aaa = (ArtilleryAttackAction) wh.waa;
                aaa.setSpotterIds(null);
            }
        }
    }

    public void creditKill(Entity target, Entity attacker) {
        
        if (target instanceof FighterSquadron) {
            return;
        }
        if ((target.isDoomed() || target.getCrew().isDoomed())
            && !target.getGaveKillCredit() && (attacker != null)) {
            attacker.addKill(target);
        }
    }

    private PhysicalResult preTreatPhysicalAttack(AbstractAttackAction aaa) {
        final Entity ae = game.getEntity(aaa.getEntityId());
        int damage = 0;
        PhysicalResult pr = new PhysicalResult();
        ToHitData toHit = new ToHitData();
        pr.roll = Compute.d6(2);
        pr.aaa = aaa;
        if (aaa instanceof BrushOffAttackAction) {
            BrushOffAttackAction baa = (BrushOffAttackAction) aaa;
            int arm = baa.getArm();
            baa.setArm(BrushOffAttackAction.LEFT);
            toHit = BrushOffAttackAction.toHit(game, aaa.getEntityId(),
                                               aaa.getTarget(game), BrushOffAttackAction.LEFT);
            baa.setArm(BrushOffAttackAction.RIGHT);
            pr.toHitRight = BrushOffAttackAction.toHit(game, aaa.getEntityId(),
                                                       aaa.getTarget(game), BrushOffAttackAction.RIGHT);
            damage = BrushOffAttackAction.getDamageFor(ae,
                                                       BrushOffAttackAction.LEFT);
            pr.damageRight = BrushOffAttackAction.getDamageFor(ae,
                                                               BrushOffAttackAction.RIGHT);
            baa.setArm(arm);
            pr.rollRight = Compute.d6(2);
        } else if (aaa instanceof ChargeAttackAction) {
            ChargeAttackAction caa = (ChargeAttackAction) aaa;
            toHit = caa.toHit(game);
            if (caa.getTarget(game) instanceof Entity) {
                Entity target = (Entity) caa.getTarget(game);
                damage = ChargeAttackAction.getDamageFor(ae, target, game
                                                                 .getOptions().booleanOption("tacops_charge_damage"),
                                                         toHit.getMoS());
            } else {
                damage = ChargeAttackAction.getDamageFor(ae);
            }
        } else if (aaa instanceof ClubAttackAction) {
            ClubAttackAction caa = (ClubAttackAction) aaa;
            toHit = caa.toHit(game);
            damage = ClubAttackAction.getDamageFor(
                    ae,
                    caa.getClub(),
                    (caa.getTarget(game) instanceof Infantry)
                    && !(caa.getTarget(game) instanceof BattleArmor));
            if (caa.getTargetType() == Targetable.TYPE_BUILDING) {
                EquipmentType clubType = caa.getClub().getType();
                if (clubType.hasSubType(MiscType.S_BACKHOE)
                        || clubType.hasSubType(MiscType.S_CHAINSAW)
                        || clubType.hasSubType(MiscType.S_MINING_DRILL)
                        || clubType.hasSubType(MiscType.S_PILE_DRIVER)) {
                    damage += Compute.d6(1);
                } else if (clubType.hasSubType(MiscType.S_DUAL_SAW)) {
                    damage += Compute.d6(2);
                } else if (clubType.hasSubType(MiscType.S_ROCK_CUTTER)) {
                    damage += Compute.d6(3);
                }
                else if (clubType.hasSubType(MiscType.S_WRECKING_BALL)) {
                    damage += Compute.d6(4);
                }
            }
        } else if (aaa instanceof DfaAttackAction) {
            DfaAttackAction daa = (DfaAttackAction) aaa;
            toHit = daa.toHit(game);
            damage = DfaAttackAction.getDamageFor(
                    ae,
                    (daa.getTarget(game) instanceof Infantry)
                    && !(daa.getTarget(game) instanceof BattleArmor));
        } else if (aaa instanceof KickAttackAction) {
            KickAttackAction kaa = (KickAttackAction) aaa;
            toHit = kaa.toHit(game);
            damage = KickAttackAction.getDamageFor(
                    ae,
                    kaa.getLeg(),
                    (kaa.getTarget(game) instanceof Infantry)
                    && !(kaa.getTarget(game) instanceof BattleArmor));
        } else if (aaa instanceof ProtomechPhysicalAttackAction) {
            ProtomechPhysicalAttackAction paa = (ProtomechPhysicalAttackAction) aaa;
            toHit = paa.toHit(game);
            damage = ProtomechPhysicalAttackAction.getDamageFor(ae,
                                                                paa.getTarget(game));
        } else if (aaa instanceof PunchAttackAction) {
            PunchAttackAction paa = (PunchAttackAction) aaa;
            int arm = paa.getArm();
            int damageRight = 0;
            paa.setArm(PunchAttackAction.LEFT);
            toHit = paa.toHit(game);
            paa.setArm(PunchAttackAction.RIGHT);
            ToHitData toHitRight = paa.toHit(game);
            damage = PunchAttackAction.getDamageFor(
                    ae,
                    PunchAttackAction.LEFT,
                    (paa.getTarget(game) instanceof Infantry)
                    && !(paa.getTarget(game) instanceof BattleArmor));
            damageRight = PunchAttackAction.getDamageFor(
                    ae,
                    PunchAttackAction.RIGHT,
                    (paa.getTarget(game) instanceof Infantry)
                    && !(paa.getTarget(game) instanceof BattleArmor));
            paa.setArm(arm);
            
            
            if (ae.isProne()) {
                double oddsLeft = Compute.oddsAbove(toHit.getValue(),
                                                    ae.getCrew().getOptions().booleanOption(OptionsConstants
                                                                                                    .PILOT_APTITUDE_PILOTING));
                double oddsRight = Compute.oddsAbove(toHitRight.getValue(),
                                                     ae.getCrew().getOptions().booleanOption(OptionsConstants
                                                                                                     .PILOT_APTITUDE_PILOTING));
                
                if ((oddsLeft * damage) > (oddsRight * damageRight)) {
                    paa.setArm(PunchAttackAction.LEFT);
                } else {
                    paa.setArm(PunchAttackAction.RIGHT);
                }
            }
            pr.damageRight = damageRight;
            pr.toHitRight = toHitRight;
            pr.rollRight = Compute.d6(2);
        } else if (aaa instanceof PushAttackAction) {
            PushAttackAction paa = (PushAttackAction) aaa;
            toHit = paa.toHit(game);
        } else if (aaa instanceof TripAttackAction) {
            TripAttackAction paa = (TripAttackAction) aaa;
            toHit = paa.toHit(game);
        } else if (aaa instanceof LayExplosivesAttackAction) {
            LayExplosivesAttackAction leaa = (LayExplosivesAttackAction) aaa;
            toHit = leaa.toHit(game);
            damage = LayExplosivesAttackAction.getDamageFor(ae);
        } else if (aaa instanceof ThrashAttackAction) {
            ThrashAttackAction taa = (ThrashAttackAction) aaa;
            toHit = taa.toHit(game);
            damage = ThrashAttackAction.getDamageFor(ae);
        } else if (aaa instanceof JumpJetAttackAction) {
            JumpJetAttackAction jaa = (JumpJetAttackAction) aaa;
            toHit = jaa.toHit(game);
            if (jaa.getLeg() == JumpJetAttackAction.BOTH) {
                damage = JumpJetAttackAction.getDamageFor(ae,
                                                          JumpJetAttackAction.LEFT);
                pr.damageRight = JumpJetAttackAction.getDamageFor(ae,
                                                                  JumpJetAttackAction.LEFT);
            } else {
                damage = JumpJetAttackAction.getDamageFor(ae, jaa.getLeg());
                pr.damageRight = 0;
            }
            ae.heatBuildup += (damage + pr.damageRight) / 3;
        } else if (aaa instanceof GrappleAttackAction) {
            GrappleAttackAction taa = (GrappleAttackAction) aaa;
            toHit = taa.toHit(game);
        } else if (aaa instanceof BreakGrappleAttackAction) {
            BreakGrappleAttackAction taa = (BreakGrappleAttackAction) aaa;
            toHit = taa.toHit(game);
        } else if (aaa instanceof RamAttackAction) {
            RamAttackAction raa = (RamAttackAction) aaa;
            toHit = raa.toHit(game);
            damage = RamAttackAction.getDamageFor((Aero) ae,
                                                  (Aero) aaa.getTarget(game));
        } else if (aaa instanceof TeleMissileAttackAction) {
            TeleMissileAttackAction taa = (TeleMissileAttackAction) aaa;
            toHit = taa.toHit(game);
            damage = TeleMissileAttackAction.getDamageFor(ae);
        } else if (aaa instanceof BAVibroClawAttackAction) {
            BAVibroClawAttackAction bvca = (BAVibroClawAttackAction) aaa;
            toHit = bvca.toHit(game);
            damage = BAVibroClawAttackAction.getDamageFor(ae);
        }
        pr.toHit = toHit;
        pr.damage = damage;
        return pr;
    }

    private void resolvePhysicalAttack(PhysicalResult pr, int cen) {
        AbstractAttackAction aaa = pr.aaa;
        if (aaa instanceof PunchAttackAction) {
            PunchAttackAction paa = (PunchAttackAction) aaa;
            if (paa.getArm() == PunchAttackAction.BOTH) {
                paa.setArm(PunchAttackAction.LEFT);
                pr.aaa = paa;
                resolvePunchAttack(pr, cen);
                cen = paa.getEntityId();
                paa.setArm(PunchAttackAction.RIGHT);
                pr.aaa = paa;
                resolvePunchAttack(pr, cen);
            } else {
                resolvePunchAttack(pr, cen);
                cen = paa.getEntityId();
            }
        } else if (aaa instanceof KickAttackAction) {
            resolveKickAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof BrushOffAttackAction) {
            BrushOffAttackAction baa = (BrushOffAttackAction) aaa;
            if (baa.getArm() == BrushOffAttackAction.BOTH) {
                baa.setArm(BrushOffAttackAction.LEFT);
                pr.aaa = baa;
                resolveBrushOffAttack(pr, cen);
                cen = baa.getEntityId();
                baa.setArm(BrushOffAttackAction.RIGHT);
                pr.aaa = baa;
                resolveBrushOffAttack(pr, cen);
            } else {
                resolveBrushOffAttack(pr, cen);
                cen = baa.getEntityId();
            }
        } else if (aaa instanceof ThrashAttackAction) {
            resolveThrashAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof ProtomechPhysicalAttackAction) {
            resolveProtoAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof ClubAttackAction) {
            resolveClubAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof PushAttackAction) {
            resolvePushAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof ChargeAttackAction) {
            resolveChargeAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof DfaAttackAction) {
            resolveDfaAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof LayExplosivesAttackAction) {
            resolveLayExplosivesAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof TripAttackAction) {
            resolveTripAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof JumpJetAttackAction) {
            resolveJumpJetAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof GrappleAttackAction) {
            resolveGrappleAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof BreakGrappleAttackAction) {
            resolveBreakGrappleAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof RamAttackAction) {
            resolveRamAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof TeleMissileAttackAction) {
            resolveTeleMissileAttack(pr, cen);
            cen = aaa.getEntityId();
        } else if (aaa instanceof BAVibroClawAttackAction) {
            resolveBAVibroClawAttack(pr, cen);
            cen = aaa.getEntityId();
        } else {
            
        }
        
        Targetable target = game.getTarget(aaa.getTargetType(),
                                           aaa.getTargetId());
        if (target instanceof Entity) {
            Entity targetEntity = (Entity) target;
            targetEntity.setStruck(true);
            targetEntity.addAttackedByThisTurn(target.getTargetId());
            creditKill(targetEntity, game.getEntity(cen));
        }
    }

    
    private void checkExtremeGravityMovement(Entity entity, MoveStep step,
                                             Coords curPos, int cachedMaxMPExpenditure) {
        PilotingRollData rollTarget;
        if (game.getPlanetaryConditions().getGravity() != 1) {
            if (entity instanceof Mech) {
                if ((step.getMovementType() == EntityMovementType.MOVE_WALK)
                    || (step.getMovementType() == EntityMovementType.MOVE_VTOL_WALK)
                    || (step.getMovementType() == EntityMovementType.MOVE_RUN)
                    || (step.getMovementType() == EntityMovementType.MOVE_SPRINT)
                    || (step.getMovementType() == EntityMovementType.MOVE_VTOL_RUN)) {
                    if (step.getMpUsed() > cachedMaxMPExpenditure) {
                        
                        
                        game.addExtremeGravityPSR(entity
                                                          .checkMovedTooFast(step));
                    }
                } else if (step.getMovementType() == EntityMovementType.MOVE_JUMP) {
                    System.err.println("gravity move check jump: "
                                       + step.getMpUsed() + "/" + cachedMaxMPExpenditure);
                    System.err.flush();
                    if (step.getMpUsed() > cachedMaxMPExpenditure) {
                        
                        game.addExtremeGravityPSR(entity
                                                          .checkMovedTooFast(step));
                    } else if (game.getPlanetaryConditions().getGravity() > 1) {
                        
                        rollTarget = entity.getBasePilotingRoll(step
                                                                        .getParent().getLastStepMovementType());
                        entity.addPilotingModifierForTerrain(rollTarget, step);
                        rollTarget.append(new PilotingRollData(entity.getId(),
                                                               0, "jumped in high gravity"));
                        game.addExtremeGravityPSR(rollTarget);
                    }
                }
            } else if (entity instanceof Tank) {
                if ((step.getMovementType() == EntityMovementType.MOVE_WALK)
                    || (step.getMovementType() == EntityMovementType.MOVE_VTOL_WALK)
                    || (step.getMovementType() == EntityMovementType.MOVE_RUN)
                    || (step.getMovementType() == EntityMovementType.MOVE_VTOL_RUN)) {
                    
                    if ((step.getMpUsed() > cachedMaxMPExpenditure)
                        && !step.isOnlyPavement()) {
                        game.addExtremeGravityPSR(entity
                                                          .checkMovedTooFast(step));
                    } else if (step.getMpUsed() > (cachedMaxMPExpenditure + 1)) {
                        
                        game.addExtremeGravityPSR(entity
                                                          .checkMovedTooFast(step));
                    } 
                }
            }
        }
    }

    
    private Vector<Report> doExtremeGravityDamage(Entity entity, int damage) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        HitData hit;
        if (entity instanceof BipedMech) {
            for (int i = 6; i <= 7; i++) {
                hit = new HitData(i);
                vPhaseReport.addAll(damageEntity(entity, hit, damage, false,
                                                 DamageType.NONE, true));
            }
        }
        if (entity instanceof QuadMech) {
            for (int i = 4; i <= 7; i++) {
                hit = new HitData(i);
                vPhaseReport.addAll(damageEntity(entity, hit, damage, false,
                                                 DamageType.NONE, true));
            }
        } else if (entity instanceof Tank) {
            hit = new HitData(Tank.LOC_FRONT);
            vPhaseReport.addAll(damageEntity(entity, hit, damage, false,
                                             DamageType.NONE, true));
            vPhaseReport.addAll(vehicleMotiveDamage((Tank) entity, 0));
        }
        return vPhaseReport;
    }

    
    public Vector<Report> ejectEntity(Entity entity, boolean autoEject) {
        return ejectEntity(entity, autoEject, false);
    }

    
    public Vector<Report> ejectEntity(Entity entity, boolean autoEject,
                                      boolean skin_of_the_teeth) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if (entity.getCrew().isEjected()) {
            return vDesc;
        }

        if (entity.isCarcass()) {
            return vDesc;
        }

        if (entity instanceof Mech) {
            PilotingRollData rollTarget = getEjectModifiers(game, entity,
                    autoEject);
            int facing = entity.getFacing();
            Coords targetCoords = entity.getPosition().translated(
                    (facing + 3) % 6);

            if (autoEject) {
                r = new Report(6395);
                r.subject = entity.getId();
                r.addDesc(entity);
                r.indent(2);
                vDesc.addElement(r);
            }
            
            r = new Report(2180);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(rollTarget.getLastPlainDesc(), true);
            r.indent();
            vDesc.addElement(r);
            
            final int diceRoll = entity.getCrew().rollPilotingSkill();
            r = new Report(2190);
            r.subject = entity.getId();
            r.add(rollTarget.getValueAsString());
            r.add(rollTarget.getDesc());
            r.add(diceRoll);
            r.indent();
            
            MechWarrior pilot = new MechWarrior(entity);
            pilot.setDeployed(true);
            pilot.setId(getFreeEntityId());
            pilot.setLanded(false);
            game.addEntity(pilot);
            send(createAddEntityPacket(pilot.getId()));
            
            pilot.setDone(true);
            if (diceRoll < rollTarget.getValue()) {
                r.choose(false);
                vDesc.addElement(r);
                Report.addNewline(vDesc);
                if ((rollTarget.getValue() - diceRoll) > 1) {
                    int damage = (rollTarget.getValue() - diceRoll) / 2;
                    if (entity.hasQuirk(OptionsConstants.QUIRK_NEG_DIFFICULT_EJECT)) {
                        damage++;
                    }
                    vDesc.addAll(damageCrew(pilot, damage));
                }

                if (skin_of_the_teeth && (pilot.getCrew().getHits() < 6)) {
                    Report.addNewline(vDesc);
                    pilot.setDoomed(true); 
                    
                    vDesc.addAll(damageCrew(pilot, 6 - pilot.getCrew()
                                                            .getHits()));
                }
            } else {
                r.choose(true);
                vDesc.addElement(r);
            }
            if (entity.getCrew().isDoomed()) {
                vDesc.addAll(destroyEntity(pilot, "deadly ejection", false,
                                           false));
            } else {
                
                if (game.getBoard().contains(targetCoords)) {
                    pilot.setPosition(targetCoords);
                    
                    r = new Report(6400);
                    r.subject = entity.getId();
                    r.indent(3);
                    vDesc.addElement(r);
                    
                    entityUpdate(pilot.getId());
                    
                    vDesc.addAll(doEntityDisplacementMinefieldCheck(pilot,
                            entity.getPosition(), targetCoords,
                            entity.getElevation()));
                } else {
                    
                    r = new Report(6410);
                    r.subject = entity.getId();
                    r.indent(3);
                    vDesc.addElement(r);
                    
                    game.removeEntity(pilot.getId(),
                                      IEntityRemovalConditions.REMOVE_IN_RETREAT);
                    send(createRemoveEntityPacket(pilot.getId(),
                                                  IEntityRemovalConditions.REMOVE_IN_RETREAT));   
                }
                if (game.getOptions().booleanOption("ejected_pilots_flee")) {
                    game.removeEntity(pilot.getId(),
                                      IEntityRemovalConditions.REMOVE_IN_RETREAT);
                    send(createRemoveEntityPacket(pilot.getId(),
                                                  IEntityRemovalConditions.REMOVE_IN_RETREAT));
                }

                if (skin_of_the_teeth && (pilot.getCrew().getHits() < 5)) {
                    Report.addNewline(vDesc);
                    vDesc.addAll(damageCrew(pilot, 5 - pilot.getCrew()
                                                            .getHits()));
                }
            } 

        } 
        else if (game.getBoard().contains(entity.getPosition())
                 && !game.getOptions().booleanOption("ejected_pilots_flee")
                 && (entity instanceof Tank)) {
            EjectedCrew crew = new EjectedCrew(entity);
            
            crew.setGame(game);
            crew.setDeployed(true);
            crew.setId(getFreeEntityId());
            
            crew.setDone(true);
            
            Coords legalPosition = null;
            if (!crew.isLocationProhibited(entity.getPosition())) {
                legalPosition = entity.getPosition();
            } else {
                for (int dir = 0; dir < 6 && legalPosition == null; dir++) {
                    Coords adjCoords = entity.getPosition().translated(dir);
                    if (!crew.isLocationProhibited(adjCoords)) {
                        legalPosition = adjCoords;
                    }
                }
            }
            
            if (legalPosition == null) {
                System.err.println("Error in Server.ejectEntity: vehicle " +
                		"crews cannot abandon if there is no legal hex!");
                return vDesc;
            }
            crew.setPosition(legalPosition);
            
            game.addEntity(crew);
            
            send(createAddEntityPacket(crew.getId()));
            
            entityUpdate(crew.getId());
            
            vDesc.addAll(doEntityDisplacementMinefieldCheck(crew,
                    entity.getPosition(), entity.getPosition(),
                    entity.getElevation()));
        }

        entity.getCrew().setEjected(true);
        if (entity instanceof VTOL) {
            vDesc.addAll(crashVTOLorWiGE((VTOL) entity));
        }
        vDesc.addAll(destroyEntity(entity, "ejection", true, true));

        if (!autoEject) {
            game.removeEntity(entity.getId(),
                    IEntityRemovalConditions.REMOVE_EJECTED);
            send(createRemoveEntityPacket(entity.getId(),
                    IEntityRemovalConditions.REMOVE_EJECTED));
        }
        return vDesc;
    }

    public static PilotingRollData getEjectModifiers(IGame game,
            Entity entity, boolean autoEject) {
        int facing = entity.getFacing();
        Coords targetCoords = entity.getPosition().translated((facing + 3) % 6);
        return getEjectModifiers(game, entity, autoEject, targetCoords,
                "ejecting");
    }

    public static PilotingRollData getEjectModifiers(IGame game, Entity entity,
            boolean autoEject, Coords targetCoords, String desc) {
        PilotingRollData rollTarget = new PilotingRollData(entity.getId(),
                entity.getCrew().getPiloting(), desc);
        if (entity.isProne()) {
            rollTarget.addModifier(5, "Mech is prone");
        }
        if (entity.getCrew().isUnconscious()) {
            rollTarget.addModifier(3, "pilot unconscious");
        }
        if (autoEject) {
            rollTarget.addModifier(1, "automatic ejection");
        }
        if ((entity instanceof Mech)
                && entity.getInternal(Mech.LOC_HEAD) < entity
                        .getOInternal(Mech.LOC_HEAD)) {
            rollTarget.addModifier(
                    entity.getOInternal(Mech.LOC_HEAD)
                            - entity.getInternal(Mech.LOC_HEAD),
                    "Head Internal Structure Damage");
        }
        IHex targetHex = game.getBoard().getHex(targetCoords);
        if (targetHex != null) {
            if ((targetHex.terrainLevel(Terrains.WATER) > 0)
                    && !targetHex.containsTerrain(Terrains.ICE)) {
                rollTarget.addModifier(-1, "landing in water");
            } else if (targetHex.containsTerrain(Terrains.ROUGH)) {
                rollTarget.addModifier(0, "landing in rough");
            } else if (targetHex.containsTerrain(Terrains.RUBBLE)) {
                rollTarget.addModifier(0, "landing in rubble");
            } else if (targetHex.terrainLevel(Terrains.WOODS) == 1) {
                rollTarget.addModifier(2, "landing in light woods");
            } else if (targetHex.terrainLevel(Terrains.WOODS) == 2) {
                rollTarget.addModifier(3, "landing in heavy woods");
            } else if (targetHex.terrainLevel(Terrains.WOODS) == 3) {
                rollTarget.addModifier(4, "landing in ultra heavy woods");
            } else if (targetHex.terrainLevel(Terrains.JUNGLE) == 1) {
                rollTarget.addModifier(3, "landing in light jungle");
            } else if (targetHex.terrainLevel(Terrains.JUNGLE) == 2) {
                rollTarget.addModifier(5, "landing in heavy jungle");
            } else if (targetHex.terrainLevel(Terrains.JUNGLE) == 3) {
                rollTarget.addModifier(7, "landing in ultra heavy jungle");
            } else if (targetHex.terrainLevel(Terrains.BLDG_ELEV) > 0) {
                rollTarget.addModifier(
                        targetHex.terrainLevel(Terrains.BLDG_ELEV),
                        "landing in a building");
            } else {
                rollTarget.addModifier(-2, "landing in clear terrain");
            }
        } else {
            rollTarget.addModifier(-2, "landing off the board");
        }

        if (game.getPlanetaryConditions().getGravity() == 0) {
            rollTarget.addModifier(3, "Zero-G");
        } else if (game.getPlanetaryConditions().getGravity() < .8) {
            rollTarget.addModifier(2, "Low-G");
        } else if (game.getPlanetaryConditions().getGravity() > 1.2) {
            rollTarget.addModifier(2, "High-G");
        }

        if (game.getPlanetaryConditions().getAtmosphere() == PlanetaryConditions.ATMO_VACUUM) {
            rollTarget.addModifier(3, "Vacuum");
        } else if (game.getPlanetaryConditions().getAtmosphere() == PlanetaryConditions.ATMO_VHIGH) {
            rollTarget.addModifier(2, "Very High Atmosphere Pressure");
        } else if (game.getPlanetaryConditions().getAtmosphere() == PlanetaryConditions.ATMO_TRACE) {
            rollTarget.addModifier(2, "Trace atmosphere");
        }

        if ((game.getPlanetaryConditions().getWeather() == PlanetaryConditions.WE_HEAVY_SNOW)
                || (game.getPlanetaryConditions().getWeather() == PlanetaryConditions.WE_ICE_STORM)
                || (game.getPlanetaryConditions().getWeather() == PlanetaryConditions.WE_DOWNPOUR)
                || (game.getPlanetaryConditions().getWindStrength() == PlanetaryConditions.WI_STRONG_GALE)) {
            rollTarget.addModifier(2, "Bad Weather");
        }

        if ((game.getPlanetaryConditions().getWindStrength() >= PlanetaryConditions.WI_STORM)
                || ((game.getPlanetaryConditions().getWeather() == PlanetaryConditions.WE_HEAVY_SNOW) && (game
                        .getPlanetaryConditions().getWindStrength() == PlanetaryConditions.WI_STRONG_GALE))) {
            rollTarget.addModifier(3, "Really Bad Weather");
        }
        return rollTarget;
    }

    
    public Vector<Report> abandonEntity(Entity entity) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;

        if (entity.getCrew().isEjected()) {
            return vDesc;
        }

        if (entity.getCrew().isDoomed() || entity.getCrew().isDead()) {
            return vDesc;
        }

        if (game.getPlanetaryConditions().isVacuum()) {
            return vDesc;
        }

        Coords targetCoords = entity.getPosition();

        if (entity instanceof Mech) {
            
            r = new Report(2027);
            r.subject = entity.getId();
            r.add(entity.getCrew().getName());
            r.addDesc(entity);
            r.indent(3);
            vDesc.addElement(r);

            MechWarrior pilot = new MechWarrior(entity);
            pilot.getCrew().setUnconscious(entity.getCrew().isUnconscious());
            pilot.setDeployed(true);
            pilot.setId(getFreeEntityId());
            game.addEntity(pilot);
            send(createAddEntityPacket(pilot.getId()));
            
            pilot.setDone(true);
            
            if (game.getBoard().contains(targetCoords)) {
                pilot.setPosition(targetCoords);
            }
            pilot.setCommander(entity.isCommander());
            
            entityUpdate(pilot.getId());
            
            vDesc.addAll(doEntityDisplacementMinefieldCheck(pilot,
                                                            entity.getPosition(), targetCoords, entity.getElevation()));
            if (game.getOptions().booleanOption("ejected_pilots_flee")) {
                game.removeEntity(pilot.getId(),
                                  IEntityRemovalConditions.REMOVE_IN_RETREAT);
                send(createRemoveEntityPacket(pilot.getId(),
                                              IEntityRemovalConditions.REMOVE_IN_RETREAT));
            }
        } 
        else if (game.getBoard().contains(entity.getPosition())
                 && !game.getOptions().booleanOption("ejected_pilots_flee")
                 && game.getOptions().booleanOption("vehicles_can_eject")
                 && (entity instanceof Tank)) {
            EjectedCrew crew = new EjectedCrew(entity);
            crew.setDeployed(true);
            crew.setId(getFreeEntityId());
            game.addEntity(crew);
            send(createAddEntityPacket(crew.getId()));
            
            crew.setDone(true);
            
            crew.setPosition(entity.getPosition());
            
            entityUpdate(crew.getId());
            
            vDesc.addAll(doEntityDisplacementMinefieldCheck(crew,
                                                            entity.getPosition(), entity.getPosition(),
                                                            entity.getElevation()));
        }

        entity.getCrew().setEjected(true);

        return vDesc;
    }

    
    private void resolveMechWarriorPickUp() {
        Report r;

        Iterator<Entity> mechWarriors = game
                .getSelectedEntities(new EntitySelector() {
                    public boolean accept(Entity entity) {
                        if (entity instanceof MechWarrior) {
                            MechWarrior mw = (MechWarrior) entity;
                            if ((mw.getPickedUpById() == Entity.NONE)
                                && !mw.isDoomed()
                                && (mw.getTransportId() == Entity.NONE)) {
                                return true;
                            }
                        }
                        return false;
                    }
                });
        
        while (mechWarriors.hasNext()) {
            boolean pickedUp = false;
            MechWarrior e = (MechWarrior) mechWarriors.next();
            
            for (Entity pe : game.getEntitiesVector(e.getPosition())) {
                if (pe.isDoomed() || pe.isShutDown()
                    || pe.getCrew().isUnconscious() || pe.isAirborne()
                    || (pe.getElevation() != e.getElevation())
                    || (pe.getOwnerId() != e.getOwnerId())
                    || (pe.getId() == e.getId())) {
                    continue;
                }
                if (!pickedUp) {
                    if (pe instanceof MechWarrior) {
                        
                        r = new Report(6415, Report.PUBLIC);
                        r.add(pe.getDisplayName());
                        addReport(r);
                        continue;
                    }
                    
                    pe.pickUp(e);
                    
                    e.setPickedUpById(pe.getId());
                    e.setPickedUpByExternalId(pe.getExternalIdAsString());
                    pickedUp = true;
                    r = new Report(6420, Report.PUBLIC);
                    r.add(e.getDisplayName());
                    r.addDesc(pe);
                    addReport(r);
                    break;
                }
            }
            
            if (!pickedUp) {
                for (Entity pe : game.getEntitiesVector(e.getPosition())) {
                    if (pe.isDoomed()
                        || pe.isShutDown()
                        || pe.getCrew().isUnconscious()
                        || pe.isAirborne()
                        || (pe.getElevation() != e.getElevation())
                        || (pe.getOwnerId() == e.getOwnerId())
                        || (pe.getId() == e.getId())
                        || (pe.getOwner().getTeam() == IPlayer.TEAM_NONE)
                        || (pe.getOwner().getTeam() != e.getOwner()
                                                        .getTeam())) {
                        continue;
                    }
                    if (!pickedUp) {
                        if (pe instanceof MechWarrior) {
                            
                            r = new Report(6415, Report.PUBLIC);
                            r.add(pe.getDisplayName());
                            addReport(r);
                            continue;
                        }
                        
                        pe.pickUp(e);
                        
                        e.setPickedUpById(pe.getId());
                        e.setPickedUpByExternalId(pe.getExternalIdAsString());
                        pickedUp = true;
                        r = new Report(6420, Report.PUBLIC);
                        r.add(e.getDisplayName());
                        r.addDesc(pe);
                        addReport(r);
                        break;
                    }
                }
            }
            
            if (!pickedUp) {
                Iterator<Entity> pickupEnemyEntities = game
                        .getEnemyEntities(e.getPosition(), e);
                while (pickupEnemyEntities.hasNext()) {
                    Entity pe = pickupEnemyEntities.next();
                    if (pe.isDoomed() || pe.isShutDown()
                        || pe.getCrew().isUnconscious() || pe.isAirborne()
                        || (pe.getElevation() != e.getElevation())) {
                        continue;
                    }
                    if (pe instanceof MechWarrior) {
                        
                        r = new Report(6415, Report.PUBLIC);
                        r.add(pe.getDisplayName());
                        addReport(r);
                        continue;
                    }
                    
                    pe.pickUp(e);
                    
                    e.setCaptured(true);
                    e.setPickedUpById(pe.getId());
                    e.setPickedUpByExternalId(pe.getExternalIdAsString());
                    pickedUp = true;
                    r = new Report(6420, Report.PUBLIC);
                    r.add(e.getDisplayName());
                    r.addDesc(pe);
                    addReport(r);
                    break;
                }
            }
            if (pickedUp) {
                
                e.setPosition(null);
                
                entityUpdate(e.getId());
            }
        }
    }

    
    private void resolveSinkVees() {
        Iterator<Entity> sinkableTanks = game
                .getSelectedEntities(new EntitySelector() {
                    public boolean accept(Entity entity) {
                        if (entity.isOffBoard()
                                || (entity.getPosition() == null)
                                || !(entity instanceof Tank)) {
                            return false;
                        }
                        final IHex hex = game.getBoard().getHex(
                                entity.getPosition());
                        final boolean onBridge = (hex
                                .terrainLevel(Terrains.BRIDGE) > 0)
                                && (entity.getElevation() == hex
                                        .terrainLevel(Terrains.BRIDGE_ELEV));
                        if (((entity.getMovementMode()
                                    == EntityMovementMode.TRACKED)
                                || (entity.getMovementMode()
                                        == EntityMovementMode.WHEELED)
                                || ((entity.getMovementMode()
                                        == EntityMovementMode.HOVER)))
                                    && entity.isImmobile()
                                    && (hex.terrainLevel(Terrains.WATER) > 0)
                                    && !onBridge
                                    && !(entity.hasWorkingMisc(
                                            MiscType.F_FULLY_AMPHIBIOUS))
                                    && !(entity.hasWorkingMisc(
                                            MiscType.F_FLOTATION_HULL))) {
                            return true;
                        }
                        return false;
                    }
                });
        while (sinkableTanks.hasNext()) {
            Entity e = sinkableTanks.next();
            addReport(destroyEntity(e, "a watery grave", false));
        }
    }
    
    private void doTryUnstuck() {
        if (game.getPhase() != IGame.Phase.PHASE_MOVEMENT) {
            return;
        }

        Report r;

        Iterator<Entity> stuckEntities = game
                .getSelectedEntities(new EntitySelector() {
                    public boolean accept(Entity entity) {
                        if (entity.isStuck()) {
                            return true;
                        }
                        return false;
                    }
                });
        PilotingRollData rollTarget;
        while (stuckEntities.hasNext()) {
            Entity entity = stuckEntities.next();
            if ((entity.getPosition() == null) && entity.isDeployed()) {
                System.out.println("Entity #" + entity.getId()
                                   + " does not know its position.");
                continue;
            }
            rollTarget = entity.getBasePilotingRoll();
            entity.addPilotingModifierForTerrain(rollTarget);
            
            IHex hex = game.getBoard().getHex(entity.getPosition());
            rollTarget.addModifier(
                    hex.getUnstuckModifier(entity.getElevation()), "terrain");
            
            r = new Report(2340);
            r.addDesc(entity);
            addReport(r);

            final int diceRoll = entity.getCrew().rollPilotingSkill();
            r = new Report(2190);
            r.subject = entity.getId();
            r.add(rollTarget.getValueAsString());
            r.add(rollTarget.getDesc());
            r.add(diceRoll);
            if (diceRoll < rollTarget.getValue()) {
                r.choose(false);
            } else {
                r.choose(true);
                entity.setStuck(false);
                entity.setCanUnstickByJumping(false);
                entity.setElevation(0);
                entityUpdate(entity.getId());
            }
            addReport(r);
        }
    }

    
    private void resolveVeeINarcPodRemoval() {
        Iterator<Entity> vees = game
                .getSelectedEntities(new EntitySelector() {
                    public boolean accept(Entity entity) {
                        if ((entity instanceof Tank) && (entity.mpUsed == 0)) {
                            return true;
                        }
                        return false;
                    }
                });
        boolean canSwipePods;
        while (vees.hasNext()) {
            canSwipePods = true;
            Entity entity = vees.next();
            for (int i = 0; i <= 5; i++) {
                if (entity.weaponFiredFrom(i)) {
                    canSwipePods = false;
                }
            }
            if (((Tank) entity).getStunnedTurns() > 0) {
                canSwipePods = false;
            }
            if (canSwipePods && entity.hasINarcPodsAttached()
                && entity.getCrew().isActive()) {
                entity.removeAllINarcPods();
                Report r = new Report(2345);
                r.addDesc(entity);
                addReport(r);
            }
        }
    }

    
    private Vector<Report> resolveIceBroken(Coords c) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        IHex hex = game.getBoard().getHex(c);
        hex.removeTerrain(Terrains.ICE);
        sendChangedHex(c);
        
        if (hex.terrainLevel(Terrains.WATER) > 0) {
            
            for (Entity e : game.getEntitiesVector(c)) {
                
                boolean isHoverOrWiGE = (e.getMovementMode() == EntityMovementMode.HOVER)
                        || (e.getMovementMode() == EntityMovementMode.WIGE);
                if ((e.getElevation() == 0)
                        && !(hex.containsTerrain(Terrains.BLDG_ELEV, 0))
                        && !(isHoverOrWiGE && (e.getRunMP() >= 0))
                        && (e.getMovementMode() != EntityMovementMode.INF_UMU)
                        && !e.hasUMU()) {
                    vPhaseReport.addAll(doEntityFallsInto(e, c,
                            new PilotingRollData(TargetRoll.AUTOMATIC_FAIL),
                            true));
                }
            }
        }
        return vPhaseReport;
    }

    private Vector<Report> meltIceAndSnow(Coords c, int entityId) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;
        IHex hex = game.getBoard().getHex(c);
        r = new Report(3069);
        r.indent(2);
        r.subject = entityId;
        vDesc.add(r);
        if (hex.containsTerrain(Terrains.SNOW)) {
            hex.removeTerrain(Terrains.SNOW);
            sendChangedHex(c);
        }
        if (hex.containsTerrain(Terrains.ICE)) {
            vDesc.addAll(resolveIceBroken(c));
        }
        
        if (!hex.containsTerrain(Terrains.MUD)
            && !hex.containsTerrain(Terrains.WATER)) {
            hex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                    Terrains.MUD, 1));
            sendChangedHex(c);
        }
        return vDesc;
    }

    private Vector<Report> checkQuickSand(Coords c) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;
        IHex hex = game.getBoard().getHex(c);
        if (hex.terrainLevel(Terrains.SWAMP) == 1) {
            if (Compute.d6(2) == 12) {
                
                hex.removeTerrain(Terrains.SWAMP);
                hex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                        Terrains.SWAMP, 2));
                sendChangedHex(c);
                r = new Report(2440);
                r.indent(1);
                vDesc.add(r);
            }
        }
        return vDesc;
    }

    public Vector<Report> checkForVehicleFire(Tank tank, boolean inferno) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        int boomroll = Compute.d6(2);
        int penalty = 0;
        switch (tank.getMovementMode()) {
            case HOVER:
                penalty = 4;
                break;
            case VTOL:
            case WHEELED:
                penalty = 2;
                break;
            default:
                break;
        }
        if (inferno) {
            boomroll = 12;
        }
        Report r = new Report(5250);
        r.subject = tank.getId();
        r.addDesc(tank);
        r.add(8 - penalty);
        r.add(boomroll);
        if ((boomroll + penalty) < 8) {
            
            r.choose(true);
            vPhaseReport.add(r);
        } else {
            
            if (!inferno) {
                r.choose(false);
                vPhaseReport.add(r);
            }
            if ((boomroll + penalty) < 10) {
                addReport(vehicleMotiveDamage(tank, penalty - 1));
            } else {
                vPhaseReport.addAll(resolveVehicleFire(tank, false));
                if ((boomroll + penalty) >= 12) {
                    r = new Report(5255);
                    r.subject = tank.getId();
                    r.indent(3);
                    vPhaseReport.add(r);
                    tank.setOnFire(inferno);
                }
            }
        }
        return vPhaseReport;
    }

    private Vector<Report> resolveVehicleFire(Tank tank, boolean existingStatus) {
        Vector<Report> vPhaseReport = new Vector<Report>();
        if (existingStatus && !tank.isOnFire()) {
            return vPhaseReport;
        }
        for (int i = 0; i < tank.locations(); i++) {
            if ((i == Tank.LOC_BODY)
                || ((tank instanceof VTOL) && (i == VTOL.LOC_ROTOR))) {
                continue;
            }
            if (existingStatus && !tank.isLocationBurning(i)) {
                continue;
            }
            HitData hit = new HitData(i);
            int damage = Compute.d6(1);
            vPhaseReport.addAll(damageEntity(tank, hit, damage));
            if ((damage == 1) && existingStatus) {
                tank.extinguishLocation(i);
            }
        }
        return vPhaseReport;
    }

    private Vector<Report> vehicleMotiveDamage(Tank te, int modifier) {
        return vehicleMotiveDamage(te, modifier, false, -1, false);
    }

    private Vector<Report> vehicleMotiveDamage(Tank te, int modifier,
                                               boolean noroll, int damagetype) {
        return vehicleMotiveDamage(te, modifier, noroll, damagetype, false);
    }

    
    private Vector<Report> vehicleMotiveDamage(Tank te, int modifier,
                                               boolean noroll, int damagetype, boolean jumpDamage) {
        Vector<Report> vDesc = new Vector<Report>();
        Report r;
        switch (te.getMovementMode()) {
            case HOVER:
            case HYDROFOIL:
                if (jumpDamage) {
                    modifier -= 1;
                } else {
                    modifier += 3;
                }
                break;
            case WHEELED:
                if (jumpDamage) {
                    modifier += 1;
                } else {
                    modifier += 2;
                }
                break;
            case WIGE:
                if (jumpDamage) {
                    modifier -= 2;
                } else {
                    modifier += 4;
                }
                break;
            case TRACKED:
                if (jumpDamage) {
                    modifier += 2;
                }
                break;
            case VTOL:
                
                if (!(te.isLocationBad(VTOL.LOC_ROTOR) || te
                        .isLocationDoomed(VTOL.LOC_ROTOR))) {
                    te.setMotiveDamage(te.getMotiveDamage() + 1);
                    if (te.getOriginalWalkMP() > te.getMotiveDamage()) {
                        r = new Report(6660);
                        r.indent(3);
                        r.subject = te.getId();
                        vDesc.add(r);
                    } else {
                        r = new Report(6670);
                        r.subject = te.getId();
                        vDesc.add(r);
                        te.immobilize();
                        
                        if (te.isAirborneVTOLorWIGE()
                           
                            && !(te.isDestroyed() || te.isDoomed())) {
                            vDesc.addAll(forceLandVTOLorWiGE(te));
                        }
                    }
                }
                
                return vDesc;
            default:
                break;
        }
        
        if (game.getOptions().booleanOption("tacops_vehicle_effective")
            && !jumpDamage) {
            modifier = Math.max(modifier - 1, 0);
        }

        if (te.hasWorkingMisc(MiscType.F_ARMORED_MOTIVE_SYSTEM)) {
            modifier -= 2;
        }
        int roll = Compute.d6(2) + modifier;
        r = new Report(6306);
        r.subject = te.getId();
        r.newlines = 0;
        r.indent(3);
        vDesc.add(r);
        if (!noroll) {
            r = new Report(6310);
            r.subject = te.getId();
            r.add(roll);
            r.newlines = 0;
            vDesc.add(r);
            r = new Report(3340);
            r.add(modifier);
            r.subject = te.getId();
            vDesc.add(r);
        }

        if ((noroll && (damagetype == 0)) || (!noroll && (roll <= 5))) {
            
            r = new Report(6005);
            r.subject = te.getId();
            r.indent(3);
            vDesc.add(r);
        } else if ((noroll && (damagetype == 1)) || (!noroll && (roll <= 7))) {
            
            r = new Report(6470);
            r.subject = te.getId();
            r.indent(3);
            vDesc.add(r);
            te.addMovementDamage(1);
        } else if ((noroll && (damagetype == 2)) || (!noroll && (roll <= 9))) {
            
            r = new Report(6471);
            r.subject = te.getId();
            r.indent(3);
            vDesc.add(r);
            te.addMovementDamage(2);
        } else if ((noroll && (damagetype == 3)) || (!noroll && (roll <= 11))) {
            
            r = new Report(6472);
            r.subject = te.getId();
            r.indent(3);
            vDesc.add(r);
            te.addMovementDamage(3);
        } else {
            r = new Report(6473);
            r.subject = te.getId();
            r.indent(3);
            vDesc.add(r);
            te.addMovementDamage(4);
        }
        
        if (((te.getMovementMode() == EntityMovementMode.HOVER) || ((te
                                                                             .getMovementMode() == EntityMovementMode
                                                                             .WIGE) && (te
                                                                                                .getElevation() == 0)))
            && (te.isMovementHitPending() || (te.getWalkMP() <= 0)) 
            
            && (game.getBoard().getHex(te.getPosition())
                    .terrainLevel(Terrains.WATER) > 0)
            && !game.getBoard().getHex(te.getPosition())
                    .containsTerrain(Terrains.ICE)) {
            vDesc.addAll(destroyEntity(te, "a watery grave", false));
        }
        
        if (((te.getMovementMode() == EntityMovementMode.WIGE) && (te
                .isAirborneVTOLorWIGE()))
            && (te.isMovementHitPending() || (te.getWalkMP() <= 0))) {
            
            vDesc.addAll(crashVTOLorWiGE(te));
        }
        return vDesc;
    }

    private void addReport(Vector<Report> reports) {
        vPhaseReport.addAll(reports);
    }

    private void addReport(Vector<Report> reports, int indents) {
        for (Report r : reports) {
            r.indent(indents);
            vPhaseReport.add(r);
        }
    }

    private void addReport(Report report) {
        vPhaseReport.addElement(report);
    }

    private void clearReports() {
        vPhaseReport.removeAllElements();
    }

    private void addNewLines() {
        Report.addNewline(vPhaseReport);
    }

    public void doAssaultDrop(Entity entity) {
        
        Report r = new Report(2380);

        entity.setAltitude(0);

        PilotingRollData psr;
        if ((entity instanceof Protomech) || (entity instanceof BattleArmor)) {
            psr = new PilotingRollData(entity.getId(), 5,
                                       "landing assault drop");

        } else if (entity instanceof Infantry) {
            psr = new PilotingRollData(entity.getId(), 4,
                                       "landing assault drop");
        } else {
            psr = entity.getBasePilotingRoll();
        }
        int roll = Compute.d6(2);
        
        addNewLines();
        r.subject = entity.getId();
        r.add(entity.getDisplayName(), true);
        r.add(psr.getValueAsString());
        r.add(roll);
        r.newlines = 1;
        r.choose(roll >= psr.getValue());
        addReport(r);

        if (game.getBoard().inAtmosphere()) {
            
        
            r = new Report(2388);
            addReport(r);
            r.subject = entity.getId();
            r.add(entity.getDisplayName(), true);
            game.removeEntity(entity.getId(),
                              IEntityRemovalConditions.REMOVE_IN_RETREAT);
            return;
        }

        if (roll < psr.getValue()) {
            int fallHeight = psr.getValue() - roll;

            
            if (fallHeight > 7) {
                addReport(destroyEntity(entity, "failed assault drop", false,
                                        false));
                entityUpdate(entity.getId());
                return;
            }

            
            Coords c = Compute.scatterAssaultDrop(entity.getPosition(),
                                                  fallHeight);
            int distance = entity.getPosition().distance(c);
            r = new Report(2385);
            r.subject = entity.getId();
            r.add(distance);
            r.indent();
            r.newlines = 0;
            addReport(r);
            if (!game.getBoard().contains(c)) {
                r = new Report(2386);
                addReport(r);
                game.removeEntity(entity.getId(),
                                  IEntityRemovalConditions.REMOVE_IN_RETREAT);
                return;
            } else {
                r = new Report(2387);
                r.add(c.getBoardNum());
                addReport(r);
            }
            entity.setPosition(c);

            IHex hex = game.getBoard().getHex(entity.getPosition());
            int bldgElev = hex.containsTerrain(Terrains.BLDG_ELEV) ? hex
                    .terrainLevel(Terrains.BLDG_ELEV) : 0;
            entity.setElevation(fallHeight + bldgElev);
            if ((entity instanceof Infantry)
                && !(entity instanceof BattleArmor)) {
                HitData hit = new HitData(Infantry.LOC_INFANTRY);
                addReport(damageEntity(entity, hit, 1));
            } else {
                addReport(doEntityFallsInto(entity, c, psr, true));
            }
        } else {
            
            IHex hex = game.getBoard().getHex(entity.getPosition());
            int bldgElev = hex.containsTerrain(Terrains.BLDG_ELEV) ? hex
                    .terrainLevel(Terrains.BLDG_ELEV) : 0;
            entity.setElevation(bldgElev);

            Building bldg = game.getBoard().getBuildingAt(entity.getPosition());
            if (bldg != null) {
                
                checkBuildingCollapseWhileMoving(bldg, entity, entity.getPosition());
            }

            
            Entity violated = Compute.stackingViolation(game, entity,
                                                        entity.getPosition(), null);
            if (violated != null) {
              
                if (violated instanceof Dropship) {
                    violated = entity;
                }
                Coords targetDest = Compute.getValidDisplacement(game,
                                                                 violated.getId(), violated.getPosition(),
                                                                 Compute.d6() - 1);
                if (null != targetDest) {
                    doEntityDisplacement(violated, violated.getPosition(),
                                         targetDest, null);
                    entityUpdate(violated.getId());
                } else {
                    
                    
                    
                    vPhaseReport.addAll(destroyEntity(entity,
                                                      "impossible displacement", entity instanceof Mech,
                                                      entity instanceof Mech));
                }
            }
        }
    }

    void doAllAssaultDrops() {
        for (Entity e : game.getEntitiesVector()) {
            if (e.isAssaultDropInProgress()) {
                doAssaultDrop(e);
                e.setLandedAssaultDrop();
            }
        }
    }

    void doMagmaDamage(Entity en, boolean eruption) {
        if ((((en.getMovementMode() == EntityMovementMode.VTOL) && (en
                                                                            .getElevation() > 0))
             || (en.getMovementMode() == EntityMovementMode.HOVER) || ((en
                                                                                .getMovementMode() ==
                                                                        EntityMovementMode.WIGE)
                                                                       && (en.getOriginalWalkMP() > 0) && !eruption))
            && !en.isImmobile()) {
            return;
        }
        Report r;
        boolean isMech = en instanceof Mech;
        if (isMech) {
            r = new Report(2405);
        } else {
            r = new Report(2400);
        }
        r.addDesc(en);
        r.subject = en.getId();
        addReport(r);
        if (isMech) {
            HitData h;
            for (int i = 0; i < en.locations(); i++) {
                if (eruption || en.locationIsLeg(i) || en.isProne()) {
                    h = new HitData(i);
                    addReport(damageEntity(en, h, Compute.d6(2)));
                }
            }
        } else {
            addReport(destroyEntity(en, "fell into magma", false, false));
        }
        addNewLines();
    }

    public void doSinkEntity(Entity en) {
        Report r;
        r = new Report(2445);
        r.addDesc(en);
        r.subject = en.getId();
        addReport(r);
        en.setElevation(en.getElevation() - 1);
        
        if (Math.abs(en.getElevation()) > en.getHeight()) {
            addReport(destroyEntity(en, "quicksand"));
        }
    }

    public Vector<Integer> artilleryDamageHex(Coords coords,
            Coords attackSource, int damage, AmmoType ammo, int subjectId,
            Entity killer, Entity exclude, boolean flak, int altitude,
            Vector<Report> vPhaseReport, boolean asfFlak,
            Vector<Integer> alreadyHit, boolean variableDamage) {

        IHex hex = game.getBoard().getHex(coords);
        if (hex == null) {
            return alreadyHit; 
        }

        Report r;

        if (!flak) {
            
            IHex h = game.getBoard().getHex(coords);
            if ((h != null) && h.hasTerrainfactor()) {
                r = new Report(3384);
                r.indent(2);
                r.subject = subjectId;
                r.add(coords.getBoardNum());
                r.add(damage * 2);
                vPhaseReport.addElement(r);
            }
            
            Vector<Report> newReports =
                    tryClearHex(coords, damage * 2, subjectId);
            for (Report nr : newReports) {
                nr.indent(3);
            }
            vPhaseReport.addAll(newReports);
        }
        Building bldg = game.getBoard().getBuildingAt(coords);
        int bldgAbsorbs = 0;
        if ((bldg != null)
            && !(flak && (((altitude > hex.terrainLevel(Terrains.BLDG_ELEV)) || (altitude > hex
                .terrainLevel(Terrains.BRIDGE_ELEV)))))) {
            bldgAbsorbs = bldg.getAbsorbtion(coords);
            if (!((ammo != null) && (ammo.getMunitionType() == AmmoType.M_FLECHETTE))) {
                
                Vector<Report> buildingReport = damageBuilding(bldg, damage,
                                                               coords);
                for (Report report : buildingReport) {
                    report.subject = subjectId;
                }
                vPhaseReport.addAll(buildingReport);
            }
        }

        if (flak
            && ((altitude <= 0)
                || (altitude <= hex.terrainLevel(Terrains.BLDG_ELEV)) || (altitude == hex
                .terrainLevel(Terrains.BRIDGE_ELEV)))) {
            
            return alreadyHit;
        }

        for (Entity entity : game.getEntitiesVector(coords)) {
            int hits = damage;
            if (variableDamage) {
                hits = Compute.d6(damage);
            }
            ToHitData toHit = new ToHitData();
            if (entity instanceof Protomech) {
                toHit.setHitTable(ToHitData.HIT_SPECIAL_PROTO);
            }
            int cluster = 5;

            
            if ((entity == exclude) || alreadyHit.contains(entity.getId())) {
                continue;
            }

            if ((bldg != null)
                && (bldgAbsorbs > 0)
                && (entity.getElevation() < hex
                    .terrainLevel(Terrains.BLDG_ELEV))) {
                cluster -= bldgAbsorbs;
                
                cluster = (int) Math.floor(bldg.getDamageToScale() * cluster);
                if (entity instanceof Infantry) {
                    continue; 
                } else if (cluster <= 0) {
                    
                    r = new Report(6426);
                    r.subject = subjectId;
                    r.addDesc(entity);
                    vPhaseReport.add(r);
                    continue;
                } else {
                    r = new Report(6425);
                    r.subject = subjectId;
                    r.add(bldgAbsorbs);
                    vPhaseReport.add(r);
                }
            }

            if (asfFlak && !(entity instanceof Aero)) {
                continue;
            }

            if (flak) {
                
                if (!((entity instanceof VTOL)
                        || (entity.getMovementMode() == EntityMovementMode.VTOL)
                        || (entity instanceof Aero))) {
                    continue;
                }
                
                if (entity.getElevation() != altitude) {
                    continue;
                }
            } else {
                
                if ((entity instanceof VTOL)
                    || (entity.getMovementMode() == EntityMovementMode.VTOL)
                    || (entity instanceof Aero)) {
                    
                    if ((entity.getElevation() != 0)
                        && (entity.getElevation() != hex
                            .terrainLevel(Terrains.BLDG_ELEV))
                        && (entity.getElevation() != hex
                            .terrainLevel(Terrains.BRIDGE_ELEV))) {
                        continue;
                    }
                }
            }

            if (attackSource != null) {
                toHit.setSideTable(entity.sideTable(attackSource));
                if ((ammo != null)
                    && (ammo.getMunitionType() == AmmoType.M_CLUSTER)
                    && attackSource.equals(coords)) {
                    if (entity instanceof Mech) {
                        toHit.setHitTable(ToHitData.HIT_ABOVE);
                    } else if (entity instanceof Tank) {
                        toHit.setSideTable(ToHitData.SIDE_FRONT);
                        toHit.addModifier(2, "cluster artillery hitting a Tank");
                    }
                }
            }

            if ((entity instanceof Infantry)
                && !(entity instanceof BattleArmor)) {
                hits *= 2;
            }
            boolean specialCaseFlechette = false;

            if (ammo != null) {
                if (ammo.getMunitionType() == AmmoType.M_CLUSTER) {
                    if (hex.containsTerrain(Terrains.FORTIFIED)
                        && (entity instanceof Infantry)
                        && !(entity instanceof BattleArmor)) {
                        hits *= 2;
                    }
                } else if (ammo.getMunitionType() == AmmoType.M_FLECHETTE) {

                    
                    if ((entity instanceof Tank)
                            && ((entity.getMovementMode() == EntityMovementMode.WHEELED) || (entity
                                    .getMovementMode() == EntityMovementMode.HOVER))) {
                        r = new Report(6480);
                        r.subject = entity.getId();
                        r.addDesc(entity);
                        r.add(toHit.getTableDesc());
                        r.add(0);
                        vPhaseReport.add(r);
                        vPhaseReport.addAll(vehicleMotiveDamage((Tank) entity,
                                0));
                        continue;
                    }
                    
                    if ((entity instanceof BattleArmor)
                            || ((entity instanceof SupportTank)
                                && !entity.hasPatchworkArmor()
                                && (entity.getBARRating(1) > 4))) {
                        continue;
                    }
                    if (entity instanceof Infantry) {
                        hits = Compute.d6(damage);
                        hits *= 2;
                    } else {
                        if ((entity.getBARRating(1) < 5)
                                && !entity.hasPatchworkArmor()) {
                            switch (ammo.getAmmoType()) {
                                case AmmoType.T_LONG_TOM:
                                    
                                    if (damage == 4) {
                                        damage = (5 - entity.getBARRating(1)) * 5;
                                    } else {
                                        continue;
                                    }
                                    break;
                                case AmmoType.T_SNIPER:
                                    
                                    if (damage == 2) {
                                        damage = (5 - entity.getBARRating(1)) * 3;
                                    } else {
                                        continue;
                                    }
                                    break;
                                case AmmoType.T_THUMPER:
                                    
                                    damage = 5 - entity.getBARRating(1);
                                    break;
                            }
                        } else {
                            
                            specialCaseFlechette = true;
                            switch (ammo.getAmmoType()) {
                                case AmmoType.T_LONG_TOM:
                                    
                                    if (damage == 4) {
                                        damage = 25;
                                    } else {
                                        continue;
                                    }
                                    break;
                                case AmmoType.T_SNIPER:
                                    
                                    if (damage == 2) {
                                        damage = 15;
                                    } else {
                                        continue;
                                    }
                                    break;
                                case AmmoType.T_THUMPER:
                                    
                                    damage = 10;
                                    break;
                            }
                        }
                    }
                }
            }

            alreadyHit.add(entity.getId());

            r = new Report(6480);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(toHit.getTableDesc());
            r.add(hits);
            vPhaseReport.add(r);
            if (entity instanceof BattleArmor) {
                
                for (int loc = 0; loc < entity.locations(); loc++) {
                    if (entity.getInternal(loc) > 0) {
                        HitData hit = new HitData(loc);
                        vPhaseReport.addAll(damageEntity(entity, hit, hits,
                                false, DamageType.NONE, false, true, false));
                    }
                }
            } else {
                while (hits > 0) {
                    int damageToDeal = Math.min(cluster, hits);
                    HitData hit = entity.rollHitLocation(toHit.getHitTable(),
                            toHit.getSideTable());
                    
                    if (specialCaseFlechette && !(entity instanceof Infantry)) {
                        damageToDeal *= (5 - entity.getBARRating(hit
                                .getLocation()));
                    }
                    vPhaseReport.addAll(damageEntity(entity, hit, damageToDeal,
                            false, DamageType.NONE, false, true, false));
                    hits -= Math.min(cluster, hits);
                }
            }
            if (killer != null) {
                creditKill(entity, killer);
            }
        }

        return alreadyHit;
    }

    public void artilleryDamageArea(Coords centre, Coords attackSource,
            AmmoType ammo, int subjectId, Entity killer, boolean flak,
            int altitude, boolean mineClear, Vector<Report> vPhaseReport,
            boolean asfFlak, int attackingBA) {
        int damage = ammo.getRackSize();
        int falloff = 10;
        if (ammo.getAmmoType() == AmmoType.T_CRUISE_MISSILE) {
            falloff = 25;
        }
        if (ammo.getAmmoType() == AmmoType.T_BA_TUBE) {
            falloff = 2 * attackingBA;
        }
        if (ammo.getMunitionType() == AmmoType.M_CLUSTER) {
            
            if (ammo.getAmmoType() != AmmoType.T_ARROW_IV) {
                damage -= 5;
            }
            
            if (ammo.getAmmoType() == AmmoType.T_THUMPER) {
                falloff = 9;
            }
            attackSource = centre;
        } else if (ammo.getMunitionType() == AmmoType.M_FLECHETTE) {
            switch (ammo.getAmmoType()) {
               
                case AmmoType.T_LONG_TOM:
                    damage = 4;
                    falloff = 2;
                    break;
                case AmmoType.T_SNIPER:
                    damage = 2;
                    falloff = 1;
                    break;
                case AmmoType.T_THUMPER:
                    damage = 1;
                    falloff = 1;
            }
        } else
            
            if (mineClear) {
                falloff = damage;
            }
        artilleryDamageArea(centre, attackSource, ammo, subjectId, killer,
                damage, falloff, flak, altitude, vPhaseReport, asfFlak);
    }
    public void artilleryDamageArea(Coords centre, Coords attackSource,
            AmmoType ammo, int subjectId, Entity killer, int damage,
            int falloff, boolean flak, int altitude,
            Vector<Report> vPhaseReport, boolean asfFlak) {
        Vector<Integer> alreadyHit = new Vector<Integer>();
        for (int ring = 0; damage > 0; ring++, damage -= falloff) {
            ArrayList<Coords> hexes = Compute.coordsAtRange(centre, ring);
            for (Coords c : hexes) {
                alreadyHit = artilleryDamageHex(c, attackSource, damage, ammo,
                        subjectId, killer, null, flak, altitude, vPhaseReport,
                        asfFlak, alreadyHit, false);
            }
            attackSource = centre; 
        }
    }

    public void deliverBombDamage(Coords centre, int type, int subjectId,
            Entity killer, Vector<Report> vPhaseReport) {
        int range = 0;
        int damage = 10;
        if (type == BombType.B_CLUSTER) {
            range = 1;
            damage = 5;
        }
        Vector<Integer> alreadyHit = new Vector<Integer>();

        alreadyHit = artilleryDamageHex(centre, centre, damage, null,
                subjectId, killer, null, false, 0, vPhaseReport, false,
                alreadyHit, false);
        if (range > 0) {
            ArrayList<Coords> hexes = Compute.coordsAtRange(centre, range);
            for (Coords c : hexes) {
                alreadyHit = artilleryDamageHex(c, centre, damage, null,
                        subjectId, killer, null, false, 0, vPhaseReport, false,
                        alreadyHit, false);
            }
        }
    }

    
    public void deliverBombInferno(Coords coords, Entity ae, int subjectId,
                                   Vector<Report> vPhaseReport) {
        IHex h = game.getBoard().getHex(coords);
        Report r;
        
        if (h.terrainLevel(Terrains.FIRE) < Terrains.FIRE_LVL_INFERNO_BOMB) {
            ignite(coords, Terrains.FIRE_LVL_INFERNO_BOMB, vPhaseReport);
        }
        
        if (h.containsTerrain(Terrains.ICE) || h.containsTerrain(Terrains.SNOW)) {
            vPhaseReport.addAll(meltIceAndSnow(coords, subjectId));
        }
        for (Entity entity : game.getEntitiesVector(coords)) {
            if (entity.isAirborne() || entity.isAirborneVTOLorWIGE()) {
                continue;
            }
            
            r = new Report(6696);
            r.indent(3);
            r.add(entity.getDisplayName());
            r.subject = entity.getId();
            r.newlines = 0;
            vPhaseReport.add(r);
            if (entity instanceof Tank) {
                Report.addNewline(vPhaseReport);
            }
            Vector<Report> vDamageReport = deliverInfernoMissiles(ae, entity, 5);
            Report.indentAll(vDamageReport, 2);
            vPhaseReport.addAll(vDamageReport);
        }
    }

    
    void resolveFortify() {
        Report r;
        for (Entity ent : game.getEntitiesVector()) {
            if (ent instanceof Infantry) {
                Infantry inf = (Infantry) ent;
                int dig = inf.getDugIn();
                if (dig == Infantry.DUG_IN_WORKING) {
                    r = new Report(5300);
                    r.addDesc(inf);
                    r.subject = inf.getId();
                    addReport(r);
                } else if (dig == Infantry.DUG_IN_FORTIFYING2) {
                    Coords c = inf.getPosition();
                    r = new Report(5305);
                    r.addDesc(inf);
                    r.add(c.getBoardNum());
                    r.subject = inf.getId();
                    addReport(r);
                    
                    IHex hex = game.getBoard().getHex(c);
                    hex.addTerrain(Terrains.getTerrainFactory().createTerrain(
                            Terrains.FORTIFIED, 1));
                    sendChangedHex(c);
                    
                    for (Entity ent2 : game.getEntitiesVector(c)) {
                        if (ent2 instanceof Infantry) {
                            Infantry inf2 = (Infantry) ent;
                            inf2.setDugIn(Infantry.DUG_IN_NONE);
                        }
                    }
                }
            }
        }
    }

    private void checkBreakSpikes(Entity e, int loc) {
        int roll = Compute.d6(2);
        Report r;
        if (roll < 9) {
            r = new Report(4445);
            r.newlines = 0;
            r.add(roll);
            r.subject = e.getId();
            addReport(r);
            return;
        }
        r = new Report(4440);
        r.newlines = 0;
        r.add(roll);
        r.subject = e.getId();
        addReport(r);
        for (Mounted m : e.getMisc()) {
            if (m.getType().hasFlag(MiscType.F_SPIKES)
                && (m.getLocation() == loc)) {
                m.setHit(true);
            }
        }
    }

    public String getHost() {
        try {
            return InetAddress.getLocalHost().getHostName();
        } catch (UnknownHostException e) {
            e.printStackTrace();
            return "";
        }
    }

    public int getPort() {
        return serverSocket.getLocalPort();
    }

    private void handleAttacks() {
        Report r;
        int lastAttackerId = -1;
        Vector<AttackHandler> currentAttacks, keptAttacks;
        currentAttacks = game.getAttacksVector();
        keptAttacks = new Vector<AttackHandler>();
        Vector<Report> handleAttackReports = new Vector<Report>();
        
        for (AttackHandler ah : currentAttacks) {
            if (!(ah instanceof TAGHandler)) {
                continue;
            }
            if (ah.cares(game.getPhase())) {
                int aId = ah.getAttackerId();
                if ((aId != lastAttackerId) && !ah.announcedEntityFiring()) {
                    
                    r = new Report(3100);
                    r.subject = aId;
                    Entity ae = game.getEntity(aId);
                    r.addDesc(ae);
                    handleAttackReports.addElement(r);
                    ah.setAnnouncedEntityFiring(true);
                    lastAttackerId = aId;
                }
                boolean keep = ah.handle(game.getPhase(), handleAttackReports);
                if (keep) {
                    keptAttacks.add(ah);
                }
                Report.addNewline(handleAttackReports);
            }
        }
        
        for (AttackHandler ah : currentAttacks) {
            if (ah instanceof TAGHandler) {
                continue;
            }
            if (ah.cares(game.getPhase())) {
                int aId = ah.getAttackerId();
                if ((aId != lastAttackerId) && !ah.announcedEntityFiring()) {
                    
                    handleAttackReports.addAll(checkFatalThresholds(aId,
                            lastAttackerId));
                    
                    if (ah.isStrafing()) {
                        r = new Report(3101);
                    } else {
                        r = new Report(3100);
                    }
                    r.subject = aId;
                    Entity ae = game.getEntity(aId);
                    
                    
                    if (ae == null) {
                        ae = game.getOutOfGameEntity(aId);
                    }
                    r.addDesc(ae);
                    handleAttackReports.addElement(r);
                    ah.setAnnouncedEntityFiring(true);
                    lastAttackerId = aId;
                }
                boolean keep = ah.handle(game.getPhase(), handleAttackReports);
                if (keep) {
                    keptAttacks.add(ah);
                }
                Report.addNewline(handleAttackReports);
            } else {
                keptAttacks.add(ah);
            }
        }
        
        handleAttackReports.addAll(checkFatalThresholds(lastAttackerId,
                lastAttackerId));
        if (handleAttackReports.size() > 0) {
            Report.addNewline(handleAttackReports);
        }
        addReport(handleAttackReports);
        
        game.setAttacksVector(keptAttacks);
    }

    public static Server getServerInstance() {
        return serverInstance;
    }

    public void createSmoke(Coords coords) {
        createSmoke(coords, 1, 0);
    }

    public void createSmoke(Coords coords, int level) {
        createSmoke(coords, level, 0);
    }

    public void createSmoke(Coords coords, int level, int duration) {
        SmokeCloud cloud = new SmokeCloud(coords, level, duration);
        game.addSmokeCloud(cloud);
        sendSmokeCloudAdded(cloud);
    }

    public void createSmoke(ArrayList<Coords> coords, int level, int duration) {
        SmokeCloud cloud = new SmokeCloud(coords, level, duration);
        game.addSmokeCloud(cloud);
        sendSmokeCloudAdded(cloud);
    }

    public void updateSmoke(SmokeCloud cloud, ArrayList<Coords> newCoords) {
        removeSmokeTerrain(cloud);
        cloud.getCoordsList().clear();
        cloud.getCoordsList().addAll(newCoords);
    }

    public void removeSmokeTerrain(SmokeCloud cloud) {
        for (Coords coords : cloud.getCoordsList()) {
            IHex nextHex = game.getBoard().getHex(coords);
            if ((nextHex != null) && nextHex.containsTerrain(Terrains.SMOKE)) {
                nextHex.removeTerrain(Terrains.SMOKE);
                sendChangedHex(coords);
            }
        }
    }

    public List<SmokeCloud> getSmokeCloudList() {
        return game.getSmokeCloudList();
    }

    private Vector<Report> resolveBlowingSandDamage() {
        Vector<Report> vFullReport = new Vector<Report>();
        vFullReport.add(new Report(5002, Report.PUBLIC));
        int damage_bonus = Math.max(0, game.getPlanetaryConditions()
                                           .getWindStrength() - PlanetaryConditions.WI_MOD_GALE);
        
        for (Enumeration<Team> loop = game.getTeams(); loop.hasMoreElements(); ) {
            Team team = loop.nextElement();
            Vector<Integer> airborne = team.getAirborneVTOL();
            if (airborne.size() > 0) {
                
                int unitsAffected = Math.min(Compute.d6(), airborne.size());
                while ((unitsAffected > 0) && (airborne.size() > 0)) {
                    int loc = Compute.randomInt(airborne.size());
                    Entity en = game.getEntity(airborne.get(loc));
                    int damage = Math.max(1, Compute.d6() / 2) + damage_bonus;
                    while (damage > 0) {
                        HitData hit = en.rollHitLocation(ToHitData.HIT_NORMAL,
                                                         ToHitData.SIDE_RANDOM);
                        vFullReport.addAll(damageEntity(en, hit, 1));
                        damage--;
                    }
                    unitsAffected--;
                    airborne.remove(loc);
                }
            }
        }
        Report.addNewline(vPhaseReport);
        return vFullReport;
    }

    private void layMine(Entity entity, int mineId, Coords coords) {
        Mounted mine = entity.getEquipment(mineId);
        Report r;
        if (!mine.isMissing()) {
            int reportId = 0;
            switch (mine.getMineType()) {
                case Mounted.MINE_CONVENTIONAL:
                    deliverThunderMinefield(coords, entity.getOwnerId(), 10,
                                            entity.getId());
                    reportId = 3500;
                    break;
                case Mounted.MINE_VIBRABOMB:
                    deliverThunderVibraMinefield(coords, entity.getOwnerId(), 10,
                                                 mine.getVibraSetting(), entity.getId());
                    reportId = 3505;
                    break;
                case Mounted.MINE_ACTIVE:
                    deliverThunderActiveMinefield(coords, entity.getOwnerId(), 10,
                                                  entity.getId());
                    reportId = 3510;
                    break;
                case Mounted.MINE_INFERNO:
                    deliverThunderInfernoMinefield(coords, entity.getOwnerId(), 10,
                                                   entity.getId());
                    reportId = 3515;
                    break;
                
            }
            mine.setShotsLeft(mine.getUsableShotsLeft() - 1);
            if (mine.getUsableShotsLeft() <= 0) {
                mine.setMissing(true);
            }
            r = new Report(reportId);
            r.subject = entity.getId();
            r.addDesc(entity);
            r.add(coords.getBoardNum());
            addReport(r);
            entity.setLayingMines(true);
        }
    }

    public void reportRoll(Roll roll) {
        Report r = new Report(1230);
        r.add(roll.getReport());
        addReport(r);
    }

    private void registerWithServerBrowser(boolean register, String urlString) {
        try {
            URL url = new URL(urlString);
            HttpURLConnection conn = (HttpURLConnection) url.openConnection();
            conn.setDoOutput(true);
            conn.setRequestProperty("Content-Type",
                                    "application/x-www-form-urlencoded");
            DataOutputStream printout = new DataOutputStream(
                    conn.getOutputStream());
            String content = "";
            content = "port="
                      + URLEncoder.encode(
                    Integer.toString(serverSocket.getLocalPort()),
                    "UTF-8");
            if (register) {
                for (IConnection iconn : connections) {
                    content += "&players[]="
                               + (getPlayer(iconn.getId()).getName());
                }
                if ((game.getPhase() != Phase.PHASE_LOUNGE)
                    && (game.getPhase() != Phase.PHASE_UNKNOWN)) {
                    content += "&close=yes";
                }
                content += "&version=" + MegaMek.VERSION;
                if (isPassworded()) {
                    content += "&pw=yes";
                }
            } else {
                content += "&delete=yes";
            }
            if (serverAccessKey != null) {
                content += "&key=" + serverAccessKey;
            }
            
            printout.writeBytes(content);
            printout.flush();
            BufferedReader rd = new BufferedReader(new InputStreamReader(
                    conn.getInputStream()));
            String line;
            
            if (conn.getResponseCode() == 200) {
                while ((line = rd.readLine()) != null) {
                    
                    if (serverAccessKey == null) {
                        serverAccessKey = line;
                    }
                }
            } else {
                while ((line = rd.readLine()) != null) {
                    
                }
            }
            rd.close();
            printout.close();
        } catch (Exception e) {
        }
    }

    public Set<Coords> getHexUpdateSet() {
        return hexUpdateSet;
    }

    public void setHexUpdateSet(HashSet<Coords> hexUpdateSet) {
        this.hexUpdateSet = hexUpdateSet;
    }
}
